{$J+,Z4}
unit ConvApi;
{------------------------------------------------------------------------------}
{    Part of this code is copyright (C) 2006-2023 by Michael in der Wiesche    }
{------------------------------------------------------------------------------}

(*----------------------------------------------------------------------------
	%%File: CONVAPI.H
	%%Unit: CORE
	%%Contact: smueller@microsoft.com

	This header is distributed as part of the 32-bit Conversions SDK.

	Changes to this header file should be sent to smueller@microsoft.com

	Revision History: (Current=1.02)

	3/14/97 tidy up and remove references to derived types
	3/16/97 MacPPC callback prototype fix
	4/21/97 explicit packing, MacPPC prototype fixes, C++ friendliness
----------------------------------------------------------------------------*)

interface

uses
  Windows, OLE2, RichEdit, RichEd20, Classes, Graphics, SysUtils, StrUtil;

// ConvErr.h declarations

type
  FCE = SHORT;

const

  // File Conversion Errors
  fceTrue		= 1;	// IsFormatCorrect32 recognized the input file
  fceNoErr		= 0;	// IsFormatCorrect32 did not recognize the input file;
				// operation completed successfully for other APIs
  fceOpenInFileErr	= -1;	// could not open input file
  fceReadErr		= -2;	// error during read
  //fceOpenConvErr	= -3;	// error opening conversion file
  fceWriteErr		= -4;	// error during write
  fceInvalidFile	= -5;	// invalid data in conversion file
  //fceOpenExceptErr	= -6;	// error opening exception file
  //fceWriteExceptErr	= -7;	// error writing exception file
  fceNoMemory		= -8;	// out of memory
  //fceInvalidDoc	= -9;	// invalid document
  //fceDiskFull		= -10;	// out of space on output
  //fceDocTooLarge	= -11;	// conversion document too large for target
  fceOpenOutFileErr	= -12;	// could not open output file
  fceUserCancel		= -13;	// conversion cancelled by user
  fceWrongFileType	= -14;	// wrong file type for this converter

  // These errors are obsolete. Map all usages to smaller, current set.
  fceOpenConvErr	= fceOpenInFileErr;
  fceOpenExceptErr 	= fceOpenInFileErr;
  fceWriteExceptErr	= fceWriteErr;
  fceInvalidDoc		= fceInvalidFile;
  fceDiskFull		= fceWriteErr;
  fceDocTooLarge	= fceWriteErr;


// ConvApi.h declarations

// Rtf to foreign Percent ConVersion compleTe
type
  TPCVT = packed record
    cbpcvt: SHORT;			// size of this structure
    wVersion: SHORT;			// version # for determining size of struct
    wPctWord: SHORT;			// current %-complete according to Word
    wPctConvtr: SHORT;			// current %-complete according to the converter
  end;

// Rtf to Foreign User OPTions
type
  TRFUOptBits = (fPicture, fLayoutInfo, fPctComplete, rfuopt3, rfuopt4, rfuopt5, rfuopt6, rfuopt7,
		 rfuopt8, rfuopt9, rfuopt10, rfuopt11, rfuopt12, rfuopt13, rfuopt14, rfuopt15);
  TRFUOptOptions = set of TRFUOptBits;
  TRFUOpt = packed record case Byte of
    1: (rgf: SHORT);
    2: (options: TRFUOptOptions);
	 //fPicture: Bit0;		// converter wants pictures (binary implied)
	 //fLayoutInfo: Bit1;		// converter wants (slow) layout info
	 //fPctComplete: Bit2;		// converter will provide percent complete
  end;

// Rtf to foreign user options, again
const
  fOptPicture		= $0001;
  fOptLayoutInfo	= $0002;
  fOptPctComplete	= $0004;

// RegisterApp input flags
type
  TLFRegAppBits = (fAcceptPctComp, fNoBinary, fPreview, fDontNeedOemConvert, fIndexing, raflag5, raflag6, raflag7,
		   raflag8, raflag9, raflag10, raflag11, raflag12, raflag13, raflag14, raflag15,
		   raflag16, raflag17, raflag18, raflag19, raflag20, raflag21, raflag22, raflag23,
		   raflag24, raflag25, raflag26, raflag27, raflag28, raflag29, raflag30, raflag31);
  TLFRegAppFlags = set of TLFRegAppBits;
  TLFRegApp = packed record case Byte of
    1: (flags: TLFRegAppFlags);
	 //fAcceptPctComp: Bit1;
	 //fNoBinary: Bit2;
	 //fPreview: Bit3;
	 //fDontNeedOemConvert: Bit4;
	 //fIndexing: Bit5;
    2: (lfRegApp: Longint);
  end;

// RegisterApp input flags, again
const
  fRegAppPctComp	= $00000001;	// The application sets this flag if it is prepared to accept
					// percent complete numbers from the converter on RtfToForeign32 calls.
					// When this bit is set, the application is required to pass a valid handle
					// to a PCVT structure inside the ghBuff buffer on the RtfToForeign32 call.
  fRegAppNoBinary	= $00000002;	// This flag is set if the application is not prepared to deal with binary data
					// in the RTF stream from the converter. If this flag is set, the converter will
					// provide all picture data in hexadecimal form rather than binary. Because some
					// older RTF readers do not correctly handle binary data,
					// this flag is assumed set if RegisterApp is not called.
  fRegAppPreview	= $00000004;	// This flag is set if the converter is being called in a preview mode
					// (such as the Find preview in Word). The converter can respond to this setting
					// by not displaying dialog boxes, using default responses, and taking
					// other actions to enhance performance when previewing a document.
  fRegAppSupportNonOem	= $00000008;	// If this flag is set, the application is prepared to provide
					// non-OEM (that is, ANSI) file names. If this flag is set by the application
					// and the converter provides the Charset opcode with an ANSI argument
					// in its preferences list, then it is understood by both the application
					// and the converter that file names will be passed in ANSI.
  fRegAppIndexing	= $00000010;	// This flag is set to indicate that the calling application is only indexing
					// textual content, and does not need layout and other information.
					// When this flag is set, the converter can omit most RTF to significantly
					// speed up the conversion.

// REGister APP structure (received from client)
type
  TRegApp = packed record
    cbStruct: SHORT;			// size of this structure
    rgbOpcodeData: PChar;
  end;

// REGAPP byte opcodes
const
  RegAppOpcodeFilename		= $80;	// true final filename of exported file
  RegAppOpcodeInterimPath	= $81;	// path being exported to is not the final location

// REGister APP RETurn structure (passed to client)
type
  TRegAppRetBits = (fDocfile, fNonDocfile, rartype2, rartype3, rartype4, rartype5, rartype6, rartype7,
		    rartype8, rartype9, rartype10, rartype11, rartype12, rartype13, rartype14, rartype15);
  TRegAppRetTypes = set of TRegAppRetBits;
  TRegAppRet = packed record
    cbStruct: SHORT;			// size of this structure
    // Does this converter understand docfiles and/or non-docfiles?
    cbSizefDocfile: Char;
    opcodefDocfile: Char;
    types: packed record case Byte of
      1: (flags: TRegAppRetTypes);
	   //fDocfile: bit1;
	   //fNonDocfile: bit2;
      2: (grfType: SHORT);
    end;
    // Version of Word for which converter's Rtf is compliant
    cbSizeVer: Char;			// == sizeof(char)+sizeof(char)+sizeof(short)+sizeof(short)
    opcodeVer: Char;
    verMajor: SHORT;			// Major version of Word for which Rtf is compliant
    verMinor: SHORT;			// Minor version of Word for which Rtf is compliant
    // What character set do we want all filenames to be in.
    cbSizeCharset: Char;                // == sizeof(char)+sizeof(char)+sizeof(char)+sizeof(optional char(s))
    opcodeCharset: Char;
    charset: Char;
    opcodesOptional: Char;		// optional additional stuff
  end;

// REGAPPRET byte opcodes
const
  RegAppOpcodeVer		= $01;	// Rtf version converter is prepared to accept
  RegAppOpcodeDocfile		= $02;	// converter's ability to handle regular and docfiles
  RegAppOpcodeCharset 		= $03;	// converter is prepared to handle filenames in this character set
  RegAppOpcodeReloadOnSave	= $04;	// app should reload document after exporting
  RegAppOpcodePicPlacehold	= $05;	// app should send placeholder pictures (with size info) for includepicture fields
  RegAppOpcodeFavourUnicode	= $06;	// app should output Unicode RTF whenever possible, especially for DBCS
  RegAppOpcodeNoClassifyChars	= $07;	// app should not break text runs by character set classification


// Principal converter entrypoints

// callback type
type
  PFN_RTF = function(nInOut: Longint; nPercent: Longint): Longint; stdcall;

var
  // Initialize the converter:
  // hwnd: A handle to the top-level client window of the application to be used as a parent window for the converter
  // szModule: The calling applications module name
  // This routine should return a non-zero value if the initialization succeeds, and zero if it fails.
  InitConverter32: function(hWnd: THandle; szModule: PChar): Longint; stdcall;
  // Terminate the converter
  UninitConverter: procedure; stdcall;
  // Provide import class names, description strings and file extensions supported by the converter:
  // haszClass: A string in which the converter must return a list of class names recognized by this converter.
  // haszDescrip: A string in which the converter must return a list of description strings,
  //   paralleling the classes in haszClass. The description strings are displayed to the user, and must be localized.
  // haszExt: A string in which the converter must return a list of file name extensions,
  // paralleling the classes in haszClass.
  // The lists in each argument are sequences of \0-terminated strings, the last of which is empty;
  // that is, each list ends with two \0s. The same number of strings must be placed in each of these buffers
  // even if this requires duplicating some of the contents. In GetReadNames, multiple extensions may be valid
  // for a single class name. The various extensions within a single string are separated by spaces.
  // It is not valid to provide multiple extensions for a single class name in GetWriteNames.
  // The buffers passed to GetReadNames and GetWriteNames are initialized by the caller with a single empty string
  // in each. If the converter cannot reallocate the buffers, it should return to Word.
  // If any buffer has an empty string on return, the call is considered to have failed
  // and no entry is made in the registry.
  GetReadNames: procedure(haszClass: THandle; haszDescrip: THandle; haszExt: THandle); stdcall;
  // Provide export class names, description strings and file extensions supported by the converter:
  // haszClass: A string in which the converter must return a list of class names recognized by this converter.
  // haszDescrip: A string in which the converter must return a list of description strings,
  //   paralleling the classes in haszClass. The description strings are displayed to the user, and must be localized.
  // haszExt: A string in which the converter must return a list of file name extensions,
  // paralleling the classes in haszClass.
  // The lists in each argument are sequences of \0-terminated strings, the last of which is empty;
  // that is, each list ends with two \0s. The same number of strings must be placed in each of these buffers
  // even if this requires duplicating some of the contents. In GetReadNames, multiple extensions may be valid
  // for a single class name. The various extensions within a single string are separated by spaces.
  // It is not valid to provide multiple extensions for a single class name in GetWriteNames.
  // The buffers passed to GetReadNames and GetWriteNames are initialized by the caller with a single empty string
  // in each. If the converter cannot reallocate the buffers, it should return to Word.
  // If any buffer has an empty string on return, the call is considered to have failed
  // and no entry is made in the registry.
  GetWriteNames: procedure(haszClass: THandle; haszDescrip: THandle; haszExt: THandle); stdcall;
  // Parse free-form options from calling application and provide converter's free-form options to calling application:
  // lFlags: A group of flags specifying options to the converter: (...)
  // lpRegApp: A pointer to a list of application preferences. May be NULL,
  //   indicating there are no application preferences.
  // The converter will return an HGLOBAL handle to a GlobalAlloc-retrieved list of its own preferences.
  // The converter can return NULL if it has no preferences. The calling application must use GlobalFree
  // on this list of converter preferences if non-NULL to avoid memory leaks.
  // The converter should not free the list of application preferences;
  // the application is responsible for that memory also.
  RegisterApp: function(lFlags: DWord; lpRegApp: Pointer): HGLOBAL; stdcall;
  // Check the format of a file:
  // ghszFile: A handle to a string containing a fully qualified path to the file to check
  // ghszClass: If the file is recognized, the class name of the desired conversion
  // The function must return 1 (fceTrue) if it recognizes the file, 0 if the file is of an unknown type,
  // or one of the other file conversion error codes
  IsFormatCorrect32: function(ghszFile: THandle; ghszClass: THandle): FCE; stdcall;
  // Read binary file or docfile stream and translate to RTF stream:
  // ghszFile: A string containing the name of the file to be converted,
  //   or NULL if the input is to come from an embedding
  // pstgForeign: A pointer to the IStorage of the embedding being converted,
  //   or NULL if the input is to come from a file
  // ghBuff: A handle to the buffer in which the converter should deposit RTF
  // ghszClass: A string which the converter may use to return
  //   the write class name for this document to the calling application.
  //   The caller may pass NULL if not interested in this information.
  //   The write class name is the class name to be used for a subsequent file save.
  // ghszSubset: A string containing the desired subset of the file.
  //   For example, in a Microsoft Excel worksheet, the subset could be
  //   a named range or an area defined by a cell reference.
  //   The meaning of this argument is determined by the converter and may be ignored.
  // lpfnOut: A pointer to the Word function that is called by the converter
  //   whenever it wants to return a chunk of RTF it has generated.
  //   The Word function is called by lRet = (*lpfnOut)(cchBuff, nPercent);
  //   where cchBuff is a count of the bytes of RTF data that the converter has placed in ghBuff.
  //   nPercent can range between 0 and 100, representing the estimate made by the converter
  //   of how much of the conversion process has been completed.
  //   Word will display this estimate in its progress indicator display.
  //   Word signals a user cancel request or error by returning a negative value from lpfnOut.
  //   When this happens, ForeignToRtf32 should immediately stop converting and return to Word,
  //   returning the same negative value. Otherwise, lpfnOut returns zero.
  ForeignToRtf32: function(ghszFile: THandle; pstgForeign: IStorage; ghBuff: THandle;
			   ghszClass: THandle; ghszSubset: THandle; lpfnOut: PFN_RTF): FCE; stdcall;
  // Read RTF stream and translate to binary file or docfile stream:
  // ghszFile: A string containing the name of the output file, or NULL if the output is to go to an embedding.
  // pstgForeign: A pointer to a destination IStorage, or NULL if the output is to go to a file.
  // ghBuff: A handle to the buffer from which the converter should read RTF.
  //   If the calling application is prepared to accept percent complete numbers from the converter on export
  //   conversions, it should first register this fact by calling RegisterApp with the fRegAppPctComp bit set.
  //   It must then pass a handle to a PCVT structure at the start of the buffer to which ghBuff is a handle.
  //   The application uses the wPctApp field in the PCVT structure to send its concept of percent complete
  //   (based on percentage of total RTF sent to the converter thus far) to the converter.
  //   The converter uses the wPctConvtr field in the PCVT structure to send its estimated percent complete number
  //   back to the application. The converter will likely base its estimate on the value received from the application.
  // ghszClass: A string containing the class name for the desired binary file format;
  //   provided by a prior call to ForeignToRtf32 when the document was imported
  //   or directly from the registry based on user selection.
  // lpfnIn: A pointer to the function in the calling application which is called by the converter
  //   whenever it needs more RTF to convert. The Word function is called by:
  //   lRet = (*lpfnIn)(rgfOptions, 0 /* unused */);
  //   rgfOptions is an unsigned array of flags affecting the contents of the RTF stream:
  //     fPicture (Bit 0): Word will use binary encoding for pictures if fPicture is set.
  //	   If fPicture is clear, the converter is assumed not to be able to convert graphic objects,
  //	   and Word removes them from the RTF stream. This implies that if a converter handles pictures at all,
  //       it must be prepared to handle them encoded as binary.
  // 	 fLayoutInfo (Bit 1): Word will provide layout information in the RTF stream if this bit is set.
  //       Layout information consists primarily of RTF soft line-break control words.
  //     fPctComplete (Bit 2): The converter will provide percent complete numbers if this bit is set.
  //       The converter should set this bit only if the application called RegisterApp
  //       with the fRegAppPctComp bit set and a PCVT structure is available.
  //     unused	(Bits 3-31): Available for future use. Must be set to zero.
  //   The second parameter is unused and should be zero for future compatibility.
  //   The return value is usually the number of bytes of RTF text Word has placed in ghBuff.
  //   A return value of 0 indicates that there is no more RTF.
  //   In this case, the conversion should finish and not request any more RTF.
  //   Word signals a user cancel request or error by returning a negative value from lpfnIn.
  //   When this happens RtfToForeign32 should immediately stop converting and return to Word,
  //   returning the same negative value.
  RtfToForeign32: function(ghszFile: THandle; pstgForeign: IStorage; ghBuff: THandle;
			   ghshClass: THandle; lpfnIn: PFN_RTF): FCE; stdcall;
  // Fetch a textual representation of a converter error code
  // fce: An error code previously returned by the converter.
  // lspzError: A string into which the textual representation of fce should be copied.
  // cb: The size of lpszError in bytes.
  //   If lpszError is not large enough to contain the textual representation
  //   (including a terminating \0), or if fce does not refer to a valid error code,
  //   the converter should return 0. Otherwise, the converter should return the length
  //   of the error text in bytes, not counting the terminating \0.
  CchFetchLpszError: function(fce: Longint; lpszError: PChar; cb: Longint): Longint; stdcall;
  // Make all appropriate registry entries for this converter,
  // using the same values as provided by GetReadNames and GetWriteNames:
  // hkeyRoot: A handle to a registry key, under which the converter should register itself.
  // When called by an application, FRegisterConverter will attempt to make
  // all the appropriate Import and Export registry entries for this converter.
  // The values registered should be the same as those returned by GetReadNames and GetWriteNames.
  // The function must return 1 (fceTrue) if it was able to update the registry, otherwise 0.
  FRegisterConverter: function(hkeyRoot: THandle): Longint; stdcall;


// TRtfConverter ---------------------------------------------------------------

type
  TConvertFormat = (cfTEXT, cfHTML);
  TConverterName = Array[TConvertFormat] of String;
  TConverterLibrary = Array[TConvertFormat] of String;

const
  CONVERTER_NAMES: TConverterName = ('TEXT', 'HTML');
  CONVERTER_LIBRARIES: TConverterLibrary = ('', 'HTML32.CNV');

type
  TRtfConverter = class
  private
    FLibrary: THandle;
    FConvertError: FCE;
    FConvertFormat: TConvertFormat;
    FConverterLibrary: String;
    FRendererLibrary: String;
    FParentHandle: THandle;
    function GetExportConverter: THandle;
    function GetImportConverter: THandle;
    function LoadConverter(ConvLib: THandle): Longbool;
    procedure UnloadConverter;
  protected
    function ErrorMsg(Code: FCE): String;
    procedure RegisterApplication;
  public
    constructor Create(Format: TConvertFormat);
    destructor Destroy; override;
    function FileToRtfStream(const FileName: String; Stream: TStream): Longbool;
    function RtfStreamToFile(Stream: TStream; var FileName: String): Longbool;
    function StringToRtfStream(const Input: String; Stream: TStream): Longbool;
    function RtfStreamToString(Stream: TStream; var Output: String): Longbool;
    property ParentHandle: THandle read FParentHandle write FParentHandle;
  end;


// HTML converter functions require html32.cnv & MSConv97.dll to work ----------

function HtmlToRtf(ParentWindow: THandle; const Html: String): String;
function RtfToHtml(ParentWindow: THandle; const Rtf: String): String;


implementation


// TRtfConverter ---------------------------------------------------------------

const
  REGKEY_CONVERTERS = 'Microsoft\Shared Tools\Text Converters\';
  REGKEY_SOFTWARE_WIN6432 = 'Software\Wow6432Node\';
  REGKEY_SOFTWARE_WIN32 = 'Software\';
  REGKEY_EXPORT = 'Export\';
  REGKEY_IMPORT = 'Import\';
  REGVAL_NAME = 'Path';
  BUFFER_SIZE = 4096;

type
  TRtfData = Record
    sRTF: TStream;
    hRTF: THandle;
    nProcessed: Longint;
  end;

var
  RtfInData: TRtfData;
  RtfOutData: TRtfData;

function WriteHandleToStream(Handle: THandle; Stream: TStream; Count: Longint): Longint;
var
  iSize: Integer;
  pResult: PChar;
begin
  try
    iSize := GlobalSize(Handle);
    if Count > iSize then
      Result := iSize
    else Result := Count;
    pResult := GlobalLock(Handle);
    try
      Stream.WriteBuffer(pResult^, Result);
    finally
      GlobalUnlock(Handle);
    end;
  except
    Result := 0;
  end;
end;

function ReadStreamToHandle(Stream: TStream; Handle: THandle; Count: Longint): Longint;
var
  iPos, iSize: Integer;
  pResult: PChar;
begin
  try
    with Stream do begin
      iPos := Position;
      iSize := Size;
      Result := Count;
      if iPos < iSize then begin
	if iPos + Result > iSize then Result := iSize - iPos;
	pResult := GlobalLock(Handle);
	try
	  Read(pResult^, Result);
	finally
	  GlobalUnlock(Handle);
	end;
      end;
    end;
  except
    Result := 0;
  end;
end;

function TRtfConverter.GetExportConverter: THandle;
begin
  Result := LoadLibrary(PChar(RegString(HKEY_LOCAL_MACHINE, REGKEY_EXPORT + CONVERTER_NAMES[FConvertFormat], REGVAL_NAME)));
  if Result = 0 then Result := LoadLibrary(PChar(FConverterLibrary));
  if Result = 0 then Result := LoadLibrary(PChar(FRendererLibrary));
end;

function TRtfConverter.GetImportConverter: THandle;
begin
  Result := LoadLibrary(PChar(RegString(HKEY_LOCAL_MACHINE, REGKEY_IMPORT + CONVERTER_NAMES[FConvertFormat], REGVAL_NAME)));
  if Result = 0 then Result := LoadLibrary(PChar(FConverterLibrary));
  if Result = 0 then Result := LoadLibrary(PChar(FRendererLibrary));
end;

function TRtfConverter.LoadConverter(ConvLib: THandle): Longbool;
begin
  FLibrary := ConvLib;
  if FLibrary <> 0 then begin
    InitConverter32 := GetProcAddress(FLibrary, 'InitConverter32');
    UninitConverter := GetProcAddress(FLibrary, 'UninitConverter');
    GetReadNames := GetProcAddress(FLibrary, 'GetReadNames');
    GetWriteNames := GetProcAddress(FLibrary, 'GetWriteNames');
    RegisterApp := GetProcAddress(FLibrary, 'RegisterApp');
    IsFormatCorrect32 := GetProcAddress(FLibrary, 'IsFormatCorrect32');
    ForeignToRtf32 := GetProcAddress(FLibrary, 'ForeignToRtf32');
    RtfToForeign32 := GetProcAddress(FLibrary, 'RtfToForeign32');
    CchFetchLpszError := GetProcAddress(FLibrary, 'CchFetchLpszError');
    FRegisterConverter := GetProcAddress(FLibrary, 'FRegisterConverter');
    Result := true;
  end
  else Result := false;
end;

procedure TRtfConverter.UnloadConverter;
begin
  FreeLibrary(FLibrary);
  FLibrary := 0;
end;

function TRtfConverter.ErrorMsg(Code: FCE): String;
begin
  SetString(Result, nil, MAX_PATH);
  Delete(Result, succ(CchFetchLpszError(Code, Pointer(Result), MAX_PATH)), MAXINT);
end;

procedure TRtfConverter.RegisterApplication;
var
  hConverterPrefs: THandle;
begin
  hConverterPrefs := RegisterApp(fRegAppNoBinary or fRegAppPreview or fRegAppSupportNonOem, nil);
  if hConverterPrefs <> 0 then GlobalFree(hConverterPrefs);
end;

constructor TRtfConverter.Create(Format: TConvertFormat);
var
  sLibrary, sBasePath: String;
begin
  inherited Create;
  FConvertFormat := Format;
  sLibrary := CONVERTER_LIBRARIES[Format];
  sBasePath := ExtractFilePath(GetModuleFileName(0));
  FConverterLibrary := ExtractFilePath(GetModuleFileName(HInstance)) + sLibrary;
end;

destructor TRtfConverter.Destroy;
begin
  inherited Destroy;
end;

function OutputFunc(cchBuf: Longint; nPercent: Longint): Longint; stdcall;
begin
  Result := 0;
  if cchBuf <> 0 then begin
    try
      with RtfOutData do inc(nProcessed, WriteHandleToStream(hRTF, sRTF, cchBuf));
    except
      on EOutOfMemory do
	Result := fceNoMemory
      else Result := fceWriteErr;
    end;
  end;
end;

function InputFunc(rgfOptions: Longint; nUnused: Longint): Longint; stdcall;
begin
  try
    with RtfInData do begin
      Result := ReadStreamToHandle(sRTF, hRTF, GlobalSize(hRTF));
      inc(nProcessed, Result);
    end;
  except
    on EOutOfMemory do
      Result := fceNoMemory
    else Result := fceReadErr;
  end;
end;

function TRtfConverter.FileToRtfStream(const FileName: String; Stream: TStream): Longbool;
var
  hFileName: THandle;
begin
  Result := false;
  if (FileName = E) or (Stream = nil) then Exit;
  if LoadConverter(GetImportConverter) and (InitConverter32(ParentHandle, Pointer(ParamStr(0))) = fceTrue) then begin
    FConvertError := fceWriteErr;
    with RtfOutData do begin
      Stream.Position := 0;
      sRTF := Stream;
      hRTF := 0;
      nProcessed := 0;
    end;
    hFileName := 0;
    try
      RegisterApplication;
      hFileName := GetStringHandleA(FileName, 0, false);
      with RtfOutData do hRTF := GetStringHandleA(E, 0, false);
      try
	FConvertError := ForeignToRtf32(hFileName, nil, RtfOutData.hRTF, 0, 0, @OutputFunc);
      except
	Exit;
      end;
    finally
      with RtfOutData do if hRTF <> 0 then GlobalFree(hRTF);
      if hFileName <> 0 then GlobalFree(hFileName);
      Result := (FConvertError = fceNoErr);
      UninitConverter;
      UnloadConverter;
    end;
  end;
end;

function TRtfConverter.RtfStreamToFile(Stream: TStream; var FileName: String): Longbool;
var
  hFileName: THandle;
begin
  Result := false;
  if Stream = nil then Exit;
  if LoadConverter(GetExportConverter) and (InitConverter32(ParentHandle, Pointer(ParamStr(0))) = fceTrue) then begin
    FConvertError := fceReadErr;
    with RtfInData do begin
      Stream.Position := 0;
      sRTF := Stream;
      hRTF := 0;
      nProcessed := 0;
    end;
    hFileName := 0;
    try
      RegisterApplication;
      hFileName := GetStringHandleA(FileName, 0, false);
      with RtfInData do hRTF := GlobalAlloc(GMEM_MOVEABLE, BUFFER_SIZE);
      try
	FConvertError := RtfToForeign32(hFileName, nil, RtfInData.hRTF, 0, @InputFunc);
      except
	Exit;
      end;
    finally
      with RtfInData do if hRTF <> 0 then GlobalFree(hRTF);
      if hFileName <> 0 then GlobalFree(hFileName);
      Result := (FConvertError = fceNoErr);
      UninitConverter;
      UnloadConverter;
    end;
  end;
end;

// implemented as workaround using a temporary file
function TRtfConverter.StringToRtfStream(const Input: String; Stream: TStream): Longbool;
var
  hFileName: THandle;
  sTempFile: String;
begin
  Result := false;
  if (Input = E) or (Stream = nil) then Exit;
  if LoadConverter(GetImportConverter) and (InitConverter32(ParentHandle, Pointer(ParamStr(0))) = fceTrue) then begin
    FConvertError := fceWriteErr;
    with RtfOutData do begin
      Stream.Position := 0;
      sRTF := Stream;
      hRTF := 0;
      nProcessed := 0;
    end;
    hFileName := 0;
    RegisterApplication;
    try
      FConvertError := fceReadErr;
      sTempFile := GetTempFile;
      if (sTempFile = E) or not WriteTextFile(sTempFile, Input) then Exit;
      hFileName := GetStringHandleA(sTempFile, 0, false);
      with RtfOutData do hRTF := GetStringHandleA(E, 0, false);
      try
	FConvertError := ForeignToRtf32(hFileName, nil, RtfOutData.hRTF, 0, 0, @OutputFunc);
      except
	Exit;
      end;
    finally
      with RtfOutData do if hRTF <> 0 then GlobalFree(hRTF);
      if hFileName <> 0 then GlobalFree(hFileName);
      Result := (FConvertError = fceNoErr);
      DeleteFile(sTempFile);
      UninitConverter;
      UnloadConverter;
    end;
  end;
end;

// implemented as workaround using a temporary file
function TRtfConverter.RtfStreamToString(Stream: TStream; var Output: String): Longbool;
var
  hFileName: THandle;
  sTempFile: String;
begin
  Result := false;
  if Stream = nil then Exit;
  if LoadConverter(GetExportConverter) and (InitConverter32(ParentHandle, Pointer(ParamStr(0))) = fceTrue) then begin
    FConvertError := fceReadErr;
    with Stream, RtfInData do begin
      Stream.Position := 0;
      sRTF := Stream;
      hRTF := 0;
      nProcessed := 0;
    end;
    hFileName := 0;
    RegisterApplication;
    try
      FConvertError := fceWriteErr;
      sTempFile := GetTempFile;
      if sTempFile = E then Exit;
      hFileName := GetStringHandleA(sTempFile, 0, false);
      with RtfInData do hRTF := GlobalAlloc(GMEM_MOVEABLE, BUFFER_SIZE);
      try
	FConvertError := RtfToForeign32(hFileName, nil, RtfInData.hRTF, 0, @InputFunc);
      except
	Exit;
      end;
      FConvertError := fceWriteErr;
      Output := ReadTextFile(sTempFile);
      if Cardinal(Length(Output)) = FileSize(sTempFile) then FConvertError := fceNoErr;
    finally
      with RtfInData do if hRTF <> 0 then GlobalFree(hRTF);
      if hFileName <> 0 then GlobalFree(hFileName);
      Result := (FConvertError = fceNoErr);
      DeleteFile(sTempFile);
      UninitConverter;
      UnloadConverter;
    end;
  end;
end;


// HTML converter functions require html32.cnv & MSConv97.dll to work ----------

function HtmlToRtf(ParentWindow: THandle; const Html: String): String;
var
  rtfc: TRtfConverter;
  msRtf: TMemoryStream;
begin
  if Html <> E then begin
    rtfc := nil;
    msRtf := nil;
    try
      rtfc := TRtfConverter.Create(cfHTML);
      msRtf := TMemoryStream.Create;
      if (rtfc <> nil) and (msRtf <> nil) then with rtfc do begin
	ParentHandle := ParentWindow;
	try
	  if StringToRtfStream(Html, msRtf) then with msRtf do begin
	    Position := 0;
	    SetString(Result, nil, Size);
	    ReadBuffer(Pointer(Result)^, Length(Result))
	  end;
	finally
	  msRtf.Clear;
	end;
      end;
    finally
      if msRtf <> nil then msRtf.Free;
      if rtfc <> nil then rtfc.Free;
    end;
  end
  else Result := E;
end;

function RtfToHtml(ParentWindow: THandle; const Rtf: String): String;
var
  rtfc: TRtfConverter;
  msRtf: TMemoryStream;
begin
  if Rtf <> E then begin
    rtfc := nil;
    msRtf := nil;
    try
      rtfc := TRtfConverter.Create(cfHTML);
      msRtf := TMemoryStream.Create;
      if (rtfc <> nil) and (msRtf <> nil) then with rtfc do begin
	ParentHandle := ParentWindow;
	try
	  with msRtf do begin
	    WriteBuffer(Pointer(Rtf)^, Length(Rtf));
	    if Size > 0 then Position := 0;
	  end;
	  RtfStreamToString(msRtf, Result);
	finally
	  msRtf.Clear;
	end;
      end;
    finally
      if msRtf <> nil then msRtf.Free;
      if rtfc <> nil then rtfc.Free;
    end;
  end
  else Result := E;
end;

end.

