{$J+,Z4}
unit DatedStrings;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Classes, SysUtils;

type
  PDatedStringItem = ^TDatedStringItem;
  TDatedStringItem = Record
    FString: String;
    FDateTime: TDateTime;
  end;

  PDatedStringItemList = ^TDatedStringItemList;
  TDatedStringItemList = Array[0..pred(MaxListSize)] of TDatedStringItem;

  TDateTimeSorted = (dtNoOrder, dtSorted, dtReverse);

  TDatedStrings = class(TStrings)
  private
    FCount: Integer;
    FCapacity: Integer;
    FDTSorted: TDateTimeSorted;
    FDTList: PDatedStringItemList;
    procedure ExchangeItems(Index1, Index2: Integer);
    procedure Grow;
    procedure InsertItem(Index: Integer; const S: String);
    procedure QuickSort(L, R: Integer);
    procedure ReverseSort(L, R: Integer);
    procedure SetDTSorted(Value: TDateTimeSorted);
  protected
    function  Get(Index: Integer): String; override;
    procedure Put(Index: Integer; const S: String); override;
    function  GetCount: Integer; override;
    function  GetDateTime(Index: Integer): TDateTime; virtual;
    procedure PutDateTime(Index: Integer; ADateTime: TDateTime); virtual;
    {$IFDEF CONDITIONALEXPRESSIONS}
    procedure SetCapacity(NewCapacity: Integer); override;
    {$ELSE}
    procedure SetCapacity(NewCapacity: Integer);
    {$ENDIF}
  public
    destructor Destroy; override;
    function  Add(const S: String): Integer; override;
    function  AddDated(const S: String; ADateTime: TDateTime): Integer; virtual;
    procedure Clear; override;
    procedure DateTimeSort(Reverse: Longbool); virtual;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function  FindDated(const DT: TDateTime; var Index: Integer): Boolean; virtual;
    procedure Insert(Index: Integer; const S: String); override;
    procedure InsertDated(Index: Integer; const S: String; ADateTime: TDateTime); virtual;
    procedure Move(CurIndex, NewIndex: Integer); override;
    property  Capacity: Integer read FCapacity write SetCapacity;
    property  DateTimes[Index: Integer]: TDateTime read GetDateTime write PutDateTime;
    property  DateTimeSorted: TDateTimeSorted read FDTSorted write SetDTSorted;
  end;

  TDatedWideStrings = class(TDatedStrings)
  private
    function  GetName(Index: Integer): String;
    function  GetValue(const Name: String): String;
    procedure SetValue(const Name, Value: String);
  protected
    function  GetCommaText: String;
    procedure SetCommaText(const Value: String);
    function  GetTextStr: String; override;
    procedure SetTextStr(const Value: String); override;
  public
    {$IFDEF CONDITIONALEXPRESSIONS}
    function  IndexOfName(const Name: String): Integer; override;
    {$ELSE}
    function  IndexOfName(const Name: String): Integer;
    {$ENDIF}
    property  CommaText: String read GetCommaText write SetCommaText;
    property  Names[Index: Integer]: String read GetName;
    property  Values[const Name: String]: String read GetValue write SetValue;
  end;


implementation

uses
  {$IFDEF CONDITIONALEXPRESSIONS}
  RtlConsts;
  {$ELSE}
  Consts;
  {$ENDIF}

const
  E = '';
  NUL = #0;
  LF = #10;
  CR = #13;
  EQ = '=';

procedure ListIndexError;
begin
  {$IFDEF CONDITIONALEXPRESSIONS}
  Raise EListError.Create(SListIndexError);
  {$ELSE}
  Raise EListError.CreateRes(SListIndexError);
  {$ENDIF}
end;


// TDatedStrings ---------------------------------------------------------------

procedure TDatedStrings.ExchangeItems(Index1, Index2: Integer);
var
  TempItem: TDatedStringItem;
begin
  TempItem := FDTList^[Index1];
  FDTList^[Index1] := FDTList^[Index2];
  FDTList^[Index2] := TempItem;
end;

procedure TDatedStrings.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 8 then
    Delta := 16
  else if FCapacity > 4 then
    Delta := 8
  else Delta := 4;
  SetCapacity(FCapacity + Delta);
end;

procedure TDatedStrings.InsertItem(Index: Integer; const S: String);
begin
  if FCount = FCapacity then Grow;
  if Index < FCount then System.Move(FDTList^[Index], FDTList^[Index + 1], (FCount - Index) * SizeOf(TDatedStringItem));
  with FDTList^[Index] do begin
    Pointer(FString) := nil;
    FDateTime := 0;
    FString := S;
  end;
  inc(FCount);
end;

procedure TDatedStrings.QuickSort(L, R: Integer);
var
  I, J: Integer;
  P: TDateTime;
begin
  repeat
    I := L;
    J := R;
    P := FDTList^[(L + R) shr 1].FDateTime;
    repeat
      while FDTList^[I].FDateTime < P do inc(I);
      while FDTList^[J].FDateTime > P do dec(J);
      if I <= J then begin
	ExchangeItems(I, J);
	Inc(I);
	Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J);
    L := I;
  until I >= R;
end;

procedure TDatedStrings.ReverseSort(L, R: Integer);
var
  I, J: Integer;
  P: TDateTime;
begin
  repeat
    I := L;
    J := R;
    P := FDTList^[(L + R) shr 1].FDateTime;
    repeat
      while FDTList^[I].FDateTime > P do inc(I);
      while FDTList^[J].FDateTime < P do dec(J);
      if I <= J then begin
	ExchangeItems(I, J);
	Inc(I);
	Dec(J);
      end;
    until I > J;
    if L < J then ReverseSort(L, J);
    L := I;
  until I >= R;
end;

procedure TDatedStrings.SetDTSorted(Value: TDateTimeSorted);
begin
  if Value <> FDTSorted then begin
    DateTimeSort(Value = dtReverse);
    FDTSorted := Value;
  end;
end;

function TDatedStrings.Get(Index: Integer): String;
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Result := FDTList^[Index].FString;
end;

procedure TDatedStrings.Put(Index: Integer; const S: String);
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  FDTList^[Index].FString := S;
end;

function TDatedStrings.GetCount: Integer;
begin
  Result := FCount;
end;

function TDatedStrings.GetDateTime(Index: Integer): TDateTime;
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Result := FDTList^[Index].FDateTime;
end;

procedure TDatedStrings.PutDateTime(Index: Integer; ADateTime: TDateTime);
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  FDTList^[Index].FDateTime := ADateTime;
end;

procedure TDatedStrings.SetCapacity(NewCapacity: Integer);
begin
  ReallocMem(FDTList, NewCapacity * SizeOf(TDatedStringItem));
  FCapacity := NewCapacity;
end;

destructor TDatedStrings.Destroy;
begin
  if FCount <> 0 then Finalize(FDTList^[0], FCount);
  SetCapacity(0);
  FCount := 0;
end;

function TDatedStrings.Add(const S: String): Integer;
begin
  Result := inherited Add(S);
  FDTSorted := dtNoOrder;
end;

function TDatedStrings.AddDated(const S: String; ADateTime: TDateTime): Integer;
begin
  if FDTSorted <> dtNoOrder then begin
    FindDated(ADateTime, Result);
    InsertItem(Result, S);
  end
  else Result := Add(S);
  PutDateTime(Result, ADateTime);
end;

procedure TDatedStrings.Clear;
begin
  if FCount <> 0 then begin
    Finalize(FDTList^[0], FCount);
    SetCapacity(0);
    FCount := 0;
  end;
end;

procedure TDatedStrings.DateTimeSort(Reverse: Longbool);
begin
  if FCount > 1 then begin
    if not Reverse then begin
      QuickSort(0, pred(FCount));
      FDTSorted := dtSorted;
    end
    else begin
      ReverseSort(0, pred(FCount));
      FDTSorted := dtReverse;
    end;
  end;
end;

procedure TDatedStrings.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then ListIndexError;
  Finalize(FDTList^[Index]);
  dec(FCount);
  if Index < FCount then System.Move(FDTList^[Index + 1], FDTList^[Index], (FCount - Index) * SizeOf(TDatedStringItem));
end;

procedure TDatedStrings.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) or (Index2 < 0) or (Index2 >= FCount) then ListIndexError;
  ExchangeItems(Index1, Index2);
  FDTSorted := dtNoOrder;
end;

function TDatedStrings.FindDated(const DT: TDateTime; var Index: Integer): Boolean;
var
  L, H, I: Integer;
  C: TDateTime;
begin
  Result := false;
  if FDTSorted <> dtNoOrder then begin
    L := 0;
    H := pred(FCount);
    if FDTSorted = dtSorted then begin
      while L <= H do begin
	I := (L + H) shr 1;
	C := FDTList^[I].FDateTime - DT;
	if C < 0 then
	  L := I + 1
	else begin
	  H := I - 1;
	  if C = 0 then Result := true;
	end;
      end;
    end
    else begin
      while L <= H do begin
	I := (L + H) shr 1;
	C := DT - FDTList^[I].FDateTime;
	if C < 0 then
	  L := I + 1
	else begin
	  H := I - 1;
	  if C = 0 then Result := true;
	end;
      end;
    end;
    Index := L;
  end
  else for I := 0 to pred(FCount) do begin
    Index := -1;
    if FDTList^[I].FDateTime = DT then begin
      Result := true;
      Index := I;
      Exit;
    end;
  end;
end;

procedure TDatedStrings.Insert(Index: Integer; const S: String);
begin
  if (Index < 0) or (Index > FCount) then ListIndexError;
  InsertItem(Index, S);
  FDTSorted := dtNoOrder;
end;

procedure TDatedStrings.InsertDated(Index: Integer; const S: String; ADateTime: TDateTime);
begin
  Insert(Index, S);
  PutDateTime(Index, ADateTime);
end;

procedure TDatedStrings.Move(CurIndex, NewIndex: Integer);
var
  TempString: String;
  TempDateTime: TDateTime;
begin
  if CurIndex <> NewIndex then begin
    TempString := Get(CurIndex);
    TempDateTime := GetDateTime(CurIndex);
    Delete(CurIndex);
    InsertDated(NewIndex, TempString, TempDateTime);
  end;
end;


// TWideStrings ----------------------------------------------------------------

function TDatedWideStrings.GetName(Index: Integer): String;
var
  I: Integer;
begin
  Result := Get(Index);
  I := Pos(EQ, Result);
  if I > 0 then begin
    inc(I);
    if Length(Result) >= I then
      System.Delete(Result, succ(I), MAXINT)
    else SetString(Result, nil, succ(I));
    PChar(Pointer(Result))[I - 1] := NUL;
    PChar(Pointer(Result))[I - 2] := NUL;
  end
  else Result := E;
end;

function TDatedWideStrings.GetValue(const Name: String): String;
var
  I: Integer;
begin
  I := IndexOfName(Name);
  if I >= 0 then begin
    Result := Get(I);
    System.Delete(Result, 1, Length(Name));
  end
  else Result := E;
end;

procedure TDatedWideStrings.SetValue(const Name, Value: String);
var
  I, Size: Integer;
  S: String;
  W: PWideChar;
begin
  I := IndexOfName(Name);
  if Value <> E then begin
    if I < 0 then I := Add(E);
    SetString(S, nil, Length(Name) + Length(Value));
    W := Pointer(S);
    Size := Length(Name) - 2;
    System.Move(Pointer(Name)^, W^, Size);
    inc(W, Size div 2);
    W^ := EQ;
    inc(W);
    System.Move(Pointer(Value)^, W^, Length(Value));
    Put(I, S);
  end
  else if I >= 0 then Delete(I);
end;

function TDatedWideStrings.GetCommaText: String;
begin
  {$IFDEF CONDITIONALEXPRESSIONS}
  Raise EStringListError.Create(SUnknownProperty);
  {$ELSE}
  Raise EStringListError.CreateRes(SUnknownProperty);
  {$ENDIF}
end;

procedure TDatedWideStrings.SetCommaText(const Value: String);
begin
  {$IFDEF CONDITIONALEXPRESSIONS}
  Raise EStringListError.Create(SUnknownProperty);
  {$ELSE}
  Raise EStringListError.CreateRes(SUnknownProperty);
  {$ENDIF}
end;

function TDatedWideStrings.GetTextStr: String;
var
  I, L, Size, Count: Integer;
  W: PWideChar;
  S: String;
begin
  Count := GetCount;
  Size := 0;
  for I := 0 to pred(Count) do inc(Size, Length(Get(I)) + 4);
  SetString(Result, nil, Size + 2);
  W := Pointer(Result);
  for I := 0 to pred(Count) do begin
    S := Get(I);
    L := Length(S) - 2;
    if L > 0 then begin
      System.Move(Pointer(S)^, W^, L);
      inc(W, L div 2);
    end;
    W^ := CR;
    inc(W);
    W^ := LF;
    inc(W);
  end;
  W^ := NUL;
  inc(W);
  System.Delete(Result, succ((W - Pointer(Result)) shl 1), MAXINT);
end;

procedure TDatedWideStrings.SetTextStr(const Value: String);
var
  W, Start: PWideChar;
  Size: Integer;
  S: String;
begin
  BeginUpdate;
  try
    Clear;
    W := Pointer(Value);
    if W <> nil then while W^ <> NUL do begin
      Start := W;
      while (W^ <> NUL) and (W^ <> LF) and (W^ <> CR) do inc(W);
      Size := W - Start;
      SetString(S, PChar(Start), Size shl 1 + 2);
      Start := Pointer(S);
      inc(Start, Size);
      Start^ := NUL;
      Add(S);
      if W^ = CR then inc(W);
      if W^ = LF then inc(W);
    end;
  finally
    EndUpdate;
  end;
end;

function TDatedWideStrings.IndexOfName(const Name: String): Integer;
var
  P: Integer;
  S: String;
begin
  if Name <> E then for Result := 0 to pred(GetCount) do begin
    S := Get(Result);
    P := Pos(EQ, S);
    if (P <> 0) and (CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
				   Pointer(Name), Length(Name) - 2,
				   Pointer(S), pred(P)) = 2) then Exit;
  end;
  Result := -1;
end;

end.

