{$J+,Z4}
unit DropEdit;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche;       }
{        the NCCalcSize and NCPaint routines have been created based on        }
{          Patchou's "Themed RichEdit Class Helper for Visual C++" in          }
{                         RichEditThemed.cpp posted at                         }
{   <http://www.codeguru.com/Cpp/W-D/dislog/miscellaneous/article.php/c8729>   }
{------------------------------------------------------------------------------}

interface

uses
  Windows, RichEdit, RichEd20, RichOle, RichOleCallback, OLE2, Messages, SysUtils, Classes, Graphics, Forms,
  Controls, ComCtrls, StdCtrls, Dialogs, Menus, Printers, PrintDlg, DatedStrings, DropFiles, ConvApi,
  BitmapObject, WindowUtils, StrUtil, ToolTip, UxThemes;

const
  MAX_LINE_LEN = $FFFF;

type // TClipboardHandler

  TClipboardAction = (caCopy, caCut, caPaste, caStore, caRetrieve);

  TClipboardHandler = class
  private
    FDataObject: IDataObject;
    FClipboardData: TVarData;
    procedure ReleaseData;
    function GetTextRange(RichEdit: TRichEdit): ITextRange;
    procedure ReleaseTextRange(var TextRange: ITextRange);
  public
    constructor Create;
    destructor Destroy; override;
    procedure AssignData(DataObject: IDataObject);
    function  Execute(RichEdit: TRichEdit; Action: TClipboardAction; Release: Longbool): IDataObject;
    function  RestoreData(RichEdit: TRichEdit): Longint;
  end;

type // TDropEdit

  TDropEdit = Class;

  TFormatOption = (foLinks, foSelFont, foTabStops);
  TFormatOptions = Set of TFormatOption;

  TLineArray = Array[0..MAX_LINE_LEN] of Char;

  TSearchType = (stDown, stWholeWord, stMatchCase, stAnsiChar, stShowOptions);
  TSearchTypes = Set of TSearchType;

  TGetSearchText = function(const Find, FontName: String; var How: TSearchTypes): String;

  TPrintEvent =  procedure(PrintEdit: TDropEdit; const PrinterPrefs: TPrinterPrefs) of Object;

  {$IFNDEF CONDITIONALEXPRESSIONS}
  TContextPopupEvent = procedure(Sender: TObject; MousePos: TPoint; var Handled: Boolean) of Object;
  {$ENDIF}

  TOnUrlClick = procedure(const Url: String; Button: TMouseButton; X, Y: Longint) of Object;
  TOnUrlDblClick = procedure(const Url: String) of Object;

  TSelection = Record
    SelBeg: Longint;
    SelEnd: Longint;
  end;

  TTextPrinterData = Record
    PD: TPrinterData;
    Utf8HeaderText: String;
    Utf8FooterText: String;
    PreformattedText: Longbool;
  end;

  TDropEdit = class(TRichEdit)
  private
    FAcceptFiles: Longbool;
    FBorderStyle: TBorderStyle;
    FBorderWidth: Longint;
    FClipboardHandler: TClipboardHandler;
    FCursor: TCursor;
    FDatedStrings: TDatedStrings;
    FDropFiles: TDropFiles;
    FFindText: String;
    FFontColor: TColor;
    FGetSearchText: TGetSearchText;
    FHideCaret: Longbool;
    FLinkDblClickPos: Longint;
    FLinkToolTipPos: TPoint;
    FLinkToolTipRect: TRect;
    FMaxLength: Longint;
    FNCClientRect: TRect;
    FOriginPos: TPoint;
    FReadOnly: Longbool;
    FReplaceAll: Longbool;
    FReSearch: Longbool;
    FRichEditV3: Longbool;
    FRichEditVersion: Longint;
    FREOCallback: IREOCallback;
    FRtfConverter: TRtfConverter;
    FRtfStream: TMemoryStream;
    FSearching: Longbool;
    FSearchOptions: TSearchTypes;
    FShowObjects: Longbool;
    FSingleLine: Longbool;
    FTabSize: Longint;
    FTheme: THandle;
    FToolTipHandle: THandle;
    FUrlDetection: Longbool;
    FWasShowing: Longbool;
    FWordWrap: Longbool;
    FOnAfterSetPrintText: TPrintEvent;
    FOnBeforeSetPrintText: TPrintEvent;
    FOnFileDropped: TNotifyEvent;
    {$IFNDEF CONDITIONALEXPRESSIONS}
    FOnContextPopup: TContextPopupEvent;
    {$ENDIF}
    FOnUrlClick: TOnUrlClick;
    FOnUrlDblClick: TOnUrlDblClick;
    procedure OpenTheme;
    procedure CloseTheme;
    function  GetState: EDITTEXTSTATES;
    procedure FillScrollBarGap(State: EDITTEXTSTATES);
    function  ClientPosFromCharPos(CharPos: Longint): TPoint;
    procedure FormatText(Options: TFormatOptions);
    function  GetCharWidth(CharIndex: Integer): Integer;
    function  GetLineHeight(LineIndex: Integer): Integer;
    function  GetMouseButton(Code: Longint): TMouseButton;
    function  GetStreamInData: TEditStream;
    function  GetStreamOutData: TEditStream;
    procedure InternalHide;
    procedure InternalShow;
    function  IsUrl(CharPos: Longint): Longbool;
    procedure LinkEmailAddresses;
    procedure ReplaceAnsi(const Ansi: String);
    procedure ReplaceUnicode(const Unicode: String);
    procedure ResetToolTip;
    procedure HideSelection;
    procedure ShowSelection;
    procedure SetDefFont;
    procedure SetSelFont;
    procedure SetTabStops;
    procedure SwapDefaultColors;
    procedure UrlReturnKeyDown(CharPos: Longint);
    function  UrlDecode(const Url: String): String;
    function  UrlExtract(const CharRange: TCharRange): String;
    function  GetLinkRect(const CharRange: TCharRange): TRect;
    function  UrlClicked(const CharRange: TCharRange; Keys: Longint; const CursorPos: TPoint): Longbool;
    function  UrlDblClicked(const CharRange: TCharRange): Longbool;
    function  UrlMouseMove(const CharRange: TCharRange; const CursorPos: TPoint): Longbool;
    procedure SetAcceptFiles(Value: Longbool);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetBorderWidth(Value: Longint);
    function  GetContainsText: Longbool;
    function  GetFont: TFont;
    procedure SetFont(const Value: TFont);
    procedure SetHideCaret(const Value: Longbool);
    function  GetLines: Integer;
    procedure SetMaxLength(Value: Longint);
    procedure SetReadOnly(const Value: Longbool);
    function  GetScrollPos: TPoint;
    function  GetSelection: TSelection;
    procedure SetSelection(const Value: TSelection);
    procedure SetSingleLine(Value: Longbool);
    procedure SetShowObjects(Value: Longbool);
    function  GetObjectCount: Integer;
    procedure SetTabSize(const Value: Longint);
    procedure SetUrlDetection(const Value: Longbool);
    procedure SetWordWrap(const Value: Longbool);
    function  GetAnsiSelText: String;
    procedure SetAnsiSelText(const Value: String);
    function  GetAnsiText: String;
    procedure SetAnsiText(const Value: String);
    function  GetRtfText: String;
    procedure SetRtfText(const Value: String);
    function  GetUtf8SelText: String;
    procedure SetUtf8SelText(const Value: String);
    function  GetUtf8Text: String;
    procedure SetUtf8Text(const Value: String);
    procedure ResetText;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnter(var Message: TMessage); message CM_ENTER;
    procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMContextMenu(var Message: TWMMouse); message WM_CONTEXTMENU;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMKeyDown(var Message: TWMKey); message WM_KEYDOWN;
    procedure WMMouseMove(var Message: TWMMouse); message WM_MOUSEMOVE;
    procedure WMLButtonDblClk(var Message: TWMMouse); message WM_LBUTTONDBLCLK;
    procedure WMRButtonDown(var Message: TWMMouse); message WM_RBUTTONDOWN;
    procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TWMNCPaint); message WM_NCPAINT;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMSysKeyDown(var Message: TWMKey); message WM_SYSKEYDOWN;
    procedure WMThemeChanged(var Message: TMessage); message WM_THEMECHANGED;
    procedure WMTimer(var Message: TWMTimer); message WM_TIMER;
  protected
    procedure Change; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure RecreateWnd;
    procedure ResetParams;
    procedure ReadFromStream(Stream: TStream); virtual;
    procedure WriteToStream(Stream: TStream); virtual;
    {$IFDEF CONDITIONALEXPRESSIONS}
    function  GetSelText: String; override;
    {$ELSE}
    function  GetSelText: String;
    {$ENDIF}
    procedure SetSelText(const Value: String);
    function  GetText: String;
    procedure SetText(const Value: String);
    property  CharWidth[CharIndex: Integer]: Integer read GetCharWidth;
    property  LineHeight[LineIndex: Integer]: Integer read GetLineHeight;
    property  ClipboardHandler: TClipboardHandler read FClipboardHandler;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Copy;
    procedure Cut;
    procedure Paste;
    function  InsertBitmap(Bitmap: TBitmap; CharPos: Integer): Longint;
    function  FindText(const SearchStr: String; StartPos, Length: Integer): Integer;
    function  SearchText: Longbool;
    function  GetLine(Index: Longint): TLineArray;
    function  GetLineStartIndex(Line: Longint): Longint;
    function  IsLineBlank(const aLine: TLineArray): Longbool;
    function  WidthAdjustText(Device: HDC; const Text: String; MaxWidth: Integer): String;
    function  GetTextLen: Integer;
    procedure SetSel(SelStart, SelLength: Integer);
    procedure SelectAll;
    procedure DeleteSel;
    procedure ResetSel;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure PrintWithMargins(var TPD: TTextPrinterData);
    procedure LoadFromFile(const FileName: String);
    procedure LoadFromRtfFile(const FileName: String);
    procedure LoadFromUtf8File(const FileName: String);
    procedure LoadFromHtmlFile(const FileName: String);
    procedure LoadFromHtmlString(const HtmlString: String);
    procedure LoadFromStream(Stream: TStream);
    procedure LoadFromRtfStream(Stream: TStream);
    procedure SaveToFile(const FileName: String);
    procedure SaveToAnsiFile(const FileName: String);
    procedure SaveToRtfFile(const FileName: String);
    procedure SaveToUtf8File(const FileName: String);
    procedure SaveToHtmlFile(const FileName: String);
    procedure SaveToHtmlString(const HtmlString: String);
    procedure SaveToStream(Stream: TStream);
    procedure SaveToAnsiStream(Stream: TStream);
    procedure SaveToRtfStream(Stream: TStream);
    procedure ReloadRtfStream;
    procedure ShowDatedStrings;
    property  AnsiText: String read GetAnsiText write SetAnsiText;
    property  AnsiSelText: String read GetAnsiSelText write SetAnsiSelText;
    property  DatedStrings: TDatedStrings read FDatedStrings;
    property  DropFiles: TDropFiles read FDropFiles;
    property  GetSearchText: TGetSearchText read FGetSearchText write FGetSearchText;
    property  ObjectCount: Integer read GetObjectCount;
    property  ReSearch: Longbool read FReSearch write FReSearch;
    property  RichEditV3: Longbool read FRichEditV3;
    property  RichEditVersion: Longint read FRichEditVersion;
    property  RtfText: String read GetRtfText write SetRtfText;
    property  Searching: Longbool read FSearching;
    property  SelText: String read GetSelText write SetSelText;
    property  Text: String read GetText write SetText;
    property  Utf8Text: String read GetUtf8Text write SetUtf8Text;
    property  Utf8SelText: String read GetUtf8SelText write SetUtf8SelText;
    property  WideText: String read GetText write SetText;
    property  WideSelText: String read GetSelText write SetSelText;
  published
    property  AutoSelect;
    property  AcceptFiles: Longbool read FAcceptFiles write SetAcceptFiles;
    property  BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle;
    property  BorderWidth: Longint read FBorderWidth write SetBorderWidth;
    property  Font: TFont read GetFont write SetFont;
    property  ContainsText: Longbool read GetContainsText;
    property  HideCaret: Longbool read FHideCaret write SetHideCaret;
    property  Lines: Integer read GetLines;
    property  MaxLength: Longint read FMaxLength write SetMaxLength;
    property  ReadOnly: Longbool read FReadOnly write SetReadOnly;
    property  SearchOptions: TSearchTypes read FSearchOptions write FSearchOptions;
    property  ShowObjects: Longbool read FShowObjects write SetShowObjects;
    property  SingleLine: Longbool read FSingleLine write SetSingleLine;
    property  TabSize: Longint read FTabSize write SetTabSize;
    property  UrlDetection: Longbool read FUrlDetection write SetUrlDetection;
    property  WordWrap: Longbool read FWordWrap write SetWordWrap;
    property  OnClick;
    property  OnDblClick;
    property  OnAfterSetPrintText: TPrintEvent read FOnAfterSetPrintText write FOnAfterSetPrintText;
    property  OnBeforeSetPrintText: TPrintEvent read FOnBeforeSetPrintText write FOnBeforeSetPrintText;
    {$IFNDEF CONDITIONALEXPRESSIONS}
    property  OnContextPopup: TContextPopupEvent read FOnContextPopup write FOnContextPopup;
    {$ENDIF}
    property  OnFileDropped: TNotifyEvent read FOnFileDropped write FOnFileDropped;
    property  OnUrlClick: TOnUrlClick read FOnUrlClick write FOnUrlClick;
    property  OnUrlDblClick: TOnUrlDblClick read FOnUrlDblClick write FOnUrlDblClick;
  end;

procedure Register;


implementation

uses
  {$IFDEF CONDITIONALEXPRESSIONS}
  RtlConsts,
  {$ENDIF}
  Consts;


// Format tools ---------------------------------------------------------------

procedure RaiseConvertError(Index: Integer; Unicode: Longbool);
const
  INVALID_FORMAT = 'Invalid %s string format argument #';
var
  sType: String;
begin
  if Unicode then
    sType := 'WS'
  else sType := 'AS';
  Raise EConvertError.Create(Format(INVALID_FORMAT, [sType]) + IntToStr(succ(Index)));
end;

procedure CheckFormat(Format: Char; const Utf8String: String; Index: Integer; Unicode: Longbool);

  function IndexedArgPos(const ArgString: String; Index: Integer): Integer; register; assembler;
  const
    PC = '%';
  asm		// EAX = @ArgString, EDX = Index
    PUSH	EBX
    PUSH	EAX
    OR		EAX,EAX
    JE		@EXIT
    OR		EDX,EDX
    JS		@EXIT
    MOV		ECX,[EAX - 04h]
    ADD		EAX,[EAX - 04h]
    XOR		EBX,EBX
    NEG		ECX
    @LOOP:
    CMP		BYTE PTR [EAX + ECX],PC
    JE		@MATCH
    INC		ECX
    JNE		@LOOP
    @EXIT:
    XOR		EAX,EAX
    POP		EDX
    POP		EBX
    RET
    @MATCH:
    CMP		BYTE PTR [EAX + ECX + 1],PC
    JNE		@CHECKINDEX
    ADD		ECX,2
    JNL		@EXIT
    JMP		@LOOP
    @CHECKINDEX:
    CMP		EDX,EBX
    JE		@FOUND
    INC		EBX
    INC		ECX
    JNL		@EXIT
    JMP		@LOOP
    @FOUND:
    INC		ECX
    POP		EDX
    POP		EBX
    MOV		EAX,ECX
    ADD		EAX,[EDX - 04h]
  end;

var
  iPos: Integer;
begin
  iPos := IndexedArgPos(Utf8String, Index);
  if (iPos = 0) or (Utf8String[succ(iPos)] <> Format) then RaiseConvertError(Index, Unicode);
end;

function GetArgList(const ArgString: String; Unicode: Longbool; const Args: Array of Const): PChar;
var
  iSize, iIndex: Integer;
  pArgs: PChar;
begin
  iSize := 0;
  for iIndex := Low(Args) to High(Args) do begin
    case Args[iIndex].VType of
      vtInteger: begin
	inc(iSize, SizeOf(Integer));
	CheckFormat('d', ArgString, iIndex, Unicode);
      end;
      vtAnsiString: begin
	inc(iSize, SizeOf(Pointer));
	CheckFormat('s', ArgString, iIndex, Unicode);
      end;
    else
      RaiseConvertError(iIndex, Unicode);
    end;
  end;
  Result := StrAlloc(iSize);
  if Result <> NIL then begin
    try
      pArgs := Result;
      for iIndex := Low(Args) to High(Args) do begin
	case Args[iIndex].VType of
	  vtInteger: iSize := SizeOf(Integer);
	  vtAnsiString: iSize := SizeOf(Pointer);
	else
	  Continue;
	end;
	Move(Args[iIndex], pArgs^, iSize);
	inc(pArgs, iSize);
      end;
    except
      StrDispose(Result);
      Raise;
    end;
  end;
end;

function wvsprintfA(Output: PChar; Format: PChar; arglist: va_list): Integer; stdcall;
	 external user32 name 'wvsprintfA';

function ASFormat(const AString: String; const Args: Array of Const): String;
var
  pArgs: PChar;
begin
  if AString <> E then begin
    pArgs := NIL;
    try
      pArgs := GetArgList(NoBomAnsiStringToUtf8(AString), false, Args);
      try
	SetString(Result, NIL, 1025);
	Delete(Result, wvsprintfA(Pointer(Result), Pointer(AString), Pointer(pArgs)) + Length(aNUL), MAXINT);
      except
	Result := E;
      end;
    finally
      StrDispose(pArgs);
    end;
  end
  else Result := E;
end;

function wvsprintfW(Output: PWideChar; Format: PWideChar; arglist: va_list): Integer; stdcall;
	 external user32 name 'wvsprintfW';

function WSFormat(const WString: String; const Args: Array of Const): String;
var
  pArgs: PChar;
begin
  if WString <> E then begin
    pArgs := NIL;
    try
      pArgs := GetArgList(NoBomWideStringToUtf8(WString), true, Args);
      try
	SetString(Result, NIL, 2050);
	Delete(Result, wvsprintfW(Pointer(Result), Pointer(WString), Pointer(pArgs)) shl 1 + succ(Length(wNUL)), MAXINT);
      except
	Result := E;
      end;
    finally
      StrDispose(pArgs);
    end;
  end
  else Result := E;
end;


// Interface utilities ---------------------------------------------------------

function GetRichEditOleInterface(RichEdit: TRichEdit): IRichEditOle;
begin
  if SendMessageA(RichEdit.Handle, EM_GETOLEINTERFACE, 0, Longint(@Result)) = 0 then Result := nil;
end;

function GetTextDocumentInterface(RichEdit: TRichEdit): ITextDocument;
var
  reoInterface: IRichEditOle;
begin
  Result := nil;
  reoInterface := GetRichEditOleInterface(RichEdit);
  if reoInterface <> nil then begin
    try
      reoInterface.QueryInterface(IID_ITextDocument, Result);
    finally
      reoInterface.Release;
    end;
  end;
end;


// TClipboardHandler -----------------------------------------------------------

constructor TClipboardHandler.Create;
begin
  inherited Create;
  with FClipboardData do begin
    VType := varUnknown or varByRef;
    VUnknown := @FDataObject;
  end;
end;

destructor TClipboardHandler.Destroy;
begin
  inherited Destroy;
end;

procedure TClipboardHandler.AssignData(DataObject: IDataObject);
begin
  FDataObject := DataObject;
end;

function TClipboardHandler.Execute(RichEdit: TRichEdit; Action: TClipboardAction; Release: Longbool): IDataObject;
var
  trInterface: ITextRange;
  lPos: Longint;
begin
  Result := nil;
  trInterface := GetTextRange(RichEdit);
  if trInterface <> nil then begin
    try
      with trInterface do begin
	lPos := -1;
	case Action of
	  caCopy: Copy(nil);
	  caCut: begin
	    Cut(nil);
	    GetEnd(lPos);
	  end;
	  caPaste: begin
	    Paste(nil, 0);
	    GetEnd(lPos);
	  end;
	  caStore: Copy(@FClipboardData);
	  caRetrieve: Paste(@FClipboardData, 0);
	end;
      end;
      if lPos > -1 then SendMessage(RichEdit.Handle, EM_SETSEL, lPos, lPos);
    finally
      ReleaseTextRange(trInterface);
      if Release then ReleaseData;
      Result := FDataObject;
    end;
  end
  else begin
    case Action of
      caCopy, caStore: SendMessage(RichEdit.Handle, WM_COPY, 0, 0);
      caCut: SendMessage(RichEdit.Handle, WM_CUT, 0, 0);
      caPaste, caRetrieve: SendMessage(RichEdit.Handle, WM_PASTE, 0, 0);
    end;
  end;
end;

function TClipboardHandler.RestoreData(RichEdit: TRichEdit): Longint;
var
  trInterface: ITextRange;
begin
  Result := E_NOINTERFACE;
  trInterface := GetTextRange(RichEdit);
  if trInterface <> nil then begin
    try
      with trInterface do begin
	Result := Copy(nil);
	if Result = S_OK then Result := Paste(nil, 0);
      end;
    finally
      ReleaseTextRange(trInterface);
      ReleaseData;
    end;
  end
  else with RichEdit do begin
    SendMessage(Handle, WM_COPY, 0, 0);
    SendMessage(Handle, WM_PASTE, 0, 0);
  end;
end;

procedure TClipboardHandler.ReleaseData;
begin
  if FDataObject <> nil then begin
    FDataObject.Release;
    FDataObject := nil;
  end;
end;

function TClipboardHandler.GetTextRange(RichEdit: TRichEdit): ITextRange;
var
  tdInterface: ITextDocument;
  lSelBeg, lSelEnd: Longint;
begin
  Result := nil;
  tdInterface := GetTextDocumentInterface(RichEdit);
  if tdInterface <> nil then begin
    try
      SendMessage(RichEdit.Handle, EM_GETSEL, Longint(@lSelBeg), Longint(@lSelEnd));
      tdInterface.Range(lSelBeg, lSelEnd, Result);
    finally
      tdInterface.Release;
    end;
  end;
end;

procedure TClipboardHandler.ReleaseTextRange(var TextRange: ITextRange);
begin
  if TextRange <> nil then begin
    TextRange.Release;
    TextRange := nil;
  end;
end;


// TDropEdit -------------------------------------------------------------------

function EditStreamLoad(dwCookie: DWord; pbBuff: PByte; cb: Longint; var pcb: Longint): DWord; stdcall;
begin
  Result := 0;
  try
    pcb := TDropEdit(dwCookie).FRtfStream.Read(pbBuff^, cb);
  except
    Result := 1;
  end;
end;

function EditStreamSave(dwCookie: DWord; pbBuff: PByte; cb: Longint; var pcb: Longint): DWord; stdcall;
begin
  Result := 0;
  try
    pcb := TDropEdit(dwCookie).FRtfStream.Write(pbBuff^, cb);
  except
    Result := 2;
  end;
end;

procedure TDropEdit.OpenTheme;
begin
  if UseThemes and (FBorderStyle = bsSingle) and ((Color = clWindow) or (Color = clBtnFace)) then begin
    FTheme := OpenThemeData(Handle, 'EDIT');
  end;
end;

procedure TDropEdit.CloseTheme;
begin
  if FTheme <> 0 then begin
    CloseThemeData(FTheme);
    FTheme := 0;
  end;
end;

function TDropEdit.GetState: EDITTEXTSTATES;
begin
  if Enabled then begin
    if FReadOnly then
      Result := ETS_READONLY
    else Result := ETS_NORMAL;
  end
  else Result := ETS_DISABLED;
end;

procedure TDropEdit.FillScrollBarGap(State: EDITTEXTSTATES);
var
  rScrollBarGap: TRect;
  hDC, hBrush: THandle;
  crDialog: COLORREF;
begin
  if (BorderStyle = bsSingle)
  and (GetWindowLong(WindowHandle, GWL_STYLE) and (WS_HSCROLL or WS_VSCROLL) = (WS_HSCROLL or WS_VSCROLL)) then begin
    with FNCClientRect do begin
      rScrollBarGap := Rect(Width - GetThemeSysSize(FTheme, SM_CXVSCROLL) - Left,
			    Height - GetThemeSysSize(FTheme, SM_CYHSCROLL) - Top,
			    Width - Right, Height - Bottom);
    end;
    hDC := GetWindowDC(WindowHandle);
    GetThemeColor(FTheme, EP_EDITTEXT, ETS_DISABLED, TMT_FILLCOLOR, crDialog);
    hBrush := CreateSolidBrush(crDialog);
    try
      FillRect(hDC, rScrollBarGap, hBrush);
    finally
      DeleteObject(hBrush);
      ReleaseDC(WindowHandle, hDC);
    end;
  end;
end;

function TDropEdit.ClientPosFromCharPos(CharPos: Longint): TPoint;
var
  iClientPos: Integer;
begin
  if FRichEditVersion = 2 then begin
    iClientPos := SendMessage(WindowHandle, EM_POSFROMCHAR, CharPos, 0);
    Result := MakePoints(iClientPos);
  end
  else SendMessage(WindowHandle, EM_POSFROMCHAR, Longint(@Result), CharPos);
end;

procedure TDropEdit.FormatText(Options: TFormatOptions);
var
  Sel: TSelection;
begin
  if not (foLinks in Options) then begin
    Sel := GetSelection;
    InternalHide;
  end;
  try
    if foLinks in Options then LinkEmailAddresses;
    SendMessage(WindowHandle, EM_SETSEL, 0, -1);
    if foSelFont in Options then SetSelFont;
    if foTabStops in Options then SetTabStops;
  finally
    if not (foLinks in Options) then begin
      SetSelection(Sel);
      InternalShow;
    end;
  end;
end;

function TDropEdit.GetCharWidth(CharIndex: Integer): Integer;
var
  ftWidth: TFont;
  cfChar: TCharFormat;
  hDC, hFont: THandle;
  tmFont: TTextMetric;
begin
  Result := 0;
  ftWidth := nil;
  try
    if CharIndex > -1 then begin
      FillChar(cfChar, SizeOf(TCharFormat), 0);
      with cfChar do begin
	cbSize := SizeOf(TCharFormat);
	dwMask := CFM_SIZE;
      end;
      SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfChar));
      ftWidth := CreateFontFromCharFormat(cfChar);
    end;
    if (CharIndex < 0) or (Result = 0) then begin
      SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_DEFAULT, Longint(@cfChar));
      ftWidth := CreateFontFromCharFormat(cfChar);
    end;
    if ftWidth <> nil then begin
      hDC := GetDC(WindowHandle);
      if hDC <> 0 then begin
	hFont := 0;
	try
	  hFont := SelectObject(hDC, Font.Handle);
	  if (hFont <> 0) and GetTextMetrics(hDC, tmFont) then Result := tmFont.tmAveCharWidth;
	finally
	  if hFont <> 0 then SelectObject(hDC, hFont);
	  ReleaseDC(WindowHandle, hDC);
	end;
      end;
    end;
  finally
    if ftWidth <> nil then ftWidth.Free;
  end;
end;

function TDropEdit.GetLineHeight(LineIndex: Integer): Integer;
var
  ftHeight: TFont;
  cfChar: TCharFormat;
  hDC, hFont: THandle;
  tmFont: TTextMetric;
begin
  Result := 0;
  ftHeight := nil;
  try
    if LineIndex > -1 then begin
      FillChar(cfChar, SizeOf(TCharFormat), 0);
      with cfChar do begin
	cbSize := SizeOf(TCharFormat);
	dwMask := CFM_SIZE;
      end;
      SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfChar));
      ftHeight := CreateFontFromCharFormat(cfChar);
    end;
    if (LineIndex < 0) or (Result = 0) then begin
      SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_DEFAULT, Longint(@cfChar));
      ftHeight := CreateFontFromCharFormat(cfChar);
    end;
    if ftHeight <> nil then begin
      hDC := GetDC(WindowHandle);
      if hDC <> 0 then begin
	hFont := 0;
	try
	  hFont := SelectObject(hDC, Font.Handle);
	  if (hFont <> 0) and GetTextMetrics(hDC, tmFont) then with tmFont do Result := tmHeight + tmExternalLeading;
	finally
	  if hFont <> 0 then SelectObject(hDC, hFont);
	  ReleaseDC(WindowHandle, hDC);
	end;
      end;
    end;
  finally
    if ftHeight <> nil then ftHeight.Free;
  end;
end;

function TDropEdit.GetMouseButton(Code: Longint): TMouseButton;
begin
  case Code of
    MK_LBUTTON: Result := mbLeft;
    MK_RBUTTON: Result := mbRight;
  else
    Result := mbMiddle;
  end;
end;

function TDropEdit.GetStreamInData: TEditStream;
begin
  with Result do begin
    dwCookie := DWord(Self);
    dwError := 0;
    pfnCallback := @EditStreamLoad;
  end;
end;

function TDropEdit.GetStreamOutData: TEditStream;
begin
  with Result do begin
    dwCookie := DWord(Self);
    dwError := 0;
    pfnCallback := @EditStreamSave;
  end;
end;

procedure TDropEdit.InternalHide;
begin
  FWasShowing := Showing;
  if FWasShowing then begin
    SendMessage(WindowHandle, WM_SETREDRAW, 0, 0);
    Visible := false;
  end;
end;

procedure TDropEdit.InternalShow;
begin
  if FWasShowing then begin
    Visible := true;
    SendMessage(WindowHandle, WM_SETREDRAW, 1, 0);
  end;
  FWasShowing := false;
end;

function TDropEdit.IsUrl(CharPos: Longint): Longbool;
var
  Sel: TSelection;
  cfLink: TCharFormat;
begin
  Sel := GetSelection;
  HideSelection;
  try
    if CharPos <> SelStart then SendMessage(WindowHandle, EM_SETSEL, CharPos, CharPos);
    FillChar(cfLink, SizeOf(TCharFormat), 0);
    with cfLink do begin
      cbSize := SizeOf(TCharFormat);
      dwMask := CFM_LINK;
    end;
    SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfLink));
    Result := (cfLink.dwEffects and CFE_LINK <> 0);
  finally
    SetSelection(Sel);
    ShowSelection;
  end;
end;

procedure TDropEdit.LinkEmailAddresses;
const
  BEG_STR = '"(<[{';
  END_STR = '")>]}';
  BEG_CHARS = ['"', '(', '<', '[', '{'];
  END_CHARS = ['"', ')', '>', ']', '}'];
  DELIMITERS = [#0..#32, #92, #127..#255];
  BEG_DELIMITERS = DELIMITERS + BEG_CHARS;
  END_DELIMITERS = DELIMITERS + END_CHARS;
var
  cfLink: TCharFormat;
  ftLink: TFindTextA;
  aLine: TLineArray;
  iAtPos, iLineNo, iLineIndex, iCharPos, iLineLen, iLinkBeg, iLinkEnd: Integer;
begin
  if FUrlDetection and (FRichEditVersion > 1) then begin
    FillChar(cfLink, SizeOf(TCharFormat), 0);
    with cfLink do begin
      cbSize := SizeOf(TCharFormat);
      dwMask := CFM_LINK;
    end;
    ftLink.lpstrText := AT;
    ftLink.chrg.cpMax := -1;
    iAtPos := 1;
    repeat
      ftLink.chrg.cpMin := iAtPos;
      iAtPos := SendMessage(WindowHandle, EM_FINDTEXT, FT_DOWN, Longint(@ftLink));
      if iAtPos > -1 then begin
	SendMessage(WindowHandle, EM_SETSEL, iAtPos, succ(iAtPos));
	SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfLink));
	if (cfLink.dwEffects and CFE_LINK) = 0 then begin
	  PWideChar(@aLine)^ := WideChar(MAX_LINE_LEN);
	  iLineNo := SendMessage(WindowHandle, EM_LINEFROMCHAR, iAtPos, 0);
	  iLineIndex := SendMessage(WindowHandle, EM_LINEINDEX, iLineNo, 0);
	  iLineLen := pred(SendMessage(WindowHandle, EM_GETLINE, iLineNo, Longint(@aLine)));
	  iLinkBeg := pred(iAtPos - iLineIndex);
	  iLinkEnd := iLinkBeg + 2;
	  while (iLinkBeg >= 0) and (iLinkEnd < iLineLen) do begin
	    if not (aLine[iLinkBeg] in BEG_DELIMITERS) and not (aLine[iLinkEnd] in END_DELIMITERS) then begin
	      while (iLinkBeg >= 0) and not (aLine[iLinkBeg] in BEG_DELIMITERS) do dec(iLinkBeg);
	      while (iLinkEnd < iLineLen) and not (aLine[iLinkEnd] in END_DELIMITERS) do inc(iLinkEnd);
	      iCharPos := Pos(aLine[iLinkBeg], BEG_STR);
	      if iCharPos > 0 then while (iLinkEnd < iLineLen) and (aLine[iLinkEnd] <> END_STR[iCharPos]) do inc(iLinkEnd);
	      inc(iLinkBeg);
	      cfLink.dwEffects := CFE_LINK;
	      SendMessage(WindowHandle, EM_SETSEL, iLineIndex + iLinkBeg, iLineIndex + iLinkEnd);
	      SendMessage(WindowHandle, EM_SETCHARFORMAT, SCF_SELECTION, Longint(@cfLink));
	    end;
	    while not (aLine[iLinkEnd] in [aNUL, AT]) do inc(iLinkEnd);
	    if iLinkEnd >= iLineLen then Break;
	    iLinkBeg := pred(iLinkEnd);
	    inc(iLinkEnd);
	  end;
	  iAtPos := iLineIndex + succ(iLineLen);
	end
	else inc(iAtPos);
	Application.ProcessMessages;
      end
      else Break;
    until iAtPos = -1;
  end;
end;

procedure TDropEdit.ReplaceAnsi(const Ansi: AnsiString);
var
  steAnsi: TSetTextEx;
begin
  if FRichEditV3 then with steAnsi do begin
    flags := ST_SELECTION;
    codepage := CP_ACP;
    SendMessage(WindowHandle, EM_SETTEXTEX, Longint(@steAnsi), Longint(Ansi));
  end
  else SendMessage(WindowHandle, EM_REPLACESEL, 1, Longint(Ansi));
end;

procedure TDropEdit.ReplaceUnicode(const Unicode: String);
var
  steUnicode: TSetTextEx;
begin
  with steUnicode do begin
    flags := ST_SELECTION;
    codepage := 1200;
    SendMessage(WindowHandle, EM_SETTEXTEX, Longint(@steUnicode), Longint(Unicode));
  end;
end;

procedure TDropEdit.ResetToolTip;
begin
  if FToolTipHandle <> 0 then begin
    KillTimer(WindowHandle, FToolTipHandle);
    FillChar(FLinkToolTipPos, SizeOf(TPoint), -1);
    FillChar(FLinkToolTipRect, SizeOf(TRect), -1);
    FToolTipHandle := DestroyTrackingToolTip(FToolTipHandle, WindowHandle, WindowHandle);
  end;
end;

procedure TDropEdit.HideSelection;
begin
  SendMessage(WindowHandle, EM_HIDESELECTION, 1, 0);
end;

procedure TDropEdit.ShowSelection;
begin
  SendMessage(WindowHandle, EM_HIDESELECTION, 0, 0);
end;

procedure TDropEdit.SetDefFont;
begin
  DefAttributes.Assign(Font);
end;

procedure TDropEdit.SetSelFont;
begin
  SelAttributes.Assign(Font);
end;

procedure TDropEdit.SetTabStops;
var
  pfTabs: TParaFormat;
  iIndex: Integer;
begin
  FillChar(pfTabs, SizeOf(TParaFormat), 0);
  with pfTabs do begin
    cbSize := SizeOf(TParaFormat);
    dwMask := PFM_TABSTOPS;
    cTabCount := MAX_TAB_STOPS;
    rgxTabs[0] := MulDiv(FTabSize * GetCharWidth(-1), TWIPS, Font.PixelsPerInch);
    for iIndex := 1 to High(rgxTabs) do rgxTabs[iIndex] := rgxTabs[pred(iIndex)] + rgxTabs[0];
  end;
  SendMessage(WindowHandle, EM_SETPARAFORMAT, 0, Longint(@pfTabs));
end;

procedure TDropEdit.SwapDefaultColors;
begin
  if (Color = clWindow) or (Color = clBtnFace) then begin
    if Enabled then begin
      Font.Color := FFontColor;
      if PlainText and not ParentColor then begin
	case Color of
	  clWindow: if FReadOnly then Color := clBtnFace;
	  clBtnFace: if not FReadOnly then Color := clWindow;
	end;
      end;
    end
    else begin
      Font.Color := clGrayText;
      if (Color = clWindow) and not ParentColor then Color := clBtnFace;
    end;
  end;
end;

procedure TDropEdit.UrlReturnKeyDown(CharPos: Longint);
begin
  if IsUrl(CharPos) then with ClientPosFromCharPos(CharPos) do begin
    PostMessage(WindowHandle, WM_LBUTTONDBLCLK, MK_LBUTTON, MakeLParam(X, Y));
  end;
end;

function TDropEdit.UrlDecode(const Url: String): String;
var
  pResult: PChar;
  iUrlLen, iResPos, iBegPos, iEndPos, iLen, iPos: Integer;
begin
  Result := E;
  iUrlLen := Length(Url);
  if iUrlLen <> 0 then begin
    SetString(Result, nil, iUrlLen);
    pResult := Pointer(Result);
    iResPos := 0;
    iBegPos := 1;
    iEndPos := 1;
    while iEndPos <= iUrlLen do begin
      if Url[iEndPos] = '%' then begin
	iLen := iEndPos - iBegPos;
	Move(Url[iBegPos], pResult[iResPos], iLen);
	inc(iResPos, iLen);
	iPos := succ(iEndPos);
	if iPos < iUrlLen then begin
	  pResult[iResPos] := HexToChar(@Url[iEndPos]);
	  if pResult[iResPos] = aNUL then begin
	    iBegPos := iEndPos;
	    inc(iEndPos);
	    Continue;
	  end
	  else iLen := 1;
	  iBegPos := iEndPos + 3;
	  iEndPos := iBegPos;
	  inc(iResPos, iLen);
	end
	else begin
	  iBegPos := succ(iUrlLen);
	  Break;
	end;
      end
      else inc(iEndPos);
    end;
    iLen := succ(iUrlLen - iBegPos);
    if iLen > 0 then begin
      Move(Url[iBegPos], pResult[iResPos], iLen);
      inc(iResPos, iLen);
    end;
    Delete(Result, succ(iResPos), MAXINT);
    Result := DecodeIdns(Result);
  end;
end;

function TDropEdit.UrlExtract(const CharRange: TCharRange): String;
var
  trLink: TTextRangeW;
begin
  with CharRange do SetWSString(Result, cpMax - cpMin);
  with trLink do begin
    chrg := CharRange;
    lpstrText := Pointer(Result);
  end;
  if SendMessage(WindowHandle, EM_GETTEXTRANGE, 0, Longint(@trLink)) > 0 then begin
    Result := WideStringToAnsi(Result);
    if (Pos(AT, Result) > 0)
    and not (LoStrAtPos('mailto:', Result, 1) or
	     LoStrAtPos('http:', Result, 1) or
	     LoStrAtPos('ftp:', Result, 1)) then begin
      Insert('mailto:', Result, 1);
    end;
  end
  else Result := E;
end;

function TDropEdit.GetLinkRect(const CharRange: TCharRange): TRect;
var
  cfLink: TCharFormat2;
  iPos, iLine, iHeight: Integer;
begin
  Result := EMPTY_RECT;
  if FUrlDetection and (FRichEditVersion > 1) then with CharRange do begin
    SendMessage(WindowHandle, EM_POSFROMCHAR, WPARAM(@Result.TopLeft), cpMin);
    FillChar(cfLink, SizeOf(TCharFormat2), 0);
    with cfLink do begin
      cbSize := SizeOf(TCharFormat2);
      dwMask := CFM_LINK or CFM_HIDDEN;
    end;
    iPos := pred(cpMax);
    repeat
      inc(iPos);
      SendMessage(WindowHandle, EM_SETSEL, iPos, iPos);
      SendMessage(WindowHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfLink));
    until (iPos >= cpMax) and ((cfLink.dwEffects and (CFE_LINK or CFE_HIDDEN)) = 0);
    SendMessage(WindowHandle, EM_POSFROMCHAR, WPARAM(@Result.BottomRight), iPos);
    iLine := SendMessage(WindowHandle, EM_LINEFROMCHAR, cpMax, 0);
    iHeight := GetLineHeight(iLine);
    inc(Result.Bottom, iHeight);
  end;
end;

function TDropEdit.UrlClicked(const CharRange: TCharRange; Keys: Longint; const CursorPos: TPoint): Longbool;
var
  sUrl: String;
begin
  sUrl := UrlExtract(CharRange);
  Result := (sUrl <> E);
  if Result and Assigned(FOnUrlClick) then with CursorPos do FOnUrlClick(sUrl, GetMouseButton(Keys), X, Y);
end;

function TDropEdit.UrlDblClicked(const CharRange: TCharRange): Longbool;
var
  sUrl: String;
begin
  FLinkDblClickPos := SelStart;
  sUrl := UrlExtract(CharRange);
  Result := (sUrl <> E);
  if Result and Assigned(FOnUrlDblClick) then FOnUrlDblClick(sUrl);
end;

function TDropEdit.UrlMouseMove(const CharRange: TCharRange; const CursorPos: TPoint): Longbool;
var
  sUrl: String;
  fWinXP: Boolean;
  td: TTitleData;
begin
  sUrl := UrlDecode(UrlExtract(CharRange));
  if not IsMailtoUrl(sUrl) then begin
    SetHideCaret(true);
    try
      if (FToolTipHandle = 0) or not PtInRect(FLinkToolTipRect, CursorPos) then begin
	if FToolTipHandle <> 0 then ResetToolTip;
	fWinXP := (MainWinVersion < 6);
	with td do begin
	  tdIcon := TTI_INFO;
	  tdTitle := 'URL-Info'
	end;
	FLinkToolTipPos := CursorPos;
	FLinkToolTipRect := GetLinkRect(CharRange);
	FToolTipHandle := CreateTrackingToolTip(WindowHandle, WindowHandle, FLinkToolTipRect,
						sUrl, ClientWidth div 2, td, nil,
						fWinXP, not fWinXP, false);
	if FToolTipHandle <> 0 then SetTimer(WindowHandle, FToolTipHandle, GetDoubleClickTime, nil);
      end;
      Result := (FToolTipHandle <> 0);
      Exit;
    finally
      SetHideCaret(false);
    end;
  end;
  Result := false;
end;

procedure TDropEdit.SetAcceptFiles(Value: Longbool);
begin
  if FDropFiles <> nil then FDropFiles.AcceptFiles := Value;
  FAcceptFiles := Value;
end;

procedure TDropEdit.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle then begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TDropEdit.SetBorderWidth(Value: Longint);
var
  rText: TRect;
begin
  FBorderWidth := Value;
  if WindowHandle <> 0 then begin
    rText := ClientRect;
    inc(rText.Left, Value);
    SendMessage(WindowHandle, EM_SETRECTNP, 0, Longint(@rText));
  end;
end;

function TDropEdit.GetContainsText: Longbool;
begin
  Result := (GetWindowTextLength(WindowHandle) > 0);
end;

function TDropEdit.GetFont: TFont;
begin
  Result := inherited Font;
end;

procedure TDropEdit.SetFont(const Value: TFont);
begin
  if Value <> inherited Font then begin
    if PlainText then
      inherited Font.Assign(Value)
    else SetDefFont;
    FFontColor := Font.Color;
  end;
end;

function TDropEdit.GetLines: Integer;
begin
  Result := SendMessage(WindowHandle, EM_GETLINECOUNT, 0, 0);
end;

procedure TDropEdit.SetHideCaret(const Value: Longbool);
begin
  FHideCaret := Value;
  SendMessage(WindowHandle, EM_HIDESELECTION, ord(FHideCaret = true), 0);
  if FHideCaret then
    Windows.HideCaret(WindowHandle)
  else Windows.ShowCaret(WindowHandle);
end;

procedure TDropEdit.SetMaxLength(Value: Longint);
begin
  if Value <> FMaxLength then begin
    SendMessage(WindowHandle, EM_EXLIMITTEXT, 0, Value);
    FMaxLength := Value;
  end;
end;

procedure TDropEdit.SetReadOnly(const Value: Longbool);
begin
  SendMessage(WindowHandle, EM_SETREADONLY, ord(Value), 0);
  SetHideCaret(Value);
  FReadOnly := Value;
  SwapDefaultColors;
end;

function TDropEdit.GetScrollPos: TPoint;
begin
  SendMessage(WindowHandle, EM_GETSCROLLPOS, 0, Longint(@Result));
end;

function TDropEdit.GetSelection: TSelection;
begin
  BeginUpdate;
  with Result do SendMessage(WindowHandle, EM_GETSEL, Longint(@SelBeg), Longint(@SelEnd));
end;

procedure TDropEdit.SetSelection(const Value: TSelection);
begin
  with Value do SendMessage(WindowHandle, EM_SETSEL, SelBeg, SelEnd);
  EndUpdate;
end;

procedure TDropEdit.SetSingleLine(Value: Longbool);
begin
  if Value <> FSingleLine then begin
    FSingleLine := Value;
    RecreateWnd;
  end;
end;

procedure TDropEdit.SetShowObjects(Value: Longbool);
begin
  if FREOCallback <> nil then begin
    FREOCallback.ShowObjects := Value;
    FShowObjects := Value;
  end;
end;

function TDropEdit.GetObjectCount: Integer;
begin
  if FREOCallback <> nil then
    Result := FREOCallback.ObjectCount
  else Result := 0;
end;

procedure TDropEdit.SetTabSize(const Value: Longint);
begin
  if not FSingleLine and (Value <> FTabSize) then begin
    FTabSize := Value;
    FormatText([foTabStops]);
  end;
end;

procedure TDropEdit.SetUrlDetection(const Value: Longbool);
begin
  if FRichEditVersion > 1 then begin
    if Value <> FUrlDetection then begin
      SendMessage(WindowHandle, EM_AUTOURLDETECT, ord(Value = true), 0);
      FUrlDetection := Value;
      ResetText;
    end;
  end
  else FUrlDetection := false;
end;

procedure TDropEdit.SetWordWrap(const Value: Longbool);
var
  hDC: THandle;
  lStyle: Longint;
begin
  BeginUpdate;
  hDC := GetDC(Parent.Handle);
  Screen.Cursor := crHourglass;
  try
    lStyle := GetWindowLong(WindowHandle, GWL_STYLE) and not ES_AUTOHSCROLL;
    SetWindowLong(WindowHandle, GWL_STYLE, lStyle or (ord(not Value) * ES_AUTOHSCROLL));
    SendMessage(WindowHandle, EM_SETTARGETDEVICE, hDC, ord(Value = false) * MAXINT);
  finally
    ReleaseDC(Parent.Handle, hDC);
    FWordWrap := Value;
    EndUpdate;
  end;
end;

function TDropEdit.GetAnsiSelText: String;
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
    Result := WideStringToAnsi(GetSelText)
  else Result := GetSelText;
end;

procedure TDropEdit.SetAnsiSelText(const Value: String);
begin
  InternalHide;
  try
    if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
      SetSelText(AnsiToWideString(Value))
    else SetSelText(Value);
  finally
    InternalShow;
  end;
end;

function TDropEdit.GetAnsiText: String;
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
    Result := WideStringToAnsi(GetText)
  else Result := GetText;
end;

procedure TDropEdit.SetAnsiText(const Value: String);
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
    SetText(AnsiToWideString(Value))
  else SetText(Value);
end;

function TDropEdit.GetRtfText: String;
var
  iSize: Integer;
begin
  SaveToRtfStream(FRtfStream);
  try
    with FRtfStream do begin
      Position := 0;
      iSize := Size;
      SetString(Result, nil, iSize);
      ReadBuffer(Pointer(Result)^, iSize);
      Delete(Result, succ(iSize), MAXINT);
    end;
  finally
    FRtfStream.Clear;
  end;
end;

procedure TDropEdit.SetRtfText(const Value: String);
begin
  FRtfStream.WriteBuffer(Pointer(Value)^, Length(Value));
  LoadFromRtfStream(FRtfStream);
end;

function TDropEdit.GetUtf8SelText: String;
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
    Result := WideStringToUtf8(GetSelText)
  else Result := GetSelText;
end;

procedure TDropEdit.SetUtf8SelText(const Value: String);
begin
  InternalHide;
  try
    if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
      SetSelText(Utf8ToWideString(Value))
    else SetSelText(Value);
  finally
    InternalShow;
  end;
end;

function TDropEdit.GetUtf8Text: String;
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
    Result := WideStringToUtf8(GetText)
  else Result := GetText;
end;

procedure TDropEdit.SetUtf8Text(const Value: String);
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
    SetText(Utf8ToWideString(Value))
  else SetText(Value);
end;

procedure TDropEdit.ResetText;
var
  Sel: TSelection;
begin
  if GetTextLen > 0 then begin
    Sel := GetSelection;
    InternalHide;
    try
      if GetObjectCount > 0 then begin
	SelectAll;
	FClipboardHandler.RestoreData(Self);
	FormatText([foLinks]);
	SetSel(0, 0);
      end
      else ReloadRtfStream;
    finally
      SetSelection(Sel);
      InternalShow;
    end;
  end;
end;

procedure TDropEdit.CMColorChanged(var Message: TMessage);
begin
  inherited;
  EnableThemes(WindowHandle, (Color = clWindow) or (Color = clBtnFace));
end;

procedure TDropEdit.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  SwapDefaultColors;
end;

procedure TDropEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if WindowHandle <> 0 then begin
    SelAttributes.Assign(Font);
    FormatText([foSelFont, foTabStops]);
  end;
end;

procedure TDropEdit.CMEnter(var Message: TMessage);
begin
  if AutoSelect and FSingleLine and not (csLButtonDown in ControlState) then SelectAll;
  DoEnter;
end;

procedure TDropEdit.CNCommand(var Message: TWMCommand);
var
  ptScrollPos: TPoint;
  rClient: TRect;
begin
  inherited;
  if not PlainText and (Message.NotifyCode = EN_UPDATE) then begin
    ptScrollPos := GetScrollPos;
    if FOriginPos.X <> ptScrollPos.X then begin
      rClient := ClientRect;
      FOriginPos := ptScrollPos;
      InvalidateRect(WindowHandle, @rClient, TRUE);
    end;
  end;
end;

procedure TDropEdit.CNNotify(var Message: TWMNotify);
begin
  inherited;
  with Message do begin
    case NMHdr^.code of
      EN_LINK: with PENLink(NMHdr)^ do begin
	case msg of
	  WM_MOUSEMOVE: UrlMouseMove(chrg, MakePoints(lParam));
	  WM_LBUTTONDOWN: UrlClicked(chrg, MK_LBUTTON, MakePoints(lParam));
	  WM_LBUTTONDBLCLK: UrlDblClicked(chrg);
	  WM_RBUTTONDOWN: UrlClicked(chrg, MK_RBUTTON, MakePoints(lParam));
	end;
      end;
    end;
  end;
end;

procedure TDropEdit.WMChar(var Message: TWMChar);
var
  ks: TKeyboardState;
  bShift, bControl: Byte;
begin
  if FReadOnly and (Message.CharCode = VK_SPACE) then with Message do begin
    GetKeyBoardState(ks);
    bShift := ks[VK_SHIFT];
    bControl := ks[VK_CONTROL];
    if (bShift or bControl) < $80 then
      CharCode := VK_NEXT
    else CharCode := VK_PRIOR;
    KeyData := (KeyData and $FF00FFFF) or (MapVirtualKey(CharCode, 0) shl 16);
    ks[VK_SHIFT] := bShift and $7F;
    ks[VK_CONTROL] := bControl and $7F;
    SetKeyboardState(ks);
    Perform(WM_KEYDOWN, CharCode, KeyData);
    Perform(WM_KEYUP, CharCode, DWord(KeyData) or $C0000000);
    ks[VK_SHIFT] := bShift;
    ks[VK_CONTROL] := bControl;
    SetKeyboardState(ks);
  end
  else begin
    if SingleLine and (SelLength = 0) and (GetTextLen = MaxLength) then Beep;
    inherited;
  end;
end;

function TDropEdit.GetSelText: String;
var
  gteUnicode: TGetTextEx;
begin
  if FRichEditV3 then begin
    SetWSString(Result, SelLength + GetLines);
    with gteUnicode do begin
      cb := Length(Result);
      flags := GT_SELECTION or GT_USECRLF;
      codepage := 1200;
      lpDefaultChar := nil;
      lpUsedDefChar := nil;
    end;
    WSDelete(Result, succ(SendMessage(WindowHandle, EM_GETTEXTEX, Longint(@gteUnicode), Longint(Result))), MAXINT);
    if (stAnsiChar in FSearchOptions) then Result := WideStringToAnsi(Result);
  end
  else begin
    SetString(Result, nil, SelLength);
    Delete(Result, succ(SendMessage(WindowHandle, EM_GETSELTEXT, 0, Longint(Result))), MAXINT);
  end;
end;

procedure TDropEdit.SetSelText(const Value: String);
var
  Sel: TSelection;
  iCharCount: Integer;
  ssText: TStringStream;
begin
  Application.ProcessMessages;
  Sel := GetSelection;
  iCharCount := 0;
  try
    if FReplaceAll then begin
      Clear;
      RecreateWnd;
      PlainText := true;
    end;
    iCharCount := Length(Value);
    if FRichEditV3 and IsWideString(Value) then begin
      iCharCount := iCharCount shr 1;
      ReplaceUnicode(Value);
    end
    else if Value <> E then begin
      if FReplaceAll then begin
	ssText := TStringStream.Create(Value);
	if ssText <> nil then begin
	  ReadFromStream(ssText);
	  ssText.Free;
	end;
      end
      else ReplaceAnsi(Value);
    end
    else iCharCount := 0;
  finally
    try
      if iCharCount > 0 then with Sel do begin
	if FReplaceAll then begin
	  SelEnd := 0;
	  SelBeg := 0;
	end
	else begin
	  SelEnd := SelBeg + iCharCount - CountChars(aLF, Value);
	  SelBeg := SelEnd;
	end;
      end;
      if PlainText and (iCharCount > 0) then FormatText([foLinks..foTabStops]);
    finally
      SetSelection(Sel);
      Application.ProcessMessages;
    end;
  end;
end;

function TDropEdit.GetText: String;
var
  gteUnicode: TGetTextEx;
begin
  if FRichEditV3 then begin
    SetWSString(Result, GetTextLen + GetLines);
    with gteUnicode do begin
      cb := Length(Result);
      flags := GT_USECRLF;
      codepage := 1200;
      lpDefaultChar := nil;
      lpUsedDefChar := nil;
    end;
    WSDelete(Result, succ(SendMessage(WindowHandle, EM_GETTEXTEX, Longint(@gteUnicode), Longint(Result))), MAXINT);
    if (stAnsiChar in FSearchOptions) then Result := WideStringToAnsi(Result);
  end
  else begin
    SetString(Result, nil, GetTextLen);
    SendMessage(WindowHandle, WM_GETTEXT, succ(Length(Result)), Longint(Result));
  end;
end;

procedure TDropEdit.SetText(const Value: String);
begin
  InternalHide;
  FReplaceAll := true;
  try
    SetSelText(Value);
  finally
    FReplaceAll := false;
    InternalShow;
  end;
end;

procedure TDropEdit.WMContextMenu(var Message: TWMMouse);
var
  fHandled: Boolean;
  ptPopup: TPoint;
  rClient: TRect;
  paMenu: TPopupAlignment;
begin
  fHandled := false;
  with Message do ptPopup := Point(XPos, YPos);
  if PopupMenu <> nil then with PopupMenu do if AutoPopup then begin
    rClient := ClientRect;
    paMenu := Alignment;
    try
      if not PtinRect(rClient, ptPopup) then with rClient, ptPopup do begin
	X := Right div 2;
	Y := Bottom div 2;
	Alignment := paCenter;
      end
      else Alignment := paLeft;
      with ClientToScreen(ptPopup) do Popup(X, Y);
    finally
      Alignment := paMenu;
    end;
  end
  else if Assigned(OnContextPopup) then OnContextPopup(Self, ptPopup, fHandled);
end;

procedure TDropEdit.WMCreate(var Message: TWMCreate);
begin
  inherited;
  if EnableThemes(WindowHandle, (Color = clWindow) or (Color = clBtnFace)) then OpenTheme;
  SendMessage(WindowHandle, EM_AUTOURLDETECT, ord(FUrlDetection = true), 0);
  if FDropFiles <> nil then with FDropFiles do begin
    AcceptFiles := FAcceptFiles;
    OnFileDropped := FOnFileDropped;
  end;
end;

procedure TDropEdit.WMDestroy(var Message: TWMDestroy);
begin
  CloseTheme;
  inherited;
end;

procedure TDropEdit.WMKeyDown(var Message: TWMKey);
begin
  ResetToolTip;
  if GetKeyState(VK_CONTROL) < 0 then with FClipboardHandler do begin
    case Char(Message.CharCode) of
      'A': SelectAll;
      'C': Copy;
      'V': Paste;
      'X': Cut;
      'Y': SendMessage(WindowHandle, EM_REDO, 0, 0);
      'Z': SendMessage(WindowHandle, EM_UNDO, 0, 0);
    else
      inherited;
    end;
  end
  else begin
    case Message.CharCode of
      VK_RETURN: UrlReturnKeyDown(SelStart);
      VK_APPS: Exit;
    else
      inherited;
    end;
  end;
end;

procedure TDropEdit.WMMouseMove(var Message: TWMMouse);
begin
  inherited;
  if Showing and IsChild(GetForegroundWindow, WindowHandle) then Windows.SetFocus(WindowHandle);
  with Message do if not PtInRect(FLinkToolTipRect, MakePoints(TMessage(Message).lParam)) then ResetToolTip;
end;

procedure TDropEdit.WMLButtonDblClk(var Message: TWMMouse);
begin
  HideSelection;
  try
    inherited;
    if FLinkDblClickPos > -1 then begin
      SendMessage(WindowHandle, EM_SETSEL, FLinkDblClickPos, FLinkDblClickPos);
      FLinkDblClickPos := -1;
    end;
  finally
    ShowSelection;
  end;
end;

procedure TDropEdit.WMRButtonDown(var Message: TWMMouse);
var
  ptMouse: TPoint;
begin
  with Message do ptMouse := Point(XPos, YPos);
  if not IsUrl(SendMessage(WindowHandle, EM_CHARFROMPOS, 0, Longint(@ptMouse))) then
    WMContextMenu(Message)
  else inherited;
end;

procedure TDropEdit.WMNCCalcSize(var Message: TWMNCCalcSize);
var
  hParent, hDC: THandle;
  rClient: TRect;
begin
  inherited;
  hParent := GetParent(WindowHandle);
  if FTheme <> 0 then begin
    hDC := GetDC(hParent);
    try
      with Message do begin
	if GetThemeBackgroundContentRect(FTheme, hDC, EP_EDITTEXT, GetState,
					 CalcSize_Params.rgrc[0], @rClient) = S_OK then begin
	  InflateRect(rClient, -1, -1);
	  with CalcSize_Params.rgrc[0] do begin
	    FNCClientRect.Left := rClient.Left - Left;
	    FNCClientRect.Top := rClient.Top - Top;
	    FNCClientRect.Right := Right - rClient.Right;
	    FNCClientRect.Bottom := Bottom - rClient.Bottom;
	  end;
	  CalcSize_Params.rgrc[0] := rClient;
	  if CalcValidRects then Result := WVR_REDRAW;
	end;
      end;
    finally
      ReleaseDC(hParent, hDC);
    end;
  end;
end;

procedure TDropEdit.WMNCPaint(var Message: TWMNCPaint);
var
  etsClient: EDITTEXTSTATES;
  rBorder, rClient: TRect;
  hDC: THandle;
begin
  inherited;
  if FTheme <> 0 then begin
    etsClient := GetState;
    with rBorder do begin
      Left := 0;
      Top := 0;
      Right := Width;
      Bottom := Height;
    end;
    rClient := rBorder;
    with rClient do begin
      inc(Left, FNCClientRect.Left);
      inc(Top, FNCClientRect.Top);
      dec(Right, FNCClientRect.Right);
      dec(Bottom, FNCClientRect.Bottom);
    end;
    hDC := GetWindowDC(WindowHandle);
    try
      with rClient do ExcludeClipRect(hDC, Left, Top, Right, Bottom);
      if IsThemeBackgroundPartiallyTransparent(FTheme, EP_EDITTEXT, etsClient) then begin
	DrawThemeParentBackground(WindowHandle, hDC, @rBorder);
      end;
      DrawThemeBackground(FTheme, hDC, EP_EDITTEXT, etsClient, rBorder, nil);
      Message.Result := 0;
    finally
      ReleaseDC(WindowHandle, hDC);
    end;
    FillScrollBarGap(etsClient);
  end;
end;

procedure TDropEdit.WMPaint(var Message: TWMPaint);
begin
  if FTheme <> 0 then
    inherited BorderStyle := bsNone
  else inherited BorderStyle := FBorderStyle;
  inherited;
end;

procedure TDropEdit.WMSysKeyDown(var Message: TWMKey);
begin
  if Message.CharCode = VK_BACK then begin
    if GetKeyState(VK_SHIFT) < 0 then
      SendMessage(WindowHandle, EM_REDO, 0, 0)
    else SendMessage(WindowHandle, EM_UNDO, 0, 0);
  end
  else inherited;
end;

procedure TDropEdit.WMThemeChanged(var Message: TMessage);
begin
  CloseTheme;
  inherited;
  OpenTheme;
end;

procedure TDropEdit.WMTimer(var Message: TWMTimer);
var
  ptShow: TPoint;
begin
  if FToolTipHandle <> 0 then with Message do begin
    if Cardinal(TimerID) = FToolTipHandle then begin
      KillTimer(WindowHandle, FToolTipHandle);
      ptShow := GetCursorPos;
      with ptShow do begin
	inc(X, GetSystemMetrics(SM_CXCURSOR) shr 1);
	inc(Y, GetSystemMetrics(SM_CYCURSOR) shr 2);
      end;
      if not PtInRect(FLinkToolTipRect, ScreenToClient(GetCursorPos))
      or not ShowTrackingToolTip(FToolTipHandle, WindowHandle, WindowHandle, E, ptShow, nil) then begin
	ResetToolTip;
      end;
    end;
  end;
end;

procedure TDropEdit.Change;
begin
  if Showing and not (csDesigning in ComponentState) then Windows.SetFocus(WindowHandle);
  inherited;
end;

procedure TDropEdit.CreateParams(var Params: TCreateParams);
var
  wcRichEdit: TWndClass;
begin
  inherited CreateParams(Params);
  with Params do begin
    Style := Style and not ES_MULTILINE or (ES_MULTILINE * DWord(not FSingleLine));
    if not (csDesigning in ComponentState) and (Align = alNone) then begin
      X := 0;
      Y := 0;
      Width := 0;
      Height := 0;
    end;
  end;
  if FRichEditV3 then begin
    if GetClassInfo(GetModuleHandle(LIB_MSFTEDIT), RICHEDIT50W, wcRichEdit) then
      CreateSubClass(Params, RICHEDIT50W)
    else if GetClassInfo(GetModuleHandle(LIB_RICHED20), RICHEDIT20W, wcRichEdit) then CreateSubClass(Params, RICHEDIT20W);
  end;
end;

procedure TDropEdit.CreateWnd;
begin
  inherited;
  ResetParams;
  if csDesigning in ComponentState then inherited Lines.Clear;
end;

procedure TDropEdit.RecreateWnd;
begin
  inherited;
  HandleNeeded;
end;

procedure TDropEdit.ResetParams;
const
  ENM_EVENTS = ENM_CHANGE or ENM_DROPFILES or ENM_KEYEVENTS or ENM_LINK or ENM_MOUSEEVENTS or
	       ENM_REQUESTRESIZE or ENM_SCROLL or ENM_SCROLLEVENTS or ENM_SELCHANGE or ENM_UPDATE;
  SES_STYLES = SES_HIDEGRIDLINES or SES_EXTENDBACKCOLOR;
begin
  SendMessage(WindowHandle, EM_SETEVENTMASK, 0, SendMessage(WindowHandle, EM_GETEVENTMASK, 0, 0) or ENM_EVENTS);
  SendMessage(WindowHandle, EM_SETOLECALLBACK, 0, Longint(FREOCallback));
  SendMessage(WindowHandle, EM_SETEDITSTYLE, SES_STYLES, SES_STYLES);
  SendMessage(WindowHandle, EM_EXLIMITTEXT, 0, FMaxLength);
  SetBorderWidth(FBorderWidth);
  SetReadOnly(FReadOnly);
end;

constructor TDropEdit.Create(AOwner: TComponent);
begin
  inherited;
  FRichEditVersion := GetRichEditVersion;
  FRichEditV3 := (FRichEditVersion >= 3);
  FClipboardHandler := TClipboardHandler.Create;
  if FRichEditV3 then
    FDatedStrings := TDatedWideStrings.Create
  else FDatedStrings := TDatedStrings.Create;
  FDropFiles := TDropFiles.Create(Self);
  FREOCallback := IREOCallback.Create;
  FRtfConverter := TRtfConverter.Create(cfHTML);
  FRtfStream := TMemoryStream.Create;
  FLinkDblClickPos := -1;
  MaxLength := MAXINT;
  FTabSize := 8;
end;

destructor TDropEdit.Destroy;
begin
  FClipboardHandler.Free;
  FDatedStrings.Free;
  FDropFiles.Free;
  FRtfConverter.Free;
  FREOCallback.Free;
  FRtfStream.Free;
  ResetToolTip;
  inherited;
end;

procedure TDropEdit.Copy;
begin
  FClipboardHandler.Execute(Self, caCopy, true);
end;

procedure TDropEdit.Cut;
var
  Sel: TSelection;
begin
  if not FReadOnly then begin
    FClipboardHandler.Execute(Self, caCut, true);
    Sel := GetSelection;
    InternalHide;
    try
      FormatText([foLinks]);
    finally
      SetSelection(Sel);
      InternalShow;
    end;
  end;
end;

procedure TDropEdit.Paste;
var
  Sel: TSelection;
begin
  if not FReadOnly then begin
    FClipboardHandler.Execute(Self, caPaste, true);
    Sel := GetSelection;
    InternalHide;
    try
      FormatText([foLinks]);
    finally
      SetSelection(Sel);
      InternalShow;
    end;
  end;
end;

function TDropEdit.InsertBitmap(Bitmap: TBitmap; CharPos: Integer): Longint;
var
  reoInterface: IRichEditOle;
  boInterface: IBitmapObject;
  ocsInterface: IOleClientSite;
  lbInterface: ILockBytes;
  stgInterface: IStorage;
  ooInterface: IOleObject;
  reoData: TREObject;
  clsidUser: TCLSID;
begin
  BeginUpdate;
  InternalHide;
  ShowObjects := true;
  Result := E_OUTOFMEMORY;
  try
    if SendMessage(WindowHandle, EM_GETOLEINTERFACE, 0, Longint(@reoInterface)) = 0 then Exit;
    boInterface := IBitmapObject.Create;
    ocsInterface := nil;
    lbInterface := nil;
    stgInterface := nil;
    ooInterface := nil;
    try
      if boInterface = nil then Exit;
      Result := reoInterface.GetClientSite(ocsInterface);
      if Result <> S_OK then Exit;
      Result := CreateILockBytesOnHGlobal(0, true, lbInterface);
      if Result <> S_OK then Exit;
      Result := StgCreateDocfileOnILockBytes(lbInterface, STGM_SHARE_EXCLUSIVE or STGM_CREATE or STGM_READWRITE,
					     0, stgInterface);
      if Result <> S_OK then Exit;
      with boInterface do begin
	Result := SetBitmap(Bitmap);
	if Result <> S_OK then Exit;
	Result := GetOleObject(ocsInterface, stgInterface, ooInterface);
      end;
      if Result <> S_OK then Exit;
      Result := OleSetContainedObject(ooInterface, true);
      if Result <> S_OK then Exit;
      Result := ooInterface.GetUserClassID(clsidUser);
      if Result <> S_OK then Exit;
      FillChar(reoData, SizeOf(TREObject), 0);
      with reoData do begin
	cbStruct := SizeOf(TREObject);
	cp := CharPos;
	Clsid := clsidUser;
	OleObj := ooInterface;
	Stg := stgInterface;
	OleSite := ocsInterface;
	dvAspect := DVASPECT_CONTENT;
	dwFlags := REO_BLANK;
	dwUser := DWord(Bitmap);
      end;
      Result := reoInterface.InsertObject(reoData);
    finally
      if ooInterface <> nil then ooInterface.Release;
      if stgInterface <> nil then stgInterface.Release;
      if lbInterface <> nil then lbInterface.Release;
      if ocsInterface <> nil then ocsInterface.Release;
      if boInterface <> nil then boInterface.Release;
      if reoInterface <> nil then reoInterface.Release;
    end;
  finally
    InternalShow;
    EndUpdate;
  end;
  if Result <> S_OK then OleException(Result);
end;

function TDropEdit.FindText(const SearchStr: String; StartPos, Length: Integer): Integer;
var
  Flags: Integer;
  FindW: TFindTextW;
  FindA: TFindTextA;
begin
  Flags := 0;
  if stDown in FSearchOptions then Flags := Flags or FT_DOWN;
  if stWholeWord in FSearchOptions then Flags := Flags or FT_WHOLEWORD;
  if stMatchCase in FSearchOptions then Flags := Flags or FT_MATCHCASE;
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then begin
    with FindW.chrg do begin
      cpMin := StartPos;
      cpMax := cpMin + Length;
    end;
    FindW.lpstrText := Pointer(SearchStr);
    Result := SendMessage(WindowHandle, EM_FINDTEXTW, Flags, Longint(@FindW));
  end
  else begin
    with FindA.chrg do begin
      cpMin := StartPos;
      cpMax := cpMin + Length;
    end;
    FindA.lpstrText := PAnsiChar(SearchStr);
    Result := SendMessage(WindowHandle, EM_FINDTEXT, Flags, Longint(@FindA));
  end;
end;

function TDropEdit.SearchText: Longbool;
var
  cDefault: TCursor;
  sSelected: String;
  iPos, iLen, iBegPos, iEndPos, iOldPos, iNewPos: Integer;
begin
  Result := true;
  if not FSearching then begin
    FSearching := true;
    cDefault := Screen.Cursor;
    try
      sSelected := GetSelText;
      if FRichEditV3 then begin
	iPos := WCPos(aCR, sSelected);
	if iPos > 0 then WSDelete(sSelected, succ(iPos), MAXINT);
      end
      else begin
	iPos := Pos(aCR, sSelected);
	if iPos > 0 then Delete(sSelected, succ(iPos), MAXINT);
      end;
      if FReSearch and (sSelected <> E) then FFindText := sSelected;
      if Assigned(FGetSearchText) and (not FReSearch or (FFindText = E)) then begin
        FFindText := FGetSearchText(sSelected, Font.Name, FSearchOptions);
      end;
      FReSearch := false;
      if FFindText <> E then begin
	Screen.Cursor := crHourGlass;
	Application.ProcessMessages;
	iLen := GetTextLen;
	SendMessage(WindowHandle, EM_GETSEL, Longint(@iBegPos), Longint(@iEndPos));
	if stDown in FSearchOptions then begin
	  iOldPos := iBegPos + (iEndPos - iBegPos);
	  if iOldPos >= iLen then iOldPos := 0;
	end
	else begin
	  iOldPos := iBegPos;
	  if iOldPos = 0 then begin
	    Result := false;
	    SetSel(0, 0);
	    Beep;
	  end;
	end;
	if (iOldPos > -1) and (iOldPos < iLen) then begin
	  iNewPos := FindText(FFindText, iOldPos, iLen);
	  if iNewPos > -1 then with inherited Lines do begin
	    if FRichEditV3 then
	      SetSel(iNewPos, (Length(FFindText) - Length(wNUL)) shr 1)
	    else SetSel(iNewPos, Length(FFindText));
	  end
	  else begin
	    if stDown in FSearchOptions then begin
	      FReSearch := true;
	      Exit;
	    end
	    else begin
	      SetSel(iBegPos, 0);
	      Result := false;
	      Beep;
	    end;
	  end;
	  Result := (SelLength > 0);
	end;
      end;
    finally
      Screen.Cursor := cDefault;
      FSearching := false;
    end;
  end
  else FReSearch := false;
end;

function TDropEdit.GetLine(Index: Longint): TLineArray;
begin
  PWideChar(@Result)^ := WideChar(MAX_LINE_LEN);
  SendMessage(WindowHandle, EM_GETLINE, Index, Longint(@Result));
end;

function TDropEdit.GetLineStartIndex(Line: Longint): Longint;
begin
  Result := SendMessage(WindowHandle, EM_LINEINDEX, Line, 0);
end;

function TDropEdit.IsLineBlank(const aLine: TLineArray): Longbool;
var
  iIndex: Integer;
begin
  Result := true;
  iIndex := Low(aLine);
  while aLine[iIndex] > aNUL do begin
    if aLine[iIndex] > SP then begin
      Result := false;
      Break;
    end
    else inc(iIndex);
  end;
end;

function TDropEdit.WidthAdjustText(Device: HDC; const Text: String; MaxWidth: Integer): String;
const
  wSP = #$2000;
  wDOTS = #$2026;
  wNULL = #$0000;
  aSP = #$20;
  aDOT = #$2E;
  aNUL = #$00;
var
  iResult: Integer;
  szText: TSize;
  pw: PWideChar;
  pa: PAnsiChar;
begin
  Result := Text;
  if IsWideString(Result) then begin
    iResult := WSLength(Result);
    GetTextExtentPoint32W(Device, Pointer(Result), iResult, szText);
    if szText.cx > MaxWidth then begin
      iResult := MulDiv(iResult, MaxWidth, szText.cx);
      WSDelete(Result, succ(iResult), MAXINT);
      pw := Pointer(Result);
      dec(iResult, 3);
      while (iResult > 0) and (pw[iResult] <= SP) do dec(iResult);
      if iResult >= 0 then begin
	inc(pw, iResult);
	pw^ := wSP;
	inc(pw);
	pw^ := wDOTS;
      end;
      inc(iResult, 2);
      WSDelete(Result, succ(iResult), MAXINT);
    end;
  end
  else begin
    iResult := Length(Result);
    GetTextExtentPoint32(Device, Pointer(Result), iResult, szText);
    if szText.cx > MaxWidth then begin
      iResult := MulDiv(iResult, MaxWidth, szText.cx);
      Delete(Result, succ(iResult), MAXINT);
      pa := Pointer(Result);
      dec(iResult, 4);
      while (iResult > 0) and (pa[iResult] <= SP) do dec(iResult);
      if iResult >= 0 then begin
	inc(pa, iResult);
	pa^ := aSP;
	inc(pa);
	pa^ := aDOT;
	inc(pa);
	pa^ := aDOT;
	inc(pa);
	pa^ := aDOT;
      end;
      inc(iResult, 4);
      Delete(Result, succ(iResult), MAXINT);
    end;
  end;
end;

function TDropEdit.GetTextLen: Integer;
var
  gtle: TGetTextLengthEx;
begin
  if FRichEditVersion > 1 then begin
    with gtle do begin
      flags := GTL_DEFAULT;
      codepage := 1200;
    end;
    Result := SendMessage(WindowHandle, EM_GETTEXTLENGTHEX, Longint(@gtle), 0);
  end
  else Result := SendMessage(WindowHandle, WM_GETTEXTLENGTH, 0, 0);
end;

procedure TDropEdit.SetSel(SelStart, SelLength: Integer);
begin
  BeginUpdate;
  try
    SendMessage(WindowHandle, EM_SETSEL, SelStart, SelStart + SelLength);
    SendMessage(WindowHandle, EM_SCROLLCARET, 0, 0);
  finally
    EndUpdate;
  end;
end;

procedure TDropEdit.SelectAll;
begin
  SendMessage(WindowHandle, EM_SETSEL, 0, GetTextLen);
end;

procedure TDropEdit.DeleteSel;
var
  iPos: Integer;
begin
  SendMessage(WindowHandle, EM_GETSEL, 0, Longint(@iPos));
  if iPos <> 0 then SendMessage(WindowHandle, EM_REPLACESEL, 1, 0);
end;

procedure TDropEdit.ResetSel;
var
  iPos: Integer;
begin
  SendMessage(WindowHandle, EM_GETSEL, 0, Longint(@iPos));
  if iPos <> 0 then SendMessage(WindowHandle, EM_SETSEL, iPos, iPos);
end;

procedure TDropEdit.BeginUpdate;
begin
  if Screen.Cursor <> crHourglass then FCursor := Screen.Cursor;
  inherited Lines.BeginUpdate;
end;

procedure TDropEdit.EndUpdate;
begin
  inherited Lines.EndUpdate;
  if Screen.Cursor = crHourglass then Screen.Cursor := FCursor;
end;

procedure TDropEdit.ReadFromStream(Stream: TStream);
var
  iSize: Integer;
  sInput, sText: String;
  esRtf: TEditStream;
begin
  if Stream <> nil then with Stream do begin
    Position := 0;
    iSize := Size;
    if iSize > Length(RTF_MARKER) then begin
      SetString(sInput, nil, Length(RTF_MARKER));
      ReadBuffer(Pointer(sInput)^, Length(RTF_MARKER));
      Position := 0;
      if LoStrAtPos(RTF_MARKER, sInput, 1) then begin
	try
	  SetWordWrap(true);
	  PlainText := false;
	  esRtf := GetStreamInData;
	  if Stream <> FRtfStream then FRtfStream.LoadFromStream(Stream);
	  SendMessage(WindowHandle, EM_STREAMIN, SF_RTF, Longint(@esRtf));
	finally
	  FRtfStream.Clear;
	end;
	Exit;
      end;
    end;
    SetString(sInput, nil, iSize);
    ReadBuffer(Pointer(sInput)^, iSize);
    if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
      sText := Utf8ToWideString(sInput)
    else sText := Utf8StringToAnsi(sInput);
  end
  else sText := E;
  if IsWideString(sText) then
    ReplaceUnicode(sText)
  else ReplaceAnsi(sText);
end;

procedure TDropEdit.WriteToStream(Stream: TStream);
var
  sOutput: String;
begin
  if Stream <> nil then with Stream do begin
    sOutput := WideStringToUtf8(GetText);
    WriteBuffer(Pointer(sOutput)^, Length(sOutput));
  end;
end;

procedure TDropEdit.PrintWithMargins(var TPD: TTextPrinterData);
type
  TCustomData = Record
    Text: String;
    Bounds: TRect;
    case NumericData: Integer of
      0: (MaxWidth: Integer);
      1: (
	PageData: Record
	  Count: Integer;
	  Index: Integer;
	end
      );
  end;

  function InvisibleDropEdit(const TextPrinterData: TTextPrinterData): TDropEdit;
  var
    aLine: TLineArray;
    iPos, iIndex: Integer;
  begin
    Result := TDropEdit.Create(nil);
    if Assigned(Result) then with Result do begin
      Parent := Self.Parent;
      BorderWidth := 0;
      MaxLength := Self.MaxLength;
      PlainText := Self.PlainText;
      ShowObjects := Self.ShowObjects;
      with TextPrinterData.PD, Font do begin
	Name := PrinterFontName;
	Size := PrinterFontSize;
      end;
      TabSize := Self.TabSize;
      SetWordWrap(false);
      if Assigned(Self.FOnBeforeSetPrintText) then Self.FOnBeforeSetPrintText(Result, TextPrinterData.PD.PrinterPrefs);
      if TPD.PD.PrintRange = prSelection then begin
	ClipboardHandler.AssignData(Self.FClipboardHandler.Execute(Self, caStore, false));
	Self.FClipboardHandler.AssignData(ClipboardHandler.Execute(Result, caRetrieve, true));
      end
      else begin
	SendMessage(WindowHandle, EM_SETSEL, 0, -1);
	SetSelText(Self.RtfText);
      end;
      FormatText([foTabStops]);
      if Assigned(Self.FOnAfterSetPrintText) then Self.FOnAfterSetPrintText(Result, TextPrinterData.PD.PrinterPrefs);
      iPos := -1;
      iIndex := pred(SendMessage(WindowHandle, EM_GETLINECOUNT, 0, 0));
      while iIndex >= 0 do begin
	PWideChar(@aLine)^ := WideChar(MAX_LINE_LEN);
	if (SendMessage(WindowHandle, EM_GETLINE, iIndex, Longint(@aLine)) > 0) and not IsLineBlank(aLine) then begin
	  if iPos > 0 then begin
	    SendMessage(WindowHandle, EM_SETSEL, pred(iPos), -1);
	    SendMessage(WindowHandle, EM_REPLACESEL, 0, 0);
	  end;
	  Break;
	end
	else begin
	  iPos := SendMessage(WindowHandle, EM_LINEINDEX, iIndex, 0);
	  dec(iIndex);
	end;
      end;
    end;
  end;

  function GetDimensions(LineHeight: Integer; Margins: TRect; var rText, rHeader, rFooter: TRect): Integer;
  var
    iLogX, iLogY, iWidth, iHeight: Integer;
  begin
    with Printer do begin
      iLogX := GetDeviceCaps(Handle, LOGPIXELSX);
      iLogY := GetDeviceCaps(Handle, LOGPIXELSY);
      iWidth := GetDeviceCaps(Handle, HORZSIZE);
      iHeight := GetDeviceCaps(Handle, VERTSIZE);
      LineHeight := MulDiv(LineHeight, PageHeight, Screen.Height);
      with Margins do begin
	Left := MulDiv(PageWidth, Left, iWidth);
	Right := PageWidth - MulDiv(PageWidth, Right, iWidth);
	Top := MulDiv(PageHeight, Top, iHeight);
	Bottom := PageHeight - MulDiv(PageHeight, Bottom, iHeight);
	Result := Right - Left;
	rText.Left := MulDiv(Left, TWIPS, iLogX);
	rText.Right := MulDiv(Right, TWIPS, iLogX);
	rText.Top := MulDiv(Top + LineHeight, TWIPS, iLogY);
	rText.Bottom := MulDiv(Bottom - LineHeight, TWIPS, iLogY);
	rHeader.Left := Left;
	rHeader.Right := Right;
	rHeader.Top := Top - LineHeight;
	rHeader.Bottom := Top;
	rFooter.Left := Left;
	rFooter.Right := Right;
	rFooter.Top := Bottom;
	rFooter.Bottom := Bottom + LineHeight;
      end;
    end;
  end;

  function GetPageCount(DropEdit: TDropEdit; FormatRange: TFormatRange): Integer;
  var
    iPrinted: Integer;
  begin
    Result := 1;
    with FormatRange.chrg do begin
      try
	iPrinted := 0;
	repeat
	  cpMin := iPrinted;
	  iPrinted := SendMessage(DropEdit.Handle, EM_FORMATRANGE, 0, Longint(@FormatRange));
	  if (iPrinted <= cpMin) or (iPrinted >= cpMax) then Break;
	  inc(Result);
	until iPrinted = cpMax;
      finally
	SendMessage(DropEdit.Handle, EM_FORMATRANGE, 0, 0);
      end;
    end;
  end;

  procedure PrintCopy(DropEdit: TDropEdit; FormatRange: TFormatRange;
		      var HeaderData, FooterData: TCustomData;
		      HeaderStyle: TFontStyles);
  type
    TLinePoints = Array[0..1] of TPoint;

    procedure PrintHeader(Device: HDC; var HeaderData: TCustomData);
    var
      iAlign: Integer;
      aLinePoints: TLinePoints;
    begin
      iAlign := SetTextAlign(Device, TA_LEFT or TA_TOP);
      try
	with HeaderData, Bounds do begin
	  aLinePoints[0] := Point(Left, Bottom);
	  aLinePoints[1] := Point(Right, Bottom);
	  if not IsWideString(Text) then Text := WidthAdjustText(Device, NoBomUtf8ToWideString(Text), MaxWidth);
	  TextOutW(Device, Left, Top, Pointer(Text), WSLength(Text));
	  PolyLine(Device, aLinePoints, succ(High(aLinePoints) - Low(aLinePoints)));
	end;
      finally
	SetTextAlign(Device, iAlign);
      end;
    end;

    procedure PrintFooter(Device: HDC; var FooterData: TCustomData);
    var
      iAlign: Integer;
      sFooterText: String;
      aLinePoints: TLinePoints;
    begin
      with FooterData, Bounds do begin
	iAlign := SetTextAlign(Device, TA_RIGHT or TA_BOTTOM);
	try
	  if not IsWideString(Text) then Text := NoBomUtf8ToWideString(ASFormat(Text, ['%d', PageData.Count]));
	  sFooterText := WSFormat(Text, [PageData.Index]);
	  aLinePoints[0] := Point(Left, Top);
	  aLinePoints[1] := Point(Right, Top);
	  PolyLine(Device, aLinePoints, succ(High(aLinePoints) - Low(aLinePoints)));
	  TextOutW(Device, Right, Bottom, Pointer(sFooterText), WSLength(sFooterText));
	finally
	  SetTextAlign(Device, iAlign);
	end;
      end;
    end;

  var
    iPrinted: Integer;
    hFont: THandle;
  begin
    iPrinted := 0;
    try
      with FormatRange do begin
	FooterData.PageData.Index := 1;
	repeat
	  chrg.cpMin := iPrinted;
	  iPrinted := SendMessage(DropEdit.Handle, EM_FORMATRANGE, 1, Longint(@FormatRange));
	  hFont := GetCurrentObject(hdc, OBJ_FONT);
	  try
	    if HeaderData.Text <> E then begin
	      with Printer.Canvas.Font do begin
		Style := HeaderStyle;
		SelectObject(hdc, Handle);
	      end;
	      PrintHeader(hdc, HeaderData);
	    end;
	    with FooterData do if (Text <> E) and (PageData.Count > 1) then begin
	      with Printer.Canvas.Font do begin
		Style := DropEdit.Font.Style;
		SelectObject(hdc, Handle);
	      end;
	      PrintFooter(hdc, FooterData);
	    end;
	  finally
	    SelectObject(hdc, hFont);
	  end;
	  if (iPrinted <= chrg.cpMin) or (iPrinted >= chrg.cpMax) then Break;
	  inc(FooterData.PageData.Index);
	  Printer.NewPage;
	until iPrinted = chrg.cpMax;
      end;
    finally
      SendMessage(DropEdit.Handle, EM_FORMATRANGE, 0, 0);
    end;
  end;

var
  iOldMapMode: Integer;
  frPrint: TFormatRange;
  dePrint: TDropEdit;
  cdHeader, cdFooter: TCustomData;
begin
  if GetTextLen > 0 then begin
    FillChar(frPrint, SizeOf(TFormatRange), 0);
    with frPrint.chrg do begin
      SendMessage(WindowHandle, EM_GETSEL, Longint(@cpMin), Longint(@cpMax));
      if cpMax > cpMin then TPD.PD.PrintRange := prSelection;
    end;
    with TPD.PD, Self.Font do begin
      if PrinterFontName = E then StrPLCopy(PrinterFontName, Name, pred(SizeOf(TPrinterName)));
      if PrinterFontSize = 0 then PrinterFontSize := Size;
    end;
    if ExecutePrintDialog(TPD.PD) then begin
      dePrint := InvisibleDropEdit(TPD);
      if Assigned(dePrint) then begin
	try
	  with Printer, frPrint do begin
	    Title := TPD.PD.PrintJobName;
	    BeginDoc;
	    hdc := Handle;
	    hdcTarget := hdc;
	    Canvas.Font.Assign(dePrint.Font);
	    iOldMapMode := SetMapMode(hdc, MM_TEXT);
	    try
	      with TPD do begin
		if moPrintHeader in PD.MarginOptions then cdHeader.Text := Utf8HeaderText;
		if moPrintFooter in PD.MarginOptions then cdFooter.Text := Utf8FooterText;
	      end;
	      cdHeader.MaxWidth := GetDimensions(dePrint.GetLineHeight(-1), TPD.PD.Margins,
						 rcPage, cdHeader.Bounds, cdFooter.Bounds);
	      rc := rcPage;
	      with chrg do begin
		cpMin := 0;
		cpMax := dePrint.GetTextLen;
	      end;
	      cdFooter.PageData.Count := GetPageCount(dePrint, frPrint);
	      PrintCopy(dePrint, frPrint, cdHeader, cdFooter, TPD.PD.MarginFontStyles);
	    finally
	      SetMapMode(hdc, iOldMapMode);
	      EndDoc;
	    end;
	  end;
	finally
	  with dePrint do begin
	    Parent := nil;
	    Free;
	  end;
	end;
      end;
    end;
  end;
end;

procedure TDropEdit.LoadFromFile(const FileName: String);
var
  fsInput: TFileStream;
begin
  fsInput := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
  if fsInput <> nil then begin
    try
      ReadFromStream(fsInput);
    finally
      fsInput.Free;
    end;
  end
  else begin
    {$IFDEF CONDITIONALEXPRESSIONS}
    Raise EOutOfMemory.Create(SOutOfResources);
    {$ELSE}
    Raise EOutOfMemory.CreateRes(SOutOfResources);
    {$ENDIF}
  end;
end;

procedure TDropEdit.LoadFromRtfFile(const FileName: String);
var
  esRtf: TEditStream;
begin
  try
    SetWordWrap(true);
    PlainText := false;
    esRtf := GetStreamInData;
    FRtfStream.LoadFromFile(FileName);
    SendMessage(WindowHandle, EM_STREAMIN, SF_RTF, Longint(@esRtf));
  finally
    FRtfStream.Clear;
  end;
end;

procedure TDropEdit.LoadFromUtf8File(const FileName: String);
var
  lError: Longint;
  tdInterface: ITextDocument;
  vdFileName: TVarData;
begin
  lError := E_NOINTERFACE;
  tdInterface := GetTextDocumentInterface(Self);
  if tdInterface <> nil then begin
    try
      PlainText := true;
      with vdFileName do begin
	VType := varOleStr;
	VOleStr := StringToOleStr(FileName);
      end;
      lError := tdInterface.Open(@vdFileName, tomOpenExisting or tomText, CP_UTF8);
    finally
      tdInterface.Release;
    end;
  end;
  if lError <> S_OK then Raise EReadError.Create(SysErrorMessage(lError));
end;

procedure TDropEdit.LoadFromHtmlFile(const FileName: String);
var
  esRtf: TEditStream;
  sError: String;
begin
  with FRtfConverter do begin
    ParentHandle := WindowHandle;
    try
      if FileToRtfStream(FileName, FRtfStream) then begin
	SetWordWrap(true);
	PlainText := false;
	SetUrlDetection(true);
	esRtf := GetStreamInData;
	FRtfStream.Position := 0;
	SendMessage(WindowHandle, EM_STREAMIN, SF_RTF, Longint(@esRtf));
      end
      else begin
	with FRtfStream do begin
	  SetString(sError, nil, Size);
	  Read(Pointer(sError), Length(sError))
	end;
	Raise EReadError.Create(sError);
      end;
    finally
      FRtfStream.Clear;
    end;
  end;
end;

procedure TDropEdit.LoadFromHtmlString(const HtmlString: String);
var
  esRtf: TEditStream;
  sError: String;
begin
  with FRtfConverter do begin
    ParentHandle := WindowHandle;
    try
      if StringToRtfStream(HtmlString, FRtfStream) then begin
	SetWordWrap(true);
	PlainText := false;
	SetUrlDetection(true);
	esRtf := GetStreamInData;
	FRtfStream.Position := 0;
	SendMessage(WindowHandle, EM_STREAMIN, SF_RTF, Longint(@esRtf));
      end
      else begin
	with FRtfStream do begin
	  SetString(sError, nil, Size);
	  Read(Pointer(sError), Length(sError))
	end;
	Raise EReadError.Create(sError);
      end;
    finally
      FRtfStream.Clear;
    end;
  end;
end;

procedure TDropEdit.LoadFromStream(Stream: TStream);
begin
  ReadFromStream(Stream);
end;

procedure TDropEdit.LoadFromRtfStream(Stream: TStream);
begin
  ReadFromStream(Stream);
end;

procedure TDropEdit.SaveToFile(const FileName: String);
var
  fsInput: TFileStream;
begin
  if FRichEditV3 and not (stAnsiChar in FSearchOptions) then begin
    fsInput := TFileStream.Create(FileName, fmCreate);
    if fsInput <> nil then begin
      try
	WriteToStream(fsInput);
      finally
	fsInput.Free;
      end;
    end
    else begin
      {$IFDEF CONDITIONALEXPRESSIONS}
      Raise EOutOfMemory.Create(SOutOfResources);
      {$ELSE}
      Raise EOutOfMemory.CreateRes(SOutOfResources);
      {$ENDIF}
    end;
  end
  else inherited Lines.SaveToFile(FileName);
end;

procedure TDropEdit.SaveToAnsiFile(const FileName: String);
begin
  inherited Lines.SaveToFile(FileName);
end;

procedure TDropEdit.SaveToRtfFile(const FileName: String);
var
  esRtf: TEditStream;
begin
  esRtf := GetStreamOutData;
  if SendMessage(WindowHandle, EM_STREAMOUT, SF_RTF, Longint(@esRtf)) <> 0 then begin
    try
      FRtfStream.SaveToFile(FileName);
    finally
      FRtfStream.Clear;
    end;
  end;
end;

procedure TDropEdit.SaveToUtf8File(const FileName: String);
var
  lError: Longint;
  tdInterface: ITextDocument;
  vdFileName: TVarData;
begin
  lError := E_NOINTERFACE;
  tdInterface := GetTextDocumentInterface(Self);
  if tdInterface <> nil then begin
    try
      with vdFileName do begin
	VType := varOleStr;
	VOleStr := StringToOleStr(FileName);
      end;
      lError := tdInterface.Save(@vdFileName, tomCreateAlways or tomText, CP_UTF8);
    finally
      tdInterface.Release;
    end;
  end;
  if lError <> S_OK then Raise EWriteError.Create(SysErrorMessage(lError));
end;

procedure TDropEdit.SaveToHtmlFile(const FileName: String);
var
  esRtf: TEditStream;
  sError: String;
begin
  try
    esRtf := GetStreamOutData;
    if SendMessage(WindowHandle, EM_STREAMOUT, SF_RTF, Longint(@esRtf)) <> 0 then begin
      with FRtfConverter do begin
	sError := FileName;
	ParentHandle := WindowHandle;
	FRtfStream.Position := 0;
	if not RtfStreamToFile(FRtfStream, sError) then Raise EReadError.Create(sError);
      end;
    end;
  finally
    FRtfStream.Clear;
  end;
end;

procedure TDropEdit.SaveToHtmlString(const HtmlString: String);
var
  esRtf: TEditStream;
  sError: String;
begin
  try
    esRtf := GetStreamOutData;
    if SendMessage(WindowHandle, EM_STREAMOUT, SF_RTF, Longint(@esRtf)) <> 0 then begin
      with FRtfConverter do begin
	sError := HtmlString;
	ParentHandle := WindowHandle;
	FRtfStream.Position := 0;
	if not RtfStreamToFile(FRtfStream, sError) then Raise EReadError.Create(sError);
      end;
    end;
  finally
    FRtfStream.Clear;
  end;
end;

procedure TDropEdit.SaveToStream(Stream: TStream);
begin
  if Stream <> nil then with Stream do begin
    if FRichEditV3 and not (stAnsiChar in FSearchOptions) then
      WriteToStream(Stream)
    else inherited Lines.SaveToStream(Stream);
  end;
end;

procedure TDropEdit.SaveToAnsiStream(Stream: TStream);
begin
  inherited Lines.SaveToStream(Stream);
end;

procedure TDropEdit.SaveToRtfStream(Stream: TStream);
var
  esRtf: TEditStream;
begin
  esRtf := GetStreamOutData;
  if (SendMessage(WindowHandle, EM_STREAMOUT, SF_RTF, Longint(@esRtf)) <> 0) and (Stream <> FRtfStream) then begin
    try
      FRtfStream.SaveToStream(Stream);
    finally
      FRtfStream.Clear;
    end;
  end;
end;

procedure TDropEdit.ReloadRtfStream;
var
  esRtf: TEditStream;
begin
  esRtf := GetStreamOutData;
  if SendMessage(WindowHandle, EM_STREAMOUT, SF_RTF, Longint(@esRtf)) <> 0 then begin
    try
      esRtf := GetStreamInData;
      FRtfStream.Position := 0;
      SendMessage(WindowHandle, EM_STREAMIN, SF_RTF, Longint(@esRtf));
    finally
      FRtfStream.Clear;
    end;
  end;
end;

procedure TDropEdit.ShowDatedStrings;
begin
  if FDatedStrings <> nil then SetText(FDatedStrings.Text);
end;

procedure Register;
begin
  RegisterComponents('New', [TDropEdit]);
end;

end.

