{$J+,Z4}
unit DropGrid;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, CommCtrl, SysUtils, Messages, Classes, Controls, ComCtrls, ExtCtrls,
  Forms, Graphics, Grids, DropFiles, ToolTip, ImgListEx, UxThemes;

// TArrowImages ----------------------------------------------------------------
type
  TArrowImages = class(TBlendImages)
  published
    procedure DrawBlended(DestCanvas: TCanvas; const DestSizeRect: TRect; ImageIndex: Longint);
  end;

// THeaderImages ---------------------------------------------------------------
type
  THeaderImages = class(TBlendImages)
  private
    FMultiple: Longint;
  published
    procedure DrawHighlighted(DestCanvas: TCanvas; const DestSizeRect: TRect; ImageIndex: Longint);
    property  Multiple: Longint read FMultiple write FMultiple;
  end;

// TSortList -------------------------------------------------------------------
type
  PLengthList = ^TLengthList;
  TLengthList = Array[0..0] of Integer;
  TSortList = class(TStringList)
  private
    FInvert: Longbool;
  public
    procedure AlphaSort(FirstItem, LastItem: Longint);
    property  Invert: Longbool read FInvert write FInvert;
  end;

// TDropGrid -------------------------------------------------------------------
type
  RGB = (R, G, B, X);
  TRGB = Array[RGB] of Byte;

  TSearchType = (stDown, stWholeWord, stMatchCase, stAnsiChar, stShowOptions);
  TSearchTypes = Set of TSearchType;

  TSpaceBarMoves = (sbmNone, sbmHorizontal, sbmVertical);

  TGetSearchText = function(const Find, FontName: String; var How: TSearchTypes): String;
  TOnGetToolTipText = procedure(ACol, ARow: Longint; Fixed: Longbool; var Text: String) of Object;
  TOnMouseActivate = procedure(MouseMsg, ACol, ARow: Longint) of Object;

  TDropGrid = class(TStringGrid)
  private
    FAcceptFiles: Longbool;
    FAllowHeaderClick: Longbool;
    FArrowImages: TArrowImages;
    FAutoAdjustColWidths: Longbool;
    FBoldHeaderText: Longbool;
    FCellTextLengths: PLengthList;
    FContextKeyPressed: Longbool;
    FCurrentCell: TGridCoord;
    FDropFiles: TDropFiles;
    FFastSearchChars: String;
    FFastSearchTimer: TTimer;
    FFindText: String;
    FHeaderClicked: Longint;
    FHeaderImages: THeaderImages;
    FHeaderTouched: Longint;
    FHintHidePause: Longint;
    FObjectColumn: Longint;
    FPrevRowWidth: Longint;
    FReSearch: Longbool;
    FSearchColumn: Longint;
    FSearching: Longbool;
    FSortedColumn: Longint;
    FSortStartColumn: Longint;
    FSpaceBarMoves: TSpaceBarMoves;
    FTheme: THandle;
    FToolTip: TToolTip;
    FToolTipCell: TGridCoord;
    FOnFileDropped: TNotifyEvent;
    FOnGetToolTipText: TOnGetToolTipText;
    FOnMouseActivate: TOnMouseActivate;
    procedure InvalidateRect(R: TRect; MouseMove: Longbool);
    procedure InvalidateHeaders(OldCol, NewCol: Longint; MouseMove: Longbool);
    procedure OpenTheme;
    procedure CloseTheme;
    function  GetRowState: HEADERITEMSTATES;
    function  GetColState(ACol: Longint): HEADERITEMSTATES;
    function  GetBottomRow: Longint;
    function  GetRightCol: Longint;
    procedure SetAcceptFiles(Value: Longbool);
    function  GetCell(ACol, ARow: Longint): String;
    procedure SetCell(ACol, ARow: Longint; const Value: String);
    function  GetIndex(ACol, ARow: Longint): Integer;
    procedure SetIndex(ACol, ARow: Longint; const Value: Integer);
    function  GetCol(Index: Longint): TStrings;
    procedure SetCol(Index: Longint; const Value: TStrings);
    function  GetRow(Index: Longint): TStrings;
    procedure SetRow(Index: Longint; const Value: TStrings);
    function  GetColTextAlignment(Index: Longint): TAlignment;
    procedure SetColTextAlignment(Index: Longint; Value: TAlignment);
    function  GetHeaderAlignment(Index: Longint): TAlignment;
    procedure SetHeaderAlignment(Index: Longint; Value: TAlignment);
    function  GetImageIndex(Index: Longint): Smallint;
    procedure SetImageIndex(Index: Longint; Value: Smallint);
    procedure DrawArrowImage(ACol: Longint; ASizeRect: TRect);
    procedure FastSearch;
    procedure OnHideToolTip(Sender: TObject);
    procedure OnShowToolTip(Sender: TObject);
    procedure ScrollIntoView(ARow: Longint);
    procedure SearchTime(Sender: TObject);
    procedure SetToolTipPos(var ptTopLeft: TPoint);
    function  SortColor(ACol: Longint; clDefault: TColor): TColor;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMParentShowHintChanged(var Message: TMessage); message CM_PARENTSHOWHINTCHANGED;
    procedure CMShowHintChanged(var Message: TMessage); message CM_SHOWHINTCHANGED;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMContextMenu(var Message: TWMMouse); message WM_CONTEXTMENU;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseActivate(var Message: TWMMouseActivate); message WM_MOUSEACTIVATE;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMSysKeyDown(var Message: TWMSysKeyDown); message WM_SYSKEYDOWN;
    procedure WMThemeChanged(var Message: TMessage); message WM_THEMECHANGED;
    property  ImageIndices[Index: Longint]: Smallint read GetImageIndex write SetImageIndex;
  protected
    procedure ColumnMoved(FromIndex, ToIndex: Longint); override;
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    procedure RowMoved(FromIndex, ToIndex: Longint); override;
  public
    GetSearchText: TGetSearchText;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AdjustColWidths(ACol, ARow: Longint);
    function  Cell(X, Y: Longint): TGridCoord;
    {$IFDEF CONDITIONALEXPRESSIONS}
    procedure DeleteRow(ARow: Longint); override;
    {$ELSE}
    procedure DeleteRow(ARow: Longint);
    {$ENDIF}
    procedure DrawCellBackground(ACol: Longint; ARect: TRect; AState: TGridDrawState; ADefColor: TColor);
    function  DrawCellImage(ACol: Longint; ARect: TRect; AState: TGridDrawState): TRect;
    function  DrawCellText(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState): TRect;
    procedure DrawGridFocus(ARect: TRect);
    procedure DrawHeaderCellContents(ACol: Longint; ARect: TRect);
    procedure DrawHeaderCell(ACol: Longint; ARect: TRect);
    function  DrawHeaderText(ACol: Longint; ARect: TRect): TRect;
    function  IsSameCell(Pt1, Pt2: TGridCoord): Longbool;
    procedure LoadArrowImages(ArrowsID: Integer);
    procedure SetColumnData(Column: Longint; HeaderAlignment, ColumnAlignment: TAlignment;
			    const Text: String; BitmapID: Integer);
    function  SearchGrid: Longbool;
    procedure SetImageProperties(Count, AWidth, AHeight: Integer);
    procedure ShowToolTip(SelectedCell: TGridCoord; Fixed: Longbool);
    function  SortRows(SortCol: Longint; GetList, Invisible: Longbool): String;
    function  VisibleRowWidth: Longint;
    property  BottomRow: Longint read GetBottomRow;
    property  RightCol: Longint read GetRightCol;
    property  Cells[ACol, ARow: Longint]: String read GetCell write SetCell;
    property  Cols[Index: Longint]: TStrings read GetCol write SetCol;
    property  Rows[Index: Longint]: TStrings read GetRow write SetRow;
    property  Indices[ACol, ARow: Longint]: Integer read GetIndex write SetIndex;
    property  ColTextAlignments[Index: Longint]: TAlignment read GetColTextAlignment write SetColTextAlignment;
    property  HeaderAlignments[Index: Longint]: TAlignment read GetHeaderAlignment write SetHeaderAlignment;
    property  DropFiles: TDropFiles read FDropFiles;
    property  ObjectColumn: Longint read FObjectColumn write FObjectColumn;
    property  ReSearch: Longbool read FReSearch write FReSearch;
    property  Searching: Longbool read FSearching write FSearching;
    property  ToolTip: TToolTip read FToolTip;
  published
    property  AllowHeaderClick: Longbool read FAllowHeaderClick write FAllowHeaderClick;
    property  AcceptFiles: Longbool read FAcceptFiles write SetAcceptFiles;
    property  AutoAdjustColWidths: Longbool read FAutoAdjustColWidths write FAutoAdjustColWidths;
    property  BoldHeaderText: Longbool read FBoldHeaderText write FBoldHeaderText;
    property  SearchColumn: Longint read FSearchColumn write FSearchColumn;
    property  SpaceBarMoves: TSpaceBarMoves read FSpaceBarMoves write FSpaceBarMoves;
    property  OnGetToolTipText: TOnGetToolTipText read FOnGetToolTipText write FOnGetToolTipText;
    property  OnMouseActivate: TOnMouseActivate read FOnMouseActivate write FOnMouseActivate;
    property  OnFileDropped: TNotifyEvent read FOnFileDropped write FOnFileDropped;
  end;

procedure Register;


implementation

{$IFDEF CONDITIONALEXPRESSIONS}
uses
  ImgList;
{$ENDIF}


// TArrowImages ----------------------------------------------------------------

procedure TArrowImages.DrawBlended(DestCanvas: TCanvas; const DestSizeRect: TRect; ImageIndex: Longint);
begin
  with Bitmap.Canvas, DestSizeRect do BitBlt(Handle, 0, 0, Width, Height, DestCanvas.Handle, Left, Top, SRCCOPY);
  Draw(Bitmap.Canvas, 0, 0, ImageIndex);
  BlendImage(DestCanvas, DestSizeRect, 80);
end;


// THeaderImages ---------------------------------------------------------------

procedure THeaderImages.DrawHighlighted(DestCanvas: TCanvas; const DestSizeRect: TRect; ImageIndex: Longint);
begin
  with Bitmap.Canvas do begin
    Brush.Assign(DestCanvas.Brush);
    FillRect(ClipRect);
  end;
  GetBitmap(ImageIndex, Bitmap);
  if Bitmap.Empty or not BlendImage(DestCanvas, DestSizeRect, 50) then begin
    if Multiple > 1 then
      inc(ImageIndex)
    else DrawingStyle := dsSelected;
    try
      with DestSizeRect do Draw(DestCanvas, Left, Top, ImageIndex);
    finally
      DrawingStyle := dsTransparent;
    end;
  end;
end;

// TSortList -------------------------------------------------------------------

function AlphaCompareStrings(const String1, String2: String; Invert: Longbool): Integer;
begin
  if not Invert then
    Result := CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE, PChar(String1), -1, PChar(String2), -1) - 2
  else Result := CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE, PChar(String2), -1, PChar(String1), -1) - 2;
end;

procedure TSortList.AlphaSort(FirstItem, LastItem: Longint);
var
  FirstIndex, LastIndex, MidIndex: Longint;
begin
  BeginUpdate;
  repeat
    FirstIndex := FirstItem;
    LastIndex := LastItem;
    MidIndex := (FirstItem + LastItem) shr 1;
    repeat
      while AlphaCompareStrings(Strings[FirstIndex], Strings[MidIndex], FInvert) < 0 do inc(FirstIndex);
      while AlphaCompareStrings(Strings[LastIndex], Strings[MidIndex], FInvert) > 0 do dec(LastIndex);
      if FirstIndex <= LastIndex then begin
	Exchange(FirstIndex, LastIndex);
	if MidIndex = FirstIndex then
	  MidIndex := LastIndex
	else if MidIndex = LastIndex then MidIndex := FirstIndex;
	inc(FirstIndex);
	dec(LastIndex);
      end;
    until FirstIndex > LastIndex;
    if FirstItem < LastIndex then AlphaSort(FirstItem, LastIndex);
    FirstItem := FirstIndex;
  until FirstIndex >= LastItem;
end;


// TDropGrid -------------------------------------------------------------------

procedure TDropGrid.InvalidateRect(R: TRect; MouseMove: Longbool);
begin
  Windows.InvalidateRect(WindowHandle, @R, false);
end;

procedure TDropGrid.InvalidateHeaders(OldCol, NewCol: Longint; MouseMove: Longbool);
begin
  if (GridLineWidth > 0) and not (goRowSelect in Options) then begin
    if OldCol > -1 then begin
      InvalidateRect(CellRect(OldCol, 0), MouseMove);
      InvalidateRect(CellRect(NewCol, 0), MouseMove);
    end
    else InvalidateRect(CellRect(NewCol, 0), MouseMove);
  end
  else InvalidateRect(Rect(0, 0, CellRect(pred(ColCount), 0).Right, RowHeights[0]), MouseMove);
end;

procedure TDropGrid.OpenTheme;
begin
  if UseThemes and (Color = clWindow) or (Color = clBtnFace) then FTheme := OpenThemeData(Handle, 'HEADER');
end;

procedure TDropGrid.CloseTheme;
begin
  if FTheme <> 0 then begin
    CloseThemeData(FTheme);
    FTheme := 0;
  end;
end;

function TDropGrid.GetRowState: HEADERITEMSTATES;
var
  ptMouse: TPoint;
  gcCell: TGridCoord;
begin
  Result := HIS_NORMAL;
  if FAllowHeaderClick and not IsSameCell(FCurrentCell, Cell(-1, -1)) then begin
    GetCursorPos(ptMouse);
    ptMouse := ScreenToClient(ptMouse);
    with gcCell do MouseToCell(ptMouse.X, ptMouse.Y, X, Y);
    if gcCell.Y = 0 then begin
      if GetKeyState(VK_LBUTTON) < 0 then
	Result := HIS_PRESSED
      else Result := HIS_HOT;
    end;
  end;
end;

function TDropGrid.GetColState(ACol: Longint): HEADERITEMSTATES;
begin
  Result := HIS_NORMAL;
  if FAllowHeaderClick then begin
    if ACol = FHeaderClicked then
      Result := HIS_PRESSED
    else if ACol = FHeaderTouched then Result := HIS_HOT;
  end;
end;

function TDropGrid.GetBottomRow: Longint;
begin
  Result := pred(TopRow + VisibleRowCount);
end;

function TDropGrid.GetRightCol: Longint;
begin
  Result := pred(LeftCol + VisibleColCount);
end;

procedure TDropGrid.SetAcceptFiles(Value: Longbool);
begin
  if FDropFiles <> nil then FDropFiles.AcceptFiles := Value;
  FAcceptFiles := Value;
end;

function TDropGrid.GetCell(ACol, ARow: Longint): String;
begin
  Result := inherited Cells[ACol, ARow];
end;

procedure TDropGrid.SetCell(ACol, ARow: Longint; const Value: String);
begin
  inherited Cells[ACol, ARow] := Value;
  if FAutoAdjustColWidths then AdjustColWidths(ACol, ARow);
end;

function TDropGrid.GetCol(Index: Longint): TStrings;
begin
  Result := inherited Cols[Index];
end;

procedure TDropGrid.SetCol(Index: Longint; const Value: TStrings);
begin
  inherited Cols[Index] := Value;
  if FAutoAdjustColWidths then AdjustColWidths(Index, RowCount);
end;

function TDropGrid.GetRow(Index: Longint): TStrings;
begin
  Result := inherited Rows[Index];
end;

procedure TDropGrid.SetRow(Index: Longint; const Value: TStrings);
begin
  inherited Rows[Index] := Value;
  if FAutoAdjustColWidths then AdjustColWidths(ColCount, Index);
end;

function TDropGrid.GetIndex(ACol, ARow: Longint): Integer;
begin
  Result := Integer(Objects[ACol, ARow]);
end;

procedure TDropGrid.SetIndex(ACol, ARow: Longint; const Value: Integer);
begin
  Objects[ACol, ARow] := Ptr(Value);
end;

function TDropGrid.GetColTextAlignment(Index: Longint): TAlignment;
begin
  Result := TAlignment((cardinal(Objects[Index, 0]) shr 16) and $FF);
end;

procedure TDropGrid.SetColTextAlignment(Index: Longint; Value: TAlignment);
begin
  Objects[Index, 0] := TObject((cardinal(Objects[Index, 0]) and $FF00FFFF) or cardinal(Value) shl 16);
end;

function TDropGrid.GetHeaderAlignment(Index: Longint): TAlignment;
begin
  Result := TAlignment((cardinal(Objects[Index, 0]) shr 24) and $FF);
end;

procedure TDropGrid.SetHeaderAlignment(Index: Longint; Value: TAlignment);
begin
  Objects[Index, 0] := TObject((cardinal(Objects[Index, 0]) and $00FFFFFF) or (cardinal(Value) shl 24));
end;

function TDropGrid.GetImageIndex(Index: Longint): Smallint;
begin
  Result := Smallint(Objects[Index, 0]);
end;

procedure TDropGrid.SetImageIndex(Index: Longint; Value: Smallint);
begin
  Objects[Index, 0] := TObject((cardinal(Objects[Index, 0]) and $FFFF0000) or word(Value));
end;

constructor TDropGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ParentShowHint := false;
  FArrowImages := TArrowImages.Create(Self);
  FArrowImages.DrawingStyle := dsTransparent;
  FAutoAdjustColWidths := true;
  FBoldHeaderText := true;
  FCurrentCell := Cell(-1, -1);
  FDropFiles := TDropFiles.Create(Self);
  FFastSearchTimer := TTimer.Create(Self);
  FFastSearchTimer.OnTimer := SearchTime;
  FHeaderImages := THeaderImages.Create(Self);
  FHeaderImages.DrawingStyle := dsTransparent;
  FHeaderClicked := -1;
  FHeaderTouched := -1;
  FObjectColumn := -1;
  FSortedColumn := 0;
  FToolTipCell := Cell(-1, -1);
end;

destructor TDropGrid.Destroy;
begin
  inherited Destroy;
end;

procedure TDropGrid.AdjustColWidths(ACol, ARow: Longint);

  function TextWidth(ACol, ARow: Integer): Integer;
  var
    fsGlobal: TFontStyles;
    hDC: THandle;
    sText: String;
    szText: TSize;
  begin
    fsGlobal := Font.Style;
    hDC := GetDC(WindowHandle);
    try
      if FBoldHeaderText and (ARow = 0) and (ARow < FixedRows) then Font.Style := Font.Style + [fsBold];
      SelectObject(hDC, Font.Handle);
      sText := Cells[ACol, ARow];
      GetTextExtentPoint32(hDC, PChar(sText), succ(Length(sText)), szText);
      Result := succ(szText.cx);
    finally
      SelectObject(hDC, Font.Handle);
      ReleaseDC(WindowHandle, hDC);
      Font.Style := fsGlobal;
    end;
  end;

  function NewColWidth(RawWidth: Integer): Integer;
  begin
    Result := RawWidth + GridLineWidth shl 1 + ord(GridLineWidth = 0);
  end;

var
  iNewWidth, iCol, iRow, iTmp: Integer;
begin
  if (ACol >= 0) and (ARow >= 0) then begin
    if (ACol < ColCount) and (ARow < RowCount) then begin
      iNewWidth := TextWidth(ACol, ARow) + 4;
      if ColWidths[ACol] < iNewWidth then ColWidths[ACol] := NewColWidth(iNewWidth);
    end
    else begin
      if (ACol = ColCount) and (ARow = RowCount) then Exit;
      // single row, all cols
      if ACol = ColCount then begin
	for iCol := 0 to pred(ColCount) do begin
	  iNewWidth := TextWidth(iCol, ARow) + 4;
	  if ColWidths[iCol] < iNewWidth then ColWidths[iCol] := NewColWidth(iNewWidth);
	end;
      end
      // single col, all rows
      else if ARow = RowCount then begin
	iNewWidth := ColWidths[ACol] - 4;
	for iRow := 0 to pred(RowCount) do begin
	  iTmp := TextWidth(ACol, iRow);
	  if iTmp > iNewWidth then iNewWidth := iTmp;
	end;
	ColWidths[ACol] := NewColWidth(iNewWidth + 4);
      end;
    end;
  end;
end;

procedure TDropGrid.DrawArrowImage(ACol: Longint; ASizeRect: TRect);
var
  fOutOfSpace: Longbool;
  lImageIndex: Longint;
  rCell: TRect;
begin
  with FArrowImages do begin
    if Count = 2 then with ASizeRect do begin
      fOutOfSpace := false;
      lImageIndex := ord(FSortedColumn > 0);
      rCell := CellRect(ACol, 0);
      inc(Top, (RowHeights[0] - Height) div 2);
      if HeaderAlignments[ACol] = taRightJustify then begin
	if Left - (Width + 4) >= rCell.Left then
	  dec(Left, Width + 2)
	else fOutOfSpace := true;
      end
      else begin
	if Left + Right + Width + 4 <= rCell.Right then
	  inc(Left, Right + 2)
	else fOutOfSpace := true;
      end;
      if fOutOfSpace then begin
	if Left + Width + 4 <= rCell.Right then begin
	  inc(Left, (Right - Width) div 2);
	  Right := Width;
	  Bottom := Height;
	  DrawBlended(Canvas, ASizeRect, ImageIndices[lImageIndex]);
	end;
      end
      else Draw(Canvas, Left, Top, ImageIndices[lImageIndex]);
    end;
  end;
end;

function LoStrAtPos(const Find, Dest: String; Pos: Integer): Boolean; register; assembler;
asm	// EAX = @Find, EDX = @Dest, ECX = Pos
  OR	EAX,EAX
  JE	@EXIT
  OR	EDX,EDX
  JE	@EXIT
  OR	ECX,ECX
  JLE	@EXIT
  PUSH	EBX
  MOV	EBX,[EDX - 04h]
  SUB	EBX,ECX
  INC	EBX
  CMP	EBX,[EAX - 04h]
  JL	@END
  ADD	EDX,ECX
  MOV	ECX,[EAX - 04h]
  DEC	EDX
  @LOOP:
  DEC	ECX
  JL	@MATCH
  MOV	BL,[EDX + ECX]
  XOR	BL,[EAX + ECX]
  JE	@LOOP
  TEST	BYTE PTR [EDX + ECX],040h
  JE	@END
  TEST	BYTE PTR [EAX + ECX],040h
  JE	@END
  AND	BL,0DFh
  JE	@LOOP
  POP	EBX
  XOR	EAX,EAX
  RET
  @MATCH:
  POP	EBX
  MOV	EAX,true
  RET
  @END:
  POP	EBX
  @EXIT:
  XOR	EAX,EAX
end;

procedure TDropGrid.FastSearch;
var
  sFastFind: String;
  lFoundRow, lRowIndex: Longint;
begin
  try
    sFastFind := FFastSearchChars;
    UniqueString(sFastFind);
    if sFastFind <> '' then begin
      SetFocus;
      FSearching := true;
      lFoundRow := Row;
      if FFastSearchTimer.Enabled and (lFoundRow >= FixedRows) then dec(lFoundRow);
      if lFoundRow >= pred(RowCount) then lFoundRow := pred(FixedRows);
      lRowIndex := FixedRows;
      while lRowIndex < RowCount do begin
	inc(lFoundRow);
	if not LoStrAtPos(sFastFind, TrimLeft(Cells[FSearchColumn, lFoundRow]), 1) then begin
	  if lFoundRow = pred(RowCount) then lFoundRow := pred(FixedRows);
	end
	else Break;
	inc(lRowIndex);
      end;
      if lRowIndex < RowCount then begin
	Selection := TGridRect(Rect(Col, lFoundRow, Col, lFoundRow));
	ScrollIntoView(lFoundRow);
      end
      else begin
	FFastSearchTimer.Enabled := false;
	FFastSearchChars := '';
	Beep;
      end;
    end;
  finally
    FSearching := false;
    SetFocus;
  end;
end;

procedure TDropGrid.OnHideToolTip(Sender: TObject);
begin
  FContextKeyPressed := false;
  with FToolTip do begin
    if FixedPos and (CenterToolTip = cttInParent) then begin
      HintHidePause := FHintHidePause;
      FHintHidePause := 0;
    end;
  end;
end;

procedure TDropGrid.OnShowToolTip(Sender: TObject);
begin
  with FToolTip do begin
    if FixedPos and (CenterToolTip = cttInParent) then begin
      FHintHidePause := HintHidePause;
      HintHidePause := MAXINT;
    end;
  end;
end;

procedure TDropGrid.ScrollIntoView(ARow: Longint);
begin
  if (ARow < TopRow) or (ARow >= (TopRow + VisibleRowCount)) then begin
    if (RowCount - ARow) >= VisibleRowCount then
      TopRow := ARow
    else TopRow := RowCount - VisibleRowCount;
  end;
end;

procedure TDropGrid.SearchTime(Sender: TObject);
begin
  FFastSearchTimer.Enabled := false;
  FFastSearchChars := '';
end;

procedure TDropGrid.SetToolTipPos(var ptTopLeft: TPoint);
var
  ptMouse: TPoint;
  lCol, lRow: Longint;
  rCell: TRect;
begin
  if FToolTip.CenterToolTip <> cttInParent then begin
    GetCursorPos(ptMouse);
    ptMouse := ScreenToClient(ptMouse);
    with ptMouse do MouseToCell(X, Y, lCol, lRow);
    rCell := CellRect(lCol, lRow);
    with rCell do begin
      dec(Top, Bottom - Top + 2);
      dec(Left);
    end;
    ptTopLeft := ClientToScreen(rCell.TopLeft);
  end;
end;

function TDropGrid.SortColor(ACol: Longint; clDefault: TColor): TColor;
var
  rgbSorted: TRGB;
begin
  Result := ColorToRGB(clDefault);
  if (pred(abs(FSortedColumn)) = ACol) then begin
    rgbSorted := TRGB(Result);
    if rgbSorted[R] + rgbSorted[G] + rgbSorted[B] < 256 then begin
      if rgbSorted[R] < 248 then
	inc(rgbSorted[R], 8)
      else rgbSorted[R] := 255;
      if rgbSorted[G] < 248 then
	inc(rgbSorted[G], 8)
      else rgbSorted[G] := 255;
      if rgbSorted[B] < 248 then
	inc(rgbSorted[B], 8)
      else rgbSorted[B] := 255;
    end
    else begin
      if rgbSorted[R] > 7 then
	dec(rgbSorted[R], 8)
      else rgbSorted[R] := 0;
      if rgbSorted[G] > 7 then
	dec(rgbSorted[G], 8)
      else rgbSorted[G] := 0;
      if rgbSorted[B] > 7 then
	dec(rgbSorted[B], 8)
      else rgbSorted[B] := 0;
    end;
    Result := TColor(rgbSorted);
  end;
end;

procedure TDropGrid.CMColorChanged(var Message: TMessage);
begin
  inherited;
  EnableThemes(WindowHandle, (Color = clWindow) or (Color = clBtnFace));
end;

procedure TDropGrid.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  SelectCell(Col, Row);
end;

procedure TDropGrid.CMParentShowHintChanged(var Message: TMessage);
begin
  ShowHint := false;
  ParentShowHint := false;
end;

procedure TDropGrid.CMShowHintChanged(var Message: TMessage);
begin
  ShowHint := false;
  ParentShowHint := false;
end;

procedure TDropGrid.CMMouseLeave(var Message: TMessage);
var
  ptMouse: TPoint;
begin
  inherited;
  GetCursorPos(ptMouse);
  FCurrentCell := Cell(-1, -1);
  if not PtInRect(GetClientRect, ScreenToClient(ptMouse)) then FToolTipCell := FCurrentCell;
  InvalidateHeaders(-1, FHeaderTouched, true);
  FHeaderClicked := -1;
  FHeaderTouched := -1;
end;

procedure TDropGrid.WMChar(var Message: TWMChar);
begin
  inherited;
  if (GetKeyState(VK_SHIFT) or GetKeyState(VK_CONTROL) or GetKeyState(VK_MENU)) >= 0 then with Message do begin
    if CharCode > VK_SPACE then begin
      CharLower(@CharCode);
      if (Length(FFastSearchChars) <> 1) or (CharCode <> ord(FFastSearchChars[1])) then
	FFastSearchChars := FFastSearchChars + chr(CharCode)
      else FFastSearchTimer.Enabled := false;
      FastSearch;
      FFastSearchTimer.Enabled := true;
    end;
  end;
end;

procedure TDropGrid.WMContextMenu(var Message: TWMMouse);
var
  ptMouse: TPoint;
  lColumn, lRow: Longint;
begin
  FToolTip.AnsiText := '';
  with Message do begin
    if (XPos = -1) or (YPos = -1) then begin
      FContextKeyPressed := true;
      ShowToolTip(Selection.TopLeft, true);
    end
    else begin
      ptMouse := ScreenToClient(Point(XPos, YPos));
      with ptMouse do MouseToCell(X, Y, lColumn, lRow);
      ShowToolTip(Cell(lColumn, lRow), true);
    end;
  end;
end;

procedure TDropGrid.WMCreate(var Message: TWMCreate);
begin
  inherited;
  if EnableThemes(WindowHandle, (Color = clWindow) or (Color = clBtnFace)) then OpenTheme;
  FToolTip := TToolTip.Create(Self);
  with FToolTip do begin
    ToolTipStyle := ttsCenterText;
    OnGetToolTipOrigin := SetToolTipPos;
    OnHideToolTip := Self.OnHideToolTip;
    OnShowToolTip := Self.OnShowToolTip;
  end;
  if FDropFiles <> nil then with FDropFiles do begin
    AcceptFiles := FAcceptFiles;
    OnFileDropped := FOnFileDropped;
  end;
end;

procedure TDropGrid.WMDestroy(var Message: TWMDestroy);
begin
  FToolTip.Free;
  CloseTheme;
  inherited;
end;

procedure TDropGrid.WMKeyDown(var Message: TWMKeyDown);
begin
  inherited;
  FToolTip.AnsiText := '';
  with Message do begin
    case CharCode of
      VK_ESCAPE: if FToolTip.Showing then begin
	FToolTip.Hide;
	CharCode := 0;
      end;
      VK_SPACE: if not (goRangeSelect in Options) then begin
	case FSpaceBarMoves of
	  sbmVertical: with Selection do begin
	    if GetKeyState(VK_SHIFT) >= 0 then begin
	      if Top < pred(RowCount) then begin
		Selection := TGridRect(Rect(Col, succ(Row), Col, succ(Row)));
		ScrollIntoView(Row);
	      end;
	    end
	    else if Top > FixedRows then begin
	      Selection := TGridRect(Rect(Col, pred(Row), Col, pred(Row)));
	      if Row < TopRow then begin
		if Row >= VisibleRowCount then
		  TopRow := Row - pred(VisibleRowCount)
		else TopRow := FixedRows;
	      end;
	    end;
	  end;
	  sbmHorizontal: begin
	    if GetKeyState(VK_SHIFT) >= 0 then begin
	      if Col < pred(ColCount) then Selection := TGridRect(Rect(succ(Col), Row, succ(Col), Row));
	    end
	    else if Col > FixedCols then Selection := TGridRect(Rect(pred(Col), Row, pred(Col), Row));
	  end;
	end;
      end;
      ord('I'): if GetKeyState(VK_CONTROL) < 0 then begin
	FContextKeyPressed := true;
	ShowToolTip(Selection.TopLeft, true);
	CharCode := 0;
      end;
    end;
  end;
end;

procedure TDropGrid.WMLButtonDown(var Message: TWMLButtonDown);
var
  lCol, lRow: Longint;
begin
  inherited;
  with Message.Pos do MouseToCell(X, Y, lCol, lRow);
  if FAllowHeaderClick and not (csDesigning in ComponentState) then begin
    if lRow = 0 then
      FHeaderClicked := lCol
    else FHeaderClicked := -1;
    InvalidateHeaders(-1, lCol, false);
  end
  else FHeaderClicked := -1;
  FPrevRowWidth := VisibleRowWidth;
end;

procedure TDropGrid.WMLButtonUp(var Message: TWMLButtonUp);
var
  lCol, lRow: Longint;
begin
  inherited;
  with Message.Pos do MouseToCell(X, Y, lCol, lRow);
  if FAllowHeaderClick and not (csDesigning in ComponentState) then begin
    if (FPrevRowWidth <> VisibleRowWidth) or not IsSameCell(Cell(FHeaderClicked, 0), Cell(lCol, lRow)) then begin
      FHeaderClicked := -1;
    end;
    SortRows(FHeaderClicked, false, false);
    InvalidateHeaders(-1, lCol, false);
    FHeaderClicked := -1;
  end
  else FHeaderClicked := -1;
end;

procedure TDropGrid.WMMouseActivate(var Message: TWMMouseActivate);
var
  ptMouse: TPoint;
  lCol, lRow: Longint;
begin
  if Assigned(FOnMouseActivate) then begin
    GetCursorPos(ptMouse);
    ptMouse := ScreenToClient(ptMouse);
    with ptMouse do MouseToCell(X, Y, lCol, lRow);
    FOnMouseActivate(Message.MouseMsg, lCol, lRow);
  end;
  inherited;
end;

procedure TDropGrid.WMMouseMove(var Message: TWMMouseMove);
var
  lCol, lRow: Longint;
begin
  inherited;
  with Message.Pos do MouseToCell(X, Y, lCol, lRow);
  if (lRow * FCurrentCell.Y = 0) and not IsSameCell(Cell(lCol, lRow), FCurrentCell) then begin
    if not (csDesigning in ComponentState) then InvalidateHeaders(FCurrentCell.X, lCol, true);
  end;
  FCurrentCell := Cell(lCol, lRow);
  if lRow = 0 then
    FHeaderTouched := lCol
  else FHeaderTouched := -1;
  if not IsSameCell(Cell(lCol, lRow), FToolTipCell) then begin
    FToolTip.AnsiText := '';
    ShowToolTip(Cell(lCol, lRow), (lRow = 0));
  end
  else if lRow = -1 then ShowToolTip(Cell(-1, -1), false);
end;

procedure TDropGrid.WMSysKeyDown(var Message: TWMSysKeyDown);
begin
  if not (goEditing in Options) and (Message.CharCode in [VK_UP, VK_DOWN]) then begin
    SortRows(Selection.Left, false, false);
  end;
end;

procedure TDropGrid.WMThemeChanged(var Message: TMessage);
begin
  CloseTheme;
  inherited;
  OpenTheme;
end;

procedure TDropGrid.ColumnMoved(FromIndex, ToIndex: Longint);
begin
  if FObjectColumn = FromIndex then FObjectColumn := ToIndex;
  FHeaderClicked := -1;
  inherited;
end;

procedure TDropGrid.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
begin
  FToolTip.ToolRect := Rect(0, 0, GridWidth, GridHeight);
  if not IsRectEmpty(ARect) then begin
    if not Assigned(OnDrawCell) then begin
      if (ARow = 0) and (FixedRows = 1) and (AState = [gdFixed]) then
	DrawHeaderCell(ACol, ARect)
      else begin
	DrawCellBackground(ACol, ARect, AState, Color);
	DrawCellText(ACol, ARow, ARect, AState);
      end;
    end
    else OnDrawCell(Self, ACol, ARow, ARect, AState);
  end;
end;

procedure TDropGrid.RowMoved(FromIndex, ToIndex: Longint);
begin
  FSortedColumn := 0;
  inherited;
end;

function TDropGrid.Cell(X, Y: Longint): TGridCoord;
begin
  Result.X := X;
  Result.Y := Y;
end;

procedure TDropGrid.DeleteRow(ARow: Longint);
var
  lTopRow: Longint;
  rRedraw: TRect;
begin
  lTopRow := TopRow;
  rRedraw := CellRect(LeftCol, ARow);
  if RowCount = succ(FixedRows) then
    Rows[FixedRows].Clear
  else begin
    for ARow := ARow to (RowCount - 2) do Rows[ARow] := Rows[succ(ARow)];
    Rows[pred(RowCount)].Clear;
    RowCount := pred(RowCount);
  end;
  if Row + VisibleRowCount > RowCount then TopRow := succ(lTopRow);
  rRedraw.BottomRight := ClientRect.BottomRight;
  InvalidateRect(rRedraw, false);
end;

procedure TDropGrid.DrawCellBackground(ACol: Longint; ARect: TRect; AState: TGridDrawState; ADefColor: TColor);
begin
  with Canvas do begin
    if gdFixed in AState then begin
      SetTextColor(Handle, ColorToRGB(Font.Color));
      Brush.Color := SortColor(ACol, FixedColor);
    end
    else begin
      if gdSelected in AState then begin
	SetTextColor(Handle, ColorToRGB(clHighlightText));
	Brush.Color := clHighlight;
      end
      else begin
	SetTextColor(Handle, ColorToRGB(Font.Color));
	Brush.Color := SortColor(ACol, ADefColor);
      end;
    end;
    FillRect(ARect);
  end;
end;

function TDropGrid.DrawCellImage(ACol: Longint; ARect: TRect; AState: TGridDrawState): TRect;
var
  taColumn: TAlignment;
  lImageIndex: Longint;
begin
  with ARect, FHeaderImages do begin
    if ((Right - Left) >= Width) and ((Bottom - Top) >= Height) then begin
      if AState = [gdFixed] then
	taColumn := HeaderAlignments[ACol]
      else taColumn := ColTextAlignments[ACol];
      case taColumn of
	taCenter: inc(Left, succ(Right - Left - Width) div 2);
	taRightJustify: Left := Right - Width;
      end;
      Right := Width;
      Bottom := Height;
      lImageIndex := ImageIndices[ACol] * Multiple;
      if gdSelected in AState then
	DrawHighlighted(Canvas, ARect, lImageIndex)
      else Draw(Canvas, Left, Top, lImageIndex);
    end;
  end;
  Result := ARect;
end;

function TDropGrid.DrawCellText(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState): TRect;
const
  AAlignments: Array[TAlignment] of Integer = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  pText: PChar;
begin
  pText := PChar(Cells[ACol, ARow]);
  if (pText[0] <> #0) and (pText <> #32) then with Canvas do begin
    dec(ARect.Right);
    InflateRect(ARect, -2, -2);
    if ColTextAlignments[ACol] <> taRightJustify then inc(ARect.Left);
    DrawText(Handle, pText, -1, ARect, AAlignments[ColTextAlignments[ACol]] or DT_NOPREFIX or DT_VCENTER);
  end;
  Result := ARect;
end;

procedure TDropGrid.DrawGridFocus(ARect: TRect);
begin
  with Canvas do begin
    if goRowSelect in Options then begin
      with ARect do ARect := Rect(0, Top, CellRect(pred(ColCount), 0).Right, Bottom);
      FrameRect(ARect);
      DrawFocusRect(ARect);
    end
    else DrawFocusRect(ARect);
  end;
end;

procedure TDropGrid.DrawHeaderCellContents(ACol: Longint; ARect: TRect);
begin
  if ImageIndices[ACol] = -1 then
    ARect := DrawHeaderText(ACol, ARect)
  else ARect := DrawCellImage(ACol, ARect, [gdFixed]);
  if (FHeaderClicked = -1) and (pred(abs(FSortedColumn)) = ACol) then DrawArrowImage(ACol, ARect);
end;

procedure TDropGrid.DrawHeaderCell(ACol: Longint; ARect: TRect);
var
  iCol: Longint;
  hisRow: HEADERITEMSTATES;
begin
  if (FTheme <> 0) and (FixedCols = 0) then begin
    if not (goRowSelect in Options) then begin
      inc(ARect.Bottom);
      with Canvas do begin
	Brush.Color := clSilver;
	InflateRect(ARect, 1, 0);
	FrameRect(ARect);
	InflateRect(ARect, -1, 0);
	DrawThemeBackground(FTheme, Handle, HP_HEADERITEM, GetColState(ACol), ARect, nil);
      end;
      dec(ARect.Bottom);
      if ACol = FHeaderClicked then OffsetRect(ARect, 1, 1);
      DrawHeaderCellContents(ACol, ARect);
    end
    else if ACol = pred(ColCount) then begin
      hisRow := GetRowState;
      ARect := Rect(0, 0, CellRect(pred(ColCount), 0).Right, RowHeights[0]);
      with Canvas do begin
	if FHeaderClicked = -1 then begin
	  inc(ARect.Bottom);
	  inc(ARect.Right);
	end;
	DrawThemeBackground(FTheme, Handle, HP_HEADERITEM, hisRow, ARect, nil);
	if FHeaderClicked = -1 then begin
	  dec(ARect.Bottom);
	  dec(ARect.Right);
	end;
	Brush.Color := clSilver;
	with ARect do FrameRect(Rect(Left, Top, succ(Right), succ(Bottom)));
      end;
      for iCol := 0 to pred(ColCount) do begin
	ARect := CellRect(iCol, 0);
	if not IsRectEmpty(ARect) then begin
	  if hisRow = HIS_PRESSED then OffsetRect(ARect, 1, 1);
	  DrawHeaderCellContents(iCol, ARect);
	end;
      end;
    end;
  end
  else begin
    with Canvas do begin
      Brush.Color := FixedColor;
      FillRect(ARect);
    end;
    if ACol = FHeaderClicked then begin
      with ARect do begin
	inc(Right);
	inc(Bottom);
      end;
      Frame3D(Canvas, ARect, clBtnShadow, clBtnShadow, 1);
      with ARect do begin
	inc(Left);
	inc(Top);
      end;
    end
    else Frame3D(Canvas, ARect, clBtnHighlight, clBtnShadow, 1);
    InflateRect(ARect, 1, 1);
    DrawHeaderCellContents(ACol, ARect);
  end;
end;

function TDropGrid.DrawHeaderText(ACol: Longint; ARect: TRect): TRect;
var
  rCalc: TRect;
  pText: PChar;
  iAlignment: Integer;
  fsGlobal: TFontStyles;
begin
  SetRectEmpty(rCalc);
  pText := PChar(Cells[ACol, 0]);
  if pText[0] <> #0 then begin
    InflateRect(ARect, -2, 0);
    iAlignment := DT_LEFT;
    case HeaderAlignments[ACol] of
      taLeftJustify: inc(ARect.Left);
      taCenter: begin
	inc(ARect.Left);
	iAlignment := DT_CENTER;
      end;
      taRightJustify: iAlignment := DT_RIGHT;
    end;
    with Canvas do begin
      fsGlobal := Font.Style;
      SetBkMode(Handle, TRANSPARENT);
      try
      if FBoldHeaderText then Font.Style := Font.Style + [fsBold];
	DrawText(Handle, pText, -1, ARect, iAlignment or DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER);
	DrawText(Handle, pText, -1, rCalc, iAlignment or DT_CALCRECT or DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER);
      finally
	SetBkMode(Handle, OPAQUE);
	Font.Style := fsGlobal;
      end;
    end;
    with ARect do begin
      case iAlignment of
	DT_CENTER: inc(Left, succ(Right - Left - rCalc.Right) div 2);
	DT_RIGHT: Left := Right - rCalc.Right;
      end;
    end;
  end;
  with ARect do begin
    Right := rCalc.Right;
    Bottom := rCalc.Bottom;
  end;
  Result := ARect;
end;

function TDropGrid.IsSameCell(Pt1, Pt2: TGridCoord): Longbool;
begin
  Result := (Pt1.X = Pt2.X) and (Pt1.Y = Pt2.Y);
end;

procedure TDropGrid.LoadArrowImages(ArrowsID: Integer);
begin
  if ArrowsID <> 0 then with FArrowImages do begin
    Bitmap.LoadFromResourceID(hInstance, ArrowsID);
    Handle := ImageList_Create(Bitmap.Width shr 1, Bitmap.Height, ILC_COLOR32 or ILC_MASK, 2, 0);
    AddMasked(Bitmap, Bitmap.TransparentColor);
    if Count = 2 then Bitmap.Width := Width;
  end;
end;

procedure TDropGrid.SetColumnData(Column: Longint; HeaderAlignment, ColumnAlignment: TAlignment;
				  const Text: String; BitmapID: Integer);
begin
  if Column < ColCount then begin
    Cells[Column, 0] := Text;
    HeaderAlignments[Column] := HeaderAlignment;
    ColTextAlignments[Column] := ColumnAlignment;
  end;
  if BitmapID <> 0 then with FHeaderImages, Bitmap do begin
    LoadFromResourceID(hInstance, BitmapID);
    Multiple := Bitmap.Width div FHeaderImages.Width;
    ImageIndices[Column] := AddMasked(Bitmap, TransparentColor);
  end
  else ImageIndices[Column] := -1;
end;

function TDropGrid.SearchGrid: Longbool;
var
  cDefault: TCursor;
  stKind: TSearchTypes;
  lOldPos, lNewPos, lRowIndex: Longint;
begin
  Result := true;
  if not FSearching then begin
    FSearching := true;
    cDefault := Screen.Cursor;
    try
      stKind := [stAnsiChar];
      if (not FReSearch or (FFindText = '')) and Assigned(GetSearchText) then begin
	FFindText := AnsiLowerCase(GetSearchText(FFindText, Font.Name, stKind));
      end;
      FReSearch := false;
      if FFindText <> '' then begin
	SetFocus;
	Screen.Cursor := crHourGlass;
	Application.ProcessMessages;
	lOldPos := Row;
	if lOldPos >= pred(RowCount) then lOldPos := pred(FixedRows);
	lNewPos := -1;
	for lRowIndex := succ(lOldPos) to pred(RowCount) do begin
	  if pos(FFindText, AnsiLowerCase(Cells[FSearchColumn, lRowIndex])) <> 0 then begin
	    lNewPos := lRowIndex;
	    Break;
	  end;
	end;
	if lNewPos <> -1 then begin
	  Selection := TGridRect(Rect(Col, lNewPos, Col, lNewPos));
	  ScrollIntoView(lNewPos);
	  Exit;
	end
	else if (lOldPos <> FixedRows) and (lOldPos <> pred(RowCount)) then begin
	  FReSearch := true;
	  Exit;
	end
	else Result := false;
      end;
    finally
      Screen.Cursor := cDefault;
      FSearching := false;
      SetFocus;
    end;
  end
  else FReSearch := false;
end;

procedure TDropGrid.SetImageProperties(Count, AWidth, AHeight: Integer);
begin
  with FHeaderImages do Handle := ImageList_Create(AWidth, AHeight, ILC_COLOR32 or ILC_MASK, Count, 0);
end;

procedure TDropGrid.ShowToolTip(SelectedCell: TGridCoord; Fixed: Longbool);
var
  ptMouse: TPoint;
  sToolTipText: String;
begin
  FToolTipCell := SelectedCell;
  if IsSameCell(FToolTipCell, Cell(-1, -1)) or FContextKeyPressed then begin
    GetCursorPos(ptMouse);
    ptMouse := ScreenToClient(ptMouse);
    with FToolTipCell do MouseToCell(ptMouse.X, ptMouse.Y, X, Y);
  end;
  with FToolTip do begin
    HMargin := 4;
    VMargin := 2;
    CenterToolTip := cttNot;
    if (FixedRows = 1) and (SelectedCell.Y = 0) then begin
      FixedPos := true;
    end
    else begin
      if Fixed then CenterToolTip := cttInParent;
      FixedPos := Fixed;
    end;
    with SelectedCell do if (X <> -1) and (Y <> -1) then begin
      if Assigned(FOnGetToolTipText) then begin
	FOnGetToolTipText(X, Y, Fixed, sToolTipText);
	AnsiText := sToolTipText;
	Show;
      end;
    end;
  end;
end;

function TDropGrid.SortRows(SortCol: Longint; GetList, Invisible: Longbool): String;

  procedure GetRowTexts(const SortList: TSortList);
  var
    lGridRow, lStartPos, lRowLength, lColumn, lRowPos: Longint;
    sRow: String;
    pRow: PChar;
  begin
    for lGridRow := FixedRows to pred(RowCount) do begin
      lStartPos := 0;
      lRowLength := 0;
      for lColumn := 0 to pred(ColCount) do begin
	FCellTextLengths^[lColumn] := Length(Cells[lColumn, lGridRow]);
	if FCellTextLengths^[lColumn] = 0 then begin
	  Cells[lColumn, lGridRow] := ' ';
	  FCellTextLengths^[lColumn] := 1;
	end;
	inc(lRowLength, FCellTextLengths^[lColumn]);
      end;
      if FObjectColumn <> -1 then inc(lRowLength, 8);
      if lRowLength > 0 then begin
	SetString(sRow, nil, lRowLength + (ColCount * Length(#10)));
	pRow := Pointer(sRow);
	lRowPos := 0;
	for lColumn := FSortStartColumn to pred(ColCount) do begin
	  Move(Pointer(Cells[lColumn, lGridRow])^, pRow[lRowPos], FCellTextLengths^[lColumn]);
	  inc(lRowPos, FCellTextLengths^[lColumn]);
	  pRow[lRowPos] := #10;
	  inc(lRowPos);
	end;
	lStartPos := succ(lRowPos);
	for lColumn := 0 to pred(FSortStartColumn) do begin
	  Move(Pointer(Cells[lColumn, lGridRow])^, pRow[lRowPos], FCellTextLengths^[lColumn]);
	  inc(lRowPos, FCellTextLengths^[lColumn]);
	  pRow[lRowPos] := #10;
	  inc(lRowPos);
	end;
	if lStartPos > lRowPos then lStartPos := 1;
	if FObjectColumn <> -1 then begin
	  Move(Pointer(IntToHex(Longint(Self.Objects[FObjectColumn, lGRidRow]), 8))^, pRow[lRowPos], 8);
	end;
      end
      else sRow := '';
      SortList.AddObject(sRow, TObject(lStartPos));
    end;
  end;

  procedure SetRowTexts(const SortList: TSortList);
  var
    lGridRow, lSortRow, lStartPos, lCopyLen: Longint;
    sRow, sBuffer: String;
    pRow: PChar;
  begin
    for lGridRow := FixedRows to pred(RowCount) do begin
      lSortRow := lGridRow - FixedRows;
      sBuffer := SortList.Strings[lSortRow];
      lStartPos := Longint(SortList.Objects[lSortRow]);
      lCopyLen := Length(sBuffer);
      if FObjectColumn <> -1 then dec(lCopyLen, 8);
      if lCopyLen > Length(#10) then begin
	SetString(sRow, nil, lCopyLen);
	pRow := Pointer(sRow);
	dec(lCopyLen, pred(lStartPos));
	Move(sBuffer[lStartPos], pRow[0], lCopyLen);
	Move(Pointer(sBuffer)^, pRow[lCopyLen], pred(lStartPos));
      end
      else sRow := '';
      Rows[lGridRow].Text := sRow;
      ValidateRect(WindowHandle, nil);
    end;
    if FObjectColumn <> -1 then begin
      for lGridRow := FixedRows to pred(RowCount) do begin
	lSortRow := lGridRow - FixedRows;
	sBuffer := '$' + Copy(SortList[lSortRow], Length(SortList[lSortRow]) - 7, 8);
	Objects[FObjectColumn, lGridRow] := TObject(StrToInt(sBuffer));
      end;
    end;
  end;

var
  lSortedCol, lSelCol, lSelRow: Longint;
  SortList: TSortList;
  sSelected: String;
  rSorted: TRect;
begin
  Result := '';
  FSortStartColumn := SortCol;
  if (RowCount > FixedRows) and (FSortStartColumn >= 0) and (FSortStartColumn < ColCount) then begin
    SortList := TSortList.Create;
    GetMem(FCellTextLengths, SizeOf(Integer) * ColCount);
    if (SortList <> nil) and (FCellTextLengths <> nil) then begin
      if Invisible then
	lSortedCol := 0
      else lSortedCol := FSortedColumn;
      try
	lSelCol := -1;
	lSelRow := -1;
	GetRowTexts(SortList);
	with SortList do begin
	  if Showing and not (goRangeSelect in Options) then begin
	    lSelCol := Selection.Left;
	    lSelRow := Selection.Top;
	    if lSelRow > -1 then dec(lSelRow, FixedRows);
	    if lSelRow < Count then sSelected := Strings[lSelRow];
	  end;
	  if (succ(SortCol) <> abs(lSortedCol)) or (lSortedCol < 0) then begin
	    Invert := false;
	    AlphaSort(0, pred(Count));
	    lSortedCol := succ(SortCol);
	  end
	  else begin
	    Invert := true;
	    AlphaSort(0, pred(Count));
	    lSortedCol := -succ(SortCol);
	  end;
	  if sSelected <> '' then begin
	    lSelRow := IndexOf(sSelected);
	    if lSelRow > -1 then inc(lSelRow, FixedRows);
	  end;
	end;
	if not Invisible then begin
	  SetRowTexts(SortList);
	  if Showing and (lSelCol <> -1) and (lSelRow <> -1) then begin
	    if goRowSelect in Options then
	      Selection := TGridRect(Rect(0, lSelRow, pred(ColCount), lSelRow))
	    else Selection := TGridRect(Rect(lSelCol, lSelRow, lSelCol, lSelRow));
	    ScrollIntoView(lSelRow);
	  end;
	  with rSorted do begin
	    TopLeft := CellRect(0, TopRow).TopLeft;
	    Bottom := ClientHeight;
	    Right := ClientWidth;
	  end;
	  InvalidateRect(rSorted, false);
	end;
	if GetList then Result := SortList.Text;
      finally
	if not Invisible then FSortedColumn := lSortedCol;
	FreeMem(FCellTextLengths);
	SortList.Free;
      end;
    end;
  end;
end;

function TDropGrid.VisibleRowWidth: Longint;
begin
  Result := CellRect(pred(FixedCols + VisibleColCount), TopRow).Right -
  	    CellRect(LeftCol, TopRow).Left + (GridLineWidth shl 1);
end;

procedure Register;
begin
  RegisterComponents('New', [TDropGrid]);
end;

end.

