{$J+,Z4}
unit ExtCtrlsEx;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, Classes, Graphics, Controls, ExtCtrls, UxThemes;

type
  TColoredFramePanel = class(TCustomPanel)
  private
    FFrameColor: TColor;
    FTabStyle: Longbool;
    FTheme: THandle;
    procedure OpenTheme;
    procedure CloseTheme;
    procedure InvalidateBkGnd;
    procedure InvalidateFrame;
    procedure RedrawFrame;
    procedure SetFrameColor(Value: TColor);
    procedure SetTabStyle(Value: Longbool);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMEraseBkGnd(var Message: TWMEraseBkGnd); message WM_ERASEBKGND;
    procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TWMNCPaint); message WM_NCPAINT;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMThemeChanged(var Message: TMessage); message WM_THEMECHANGED;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
  published
    property Align;
    property Alignment;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Caption;
    property Color;
    property Ctl3D;
    property Font;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
    property FrameColor: TColor read FFrameColor write SetFrameColor;
    property TabStyle: Longbool read FTabStyle write SetTabStyle;
  end;

type
  TAlignedShape = class(TShape)
  private
  protected
  public
  published
    property Align;
  end;

procedure Register;


implementation

procedure TColoredFramePanel.OpenTheme;
begin
  if UseThemes and FTabStyle and (Color = clBtnFace) then FTheme := OpenThemeData(Handle, 'TAB');
end;

procedure TColoredFramePanel.CloseTheme;
begin
  if FTheme <> 0 then begin
    CloseThemeData(FTheme);
    FTheme := 0;
  end;
end;

procedure TColoredFramePanel.InvalidateBkGnd;
var
  rClient: TRect;
begin
  if WindowHandle <> 0 then begin
    rClient := ClientRect;
    Perform(WM_THEMECHANGED, 0, 0);
    InvalidateRect(WindowHandle, @rClient, true);
  end;
end;

procedure TColoredFramePanel.InvalidateFrame;
begin
  if WindowHandle <> 0 then begin
    RedrawWindow(WindowHandle, nil, 0, RDW_FRAME or RDW_INVALIDATE);
  end;
end;

procedure TColoredFramePanel.RedrawFrame;
begin
  if WindowHandle <> 0 then begin
    SetWindowPos(WindowHandle, 0, 0, 0, 0, 0, SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER or SWP_DRAWFRAME);
  end;
end;

procedure TColoredFramePanel.SetFrameColor(Value: TColor);
begin
  if Value <> FFrameColor then begin
    FFrameColor := Value;
    RedrawFrame;
  end;
end;

procedure TColoredFramePanel.SetTabStyle(Value: Longbool);
begin
  if Value <> FTabStyle then begin
    FTabStyle := Value;
    InvalidateBkGnd;
  end;
end;

procedure TColoredFramePanel.CMColorChanged(var Message: TMessage);
begin
  inherited;
  InvalidateBkGnd;
end;

procedure TColoredFramePanel.CMTextChanged(var Message: TMessage);
begin
  if FTheme <> 0 then
    InvalidateBkGnd
  else inherited;
end;

procedure TColoredFramePanel.WMCreate(var Message: TWMCreate);
begin
  inherited;
  OpenTheme;
end;

procedure TColoredFramePanel.WMDestroy(var Message: TWMDestroy);
begin
  CloseTheme;
  inherited;
end;

procedure TColoredFramePanel.WMEraseBkGnd(var Message: TWMEraseBkGnd);
const
  ATabItemStates: Array[Boolean] of TABITEMSTATES = (TIS_DISABLED, TIS_NORMAL);
var
  iState: Integer;
begin
  if FTheme <> 0 then begin
    iState := ATabItemStates[Enabled];
    if IsThemeBackgroundPartiallyTransparent(FTheme, TABP_BODY, iState) then begin
      DrawThemeParentBackground(WindowHandle, Message.DC, nil);
    end;
    DrawThemeBackground(FTheme, Message.DC, TABP_BODY, iState, ClientRect, nil);
    Message.Result := -1;
  end
  else inherited;
end;

procedure TColoredFramePanel.WMNCCalcSize(var Message: TWMNCCalcSize);
var
  hDC: THandle;
begin
  inherited;
  if FFrameColor = clNone then Exit;
  hDC := GetDC(GetParent(WindowHandle));
  try
    with Message do begin
      InflateRect(CalcSize_Params.rgrc[0], -1, -1);
      if CalcValidRects then Result := WVR_REDRAW;
    end;
  finally
    ReleaseDC(GetParent(WindowHandle), hDC);
  end;
end;

procedure TColoredFramePanel.WMNCPaint(var Message: TWMNCPaint);
var
  rFrame: TRect;
  hDC, hBrush: THandle;
begin
  inherited;
  if FFrameColor <> clNone then begin
    with rFrame do begin
      Left := 0;
      Top := 0;
      Right := Width;
      Bottom := Height;
    end;
    hDC := GetWindowDC(WindowHandle);
    hBrush := CreateSolidBrush(ColorToRGB(FFrameColor));
    try
      FrameRect(hDC, rFrame, hBrush);
      Message.Result := 0;
    finally
      DeleteObject(hBRush);
      ReleaseDC(WindowHandle, hDC);
    end;
  end;
end;

procedure TColoredFramePanel.WMSize(var Message: TWMSize);
begin
  InvalidateFrame;
  inherited;
end;

procedure TColoredFramePanel.WMThemeChanged(var Message: TMessage);
begin
  CloseTheme;
  inherited;
  OpenTheme;
  Invalidate;
end;

constructor TColoredFramePanel.Create(AOwner: TComponent);
begin
  inherited;
  BevelInner := bvNone;
  BevelOuter := bvNone;
  FFrameColor := clWindowFrame;
end;

destructor TColoredFramePanel.Destroy;
begin
  inherited;
end;

procedure TColoredFramePanel.Paint;
const
  AAlignments: Array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  rText: TRect;
  iFontHeight: Integer;
begin
  with Canvas do begin
    Font.Assign(Self.Font);
    Brush.Style := bsClear;
    iFontHeight := TextHeight('W');
    rText := ClientRect;
    with rText do begin
      Top := ((Bottom + Top) - iFontHeight) div 2;
      Bottom := Top + iFontHeight;
    end;
    DrawText(Handle, PChar(Caption), -1, rText, (DT_EXPANDTABS or DT_VCENTER) or AAlignments[Alignment]);
  end;
end;


procedure Register;
begin
  RegisterComponents('Modified', [TColoredFramePanel, TAlignedShape]);
end;

end.

