{$J+,Z4}
unit FileDialogs;

{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2001-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, CommDlg, Classes, Dialogs, StrTools, SysUtils;

const
  MAX_PATH_LENGTH	= $FFFF;
  ALLFILE_FILTER	= '*.*|*.*';
  PUBRING_FILTER	= 'pgp, pkr, pubkr, gpg|*.pgp;*.pkr;*.pubkr;*.gpg' + LF + ALLFILE_FILTER;
  SECRING_FILTER	= 'pgp, skr, prvkr, gpg|*.pgp;*.skr;*.prvkr;*.gpg' + LF + ALLFILE_FILTER;
  GROUPS_FILTER		= 'pgr|*.pgr' + LF + ALLFILE_FILTER;
  RANDSEED_FILTER	= 'bin, rnd|*.bin;*.rnd' + LF + ALLFILE_FILTER;
  CERTEXPORT_FILTER	= 'cer, pem|*.cer;*.pem';
  CERTIMPORT_FILTER	= 'cer, der, pem, pfx, p7b, p12|*.cer;*.der;*.pem;*.pfx;*.p7b;*.p12';
  KEYEXPORT_FILTER	= 'asc, gpg, pgp|*.asc;*.gpg;*.pgp';
  KEYIMPORT_FILTER	= 'asc, gpg, pgp, pkr, pubkr|*.asc;*.gpg;*.pgp;*.pkr;*.pubkr';
  CKIMPORT_FILTER	= 'asc, gpg, pgp, pkr, pubkr, cer, der, pem, pfx, p7b, p12|' +
			  '*.asc;*.gpg;*.pgp;*.pkr;*.pubkr;*.cer;*.der;*.pem;*.pfx;*.p7b;*.p12';
  ENCODE_FILTER		= 'txt, rtf, doc, wri|*.txt;*.rtf;*.doc;*.wri';
  DECODE_FILTER		= 'asc, gpg, pgp, sig|*.asc;*.gpg;*.pgp;*.sig';
  CLIP_NAME		= ':CLIP:';

type
  TOpenFileName = packed record
    lStructSize: DWORD;
    hWndOwner: HWND;
    hInstance: HINST;
    lpstrFilter: PChar;
    lpstrCustomFilter: PChar;
    nMaxCustFilter: DWORD;
    nFilterIndex: DWORD;
    lpstrFile: PChar;
    nMaxFile: DWORD;
    lpstrFileTitle: PChar;
    nMaxFileTitle: DWORD;
    lpstrInitialDir: PChar;
    lpstrTitle: PChar;
    Flags: DWORD;
    nFileOffset: Word;
    nFileExtension: Word;
    lpstrDefExt: PChar;
    lCustData: LPARAM;
    lpfnHook: function(Wnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT stdcall;
    lpTemplateName: PChar;
    pvReserved: Pointer;
    dwReserved: DWORD;
    FlagsEx: DWORD;
  end;

type
  TOpenDialogEx = class(TOpenDialog)
  private
    FHandle: THandle;
    FParentHandle: THandle;
    FFilterList: TStringList;
    FDefaultExtension: String;
    FInitialDir: String;
    FOpenFileName: TOpenFileName;
    FMsgProc: Pointer;
    FShortCut: Byte;
    FBtnText: String;
    FFileName: String;
    FFromClipboard: Longbool;
    function GetFlags: Integer;
    procedure SetBtnText(const Value: String);
    procedure SetDefaultExtension(const Value: String);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: Boolean; override;
    property Handle: THandle
      read FHandle;
  published
    property ClipBtnText: String
      read FBtnText
      write SetBtnText;
    property DefaultExtension: String
      read FDefaultExtension
      write SetDefaultExtension;
    property Filter: TStringList
      read FFilterList
      write FFilterList;
    property InitialDir: String
      read FInitialDir
      write FInitialDir;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  end;

type
  TSaveDialogEx = class(TSaveDialog)
  private
    FHandle: THandle;
    FParentHandle: THandle;
    FFilterList: TStringList;
    FDefaultExtension: String;
    FInitialDir: String;
    FSaveFileName: TOpenFileName;
    FMsgProc: Pointer;
    FShortCut: Byte;
    FBtnText: String;
    FFileName: String;
    FToClipboard: Longbool;
    function GetFlags: Integer;
    procedure SetBtnText(const Value: String);
    procedure SetDefaultExtension(const Value: String);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: Boolean; override;
    property Handle: THandle
      read FHandle;
  published
    property ClipBtnText: String
      read FBtnText
      write SetBtnText;
    property DefaultExtension: String
      read FDefaultExtension
      write SetDefaultExtension;
    property Filter: TStringList
      read FFilterList
      write FFilterList;
    property InitialDir: String
      read FInitialDir
      write FInitialDir;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  end;


implementation

{$R *.RES}
{$I *.INC}

const
  OPENFILENAME_SIZE_VERSION_400 = SizeOf(TOpenFileName) - 12;
  FSTANDARD = OFN_ENABLEHOOK or OFN_ENABLETEMPLATE or OFN_EXPLORER or OFN_HIDEREADONLY or
	      OFN_NOCHANGEDIR or OFN_PATHMUSTEXIST or OFN_OVERWRITEPROMPT;
  FOPEN = FSTANDARD or OFN_FILEMUSTEXIST;
  FSAVE = FSTANDARD or OFN_NOREADONLYRETURN;

function PCharList2StringList(pText: PChar; var Count: Integer): Integer; register; assembler;
asm	// EAX = pText, EDX = @Count
  PUSH	EDI
  PUSH	EAX
  MOV	DWORD PTR [EDX],0
  MOV	EDI,EAX
  OR	EAX,EAX
  JZ	@END
  MOV	ECX,MAX_PATH_LENGTH
  XOR	EAX,EAX
  @LOOP:
  CLD
  REPNZ	SCASB
  JNZ	@END
  MOV	BYTE PTR [EDI - 1],LF
  INC	DWORD PTR [EDX]
  OR	BYTE PTR [EDI],0
  JNZ	@LOOP
  @END:
  POP	EDX
  MOV	EAX,EDI
  SUB	EAX,EDX
  POP	EDI
end;

function GetInsertPath(var FileList: String): String;
var
  iCount, iPos, iRes, iLen: Integer;
  NewList: String;
  pList: PChar;
begin
  Delete(FileList, succ(PCharList2StringList(Pointer(FileList), iCount)), MAXINT);
  iPos := FirstCharPos(LF, FileList);
  Result := ExtractStr(FileList, 1, iPos);
  iRes := Length(Result);
  if iRes > 0 then begin
    if iCount = 1 then begin
      FileList := Result;
      Delete(FileList, iRes, 1);
      Result := ExtractFilePath(FileList);
    end
    else begin
      Result[iRes] := BS;
      SetString(NewList, nil, Length(FileList) + pred(iCount) * iRes);
      pList := Pointer(NewList);
      iPos := 0;
      for iCount := pred(iCount) downto 1 do begin
	Move(Pointer(Result)^, pList^, iRes);
	inc(pList, iRes);
	iLen := NextCharPos(LF, FileList, succ(iPos)) - iPos;
	Move(FileList[succ(iPos)], pList^, iLen);
	inc(pList, iLen);
	inc(iPos, iLen);
      end;
      Delete(NewList, pList - Pointer(NewList), 1);
      FileList := NewList;
    end;
  end;
end;

procedure ProcessMessages;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;
end;

procedure CenterDialog(hDialog, hParent: THandle);
var
  rDialog, rParent: TRect;
begin
  if hParent = 0 then hParent := GetDesktopWindow;
  GetWindowRect(hParent, rParent);
  with rParent do begin
    dec(Right, pred(Left));
    dec(Bottom, pred(Top));
  end;
  GetWindowRect(hDialog, rDialog);
  with rDialog do begin
    dec(Right, pred(Left));
    dec(Bottom, pred(Top));
    Left := rParent.Left + (rParent.Right - Right) div 2;
    Top := rParent.Top + (rParent.Bottom - Bottom) div 2;
    SetWindowPos(hDialog, HWND_TOP, Left, Top, 0, 0, SWP_DRAWFRAME or SWP_NOSIZE);
  end;
end;

function foFakeEdit(hWnd: THandle; lParam: LPARAM): BOOL; stdcall;
var
  aFoundText: Array[0..MAX_PATH_LENGTH] of Char;
begin
  Result := true;
  if (GetClassName(hWnd, aFoundText, MAX_PATH) <> 0) and (UpperCase(aFoundText) = 'EDIT') then begin
    GetWindowText(hWnd, aFoundText, MAX_PATH_LENGTH);
    SetWindowText(hWnd, aFoundText);
    Result := false;
  end;
end;

function foMsgHook(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT; stdcall;
var
  OpenDialogEx: TOpenDialogEx;
begin
  Result := 0;
  OpenDialogEx := TOpenDialogEx(GetProp(hWnd, 'TOpenDialogEx'));
  if OpenDialogEx <> nil then Result := CallWindowProc(OpenDialogEx.FMsgProc, hWnd, Msg, wParam, lParam);
  case Msg of
    WM_DESTROY: begin
      if OpenDialogEx <> nil then OpenDialogEx.FHandle := 0;
      BringWindowToTop(GetParent(hWnd));
      RemoveProp(hWnd, 'TOpenDialogEx');
    end;
    WM_SHOWWINDOW: if wParam <> 0 then begin
      if OpenDialogEx <> nil then OpenDialogEx.FHandle := hWnd;
      CenterDialog(hWnd, GetParent(hWnd));
    end;
    WM_COMMAND: if OpenDialogEx <> nil then with OpenDialogEx do begin
      if (lParam = 0) and (HiWord(wParam) = 1) and ((LoWord(wParam) and $DF) = FShortCut) then begin
	Result := 1;
	FFromClipboard := true;
	PostMessage(hWnd, WM_CLOSE, 0, 0);
      end;
    end;
  end;
end;

function foMsgProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT; stdcall;
var
  OpenDialogEx: TOpenDialogEx;
  hParent, hButton: THandle;
  rDialog, rButton, rPanel: TRect;
begin
  Result := 0;
  case Msg of
    WM_DESTROY: RemoveProp(hWnd, 'TOpenDialogEx');
    WM_NOTIFY: if PNMHdr(lParam)^.code = CDN_TYPECHANGE then EnumChildWindows(GetParent(hWnd), @foFakeEdit, 0);
    WM_INITDIALOG: begin
      OpenDialogEx := TOpenDialogEx(pOpenFileName(lParam)^.lCustData);
      SetProp(hWnd, 'TOpenDialogEx', THandle(OpenDialogEx));
      hParent := GetParent(hWnd);
      SetProp(hParent, 'TOpenDialogEx', THandle(OpenDialogEx));
      with OpenDialogEx do begin
	FMsgProc := ptr(SetWindowLong(hParent, GWL_WNDPROC, Longint(@foMsgHook)));
	if FBtnText <> E then begin
	  hButton := GetDlgItem(hWnd, OPEN_CLIP);
	  GetClientRect(hParent, rDialog);
	  GetClientRect(hButton, rButton);
	  GetClientRect(hWnd, rPanel);
	  SetWindowPos(hWnd, 0, 0, 0, rDialog.Right, rPanel.Bottom,
		       SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOREDRAW or SWP_NOZORDER);
	  dec(rDialog.Right, rButton.Right);
	  SetWindowPos(hButton, 0, rDialog.Right div 2, 0, 0, 0,
		       SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOREDRAW or SWP_NOZORDER);
	  SetDlgItemText(hWnd, OPEN_CLIP, PChar(FBtnText));
	end;
      end;
    end;
    WM_COMMAND: begin
      if (lParam <> 0) and (LoWord(wParam) = OPEN_CLIP) and (HiWord(wParam) = BN_CLICKED) then begin
	OpenDialogEx := TOpenDialogEx(GetProp(hWnd, 'TOpenDialogEx'));
	if OpenDialogEx <> nil then OpenDialogEx.FFromClipboard := true;
	PostMessage(GetParent(hWnd), WM_CLOSE, 0, 0);
	Result := 1;
      end;
    end;
  end;
end;

function GetOpenFileName(var OpenFile: TOpenFilename): Bool; stdcall; external 'comdlg32.dll' name 'GetOpenFileNameA';

constructor TOpenDialogEx.Create(AOwner: TComponent);
begin
  inherited Create(nil);
  FillChar(FOpenFileName, SizeOf(TOpenFileName), 0);
  if lo(GetVersion) >= 5 then
    FOpenFileName.lStructSize := SizeOf(TOpenFileName)
  else FOpenFileName.lStructSize := OPENFILENAME_SIZE_VERSION_400;
  FOpenFileName.hInstance := hInstance;
  FFilterList := TStringList.Create;
  FilterIndex := 0;
end;

destructor TOpenDialogEx.Destroy;
begin
  FFilterList.Free;
  inherited;
end;

function TOpenDialogEx.GetFlags: Integer;
begin
  Result:=
    ord(ofAllowMultiSelect in Options) * OFN_ALLOWMULTISELECT or
    ord(ofNoTestFileCreate in Options) * OFN_NOTESTFILECREATE or
    ord(ofShareAware in Options) * OFN_SHAREAWARE or
    ord(ofShowHelp in Options) * OFN_SHOWHELP or
    FOPEN;
end;

procedure TOpenDialogEx.SetBtnText(const Value: String);
var
  iPos: Integer;
begin
  FShortCut := 0;
  FBtnText := Value;
  iPos := succ(FirstCharPos(AM, Value));
  if iPos > 1 then FShortCut := ord(Value[iPos]) and $DF;
end;

procedure TOpenDialogEx.SetDefaultExtension(const Value: String);
begin
  if Value <> E then begin
    if Value[1] = DOT then
      FDefaultExtension := Copy(Value, 2, 3)
    else FDefaultExtension := Copy(Value, 1, 3);
  end
  else FDefaultExtension := E;
end;

function TOpenDialogEx.Execute: Boolean;
var
  sFName, sFilter: String;
  iIndex, iLength: Integer;
  pFilter: PChar;
begin
  try
    sFName := ExtractFileName(FileName);
    SetString(FFileName, nil, MAX_PATH_LENGTH);
    FillChar(Pointer(FFileName)^, MAX_PATH_LENGTH, 0);
    if FileName <> E then Move(Pointer(sFName)^, Pointer(FFileName)^, Length(sFName));
    FFromClipboard := false;
    with FOpenFileName do begin
      hWndOwner := FParentHandle;
      if FFilterList.Count <> 0 then begin
	iLength := 0;
	with FFilterList do for iIndex := 0 to pred(Count) do inc(iLength, succ(Length(Strings[iIndex])));
	SetString(sFilter, nil, iLength);
	pFilter := Pointer(sFilter);
	with FFilterList do for iIndex := 0 to pred(Count) do begin
	  iLength := Length(Strings[iIndex]);
	  System.Move(Pointer(Strings[iIndex])^, pFilter^, iLength);
	  inc(pFilter, iLength);
	  pFilter^ := #0;
	  inc(pFilter);
	end;
	XChgChars(sFilter, '|', #0);
	lpstrFilter := Pointer(sFilter);
      end
      else lpstrFilter := nil;
      lpstrFile := Pointer(FFileName);
      nFilterIndex := FilterIndex;
      nMaxFile := MAX_PATH_LENGTH;
      if FInitialDir = E then FInitialDir := ExtractFilePath(FileName);
      if FInitialDir = E then begin
	SetString(FInitialDir, nil, GetCurrentDirectory(0, nil));
	GetCurrentDirectory(Length(FInitialDir), Pointer(FInitialDir));
      end;
      lpstrInitialDir := Pointer(FInitialDir);
      lpstrTitle := PChar(Title);
      Flags := GetFlags;
      if FDefaultExtension <> E then
	lpstrDefExt := PChar(FDefaultExtension)
      else lpstrDefExt := nil;
      lCustData := Longint(Self);
      lpfnHook := @foMsgProc;
      if FBtnText <> E then
	lpTemplateName := PChar(OPEN_CLIP)
      else lpTemplateName := PChar(OPEN_NORM);
      try
	if GetOpenFileName(FOpenFileName) then begin
	  FInitialDir := GetInsertPath(FFileName);
	  SetCurrentDirectory(PChar(FInitialDir));
	  FilterIndex := nFilterIndex;
	end
	else FFileName := E;
      finally
	ProcessMessages;
      end;
    end;
  except
    Result := false;
    FileName := E;
    FFileName := E;
    Exit;
  end;
  if FFromClipboard then FFileName := CLIP_NAME;
  Result := FFileName <> E;
  FileName := PChar(FFileName);
end;

function fsFakeEdit(hWnd: THandle; lParam: LPARAM): BOOL; stdcall;
var
  aFoundText: Array[0..MAX_PATH_LENGTH] of Char;
begin
  Result := true;
  if (GetClassName(hWnd, aFoundText, MAX_PATH) <> 0) and (UpperCase(aFoundText) = 'EDIT') then begin
    GetWindowText(hWnd, aFoundText, MAX_PATH_LENGTH);
    SetWindowText(hWnd, aFoundText);
    Result := false;
  end;
end;

function fsMsgHook(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT; stdcall;
var
  SaveDialogEx: TSaveDialogEx;
begin
  Result := 0;
  SaveDialogEx := TSaveDialogEx(GetProp(hWnd, 'TSaveDialogEx'));
  if SaveDialogEx <> nil then Result := CallWindowProc(SaveDialogEx.FMsgProc, hWnd, Msg, wParam, lParam);
  case Msg of
    WM_DESTROY: begin
      if SaveDialogEx <> nil then SaveDialogEx.FHandle := 0;
      BringWindowToTop(GetParent(hWnd));
      RemoveProp(hWnd, 'TSaveDialogEx');
    end;
    WM_SHOWWINDOW: if wParam <> 0 then begin
      if SaveDialogEx <> nil then SaveDialogEx.FHandle := hWnd;
      CenterDialog(hWnd, GetParent(hWnd));
    end;
    WM_COMMAND: if SaveDialogEx <> nil then with SaveDialogEx do begin
      if (lParam = 0) and (HiWord(wParam) = 1) and ((LoWord(wParam) and $DF) = FShortCut) then begin
	Result := 1;
	FToClipboard := true;
	PostMessage(hWnd, WM_CLOSE, 0, 0);
      end;
    end;
  end;
end;

function fsMsgProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): UINT; stdcall;
var
  SaveDialogEx: TSaveDialogEx;
  hParent, hButton: THandle;
  rDialog, rButton, rPanel: TRect;
begin
  Result := 0;
  case Msg of
    WM_DESTROY: RemoveProp(hWnd, 'TSaveDialogEx');
    WM_NOTIFY: if PNMHdr(lParam)^.code = CDN_TYPECHANGE then EnumChildWindows(GetParent(hWnd), @fsFakeEdit, 0);
    WM_INITDIALOG: begin
      SaveDialogEx := TSaveDialogEx(pOpenFileName(lParam)^.lCustData);
      SetProp(hWnd, 'TSaveDialogEx', THandle(SaveDialogEx));
      hParent := GetParent(hWnd);
      SetProp(hParent, 'TSaveDialogEx', THandle(SaveDialogEx));
      with SaveDialogEx do begin
	FMsgProc := ptr(SetWindowLong(hParent, GWL_WNDPROC, Longint(@fsMsgHook)));
	if FBtnText <> E then begin
	  hButton := GetDlgItem(hWnd, SAVE_CLIP);
	  GetClientRect(hParent, rDialog);
	  GetClientRect(hButton, rButton);
	  GetClientRect(hWnd, rPanel);
	  SetWindowPos(hWnd, 0, 0, 0, rDialog.Right, rPanel.Bottom,
		       SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOREDRAW or SWP_NOZORDER);
	  dec(rDialog.Right, rButton.Right);
	  SetWindowPos(hButton, 0, rDialog.Right div 2, 0, 0, 0,
		       SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOREDRAW or SWP_NOZORDER);
	  SetDlgItemText(hWnd, SAVE_CLIP, PChar(FBtnText));
	end;
      end;
    end;
    WM_COMMAND: begin
      if (lParam <> 0) and (LoWord(wParam) = SAVE_CLIP) and (HiWord(wParam) = BN_CLICKED) then begin
	SaveDialogEx := TSaveDialogEx(GetProp(hWnd, 'TSaveDialogEx'));
	if SaveDialogEx <> nil then SaveDialogEx.FToClipboard := true;
	PostMessage(GetParent(hWnd), WM_CLOSE, 0, 0);
	Result := 1;
      end;
    end;
  end;
end;

function GetSaveFileName(var OpenFile: TOpenFilename): Bool; stdcall; external 'comdlg32.dll' name 'GetSaveFileNameA';

constructor TSaveDialogEx.Create(AOwner: TComponent);
begin
  inherited Create(nil);
  FillChar(FSaveFileName, SizeOf(TOpenFileName), 0);
  if lo(GetVersion) >= 5 then
    FSaveFileName.lStructSize := SizeOf(TOpenFileName)
  else FSaveFileName.lStructSize := OPENFILENAME_SIZE_VERSION_400;
  FSaveFileName.hInstance := hInstance;
  FFilterList := TStringList.Create;
  FilterIndex := 0;
end;

destructor TSaveDialogEx.Destroy;
begin
  FFilterList.Free;
  inherited;
end;

function TSaveDialogEx.GetFlags: Integer;
begin
  Result:=
    ord(ofNoTestFileCreate in Options) * OFN_NOTESTFILECREATE or
    ord(ofShareAware in Options) * OFN_SHAREAWARE or
    ord(ofShowHelp in Options) * OFN_SHOWHELP or
    FSAVE;
end;

procedure TSaveDialogEx.SetBtnText(const Value: String);
var
  iPos: Integer;
begin
  FShortCut := 0;
  FBtnText := Value;
  iPos := succ(FirstCharPos(AM, Value));
  if iPos > 1 then FShortCut := ord(Value[iPos]) and $DF;
end;

procedure TSaveDialogEx.SetDefaultExtension(const Value: String);
begin
  if Value <> E then begin
    if Value[1] = DOT then
      FDefaultExtension := Copy(Value, 2, 3)
    else FDefaultExtension := Copy(Value, 1, 3);
  end
  else FDefaultExtension := E;
end;

function TSaveDialogEx.Execute: Boolean;
var
  sFName, sFilter: String;
  iIndex, iLength: Integer;
  pFilter: PChar;
begin
  try
    sFName := ExtractFileName(FileName);
    SetString(FFileName, nil, MAX_PATH_LENGTH);
    FillChar(Pointer(FFileName)^, MAX_PATH_LENGTH, 0);
    if FileName <> E then Move(Pointer(sFName)^, Pointer(FFileName)^, Length(sFName));
    fToClipboard := false;
    with FSaveFileName do begin
      hWndOwner := FParentHandle;
      if FFilterList.Count <> 0 then begin
	iLength := 0;
	with FFilterList do for iIndex := 0 to pred(Count) do inc(iLength, succ(Length(Strings[iIndex])));
	SetString(sFilter, nil, iLength);
	pFilter := Pointer(sFilter);
	with FFilterList do for iIndex := 0 to pred(Count) do begin
	  iLength := Length(Strings[iIndex]);
	  System.Move(Pointer(Strings[iIndex])^, pFilter^, iLength);
	  inc(pFilter, iLength);
	  pFilter^ := #0;
	  inc(pFilter);
	end;
	XChgChars(sFilter, '|', #0);
	lpstrFilter := Pointer(sFilter);
      end
      else lpstrFilter := nil;
      lpstrFile := Pointer(FFileName);
      nFilterIndex := FilterIndex;
      nMaxFile := MAX_PATH_LENGTH;
      if FInitialDir = E then FInitialDir := ExtractFilePath(FileName);
      if FInitialDir = E then begin
	SetString(FInitialDir, nil, GetCurrentDirectory(0, nil));
	GetCurrentDirectory(Length(FInitialDir), Pointer(FInitialDir));
      end;
      lpstrInitialDir := Pointer(FInitialDir);
      lpstrTitle := PChar(Title);
      Flags := GetFlags;
      if FDefaultExtension <> E then
	lpstrDefExt := PChar(FDefaultExtension)
      else lpstrDefExt := nil;
      lCustData := Longint(Self);
      lpfnHook := @fsMsgProc;
      if FBtnText <> E then
	lpTemplateName := PChar(SAVE_CLIP)
      else lpTemplateName := PChar(SAVE_NORM);
      try
	if GetSaveFileName(FSaveFileName) then begin
	  FInitialDir := ExtractFilePath(FFileName);
	  SetCurrentDirectory(PChar(FInitialDir));
	  FilterIndex := nFilterIndex;
	end
	else FFileName := E;
      finally
	ProcessMessages;
      end;
    end;
  except
    Result := false;
    FileName := E;
    FFileName := E;
    Exit;
  end;
  if fToClipboard then FFileName := CLIP_NAME;
  Result := FFileName <> E;
  FileName := PChar(FFileName);
end;


end.

