{$J+,Z4}
unit Gauges3d;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, Consts, SysUtils, Classes, Graphics, Forms, Controls, ExtCtrls, UxThemes;

type
  TShape = (bsBox, bsFrame);
  TStyle = (bsLowered, bsRaised);
  TTextKind = (tkNone, tkAbsValue, tkRelValue);
  TGaugeKind = (gkText, gkHorizontalBar, gkVerticalBar, gkPie, gkNeedle);

  TBltBitmap = class(TBitmap)
    procedure MakeLike(ATemplate: TBitmap);
  end;

  TGauge3d = class(TBevel)
  private
    FAbsValue: Longint;
    FBackColor: TColor;
    FBorderWidth: TBorderWidth;
    FCurValue: Longint;
    FForeColor: TColor;
    FHandle: THandle;
    FImage: TBitmap;
    FKind: TGaugeKind;
    FMaxValue: Longint;
    FMinValue: Longint;
    FOverlay: TBltBitmap;
    FPrevValue: Longint;
    FRelValue: Longint;
    FShape: TShape;
    FStyle: TStyle;
    FText: String;
    FTextKind: TTextKind;
    FTextOverlay: TBltBitmap;
    FTheme: THandle;
    procedure OpenTheme;
    procedure CloseTheme;
    procedure PaintBackground(const AnImage: TBitmap);
    procedure PaintAsBar(const AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsNothing(const AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsNeedle(const AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsPie(const AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsText(const AnImage: TBitmap; PaintRect: TRect);
    procedure SetBackColor(Value: TColor);
    procedure SetBorderWidth(Value: TBorderWidth);
    procedure SetForeColor(Value: TColor);
    procedure SetGaugeKind(Value: TGaugeKind);
    procedure SetMaxValue(Value: Longint);
    procedure SetMinValue(Value: Longint);
    procedure SetProgress(Value: Longint);
    procedure SetShape(Value: TShape);
    procedure SetTextKind(Value: TTextKind);
    procedure SetStyle(Value: TStyle);
    procedure SetText(const Value: String);
    function  Evaluate: Longbool;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
  protected
    procedure Paint; override;
    procedure WndMethod(var Message: TMessage); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddProgress(Value: Longint);
    procedure Reset;
  published
    property  Color;
    property  Enabled;
    property  Font;
    property  ParentColor;
    property  BackColor: TColor read FBackColor write SetBackColor;
    property  BorderWidth: TBorderWidth read FBorderWidth write SetBorderWidth;
    property  ForeColor: TColor read FForeColor write SetForeColor;
    property  Kind: TGaugeKind read FKind write SetGaugeKind;
    property  MaxValue: Longint read FMaxValue write SetMaxValue;
    property  MinValue: Longint read FMinValue write SetMinValue;
    property  Progress: Longint read FCurValue write SetProgress;
    property  Shape: TShape read FShape write SetShape;
    property  TextKind: TTextKind read FTextKind write SetTextKind;
    property  Style: TStyle read FStyle write SetStyle;
    property  Text: String read FText write SetText;
  end;

procedure Register;


implementation

// TBltBitmap ------------------------------------------------------------------

procedure TBltBitmap.MakeLike(ATemplate: TBitmap);
begin
  Width := ATemplate.Width;
  Height := ATemplate.Height;
  with Canvas do begin
    Brush.Style := bsSolid;
    Brush.Color := clWindowFrame;
    FillRect(Rect(0, 0, Width, Height));
  end;
end;

// TGauge3d --------------------------------------------------------------------

function SolveForX(Y, Z: Longint): Longint;
begin
  Result := Round(Z * (Y * 0.01));
end;

function SolveForY(X, Z: Longint): Longint;
begin
  if Z = 0 then Result := 0 else Result := Round((X * 100.0) / Z);
end;

procedure TGauge3d.OpenTheme;
begin
  if UseThemes and (FKind in [gkHorizontalBar, gkVerticalBar])
  and ((Color = clWindow) or (Color = clBtnFace)) then begin
    FTheme := OpenThemeData(FHandle, 'PROGRESS');
  end;
end;

procedure TGauge3d.CloseTheme;
begin
  if FTheme <> 0 then begin
    CloseThemeData(FTheme);
    FTheme := 0;
  end;
end;

procedure TGauge3d.PaintBackground(const AnImage: TBitmap);
var
  ARect: TRect;
begin
  with AnImage.Canvas do begin
    ARect := Rect(0, 0, Width, Height);
    Brush.Color := clBlack;
    FillRect(ARect);
  end;
end;

procedure TGauge3d.PaintAsBar(const AnImage: TBitmap; PaintRect: TRect);
var
  FillSize: Longint;
  W, H, O: Integer;
  rChunk: TRect;
begin
  with AnImage.Canvas do begin
    if FTheme = 0 then begin
      Brush.Color := BackColor;
      FillRect(PaintRect);
    end;
    Brush.Color := ForeColor;
    Pen.Color := ForeColor;
    Pen.Width := 1;
    case FKind of
      gkHorizontalBar: begin
	if FTheme <> 0 then begin
	  DrawThemeBackground(FTheme, Handle, PP_BAR, 0, PaintRect, nil);
	  GetThemeBackgroundContentRect(FTheme, Handle, PP_BAR, 0, PaintRect, @rChunk);
	  if not EqualRect(rChunk, PaintRect) then with rChunk do begin
	    W := (PaintRect.Right - (Right - Left)) div 2;
	    H := (PaintRect.Bottom - (Bottom - Top)) div 2;
	    inc(PaintRect.Left, W);
	    inc(PaintRect.Top, H);
	  end;
	  with PaintRect do begin
	    W := Right - Left - FBorderWidth;
	    H := Bottom - Top - FBorderWidth;
	  end;
	  FillSize := SolveForX(FRelValue, W);
	  if FillSize > W then FillSize := W;
	  with PaintRect do begin
	    Right := FillSize;
	    Bottom := H;
	  end;
	  O := FBorderWidth div 2;
	  OffsetRect(PaintRect, O, O);
	  if FillSize > 0 then DrawThemeBackground(FTheme, Handle, PP_CHUNK, 0, PaintRect, nil);
	end
	else begin
	  with PaintRect do begin
	    W := Right - Left - FBorderWidth;
	    H := Bottom - Top - FBorderWidth;
	  end;
	  FillSize := SolveForX(FRelValue, W);
	  if FillSize > W then FillSize := W;
	  if FillSize > 0 then with PaintRect do FillRect(Rect(Left, Top, FillSize, H));
	end;
      end;
      gkVerticalBar: begin
	if FTheme <> 0 then begin
	  DrawThemeBackground(FTheme, Handle, PP_BARVERT, 0, PaintRect, nil);
	  GetThemeBackgroundContentRect(FTheme, Handle, PP_BARVERT, 0, PaintRect, @rChunk);
	  if not EqualRect(rChunk, PaintRect) then with rChunk do begin
	    W := (PaintRect.Right - (Right - Left)) div 2;
	    H := (PaintRect.Bottom - (Bottom - Top)) div 2;
	    inc(PaintRect.Left, W);
	    inc(PaintRect.Top, H);
	  end;
	  with PaintRect do begin
	    W := Right - Left - FBorderWidth;
	    H := Bottom - Top - FBorderWidth;
	  end;
	  FillSize := SolveForX(FRelValue, H);
	  if FillSize >= H then FillSize := H - 1;
	  with PaintRect do begin
	    Top := H - FillSize;
	    Right := W;
	    Bottom := H;
	  end;
	  O := FBorderWidth div 2;
	  OffsetRect(PaintRect, O, O);
	  DrawThemeBackground(FTheme, Handle, PP_CHUNKVERT, 0, PaintRect, nil);
	end
	else begin
	  with PaintRect do begin
	    W := Right - Left - FBorderWidth;
	    H := Bottom - Top - FBorderWidth;
	  end;
	  FillSize := SolveForX(FRelValue, H);
	  if FillSize >= H then FillSize := H - 1;
	  with PaintRect do FillRect(Rect(Left, H - FillSize, W, H));
	end;
      end;
    end;
  end;
end;

procedure TGauge3d.PaintAsNeedle(const AnImage: TBitmap; PaintRect: TRect);
var
  MiddleX: Integer;
  Angle: Double;
  X, Y, W, H: Integer;
begin
  with PaintRect do begin
    X := Left;
    Y := Top;
    W := Right - Left - FBorderWidth;
    H := Bottom - Top - FBorderWidth;
  end;
  with AnImage.Canvas do begin
    Brush.Color := Color;
    FillRect(PaintRect);
    Brush.Color := BackColor;
    Pen.Color := ForeColor;
    Pen.Width := 1;
    with PaintRect do begin
      Pie(X, Y, W, H * 2 - 1, X + W, pred(Bottom), X, pred(Bottom));
      MoveTo(X, Bottom);
      LineTo(X + W, Bottom);
    end;
    if FRelValue > 0 then begin
      Pen.Color := ForeColor;
      MiddleX := Width div 2;
      with PaintRect do begin
	dec(Bottom, 1);
	MoveTo(MiddleX, Bottom);
	Angle := (Pi * ((FRelValue / 100)));
	LineTo(Round(MiddleX * (1 - Cos(Angle))), Round((Bottom) * (1 - Sin(Angle))));
      end;
    end;
  end;
end;

procedure TGauge3d.PaintAsNothing(const AnImage: TBitmap; PaintRect: TRect);
begin
  with AnImage do begin
    Canvas.Brush.Color := BackColor;
    Canvas.FillRect(PaintRect);
  end;
end;

procedure TGauge3d.PaintAsPie(const AnImage: TBitmap; PaintRect: TRect);
var
  MiddleX, MiddleY: Integer;
  Angle: Double;
  W, H: Integer;
begin
  with PaintRect do begin
    W := Right - Left - FBorderWidth;
    H := Bottom - Top - FBorderWidth;
  end;
  with AnImage.Canvas do begin
    Brush.Color := Color;
    FillRect(PaintRect);
    Brush.Color := BackColor;
    Pen.Color := ForeColor;
    Pen.Width := 1;
    with PaintRect do Ellipse(Left, Top, W, H);
    if FRelValue > 0 then begin
      Brush.Color := ForeColor;
      MiddleX := W div 2;
      MiddleY := H div 2;
      Angle := (Pi * ((FRelValue / 50) + 0.5));
      with PaintRect do begin
	Pie(Left, Top, W, H,
	    Round(MiddleX * (1 - Cos(Angle))),
	    Round(MiddleY * (1 - Sin(Angle))),
	    MiddleX, 0);
      end;
    end;
  end;
end;

procedure TGauge3d.PaintAsText(const AnImage: TBitmap; PaintRect: TRect);
var
  S: String;
begin
  case FTextKind of
    tkAbsValue: S := Format('%.4d%s', [FAbsValue, FText]);
    tkRelValue: S := Format('%d%%', [FRelValue]);
  end;
  FTextOverlay.MakeLike(AnImage);
  PaintBackground(FTextOverlay);
  with FTextOverlay.Canvas do begin
    Brush.Style := bsClear;
    Font.Assign(Self.Font);
    Font.Color := clWhite;
    DrawText(Handle, PChar(S), -1, PaintRect, DT_CENTER or DT_NOPREFIX or DT_SINGLELINE or DT_VCENTER)
  end;
  with AnImage.Canvas do begin
    CopyMode := cmSrcInvert;
    Draw(0, 0, FTextOverlay);
  end;
end;

procedure TGauge3d.SetBackColor(Value: TColor);
begin
  if Value <> FBackColor then begin
    FBackColor := Value;
    CloseTheme;
    Refresh;
  end;
end;

procedure TGauge3d.SetBorderWidth(Value: TBorderWidth);
begin
  if Value <> FBorderWidth then begin
    FBorderWidth := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetForeColor(Value: TColor);
begin
  if Value <> FForeColor then begin
    FForeColor := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetGaugeKind(Value: TGaugeKind);
begin
  if Value <> FKind then begin
    FKind := Value;
    CloseTheme;
    Refresh;
  end;
end;

procedure TGauge3d.SetMaxValue(Value: Longint);
begin
  if Value <> FMaxValue then begin
    {$IFDEF CONDITIONALEXPRESSIONS}
    if Value < FMinValue then raise EInvalidOperation.CreateResFmt(@SOutOfRange, [FMinValue, MAXINT]);
    {$ELSE}
    if Value < FMinValue then raise EInvalidOperation.CreateResFmt(SOutOfRange, [FMinValue, MAXINT]);
    {$ENDIF}
    FMaxValue := Value;
    if FCurValue > Value then FCurValue := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetMinValue(Value: Longint);
begin
  if Value <> FMinValue then begin
    {$IFDEF CONDITIONALEXPRESSIONS}
    if Value > FMaxValue then raise EInvalidOperation.CreateResFmt(@SOutOfRange, [0, FMaxValue]);
    {$ELSE}
    if Value < FMinValue then raise EInvalidOperation.CreateResFmt(SOutOfRange, [0, FMaxValue]);
    {$ENDIF}
    FMinValue := Value;
    if FCurValue < Value then FCurValue := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetProgress(Value: Longint);
begin
  if Value < FMinValue then
    Value := FMinValue
  else if Value > FMaxValue then Value := FMaxValue;
  if Value <> FCurValue then begin
    FCurValue := Value;
    if Evaluate then Refresh;
  end;
end;

procedure TGauge3d.SetShape(Value: TShape);
begin
  if Value <> FShape then begin
    FShape := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetTextKind(Value: TTextKind);
begin
  if Value <> FTextKind then begin
    FTextKind := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetStyle(Value: TStyle);
begin
  if Value <> FStyle then begin
    FStyle := Value;
    Refresh;
  end;
end;

procedure TGauge3d.SetText(const Value: String);
begin
  if Value <> FText then FText := Value;
  Refresh;
end;

function TGauge3d.Evaluate: Longbool;
begin
  case FTextKind of
    tkAbsValue: FPrevValue := FAbsValue;
    tkRelValue: FPrevValue := FRelValue;
  else
    FPrevValue := pred(FMinValue);
  end;
  FAbsValue := FCurValue - FMinValue;
  FRelValue := SolveForY(FAbsValue, FMaxValue - FMinValue);
  Result := ((FTextKind = tkAbsValue) and (FAbsValue <> FPrevValue)) or
	    ((FTextKind = tkRelValue) and (FRelValue <> FPrevValue)) or
	    (FTextKind = tkNone) or (FPrevValue = FMinValue);
end;

procedure TGauge3d.CMColorChanged(var Message: TMessage);
begin
  inherited;
  SetBackColor(Color);
  EnableThemes(FHandle, (Color = clBtnFace));
end;

procedure TGauge3d.Paint;
var
  PaintRect: TRect;
begin
  Evaluate;
  with Canvas do begin
    Brush.Color := Color;
    PaintRect := ClientRect;
    FImage.Height := Height;
    FImage.Width := Width;
    PaintBackground(FImage);
    FOverlay.MakeLike(FImage);
    PaintBackground(FOverlay);
    case FKind of
      gkText: PaintAsNothing(FOverlay, PaintRect);
      gkHorizontalBar, gkVerticalBar: PaintAsBar(FOverlay, PaintRect);
      gkPie: PaintAsPie(FOverlay, PaintRect);
      gkNeedle: PaintAsNeedle(FOverlay, PaintRect);
    end;
    if (FTheme = 0) and (FBorderWidth <> 0) then begin
      case FShape of
	bsFrame: begin
	  case FStyle of
	    bsLowered: Frame3D(FOverlay.Canvas, PaintRect, clBtnShadow, clBtnHighlight, FBorderWidth);
	    bsRaised: Frame3D(FOverlay.Canvas, PaintRect, clBtnHighlight, clBtnShadow, FBorderWidth);
	  end;
	end;
	bsBox: Frame3D(FOverlay.Canvas, PaintRect, clWindowFrame, clWindowFrame, FBorderWidth);
      end;
    end;
    with FImage.Canvas do begin
      CopyMode := cmSrcInvert;
      Draw(0, 0, FOverlay);
      CopyMode := cmSrcCopy;
    end;
    if FTextKind <> tkNone then PaintAsText(FImage, PaintRect);
    CopyMode := cmSrcCopy;
    Draw(0, 0, FImage);
  end;
  Application.ProcessMessages;
end;

procedure TGauge3d.WndMethod(var Message: TMessage);
begin
  inherited;
  case Message.Msg of
    WM_THEMECHANGED: begin
      CloseTheme;
      inherited;
      OpenTheme;
      Invalidate;
    end;
  end;
end;

constructor TGauge3d.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FImage := TBitmap.Create;
  FOverlay := TBltBitmap.Create;
  FTextOverlay := TBltBitmap.Create;
  {$IFDEF CONDITIONALEXPRESSIONS}
  FHandle := Classes.AllocateHWnd(WndMethod);
  {$ELSE}
  FHandle := AllocateHWnd(WndMethod);
  {$ENDIF}
  ParentColor := true;
  ControlStyle := ControlStyle + [csFramed, csOpaque];
  if EnableThemes(FHandle, (Color = clBtnFace)) then OpenTheme;
  Height := 100;
  Width := 100;
  FBackColor := clBtnFace;
  FCurValue := 0;
  FForeColor := clHighLight;
  FKind := gkHorizontalBar;
  FMinValue := 0;
  FMaxValue := 100;
  FShape := bsBox;
  FStyle := bsLowered;
  FTextKind := tkNone;
end;

destructor TGauge3d.Destroy;
begin
  CloseTheme;
  FImage.Free;
  FOverlay.Free;
  FTextOverlay.Free;
  {$IFDEF CONDITIONALEXPRESSIONS}
  Classes.DeallocateHWnd(FHandle);
  {$ELSE}
  DeallocateHWnd(FHandle);
  {$ENDIF}
  inherited Destroy;
end;

procedure TGauge3d.AddProgress(Value: Longint);
begin
  SetProgress(FCurValue + Value);
end;

procedure TGauge3d.Reset;
begin
  SetProgress(FMinValue);
end;

procedure Register;
begin
  RegisterComponents('New', [TGauge3d]);
end;

end.

