{$J+,Z4}
unit GdiPlus;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, OLE2, Clipbrd, SysUtils, Graphics, Classes, Dialogs, Printers, PrintDlg, StrUtil, WindowUtils;

type
  TGpStatus = (
    Ok,
    GenericError,
    InvalidParameter,
    OutOfMemory,
    ObjectBusy,
    InsufficientBuffer,
    NotImplemented,
    Win32Error,
    WrongState,
    Aborted,
    FileNotFound,
    ValueOverflow,
    AccessDenied,
    UnknownImageFormat,
    FontFamilyNotFound,
    FontStyleNotFound,
    NotTrueTypeFont,
    UnsupportedGdiplusVersion,
    GdiplusNotInitialized,
    PropertyNotFound,
    PropertyNotSupported
  );

type
  TDebugEventLevel = (
    DebugEventLevelFatal,
    DebugEventLevelWarning
  );

  // Callback function that GDI+ can call, on debug builds, for assertions
  // and warnings.

type
  DebugEventProc = procedure(Level: TDebugEventLevel; Message: PChar); stdcall;

  // Notification functions which the user must call appropriately if
  // "SuppressBackgroundThread" (below) is set.

type
  TGdiplusStartupInput = packed record
    GdiplusVersion: Cardinal;				// Must be 1
    DebugEventCallback: DebugEventProc;			// Ignored on free builds
    SuppressBackgroundThread: BOOL;			// FALSE unless you're prepared to call
							// the hook/unhook functions properly
    SuppressExternalCodecs: BOOL;			// FALSE unless you want GDI+ only to use
							// its internal image codecs.
  end;

type
  NotificationHookProc = function(var Token: ULONG): TGpStatus; stdcall;
  NotificationUnhookProc = procedure(Token: ULONG); stdcall;

type
  TGdiplusStartupOutput = packed record
    // The following 2 fields are NULL if SuppressBackgroundThread is FALSE.
    // Otherwise, they are functions which must be called appropriately to
    // replace the background thread.
    //
    // These should be called on the application's main message loop - i.e.
    // a message loop which is active for the lifetime of GDI+.
    // "NotificationHook" should be called before starting the loop,
    // and "NotificationUnhook" should be called after the loop ends.
    NotificationHook: NotificationHookProc;
    NotificationUnhook: NotificationUnhookProc;
  end;

type
  TImageCodecInfo = packed record
    Clsid: TGUID;
    FormatID: TGUID;
    CodecName: PWCHAR;
    DllName: PWCHAR;
    FormatDescription: PWCHAR;
    FilenameExtension: PWCHAR;
    MimeType: PWCHAR;
    Flags: DWORD;
    Version: DWORD;
    SigCount: DWORD;
    SigSize: DWORD;
    SigPattern: PBYTE;
    SigMask: PBYTE;
  end;
  PImageCodecInfo = ^TImageCodecInfo;

const
  EncoderCompression			= '{e09d739d-ccd4-44ee-8eba-3fbf8be4fc58}';
  EncoderColorDepth			= '{66087055-ad66-4c7c-9a18-38a2310b8337}';
  EncoderScanMethod			= '{3a4e2661-3109-4e56-8536-42c156e7dcfa}';
  EncoderVersion			= '{24d18c76-814a-41a4-bf53-1c219cccf797}';
  EncoderRenderMethod			= '{6d42c53a-229a-4825-8bb7-5c99e2b9a8b8}';
  EncoderQuality			= '{1d5be4b5-fa4a-452d-9cdd-5db35105e7eb}';
  EncoderTransformation			= '{8d0eb2d1-a58e-4ea8-aa14-108074b7b6f9}';
  EncoderLuminanceTable			= '{edb33bce-0266-4a77-b904-27216099e717}';
  EncoderChrominanceTable		= '{f2e455dc-09b3-4316-8260-676ada32481c}';
  EncoderSaveFlag			= '{292266fc-ac40-47bf-8cfc-a85b89a655de}';

  EncoderParameterValueTypeByte		= 1;		// 8-bit unsigned int
  EncoderParameterValueTypeASCII	= 2;		// 8-bit byte containing one 7-bit ASCII
							// code. NULL terminated.
  EncoderParameterValueTypeShort	= 3;		// 16-bit unsigned int
  EncoderParameterValueTypeLong		= 4;		// 32-bit unsigned int
  EncoderParameterValueTypeRational	= 5;		// Two Longs. The first Long is the
							// numerator, the second Long expresses the
							// denomintor.
  EncoderParameterValueTypeLongRange	= 6;    	// Two longs which specify a range of
							// integer values. The first Long specifies
							// the lower end and the second one
							// specifies the higher end. All values
							// are inclusive at both ends
  EncoderParameterValueTypeUndefined	= 7;    	// 8-bit byte that can take any value
							// depending on field definition
  EncoderParameterValueTypeRationalRange= 8;    	// Two Rationals. The first Rational
							// specifies the lower end and the second
							// specifies the higher end. All values
							// are inclusive at both ends

type
  TEncoderValue = (
    EncoderValueColorTypeCMYK,
    EncoderValueColorTypeYCCK,
    EncoderValueCompressionLZW,
    EncoderValueCompressionCCITT3,
    EncoderValueCompressionCCITT4,
    EncoderValueCompressionRle,
    EncoderValueCompressionNone,
    EncoderValueScanMethodInterlaced,
    EncoderValueScanMethodNonInterlaced,
    EncoderValueVersionGif87,
    EncoderValueVersionGif89,
    EncoderValueRenderProgressive,
    EncoderValueRenderNonProgressive,
    EncoderValueTransformRotate90,
    EncoderValueTransformRotate180,
    EncoderValueTransformRotate270,
    EncoderValueTransformFlipHorizontal,
    EncoderValueTransformFlipVertical,
    EncoderValueMultiFrame,
    EncoderValueLastFrame,
    EncoderValueFlush,
    EncoderValueFrameDimensionTime,
    EncoderValueFrameDimensionResolution,
    EncoderValueFrameDimensionPage
  );

type
  TEncoderParameter = packed record
    Guid: TGUID;					// GUID of the parameter
    NumberOfValues: ULONG;				// Number of the parameter values
    Type_: ULONG;					// Value type, like ValueTypeLONG  etc.
    Value: Pointer;					// A pointer to the parameter values
  end;

type
  TEncoderParameters = packed record
    Count: UINT;					// Number of parameters in this structure
    Parameter: Array[0..ord(High(TEncoderValue))] of TEncoderParameter;
  end;
  PEncoderParameters = ^TEncoderParameters;

type
  TGpImage = Pointer;

const
  ImageFormatUndefined			= '{b96b3ca9-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatMemoryBMP			= '{b96b3caa-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatBMP			= '{b96b3cab-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatEMF			= '{b96b3cac-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatWMF			= '{b96b3cad-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatJPEG			= '{b96b3cae-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatPNG			= '{b96b3caf-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatGIF			= '{b96b3cb0-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatTIFF			= '{b96b3cb1-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatEXIF			= '{b96b3cb2-0728-11d3-9d7b-0000f81ef32e}';
  ImageFormatIcon			= '{b96b3cb5-0728-11d3-9d7b-0000f81ef32e}';

type
  TGpBitmap = Pointer;

type
  ARGB = DWORD;

type
  TGpGraphics = Pointer;

type
  TGpUnit = (
    UnitWorld,						// 0 -- World coordinate (non-physical unit)
    UnitDisplay,					// 1 -- Variable -- for PageTransform only
    UnitPixel,						// 2 -- Each unit is one device pixel.
    UnitPoint,						// 3 -- Each unit is a printer's point, or 1/72 inch.
    UnitInch,						// 4 -- Each unit is 1 inch.
    UnitDocument,					// 5 -- Each unit is 1/300 inch.
    UnitMillimeter					// 6 -- Each unit is 1 millimeter.
  );

type
  TGpImageAttributes = Pointer;

type
  TImageAbort				= function: BOOL; stdcall;
  TDrawImageAbort			= TImageAbort;
  TGetThumbnailImageAbort		= TImageAbort;

const
  EmfToWmfBitsFlagsDefault		= $00000000;	// default conversion
  EmfToWmfBitsFlagsEmbedEmf		= $00000001;    // the source EMF metafile is embedded
  							// as a comment in the resulting WMF metafile
  EmfToWmfBitsFlagsIncludePlaceable	= $00000002;	// the resulting WMF metafile is
							// in the placeable metafile format
  EmfToWmfBitsFlagsNoXORClip		= $00000004;    // the clipping region is stored
  							// in the traditional way

type
  TImageType = (itNONE, itBMP, itGIF, itJPEG, itPNG, itTIFF, itWMF);
  TImageTypes = Array[TImageType] of String;
  TMimeTypes = Array[TImageType] of String;
  TJpegQuality = 0..100;

const
  IMAGE_TYPES: TImageTypes = (E, 'BMP', 'GIF', 'JPEG', 'PNG', 'TIFF', 'WMF');
  MIME_TYPES: TMimeTypes = (E, 'image/bmp', 'image/gif', 'image/jpeg', 'image/png',  'image/tiff',  'image/wmf');

type
  TBitmapPrinterData = Record
    PD: TPrinterData;
    Bitmap: TBitmap;
  end;

type
  TImageData = Record
    DataPtr: Pointer;
    DataSize: Integer;
  end;

function CreateImgDataRec(const Data: Pointer; Size: Integer): TImageData;

type
  TIStream = class(TStream)
  private
    FStreamError: HResult;
    FILockBytes: ILockBytes;
    FIStorage: IStorage;
    FIStream: IStream;
  protected
    function GetSize: Longint; {$IFNDEF VER90} reintroduce; overload; {$ENDIF}
    procedure SetSize(Value: Longint); {$IFNDEF VER90} reintroduce; overload; {$ENDIF}
  public
    constructor Create;
    destructor Destroy; override;
    procedure Reset;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    property StreamError: HResult read FStreamError;
    property Stream: IStream read FIStream;
    property Size: Longint read GetSize write SetSize;
  end;

type
  TEncoderIDs = class(TStringList)
  private
    FImageTypes: TImageTypes;
    FMimeTypes: TMimeTypes;
  public
    constructor Create;
    destructor Destroy; override;
    function GetFormatIndex(const FormatDescription: String): TImageType;
    function GetGUIDString(const GUID: TGUID): String;
    function GetGUIDRecord(const GUIDStr: String): TGUID;
    function GetGUIDImageType(GUID: TGUID): TImageType;
    function GetImageTypeGUID(ImageType: TImageType): TGUID;
    function GetImageMimeType(ImageType: TImageType): String;
    function GetImageTypeFileExt(ImageType: TImageType): String;
  end;

type
  TGdipImage = Class
  private
    FEncoderIDs: TEncoderIDs;
    FReadStream: TIStream;
    FWriteStream: TIStream;
    FThumbnail: TBitmap;
    FStatus: TGpStatus;
    FImage: TGdipImage;
    FFromFile: Longbool;
    FImageWidth: Integer;
    FImageHeight: Integer;
    FImageType: TImageType;
    FImageTypeFileExt: String;
    FOriginalFileName: String;
    FToken: ULONG;
    procedure Clear;
    procedure GetImageSizeAndFileExtension;
    function GetImageTypeFromRawFormat: TImageType;
    function GetImageTypeFromFileExtension(const FileName: String): TImageType;
    procedure SetJpegQuality(QualityLevel: ULONG; var Quality: TEncoderParameter; var Parameters: TEncoderParameters);
    function LoadImageFile(const FileName: String): Longbool;
    function LoadImageBuffer(const ImageData: Array of TImageData): Longbool;
    function SaveImageTypeFile(const FileName: String; ImageType: TImageType): Longbool;
    function SaveImageTypeBuffer(var ImageData: String; ImageType: TImageType): Longbool;
  protected
    function ErrorMsg(Code: Integer): String;
    function GetCodecIDs: Integer;
    function GetBitmap: TBitmap;
    function GetRawData: String;
  public
    constructor Create;
    destructor Destroy; override;
    function LoadFromFile(const FileName: String): Longbool;
    function LoadFromStrBuffer(const ImageData: String): Longbool;
    function LoadFromPtrBuffer(ImageData: Pointer; DataSize: Integer): Longbool;
    function LoadResourceBitmap(ModuleHandle: THandle; const ResourceName: String; RtHtml: Longbool): Longbool;
    function SaveToFile(const FileName: String): Longbool;
    function SaveToBuffer(var ImageData: String): Longbool;
    function SaveAsBmpFile(const FileName: String): Longbool;
    function SaveAsBmpBuffer(var ImageData: String): Longbool;
    function SaveAsGifFile(const FileName: String): Longbool;
    function SaveAsGifBuffer(var ImageData: String): Longbool;
    function SaveAsJpegFile(const FileName: String): Longbool;
    function SaveAsJpegBuffer(var ImageData: String): Longbool;
    function SaveAsPngFile(const FileName: String): Longbool;
    function SaveAsPngBuffer(var ImageData: String): Longbool;
    function SaveAsTiffFile(const FileName: String): Longbool;
    function SaveAsTiffBuffer(var ImageData: String): Longbool;
    function GetDIBitstream(var WPixels, HPixels, WTwips, HTwips: Longint): String;
    function GetWmfBitstream(var WPixels, HPixels, WTwips, HTwips: Longint): String;
    function CreateThumbnail(Width, Height: Integer; KeepAspectRatio: Longbool): Longbool;
    function RotateImage(Bitmap: TBitmap; ToLeft: Longbool): Longbool;
    function PrintImage(var BPD: TBitmapPrinterData): Longbool;
    function DrawImage(Bitmap: TBitmap): Longbool;
    function CopyImage: Longbool;
    property ImageWidth: Integer read FImageWidth;
    property ImageHeight: Integer read FImageHeight;
    property ImageType: TImageType read FImageType;
    property ImageTypeFileExt: String read FImageTypeFileExt;
    property RawData: String read GetRawData;
    property OriginalFileName: String read FOriginalFileName write FOriginalFileName;
    property Thumbnail: TBitmap read FThumbnail;
  end;

var
  HGdiPlus: THandle; // since WinXP
  GdiplusStartup: function(var Token: ULONG; const Input: TGdiplusStartupInput;
			   var Output: TGdiplusStartupOutput): TGpStatus; stdcall;
  GdiplusShutdown: procedure(Token: ULONG); stdcall;
  GdipGetImageDecodersSize: function(var numDecoders: UINT; var Size: UINT): TGpStatus; stdcall;
  GdipGetImageDecoders: function(numDecoders: UINT; Size: UINT; Decoders: PImageCodecInfo): TGpStatus; stdcall;
  GdipGetImageEncodersSize: function(var numEncoders: UINT; var Size: UINT): TGpStatus; stdcall;
  GdipGetImageEncoders: function(numEncoders: UINT; Size: UINT; Encoders: PImageCodecInfo): TGpStatus; stdcall;
  GdipLoadImageFromFile: function(Filename: PWCHAR; var Image: TGdipImage): TGpStatus; stdcall;
  GdipLoadImageFromStream: function (Stream: ISTREAM; var Image: TGdipImage): TGpStatus; stdcall;
  GdipSaveImageToFile: function(Image: TGdipImage; Filename: PWCHAR; const clsidEncoder: TGUID;
				EncoderParams: PEncoderParameters): TGpStatus; stdcall;
  GdipSaveImageToStream: function(Image: TGdipImage; Stream: ISTREAM; const clsidEncoder: TGUID;
				  EncoderParams: PEncoderParameters): TGpStatus; stdcall;
  GdipCloneImage: function(Image: TGdipImage; var CloneImage: TGdipImage): TGpStatus; stdcall;
  GdipDisposeImage: function(Image: TGdipImage): TGpStatus; stdcall;
  GdipGetImageRawFormat: function(Image: TGdipImage; var Format: TGUID): TGpStatus; stdcall;
  GdipGetImageWidth: function(Image: TGdipImage; var Width: UINT): TGpStatus; stdcall;
  GdipGetImageHeight: function(Image: TGdipImage; var Height: UINT): TGpStatus; stdcall;
  GdipGetImageThumbnail: function(Image: TGdipImage; ThumbWidth: UINT; ThumbHeight: UINT;
				  var ThumbImage: TGdipImage; Callback: TGetThumbnailImageAbort;
				  CallbackData: Pointer): TGpStatus; stdcall;
  GdipCreateBitmapFromFile: function(Filename: PWCHAR; var Bitmap: TGpBitmap): TGpStatus; stdcall;
  GdipCreateBitmapFromStream: function(Stream: ISTREAM; var Bitmap: TGpBitmap): TGpStatus; stdcall;
  GdipCreateHBITMAPFromBitmap: function(Bitmap: TGpBitmap; var hbmReturn: HBITMAP; Background: ARGB): TGpStatus; stdcall;
  GdipCreateBitmapFromHBITMAP: function(hbm: HBITMAP; hpal: HPALETTE; var Bitmap: TGpBitmap): TGpStatus; stdcall;
  GdipCreateHICONFromBitmap: function(Bitmap: TGpBitmap; var hbmReturn: HICON): TGpStatus; stdcall;
  GdipCreateBitmapFromHICON: function(hicon: HICON; var Bitmap: TGpBitmap): TGpStatus; stdcall;
  GdipCreateFromHDC: function(hdc: THandle; var Graphics: TGpGraphics): TGpStatus; stdcall;
  GdipDeleteGraphics: function(Graphics: TGpGraphics): TGpStatus; stdcall;
  GdipDrawImageRectRect: function(Graphics: TGpGraphics; Image: TGdipImage;
				  dstX: Single; dstY: Single; dstWidth: Single; dstHeight: Single;
				  srcX: Single; srcY: Single; srcWidth: Single; srcHeight: Single;
				  srcUnit: TGpUnit; ImageAttributes: TGpImageAttributes;
				  Callback: TDrawImageAbort; CallbackData: Pointer): TGpStatus; stdcall;
  GdipEmfToWmfBits: function(hemf: HENHMETAFILE; cbData16: UINT; pData16: Pointer;
			     iMapMode: Integer; eFlags: Integer): UINT; stdcall;


function ImagePreviewFileExtensions: String;
function ImageConvertFileExtensions: String;
function IconToBitmap(Icon: THandle; BGColor: ARGB): TBitmap;
function GetIconFromImageFile(const FileName: String): THandle;
function GetIconFromImageBuffer(const ImageData: String): THandle;
function CreateBitmapFileHeader(DIBSize: Cardinal): TBitmapFileHeader;
function CopyImageToClipboard(const ImageFile: String; const ImageData: String): Longbool;
function CaptureClientRect(WindowHandle: THandle; const Target: TRect; ToClipboard: Longbool): TBitmap;
function GetImageDimensions(const FileName: String; ImageData: Pointer; Size: Integer): TSize;
function GetMMDimensions(DC: THandle; const PixelDims: TSize): TSize;
function GetMMImageWidth(DC: THandle; Pixels: Integer): Cardinal;
function GetMMImageHeight(DC: THandle; Pixels: Integer): Cardinal;
function GetTwipsDimensions(DC: THandle; const PixelDims: TSize): TSize;
function GetTwipsImageWidth(DC: THandle; Pixels: Integer): Cardinal;
function GetTwipsImageHeight(DC: THandle; Pixels: Integer): Cardinal;
function GetPixDimsFromTwips(DC: THandle; const TwipsDims: TSize): TSize;
function GetPixWidthFromTwips(DC: THandle; Twps: Integer): Cardinal;
function GetPixHeightFromTwips(DC: THandle; Twps: Integer): Cardinal;


implementation

const
  GDIP = 'GdiPlus.dll';
  TWIPS = 1440;
  POINTS = 72;


// TImageData constructor ------------------------------------------------------
function CreateImgDataRec(const Data: Pointer; Size: Integer): TImageData;
begin
  with Result do begin
    DataPtr := Data;
    DataSize := Size;
  end;
end;


// WideChar handling -----------------------------------------------------------

procedure WideCharAlloc(var Wide: PWideChar; Size: Cardinal);
begin
  Size := succ(Size) shl 1;
  if (Wide = nil) or (StrBufSize(PChar(Wide)) < Size) then begin
    if Wide <> nil then StrDispose(PChar(Wide));
    Wide := PWideChar(StrAlloc(Size));
  end;
end;

procedure WideCharDispose(var Wide: PWideChar);
begin
  if Wide <> nil then StrDispose(PChar(Wide));
  Wide := nil;
end;

function Str2WideChar(const Str: String): PWideChar;
var
  CharCount: Cardinal;
begin
  Result := nil;
  CharCount := Length(Str);
  WideCharAlloc(Result, CharCount);
  if Result <> nil then StringToWideChar(Str, Result, succ(CharCount));
end;


// Basic GdiPlus functions -----------------------------------------------------

function LoadGdiPlus: THandle;
begin
  Result := GetModuleHandle(GDIP);
  if Result = 0 then Result := LoadLibrary(GDIP);
  if Result <> 0 then begin
    GdiplusStartup := GetProcAddress(Result, 'GdiplusStartup');
    GdiplusShutdown := GetProcAddress(Result, 'GdiplusShutdown');
    GdipGetImageDecodersSize := GetProcAddress(Result, 'GdipGetImageDecodersSize');
    GdipGetImageDecoders := GetProcAddress(Result, 'GdipGetImageDecoders');
    GdipGetImageEncodersSize := GetProcAddress(Result, 'GdipGetImageEncodersSize');
    GdipGetImageEncoders := GetProcAddress(Result, 'GdipGetImageEncoders');
    GdipLoadImageFromFile := GetProcAddress(Result, 'GdipLoadImageFromFile');
    GdipLoadImageFromStream := GetProcAddress(Result, 'GdipLoadImageFromStream');
    GdipSaveImageToFile := GetProcAddress(Result, 'GdipSaveImageToFile');
    GdipSaveImageToStream := GetProcAddress(Result, 'GdipSaveImageToStream');
    GdipCloneImage := GetProcAddress(Result, 'GdipCloneImage');
    GdipDisposeImage := GetProcAddress(Result, 'GdipDisposeImage');
    GdipGetImageRawFormat := GetProcAddress(Result, 'GdipGetImageRawFormat');
    GdipGetImageWidth := GetProcAddress(Result, 'GdipGetImageWidth');
    GdipGetImageHeight := GetProcAddress(Result, 'GdipGetImageHeight');
    GdipGetImageThumbnail := GetProcAddress(Result, 'GdipGetImageThumbnail');
    GdipCreateBitmapFromFile := GetProcAddress(Result, 'GdipCreateBitmapFromFile');
    GdipCreateBitmapFromStream := GetProcAddress(Result, 'GdipCreateBitmapFromStream');
    GdipCreateHBITMAPFromBitmap := GetProcAddress(Result, 'GdipCreateHBITMAPFromBitmap');
    GdipCreateBitmapFromHBITMAP := GetProcAddress(Result, 'GdipCreateBitmapFromHBITMAP');
    GdipCreateHICONFromBitmap := GetProcAddress(Result, 'GdipCreateHICONFromBitmap');
    GdipCreateBitmapFromHICON := GetProcAddress(Result, 'GdipCreateBitmapFromHICON');
    GdipCreateFromHDC := GetProcAddress(Result, 'GdipCreateFromHDC');
    GdipDeleteGraphics := GetProcAddress(Result, 'GdipDeleteGraphics');
    GdipDrawImageRectRect := GetProcAddress(Result, 'GdipDrawImageRectRect');
    GdipEmfToWmfBits := GetProcAddress(Result, 'GdipEmfToWmfBits');
  end;
end;

function ImagePreviewFileExtensions: String;
var
  Token: ULONG;
  StartIn: TGdiplusStartupInput;
  StartOut: TGdiplusStartupOutput;
  DecoderCount, Size: UINT;
  ImageCodecStart, ImageCodecInfo: PImageCodecInfo;
  FileExtensions: String;
  pResult: PChar;
begin
  Token := 0;
  Result := E;
  if HGdiPlus <> 0 then begin
    ImageCodecStart := nil;
    FillChar(StartIn, SizeOf(TGdiplusStartupInput), 0);
    StartIn.GdiplusVersion := 1;
    try
      if (GdiplusStartup(Token, StartIn, StartOut) = Ok)
      and (GdipGetImageDecodersSize(DecoderCount, Size) = Ok) and (DecoderCount > 0) then begin
	GetMem(ImageCodecStart, Size);
	if (ImageCodecStart <> nil) and (GdipGetImageDecoders(DecoderCount, Size, ImageCodecStart) = Ok) then begin
	  SetString(Result, nil, Size);
	  pResult := Pointer(Result);
	  ImageCodecInfo := ImageCodecStart;
	  for DecoderCount := DecoderCount downto 1 do with ImageCodecInfo^ do begin
	    FileExtensions := WideCharToString(FilenameExtension);
	    Size := Length(FileExtensions);
	    Move(Pointer(FileExtensions)^, pResult^, Size);
	    inc(pResult, Size);
	    pResult^ := SC;
	    inc(pResult);
	    inc(ImageCodecInfo);
	  end;
	  Delete(Result, pResult - Pointer(Result), MAXINT);
	end;
      end;
    finally
      FreeMem(ImageCodecStart);
      GdiplusShutdown(Token);
    end;
  end;
end;

function ImageConvertFileExtensions: String;
var
  Token: ULONG;
  StartIn: TGdiplusStartupInput;
  StartOut: TGdiplusStartupOutput;
  EncoderCount, Size: UINT;
  ImageCodecStart, ImageCodecInfo: PImageCodecInfo;
  FileExtensions: String;
  pResult: PChar;
begin
  Token := 0;
  Result := E;
  if HGdiPlus <> 0 then begin
    ImageCodecStart := nil;
    FillChar(StartIn, SizeOf(TGdiplusStartupInput), 0);
    StartIn.GdiplusVersion := 1;
    try
      if (GdiplusStartup(Token, StartIn, StartOut) = Ok)
      and (GdipGetImageEncodersSize(EncoderCount, Size) = Ok) and (EncoderCount > 0) then begin
	GetMem(ImageCodecStart, Size);
	if (ImageCodecStart <> nil) and (GdipGetImageEncoders(EncoderCount, Size, ImageCodecStart) = Ok) then begin
	  SetString(Result, nil, Size);
	  pResult := Pointer(Result);
	  ImageCodecInfo := ImageCodecStart;
	  for EncoderCount := EncoderCount downto 1 do with ImageCodecInfo^ do begin
	    FileExtensions := WideCharToString(FilenameExtension);
	    Size := Length(FileExtensions);
	    Move(Pointer(FileExtensions)^, pResult^, Size);
	    inc(pResult, Size);
	    pResult^ := SC;
	    inc(pResult);
	    inc(ImageCodecInfo);
	  end;
	  Delete(Result, pResult - Pointer(Result), MAXINT);
	end;
      end;
    finally
      FreeMem(ImageCodecStart);
      GdiplusShutdown(Token);
    end;
  end;
end;

function IconToBitmap(Icon: THandle; BGColor: ARGB): TBitmap;
var
  Token: ULONG;
  StartIn: TGdiplusStartupInput;
  StartOut: TGdiplusStartupOutput;
  GpBitmap: TGpBitmap;
  Bitmap: THandle;
begin
  Token := 0;
  Result := nil;
  if (Icon <> 0) and (HGdiPlus <> 0) then begin
    FillChar(StartIn, SizeOf(TGdiplusStartupInput), 0);
    StartIn.GdiplusVersion := 1;
    try
      if (GdiplusStartup(Token, StartIn, StartOut) = Ok)
      and (GdipCreateBitmapFromHICON(Icon, GpBitmap) = Ok)
      and (GdipCreateHBITMAPFromBitmap(GpBitmap, HBITMAP(Bitmap), BGColor) = Ok) then begin
        Result := TBitmap.Create;
	if Result = nil then Exit;
	Result.Handle := Bitmap;
      end;
    finally
      GdiplusShutdown(Token);
    end;
  end;
end;

function GetIconFromImageFile(const FileName: String): THandle;
var
  Token: ULONG;
  StartIn: TGdiplusStartupInput;
  StartOut: TGdiplusStartupOutput;
  ImageFileName: PWideChar;
  GpBitmap: TGpBitmap;
begin
  Token := 0;
  Result := 0;
  ImageFileName := nil;
  if (FileName <> E) and (HGdiPlus <> 0) then begin
    FillChar(StartIn, SizeOf(TGdiplusStartupInput), 0);
    StartIn.GdiplusVersion := 1;
    try
      ImageFileName := Str2WideChar(FileName);
      if (GdiplusStartup(Token, StartIn, StartOut) <> Ok)
      or (GdipCreateBitmapFromFile(ImageFileName, GpBitmap) <> Ok)
      or (GdipCreateHICONFromBitmap(GpBitmap, HICON(Result)) <> Ok) then begin
	Result := 0;
      end;
    finally
      WideCharDispose(ImageFileName);
      GdiplusShutdown(Token);
    end;
  end;
end;

function GetIconFromImageBuffer(const ImageData: String): THandle;
var
  Token: ULONG;
  StartIn: TGdiplusStartupInput;
  StartOut: TGdiplusStartupOutput;
  ImageDataStream: TIStream;
  ImageDataSize: Integer;
  GpBitmap: TGpBitmap;
begin
  Token := 0;
  Result := 0;
  ImageDataStream := nil;
  if (ImageData <> E) and (HGdiPlus <> 0) then begin
    FillChar(StartIn, SizeOf(TGdiplusStartupInput), 0);
    StartIn.GdiplusVersion := 1;
    try
      ImageDataSize := Length(ImageData);
      ImageDataStream := TIStream.Create;
      if ImageDataStream <> nil then with ImageDataStream do begin
	Size := ImageDataSize;
	if (Write(Pointer(ImageData), ImageDataSize) <> ImageDataSize)
	or (GdiplusStartup(Token, StartIn, StartOut) <> Ok)
	or (GdipCreateBitmapFromStream(Stream, GpBitmap) <> Ok)
	or (GdipCreateHICONFromBitmap(GpBitmap, HICON(Result)) <> Ok) then begin
	  Result := 0;
	end;
      end;
    finally
      ImageDataStream.Free;
      GdiplusShutdown(Token);
    end;
  end;
end;

function CreateBitmapFileHeader(DIBSize: Cardinal): TBitmapFileHeader;
begin
  with Result do begin
    bfType := $4D42;
    bfSize := DIBSize + SizeOf(TBitmapFileHeader);
    bfReserved1 := 0;
    bfReserved2 := 0;
    bfOffBits := SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader);
  end;
end;

function CopyImageToClipboard(const ImageFile: String; const ImageData: String): Longbool;
var
  gdipi: TGdipImage;
begin
  Result := false;
  gdipi := TGdipImage.Create;
  if gdipi <> nil then with gdipi do begin
    try
      if LoadFromFile(ImageFile) or LoadFromStrBuffer(ImageData) then begin
	Result := CopyImage;
	Exit;
      end;
    finally
      Free;
    end;
  end;
end;

function CaptureClientRect(WindowHandle: THandle; const Target: TRect; ToClipboard: Longbool): TBitmap;
var
  hWindowDC: HDC;
  rClient: TRect;
begin
  Result := nil;
  if WindowHandle = 0 then WindowHandle := GetDesktopWindow;
  if not IsRectEmpty(Target) then
    rClient := Target
  else if not Windows.GetClientRect(WindowHandle, rClient) then Exit;
  hWindowDC := GetDC(WindowHandle);
  if hWindowDC = 0 then Exit;
  try
    Result := TBitmap.Create;
    if Result = nil then Exit;
    with Result, rClient do begin
      Height := Bottom - Top;
      Width := Right - Left;
      if not BitBlt(Canvas.Handle, 0, 0, Width, Height, hWindowDC, Left, Top, SRCCOPY) then begin
	Free;
	Result := nil;
	Exit;
      end;
      if ToClipboard then begin
	Clipboard.Assign(Result);
	Free;
	Result := nil;
	Exit;
      end;
    end;
  finally
    ReleaseDC(WindowHandle, hWindowDC);
  end;
end;

function GetImageDimensions(const FileName: String; ImageData: Pointer; Size: Integer): TSize;
var
  gdipi: TGdipImage;
begin
  with Result do begin
    cx := 0;
    cy := 0;
  end;
  gdipi := TGdipImage.Create;
  if gdipi <> nil then with gdipi do begin
    try
      if LoadFromFile(FileName) or LoadFromPtrBuffer(ImageData, Size) then with Result do begin
	cx := FImageWidth;
	cy := FImageHeight;
      end;
    finally
      Free;
    end;
  end;
end;

function GetMMDimensions(DC: THandle; const PixelDims: TSize): TSize;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result.cx := MulDiv(PixelDims.cx, GetDeviceCaps(DC, HORZSIZE) * 100, GetDeviceCaps(DC, HORZRES));
    Result.cy := MulDiv(PixelDims.cy, GetDeviceCaps(DC, VERTSIZE) * 100, GetDeviceCaps(DC, VERTRES));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetMMImageWidth(DC: THandle; Pixels: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(Pixels, GetDeviceCaps(DC, HORZSIZE) * 100, GetDeviceCaps(DC, HORZRES));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetMMImageHeight(DC: THandle; Pixels: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(Pixels, GetDeviceCaps(DC, VERTSIZE) * 100, GetDeviceCaps(DC, VERTRES));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetTwipsDimensions(DC: THandle; const PixelDims: TSize): TSize;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result.cx := MulDiv(PixelDims.cx, TWIPS, GetDeviceCaps(DC, LOGPIXELSX));
    Result.cy := MulDiv(PixelDims.cy, TWIPS, GetDeviceCaps(DC, LOGPIXELSY));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetTwipsImageWidth(DC: THandle; Pixels: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(Pixels, TWIPS, GetDeviceCaps(DC, LOGPIXELSX));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetTwipsImageHeight(DC: THandle; Pixels: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(Pixels, TWIPS, GetDeviceCaps(DC, LOGPIXELSY));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetPixDimsFromTwips(DC: THandle; const TwipsDims: TSize): TSize;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result.cx := MulDiv(TwipsDims.cx, GetDeviceCaps(DC, LOGPIXELSX), TWIPS);
    Result.cy := MulDiv(TwipsDims.cy, GetDeviceCaps(DC, LOGPIXELSY), TWIPS);
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetPixWidthFromTwips(DC: THandle; Twps: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(Twps, GetDeviceCaps(DC, LOGPIXELSX), TWIPS);
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetPixHeightFromTwips(DC: THandle; Twps: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(Twps, GetDeviceCaps(DC, LOGPIXELSY), TWIPS);
  finally
    ReleaseDC(0, DC);
  end;
end;


// TIStream --------------------------------------------------------------------

constructor TIStream.Create;
begin
  inherited Create;
  FILockBytes := nil;
  FIStorage := nil;
  FIStream := nil;
  try
    FStreamError := CreateILockBytesOnHGlobal(0, true, FILockBytes);
    if FStreamError <> S_OK then Exit;
    FStreamError := StgCreateDocfileOnILockBytes(FILockBytes, STGM_SHARE_EXCLUSIVE or STGM_CREATE or STGM_READWRITE,
						 0, FIStorage);
    if FStreamError <> S_OK then Exit;
    FStreamError := FIStorage.CreateStream('ISTREAM', STGM_SHARE_EXCLUSIVE or STGM_CREATE or STGM_READWRITE or STGM_DIRECT,
					   0, 0, FIStream);
  finally
    if FStreamError <> S_OK then Raise EAbort.Create(E);
  end;
end;

destructor TIStream.Destroy;
begin
  if FIStream <> nil then FIStream.Release;
  if FIStorage <> nil then FIStorage.Release;
  if FILockBytes <> nil then FILockBytes.Release;
  inherited Destroy;
end;

function TIStream.GetSize: Longint;
begin
  Result := inherited Size;
end;

procedure TIStream.SetSize(Value: Longint);
begin
  FIStream.SetSize(Value);
end;

procedure TIStream.Reset;
begin
  Seek(0, soFromBeginning);
  SetSize(0);
end;

function TIStream.Read(var Buffer; Count: Longint): Longint;
begin
  FStreamError := FIStream.Read(Pointer(Buffer), Count, @Result);
  if (FStreamError = S_OK) and (Result <> Count) and (Count <= Size - Position) then FStreamError := STG_E_READFAULT;
end;

function TIStream.Write(const Buffer; Count: Longint): Longint;
begin
  FStreamError := FIStream.Write(Pointer(Buffer), Count, @Result);
  if (FStreamError = S_OK) and (Result <> Count) then FStreamError := STG_E_WRITEFAULT;
end;

function TIStream.Seek(Offset: Longint; Origin: Word): Longint;
var
  DWOrigin: DWord;
  Position: Largeint;
begin
  Result := 0;
  FStreamError := E_INVALIDARG;
  case Origin of
    soFromBeginning: dwOrigin := STREAM_SEEK_SET;
    soFromCurrent: dwOrigin := STREAM_SEEK_CUR;
    soFromEnd: dwOrigin := STREAM_SEEK_END;
  else
    Exit;
  end;
  FStreamError := FIStream.Seek(Offset, DWOrigin, Position);
  if FStreamError = S_OK then Result := Trunc(Position);
end;


// TEncoderIDs -----------------------------------------------------------------

constructor TEncoderIDs.Create;
begin
  inherited Create;
  FImageTypes := IMAGE_TYPES;
  FMimeTypes := MIME_TYPES;
end;

destructor TEncoderIDs.Destroy;
begin
  inherited Destroy;
end;

function TEncoderIDs.GetFormatIndex(const FormatDescription: String): TImageType;
begin
  for Result := High(TImageType) downto Low(TImageType) do begin
    if FImageTypes[Result] = FormatDescription then Break;
  end;
end;

function TEncoderIDs.GetGUIDString(const GUID: TGUID): String;
var
  PWGUID: POleStr;
begin
  Result := E;
  if StringFromCLSID(GUID, PWGUID) = S_OK then begin
    try
      Result := WideCharToString(PWGUID);
    finally
      CoTaskMemFree(PWGUID);
    end;
  end;
end;

function TEncoderIDs.GetGUIDRecord(const GUIDStr: String): TGUID;
var
  StrLen: Cardinal;
  PWGUID: POleStr;
begin
  FillChar(Result, SizeOf(TGUID), 0);
  StrLen := succ(Length(GUIDStr));
  PWGUID := CoTaskMemAlloc(StrLen shl 1);
  if PWGUID <> nil then begin
    try
      StringToWideChar(GUIDStr, PWGUID, StrLen);
      CLSIDFromString(PWGUID, Result);
    finally
      CoTaskMemFree(PWGUID);
    end;
  end;
end;

function TEncoderIDs.GetGUIDImageType(GUID: TGUID): TImageType;
var
  GUIDString: String;
begin
  GUIDString := LowerCase(GetGUIDString(GUID));
  if GUIDString = ImageFormatBMP then
    Result := itBMP
  else if GUIDString = ImageFormatGIF then
    Result := itGIF
  else if GUIDString = ImageFormatJPEG then
    Result := itJPEG
  else if GUIDString = ImageFormatPNG then
    Result := itPNG
  else if GUIDString = ImageFormatTIFF then
    Result := itTIFF
  else Result := itNONE;
end;

function TEncoderIDs.GetImageTypeGUID(ImageType: TImageType): TGUID;
begin
  Result := GetGUIDRecord(Values[FImageTypes[ImageType]]);
end;

function TEncoderIDs.GetImageMimeType(ImageType: TImageType): String;
begin
  Result := FMimeTypes[ImageType];
end;

function TEncoderIDs.GetImageTypeFileExt(ImageType: TImageType): String;
begin
  Result := '.' + LowerCase(FImageTypes[ImageType]);
end;


// TGdipImage --------------------------------------------------------------------

constructor TGdipImage.Create;
var
  StartInput: TGdiplusStartupInput;
  StartOutput: TGdiplusStartupOutput;
begin
  inherited Create;
  try
    if HGdiPlus <> 0 then begin
      FEncoderIDs := TEncoderIDs.Create;
      FReadStream := TIStream.Create;
      FWriteStream := TIStream.Create;
      FThumbnail := TBitmap.Create;
      FillChar(StartInput, SizeOf(TGdiplusStartupInput), 0);
      StartInput.GdiplusVersion := 1;
      FStatus := GdiplusStartup(FToken, StartInput, StartOutput);
      if FStatus = Ok then
	GetCodecIDs
      else Raise EAbort.Create(E);
    end
    else Raise EAbort.Create(E);
  except
    Free;
  end;
end;

destructor TGdipImage.Destroy;
begin
  if FImage <> nil then begin
    GdipDisposeImage(FImage);
    FImage := nil;
  end;
  if FToken <> 0 then begin
    GdiplusShutdown(FToken);
    FToken := 0;
  end;
  FThumbnail.Free;
  FWriteStream.Free;
  FReadStream.Free;
  FEncoderIDs.Free;
  inherited Destroy;
end;

procedure TGdipImage.Clear;
begin
  if FImage <> nil then begin
    GdipDisposeImage(FImage);
    FImage := nil;
  end;
end;

procedure TGdipImage.GetImageSizeAndFileExtension;
begin
  GdipGetImageWidth(FImage, UINT(FImageWidth));
  GdipGetImageHeight(FImage, UINT(FImageHeight));
  FImageTypeFileExt := FEncoderIDs.GetImageTypeFileExt(GetImageTypeFromRawFormat);
end;

function TGdipImage.GetImageTypeFromRawFormat: TImageType;
var
  FormatID: TGUID;
begin
  Result := itNONE;
  FStatus := GdipGetImageRawFormat(FImage, FormatID);
  if FStatus = Ok then Result := FEncoderIDs.GetGUIDImageType(FormatID);
  if Result = itNONE then FStatus := UnknownImageFormat;
end;

function TGdipImage.GetImageTypeFromFileExtension(const FileName: String): TImageType;
var
  Index: Integer;
begin
  Result := itNONE;
  Index := Length(FileName);
  while (Index > 0) and (FileName[Index] <> '.') do dec(Index);
  if Index = 0 then Exit;
  Index := FEncoderIDs.IndexOfName(Copy(FileName, succ(Index), MAXINT));
  if Index = -1 then Exit;
  Index := Integer(FEncoderIDs.Objects[Index]);
  if Index = -1 then Exit;
  Result := TImageType(Index);
end;

procedure TGdipImage.SetJpegQuality(QualityLevel: ULONG; var Quality: TEncoderParameter; var Parameters: TEncoderParameters);
begin
  with Quality do begin
    Guid := FEncoderIDs.GetGUIDRecord(EncoderQuality);
    NumberOfValues := 1;
    Type_ := EncoderParameterValueTypeLong;
    Value := @QualityLevel;
  end;
  with Parameters do begin
    Count := 1;
    Parameter[0] := Quality;
  end;
end;

function TGdipImage.LoadImageFile(const FileName: String): Longbool;
var
  NewFileName: PWideChar;
begin
  try
    NewFileName := Str2WideChar(FileName);
    if NewFileName <> nil then begin
      try
	Clear;
	FStatus := GdipLoadImageFromFile(NewFileName, FImage);
	if FStatus = Ok then GetImageSizeAndFileExtension;
      finally
	WideCharDispose(NewFileName);
      end;
    end
    else FStatus := OutOfMemory;
  finally
    Result := (FStatus = Ok);
  end;
end;

function TGdipImage.LoadImageBuffer(const ImageData: Array of TImageData): Longbool;
var
  Index, Bytes: Integer;
begin
  Result := false;
  try
    FReadStream.Reset;
    with FReadStream do begin
      Bytes := 0;
      for Index := Low(ImageData) to High(ImageData) do inc(Bytes, ImageData[Index].DataSize);
      Size := Bytes;
      Bytes := 0;
      for Index := Low(ImageData) to High(ImageData) do with ImageData[Index] do if DataSize <> 0 then inc(Bytes, Write(DataPtr, DataSize));
      if Bytes <> Size then begin
	Clear;
	Exit;
      end;
    end;
    Clear;
    FStatus := GdipLoadImageFromStream(FReadStream.Stream, FImage);
    if FStatus = Ok then GetImageSizeAndFileExtension;
  finally
    if FReadStream.StreamError <> S_OK then FStatus := Win32Error;
    Result := (FStatus = Ok);
  end;
end;

function TGdipImage.SaveImageTypeFile(const FileName: String; ImageType: TImageType): Longbool;
var
  NewFileName: PWideChar;
  Quality: TEncoderParameter;
  Parameters: TEncoderParameters;
  PParameters: PEncoderParameters;
begin
  try
    NewFileName := Str2WideChar(FileName);
    if NewFileName <> nil then begin
      try
	PParameters := nil;
	if ImageType = itJPEG then begin
	  SetJpegQuality(100, Quality, Parameters);
	  PParameters := @Parameters;
	end;
	FStatus := GdipSaveImageToFile(FImage, NewFileName, FEncoderIDs.GetImageTypeGUID(ImageType), PParameters);
      finally
	WideCharDispose(NewFileName);
      end;
    end
    else FStatus := OutOfMemory;
  finally
    Result := (FStatus = Ok);
  end;
end;

function TGdipImage.SaveImageTypeBuffer(var ImageData: String; ImageType: TImageType): Longbool;
var
  Quality: TEncoderParameter;
  Parameters: TEncoderParameters;
  PParameters: PEncoderParameters;
  DataSize: Longint;
begin
  Result := false;
  try
    PParameters := nil;
    if ImageType = itJPEG then begin
      SetJpegQuality(100, Quality, Parameters);
      PParameters := @Parameters;
    end;
    FWriteStream.Size := 0;
    FStatus := GdipSaveImageToStream(FImage, FWriteStream.Stream, FEncoderIDs.GetImageTypeGUID(ImageType), PParameters);
    if FStatus <> Ok then Exit;
    with FWriteStream do begin
      Position := 0;
      DataSize := Size;
      try
	SetString(ImageData, nil, DataSize);
      except
	FStatus := OutOfMemory;
	Exit;
      end;
      if Read(Pointer(ImageData), DataSize) <> DataSize then Exit;
      Size := 0;
    end;
  finally
    if FWriteStream.StreamError <> S_OK then FStatus := Win32Error;
    Result := (FStatus = Ok);
  end;
end;

function TGdipImage.ErrorMsg(Code: Integer): String;
begin
  case TGpStatus(Code) of
    GenericError		: Result := 'Generic error';
    InvalidParameter		: Result := 'Invalid Parameter';
    OutOfMemory			: Result := 'Out of memory';
    ObjectBusy			: Result := 'Object busy';
    InsufficientBuffer		: Result := 'Insufficient buffer';
    NotImplemented		: Result := 'Not implemented';
    Win32Error			: Result := 'Win32 error';
    WrongState			: Result := 'Wrong state';
    Aborted			: Result := 'Aborted';
    FileNotFound		: Result := 'File not found';
    ValueOverflow		: Result := 'Value overflow';
    AccessDenied		: Result := 'Access denied';
    UnknownImageFormat		: Result := 'Unknown image format';
    FontFamilyNotFound		: Result := 'Font family not found';
    FontStyleNotFound		: Result := 'Font style not found';
    NotTrueTypeFont		: Result := 'Not TrueType font';
    UnsupportedGdiplusVersion	: Result := 'Unsupported GDIPlus version';
    GdiplusNotInitialized	: Result := 'GDIPlus not Initialized';
    PropertyNotFound		: Result := 'Property not found';
    PropertyNotSupported	: Result := 'Property not supported';
  else
    if Code = 0 then
      Result := 'No error'
    else Result := 'Unknown error';
  end;
end;

function TGdipImage.GetCodecIDs: Integer;
var
  EncoderCount, Size: UINT;
  ImageCodecStart, ImageCodecInfo: PImageCodecInfo;
  FormatClsid, FileExtensions, FileExtension: String;
  ImageType: TImageType;
  StrLen, BegPos, EndPos: Integer;
begin
  FStatus := GdipGetImageEncodersSize(EncoderCount, Size);
  if (FStatus = Ok) and (EncoderCount > 0) then begin
    GetMem(ImageCodecStart, Size);
    if ImageCodecStart <> nil then begin
      try
	FStatus := GdipGetImageEncoders(EncoderCount, Size, ImageCodecStart);
	if FStatus = Ok then begin
	  ImageCodecInfo := ImageCodecStart;
	  for EncoderCount := EncoderCount downto 1 do with ImageCodecInfo^ do begin
	    FormatClsid := FEncoderIDs.GetGUIDString(Clsid);
	    ImageType := FEncoderIDs.GetFormatIndex(WideCharToString(FormatDescription));
	    if ImageType > itNONE then begin
	      FileExtensions := WideCharToString(FilenameExtension);
	      StrLen := Length(FileExtensions);
	      BegPos := 1;
	      repeat
		EndPos := NextCharPos(SC, FileExtensions, succ(BegPos));
		if EndPos = 0 then EndPos := succ(StrLen);
		if EndPos > BegPos then begin
		  inc(BegPos, 2);
		  FileExtension := Copy(FileExtensions, BegPos, EndPos - BegPos);
		  if FileExtension <> E then FEncoderIDs.AddObject(FileExtension + '=' + FormatClsid, ptr(ord(ImageType)));
		end;
		BegPos := succ(EndPos);
	      until EndPos > StrLen;
	    end;
	    inc(ImageCodecInfo);
	  end;
	end;
      finally
	FreeMem(ImageCodecStart);
      end;
    end;
  end;
  Result := FEncoderIDs.Count;
end;

function TGdipImage.GetBitmap: TBitmap;
var
  Graphics: TGpGraphics;
begin
  Result := TBitmap.Create;
  if Result <> nil then begin
    with Result do begin
      Width := FImageWidth;
      Height := FImageHeight;
    end;
    FStatus := GdipCreateFromHDC(Result.Canvas.Handle, Graphics);
    if FStatus <> Ok then Exit;
    try
      with Result.Canvas.ClipRect do begin
	FStatus := GdipDrawImageRectRect(Graphics, FImage, Left, Top, Right, Bottom,
					 0, 0, FImageWidth, FImageHeight, UnitPixel, nil, nil, nil);
      end;
    finally
      GdipDeleteGraphics(Graphics);
    end;
  end
  else FStatus := OutOfMemory;
end;

function TGdipImage.GetRawData: String;
var
  FileStream: TFileStream;
  DataSize: Integer;
begin
  Result := E;
  if FFromFile then begin
    try
      FileStream := TFileStream.Create(FOriginalFileName, fmOpenRead or fmShareDenyNone);
      if FileStream <> nil then with FileStream do begin
	try
	  DataSize := Size;
	  if DataSize > 0 then begin
	    try
	      SetString(Result, nil, DataSize);
	    except
	      FStatus := OutOfMemory;
	      Exit;
	    end;
	    if Read(Pointer(Result), DataSize) = DataSize then FStatus := Ok;
	  end;
	finally
	  FileStream.Free;
	end;
      end
      else FStatus := OutOfMemory;
    except
      on EFOpenError do
	FStatus := AccessDenied
      else FStatus := GenericError;
    end;
  end
  else with FReadStream do begin
    Position := 0;
    DataSize := Size;
    try
      SetString(Result, nil, DataSize);
    except
      FStatus := OutOfMemory;
      Exit;
    end;
    if Read(Pointer(Result), DataSize) <> DataSize then FStatus := Win32Error;
  end;
end;

function TGdipImage.LoadFromFile(const FileName: String): Longbool;
begin
  Result := LoadImageFile(FileName);
  if Result then begin
    FOriginalFileName := FileName;
    FReadStream.Size := 0;
    FFromFile := true;
  end;
end;

function TGdipImage.LoadFromStrBuffer(const ImageData: String): Longbool;
var
  ImgData: TImageData;
begin
  ImgData := CreateImgDataRec(Pointer(ImageData), Length(ImageData));
  Result := LoadImageBuffer(ImgData);
  if Result then FFromFile := false;
end;

function TGdipImage.LoadFromPtrBuffer(ImageData: Pointer; DataSize: Integer): Longbool;
var
  ImgData: TImageData;
begin
  ImgData := CreateImgDataRec(ImageData, DataSize);
  Result := LoadImageBuffer(ImgData);
  if Result then FFromFile := false;
end;

function TGdipImage.LoadResourceBitmap(ModuleHandle: THandle; const ResourceName: String; RtHtml: Longbool): Longbool;
type
  TImgDataArray = Array[0..1] of TImageData;
var
  ResourceType: PChar;
  ResourceHandle, DataHandle: THandle;
  ResourceSize: Cardinal;
  DataPointer: Pointer;
  BitmapFileHeader: TBitmapFileHeader;
  ImgDataArray: TImgDataArray;
begin
  Result := false;
  if ResourceName <> E then begin
    if RtHtml then
      ResourceType := RT_HTML
    else ResourceType := RT_BITMAP;
    if ModuleHandle = 0 then ModuleHandle := HInstance;
    ResourceHandle := FindResource(ModuleHandle, PChar(ResourceName), ResourceType);
    if ResourceHandle = 0 then Exit;
    ResourceSize := SizeOfResource(ModuleHandle, ResourceHandle);
    if ResourceSize = 0 then Exit;
    DataHandle := LoadResource(ModuleHandle, ResourceHandle);
    if DataHandle = 0 then Exit;
    DataPointer := LockResource(DataHandle);
    if DataPointer = nil then Exit;
    if not RTHtml then begin
      BitmapFileHeader := CreateBitmapFileHeader(ResourceSize);
      ImgDataArray[0] := CreateImgDataRec(@BitmapFileHeader, SizeOf(TBitmapFileHeader));
      ImgDataArray[1] := CreateImgDataRec(DataPointer, ResourceSize);
      Result := LoadImageBuffer(ImgDataArray);
    end
    else begin
      ImgDataArray[0] := CreateImgDataRec(DataPointer, ResourceSize);
      ImgDataArray[1] := CreateImgDataRec(nil, 0);
      Result := LoadImageBuffer(ImgDataArray);
    end;
    if Result then SetLastError(S_OK);
  end;
end;

function TGdipImage.SaveToFile(const FileName: String): Longbool;
var
  ImageType: TImageType;
begin
  Result := false;
  ImageType := GetImageTypeFromFileExtension(FileName);
  if ImageType <> itNONE then Result := SaveImageTypeFile(FileName, ImageType);
end;

function TGdipImage.SaveToBuffer(var ImageData: String): Longbool;
var
  itRaw: TImageType;
begin
  Result := false;
  itRaw := GetImageTypeFromRawFormat;
  if itRaw <> itNONE then Result := SaveImageTypeBuffer(ImageData, itRaw);
end;

function TGdipImage.SaveAsBmpFile(const FileName: String): Longbool;
begin
  Result := SaveImageTypeFile(FileName, itBMP);
end;

function TGdipImage.SaveAsBmpBuffer(var ImageData: String): Longbool;
begin
  Result := SaveImageTypeBuffer(ImageData, itBMP);
end;

function TGdipImage.SaveAsGifFile(const FileName: String): Longbool;
begin
  Result := SaveImageTypeFile(FileName, itGIF);
end;

function TGdipImage.SaveAsGifBuffer(var ImageData: String): Longbool;
begin
  Result := SaveImageTypeBuffer(ImageData, itGIF);
end;

function TGdipImage.SaveAsJpegFile(const FileName: String): Longbool;
begin
  Result := SaveImageTypeFile(FileName, itJPEG);
end;

function TGdipImage.SaveAsJpegBuffer(var ImageData: String): Longbool;
begin
  Result := SaveImageTypeBuffer(ImageData, itJPEG);
end;

function TGdipImage.SaveAsPngFile(const FileName: String): Longbool;
begin
  Result := SaveImageTypeFile(FileName, itPNG);
end;

function TGdipImage.SaveAsPngBuffer(var ImageData: String): Longbool;
begin
  Result := SaveImageTypeBuffer(ImageData, itPNG);
end;

function TGdipImage.SaveAsTiffFile(const FileName: String): Longbool;
begin
  Result := SaveImageTypeFile(FileName, itTIFF);
end;

function TGdipImage.SaveAsTiffBuffer(var ImageData: String): Longbool;
begin
  Result := SaveImageTypeBuffer(ImageData, itTIFF);
end;

function TGdipImage.GetDIBitstream(var WPixels, HPixels, WTwips, HTwips: Longint): String;
var
  Bitmap: TBitmap;
  Graphics: TGpGraphics;
  HeaderSize, ImageSize: DWord;
  PPixels: PChar;
begin
  Result := E;
  Bitmap := TBitmap.Create;
  if Bitmap <> nil then with Bitmap do begin
    Graphics := nil;
    try
      Width := FImageWidth;
      Height := FImageHeight;
      with Canvas do begin
	FStatus := GdipCreateFromHDC(Handle, Graphics);
	if FStatus <> Ok then Exit;
	with ClipRect do begin
	  FStatus := GdipDrawImageRectRect(Graphics, FImage, Left, Top, Right, Bottom,
					   0, 0, FImageWidth, FImageHeight, UnitPixel, nil, nil, nil);
	end;
	WPixels := FImageWidth;
	HPixels := FImageHeight;
	WTwips := MulDiv(WPixels, TWIPS, GetDeviceCaps(Handle, LOGPIXELSX));
	HTwips := MulDiv(HPixels, TWIPS, GetDeviceCaps(Handle, LOGPIXELSY));
      end;
      if FStatus <> Ok then Exit;
      GetDIBSizes(Handle, HeaderSize, ImageSize);
      if (HeaderSize > 0) and (ImageSize > 0) then begin
	SetString(Result, nil, HeaderSize + ImageSize);
	PPixels := Pointer(Result);
	inc(PPixels, HeaderSize);
	GetDIB(Handle, 0, Pointer(Result)^, PPixels^);
      end;
    finally
      GdipDeleteGraphics(Graphics);
      Bitmap.Free;
    end;
  end
  else FStatus := OutOfMemory;
end;

function TGdipImage.GetWmfBitstream(var WPixels, HPixels, WTwips, HTwips: Longint): String;
var
  Metafile: TMetafile;
  MetafileCanvas: TMetafileCanvas;
  Graphics: TGpGraphics;
begin
  Result := E;
  Metafile := TMetafile.Create;
  if Metafile <> nil then begin
    MetafileCanvas := nil;
    Graphics := nil;
    try
      Metafile.Width := FImageWidth;
      Metafile.Height := FImageHeight;
      MetafileCanvas := TMetafileCanvas.Create(MetaFile, 0);
      if MetafileCanvas <> nil then with MetafileCanvas do begin
	FStatus := GdipCreateFromHDC(Handle, Graphics);
	if FStatus <> Ok then Exit;
	with ClipRect do begin
	  FStatus := GdipDrawImageRectRect(Graphics, FImage, Left, Top, Right, Bottom,
					   0, 0, FImageWidth, FImageHeight, UnitPixel, nil, nil, nil);
	end;
	WPixels := FImageWidth;
	HPixels := FImageHeight;
	WTwips := MulDiv(WPixels, TWIPS, GetDeviceCaps(Handle, LOGPIXELSX));
	HTwips := MulDiv(HPixels, TWIPS, GetDeviceCaps(Handle, LOGPIXELSY));
      end
      else FStatus := OutOfMemory;
      if FStatus <> Ok then Exit;
      SetString(Result, nil, GdipEmfToWmfBits(Metafile.Handle, 0, nil, MM_ANISOTROPIC, EmfToWmfBitsFlagsDefault));
      GdipEmfToWmfBits(Metafile.Handle, Length(Result), Pointer(Result), MM_ANISOTROPIC, EmfToWmfBitsFlagsDefault);
    finally
      GdipDeleteGraphics(Graphics);
      MetafileCanvas.Free;
      MetaFile.Free;
    end;
  end
  else FStatus := OutOfMemory;
end;

function TGdipImage.CreateThumbnail(Width, Height: Integer; KeepAspectRatio: Longbool): Longbool;
var
  ThumbImage: TGdipImage;
  Graphics: TGpGraphics;
begin
  Result := false;
  if FThumbnail <> nil then begin
    ThumbImage := nil;
    Graphics := nil;
    try
      if KeepAspectRatio and (Width > FImageWidth) and (Height > FImageHeight) then begin
	FThumbnail.Width := FImageWidth;
	FThumbnail.Height := FImageHeight;
      end
      else begin
	if Width > Height then begin
	  FThumbnail.Height := Height;
	  if KeepAspectRatio then
	    FThumbnail.Width := MulDiv(FImageWidth, Height, FImageHeight)
	  else FThumbnail.Width := Width;
	end
	else begin
	  FThumbnail.Width := Width;
	  if KeepAspectRatio then
	    FThumbnail.Height := MulDiv(FImageHeight, Width, FImageWidth)
	  else FThumbnail.Height := Height;
	end;
      end;
      FStatus := GdipGetImageThumbnail(FImage, FThumbnail.Width, FThumbnail.Height, ThumbImage, nil, nil);
      if FStatus <> Ok then Exit;
      FStatus := GdipCreateFromHDC(FThumbnail.Canvas.Handle, Graphics);
      if FStatus <> Ok then Exit;
      with FThumbnail.Canvas.ClipRect do begin
	FStatus := GdipDrawImageRectRect(Graphics, ThumbImage, Left, Top, Right, Bottom,
					 0, 0, FThumbnail.Width, FThumbnail.Height, UnitPixel, nil, nil, nil);
      end;
    finally
      Result := (FStatus = Ok);
      GdipDeleteGraphics(Graphics);
      GdipDisposeImage(ThumbImage);
    end;
  end
  else FStatus := OutOfMemory;
end;

function TGdipImage.RotateImage(Bitmap: TBitmap; ToLeft: Longbool): Longbool;
type
  TDestination = (OrgTopLeft, OrgTopRight, OrgBottomLeft);
var
  DestBitmap: TBitmap;
  DestPoints: Array[TDestination] of TPoint;
begin
  Result := false;
  DestBitmap := TBitmap.Create;
  if DestBitmap <> nil then begin
    try
      with Bitmap do begin
	DestBitmap.Width := Height;
	DestBitmap.Height := Width;
      end;
      with DestBitmap.Canvas.ClipRect do begin
	if ToLeft then begin
	  DestPoints[OrgTopLeft] := Point(Left, Bottom);
	  DestPoints[OrgTopRight] := TopLeft;
	  DestPoints[OrgBottomLeft] := BottomRight;
	end
	else begin
	  DestPoints[OrgTopLeft] := Point(Right, Top);
	  DestPoints[OrgTopRight] := BottomRight;
	  DestPoints[OrgBottomLeft] := TopLeft;
	end;
      end;
      with Bitmap do begin
	if not PlgBlt(DestBitmap.Canvas.Handle, DestPoints, Canvas.Handle, 0, 0, Width, Height, 0, 0, 0) then begin
	  FStatus := Win32Error;
	  Exit;
	end;
      end;
      Bitmap.Assign(DestBitmap);
    finally
      Result := (FStatus = Ok);
      DestBitmap.Free;
    end;
  end
  else FStatus := OutOfMemory;
end;

function TGdipImage.PrintImage(var BPD: TBitmapPrinterData): Longbool;
var
  rPrintMargins: TRect;
  BMPInfo: PBitmapInfo;
  Image: Pointer;
  BMPHandle: HBITMAP;
  InfoSize, ImageSize: DWord;
  DIBWidth, DIBHeight, LogPixX, LogPixY, MMWidth, MMHeight, MaxWidth, MaxHeight, PrintWidth, PrintHeight: Integer;
begin
  Result := false;
  BPD.PD.PrintRange := prAllPages;
  if ExecutePrintDialog(BPD.PD) then begin
    rPrintMargins := BPD.PD.Margins;
    BMPInfo := nil;
    Image := nil;
    try
      BMPHandle := BPD.Bitmap.Handle;
      GetDIBSizes(BMPHandle, InfoSize, ImageSize);
      BMPInfo := AllocMem(InfoSize);
      Image := AllocMem(ImageSize);
      if (BMPInfo <> nil) and (Image <> nil) then with Printer do begin
	Title := BPD.PD.PrintJobName;
	GetDIB(BMPHandle, 0, BMPInfo^, Image^);
	with BMPInfo^.bmiHeader do begin
	  DIBWidth := biWidth;
	  DIBHeight := biHeight;
	end;
	BeginDoc;
	try
	  with Canvas do begin
	    MMWidth := GetDeviceCaps(Handle, HORZSIZE);
	    MMHeight := GetDeviceCaps(Handle, VERTSIZE);
	    LogPixX := GetDeviceCaps(Handle, LOGPIXELSX);
	    LogPixY := GetDeviceCaps(Handle, LOGPIXELSY);
	  end;
	  with rPrintMargins do begin
	    Left := MulDiv(PageWidth, Left, MMWidth);
	    Right := MulDiv(PageWidth, Right, MMWidth);
	    Top := MulDiv(PageHeight, Top, MMHeight);
	    Bottom := MulDiv(PageHeight, Bottom, MMHeight);
	    MaxWidth := PageWidth - Left - Right;
	    MaxHeight := PageHeight - Top - Bottom;
	  end;
	  PrintWidth := MulDiv(DIBWidth, LogPixX, GetDeviceCaps(BPD.Bitmap.Canvas.Handle, LOGPIXELSX));
	  PrintHeight := MulDiv(DIBHeight, LogPixY, GetDeviceCaps(BPD.Bitmap.Canvas.Handle, LOGPIXELSY));
	  if (PrintWidth > MaxWidth) or (PrintHeight > MaxHeight) then begin
	    if MulDiv(PrintWidth, MaxHeight, MaxWidth) > PrintHeight then begin
	      PrintHeight := MulDiv(PrintHeight, MaxWidth, PrintWidth);
	      PrintWidth := MaxWidth;
	    end
	    else begin
	      PrintWidth := MulDiv(PrintWidth, MaxHeight, PrintHeight);
	      PrintHeight := MaxHeight;
	    end;
	  end;
	  with rPrintMargins do begin
	    Left := (PageWidth - PrintWidth) div 2;
	    Top := (PageHeight - PrintHeight) div 2;
	    Result := (StretchDIBits(Canvas.Handle, Left, Top, PrintWidth, PrintHeight, 0, 0,
				     DIBWidth, DIBHeight, Image, BMPInfo^, DIB_RGB_COLORS, SRCCOPY) <> Int(GDI_ERROR));
	  end;
	finally
	  EndDoc;
	end;
      end;
    finally
      FreeMem(Image, ImageSize);
      FreeMem(BMPInfo, InfoSize);
    end;
  end;
end;

function TGdipImage.DrawImage(Bitmap: TBitmap): Longbool;
var
  DestBitmap: TBitmap;
begin
  Result := false;
  DestBitmap := GetBitmap;
  if DestBitmap <> nil then begin
    try
      Bitmap.Assign(DestBitmap);
    finally
      Result := (FStatus = Ok);
      DestBitmap.Free;
    end;
  end
  else FStatus := OutOfMemory;
end;

function TGdipImage.CopyImage: Longbool;
var
  DestBitmap: TBitmap;
begin
  Result := false;
  DestBitmap := GetBitmap;
  if DestBitmap <> nil then begin
    try
      Clipboard.Assign(DestBitmap);
    finally
      Result := (FStatus = Ok);
      DestBitmap.Free;
    end;
  end
  else FStatus := OutOfMemory;
end;

begin
  HGdiPlus := LoadGdiPlus;
end.

