{$J+,Z4}
unit HelpCtrl;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, SysUtils, StrUtil, FileUtils, Messages, Graphics, Classes;

// THelpClass ------------------------------------------------------------------
type
  TConverter = function(const Text: String; InputValue: Integer; var OutputValue: Integer): String of Object;
  THelpClass = Class
    FCaller: THandle;
    FDontClose: Longbool;
    FHelpFile: String;
    FHelpTitle: String;
    FTopicClass: String;
    FTopicChild: String;
    FOldWndProc: Pointer;
    function GetHelpWindow: THandle;
  protected
    procedure SubClassHelpWindow;
    procedure Init; virtual; abstract;
    procedure Cleanup; virtual; abstract;
    property  HelpWindow: THandle read GetHelpWindow;
  public
    constructor Create(Caller: THandle; const HelpFile: String);
    destructor Destroy; override;
    procedure Close(Thread: THandle);
    property  DontClose: Longbool read FDontClose write FDontClose;
    property  HelpTitle: String read FHelpTitle write FHelpTitle;
  end;


// TWinHelp --------------------------------------------------------------------
const
  WH_TOPIC_CLASS = 'MS_WINHELP';
  WH_TOPIC_CHILD = 'MS_WINTOPIC';
type
  TWinHelp = Class(THelpClass)
  private
  protected
    procedure Init; override;
    procedure Cleanup; override;
  public
    function Popup(ID: Integer; Position: TPoint): Longbool;
    function Topic(ID: Integer): Longbool;
  end;


// THtmlHelp -------------------------------------------------------------------
const
  MSHTML = 'mshtml.dll';
  HH_CTRL_OCX = 'hhctrl.ocx';
  HH_TOPIC_CLASS = 'HH PARENT';
  HH_TOPIC_CHILD = 'INTERNET EXPLORER_SERVER';
  HH_LIB_REG_KEY = 'CLSID\{ADB880A6-D8FF-11CF-9377-00AA003B7A11}\InprocServer32';
const
  HH_DISPLAY_TOPIC		= $0000;
  HH_HELP_FINDER		= $0000;  // WinHelp equivalent
  HH_DISPLAY_TOC		= $0001;
  HH_DISPLAY_INDEX		= $0002;
  HH_DISPLAY_SEARCH		= $0003;
  HH_SET_WIN_TYPE		= $0004;
  HH_GET_WIN_TYPE		= $0005;
  HH_GET_WIN_HANDLE		= $0006;
  HH_ENUM_INFO_TYPE		= $0007;  // Get Info type name, call repeatedly to enumerate, -1 at end
  HH_SET_INFO_TYPE		= $0008;  // Add Info type to filter.
  HH_SYNC			= $0009;
  HH_RESERVED1			= $000A;
  HH_RESERVED2			= $000B;
  HH_RESERVED3			= $000C;
  HH_KEYWORD_LOOKUP		= $000D;
  HH_DISPLAY_TEXT_POPUP		= $000E;  // display string resource id or text in a popup window
  HH_HELP_CONTEXT		= $000F;  // display mapped numeric value in dwData
  HH_TP_HELP_CONTEXTMENU	= $0010;  // text popup help, same as WinHelp HELP_CONTEXTMENU
  HH_TP_HELP_WM_HELP		= $0011;  // text popup help, same as WinHelp HELP_WM_HELP
  HH_CLOSE_ALL			= $0012;  // close all windows opened directly or indirectly by the caller
  HH_ALINK_LOOKUP		= $0013;  // ALink version of HH_KEYWORD_LOOKUP
  HH_GET_LAST_ERROR		= $0014;  // not currently implemented // See HHERROR.h
  HH_ENUM_CATEGORY		= $0015;  // Get category name, call repeatedly to enumerate, -1 at end
  HH_ENUM_CATEGORY_IT		= $0016;  // Get category info type members, call repeatedly to enumerate, -1 at end
  HH_RESET_IT_FILTER		= $0017;  // Clear the info type filter of all info types
  HH_SET_INCLUSIVE_FILTER	= $0018;  // set inclusive filtering method for untyped topics to be included in display
  HH_SET_EXCLUSIVE_FILTER	= $0019;  // set exclusive filtering method for untyped topics to be excluded from display
  HH_INITIALIZE			= $001C;  // Initializes the help system
  HH_UNINITIALIZE		= $001D;  // Uninitializes the help system
  HH_SET_QUERYSERVICE		= $001E;  // Set the Host IQueryService interface
  HH_PRETRANSLATEMESSAGE	= $00FD;  // Pumps messages (NULL, NULL, MSG*)
  HH_SET_GLOBAL_PROPERTY	= $00FC;  // Set a global property (NULL, NULL, HH_GPROP) HH_SAFE_DISPLAY_TOPIC		= $0020;  // private addition to the interface for InternetExplorer

type
  THHPopup = Record
    cbStruct: Integer;			  // sizeof this structure
    hinst: THandle;			  // instance handle for string resource
    idString: UINT;			  // string resource id, or text id if pszFile is specified in HtmlHelp call
    pszText: PChar;			  // used if idString is zero
    pt: TPoint;				  // top center of popup window
    clrForeground: COLORREF;		  // use -1 for default
    clrBackground: COLORREF;		  // use -1 for default
    rcMargins: TRect;			  // amount of space between edges of window and text, -1 for each member to ignore
    pszFont: PChar;			  // facename, point size, char set, BOLD ITALIC UNDERLINE
  end;

type
  THtmlHelp = Class(THelpClass)
  private
    FMSHTML: HMODULE;
    FHHLibrary: HMODULE;
    FInitializeCookie: DWORD;
  protected
    procedure Init; override;
    procedure Cleanup; override;
  public
    function PreTranslateMessage(PMsg: PMsg): Longbool;
    function Popup(ID: Integer; const FileName: String; const Position: TPoint): Longbool;
    function Topic(const Name, Window: String): Longbool;
  end;


type
  THelpHandler = Class;


// TTopicList ------------------------------------------------------------------
  TTopicList = Class(TStringList)
  private
    FHelpHandler: THelpHandler;
    function  GetTopic(ID: Integer): String;
    procedure SetTopic(ID: Integer; const Value: String);
  public
    constructor Create(HelpHandler: THelpHandler);
    destructor Destroy; override;
    property  Topics[ID: Integer]: String read GetTopic write SetTopic; default;
  end;


// TPopupList ------------------------------------------------------------------
  TPopupList = Class(TStringList)
  private
    FHelpHandler: THelpHandler;
    FHelpIDMarker: String;
    FRTFHeader: String;
    function  GetValue(const ID: String): String;
    procedure SetValue(const ID, Value: String);
    function  GetTextHeight(const ID: String): Integer;
    procedure SetTextHeight(const ID: String; Value: Integer);
  protected
    procedure SetTextStr(const Value: String); override;
  public
    constructor Create(HelpHandler: THelpHandler);
    destructor Destroy; override;
    property  HelpIDMarker: String read FHelpIDMarker write FHelpIDMarker;
    property  Values[const Name: String]: String read GetValue write SetValue; default;
    property  Heights[const Name: String]: Integer read GetTextHeight write SetTextHeight;
  end;


// THelpHandler ----------------------------------------------------------------
  THelpType = (htNone, htWin, htHtml);
  THelpHandler = Class
  private
    FWinHelp: TWinHelp;
    FHtmlHelp: THtmlHelp;
    FHelpType: THelpType;
    FConverter: TConverter;
    FMaxTableWidth: Integer;
    FPopupFile: String;
    FPopupList: TPopupList;
    FTopicList: TTopicList;
    FTopicWindow: String;
    function GetDontClose: Longbool;
    procedure SetDontClose(Value: Longbool);
    function  GetHelpTitle: String;
    procedure SetHelpTitle(const Value: String);
  public
    constructor Create(Caller: THandle; const HelpFile, PopupFile: String; Converter: TConverter; MaxTableWidth: Integer);
    destructor Destroy; override;
    procedure CloseHelpWindow;
    function  IsHelpWindow(Window: THandle): Longbool;
    function  GetRichTextByID(const ID: Integer): String;
    function  GetRichTextByName(const Name: String): String;
    function  GetTextHeightByID(const ID: Integer): Integer;
    function  GetTextHeightByName(const Name: String): Integer;
    function  PreTranslateHtmlHelpMsg(PMsg: PMsg): Longbool;
    function  Popup(Control: THandle; ID: Integer): Longbool;
    function  Topic(Control: THandle; ID: Integer): Longbool;
    property  DontClose: Longbool read GetDontClose write SetDontClose;
    property  HelpTitle: String read GetHelpTitle write SetHelpTitle;
    property  TopicList: TTopicList read FTopicList write FTopicList;
    property  TopicWindow: String read FTopicWindow write FTopicWindow;
  end;


implementation

function FirstRangeCharPos(Find: Char; const Dest: String; LoPos, HiPos: Integer): Integer; register; assembler;
asm	// EAX = Find, EDX = @Dest, ECX = LoPos
  PUSH	EBX
  OR	EDX,EDX
  JE	@EXIT
  OR	ECX,ECX
  JLE	@EXIT
  MOV	EBX,[EDX - 04h]
  CMP	EBX,HiPos
  JGE	@CONTINUE
  MOV	HiPos,EBX
  @CONTINUE:
  SUB	ECX,HiPos
  JG	@EXIT
  DEC	ECX
  ADD	EDX,HiPos
  @LOOP:
  CMP	[EDX + ECX],AL
  JE	@MATCH
  INC	ECX
  JNE	@LOOP
  @EXIT:
  POP	EBX
  POP	EBP
  XOR	EAX,EAX
  RET	TYPE HiPos
  @MATCH:
  POP	EBX
  INC	ECX
  MOV	EAX,ECX
  ADD	EAX,HiPos
end;

function LastRangeCharPos(Find: Char; const Dest: String; HiPos, LoPos: Integer): Integer; register; assembler;
asm	// EAX = Find, EDX = @Dest, ECX = HiPos
  OR	EDX,EDX
  JE	@EXIT
  CMP	ECX,[EDX - 04h]
  JG	@EXIT
  OR	LoPos,0
  JG	@CONTINUE
  MOV	LoPos,1
  @CONTINUE:
  SUB	ECX,LoPos
  JL	@EXIT
  ADD	EDX,LoPos
  DEC	EDX
  @LOOP:
  CMP	[EDX + ECX],AL
  JE	@MATCH
  DEC	ECX
  JNL	@LOOP
  @EXIT:
  POP	EBP
  XOR	EAX,EAX
  RET	TYPE LoPos
  @MATCH:
  MOV	EAX,ECX
  ADD	EAX,LoPos
end;

function FirstRangeLoStrPos(const Find, Dest: String; LoPos, HiPos: Integer): Integer; register; assembler;
asm	// EAX = @Find, EDX = @Dest, ECX = LoPos
  PUSH	EBX
  OR	EAX,EAX
  JE	@EXIT
  OR	EDX,EDX
  JE	@EXIT
  OR	ECX,ECX
  JLE	@EXIT
  MOV	EBX,[EDX - 04h]
  CMP	EBX,HiPos
  JGE	@CONTINUE
  MOV	HiPos,EBX
  @CONTINUE:
  MOV	EBX,[EAX - 04h]
  SUB	ECX,HiPos
  ADD	EBX,ECX
  CMP	EBX,1
  JG	@EXIT
  CMP	DWORD PTR [EAX - 04h],1
  JE	@CHAR
  PUSH	EDI
  PUSH	ESI
  MOV	EDI,EDX
  MOV	ESI,EAX
  ADD	EDI,HiPos
  MOV	AL,[ESI]
  DEC	EDI
  @LOOP:
  MOV	AH,[EDI + ECX]
  XOR	AH,AL
  JE	@MATCH
  TEST	AL,040h
  JE	@NEXT
  TEST	BYTE PTR [EDI + ECX],040h
  JE	@NEXT
  AND	AH,0DFh
  JE	@MATCH
  @NEXT:
  INC	ECX
  JNE	@LOOP
  @NOT:
  XOR	EAX,EAX
  POP	ESI
  POP	EDI
  POP	EBX
  POP	EBP
  RET	TYPE HiPos
  @MATCH:
  MOV	EBX,[ESI - 04h]
  ADD	EBX,ECX
  CMP	EBX,1
  JG	@NOT
  PUSH	EDI
  PUSH	ECX
  ADD	EDI,ECX
  MOV	EBX,[ESI - 04h]
  MOV	ECX,ESI
  ADD	EDI,EBX
  ADD	ECX,EBX
  NEG	EBX
  @CMP:
  INC	EBX
  JE	@YES
  MOV	AH,[EDI + EBX]
  XOR	AH,[ECX + EBX]
  JE	@CMP
  TEST	BYTE PTR [EDI + EBX],040h
  JE	@BREAK
  TEST	BYTE PTR [ECX + EBX],040h
  JE	@BREAK
  AND	AH,0DFh
  JE	@CMP
  @BREAK:
  POP	ECX
  POP	EDI
  INC	ECX
  JNE	@LOOP
  XOR	EAX,EAX
  POP	ESI
  POP	EDI
  POP	EBX
  POP	EBP
  RET	TYPE HiPos
  @YES:
  POP	ECX
  POP	EDI
  POP	ESI
  POP	EDI
  ADD	ECX,HiPos
  MOV	EAX,ECX
  POP	EBX
  POP	EBP
  RET	TYPE HiPos
  @EXIT:
  POP	EBX
  POP	EBP
  XOR	EAX,EAX
  RET	TYPE HiPos
  @CHAR:
  DEC	ECX
  ADD	EDX,HiPos
  MOV	AL,[EAX]
  @@LOOP:
  MOV	AH,[EDX + ECX]
  XOR	AH,AL
  JE	@@MATCH
  TEST	AL,040h
  JE	@@NEXT
  TEST	BYTE PTR [EDX + ECX],040h
  JE	@@NEXT
  AND	AH,0DFh
  JE	@@MATCH
  @@NEXT:
  INC	ECX
  JNE	@@LOOP
  POP	EBX
  POP	EBP
  XOR	EAX,EAX
  RET	TYPE HiPos
  @@MATCH:
  POP	EBX
  INC	ECX
  MOV	EAX,ECX
  ADD	EAX,HiPos
end;

function LastRangeLoStrPos(const Find, Dest: String; HiPos, LoPos: Integer): Integer; register; assembler;
asm	// EAX = @Find, EDX = @Dest, ECX = HiPos
  PUSH	EBX
  OR	EAX,EAX
  JE	@EXIT
  OR	EDX,EDX
  JE	@EXIT
  OR	LoPos,0
  JG	@CONTINUE
  MOV	LoPos,1
  @CONTINUE:
  CMP	ECX,[EDX - 04h]
  JG	@EXIT
  SUB	ECX,LoPos
  JL	@EXIT
  ADD	EDX,LoPos
  CMP	DWORD PTR [EAX - 04h],1
  JE	@CHAR
  PUSH	EDI
  PUSH	ESI
  MOV	ESI,EAX
  MOV	EBX,[EAX - 04h]
  MOV	EDI,EDX
  MOV	AL,[ESI + EBX - 01h]
  @LOOP:
  MOV	AH,[EDI + ECX - 01h]
  XOR	AH,AL
  JE	@MATCH
  TEST	AL,040h
  JE	@NEXT
  TEST	BYTE PTR [EDI + ECX - 01h],040h
  JE	@NEXT
  AND	AH,0DFh
  JE	@MATCH
  @NEXT:
  DEC	ECX
  JNE	@LOOP
  @NOT:
  XOR	EAX,EAX
  POP	ESI
  POP	EDI
  POP	EBX
  POP	EBP
  RET	TYPE LoPos
  @MATCH:
  MOV	EBX,[ESI - 04h]
  SUB	EBX,ECX
  CMP	EBX,1
  JG	@NOT
  PUSH	EDI
  PUSH	ECX
  ADD	EDI,ECX
  MOV	EBX,[ESI - 04h]
  MOV	ECX,ESI
  ADD	ECX,EBX
  NEG	EBX
  @CMP:
  INC	EBX
  JE	@YES
  MOV	AH,[EDI + EBX - 01h]
  XOR	AH,[ECX + EBX - 01h]
  JE	@CMP
  TEST	BYTE PTR [EDI + EBX - 01h],040h
  JE	@BREAK
  TEST	BYTE PTR [ECX + EBX - 01h],040h
  JE	@BREAK
  AND	AH,0DFh
  JE	@CMP
  @BREAK:
  POP	ECX
  POP	EDI
  DEC	ECX
  JNE	@LOOP
  XOR	EAX,EAX
  POP	ESI
  POP	EDI
  POP	EBX
  POP	EBP
  RET	TYPE LoPos
  @YES:
  POP	ECX
  SUB	ECX,[ESI - 04h]
  ADD	ECX,LoPos
  INC	ECX
  POP	EDI
  POP	ESI
  POP	EDI
  MOV	EAX,ECX
  POP	EBX
  POP	EBP
  RET	TYPE LoPos
  @EXIT:
  POP	EBX
  POP	EBP
  XOR	EAX,EAX
  RET	TYPE LoPos
  @CHAR:
  DEC	EDX
  MOV	AL,[EAX]
  @@LOOP:
  MOV	AH,[EDX + ECX]
  XOR	AH,AL
  JE	@@MATCH
  TEST	AL,040h
  JE	@@NEXT
  TEST	BYTE PTR [EDX + ECX],040h
  JE	@@NEXT
  AND	AH,0DFh
  JE	@@MATCH
  @@NEXT:
  DEC	ECX
  JNL	@@LOOP
  @@EXIT:
  POP	EBX
  POP	EBP
  XOR	EAX,EAX
  RET	TYPE LoPos
  @@MATCH:
  MOV	EAX,LoPos
  ADD	EAX,ECX
  POP	EBX
end;

procedure ProcessMessages;
var
  Msg: TMsg;
begin
  while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) and (Msg.Message <> WM_QUIT) do begin
    TranslateMessage(Msg);
    DispatchMessage(Msg);
  end;
end;

// THelpClass ------------------------------------------------------------------

const
  WINDOW_PROC_PROP = 'WindowProcProp';

constructor THelpClass.Create(Caller: THandle; const HelpFile: String);
begin
  inherited Create;
  Init;
  FCaller := Caller;
  FHelpFile := HelpFile;
end;

destructor THelpClass.Destroy;
begin
  Close(GetCurrentThread);
  Cleanup;
  inherited;
end;

procedure THelpClass.Close(Thread: THandle);
var
  hWindow: THandle;
begin
  hWindow := GetHelpWindow;
  if (Thread = 0) or (Thread = GetWindowThreadProcessId(hWindow, nil)) then PostMessage(hWindow, WM_CLOSE, 0, 0);
end;

function NewWndProc(hHelpWindow: THandle; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  with THelpClass(GetProp(hHelpWindow, WINDOW_PROC_PROP)) do begin
    case Msg of
      WM_DESTROY: begin
        FDontClose := false;
	RemoveProp(hHelpWindow, WINDOW_PROC_PROP);
	SetWindowLong(hHelpWindow, GWL_WNDPROC, Longint(FOldWndProc));
      end;
      WM_CONTEXTMENU: FDontClose := true;
      WM_MENUSELECT: if (lParam = 0) and (HiWord(wParam) = $FFFF) then FDontClose := false;
      WM_KEYDOWN: if (wParam = VK_ESCAPE) and not FDontClose then PostMessage(GetActiveWindow, WM_CLOSE, 0, 0);
    end;
    Result := CallWindowProc(FOldWndProc, hHelpWindow, Msg, wParam, lParam);
  end;
end;

function THelpClass.GetHelpWindow: THandle;
begin
  Result := FindWindow(PChar(FTopicClass), PChar(FHelpTitle));
end;

procedure THelpClass.SubClassHelpWindow;

type
  PChildData = ^TChildData;
  TChildData = Record
    ClassName: String;
    Handle: THandle;
  end;

  function CheckHelpChildren(FoundWindow: THandle; ChildData: PChildData): Bool; stdcall;
  begin
    Result := true;
    if ClassString(FoundWindow) = ChildData^.ClassName then begin
      ChildData^.Handle := FoundWindow;
      SetFocus(ChildData^.Handle);
      Result := false;
    end;
  end;

var
  ChildData: TChildData;
  hWindow: THandle;
begin
  with ChildData do begin
    ClassName := FTopicChild;
    Handle := 0;
  end;
  repeat
    Sleep(0);
    ProcessMessages;
    hWindow := GetHelpWindow;
    if hWindow <> 0 then EnumChildWindows(hWindow, @CheckHelpChildren, LPARAM(@ChildData));
  until IsWindowVisible(hWindow) and (ChildData.Handle <> 0);
  SetProp(ChildData.Handle, WINDOW_PROC_PROP, THandle(Self));
  FOldWndProc := Ptr(SetWindowLong(ChildData.Handle, GWL_WNDPROC, Longint(@NewWndProc)));
end;


// TWinHelp --------------------------------------------------------------------

procedure TWinHelp.Init;
begin
  FTopicClass := WH_TOPIC_CLASS;
  FTopicChild := WH_TOPIC_CHILD;
end;

procedure TWinHelp.Cleanup;
begin
  WinHelp(FCaller, nil, HELP_QUIT, 0);
end;

function TWinHelp.Popup(ID: Integer; Position: TPoint): Longbool;
begin
  with Position do WinHelp(FCaller, nil, HELP_SETPOPUP_POS, Y shl 16 + X);
  Result := WinHelp(FCaller, PChar(FHelpFile), HELP_CONTEXTPOPUP or HELP_TCARD, ID);
end;

function TWinHelp.Topic(ID: Integer): Longbool;
begin
  Result := WinHelp(FCaller, PChar(FHelpFile), HELP_CONTEXT or HELP_TCARD, ID);
  SubClassHelpWindow;
end;


// THtmlHelp -------------------------------------------------------------------

var
  HtmlHelp: function(hwndCaller: HWND; pszFile: PChar; uCommand: UINT; dwData: DWORD): HWND; stdcall;

procedure THtmlHelp.Init;
begin
  FTopicClass := HH_TOPIC_CLASS;
  FTopicChild := HH_TOPIC_CHILD;
  FMSHTML := LoadLibrary(MSHTML);
  FHHLibrary := LoadLibrary(PChar(ExpandEnvironmentStrings(RegString(HKEY_CLASSES_ROOT, HH_LIB_REG_KEY, E))));
  if FHHLibrary <> 0 then begin
    HtmlHelp := GetProcAddress(FHHLibrary, 'HtmlHelpA');
    HtmlHelp(0, nil, HH_INITIALIZE, DWORD(@FInitializeCookie));
  end
  else HtmlHelp := nil;
end;

procedure THtmlHelp.Cleanup;
begin
  if Assigned(HtmlHelp) then begin
    HtmlHelp(0, nil, HH_CLOSE_ALL, 0);
    HtmlHelp(0, nil, HH_UNINITIALIZE, FInitializeCookie);
  end;
end;

function THtmlHelp.PreTranslateMessage(PMsg: PMsg): Longbool;
begin
  Result := (GetActiveWindow = GetHelpWindow) and
	    (
	     (HtmlHelp(0, nil, HH_PRETRANSLATEMESSAGE, DWORD(PMsg)) <> 0) or
	     ((PMsg^.Message >= WM_KEYFIRST) and (PMsg^.Message <= WM_KEYLAST) and (PMsg^.wParam = VK_TAB))
	    );
end;

function THtmlHelp.Popup(ID: Integer; const FileName: String; const Position: TPoint): Longbool;
var
  hhpWindow: THHPopup;
begin
  Result := false;
  if not Assigned(HtmlHelp) then Exit;
  with hhpWindow do begin
    cbStruct := SizeOf(THHPopup);
    hinst := 0;
    idString := ID;
    pszText := E;
    pt := Position;
    clrForeground := ColorToRGB(clInfoText);
    clrBackground := ColorToRGB(clInfoBk);
    rcMargins := Rect(-1, -1, -1, -1);
    pszFont := 'Arial, 8, , ';
  end;
  Result := (HtmlHelp(FCaller, PChar(FHelpFile + '::/' + FileName), HH_DISPLAY_TEXT_POPUP, DWORD(@hhpWindow)) <> 0);
end;

function THtmlHelp.Topic(const Name, Window: String): Longbool;
var
  sWindow: String;
  hWindow: THandle;
  rDesktop, rWindow: TRect;
begin
  Result := false;
  if not Assigned(HtmlHelp) then Exit;
  if Window <> E then sWindow := GT + Window;
  hWindow := HtmlHelp(FCaller, PChar(FHelpFile + sWindow), HH_DISPLAY_TOPIC, DWord(Name));
  Result := (hWindow <> 0);
  if Result and GetWindowRect(GetDesktopWindow, rDesktop) then with rDesktop do begin
    if GetClientRect(hWindow, rWindow)
    and ((rWindow.Right < (Right - Left) shr 1) or (rWindow.Bottom < (Bottom - Top) shr 1)) then begin
      Left := (Right - Left) shr 3;
      Right := Right - (Left shl 1);
      SetWindowPos(hWindow, 0, Left, Top, Right, Bottom, SWP_NOZORDER);
    end;
  end;
  SubClassHelpWindow;
end;


// TTopicList ------------------------------------------------------------------

function TTopicList.GetTopic(ID: Integer): String;
begin
  ID := IndexOfObject(Ptr(ID));
  if ID > -1 then Result := Strings[ID];
end;

procedure TTopicList.SetTopic(ID: Integer; const Value: String);
begin
  AddObject(Value, Ptr(ID));
end;

constructor TTopicList.Create(HelpHandler: THelpHandler);
begin
  inherited Create;
  FHelpHandler := HelpHandler;
end;

destructor TTopicList.Destroy;
begin
  inherited;
end;


// TPopupList ------------------------------------------------------------------

constructor TPopupList.Create(HelpHandler: THelpHandler);
begin
  inherited Create;
  FHelpIDMarker := 'IDH_';
  FHelpHandler := HelpHandler;
end;

destructor TPopupList.Destroy;
begin
  inherited;
end;

procedure TPopupList.SetTextStr(const Value: String);
const
  E		= '';
  LF		= #10;
  CR		= #13;
  QU		= '"';
  EQ		= '=';
  BODY_BEG      = '<body';
  BODY_END	= '</body';
  ID_TAG	= '<p id="';
var
  iLen, iBodyBeg, iBodyEnd, iTag, iBeg, iEnd, iName, iTextHeight: Integer;
  sTag, sText, sName, sConverted: String;
  pText: PChar;
begin
  if (Value <> E) and Assigned(FHelpHandler.FConverter) then begin
    Clear;
    iLen := Length(Value);
    sTag := ID_TAG + FHelpIDMarker;
    iBodyBeg := FirstRangeLoStrPos(BODY_BEG, Value, 1, iLen) + succ(Length(BODY_BEG));
    iBodyEnd := LastRangeLoStrPos(BODY_END, Value, iLen, 1);
    iTag := Length(sTag);
    if iBodyEnd > iBodyBeg then begin
      iBeg := iBodyBeg;
      iEnd := iBodyEnd;
      repeat
	iBeg := FirstRangeLoStrPos(sTag, Value, iBeg, iBodyEnd);
	if iBeg = 0 then Break;
	iEnd := FirstRangeLoStrPos(sTag, Value, iBeg + iTag, iBodyEnd);
	if iEnd = 0 then iEnd := iBodyEnd;
	sText := Copy(Value, iBeg, iEnd - iBeg);
	if sText = E then Break;
	inc(iBeg, Length(ID_TAG));
	sName := Copy(Value, iBeg, FirstRangeCharPos(QU, Value, iBeg, iEnd) - iBeg);
	if sName = E then Break;
	iTextHeight := 0;
	with FHelpHandler do sConverted := FConverter(sText, FMaxTableWidth, iTextHeight);
	if sConverted <> E then begin
	  iLen := Length(sConverted);
	  if iLen > 0 then begin
	    iName := Length(sName);
	    SetString(sText, nil, iName + Length(EQ) + iLen);
	    pText := Pointer(sText);
	    System.Move(Pointer(sName)^, pText^, iName);
	    inc(pText, iName);
	    pText^ := EQ;
	    inc(pText);
	    System.Move(Pointer(sConverted)^, pText^, iLen);
	  end;
	  AddObject(sText, TObject(iTextHeight));
	end;
	iBeg := iEnd;
      until iEnd = iBodyEnd;
    end;
  end;
end;

function TPopupList.GetValue(const ID: String): String;
begin
  Result := FRtfHeader + inherited Values[ID];
end;

procedure TPopupList.SetValue(const ID, Value: String);
begin
  inherited Values[ID] := Value;
end;

function TPopupList.GetTextHeight(const ID: String): Integer;
var
  iIndex: Integer;
begin
  iIndex := IndexOfName(ID);
  if (iIndex > -1) and (iIndex < Count) then begin
    Result := Integer(Objects[iIndex]);
    Exit;
  end;
  Result := 0;
end;

procedure TPopupList.SetTextHeight(const ID: String; Value: Integer);
var
  iIndex: Integer;
begin
  iIndex := IndexOfName(ID);
  if (iIndex > -1) and (iIndex < Count) then Objects[iIndex] := TObject(Value);
end;


// THelpHandler ----------------------------------------------------------------

constructor THelpHandler.Create(Caller: THandle; const HelpFile, PopupFile: String;
				Converter: TConverter; MaxTableWidth: Integer);
var
  sFileExt: String;
begin
  inherited Create;
  FConverter := Converter;
  FMaxTableWidth := MaxTableWidth;
  FPopupList := TPopupList.Create(Self);
  FTopicList := TTopicList.Create(Self);
  sFileExt := LowerCase(ExtractFileExt(HelpFile));
  if sFileExt = '.chm' then begin
    FHtmlHelp := THtmlHelp.Create(Caller, HelpFile);
    if FHtmlHelp <> nil then FHelpType := htHtml;
  end
  else if sFileExt = '.hlp' then begin
    FWinHelp := TWinHelp.Create(Caller, HelpFile);
    if FWinHelp <> nil then FHelpType := htWin;
  end;
  if FileExists(PopupFile) and (LowerCase(ExtractFileExt(PopupFile)) = '.htm') then FPopupList.LoadFromFile(PopupFile);
end;

destructor THelpHandler.Destroy;
begin
  FPopupList.Free;
  FTopicList.Free;
  FHtmlHelp.Free;
  FWinHelp.Free;
  inherited;
end;

procedure THelpHandler.CloseHelpWindow;
begin
  case FHelpType of
    htWin: FWinHelp.Close(0);
    htHtml: FHtmlHelp.Close(0);
  end;
end;

function THelpHandler.IsHelpWindow(Window: THandle): Longbool;
var
  hTopic: THandle;
begin
  Result := false;
  if Window = 0 then Exit;
  hTopic := 0;
  case FHelpType of
    htWin: hTopic := FWinHelp.HelpWindow;
    htHtml: hTopic := FHtmlHelp.HelpWindow;
  end;
  if hTopic = 0 then Exit;
  Result := (Window = hTopic) or (GetParent(Window) = hTopic) or IsChild(hTopic, Window);
end;

function THelpHandler.GetRichTextByID(const ID: Integer): String;
begin
  with FPopupList do Result := Values[Names[ID]];
end;

function THelpHandler.GetRichTextByName(const Name: String): String;
begin
  with FPopupList do Result := Values[Name];
end;

function THelpHandler.GetTextHeightByID(const ID: Integer): Integer;
begin
  with FPopupList do Result := Heights[Names[ID]];
end;

function THelpHandler.GetTextHeightByName(const Name: String): Integer;
begin
  with FPopupList do Result := Heights[Name];
end;

function THelpHandler.PreTranslateHtmlHelpMsg(PMsg: PMsg): Longbool;
begin
  if FHtmlHelp <> nil then
    Result := FHtmlHelp.PreTranslateMessage(PMsg)
  else Result := false;
end;

function THelpHandler.Popup(Control: THandle; ID: Integer): Longbool;
var
  ptScreen, ptClient: TPoint;
  rClient, rScreen: TRect;
begin
  Result := false;
  GetCursorPos(ptScreen);
  ptClient := ptScreen;
  GetClientRect(Control, rClient);
  ScreenToClient(Control, ptClient);
  if not PtInRect(rClient, ptClient) then begin
    GetWindowRect(Control, rScreen);
    with rScreen do ptScreen := Point(Left + (Right - Left) div 2, Top + (Bottom - Top) div 2 + 4);
  end;
  case FHelpType of
    htWin: Result := FWinHelp.Popup(ID, ptScreen);
    htHtml: Result := FHtmlHelp.Popup(ID, FPopupFile, ptScreen);
  end;
end;

function THelpHandler.Topic(Control: THandle; ID: Integer): Longbool;
begin
  Result := false;
  case FHelpType of
    htWin: Result := FWinHelp.Topic(ID);
    htHtml: if (FTopicList <> nil) then Result := FHtmlHelp.Topic(FTopicList[ID], FTopicWindow);
  end;
end;

function THelpHandler.GetDontClose: Longbool;
begin
  Result := false;
  case FHelpType of
    htNone: Exit;
    htWin: Result := FWinHelp.DontClose;
    htHtml: Result := FHtmlHelp.DontClose;
  end;
end;

procedure THelpHandler.SetDontClose(Value: Longbool);
begin
  case FHelpType of
    htNone: Exit;
    htWin: FWinHelp.DontClose := Value;
    htHtml: FHtmlHelp.DontClose := Value;
  end;
end;

function THelpHandler.GetHelpTitle: String;
begin
  case FHelpType of
    htNone: Exit;
    htWin: Result := FWinHelp.HelpTitle;
    htHtml: Result := FHtmlHelp.HelpTitle;
  end;
end;

procedure THelpHandler.SetHelpTitle(const Value: String);
begin
  case FHelpType of
    htNone: Exit;
    htWin: FWinHelp.HelpTitle := Value;
    htHtml: FHtmlHelp.HelpTitle := Value;
  end;
end;


end.

