{$J+,Z4}
unit HotLabel;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls;

const // Cursor IDs
  crHand		= 1;
  IDC_HAND		= 32649;
  IDW_CURSOR_HAND	= IDC_HAND;

type
  THotLabel = Class;
  TVerticalAlignment = (vaTop, vaBottom, vaCenter);
  TOnTextHover = procedure(Sender: THotLabel; Active: Longbool) of Object;
  THotLabel = class(TCustomLabel)
  private
    FAutoSize: Boolean;
    FCaption: String;
    FHorizontalMargin: Integer;
    FHovering: Longbool;
    FHotColor: TColor;
    FLinkColor: TColor;
    FLinkLabel: Longbool;
    FVerticalAlignment: TVerticalAlignment;
    FUnicodeSystem: Longbool;
    FOnTextHover: TOnTextHover;
    procedure AdjustLinkFont;
    function  GetMinimumWidth: Integer;
    function  GetTextSize: TSize;
    function  GetCaption: String;
    procedure SetCaption(const Value: String);
    procedure SetHorizontalMargin(Value: Integer);
    procedure SetLinkLabel(Value: Longbool);
    procedure SetVerticalAlignment(Value: TVerticalAlignment);
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMRButtonUp(var Message: TWMMouse); message WM_RBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouse); message WM_MOUSEMOVE;
  protected
    procedure SetAutoSize(Value: Boolean); override;
    procedure Click; override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    property  MinimumWidth: Integer read GetMinimumWidth;
    property  Hovering: Longbool read FHovering;
  published
    property  Align;
    property  Alignment;
    property  Color;
    property  DragCursor;
    property  DragMode;
    property  Enabled;
    property  FocusControl;
    property  Font;
    property  ParentColor;
    property  ParentFont;
    property  ParentShowHint;
    property  PopupMenu;
    property  ShowHint;
    property  ShowAccelChar;
    property  Transparent;
    property  Visible;
    property  AutoSize: Boolean read FAutoSize write SetAutoSize;
    property  Caption: String read GetCaption write SetCaption;
    property  HorizontalMargin: Integer read FHorizontalMargin write SetHorizontalMargin;
    property  HotColor: TColor read FHotColor write FHotColor;
    property  LinkColor: TColor read FLinkColor write FLinkColor;
    property  LinkLabel: Longbool read FLinkLabel write SetLinkLabel;
    property  VerticalAlignment: TVerticalAlignment read FVerticalAlignment write SetVerticalAlignment;
    property  OnClick;
    property  OnDblClick;
    property  OnDragDrop;
    property  OnDragOver;
    property  OnEndDrag;
    property  OnMouseDown;
    property  OnMouseMove;
    property  OnMouseUp;
    property  OnStartDrag;
    property  OnTextHover: TOnTextHover read FOnTextHover write FOnTextHover;
  end;

procedure Register;

implementation

{$R *.RES}

const
  E = '';
  NULL = #0;
  dtAlignments: Array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  dtVAlignments: Array[TVerticalAlignment] of Word = (DT_TOP, DT_BOTTOM, DT_VCENTER);

procedure THotLabel.AdjustLinkFont;
begin
  if FLinkLabel then with Font do begin
    Style := Style - [fsBold] + [fsUnderline];
    Color := FLinkColor;
  end;
end;

function THotLabel.GetMinimumWidth: Integer;
begin
  Result := GetTextSize.cx;
  if Alignment = taCenter then inc(Result, FHorizontalMargin * 2);
end;

function THotLabel.GetTextSize: TSize;
var
  iLen, iMax: Integer;
  rClient: TRect;
begin
  with Result do begin
    cx := 0;
    cy := 0;
    if FCaption = E then Exit;
    iLen := Length(FCaption);
    rClient := ClientRect;
    if FCaption[iLen] = NULL then begin
      if FUnicodeSystem then begin
	DrawTextW(Canvas.Handle, Pointer(FCaption), -1, rClient,
		  dtAlignments[Alignment] or dtVAlignments[FVerticalAlignment] or
		  DT_CALCRECT or DT_EXTERNALLEADING or DT_SINGLELINE or
		  (ord(ShowAccelChar = false) * DT_NOPREFIX));
	with rClient do begin
	  cx := Right - Left;
	  cy := Bottom - Top;
	end;
      end
      else GetTextExtentPoint32W(Canvas.Handle, Pointer(FCaption), pred(iLen shr 1), Result);
    end
    else begin
      DrawText(Canvas.Handle, Pointer(FCaption), -1, rClient,
	       dtAlignments[Alignment] or dtVAlignments[FVerticalAlignment] or
	       DT_CALCRECT or DT_EXTERNALLEADING or DT_SINGLELINE or
	       (ord(ShowAccelChar = false) * DT_NOPREFIX));
      with rClient do begin
	cx := Right - Left;
	cy := Bottom - Top;
      end;
    end;
    if Alignment = taCenter then begin
      iMax := Width - FHorizontalMargin * 2;
      if cx > iMax then cx := iMax;
    end;
  end;
end;

function THotLabel.GetCaption: String;
begin
  if (FCaption <> E) and (FCaption[Length(FCaption)] = NULL) then
    Result := WideCharToString(Pointer(FCaption))
  else Result := FCaption;
end;

procedure THotLabel.SetCaption(const Value: String);
var
  iLen: Integer;
begin
  if Value <> E then begin
    iLen := Length(Value);
    if Value[iLen] = NULL then begin
      SetString(FCaption, nil, iLen);
      Move(Pointer(Value)^, Pointer(FCaption)^, iLen);
    end
    else FCaption := Value;
  end
  else FCaption := E;
  Invalidate;
end;

procedure THotLabel.SetHorizontalMargin(Value: Integer);
begin
  if Value < Width div 2 then begin
    FHorizontalMargin := abs(Value);
    Invalidate;
  end;
end;

procedure THotLabel.SetLinkLabel(Value: Longbool);
begin
  if Value <> FLinkLabel then begin
    FLinkLabel := Value;
    AdjustLinkFont;
  end;
end;

procedure THotLabel.SetVerticalAlignment(Value: TVerticalAlignment);
begin
  if Value <> FVerticalAlignment then begin
    FVerticalAlignment := Value;
    Invalidate;
  end;
end;

procedure THotLabel.CMDialogChar(var Message: TCMDialogChar);
begin
  if Enabled and ShowAccelChar and IsAccel(Message.CharCode, GetCaption)
  and (FocusControl <> nil) and FocusControl.CanFocus then begin
    FocusControl.SetFocus;
    Message.Result := 1;
  end
  else inherited;
end;

procedure THotLabel.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FLinkLabel then begin
    AdjustLinkFont;
    Screen.Cursor := crDefault;
  end;
  if FHovering then begin
    FHovering := false;
    if Assigned(FOnTextHover) then FOnTextHover(Self, false);
  end;
end;

procedure THotLabel.WMRButtonUp(var Message: TWMMouse);
var
  ptMouse: TPoint;
  hiLabel: THelpInfo;
begin
  if (PopupMenu = nil) and (FocusControl <> nil) then with FocusControl do begin
    GetCursorPos(ptMouse);
    with hiLabel do begin
      cbSize := SizeOf(THelpInfo);
      iContextType := HELPINFO_WINDOW;
      iCtrlId := GetWindowLong(Handle, GWL_ID);
      hItemHandle := Handle;
      dwContextId := HelpContext;
      MousePos := ptMouse;
    end;
    with TMessage(Message) do begin
      Msg := WM_HELP;
      lParam := Longint(@hiLabel);
    end;
    with ptMouse do DefaultHandler(Message);
  end
  else inherited;
end;

procedure THotLabel.WMMouseMove(var Message: TWMMouse);
var
  rText: TRect;
begin
  inherited;
  rText := ClientRect;
  with rText do begin
    with GetTextSize do begin
      Right := cx;
      Bottom := cy;
    end;
    case Alignment of
      taLeftJustify: begin
	inc(Right, Left);
      end;
      taCenter: begin
	Left := (Width - Right) div 2;
	Right := pred(Width - Left);
      end;
      taRightJustify: begin
	Left := Width - Right;
	Right := Width;
      end;
    end;
    case FVerticalAlignment of
      vaTop: begin
	inc(Bottom, Top);
      end;
      vaCenter: begin
	Top := (Height - Bottom) div 2;
	Bottom := pred(Height - Top);
      end;
      vaBottom: begin
	Top := Height - Bottom;
	Bottom := Height;
      end;
    end;
    if PtInRect(rText, Point(Message.XPos, Message.YPos)) then begin
      if not FHovering then begin
	FHovering := true;
	if FLinkLabel then begin
	  Screen.Cursor := crHand;
	  Font.Style := [fsBold, fsUnderline];
	end;
	if Assigned(FOnTextHover) then FOnTextHover(Self, true);
      end;
    end
    else begin
      if FHovering then begin
	FHovering := false;
	if FLinkLabel then begin
	  Screen.Cursor := crDefault;
	  Font.Style := [fsUnderline];
	end;
	if Assigned(FOnTextHover) then FOnTextHover(Self, false);
      end;
    end;
  end;
end;

procedure THotLabel.SetAutoSize(Value: Boolean);
begin
  if FLinkLabel then begin
    inherited AutoSize := false;
    FAutoSize := Value;
  end
  else begin
    inherited;
    FAutoSize := inherited AutoSize;
  end;
end;

procedure THotLabel.Click;
begin
  if FLinkLabel then begin
    Font.Color := FHotColor;
    Refresh;
  end;
  try
    inherited;
  finally
    if FLinkLabel then begin
      Font.Color := FLinkColor;
      Refresh;
    end;
  end;
end;

procedure THotLabel.Paint;

  function ResizeText: String;
  const
    ELLIPSIS = #$2026;
  var
    iMaxWidth, iTextLen: Integer;
    dQuotient: Double;
    szText: TSize;
  begin
    Result := FCaption;
    if (Result <> E) and (Alignment = taCenter) then begin
      iMaxWidth := Width - FHorizontalMargin shl 1;
      iTextLen := pred(Length(FCaption) shr 1);
      repeat
	GetTextExtentPoint32W(Canvas.Handle, Pointer(Result), iTextLen, szText);
	dQuotient := szText.cx / iMaxWidth;
	if dQuotient >= 1.5 then
	  iTextLen := Round(iTextLen / dQuotient)
	else begin
	  if iTextLen > 0 then begin
	    inc(iTextLen);
	    UniqueString(Result);
	    Delete(Result, succ(iTextLen shl 1), MAXINT);
	    repeat
	      {$WARNINGS OFF}
	      PWideChar(Result)[iTextLen] := NULL;
	      if iTextLen > 0 then PWideChar(Result)[pred(iTextLen)] := ELLIPSIS;
	      {$WARNINGS ON}
	      GetTextExtentPoint32W(Canvas.Handle, Pointer(Result), iTextLen, szText);
	      if szText.cx > iMaxWidth then begin
		Delete(Result, succ(iTextLen shl 1), MAXINT);
		dec(iTextLen);
	      end
	      else Break;
	    until szText.cx <= iMaxWidth;
	  end
	  else Result := E;
	  Exit;
	end;
      until dQuotient < 1.5;
    end
  end;

var
  rClient, rText: TRect;
  iLen: Integer;
  szText: TSize;
  sText: String;
begin
  with Canvas do begin
    Brush.Color := Self.Color;
    Brush.Style := bsSolid;
    rClient := ClientRect;
    rText := ClientRect;
    if not Transparent then begin
      SetBkMode(Handle, Windows.OPAQUE);
      FillRect(rClient);
    end
    else SetBkMode(Handle, Windows.TRANSPARENT);
    if Alignment = taCenter then InflateRect(rText, -FHorizontalMargin, 0);
    Font.Assign(Self.Font);
    if not Enabled then
      SetTextColor(Handle, ColorToRGB(clGrayText))
    else SetTextColor(Handle, ColorToRGB(Font.Color));
    Brush.Style := bsClear;
    iLen := Length(FCaption);
    if iLen > 0 then begin
      if FCaption[iLen] = NULL then begin
	if FUnicodeSystem then begin
	  Windows.DrawTextW(Handle, Pointer(FCaption), -1, rText,
			    dtAlignments[Alignment] or dtVAlignments[FVerticalAlignment] or
			    DT_SINGLELINE or (ord(ShowAccelChar = false) * DT_NOPREFIX or DT_END_ELLIPSIS));
	end
	else begin
	  SetTextAlign(Handle, TA_LEFT or TA_TOP);
	  szText := GetTextSize;
	  with rClient.TopLeft do begin
	    case Alignment of
	      taRightJustify: X := Width - szText.cx;
	      taCenter: if szText.cx < Width then begin
		X := (Width - szText.cx) div 2;
		dec(rClient.Right, X);
	      end;
	    end;
	    case VerticalAlignment of
	      vaBottom: Y := Height - szText.cy;
	      vaCenter: if szText.cy < Height then begin
		Y := (Height - szText.cy) div 2;
		dec(rClient.Bottom, Y);
	      end;
	    end;
	    sText := ResizeText;
	    iLen := Length(sText);
	    Windows.ExtTextOutW(Handle, X, Y, ETO_CLIPPED, @rClient, Pointer(sText), pred(iLen shr 1), nil);
	  end;
	end;
      end
      else begin
	Windows.DrawText(Handle, Pointer(FCaption), -1, rText,
			 dtAlignments[Alignment] or dtVAlignments[FVerticalAlignment] or
			 DT_SINGLELINE or (ord(ShowAccelChar = false) * DT_NOPREFIX or DT_END_ELLIPSIS));
      end;
    end;
  end;
end;

constructor THotLabel.Create(AOwner: TComponent);
begin
  inherited;
  AutoSize := false;
  ShowAccelChar := true;
  FHotColor := clRed;
  FLinkColor := clBlue;
  FUnicodeSystem := (Win32Platform = VER_PLATFORM_WIN32_NT);
  FVerticalAlignment := vaCenter;
  with Screen do begin
    Cursors[crHand] := LoadCursor(0, Ptr(IDC_HAND));
    if Cursors[crHand] = Cursors[crDefault] then Cursors[crHand] := LoadCursor(HInstance, 'HANDCURSOR');
  end;
end;

destructor THotLabel.Destroy;
begin
  inherited;
end;

procedure THotLabel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if (Alignment = taCenter) and (AWidth div 2 < FHorizontalMargin) then Exit;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

procedure Register;
begin
  RegisterComponents('Modified', [THotLabel]);
end;

end.

