{$J+,Z4}
unit ImgListEx;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, CommCtrl, Classes, Graphics, Controls;


// TBlendImages ----------------------------------------------------------------
type
  TAlpha = 0..100;
  TBlendImages = class(TImageList)
  private
    FAlphaBlendAssigned: Longbool;
    FBitmap: TBitmap;
  protected
    function BlendImage(DestCanvas: TCanvas; const DestSizeRect: TRect; Alpha: TAlpha): Longbool;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property  AlphaBlendAssigned: Longbool read FAlphaBlendAssigned;
    property  Bitmap: TBitmap read FBitmap;
  published
  end;


// T32BitImages ----------------------------------------------------------------
type
  T32BitImages = class(TImageList)
  private
    FImage, FMask: TBitmap;
    FControlBrush: TBrush;
    FThemed: Longbool;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateHandle(AWidth, AHeight, InitialCount: Integer; Brush: TBrush; Themed: Longbool);
    function AddIconByHandle(Icon: HICON; IncludeBlended, DestroyIcon: Longbool): Integer;
  published
    property ControlBrush: TBrush read FControlBrush write FControlBrush;
  end;


implementation

uses
  {$IFDEF CONDITIONALEXPRESSIONS}
  RtlConsts,
  {$ENDIF}
  Consts;


// coming with W98 / W2k and later ---------------------------------------------
const
  AC_SRC_OVER	= $00;
  AC_SRC_ALPHA	= $01;

type
  TBlendFunction = packed record
    BlendOp: Byte;
    BlendFlags: Byte;
    SourceConstantAlpha: Byte;
    AlphaFormat: Byte;
  end;

var
  AlphaBlend: function(hdcDest: HDC; nXOriginDest, nYOriginDest, nWidthDest, nHeightDest: Integer;
		       hdcSrc: HDC; nXOriginSrc, nYOriginSrc, nWidthSrc, nHeightSrc: Integer;
		       const BlendFunction: TBlendFunction): BOOL; stdcall;


// TBlendImages -----------------------------------------------------------------

constructor TBlendImages.Create(AOwner: TComponent);
begin
  inherited;
  FAlphaBlendAssigned := Assigned(AlphaBlend);
  FBitmap := TBitmap.Create;
end;

destructor TBlendImages.Destroy;
begin
  FBitmap.Free;
  inherited;
end;

function TBlendImages.BlendImage(DestCanvas: TCanvas; const DestSizeRect: TRect; Alpha: TAlpha): Longbool;
var
  bfImage: TBlendFunction;
begin
  Result := false;
  if AlphaBlendAssigned then begin
    with bfImage do begin
      BlendOp := AC_SRC_OVER;
      BlendFlags := 0;
      SourceConstantAlpha := ($FF * Alpha) div 100;
      AlphaFormat := 0;
    end;
    with DestSizeRect, FBitmap do begin
      Result := AlphaBlend(DestCanvas.Handle, Left, Top, Right, Bottom, Canvas.Handle, 0, 0, Width, Height, bfImage);
    end;
  end;
end;


// T32BitImages ----------------------------------------------------------------

constructor T32BitImages.Create;
begin
  inherited;
  FImage := TBitmap.Create;
  FMask := TBitmap.Create;
end;

destructor T32BitImages.Destroy;
begin
  FImage.Free;
  FMask.Free;
  inherited;
end;

procedure T32BitImages.CreateHandle(AWidth, AHeight, InitialCount: Integer; Brush: TBrush; Themed: Longbool);
begin
  Handle := ImageList_Create(AWidth, AHeight, ILC_COLOR32 or ILC_MASK, InitialCount, AllocBy);
  if (Handle = 0) or (Brush = nil) then begin
    {$IFDEF CONDITIONALEXPRESSIONS}
    Raise EInvalidOperation.Create(SInvalidImageList);
    {$ELSE}
    Raise EInvalidOperation.CreateRes(SInvalidImageList);
    {$ENDIF}
  end;
  FControlBrush := Brush;
  if FImage <> nil then with FImage do begin
    Width := AWidth;
    Height := AHeight;
  end;
  if FMask <> nil then with FMask do begin
    Width := AWidth;
    Height := AHeight;
  end;
  FThemed := Themed;
  Masked := true;
end;

function T32BitImages.AddIconByHandle(Icon: HICON; IncludeBlended, DestroyIcon: Longbool): Integer;
begin
  if Icon <> 0 then begin
    try
      if not FThemed then begin
	with FImage do DrawIconEx(Canvas.Handle, 0, 0, Icon, Width, Height, 0, FControlBrush.Handle, DI_NORMAL);
	with FMask do DrawIconEx(Canvas.Handle, 0, 0, Icon, Width, Height, 0, 0, DI_MASK);
	Result := ImageList_Add(Self.Handle, FImage.Handle, FMask.Handle);
      end
      else Result := ImageList_AddIcon(Self.Handle, Icon);
    finally
      if DestroyIcon then Windows.DestroyIcon(Icon);
    end;
    if IncludeBlended then begin
      Icon := ImageList_GetIcon(Self.Handle, Result, ILD_SELECTED);
      try
	if not FThemed then begin
	  with FImage do DrawIconEx(Canvas.Handle, 0, 0, Icon, Width, Height, 0, FControlBrush.Handle, DI_NORMAL);
	  ImageList_Add(Self.Handle, FImage.Handle, FMask.Handle);
	end
	else ImageList_AddIcon(Self.Handle, Icon);
      finally
	Windows.DestroyIcon(Icon);
      end;
    end;
  end
  else Result := -1;
end;


begin
  AlphaBlend := GetProcAddress(GetModuleHandle('MsImg32.dll'), 'AlphaBlend');
end.

