{$J+,Z4}
unit MaskEdEx;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$IFDEF CONDITIONALEXPRESSIONS}
{$WARN SUSPICIOUS_TYPECAST OFF}
{$ENDIF}

interface

uses
  Windows, Messages, CommCtrl, SysUtils, Classes, Graphics, Forms,
  Controls, ComCtrls, StdCtrls, Mask, ToolTip, UTF8, UxThemes;


// TEditEx ---------------------------------------------------------------------
type
  TEditEx = class(TEdit)
  private
    FToolTip: TToolTip;
    FEditTheme: THandle;
    FTextChanged: Longbool;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMContextMenu(var Message: TWMMouse); message WM_CONTEXTMENU;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMSetText(var Message: TWMSetText); message WM_SETTEXT;
    function  GetOversizedText(const Text: String): String;
    procedure GetToolTipOrigin(var TopLeft: TPoint);
  protected
    procedure Change; override;
  public
  published
  end;


// TMaskEdEx -------------------------------------------------------------------
type
  TOnEditLeave = procedure(Sender: TObject; Trigger: Integer) of Object;
  TMaskEdEx = class(TMaskEdit)
  private
    FToolTip: TToolTip;
    FEditTheme: THandle;
    FTextChanged: Longbool;
    FOnCopy: TNotifyEvent;
    FOnPaste: TNotifyEvent;
    FOnEditLeave: TOnEditLeave;
    procedure EditLeave(Trigger: Integer);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMExit(var Message: TMessage); message CM_EXIT;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CNKeyDown(var Message: TWMKeyDown); message CN_KEYDOWN;
    procedure WMContextMenu(var Message: TWMMouse); message WM_CONTEXTMENU;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMSetText(var Message: TWMSetText); message WM_SETTEXT;
    procedure WMCopy(var Message: TMessage); message WM_COPY;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    function  GetOversizedText(const Text: String): String;
    procedure GetToolTipOrigin(var TopLeft: TPoint);
  protected
    procedure Change; override;
  public
  published
    property  OnEditLeave: TOnEditLeave read FOnEditLeave write FOnEditLeave;
    property  OnCopy: TNotifyEvent read FOnCopy write FOnCopy;
    property  OnPaste: TNotifyEvent read FOnPaste write FOnPaste;
  end;


// TSpinEdEx -------------------------------------------------------------------
type
  TOnOutOfRange = procedure(Sender: TObject; OutOfRangeValue: Integer; var ReplaceWithValue: Integer) of Object;
  TSpinEdEx = class(TMaskEdEx)
  private
    FUDButton: TUpDown;
    procedure SetMaxLength;
    procedure DecValue;
    procedure IncValue;
    function  GetIncrement: Integer;
    procedure SetIncrement(NewValue: Integer);
    function  GetMaxValue: Integer;
    procedure SetMaxValue(NewValue: Integer);
    function  GetMinValue: Integer;
    procedure SetMinValue(NewValue: Integer);
    function  GetValue: Integer;
    procedure SetValue(NewValue: Integer);
    procedure OnUDClick(Sender: TObject; Button: TUDBtnType);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure WMContextMenu(var Message: TWMMouse); message WM_CONTEXTMENU;
    procedure WMKeyDown(var Message: TWMKey); message WM_KEYDOWN;
    procedure WMNotify(var Message: TWMNotify); message WM_NOTIFY;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Change; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property  Increment: Integer read GetIncrement write SetIncrement;
    property  MinValue: Integer read GetMinValue write SetMinValue;
    property  MaxValue: Integer read GetMaxValue write SetMaxValue;
    property  Value: Integer read GetValue write SetValue;
  end;


procedure Register;


implementation

uses
  Consts;

const
  DRAWSTYLE = DT_EXTERNALLEADING or DT_NOCLIP or DT_HIDEPREFIX or DT_SINGLELINE or DT_VCENTER;

function OpenEditTheme(Window, Theme: THandle; Color: TColor): THandle;
begin
  if EnableThemes(Window, (Color = clWindow) or (Color = clBtnFace)) then begin
    if Theme = 0 then Theme := OpenThemeData(Window, 'EDIT');
    Result := Theme;
    Exit;
  end;
  CloseThemeData(Theme);
  Result := 0;
end;

function CloseEditTheme(Theme: THandle): THandle;
begin
  CloseThemeData(Theme);
  Result := 0;
end;


// TEditEx ---------------------------------------------------------------------

procedure TEditEx.CMColorChanged(var Message: TMessage);
begin
  inherited;
  FEditTheme := OpenEditTheme(Handle, FEditTheme, Color);
end;

procedure TEditEx.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if FToolTip <> nil then FToolTip.AnsiText := Text;
end;

procedure TEditEx.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FTextChanged then begin
    FTextChanged := false;
    if FToolTip <> nil then FToolTip.AnsiText := Text;
  end;
end;

procedure TEditEx.WMContextMenu(var Message: TWMMouse);
var
  ptMouse: TPoint;
begin
  GetCursorPos(ptMouse);
  with Message do if PtInRect(ClientRect, ScreenToClient(ptMouse)) then
    inherited
  else Parent.DefaultHandler(Message);
end;

procedure TEditEx.WMCreate(var Message: TWMCreate);
begin
  inherited;
  Perform(CM_COLORCHANGED, 0, 0);
  if not (csDesigning in ComponentState) then FToolTip := TToolTip.Create(Self);
  if FToolTip <> nil then with FToolTip do begin
    FixedPos := true;
    ParentFont := true;
    OverSizeOnly := true;
    AnsiText := Text;
    OnGetOversizedText := GetOversizedText;
    OnGetToolTipOrigin := GetToolTipOrigin;
  end;
end;

procedure TEditEx.WMDestroy(var Message: TWMDestroy);
begin
  CloseEditTheme(FEditTheme);
  FToolTip.Free;
  inherited;
end;

procedure TEditEx.WMSetText(var Message: TWMSetText);
begin
  inherited;
  if FToolTip <> nil then FToolTip.AnsiText := Message.Text;
end;

function TEditEx.GetOversizedText(const Text: String): String;
var
  hDC, hFont: THandle;
  rText: TRect;
begin
  hDC := GetDC(Handle);
  hFont := SelectObject(hDC, Font.Handle);
  try
    rText := ClientRect;
    if FEditTheme <> 0 then begin
      GetThemeTextExtent(FEditTheme, hDC, EP_EDITTEXT, ETS_NORMAL,
			 PWideChar(AnsiToWideString(Text)), -1, DRAWSTYLE, @rText, rText);
    end
    else DrawText(hDC, PChar(Text), -1, rText, DRAWSTYLE or DT_CALCRECT);
    if rText.Right > ClientWidth then
      Result := Text
    else Result := '';
  finally
    if hFont <> 0 then SelectObject(hDC, hFont);
    if hDC <> 0 then ReleaseDC(Handle, hDC);
  end;
end;

procedure TEditEx.GetToolTipOrigin(var TopLeft: TPoint);
begin
  if BorderStyle = bsNone then with TopLeft do begin
    dec(X, GetSystemMetrics(SM_CXBORDER));
    dec(Y, GetSystemMetrics(SM_CYBORDER));
  end;
end;

procedure TEditEx.Change;
begin
  inherited;
  Invalidate;
  FTextChanged := true;
end;


// TMaskEdEx -------------------------------------------------------------------

procedure TMaskEdEx.EditLeave(Trigger: Integer);
begin
  if Assigned(FOnEditLeave) then FOnEditLeave(Self, Trigger);
end;

procedure TMaskEdEx.CMColorChanged(var Message: TMessage);
begin
  inherited;
  FEditTheme := OpenEditTheme(Handle, FEditTheme, Color);
end;

procedure TMaskEdEx.CMExit(var Message: TMessage);
begin
  EditLeave(0);
  inherited;
end;

procedure TMaskEdEx.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if FToolTip <> nil then FToolTip.AnsiText := Text;
end;

procedure TMaskEdEx.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FTextChanged then begin
    FTextChanged := false;
    if FToolTip <> nil then FToolTip.AnsiText := Text;
  end;
end;

procedure TMaskEdEx.CNKeyDown(var Message: TWMKeyDown);
begin
  with Message do begin
    if CharCode = VK_RETURN then CharCode := VK_TAB;
    if CharCode = VK_TAB then EditLeave(CharCode);
  end;
  inherited;
end;

procedure TMaskEdEx.WMContextMenu(var Message: TWMMouse);
var
  ptMouse: TPoint;
begin
  GetCursorPos(ptMouse);
  with Message do if PtInRect(ClientRect, ScreenToClient(ptMouse)) then
    inherited
  else Parent.DefaultHandler(Message);
end;

procedure TMaskEdEx.WMCreate(var Message: TWMCreate);
begin
  inherited;
  Perform(CM_COLORCHANGED, 0, 0);
  if not (csDesigning in ComponentState) then FToolTip := TToolTip.Create(Self);
  if FToolTip <> nil then with FToolTip do begin
    FixedPos := true;
    ParentFont := true;
    OverSizeOnly := true;
    AnsiText := Text;
    OnGetOversizedText := GetOversizedText;
    OnGetToolTipOrigin := GetToolTipOrigin;
  end;
end;

procedure TMaskEdEx.WMDestroy(var Message: TWMDestroy);
begin
  CloseEditTheme(FEditTheme);
  FToolTip.Free;
  inherited;
end;

procedure TMaskEdEx.WMSetText(var Message: TWMSetText);
begin
  inherited;
  if FToolTip <> nil then FToolTip.AnsiText := Message.Text;
end;

procedure TMaskEdEx.WMCopy(var Message: TMessage);
begin
  if Assigned(FOnCopy) then
    FOnCopy(Self)
  else inherited;
end;

procedure TMaskEdEx.WMPaste(var Message: TMessage);
begin
  if Assigned(FOnPaste) then
    FOnPaste(Self)
  else inherited;
end;

function TMaskEdEx.GetOversizedText(const Text: String): String;
var
  hDC, hFont: THandle;
  rText: TRect;
begin
  hDC := GetDC(Handle);
  hFont := SelectObject(hDC, Font.Handle);
  try
    rText := ClientRect;
    if FEditTheme <> 0 then begin
      GetThemeTextExtent(FEditTheme, hDC, EP_EDITTEXT, ETS_NORMAL,
			 PWideChar(AnsiToWideString(Text)), -1, DRAWSTYLE, @rText, rText);
    end
    else DrawText(hDC, PChar(Text), -1, rText, DRAWSTYLE or DT_CALCRECT);
    if rText.Right > ClientWidth then
      Result := Text
    else Result := '';
  finally
    if hFont <> 0 then SelectObject(hDC, hFont);
    if hDC <> 0 then ReleaseDC(Handle, hDC);
  end;
end;

procedure TMaskEdEx.GetToolTipOrigin(var TopLeft: TPoint);
begin
  if BorderStyle = bsNone then with TopLeft do begin
    dec(X, GetSystemMetrics(SM_CXBORDER));
    dec(Y, GetSystemMetrics(SM_CYBORDER));
  end;
end;

procedure TMaskEdEx.Change;
begin
  inherited;
  FTextChanged := true;
end;


// TSpinEdEx -------------------------------------------------------------------

procedure TSpinEdEx.SetMaxLength;
var
  iMin, iMax: Integer;
begin
  with FUDButton do begin
    iMin := Length(IntToStr(Min));
    iMax := Length(IntToStr(Max));
  end;
  if iMax > iMin then
    MaxLength := iMax
  else MaxLength := iMin;
end;

procedure TSpinEdEx.DecValue;
begin
  SetValue(FUDButton.Position - Increment);
end;

procedure TSpinEdEx.IncValue;
begin
  SetValue(FUDButton.Position + Increment);
end;

function TSpinEdEx.GetIncrement: Integer;
begin
  Result := FUDButton.Increment;
end;

procedure TSpinEdEx.SetIncrement(NewValue: Integer);
begin
  if NewValue <> FUDButton.Increment then FUDButton.Increment := NewValue;
end;

function TSpinEdEx.GetMaxValue: Integer;
begin
  Result := FUDButton.Max;
end;

procedure TSpinEdEx.SetMaxValue(NewValue: Integer);
begin
  if NewValue <> FUDButton.Max then begin
    FUDButton.Max := NewValue;
    SetMaxLength;
  end;
end;

function TSpinEdEx.GetMinValue: Integer;
begin
  Result := FUDButton.Min;
end;

procedure TSpinEdEx.SetMinValue(NewValue: Integer);
begin
  if NewValue <> FUDButton.Min then begin
    FUDButton.Min := NewValue;
    SetMaxLength;
  end;
end;

function TSpinEdEx.GetValue: Integer;
begin
  Result := FUDButton.Position;
end;

procedure TSpinEdEx.SetValue(NewValue: Integer);
var
  iStart, iLength, iTextLen: Integer;
begin
  if NewValue <> FUDButton.Position then begin
    iStart := SelStart;
    iLength := SelLength;
    iTextLen := Length(Text);
    Text := IntToStr(NewValue);
    if iLength < iTextLen then begin
      SelStart := iStart;
      SelLength := iLength;
    end
    else SelectAll;
    Change;
  end;
end;

procedure TSpinEdEx.OnUDClick(Sender: TObject; Button: TUDBtnType);
begin
  if ReadOnly then Exit;
  if not Focused then SetFocus;
end;

procedure TSpinEdEx.CMColorChanged(var Message: TMessage);
begin
  inherited;
  EnableThemes(FUDButton.Handle, (Color = clWindow) or (Color = clBtnFace));
end;

procedure TSpinEdEx.CMEnabledChanged(var Message: TMessage);
begin
  FUDButton.Enabled := Enabled;
  inherited;
end;

procedure TSpinEdEx.CMShowingChanged(var Message: TMessage);
begin
  if Showing then Change;
  inherited;
end;

procedure TSpinEdEx.WMChar(var Message: TWMChar);
var
  iStart, iLength: Integer;
  sBuffer: String;
begin
  with Message, FUDButton do begin
    case Char(CharCode) of
      '+': IncValue;
      '-': DecValue;
    else
      if (KeyDataToShiftState(KeyData) * [ssCtrl, ssAlt] = []) and (CharCode in [Ord('0')..Ord('9')]) then begin
	iStart := SelStart;
	iLength := SelLength;
	if iLength > 0 then begin
	  if iStart < Length(Text) then begin
	    sBuffer := Text;
	    sBuffer[succ(iStart)] := Char(CharCode);
	    Text := sBuffer;
	    SelStart := succ(iStart);
	    SelLength := pred(iLength);
	    Change;
	    Exit;
	  end;
	end;
      end;
      inherited;
    end;
  end;
end;

procedure TSpinEdEx.WMContextMenu(var Message: TWMMouse);
var
  ptMouse: TPoint;
begin
  GetCursorPos(ptMouse);
  with Message do if PtInRect(ClientRect, ScreenToClient(ptMouse)) then
    inherited
  else Parent.DefaultHandler(Message);
end;

procedure TSpinEdEx.WMKeyDown(var Message: TWMKey);
begin
  with FUDButton do begin
    case Message.CharCode of
      VK_PRIOR: if not ReadOnly then SetValue(Max);
      VK_NEXT: if not ReadOnly then SetValue(Min);
      VK_UP: if not ReadOnly then IncValue;
      VK_DOWN: if not ReadOnly then DecValue;
    else
      inherited;
    end;
  end;
end;

procedure TSpinEdEx.WMNotify(var Message: TWMNotify);
begin
  inherited;
  with PNMUpDown(Message.NMHdr)^ do if Hdr.code = UDN_DELTAPOS then begin
    if iDelta < 0
      then DecValue
    else IncValue;
    Message.Result := -1;
  end
  else inherited;
end;

procedure TSpinEdEx.WMSize(var Message: TWMSize);
var
  rFUD: TRect;
begin
  inherited;
  with rFUD do begin
    Left := ClientWidth - FUDButton.Width;
    Top := ClientHeight - FUDButton.Height;
    if Left < 0 then Left := 0;
    if Top < 0 then Top := 0;
    Right := FUDButton.Width;
    Bottom := FUDButton.Height;
    if Right > ClientWidth then Right := ClientWidth;
    if Bottom > ClientHeight then Bottom := ClientHeight;
    FUDButton.SetBounds(Left, Top, Right, Bottom);
  end;
end;

procedure TSpinEdEx.Change;
var
  iNewPosition: Integer;
begin
  with FUDButton do begin
    iNewPosition := StrToInt(Text);
    if iNewPosition < MinValue then iNewPosition := MinValue;
    if iNewPosition > MaxValue then iNewPosition := MaxValue;
    Text := IntToStr(iNewPosition);
    Position := iNewPosition;
  end;
  inherited;
end;

procedure TSpinEdEx.CreateParams(var Params: TCreateParams);
begin
  inherited;
  with Params do ExStyle := ExStyle or WS_EX_CONTROLPARENT;
end;

constructor TSpinEdEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle :=  ControlStyle + [csAcceptsControls];
  FUDButton := TUpDown.Create(Self);
  if FUDButton <> nil then with FUDButton do begin
    OnClick := OnUDClick;
    Parent := Self;
    SetMaxLength;
  end;
end;

destructor TSpinEdEx.Destroy;
begin
  if FUDButton <> nil then FUDButton.Free;
  inherited Destroy;
end;


procedure Register;
begin
  RegisterComponents('Modified', [TEditEx, TMaskEdEx, TSpinEdEx]);
end;


end.
