{$J+,Z4}
unit MenuControl;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, SysUtils, Messages, Classes, Graphics, Forms, Controls, Menus, UxThemes;

const
  WM_CHANGEUISTATE	= $0127;
  WM_UPDATEUISTATE	= $0128;
  WM_QUERYUISTATE	= $0129;
  UISF_HIDEFOCUS	= $1;
  UISF_HIDEACCEL	= $2;

type
  TMCPopupMenu = class(TPopupMenu)
  private
    function  GetAutoPopup: Longbool;
    procedure SetAutoPopup(Value: Longbool);
  public
    procedure Popup(X, Y: Integer); override;
  published
    property AutoPopup: Longbool read GetAutoPopup write SetAutoPopup;
  end;

type
  TOnShowPopupMenu = procedure(const Position: TPoint; var Cancel: Longbool) of Object;
  TMenuControl = class(TCustomControl)
  private
    FAppOnMessage: TMessageEvent;
    FFormProc: TFNWndProc;
    FHovering: Longbool;
    FKeyTriggered: Longbool;
    FMouseClicked: Longbool;
    FParentForm: TForm;
    FPopupMenu: TMCPopupMenu;
    FOnShowPopupMenu: TOnShowPopupMenu;
    procedure AdjustBounds;
    function  GetTextSize: TSize;
    function  ShowAccelChar: Longbool;
    procedure SetHovering(Value: Longbool);
    procedure SetKeyTriggered(Value: Longbool);
    procedure SetMouseClicked(Value: Longbool);
    procedure SetPopupMenu(Value: TMCPopupMenu);
    function  CheckAccel(CharCode: Word): Longbool;
    procedure SubclassParentForm;
    procedure UnsubclassParentForm;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMEraseBkGnd(var Message: TWMEraseBkGnd); message WM_ERASEBKGND;
    procedure WMExitMenuLoop(var Message: TWMExitMenuLoop); message WM_EXITMENULOOP;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMMenuSelect(var Message: TWMMenuSelect); message WM_MENUSELECT;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMUpdateUIState(var Message: TMessage); message WM_UPDATEUISTATE;
    procedure AppOnMessage(var Msg: TMsg; var Handled: Boolean);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Click; override;
  published
    property Caption;
    property Enabled;
    property Font;
    property ParentFont;
    property Visible;
    property PopupMenu: TMCPopupMenu read FPopupMenu write SetPopupMenu;
    property OnShowPopupMenu: TOnShowPopupMenu read FOnShowPopupMenu write FOnShowPopupMenu;
  end;

procedure Register;

implementation

const
  MARGIN = 6;
  EDGE_WIDTH = 2;
  DT_HIDEPREFIX = $00100000;
  MENU_LABEL_PROP = 'MenuLabelProp';
  DRAW_TEXT_FLAGS = DT_CENTER or DT_SINGLELINE or DT_VCENTER;

function TMCPopupMenu.GetAutoPopup: Longbool;
begin
  Result := inherited AutoPopup;
end;

procedure TMCPopupMenu.SetAutoPopup(Value: Longbool);
begin
  inherited AutoPopup := false;
end;

procedure TMCPopupMenu.Popup(X, Y: Integer);
var
  wFlags: Word;
  fCommand: Longbool;
  miClicked: TMenuItem;
const
  AAlignments: Array[TPopupAlignment] of Word = (TPM_LEFTALIGN, TPM_RIGHTALIGN, TPM_CENTERALIGN);
begin
  if Assigned(OnPopup) then OnPopup(Self);
  wFlags := AAlignments[Alignment] or TPM_RIGHTBUTTON or TPM_RETURNCMD;
  fCommand := TrackPopupMenu(Handle, wFlags, X, Y, 0, TWinControl(PopupComponent).Handle, nil);
  miClicked := FindItem(ord(fCommand), fkCommand);
  if miClicked <> nil then miClicked.Click;
end;

procedure TMenuControl.AdjustBounds;
begin
  SetBounds(0, 0, Width, GetTextSize.cy + EDGE_WIDTH);
end;

function TMenuControl.GetTextSize: TSize;
var
  rClient: TRect;
begin
  SetRectEmpty(rClient);
  with Result do begin
    cx := Parent.ClientWidth;
    cy := MARGIN;
    if Caption <> '' then begin
      DrawText(Canvas.Handle, Pointer(Caption), -1, rClient, DRAW_TEXT_FLAGS or DT_EXTERNALLEADING or DT_CALCRECT);
      with rClient do begin
	cx := Right - Left + (MARGIN shl 1);
	cy := Bottom - Top + (MARGIN shl 1);
      end;
    end;
  end;
end;

function TMenuControl.ShowAccelChar: Longbool;
begin
  Result := (FParentForm <> nil) and (SendMessage(GetDesktopWindow, WM_QUERYUISTATE, 0, 0) and UISF_HIDEACCEL = 0);
end;

procedure TMenuControl.SetHovering(Value: Longbool);
begin
  if Value <> FHovering then begin
    if not Value then FMouseClicked := false;
    FHovering := Value;
    Invalidate;
  end;
end;

procedure TMenuControl.SetKeyTriggered(Value: Longbool);
begin
  if Value <> FKeyTriggered then begin
    FKeyTriggered := Value;
    Invalidate;
  end;
end;

procedure TMenuControl.SetMouseClicked(Value: Longbool);
begin
  if Value <> FMouseClicked then begin
    FMouseClicked := Value;
    Invalidate;
  end;
end;

procedure TMenuControl.SetPopupMenu(Value: TMCPopupMenu);
begin
  FPopupMenu := Value;
  if FPopupMenu <> nil then FPopupMenu.PopupComponent := Self;
end;

function TMenuControl.CheckAccel(CharCode: Word): Longbool;
begin
  Result := false;
  if IsAccel(CharCode, Caption) and CanFocus then begin
    SetKeyTriggered(true);
    Result := true;
    Click;
  end;
end;

function FormProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  with TMenuControl(GetProp(hWnd, MENU_LABEL_PROP)) do begin
    case uMsg of
      WM_NCLBUTTONDOWN, WM_NCRBUTTONDOWN, WM_LBUTTONDOWN, WM_RBUTTONDOWN: if hWnd <> Handle then SetKeyTriggered(false);
      WM_MOUSEACTIVATE: if (hWnd <> Handle) and (HiWord(lParam) <> WM_MOUSEMOVE) then SetKeyTriggered(false);
      WM_SYSCOMMAND: if (wParam = SC_KEYMENU) and (lParam = 0) then begin
	SetKeyTriggered(not FKeyTriggered);
	Result := 0;
	Exit;
      end;
    end;
    Result := CallWindowProc(FFormProc, hWnd, uMsg, wParam, lParam);
  end;
end;

procedure TMenuControl.SubclassParentForm;
begin
  if not (Parent is TForm) or (GetProp(FParentForm.Handle, MENU_LABEL_PROP) <> 0) then Raise EAbort.Create('');
  with FParentForm do begin
    SetProp(Handle, MENU_LABEL_PROP, THandle(Self));
    FFormProc := TFnWndProc(SetWindowLong(Handle, GWL_WNDPROC, Longint(@FormProc)));
  end;
end;

procedure TMenuControl.UnsubclassParentForm;
begin
  if FFormProc <> nil then with FParentForm do begin
    SetWindowLong(Handle, GWL_WNDPROC, Longint(FFormProc));
    RemoveProp(Handle, MENU_LABEL_PROP);
    FFormProc := nil;
  end;
end;

procedure TMenuControl.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do begin
    if CheckAccel(CharCode) then
      Result := 1
    else inherited;
  end;
end;

procedure TMenuControl.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TMenuControl.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FHovering then begin
    FMouseClicked := false;
    FKeyTriggered := false;
    FHovering := false;
    Invalidate;
  end;
end;

procedure TMenuControl.CMTextChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TMenuControl.WMCommand(var Message: TWMCommand);
begin
  inherited;
  with Message do if (Ctl = 0) and (NotifyCode = 0) then FParentForm.Dispatch(Message);
end;

procedure TMenuControl.WMCreate(var Message: TWMCreate);
const
  MS_SHELL_DLG_2 = 'MS Shell Dlg 2';

  function FindMSShellDlg2: Longbool;

    function EnumFontsProc(const ELFE: TEnumLogFontEx; const TM: TTextMetric;
			   FontType: Cardinal; var Match: Longbool): Integer; stdcall;
    begin
      Result := CompareText(ELFE.elfLogFont.lfFaceName, MS_SHELL_DLG_2);
      Match := (Result = 0);
    end;

  var
    lfEnum: TLogFont;
  begin
    Result := false;
    FillChar(lfEnum, SizeOf(TLogFont), 0);
    with lfEnum do begin
      lfCharSet := DEFAULT_CHARSET;
      lfFaceName := MS_SHELL_DLG_2;
    end;
    EnumFontFamiliesEx(Canvas.Handle, lfEnum, @EnumFontsProc, Longint(@Result), 0);
  end;

begin
  inherited;
  AdjustBounds;
  try
    if FindMSShellDlg2 then Font.Name := MS_SHELL_DLG_2;
    FParentForm := TForm(GetParentForm(Self));
    SubclassParentForm;
  except
    Message.Result := -1;
    Raise;
  end;
end;

procedure TMenuControl.WMDestroy(var Message: TWMDestroy);
begin
  inherited;
  UnsubclassParentForm;
end;

procedure TMenuControl.WMEraseBkGnd(var Message: TWMEraseBkGnd);
var
  rClient, rText: TRect;
begin
  AdjustBounds;
  rClient := ClientRect;
  Canvas.Handle := Message.DC;
  with Canvas do begin
    Brush.Style := bsSolid;
    Brush.Color := Self.Color;
    DrawEdge(Handle, rClient, EDGE_ETCHED, BF_BOTTOM or BF_ADJUST);
    rText := Rect(0, 0, GetTextSize.cx, ClientHeight - EDGE_WIDTH);
    if FHovering or FKeyTriggered then begin
      SubtractRect(rClient, rClient, rText);
      if UseThemes then begin
	Brush.Color := clHighlight;
	FillRect(rText);
      end
      else begin
	FillRect(rText);
	dec(rText.Bottom, 1);
	if FHovering and not (FKeyTriggered or FMouseClicked) then
	  DrawEdge(Handle, rText, BDR_RAISEDINNER, BF_RECT)
	else DrawEdge(Handle, rText, BDR_SUNKENOUTER, BF_RECT);
      end;
      Brush.Color := Self.Color;
    end;
    FillRect(rClient);
  end;
  Message.Result := -1;
end;

procedure TMenuControl.WMExitMenuLoop(var Message: TWMExitMenuLoop);
begin
  FKeyTriggered := (GetKeyState(VK_ESCAPE) < 0);
  if not FKeyTriggered then begin
    FMouseClicked := false;
    FHovering := false;
  end;
  Invalidate;
  inherited;
end;

procedure TMenuControl.WMKillFocus(var Message: TWMKillFocus);
begin
  FMouseClicked := false;
  FKeyTriggered := false;
  FHovering := false;
  inherited;
end;

procedure TMenuControl.WMLButtonDown(var Message: TWMLButtonDown);
begin
  with Message do begin
    if PtInRect(Rect(0, 0, GetTextSize.cx, ClientHeight - succ(EDGE_WIDTH)), Point(XPos, YPos)) then begin
      SetMouseClicked(true);
      inherited;
    end
    else Result := -1;
  end;
end;

procedure TMenuControl.WMMenuSelect(var Message: TWMMenuSelect);
var
  fkSelected: TFindItemKind;
  miSelected: TMenuItem;
begin
  with Message do begin
    fkSelected := fkCommand;
    if MenuFlag and MF_POPUP <> 0 then fkSelected := fkHandle;
    miSelected := FPopupMenu.FindItem(IDItem, fkSelected);
    if miSelected <> nil then begin
      Application.Hint := miSelected.Hint;
      Exit;
    end;
    Application.Hint := '';
  end;
end;

procedure TMenuControl.WMMouseMove(var Message: TWMMouseMove);
var
  rText: TRect;
begin
  inherited;
  rText := Rect(0, 0, GetTextSize.cx, ClientHeight - succ(EDGE_WIDTH));
  with Message do if PtInRect(rText, Point(XPos, YPos)) <> FHovering then begin
    if FHovering then FKeyTriggered := false;
    SetHovering(not FHovering);
  end;
end;

procedure TMenuControl.WMUpdateUIState(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TMenuControl.AppOnMessage(var Msg: TMsg; var Handled: Boolean);
begin
  if TMethod(FAppOnMessage).Code <> nil then FAppOnMessage(Msg, Handled);
  if (FHovering or FKeyTriggered) and (Msg.Message = WM_KEYDOWN) then begin
    case Msg.wParam of
      VK_ESCAPE: begin
	if FHovering then FMouseClicked := false;
	FKeyTriggered := false;
	Invalidate;
      end;
      VK_UP, VK_DOWN: Click;
    else
      CheckAccel(msg.wParam);
    end;
    Handled := true;
  end;
end;

procedure TMenuControl.Paint;
var
  iOffset: Integer;
  rText: TRect;
begin
  with Canvas do begin
    if Caption <> '' then begin
      iOffset := ord(not UseThemes and ((FHovering and FMouseClicked) or FKeyTriggered));
      rText := Rect(0, 0, GetTextSize.cx, ClientHeight - succ(EDGE_WIDTH));
      OffsetRect(rText, iOffset, iOffset);
      if (FHovering or FKeyTriggered) and UseThemes then
	Brush.Color := clHighlight
      else Brush.Color := Self.Color;
      Font.Assign(Self.Font);
      if not Enabled then
	SetTextColor(Handle, ColorToRGB(clGrayText))
      else SetTextColor(Handle, ColorToRGB(Font.Color));
      DrawText(Handle, Pointer(Caption), -1, rText, DRAW_TEXT_FLAGS or DT_END_ELLIPSIS or
						    ord(ShowAccelChar = false) * DT_HIDEPREFIX);
    end;
  end;
end;

constructor TMenuControl.Create(AOwner: TComponent);
begin
  inherited;
  Align := alTop;
  ParentColor := true;
  Font.Color := clMenuText;
  FAppOnMessage := Application.OnMessage;
  Application.OnMessage := AppOnMessage;
end;

destructor TMenuControl.Destroy;
begin
  Application.OnMessage := FAppOnMessage;
  FAppOnMessage := nil;
  inherited;
end;

procedure TMenuControl.Click;
var
  ptPopup: TPoint;
  fCancel: Longbool;
begin
  inherited;
  if PopupMenu <> nil then begin
    ptPopup := ClientOrigin;
    with ptPopup do begin
      inc(Y, ClientHeight - succ(EDGE_WIDTH));
      SetHovering(true);
      try
	if Assigned(FOnShowPopupMenu) then begin
	  fCancel := false;
	  FOnShowPopupMenu(ptPopup, fCancel);
	  if fCancel then Exit;
	end;
	FPopupMenu.Popup(X, Y);
      finally
	SetHovering(false);
      end;
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('New', [TMCPopupMenu, TMenuControl]);
end;

end.

