{$J+,Z4}
unit Processes;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, ShellApi, SysUtils, StrUtil;

type
  PProcessInfo = ^TProcessInfo;
  TProcessInfo = Record
    piProcName: String;
    piVisible: Longbool;
    piWindow: THandle;
    piThread: Cardinal;
    piID: Cardinal;
  end;

function ShellExecuteAndWait(WindowHandle: THandle; const Operation, FileName, Parameters, Directory: String;
			     ShowCmd, WaitSeconds: Integer): Cardinal;
function CreateProcessID(const CommandLine, PathName: String; Show, WaitForExit: Longbool): Cardinal;
function GetProcessNameByID(ProcessId: DWord): String;
function FindProcessByID(const PI: TProcessInfo; ProcessId: DWord): Longbool;
function FindProcessInfo(const ProcessName: String; Visible: Longbool): TProcessInfo;
function FindParentProcess(const ProcessName: String): String;
function FindProcessWindow(ProcessID: Cardinal; const ClassName: String): THandle;
function GetProcessModuleFileName(Process, Module: THandle): String;
function GetModulePathByFileHandle(FileHandle: THandle): String;
function GetShellProcessId: Cardinal;
function GetShellAccessToken: THandle;
function RunningAsAdmin(ProcessId: Cardinal): Longbool;
function UserProfileLoad(Token: THandle; const User: String): THandle;
function UserProfileUnload(Token, Profile: THandle): Longbool;
function UserImpersonate(Token: THandle): Longbool;
function UserRevert: Longbool;
function NewInstance: Longbool;
procedure ProcessMessages;

implementation

// USERENV stuff
const
  PI_NOUI		= $00000001;
  PI_APPLYPOLICY	= $00000002;

type
  PProfileInfo = ^TProfileInfo;
  TProfileInfo = Record
    dwSize: DWord;
    dwFlags: DWord;
    lpUserName: PChar;
    lpProfilePath: PChar;
    lpDefaultPath: PChar;
    lpServerName: PChar;
    lpPolicyPath: PChar;
    hProfile: THandle;
  end;

// USER32 stuff
type
  TTokenInformationClass = (
    TokenNull,
    TokenUser,
    TokenGroups,
    TokenPrivileges,
    TokenOwner,
    TokenPrimaryGroup,
    TokenDefaultDacl,
    TokenSource,
    TokenType,
    TokenImpersonationLevel,
    TokenStatistics,
    TokenRestrictedSids,
    TokenSessionId,
    TokenGroupsAndPrivileges,
    TokenSessionReference,
    TokenSandBoxInert,
    TokenAuditPolicy,
    TokenOrigin,
    TokenElevationType,
    TokenLinkedToken,
    TokenElevation,
    TokenHasRestrictions,
    TokenAccessInformation,
    TokenVirtualizationAllowed,
    TokenVirtualizationEnabled,
    TokenIntegrityLevel,
    TokenUIAccess,
    TokenMandatoryPolicy,
    TokenLogonSid,
    vMaxTokenInfoClass
  );
  TTokenElevationType = (
    TokenElevationTypeNull,
    TokenElevationTypeDefault,
    TokenElevationTypeFull,
    TokenElevationTypeLimited
  );
  PTokenElevation = ^TTokenElevation;
  TTokenElevation = Packed Record
    TokenIsElevated : DWORD;
  end;

// TLHELP32 stuff
const
  PROCESS_TERMINATE			= $0001;
  PROCESS_CREATE_THREAD			= $0002;
  PROCESS_SET_SESSIONID			= $0004;
  PROCESS_VM_OPERATION			= $0008;
  PROCESS_VM_READ			= $0010;
  PROCESS_VM_WRITE			= $0020;
  PROCESS_DUP_HANDLE			= $0040;
  PROCESS_CREATE_PROCESS		= $0080;
  PROCESS_SET_QUOTA			= $0100;
  PROCESS_SET_INFORMATION		= $0200;
  PROCESS_QUERY_INFORMATION		= $0400;
  PROCESS_SUSPEND_RESUME		= $0800;
  PROCESS_QUERY_LIMITED_INFORMATION	= $1000;

const
  TH32CS_SNAPHEAPLIST			= $00000001;
  TH32CS_SNAPPROCESS 			= $00000002;
  TH32CS_SNAPTHREAD			= $00000004;
  TH32CS_SNAPMODULE			= $00000008;

type
  PProcessEntry32 = ^TProcessEntry32;
  TProcessEntry32 = Record
    dwSize: DWord;
    cntUsage: DWord;
    th32ProcessID: DWord;
    th32DefaultHeapID: DWord;
    th32ModuleID: DWord;
    cntThreads: DWord;
    th32ParentProcessID: DWord;
    pcPriClassBase: Longint;
    dwFlags: DWord;
    szExeFile: Array[0..pred(MAX_PATH)] of Char;
  end;

var
  GetModuleFileNameEx: function(hProcess, hModule: THandle; lpFilename: PChar; nSize: DWord): DWord; stdcall;
  GetMappedFileName: function(hProcess: THandle; lpFileView: Pointer; lpFilename: PChar; nSize: DWord): DWord; stdcall;
  GetModuleBaseName: function(hProcess, hModule: THandle; lpBaseName: PChar; nSize: DWord): DWord; stdcall;
  GetTokenInformation: function(TokenHandle: THandle; TokenInformationClass: TTokenInformationClass;
			       TokenInformation: Pointer; TokenInformationLength: DWORD;
			       var ReturnLength: DWORD): BOOL; stdcall;
  LoadUserProfile: function(hToken: THandle; var lpProfileInfo: TProfileInfo): Bool; stdcall;
  UnloadUserProfile: function(hToken, hProfile: THandle): Bool; stdcall;
  GetShellWindow: function: THandle; stdcall;
  CreateToolhelp32Snapshot: function(dwFlags, th32ProcessID: DWord): THandle; stdcall;
  Process32First: function(hSnapshot: THandle; var lppe: TProcessEntry32): Bool; stdcall;
  Process32Next: function (hSnapshot: THandle; var lppe: TProcessEntry32): Bool; stdcall;

function InitProcessInfoFunctions: THandle;
const
  USERENV = 'UserEnv.dll';
  PSAPI = 'PsApi.dll';
begin
  Result := LoadLibrary(USERENV);
  LoadUserProfile := GetProcAddress(Result, 'LoadUserProfileA');
  UnloadUserProfile := GetProcAddress(Result, 'UnloadUserProfile');
  GetShellWindow := GetProcAddress(GetModuleHandle(USER32), 'GetShellWindow');
  GetTokenInformation := GetProcAddress(GetModuleHandle(ADVAPI32), 'GetTokenInformation');
  Result := GetModuleHandle(PSAPI);
  if Result = 0 then Result := LoadLibrary(PSAPI);
  if Result = 0 then begin
    Result := GetModuleHandle(KERNEL32);
    CreateToolhelp32Snapshot := GetProcAddress(Result, 'CreateToolhelp32Snapshot');
    Process32First := GetProcAddress(Result, 'Process32First');
    Process32Next := GetProcAddress(Result, 'Process32Next');
  end
  else begin
    GetModuleFileNameEx := GetProcAddress(Result, 'GetModuleFileNameExA');
    GetMappedFileName := GetProcAddress(Result, 'GetMappedFileNameA');
    GetModuleBaseName := GetProcAddress(Result, 'GetModuleBaseNameA');
  end;
end;

function ShellExecuteAndWait(WindowHandle: THandle; const Operation, FileName, Parameters, Directory: String;
			     ShowCmd, WaitSeconds: Integer): Cardinal;
var
  sei: TShellExecuteInfo;
begin
  Result := 0;
  with sei do begin
    cbSize := SizeOf(TShellExecuteInfo);
    fMask := SEE_MASK_FLAG_NO_UI or SEE_MASK_NOCLOSEPROCESS or SEE_MASK_FLAG_DDEWAIT;
    Wnd := WindowHandle;
    lpVerb := PChar(Operation);
    lpFile := PChar(FileName);
    lpParameters := PChar(Parameters);
    lpDirectory := PChar(Directory);
    nShow := ShowCmd;
    hInstApp := 0;
    lpIDList := NIL;
    lpClass := NIL;
    hkeyClass := 0;
    dwHotKey := 0;
    hIcon := 0;
    hProcess := 0;
  end;
  if ShellExecuteEx(@sei) then with sei do begin
    try
      WaitForInputIdle(hProcess, WaitSeconds * 1000);
      Result := hInstApp;
    finally
      CloseHandle(hProcess);
    end;
  end;
end;

function CreateProcessID(const CommandLine, PathName: String; Show, WaitForExit: Longbool): Cardinal;
var
  si: TStartupInfo;
  pi: TProcessInformation;
begin
  Result := 0;
  FillChar(si, SizeOf(TStartupInfo), 0);
  with si do begin
    cb := SizeOf(TStartupInfo);
    dwFlags := STARTF_USESHOWWINDOW;
    if Show then
      wShowWindow := SW_SHOWNORMAL
    else wShowWindow := SW_HIDE;
  end;
  FillChar(pi, SizeOf(TProcessInfo), 0);
  if Windows.CreateProcess(nil, Pointer(CommandLine), nil, nil, false,
			   0, nil, Pointer(PathName), si, pi) then begin
    with pi do begin
      CloseHandle(hThread);
      try
	if not WaitForExit then begin
	  while Show and (FindProcessWindow(dwProcessID, E) = 0) do begin
	    if WaitForInputIdle(hProcess, 100) = WAIT_FAILED then Break;
	  end;
	end
	else if WaitForSingleObject(hProcess, INFINITE) = WAIT_FAILED then Result := GetLastError;
      finally
	CloseHandle(hProcess);
	SetLastError(Result);
	Result := dwProcessID;
      end;
    end;
  end;
end;

function GetProcessNameByID(ProcessId: DWord): String;

  function GetProcessModuleBaseName(Process, Module: THandle): String;
  begin
    SetString(Result, NIL, MAX_PATH);
    Delete(Result, succ(GetModuleBaseName(Process, Module, PChar(Result), MAX_PATH)), MAXINT);
  end;

var
  hProcess: THandle;
begin
  hProcess := OpenProcess(PROCESS_QUERY_INFORMATION or PROCESS_VM_OPERATION or PROCESS_VM_READ or PROCESS_VM_WRITE, FALSE, ProcessId);
  try
    Result := GetProcessModuleBaseName(hProcess, 0);
  finally
    CloseHandle(hProcess);
  end;
end;

function FindProcessByID(const PI: TProcessInfo; ProcessId: DWord): Longbool;
begin
  Result := (CompareText(PI.piProcName, GetProcessNameByID(ProcessId)) = 0);
end;

function FindProcessInfo(const ProcessName: String; Visible: Longbool): TProcessInfo;

  function EnumWindowsProc(FoundWindow: THandle; PPI: PProcessInfo): Bool; stdcall;
  var
    dwProcessId: DWord;
  begin
    Result := TRUE;
    with PPI^ do begin
      if not piVisible or IsWindowVisible(FoundWindow) then begin
	piThread := GetWindowThreadProcessId(FoundWindow, @dwProcessId);
	if (dwProcessId <> GetCurrentProcessId) and FindProcessByID(PPI^, dwProcessId) then begin
	  piWindow := FoundWindow;
	  piID := dwProcessId;
	  Result := FALSE;
	end
	else piThread := 0;
      end;
    end;
  end;

begin
  with Result do begin
    piProcName := ProcessName;
    piVisible := Visible;
    piWindow := 0;
    piThread := 0;
    piID := 0;
  end;
  EnumWindows(@EnumWindowsProc, LPARAM(@Result));
end;

function FindParentProcess(const ProcessName: String): String;
var
  hSnapshot: THandle;
  pe: TProcessEntry32;
begin
  hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if hSnapshot <> INVALID_HANDLE_VALUE then begin
    try
      pe.dwSize := SizeOf(TProcessEntry32);
      if Process32First(hSnapshot, pe) then begin
	repeat
	  if CompareText(ProcessName, pe.szExeFile) = 0 then begin
	    Result := GetProcessNameByID(pe.th32ParentProcessID);
	    Exit;
	  end;
	until not Process32Next(hSnapshot, pe);
      end;
    finally
      CloseHandle(hSnapshot);
    end;
  end;
  Result := E;
end;

function FindProcessId(const ProcessName: String): Cardinal;
var
  hSnapshot: THandle;
  pe: TProcessEntry32;
begin
  hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if hSnapshot <> INVALID_HANDLE_VALUE then begin
    try
      pe.dwSize := SizeOf(TProcessEntry32);
      if Process32First(hSnapshot, pe) then begin
	repeat
	  if CompareText(ProcessName, pe.szExeFile) = 0 then begin
	    Result := pe.th32ProcessId;
	    Exit;
	  end;
	until not Process32Next(hSnapshot, pe);
      end;
    finally
      CloseHandle(hSnapshot);
    end;
  end;
  Result := 0;
end;

function FindProcessWindow(ProcessId: Cardinal; const ClassName: String): THandle;
type
  PWindowData = ^TWindowData;
  TWindowData = Record
    ProcessId: Cardinal;
    ClassName: String;
    FoundWindow: THandle;
  end;

  function EnumWindowsProc(Window: THandle; PWD: PWindowData): Bool; stdcall;
  var
    dwProcessId: DWord;
  begin
    GetWindowThreadProcessId(Window, @dwProcessId);
    with PWD^ do begin
      if (ProcessId = dwProcessId) and ((ClassName = E) or (ClassName = ClassString(Window))) then begin
	FoundWindow := Window;
	Result := FALSE;
      end
      else begin
	FoundWindow := 0;
	Result := TRUE;
      end;
    end;
  end;

var
  wd: TWindowData;
begin
  wd.ProcessId := ProcessId;
  wd.ClassName := ClassName;
  EnumWindows(@EnumWindowsProc, LPARAM(@wd));
  Result := wd.FoundWindow;
end;

function GetProcessModuleFileName(Process, Module: THandle): String;
var
  iLen: Integer;
begin
  if Assigned(GetModuleFileNameEx) then begin
    SetString(Result, nil, MAX_PATH);
    iLen := GetModuleFileNameEx(Process, Module, Pointer(Result), MAX_PATH);
    Delete(Result, succ(iLen), MAXINT);
  end
  else Result := E;
end;

function GetModulePathByFileHandle(FileHandle: THandle): String;
var
  dwSizeLo, dwSizeHi, dwSize: DWord;
  hFileMap: THandle;
  pFileView: Pointer;
  sDriveList, sDrive: String;
  pDriveList: PChar;
begin
  if (FileHandle <> 0) and Assigned(GetMappedFileName) then begin
    dwSizeLo := GetFileSize(FileHandle, @dwSizeHi);
    if ((dwSizeLo <> INVALID_FILE_SIZE) or (GetLastError = NO_ERROR))
    and ((dwSizeLo <> 0) or (dwSizeHi <> 0)) then begin
      hFileMap := CreateFileMapping(FileHandle, nil, PAGE_READONLY, 0, 1, nil);
      if hFileMap <> 0 then begin
	pFileView := nil;
	try
	  pFileView := MapViewOfFile(hFileMap, FILE_MAP_READ, 0, 0, 1);
	  if pFileView <> nil then begin
	    SetString(Result, nil, MAX_PATH);
	    Delete(Result, succ(GetMappedFileName(GetCurrentProcess, pFileView, Pointer(Result), MAX_PATH)), MAXINT);
	    dwSizeHi := GetLogicalDriveStrings(0, nil);
	    if dwSizeHi > 0 then begin
	      SetString(sDriveList, nil, dwSizeHi);
	      Delete(sDriveList, succ(GetLogicalDriveStrings(dwSizeHi, Pointer(sDriveList))), MAXINT);
	      pDriveList := Pointer(sDriveList);
	      SetString(sDrive, nil, MAX_PATH);
	      dwSizeLo := 0;
	      repeat
		if succ(dwSizeLo) < dwSizeHi then pDriveList[dwSizeLo + 2] := NULL;
		if (QueryDosDevice(@pDriveList[dwSizeLo], Pointer(sDrive), MAX_PATH) > 0) then begin
		  dwSize := FastStrLen(Pointer(sDrive));
		  if (CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
		      Pointer(Result), dwSize, Pointer(sDrive), dwSize) = 2) then begin
		    inc(pDriveList, dwSizeLo);
		    Result := pDriveList + Copy(Result, succ(dwSize), MAXINT);
		    Break;
		  end;
		end;
		while pDriveList[dwSizeLo] <> NULL do inc(dwSizeLo);
		while (dwSizeLo < dwSizeHi) and (pDriveList[dwSizeLo] = NULL) do inc(dwSizeLo);
		if dwSizeLo < dwSizeHi then
		  Continue
		else Break;
	      until dwSizeLo = dwSizeHi;
	    end;
	    Exit;
	  end;
	finally
	  UnmapViewOfFile(pFileView);
	  CloseHandle(hFileMap);
	end;
      end;
    end;
  end;
  Result := E;
end;

function GetShellProcessId: Cardinal;
begin
  if Assigned(GetShellWindow) then
    GetWindowThreadProcessId(GetShellWindow, @Result)
  else Result := 0;
end;

function GetShellAccessToken: THandle;
var
  dwProcessId: DWord;
  hProcess: THandle;
begin
  dwProcessId := GetShellProcessId;
  if dwProcessId <> 0 then begin
    hProcess := OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwProcessId);
    if hProcess <> 0 then begin
      try
	{$IFNDEF VER90}
	if OpenProcessToken(hProcess, TOKEN_QUERY or TOKEN_IMPERSONATE or TOKEN_DUPLICATE, Result) then Exit;
	{$ELSE}
	if OpenProcessToken(hProcess, TOKEN_QUERY or TOKEN_IMPERSONATE or TOKEN_DUPLICATE, @Result) then Exit;
	{$ENDIF}
      finally
	CloseHandle(hProcess);
      end;
    end;
  end;
  Result := INVALID_HANDLE_VALUE;
end;

function RunningAsAdmin(ProcessId: Cardinal): Longbool;
var
  hProcess, hToken: THandle;
  te: TTokenElevation;
  dwSize: DWord;
begin
  if ProcessId <> 0 then begin
    hProcess := OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, ProcessId);
    if hProcess <> 0 then begin
      hToken := INVALID_HANDLE_VALUE;
      try
	{$IFNDEF VER90}
	if OpenProcessToken(hProcess, TOKEN_QUERY, hToken)
	{$ELSE}
	if OpenProcessToken(hProcess, TOKEN_QUERY, @hToken)
	{$ENDIF}
	and GetTokenInformation(hToken, TokenElevationType, @te, SizeOf(TTokenElevation), dwSize) then begin
	  Result := (TTokenElevationType(te.TokenIsElevated) = TokenElevationTypeFull);
	  Exit;
	end;
      finally
	CloseHandle(hToken);
	CloseHandle(hProcess);
      end;
    end;
  end;
  Result := FALSE;
end;

function UserProfileLoad(Token: THandle; const User: String): THandle;
var
  pi: TProfileInfo;
begin
  if Assigned(LoadUserProfile) then begin
    FillChar(pi, SizeOf(TProfileInfo), 0);
    with pi do begin
      dwSize := SizeOf(TProfileInfo);
      dwFlags := PI_NOUI;
      lpUserName := PChar(User);
    end;
    if LoadUserProfile(Token, pi) then begin
      Result := pi.hProfile;
      Exit;
    end;
  end;
  Result := HKEY_CURRENT_USER;
end;

function UserProfileUnload(Token, Profile: THandle): Longbool;
begin
  Result := (Profile <> HKEY_CURRENT_USER) and Assigned(UnloadUserProfile) and UnloadUserProfile(Token, Profile);
end;

function UserImpersonate(Token: THandle): Longbool;
begin
  Result := ImpersonateLoggedOnUser(Token);
end;

function UserRevert: Longbool;
begin
  Result := RevertToSelf;
end;

function NewInstance: Longbool;
var
  Mutex: THandle;
begin
  Result := false;
  Mutex := CreateMutex(nil, true, PChar(ParamStr(0)));
  if (Mutex <> 0) and (GetLastError = ERROR_ALREADY_EXISTS) then begin
    CloseHandle(Mutex);
    Result := true;
  end;
end;

procedure ProcessMessages;
var
  msg: TMsg;
begin
  while PeekMessage(msg, 0, 0, 0, PM_REMOVE) do begin
    TranslateMessage(msg);
    DispatchMessage(msg);
  end;
end;

initialization
  InitProcessInfoFunctions;

finalization

end.

