{------------------------------------------------------------------------------}
{        This code is copyright (C) 2010-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit RawByteStringList;

interface

uses
  Windows, SysUtils, Classes;

const
  MaxListSize = MaxInt div 16;


// TRawByteStrings -------------------------------------------------------------

type
  TRawByteStrings = Class;
  RawByteString = AnsiString;

  PStringItem = ^TStringItem;
  TStringItem = Record
    FString: RawByteString;
    FObject: TObject;
  end;

  PStringItemList = ^TStringItemList;
  TStringItemList = Array[0..MaxListSize] of TStringItem;
  TRawByteStringsSortCompare = function(List: TRawByteStrings; Index1, Index2: Integer): Integer;

  TRawByteStrings = class(TStrings)
  private
    FCount: Integer;
    FCapacity: Integer;
    FSorted: Boolean;
    FDelimiter: AnsiChar;
    FQuoteChar: AnsiChar;
    FNameValueSeparator: AnsiChar;
    FLineBreak: RawByteString;
    FCaseSensitive: Boolean;
    FDuplicates: TDuplicates;
    FList: PStringItemList;
    FOwnsObjects: Boolean;
    FOnChange: TNotifyEvent;
    FOnChanging: TNotifyEvent;
    FStrictDelimiter: Boolean;
    procedure Grow;
    procedure QuickSort(L, R: Integer; SCompare: TRawByteStringsSortCompare);
    procedure SetCaseSensitive(const Value: Boolean);
    procedure SetSorted(Value: Boolean);
    function GetCommaText: RawByteString;
    procedure SetCommaText(const Value: RawByteString);
    function GetName(Index: Integer): RawByteString;
    function GetValue(const Name: RawByteString): RawByteString;
    procedure SetValue(const Name, Value: RawByteString);
    function GetValueFromIndex(Index: Integer): RawByteString;
    procedure SetValueFromIndex(Index: Integer; const Value: RawByteString);
  protected
    function ExtractName(const S: RawByteString): RawByteString;
    function GetDelimitedText: RawByteString; virtual;
    procedure SetDelimitedText(const Value: RawByteString); virtual;
    function Get(Index: Integer): RawByteString; override;
    procedure Put(Index: Integer; const S: RawByteString); override;
    {$IFDEF CONDITIONALEXPRESSIONS}
    function GetCapacity: Integer; override;
    procedure SetCapacity(NewCapacity: Integer); override;
    function CompareStrings(const S1, S2: RawByteString): Integer; override;
    {$ELSE}
    function GetCapacity: Integer; virtual;
    procedure SetCapacity(NewCapacity: Integer); virtual;
    function CompareStrings(const S1, S2: RawByteString): Integer; virtual;
    {$ENDIF}
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    function GetString(Index: Integer): RawByteString; virtual;
    procedure PutString(Index: Integer; const S: RawByteString); virtual;
    function GetTextStr: RawByteString; override;
    procedure SetTextStr(const Value: RawByteString); override;
    procedure ExchangeItems(Index1, Index2: Integer); virtual;
    procedure InsertItem(Index: Integer; const S: RawByteString; AObject: TObject); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    function GetText: PAnsiChar; override;
    procedure SetText(Text: PAnsiChar); override;
    function Add(const S: RawByteString): Integer; override;
    function AddObject(const S: RawByteString; AObject: TObject): Integer; override;
    procedure Append(const S: RawByteString);
    procedure Changed; virtual;
    procedure Changing; virtual;
    procedure Clear; override;
    procedure CustomSort(Compare: TRawByteStringsSortCompare); virtual;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function Find(const S: RawByteString; var Index: Integer): Boolean; virtual;
    function IndexOf(const S: RawByteString): Integer; override;
    procedure Insert(Index: Integer; const S: RawByteString); override;
    {$IFDEF CONDITIONALEXPRESSIONS}
    function IndexOfName(const Name: RawByteString): Integer; override;
    procedure InsertObject(Index: Integer; const S: RawByteString; AObject: TObject); override;
    {$ELSE}
    function IndexOfName(const Name: RawByteString): Integer; virtual;
    procedure InsertObject(Index: Integer; const S: RawByteString; AObject: TObject); virtual;
    {$ENDIF}
    procedure Move(CurIndex, NewIndex: Integer); override;
    procedure Sort; virtual;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    function SetListString(Chars: PAnsiChar; Length: Integer): Integer;
    property Delimiter: AnsiChar read FDelimiter write FDelimiter;
    property QuoteChar: AnsiChar read FQuoteChar write FQuoteChar;
    property NameValueSeparator: AnsiChar read FNameValueSeparator write FNameValueSeparator;
    property LineBreak: RawByteString read FLineBreak write FLineBreak;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive;
    property DelimitedText: RawByteString read GetDelimitedText write SetDelimitedText;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
    property OwnsObjects: Boolean read FOwnsObjects write FOwnsObjects;
    property Sorted: Boolean read FSorted write SetSorted;
    property StrictDelimiter: Boolean read FStrictDelimiter write FStrictDelimiter;
    property Text: RawByteString read GetTextStr write SetTextStr;
    property CommaText: RawByteString read GetCommaText write SetCommaText;
    property Strings[Index: Integer]: RawByteString read GetString write PutString; default;
    property Names[Index: Integer]: RawByteString read GetName;
    property Values[const Name: RawByteString]: RawByteString read GetValue write SetValue;
    property ValueFromIndex[Index: Integer]: RawByteString read GetValueFromIndex write SetValueFromIndex;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;


implementation

uses
  {$IFDEF CONDITIONALEXPRESSIONS}
  RtlConsts,
  {$ELSE}
  Consts,
  {$ENDIF}
  StrUtil;


{$IFDEF CONDITIONALEXPRESSIONS}
procedure ListError(const Error: String);
begin
  Raise EListError.Create(Error);
end;
{$ELSE}
procedure ListError(Error: Integer);
begin
  Raise EListError.CreateRes(Error);
end;
{$ENDIF}

// adapted from Delphi 2010
function QuotedStr(const S: String; Quote: Char): string;
var
  P, Src, Dest: PChar;
  AddCount: Integer;
begin
  AddCount := 0;
  P := StrScan(PChar(S), Quote);
  while P <> nil do begin
    Inc(P);
    Inc(AddCount);
    P := StrScan(P, Quote);
  end;
  if AddCount = 0 then begin
    Result := Quote + S + Quote;
    Exit;
  end;
  SetLength(Result, Length(S) + AddCount + 2);
  Dest := PChar(Result);
  Dest^ := Quote;
  Inc(Dest);
  Src := PChar(S);
  P := StrScan(Src, Quote);
  repeat
    Inc(P);
    Move(Src^, Dest^, (P - Src) * SizeOf(Char));
    Inc(Dest, P - Src);
    Dest^ := Quote;
    Inc(Dest);
    Src := P;
    P := StrScan(Src, Quote);
  until P = nil;
  P := StrEnd(Src);
  Move(Src^, Dest^, (P - Src) * SizeOf(Char));
  Inc(Dest, P - Src);
  Dest^ := Quote;
end;

// adapted from Delphi 2010
function ExtractQuotedStr(var Src: PAnsiChar; Quote: AnsiChar): AnsiString;
var
  P, Dest: PAnsiChar;
  DropCount: Integer;
begin
  Result := '';
  if (Src = nil) or (Src^ <> Quote) then Exit;
  Inc(Src);
  DropCount := 1;
  P := Src;
  Src := StrScan(Src, Quote);
  // count adjacent pairs of quote chars
  while Src <> nil do begin
    Inc(Src);
    if Src^ <> Quote then Break;
    Inc(Src);
    Inc(DropCount);
    Src := StrScan(Src, Quote);
  end;
  if Src = nil then Src := StrEnd(P);
  if ((Src - P) <= 1) or ((Src - P - DropCount) = 0) then Exit;
  if DropCount = 1 then
    SetString(Result, P, Src - P - 1)
  else begin
    SetLength(Result, Src - P - DropCount);
    Dest := PAnsiChar(Result);
    Src := StrScan(P, Quote);
    while Src <> nil do begin
      Inc(Src);
      if Src^ <> Quote then Break;
      Move(P^, Dest^, Src - P);
      Inc(Dest, Src - P);
      Inc(Src);
      P := Src;
      Src := StrScan(Src, Quote);
    end;
    if Src = nil then Src := StrEnd(P);
    Move(P^, Dest^, Src - P - 1);
  end;
end;


// TRawByteStrings -------------------------------------------------------------

function RawByteStringsCompare(List: TRawByteStrings; Index1, Index2: Integer): Integer;
begin
  Result := List.CompareStrings(List.FList^[Index1].FString, List.FList^[Index2].FString);
end;

procedure TRawByteStrings.Grow;
var
  iDelta: Integer;
begin
  if FCapacity > 64 then
    iDelta := FCapacity div 4
  else if FCapacity > 8 then
    iDelta := 16
  else iDelta := 4;
  SetCapacity(FCapacity + iDelta);
end;

procedure TRawByteStrings.QuickSort(L, R: Integer; SCompare: TRawByteStringsSortCompare);
var
  I, J, M: Integer;
begin
  repeat
    I := L;
    J := R;
    M := (L + R) shr 1;
    repeat
      while SCompare(Self, I, M) < 0 do inc(I);
      while SCompare(Self, J, M) > 0 do dec(J);
      if I <= J then begin
	if I <> J then ExchangeItems(I, J);
	if M = I then
	  M := J
	else if M = J then
	  M := I;
	inc(I);
	dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TRawByteStrings.SetCaseSensitive(const Value: Boolean);
begin
  if Value <> FCaseSensitive then begin
    FCaseSensitive := Value;
    if FSorted then begin
      Sorted := FALSE;
      Sorted := TRUE;
    end;
  end;
end;

procedure TRawByteStrings.SetSorted(Value: Boolean);
begin
  if FSorted <> Value then begin
    if Value then Sort;
    FSorted := Value;
  end;
end;

function TRawByteStrings.GetCommaText: RawByteString;
var
  chrDelimiter: AnsiChar;
  chrQuoteChar: AnsiChar;
begin
  chrDelimiter := FDelimiter;
  chrQuoteChar := FQuoteChar;
  FDelimiter := CM;
  FQuoteChar := QU;
  try
    Result := GetDelimitedText;
  finally
    FDelimiter := chrDelimiter;
    FQuoteChar := chrQuoteChar;
  end;
end;

procedure TRawByteStrings.SetCommaText(const Value: RawByteString);
begin
  FDelimiter := CM;
  FQuoteChar := QU;
  SetDelimitedText(Value);
end;

function TRawByteStrings.GetName(Index: Integer): RawByteString;
begin
  Result := ExtractName(GetString(Index));
end;

function TRawByteStrings.GetValue(const Name: RawByteString): RawByteString;
var
  iIndex: Integer;
begin
  iIndex := IndexOfName(Name);
  if iIndex >= 0 then
    Result := Copy(GetString(iIndex), Length(Name) + 2, MAXINT)
  else Result := E;
end;

procedure TRawByteStrings.SetValue(const Name, Value: RawByteString);
var
  iIndex: Integer;
begin
  iIndex := IndexOfName(Name);
  if Value <> E then begin
    if iIndex < 0 then iIndex := Add(E);
    PutString(iIndex, Name + FNameValueSeparator + Value);
  end
  else if iIndex >= 0 then Delete(iIndex);
end;

function TRawByteStrings.GetValueFromIndex(Index: Integer): RawByteString;
var
  iLen: Integer;
begin
  if Index >= 0 then begin
    iLen := Length(Names[Index]);
    if iLen > 0 then begin
      Result := Copy(GetString(Index), iLen + 2, MAXINT);
      Exit;
    end;
  end;
  Result := E;
end;

procedure TRawByteStrings.SetValueFromIndex(Index: Integer; const Value: RawByteString);
begin
  if Value <> E then begin
    if Index < 0 then Index := Add(E);
    PutString(Index, Names[Index] + FNameValueSeparator + Value);
  end
  else if Index >= 0 then Delete(Index);
end;

function TRawByteStrings.ExtractName(const S: RawByteString): RawByteString;
var
  iPos: Integer;
begin
  Result := S;
  iPos := Pos(FNameValueSeparator, Result);
  if iPos <> 0 then
    SetLength(Result, pred(iPos))
  else SetLength(Result, 0);
end;

function TRawByteStrings.GetDelimitedText: RawByteString;
var
  iIndex: Integer;
  scsDelimiters: Set of Char;
  s: RawByteString;
  p: PAnsiChar;
begin
  if (FCount = 1) and (Get(0) = E) then
    Result := FQuoteChar + FQuoteChar
  else begin
    Result := E;
    scsDelimiters := [NULL, FQuoteChar, FDelimiter];
    if not FStrictDelimiter then scsDelimiters := scsDelimiters + TRIMCHARS;
    for iIndex := 0 to pred(FCount) do begin
      s := GetString(iIndex);
      p := PAnsiChar(s);
      while not (p^ in scsDelimiters) do inc(p);
      if (p^ <> NULL) then s := QuotedStr(s, FQuoteChar);
      Result := Result + s + FDelimiter;
    end;
    System.Delete(Result, Length(Result), 1);
  end;
end;

procedure TRawByteStrings.SetDelimitedText(const Value: RawByteString);
var
  p1, p2: PAnsiChar;
  s: RawByteString;
begin
  BeginUpdate;
  try
    Clear;
    p1 := PAnsiChar(Value);
    if not FStrictDelimiter then while (p1^ in TRIMCHARS) do inc(p1);
    while p1^ <> NULL do begin
      if p1^ = FQuoteChar then
	s := ExtractQuotedStr(p1, FQuoteChar)
      else begin
	p2 := p1;
	while ((not FStrictDelimiter and (p1^ > SP)) or (FStrictDelimiter and (p1^ <> NULL))) and (p1^ <> FDelimiter) do inc(p1);
	SetString(s, p2, p1 - p2);
      end;
      Add(s);
      if not FStrictDelimiter then while (p1^ in TRIMCHARS) do inc(p1);
      if p1^ = FDelimiter then begin
	p2 := p1;
	inc(p2);
	if p2^ = NULL then Add(E);
	repeat
	  inc(p1);
	until not (not FStrictDelimiter and (p1^ in TRIMCHARS));
      end;
    end;
  finally
    EndUpdate;
  end;
end;

function TRawByteStrings.Get(Index: Integer): RawByteString;
begin
  if Cardinal(Index) >= Cardinal(FCount) then ListError(SListIndexError);
  Result := FList^[Index].FString;
end;

procedure TRawByteStrings.Put(Index: Integer; const S: RawByteString);
begin
  if FSorted then ListError(SSortedListError);
  if Cardinal(Index) >= Cardinal(FCount) then ListError(SListIndexError);
  Changing;
  FList^[Index].FString := S;
  Changed;
end;

function TRawByteStrings.GetCapacity: Integer;
begin
  Result := FCapacity;
end;

procedure TRawByteStrings.SetCapacity(NewCapacity: Integer);
begin
  if NewCapacity <> FCapacity then begin
    ReallocMem(FList, NewCapacity * SizeOf(TStringItem));
    FCapacity := NewCapacity;
  end;
end;

function TRawByteStrings.GetCount: Integer;
begin
  Result := FCount;
end;

function TRawByteStrings.GetObject(Index: Integer): TObject;
begin
  if Cardinal(Index) >= Cardinal(FCount) then ListError(SListIndexError);
  Result := FList^[Index].FObject;
end;

procedure TRawByteStrings.PutObject(Index: Integer; AObject: TObject);
begin
  if Cardinal(Index) >= Cardinal(FCount) then ListError(SListIndexError);
  Changing;
  FList^[Index].FObject := AObject;
  Changed;
end;

function TRawByteStrings.GetString(Index: Integer): RawByteString;
begin
  if Cardinal(Index) >= Cardinal(FCount) then ListError(SListIndexError);
  Result := FList^[Index].FString;
end;

procedure TRawByteStrings.PutString(Index: Integer; const S: RawByteString);
begin
  if FSorted then ListError(SSortedListError);
  if Cardinal(Index) >= Cardinal(FCount) then ListError(SListIndexError);
  Changing;
  FList^[Index].FString := S;
  Changed;
end;

function TRawByteStrings.GetTextStr: RawByteString;
var
  sItem: RawByteString;
  iSize, iIndex, iLen: Integer;
  pResult: PAnsiChar;
begin
  iSize := 0;
  for iIndex := 0 to pred(FCount) do inc(iSize, Length(Get(iIndex)) + Length(FLineBreak));
  SetString(Result, NIL, iSize);
  pResult := Pointer(Result);
  for iIndex := 0 to pred(FCount) do begin
    sItem := GetString(iIndex);
    iLen := Length(sItem);
    if iLen <> 0 then begin
      System.Move(Pointer(sItem)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    iLen := Length(FLineBreak);
    if iLen <> 0 then begin
      System.Move(Pointer(FLineBreak)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
  end;
end;

procedure TRawByteStrings.SetTextStr(const Value: RawByteString);
var
  pLineBreak, pValue, pStart, pLB: PAnsiChar;
  sItem: RawByteString;
  iLen: Integer;
begin
  BeginUpdate;
  try
    Clear;
    pLineBreak := PAnsiChar(FLineBreak);
    pValue := Pointer(Value);
    if pValue <> NIL then
      if AnsiCompareStr(FLineBreak, CRLF) = 0 then begin
	while pValue^ <> NULL do begin
	  pStart := pValue;
	  while not (pValue^ in [NULL, LF, CR]) do inc(pValue);
	  SetString(sItem, pStart, pValue - pStart);
	  Add(sItem);
	  if pValue^ = CR then inc(pValue);
	  if pValue^ = LF then inc(pValue);
	end;
      end
      else begin
	iLen := Length(FLineBreak);
	while pValue^ <> NULL do begin
	  pStart := pValue;
	  pLB := StrPos(pValue, pLineBreak);
	  while (pValue^ <> NULL) and (pValue <> pLB) do inc(pValue);
	  SetString(sItem, pStart, pValue - pStart);
	  Add(sItem);
	  if pValue = pLB then inc(pValue, iLen);
	end;
      end;
  finally
    EndUpdate;
  end;
end;

function TRawByteStrings.CompareStrings(const S1, S2: RawByteString): Integer;
begin
  if FCaseSensitive then
    Result := AnsiCompareStr(S1, S2)
  else Result := AnsiCompareText(S1, S2);
end;

procedure TRawByteStrings.ExchangeItems(Index1, Index2: Integer);
var
  si1, si2: PStringItem;
  p: Pointer;
begin
  si1 := @FList^[Index1];
  si2 := @FList^[Index2];
  p := Pointer(si1^.FString);
  Pointer(si1^.FString) := Pointer(si2^.FString);
  Pointer(si2^.FString) := p;
  p := si1^.FObject;
  si1^.FObject := si2^.FObject;
  si2^.FObject := p;
end;

procedure TRawByteStrings.InsertItem(Index: Integer; const S: RawByteString; AObject: TObject);
begin
  Changing;
  if FCount = FCapacity then Grow;
  if Index < FCount then System.Move(FList^[Index], FList^[Index + 1], (FCount - Index) * SizeOf(TStringItem));
  with FList^[Index] do begin
    Pointer(FString) := NIL;
    FObject := AObject;
    FString := S;
  end;
  inc(FCount);
  Changed;
end;

constructor TRawByteStrings.Create;
begin
  inherited Create;
  FDelimiter := CM;
  FQuoteChar := QU;
  FNameValueSeparator := EQ;
  FLineBreak := CRLF;
  FStrictDelimiter := TRUE;
end;

destructor TRawByteStrings.Destroy;
var
  iIndex: Integer;
begin
  FOnChanging := NIL;
  FOnChange := NIL;
  if FCount <> 0 then begin
    if FOwnsObjects then for iIndex := 0 to pred(FCount) do GetObject(iIndex).Free;
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
  end;
  inherited Destroy;
end;

function TRawByteStrings.GetText: PAnsiChar;
begin
  Result := StrNew(PAnsiChar(GetTextStr));
end;

procedure TRawByteStrings.SetText(Text: PAnsiChar);
begin
  SetTextStr(Text);
end;

function TRawByteStrings.Add(const S: RawByteString): Integer;
begin
  Result := AddObject(S, NIL);
end;

function TRawByteStrings.AddObject(const S: RawByteString; AObject: TObject): Integer;
begin
  if not FSorted then
    Result := FCount
  else if Find(S, Result) then begin
    case FDuplicates of
      dupIgnore: Exit;
      dupError: ListError(SDuplicateString);
    end;
  end;
  InsertItem(Result, S, AObject);
end;

procedure TRawByteStrings.Append(const S: RawByteString);
begin
  Add(S);
end;

procedure TRawByteStrings.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TRawByteStrings.Changing;
begin
  if Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TRawByteStrings.Clear;
var
  iIndex: Integer;
begin
  if FCount <> 0 then begin
    Changing;
    if FOwnsObjects then for iIndex := 0 to pred(FCount) do GetObject(iIndex).Free;
    Finalize(FList^[0], FCount);
    FCount := 0;
    SetCapacity(0);
    Changed;
  end;
end;

procedure TRawByteStrings.CustomSort(Compare: TRawByteStringsSortCompare);
begin
  if not FSorted and (FCount > 1) then begin
    Changing;
    QuickSort(0, FCount - 1, Compare);
    Changed;
  end;
end;

procedure TRawByteStrings.Delete(Index: Integer);
begin
  if (Index < 0) or (Index >= FCount) then ListError(SListIndexError);
  Changing;
  if FOwnsObjects then GetObject(Index).Free;
  Finalize(FList^[Index]);
  dec(FCount);
  if Index < FCount then System.Move(FList^[Index + 1], FList^[Index], (FCount - Index) * SizeOf(TStringItem));
  Changed;
end;

procedure TRawByteStrings.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) then ListError(SListIndexError);
  if (Index2 < 0) or (Index2 >= FCount) then ListError(SListIndexError);
  Changing;
  ExchangeItems(Index1, Index2);
  Changed;
end;

function TRawByteStrings.Find(const S: RawByteString; var Index: Integer): Boolean;
var
  L, H, M, C: Integer;
begin
  Result := FALSE;
  L := 0;
  H := pred(FCount);
  while L <= H do begin
    M := (L + H) shr 1;
    C := CompareStrings(FList^[M].FString, S);
    if C < 0 then
      L := succ(M)
    else begin
      H := pred(M);
      if C = 0 then begin
	Result := TRUE;
	if FDuplicates <> dupAccept then L := M;
      end;
    end;
  end;
  Index := L;
end;

function TRawByteStrings.IndexOf(const S: RawByteString): Integer;
begin
  if FSorted then begin
    if Find(S, Result) then Exit;
  end
  else for Result := 0 to pred(FCount) do if CompareStrings(GetString(Result), S) = 0 then Exit;
  Result := -1;
end;

function TRawByteStrings.IndexOfName(const Name: RawByteString): Integer;
var
  sItem: RawByteString;
  iPos: Integer;
begin
  for Result := 0 to pred(FCount) do begin
    sItem := GetString(Result);
    iPos := Pos(FNameValueSeparator, sItem);
    if (iPos <> 0) and (CompareStrings(Copy(sItem, 1, pred(iPos)), Name) = 0) then Exit;
  end;
  Result := -1;
end;

procedure TRawByteStrings.Insert(Index: Integer; const S: RawByteString);
begin
  InsertObject(Index, S, NIL);
end;

procedure TRawByteStrings.InsertObject(Index: Integer; const S: RawByteString; AObject: TObject);
begin
  if FSorted then ListError(SSortedListError);
  if (Index < 0) or (Index > FCount) then ListError(SListIndexError);
  InsertItem(Index, S, AObject);
end;

procedure TRawByteStrings.Move(CurIndex, NewIndex: Integer);
var
  sTemp: RawByteString;
  objTemp: TObject;
begin
  if CurIndex <> NewIndex then begin
    BeginUpdate;
    try
      sTemp := GetString(CurIndex);
      objTemp := GetObject(CurIndex);
      Delete(CurIndex);
      InsertObject(NewIndex, sTemp, objTemp);
    finally
      EndUpdate;
    end;
  end;
end;

procedure TRawByteStrings.Sort;
begin
  CustomSort(RawByteStringsCompare);
end;

procedure TRawByteStrings.LoadFromStream(Stream: TStream);
var
  iSize: Integer;
  sBuffer: RawByteString;
begin
  with Stream do iSize := Size - Position;
  if iSize > 0 then begin
    SetLength(sBuffer, iSize);
    Stream.ReadBuffer(Pointer(sBuffer)^, iSize);
    SetTextStr(sBuffer);
  end;
end;

procedure TRawByteStrings.SaveToStream(Stream: TStream);
var
  sBuffer: RawByteString;
begin
  sBuffer := GetTextStr;
  if sBuffer <> E then Stream.WriteBuffer(Pointer(sBuffer)^, Length(sBuffer));
end;

function TRawByteStrings.SetListString(Chars: PAnsiChar; Length: Integer): Integer;
var
  sNew: RawByteString;
begin
  SetString(sNew, Chars, Length);
  Result := Add(sNew);
end;


end.

