{$J+,Z4}
unit Tidy;
{------------------------------------------------------------------------------}
{    Part of this code is copyright (C) 2006-2023 by Michael in der Wiesche    }
{------------------------------------------------------------------------------}

interface

uses
  Windows,
  StrUtil,
  TidyEnum;


(** @file buffio.h - Treat buffer as an I/O stream.

  (c) 1998-2007 (W3C) MIT, ERCIM, Keio University
  See tidy.h for the copyright notice.
  Requires buffer to automatically grow as bytes are added.
  Must keep track of current read and write points.
*)

type

(** Prototype for the allocator's function table *)
(** The allocators function table *)
  TTidyAllocatorVtbl = Record
    AllocProc: TFarProc;
    ReallocProc: TFarProc;
    FreeProc: TFarProc;
    PanicProc: TFarProc;
  end;

(** Prototype for the allocator *)
(** The allocator **)
  TTidyAllocator = Record
    vtbl: TTidyAllocatorVtbl;
  end;

(** TidyBuffer - A chunk of memory *)
  TTidyBuffer = Record
    allocator: TTidyAllocator;		(**< Memory allocator *)
    bp: Pointer;			(**< Pointer to bytes *)
    size: UInt;				(**< # bytes currently in use *)
    allocated: UInt;			(**< # bytes allocated *)
    next: UInt;				(**< Offset of current input position *)
  end;

var

(** Initialize data structure using the default allocator *)
  tidyBufInit: procedure( var buf: TTidyBuffer ); cdecl;

(** Initialize data structure using the given custom allocator *)
  tidyBufInitWithAllocator: procedure( var buf: TTidyBuffer; const allocator: TTidyAllocator ); cdecl;

(** Free current buffer, allocate given amount, reset input pointer, use the default allocator *)
  tidyBufAlloc: procedure( var buf: TTidyBuffer; allocSize: UInt ); cdecl;

(** Free current buffer, allocate given amount, reset input pointer, use the given custom allocator *)
  tidyBufAllocWithAllocator: procedure( var buf: TTidyBuffer;
					const allocator: TTidyAllocator;
					allocSize: UInt ); cdecl;

(** Expand buffer to given size.
**  Chunk size is minimum growth. Pass 0 for default of 256 bytes.
*)
  tidyBufCheckAlloc: procedure( var buf: TTidyBuffer;
				allocSize: UInt;
				chunkSize: UInt ); cdecl;

(** Free current contents and zero out *)
  tidyBufFree: procedure( var buf: TTidyBuffer ); cdecl;

(** Set buffer bytes to 0 *)
  tidyBufClear: procedure( var buf: TTidyBuffer ); cdecl;

(** Attach to existing buffer *)
  tidyBufAttach: procedure( var buf: TTidyBuffer; bp: Pointer; size: UInt ); cdecl;

(** Detach from buffer.  Caller must free. *)
  tidyBufDetach: procedure( var buf: TTidyBuffer ); cdecl;

(** Append bytes to buffer.  Expand if necessary. *)
  tidyBufAppend: procedure( var buf: TTidyBuffer; vp: Pointer; size: UInt ); cdecl;

(** Append one byte to buffer.  Expand if necessary. *)
  tidyBufPutByte: procedure( var buf: TTidyBuffer; bv: Byte ); cdecl;

(** Get byte from end of buffer *)
  tidyBufPopByte: function( var buf: TTidyBuffer ): Integer; cdecl;

(** Get byte from front of buffer.  Increment input offset. *)
  tidyBufGetByte: function( var buf: TTidyBuffer ): Integer; cdecl;

(** At end of buffer? *)
  tidyBufEndOfInput: function( var buf: TTidyBuffer ): Bool; cdecl;

(** Put a byte back into the buffer.  Decrement input offset. *)
  tidyBufUngetByte: procedure( var buf: TTidyBuffer; bv: Byte ); cdecl;


(** @file tidy.h - Defines HTML Tidy API implemented by tidy library.

  Public interface is const-correct and doesn't explicitly depend
  on any globals.  Thus, thread-safety may be introduced w/out
  changing the interface.

  Looking ahead to a C++ wrapper, C functions always pass
  this-equivalent as 1st arg.


  Copyright (c) 1998-2006 World Wide Web Consortium
  (Massachusetts Institute of Technology, European Research
  Consortium for Informatics and Mathematics, Keio University).
  All Rights Reserved.

  COPYRIGHT NOTICE:

  This software and documentation is provided "as is," and
  the copyright holders and contributing author(s) make no
  representations or warranties, express or implied, including
  but not limited to, warranties of merchantability or fitness
  for any particular purpose or that the use of the software or
  documentation will not infringe any third party patents,
  copyrights, trademarks or other rights.

  The copyright holders and contributing author(s) will not be held
  liable for any direct, indirect, special or consequential damages
  arising out of any use of the software or documentation, even if
  advised of the possibility of such damage.

  Permission is hereby granted to use, copy, modify, and distribute
  this source code, or portions hereof, documentation and executables,
  for any purpose, without fee, subject to the following restrictions:

  1. The origin of this source code must not be misrepresented.
  2. Altered versions must be plainly marked as such and must
     not be misrepresented as being the original source.
  3. This Copyright notice may not be removed or altered from any
     source or altered source distribution.

  The copyright holders and contributing author(s) specifically
  permit, without fee, and encourage the use of this source code
  as a component for supporting the Hypertext Markup Language in
  commercial products. If you use this source code in a product,
  acknowledgment is not required but would be appreciated.


  Created 2001-05-20 by Charles Reitzel
  Updated 2002-07-01 by Charles Reitzel - 1st Implementation

*)


(** @defgroup Opaque Opaque Types
**
** Cast to implementation types within lib.
** Reduces inter-dependencies/conflicts w/ application code.
** @{
*)

type

(** @struct TidyDoc
**  Opaque document datatype
*)
  TTidyDoc = Pointer;

(** @struct TidyOption
**  Opaque option datatype
*)
  TTidyOption = Pointer;

(** @struct TidyNode
**  Opaque node datatype
*)
  TTidyNode = Pointer;

(** @struct TidyAttr
**  Opaque attribute datatype
*)
  TTidyAttr = Pointer;

(** @} *)


(** @defgroup Basic Basic Operations
**
** Tidy public interface
**
** Several functions return an integer document status:
**
** <pre>
** 0    -> SUCCESS
** >0   -> 1 == TIDY WARNING, 2 == TIDY ERROR
** <0   -> SEVERE ERROR
** </pre>
**
The following is a short example program.

<pre>
#include <tidy.h>
#include <buffio.h>
#include <stdio.h>
#include <errno.h>


Integer main(Integer argc, char **argv )
{
  const char* input = "<title>Foo</title><p>Foo!";
  TidyBuffer output = {0};
  TidyBuffer errbuf = {0};
  Integer rc = -1;
  Bool ok;

  tdoc: TidyDoc = tidyCreate();                     // Initialize "document"
  printf( "Tidying:\t\%s\\n", input );

  ok = tidyOptSetBool( tdoc, TidyXhtmlOut, yes );  // Convert to XHTML
  if ( ok )
    rc = tidySetErrorBuffer( tdoc, &errbuf );      // Capture diagnostics
  if ( rc >= 0 )
    rc = tidyParseString( tdoc, input );           // Parse the input
  if ( rc >= 0 )
    rc = tidyCleanAndRepair( tdoc );               // Tidy it up!
  if ( rc >= 0 )
    rc = tidyRunDiagnostics( tdoc );               // Kvetch
  if ( rc > 1 )                                    // If error, force output.
    rc = ( tidyOptSetBool(tdoc, TidyForceOutput, yes) ? rc : -1 );
  if ( rc >= 0 )
    rc = tidySaveBuffer( tdoc, &output );          // Pretty Print

  if ( rc >= 0 )
  {
    if ( rc > 0 )
      printf( "\\nDiagnostics:\\n\\n\%s", errbuf.bp );
    printf( "\\nAnd here is the result:\\n\\n\%s", output.bp );
  }
  else
    printf( "A severe error (\%d) occurred.\\n", rc );

  tidyBufFree( &output );
  tidyBufFree( &errbuf );
  tidyRelease( tdoc );
  return rc;
}
</pre>
** @{
*)

var

  tidyCreate: function: TTidyDoc; cdecl;
  tidyCreateWithAllocator: function( const allocator: TTidyAllocator ): TTidyDoc; cdecl;
  tidyRelease: procedure( tdoc: TTidyDoc ); cdecl;

(** Let application store a chunk of data w/ each Tidy instance.
**  Useful for callbacks.
*)
  tidySetAppData: procedure( tdoc: TTidyDoc; appData: Pointer ); cdecl;

(** Get application data set previously *)
  tidyGetAppData: function( tdoc: TTidyDoc ): Pointer; cdecl;

(** Get release date (version) for current library *)
  tidyReleaseDate: function: PChar; cdecl;

(* Diagnostics and Repair
*)

(** Get status of current document. *)
  tidyStatus: function( tdoc: TTidyDoc ): Integer; cdecl;

(** Detected HTML version: 0, 2, 3 or 4 *)
  tidyDetectedHtmlVersion: function( tdoc: TTidyDoc ): Integer; cdecl;

(** Input is XHTML? *)
  tidyDetectedXhtml: function( tdoc: TTidyDoc ): Bool; cdecl;

(** Input is generic XML (not HTML or XHTML)? *)
  tidyDetectedGenericXml: function( tdoc: TTidyDoc ): Bool; cdecl;

(** Number of Tidy errors encountered.  If > 0, output is suppressed
**  unless TidyForceOutput is set.
*)
  tidyErrorCount: function( tdoc: TTidyDoc ): UInt; cdecl;

(** Number of Tidy warnings encountered. *)
  tidyWarningCount: function( tdoc: TTidyDoc ): UInt; cdecl;

(** Number of Tidy accessibility warnings encountered. *)
  tidyAccessWarningCount: function( tdoc: TTidyDoc ): UInt; cdecl;

(** Number of Tidy configuration errors encountered. *)
  tidyConfigErrorCount: function( tdoc: TTidyDoc ): UInt; cdecl;

(* Get/Set configuration options
*)
(** Load an ASCII Tidy configuration file *)
  tidyLoadConfig: function( tdoc: TTidyDoc; configFile: PChar ): Integer; cdecl;

(** Load a Tidy configuration file with the specified character encoding *)
  tidyLoadConfigEnc: function( tdoc: TTidyDoc; configFile: PChar; charenc: PChar ): Integer; cdecl;

  tidyFileExists: function( filename: PChar ): Bool; cdecl;


(** Set the input/output character encoding for parsing markup.
**  Values include: ascii, latin1, raw, utf8, iso2022, mac,
**  win1252, utf16le, utf16be, utf16, big5 and shiftjis.  Case in-sensitive.
*)
  tidySetCharEncoding: function( tdoc: TTidyDoc; encnam: PChar ): Integer; cdecl;

(** Set the input encoding for parsing markup.
** As for tidySetCharEncoding but only affects the input encoding
*)
  tidySetInCharEncoding: function( tdoc: TTidyDoc; encnam: PChar ): Integer; cdecl;

(** Set the output encoding.
*)
  tidySetOutCharEncoding: function( tdoc: TTidyDoc; encnam: PChar ): Integer; cdecl;

(** @} end Basic group *)


(** @defgroup Configuration Configuration Options
**
** Functions for getting and setting Tidy configuration options.
** @{
*)

(** Applications using TidyLib may want to augment command-line and
**  configuration file options.  Setting this callback allows an application
**  developer to examine command-line and configuration file options after
**  TidyLib has examined them and failed to recognize them.
*)

type

  PTidyOptCallback = ^TTidyOptCallback;
  TTidyOptCallback = function ( Option: PChar; Value: PChar ): Bool; cdecl;

var

  tidySetOptionCallback: function( tdoc: TTidyDoc; pOptCallback: PTidyOptCallback ): Bool; cdecl;

(** Get option ID by name *)
  tidyOptGetIdForName: function( optnam: PChar ): TTidyOptionId; cdecl;

(** Get iterator for list of option *)
(**
Example:
<pre>
PTidyIterator itOpt = tidyGetOptionList( tdoc );
while ( itOpt )
{
  opt: PTidyOption = tidyGetNextOption( tdoc, &itOpt );
  .. get/set option values ..
}
</pre>
*)

type

  TTidyIterator = Pointer;

var

  tidyGetOptionList: function( tdoc: TTidyDoc ): TTidyIterator; cdecl;
(** Get next Option *)
  tidyGetNextOption: function( tdoc: TTidyDoc; pos: TTidyIterator ): TTidyOption; cdecl;

(** Lookup option by ID *)
  tidyGetOption: function( tdoc: TTidyDoc; optId: TTidyOptionId ): TTidyOption; cdecl;
(** Lookup option by name *)
  tidyGetOptionByName: function( tdoc: TTidyDoc; optnam: PChar ): TTidyOption; cdecl;

(** Get ID of given Option *)
  tidyOptGetId: function( opt: TTidyOption ): TTidyOptionId; cdecl;

(** Get name of given Option *)
  tidyOptGetName: function( opt: TTidyOption ): PChar; cdecl;

(** Get datatype of given Option *)
  tidyOptGetType: function( opt: TTidyOption ): TTidyOptionType; cdecl;

(** Is Option read-only? *)
  tidyOptIsReadOnly: function( opt: TTidyOption ): Bool; cdecl;

(** Get category of given Option *)
  tidyOptGetCategory: function( opt: TTidyOption ): TTidyConfigCategory; cdecl;

(** Get default value of given Option as a string *)
  tidyOptGetDefault: function( opt: TTidyOption ): PChar; cdecl;

(** Get default value of given Option as an unsigned integer *)
  tidyOptGetDefaultInt: function( opt: TTidyOption ): DWord; cdecl;

(** Get default value of given Option as a Boolean value *)
  tidyOptGetDefaultBool: function( opt: TTidyOption ): Bool; cdecl;

(** Iterate over Option "pick list" *)
  tidyOptGetPickList: function( opt: TTidyOption ): TTidyIterator; cdecl;
(** Get next string value of Option "pick list" *)
  tidyOptGetNextPick: function( opt: TTidyOption; pos: TTidyIterator ): PChar; cdecl;

(** Get current Option value as a string *)
  tidyOptGetValue: function( tdoc: TTidyDoc; optId: TTidyOptionId ): PChar; cdecl;
(** Set Option value as a string *)
  tidyOptSetValue: function( tdoc: TTidyDoc; optId: TTidyOptionId; val: PChar ): Bool; cdecl;
(** Set named Option value as a string.  Good if not sure of type. *)
  tidyOptParseValue: function( tdoc: TTidyDoc; optnam: PChar; val: PChar ): Bool; cdecl;

(** Get current Option value as an integer *)
  tidyOptGetInt: function( tdoc: TTidyDoc; optId: TTidyOptionId ): DWord; cdecl;
(** Set Option value as an integer *)
  tidyOptSetInt: function( tdoc: TTidyDoc; optId: TTidyOptionId; val: DWord ): Bool; cdecl;

(** Get current Option value as a Boolean flag *)
  tidyOptGetBool: function( tdoc: TTidyDoc; optId: TTidyOptionId ): Bool; cdecl;
(** Set Option value as a Boolean flag *)
  tidyOptSetBool: function( tdoc: TTidyDoc; optId: TTidyOptionId; val: Bool ): Bool; cdecl;

(** Reset option to default value by ID *)
  tidyOptResetToDefault: function( tdoc: TTidyDoc; optId: TTidyOptionId ): Bool; cdecl;
(** Reset all options to their default values *)
  tidyOptResetAllToDefault: function( tdoc: TTidyDoc ): Bool; cdecl;

(** Take a snapshot of current config settings *)
  tidyOptSnapshot: function( tdoc: TTidyDoc ): Bool; cdecl;
(** Reset config settings to snapshot (after document processing) *)
  tidyOptResetToSnapshot: function( tdoc: TTidyDoc ): Bool; cdecl;

(** Any settings different than default? *)
  tidyOptDiffThanDefault: function( tdoc: TTidyDoc ): Bool; cdecl;
(** Any settings different than snapshot? *)
  tidyOptDiffThanSnapshot: function( tdoc: TTidyDoc ): Bool; cdecl;

(** Copy current configuration settings from one document to another *)
  tidyOptCopyConfig: function( tdocTo: TTidyDoc; tdocfrom: TTidyDoc ): Bool; cdecl;

(** Get character encoding name.  Used with TidyCharEncoding,
**  TidyOutCharEncoding, TidyInCharEncoding *)
  tidyOptGetEncName: function( tdoc: TTidyDoc; optId: TTidyOptionId ): PChar; cdecl;

(** Get current pick list value for option by ID.  Useful for enum types. *)
  tidyOptGetCurrPick: function( tdoc: TTidyDoc; optId: TTidyOptionId): PChar; cdecl;

(** Iterate over user declared tags *)
  tidyOptGetDeclTagList: function( tdoc: TTidyDoc ): TTidyIterator; cdecl;
(** Get next declared tag of specified type: TidyInlineTags, TidyBlockTags,
**  TidyEmptyTags, TidyPreTags *)
  tidyOptGetNextDeclTag: function( tdoc: TTidyDoc; optId: TTidyOptionId; iter: TTidyIterator ): PChar; cdecl;
(** Get option description *)
  tidyOptGetDoc: function( tdoc: TTidyDoc; opt: TTidyOption ): PChar; cdecl;

(** Iterate over a list of related options *)
  tidyOptGetDocLinksList: function( tdoc: TTidyDoc; opt: TTidyOption ): TTidyIterator; cdecl;
(** Get next related option *)
  tidyOptGetNextDocLinks: function( tdoc: TTidyDoc; pos: TTidyIterator ): TTidyOption; cdecl;


(** @} end Configuration group *)

(** @defgroup IO  I/O and Messages
**
** By default, Tidy will define, create and use
** instances of input and output handlers for
** standard C buffered I/O (i.e. FILE* stdin,
** FILE* stdout and FILE* stderr for content
** input, content output and diagnostic output,
** respectively.  A FILE* cfgFile input handler
** will be used for config files.  Command line
** options will just be set directly.
**
** @{
*)

(*****************
   Input Source
*****************)

type

(** Input Callback: get next byte of input *)
  PTidyGetByteFunc = ^TTidyGetByteFunc;
  TTidyGetByteFunc = function( sourceData: Pointer ): Integer; cdecl;

(** Input Callback: unget a byte of input *)
  PTidyUngetByteFunc = ^TTidyUngetByteFunc;
  TTidyUngetByteFunc = procedure( sourceData: Pointer; bt: Byte ); cdecl;

(** Input Callback: is end of input? *)
  PTidyEOFFunc = ^TTidyEOFFunc;
  TTidyEOFFunc = function( sourceData: Pointer ): Bool; cdecl;

const

(** End of input "character" *)
  EndOfStream = not (0);

type

(** TidyInputSource - Delivers raw bytes of input
*)
  TTidyInputSource = Record

    (* Instance data *)
    sourceData: Pointer;		(**< Input context.  Passed to callbacks *)

    (* Methods *)
    getByte: TTidyGetByteFunc;		(**< Pointer to "get byte" callback *)
    ungetByte: TTidyUngetByteFunc;	(**< Pointer to "unget" callback *)
    eof: TTidyEOFFunc;			(**< Pointer to "eof" callback *)

  end;

var

(** Initialize a buffer input source *)
  tidyInitInputBuffer: procedure( const inp: TTidyInputSource; var buf: TTidyBuffer ); cdecl;

(** Facilitates user defined source by providing
**  an entry point to marshal pointers-to-functions.
**  Needed by .NET and possibly other language bindings.
*)
  tidyInitSource: function( const source: TTidyInputSource; srcData: Pointer;
			    gbFunc: TTidyGetByteFunc;
			    ugbFunc: TTidyUngetByteFunc;
			    endFunc: TTidyEOFFunc ): Bool; cdecl;

(** Helper: get next byte from input source *)
  tidyGetByte: function( const source: TTidyInputSource ): UInt; cdecl;

(** Helper: unget byte back to input source *)
  tidyUngetByte: procedure( const source: TTidyInputSource; byteValue: UInt ); cdecl;

(** Helper: check if input source at end *)
  tidyIsEOF: function( const source: TTidyInputSource ): Bool; cdecl;


(****************
   Output Sink
****************)

type

(** Output callback: send a byte to output *)
  PTidyPutByteFunc = ^TTidyPutByteFunc;
  TTidyPutByteFunc = procedure( sinkData: Pointer; bt: Byte ); cdecl;

(** TidyOutputSink - accepts raw bytes of output
*)
  TTidyOutputSink = Record

    (* Instance data *)
    sinkData: Pointer;			(**< Output context.  Passed to callbacks *)

    (* Methods *)
    putByte: TTidyPutByteFunc;		(**< Pointer to "put byte" callback *)

  end;

var

(** Initialize a buffer output sink *)
  tidyInitOutputBuffer: procedure( const outp : TTidyOutputSink; var buf: TTidyBuffer ); cdecl;

(** Facilitates user defined sinks by providing
**  an entry point to marshal pointers-to-functions.
**  Needed by .NET and possibly other language bindings.
*)
  tidyInitSink: function( const sink: TTidyOutputSink; snkData: Pointer; pbFunc: TTidyPutByteFunc ): Bool; cdecl;

(** Helper: send a byte to output *)
  tidyPutByte: procedure( const sink: TTidyOutputSink; byteValue: UInt ); cdecl;

type

(** Callback to filter messages by diagnostic level:
**  info, warning, etc.  Just set diagnostic output
**  handler to redirect all diagnostics output.  Return true
**  to proceed with output, false to cancel.
*)
  TTidyReportFilter = function( tdoc: TTidyDoc; lvl: TTidyReportLevel; line: UInt; col: UInt; mssg: PChar ): Bool; cdecl;

var

(** Give Tidy a filter callback to use *)
  tidySetReportFilter: function( tdoc: TTidyDoc; filtCallback: TTidyReportFilter ): Bool; cdecl;

(** Set error sink to named file *)
  tidySetErrorFile: function( tdoc: TTidyDoc; errfilnam: PChar ): Pointer; cdecl;
(** Set error sink to given buffer *)
  tidySetErrorBuffer: function( tdoc: TTidyDoc; const errbuf: TTidyBuffer ): Integer; cdecl;
(** Set error sink to given generic sink *)
  tidySetErrorSink: function( tdoc: TTidyDoc; const sink: TTidyOutputSink ): Integer; cdecl;

(** @} end IO group *)


(** @defgroup Memory  Memory Allocation
**
** By default, Tidy will use its own wrappers
** around standard C malloc/free calls.
** These wrappers will abort upon any failures.
** If any are set, all must be set.
** Pass NULL to clear previous setting.
**
** May be used to set environment-specific allocators
** such as used by web server plugins, etc.
**
** @{
*)

type

(** Callback for "malloc" replacement *)
  TTidyMalloc = procedure( len: DWord ); cdecl;
(** Callback for "realloc" replacement *)
  TTidyRealloc = procedure( buf: Pointer; len: DWord ); cdecl;
(** Callback for "free" replacement *)
  TTidyFree = procedure( buf: Pointer ); cdecl;
(** Callback for "out of memory" panic state *)
  TTidyPanic = procedure( mssg: PChar ); cdecl;

var

(** Give Tidy a malloc() replacement *)
  tidySetMallocCall: function( fmalloc: TTidyMalloc ): Bool; cdecl;
(** Give Tidy a realloc() replacement *)
  tidySetReallocCall: function( frealloc: TTidyRealloc ): Bool; cdecl;
(** Give Tidy a free() replacement *)
  tidySetFreeCall: function( ffree: TTidyFree ): Bool; cdecl;
(** Give Tidy an "out of memory" handler *)
  tidySetPanicCall: function( fpanic: TTidyPanic ): Bool; cdecl;

(** @} end Memory group *)

(* TODO: Catalog all messages for easy translation
  PChar     tidyLookupMessage( Integer errorNo );
*)


(** @defgroup Parse Document Parse
**
** Parse markup from a given input source.  String and filename
** functions added for convenience.  HTML/XHTML version determined
** from input.
** @{
*)

var

(** Parse markup in named file *)
  tidyParseFile: function( tdoc: TTidyDoc; filename: PChar ): Integer; cdecl;

(** Parse markup from the standard input *)
  tidyParseStdin: function( tdoc: TTidyDoc ): Integer; cdecl;

(** Parse markup in given string *)
  tidyParseString: function( tdoc: TTidyDoc; content: PChar ): Integer; cdecl;

(** Parse markup in given buffer *)
  tidyParseBuffer: function( tdoc: TTidyDoc; const buf: TTidyBuffer ): Integer; cdecl;

(** Parse markup in given generic input source *)
  tidyParseSource: function( tdoc: TTidyDoc; const source: TTidyInputSource): Integer; cdecl;

(** @} End Parse group *)

var

(** @defgroup Clean Diagnostics and Repair
**
** @{
*)
(** Execute configured cleanup and repair operations on parsed markup *)
  tidyCleanAndRepair: function( tdoc: TTidyDoc ): Integer; cdecl;

(** Run configured diagnostics on parsed and repaired markup.
**  Must call tidyCleanAndRepair() first.
*)
  tidyRunDiagnostics: function( tdoc: TTidyDoc ): Integer; cdecl;

(** @} end Clean group *)

var

(** @defgroup Save Document Save Functions
**
** Save currently parsed document to the given output sink.  File name
** and string/buffer functions provided for convenience.
** @{
*)

(** Save to named file *)
  tidySaveFile: function( tdoc: TTidyDoc; filename: PChar ): Integer; cdecl;

(** Save to standard output *)
  tidySaveStdout: function( tdoc: TTidyDoc ): Integer; cdecl;

(** Save to given TTidyBuffer object *)
  tidySaveBuffer: function( tdoc: TTidyDoc; var buf: TTidyBuffer ): Integer; cdecl;

(** Save document to application buffer.  If buffer is not big enough,
**  ENOMEM will be returned and the necessary buffer size will be placed
**  in *buflen.
*)
  tidySaveString: function( tdoc: TTidyDoc; buffer: PChar; var buflen: UInt ): Integer; cdecl;

(** Save to given generic output sink *)
  tidySaveSink: function( tdoc: TTidyDoc; var sink: TTidyOutputSink ): Integer; cdecl;

(** @} end Save group *)

var

(** @addtogroup Basic
** @{
*)
(** Save current settings to named file.
    Only non-default values are written. *)
  tidyOptSaveFile: function( tdoc: TTidyDoc; cfgfil: PChar ): Integer; cdecl;

(** Save current settings to given output sink.
    Only non-default values are written. *)
  tidyOptSaveSink: function( tdoc: TTidyDoc; const sink: TTidyOutputSink ): Integer; cdecl;

(* Error reporting functions
*)

(** Write more complete information about errors to current error sink. *)
  tidyErrorSummary: procedure( tdoc: TTidyDoc ); cdecl;

(** Write more general information about markup to current error sink. *)
  tidyGeneralInfo: procedure( tdoc: TTidyDoc ); cdecl;

(** @} end Basic group (again) *)


(** @defgroup Tree Document Tree
**
** A parsed and, optionally, repaired document is
** represented by Tidy as a Tree, much like a W3C DOM.
** This tree may be traversed using these functions.
** The following snippet gives a basic idea how these
** functions can be used.
**
<pre>
void dumpNode( tnod: TTidyNode, Integer indent )
{
  TidyNode child;

  for ( child = tidyGetChild(tnod); child; child = tidyGetNext(child) )
  {
    PChar name;
    switch ( tidyNodeGetType(child) )
    {
    case TidyNode_Root:      name = "Root";                    break;
    case TidyNode_DocType:   name = "DOCTYPE";                 break;
    case TidyNode_Comment:   name = "Comment";                 break;
    case TidyNode_ProcIns:   name = "Processing Instruction";  break;
    case TidyNode_Text:      name = "Text";                    break;
    case TidyNode_CDATA:     name = "CDATA";                   break;
    case TidyNode_Section:   name = "XML Section";             break;
    case TidyNode_Asp:       name = "ASP";                     break;
    case TidyNode_Jste:      name = "JSTE";                    break;
    case TidyNode_Php:       name = "PHP";                     break;
    case TidyNode_XmlDecl:   name = "XML Declaration";         break;

    case TidyNode_Start:
    case TidyNode_End:
    case TidyNode_StartEnd:
    default:
      name = tidyNodeGetName( child );
      break;
    }
    assert( name != NULL );
    printf( "\%*.*sNode: \%s\\n", indent, indent, " ", name );
    dumpNode( child, indent + 4 );
  }
}

void dumpDoc( tdoc: TTidyDoc )
{
  dumpNode( tidyGetRoot(tdoc), 0 );
}

void dumpBody( tdoc: TTidyDoc )
{
  dumpNode( tidyGetBody(tdoc), 0 );
}
</pre>

@{

*)

var

  tidyGetRoot: function( tdoc: TTidyDoc ): TTidyNode; cdecl;
  tidyGetHtml: function( tdoc: TTidyDoc ): TTidyNode; cdecl;
  tidyGetHead: function( tdoc: TTidyDoc ): TTidyNode; cdecl;
  tidyGetBody: function( tdoc: TTidyDoc ): TTidyNode; cdecl;

(* parent / child *)
  tidyGetParent: function( tnod: TTidyNode ): TTidyNode; cdecl;
  tidyGetChild: function( tnod: TTidyNode ): TTidyNode; cdecl;

(* siblings *)
  tidyGetNext: function( tnod: TTidyNode ): TTidyNode; cdecl;
  tidyGetPrev: function( tnod: TTidyNode ): TTidyNode; cdecl;

(* Null for non-element nodes and all pure HTML
  PChar     tidyNodeNsLocal( tnod: TTidyNode );
  PChar     tidyNodeNsPrefix( tnod: TTidyNode );
  PChar     tidyNodeNsUri( tnod: TTidyNode );
*)

(* Iterate over attribute values *)
  tidyAttrFirst: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrNext: function( tattr: TTidyAttr ): TTidyAttr; cdecl;

  tidyAttrName: function( tattr: TTidyAttr ): PChar; cdecl;
  tidyAttrValue: function( tattr: TTidyAttr ): PChar; cdecl;

(* Null for pure HTML
  PChar     tidyAttrNsLocal( tattr: TTidyAttr );
  PChar     tidyAttrNsPrefix( tattr: TTidyAttr );
  PChar     tidyAttrNsUri( tattr: TTidyAttr );
*)

(** @} end Tree group *)

var

(** @defgroup NodeAsk Node Interrogation
**
** Get information about any givent node.
** @{
*)

(* Node info *)
  tidyNodeGetName: function( tnod: TTidyNode ): PChar; cdecl;
  tidyNodeGetType: function( tnod: TTidyNode ): TTidyNodeType; cdecl;
  tidyNodeGetValue: function( tdoc: TTidyDoc; tnod: TTidyNode; var buf: TTidyBuffer ): Bool; cdecl;

  tidyNodeIsText: function( tnod: TTidyNode ): Bool; cdecl;
  tidyNodeIsProp: function( tdoc: TTidyDoc; tnod: TTidyNode ): Bool; cdecl;
  tidyNodeIsHeader: function( tnod: TTidyNode ): Bool; (* h1, h2, ... *) cdecl;

  tidyNodeHasText: function( tdoc: TTidyDoc; tnod: TTidyNode ): Bool; cdecl;
  tidyNodeGetText: function( tdoc: TTidyDoc; tnod: TTidyNode; var buf: TTidyBuffer ): Bool; cdecl;

  tidyNodeGetId: function( tnod: TTidyNode ): TTidyTagId; cdecl;

  tidyNodeLine: function( tnod: TTidyNode ): UInt; cdecl;
  tidyNodeColumn: function( tnod: TTidyNode ): UInt; cdecl;

  tidyNodeIsHTML: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsHEAD: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsTITLE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBASE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsMETA: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBODY: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsFRAMESET: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsFRAME: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsIFRAME: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsNOFRAMES: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsHR: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsH1: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsH2: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsPRE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsLISTING: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsP: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsUL: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsOL: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsDL: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsDIR: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsLI: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsDT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsDD: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsTABLE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsCAPTION: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsTD: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsTH: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsTR: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsCOL: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsCOLGROUP: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBR: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsA: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsLINK: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsB: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsI: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSTRONG: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsEM: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBIG: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSMALL: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsPARAM: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsOPTION: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsOPTGROUP: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsIMG: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsMAP: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsAREA: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsNOBR: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsWBR: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsFONT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsLAYER: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSPACER: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsCENTER: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSTYLE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSCRIPT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsNOSCRIPT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsFORM: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsTEXTAREA: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBLOCKQUOTE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsAPPLET: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsOBJECT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsDIV: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSPAN: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsINPUT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsQ: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsLABEL: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsH3: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsH4: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsH5: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsH6: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsADDRESS: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsXMP: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSELECT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBLINK: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsMARQUEE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsEMBED: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsBASEFONT: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsISINDEX: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsS: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsSTRIKE: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsU: function(tnod: TTidyNode): Bool; cdecl;
  tidyNodeIsMENU: function(tnod: TTidyNode): Bool; cdecl;

(** @} End NodeAsk group *)

var

(** @defgroup Attribute Attribute Interrogation
**
** Get information about any given attribute.
** @{
*)

  tidyAttrGetIdOf: function( tattr: TTidyAttr ): TTidyAttrId; cdecl;
  tidyAttrIsEvent: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsProp: function( tattr: TTidyAttr ): Bool; cdecl;

  tidyAttrIsHREF: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsSRC: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsID: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsNAME: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsSUMMARY: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsALT: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsLONGDESC: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsUSEMAP: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsISMAP: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsLANGUAGE: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsTYPE: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsVALUE: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsCONTENT: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsTITLE: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsXMLNS: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsDATAFLD: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsWIDTH: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsHEIGHT: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsFOR: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsSELECTED: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsCHECKED: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsLANG: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsTARGET: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsHTTP_EQUIV: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsREL: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnMOUSEMOVE: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnMOUSEDOWN: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnMOUSEUP: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnCLICK: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnMOUSEOVER: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnMOUSEOUT: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnKEYDOWN: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnKEYUP: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnKEYPRESS: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnFOCUS: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsOnBLUR: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsBGCOLOR: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsLINK: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsALINK: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsVLINK: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsTEXT: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsSTYLE: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsABBR: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsCOLSPAN: function( tattr: TTidyAttr ): Bool; cdecl;
  tidyAttrIsROWSPAN: function( tattr: TTidyAttr ): Bool; cdecl;

(** @} end AttrAsk group *)

var

(** @defgroup AttrGet Attribute Retrieval
**
** Lookup an attribute from a given node
** @{
*)

  tidyAttrGetById: function( tnod: TTidyNode; attId: TTidyAttrId ): TTidyAttr; cdecl;
  tidyAttrGetHREF: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetSRC: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetID: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetNAME: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetSUMMARY: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetALT: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetLONGDESC: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetUSEMAP: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetISMAP: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetLANGUAGE: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetTYPE: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetVALUE: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetCONTENT: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetTITLE: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetXMLNS: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetDATAFLD: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetWIDTH: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetHEIGHT: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetFOR: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetSELECTED: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetCHECKED: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetLANG: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetTARGET: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetHTTP_EQUIV: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetREL: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnMOUSEMOVE: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnMOUSEDOWN: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnMOUSEUP: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnCLICK: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnMOUSEOVER: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnMOUSEOUT: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnKEYDOWN: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnKEYUP: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnKEYPRESS: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnFOCUS: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetOnBLUR: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetBGCOLOR: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetLINK: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetALINK: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetVLINK: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetTEXT: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetSTYLE: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetABBR: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetCOLSPAN: function( tnod: TTidyNode ): TTidyAttr; cdecl;
  tidyAttrGetROWSPAN: function( tnod: TTidyNode ): TTidyAttr; cdecl;

(** @} end AttrGet group *)

type
  THtmlData = Record
    Body: String;
    Base: String;
  end;

function LoadTidy(const Path: String): THandle;
function Utf8ToUnicode(const Input: String): String;
function BeautifyHtmlFile(const FileName: String): Longbool;
function TidyConvertEntities(const Input: String): String;
function TidyHtml(const Input, Font: String; FromFile, DropFont, Utf8: Longbool): THtmlData;

implementation

const
  HTidy: THandle = 0;

const
  E = '';
  SP = ' ';
  CR = #13;
  LF = #10;
  CRLF = #13#10;

function LoStrAtPos(const Find, Dest: String; Pos: Integer): Boolean; register; assembler;
const
  SLO = 4;
asm	// EAX = @Find, EDX = @Dest, ECX = Pos
  OR	EAX,EAX
  JE	@EXIT
  OR	EDX,EDX
  JE	@EXIT
  OR	ECX,ECX
  JLE	@EXIT
  PUSH	EBX
  MOV	EBX,[EDX - SLO]
  SUB	EBX,ECX
  INC	EBX
  CMP	EBX,[EAX - SLO]
  JL	@END
  ADD	EDX,ECX
  MOV	ECX,[EAX - SLO]
  DEC	EDX
  @LOOP:
  DEC	ECX
  JL	@MATCH
  MOV	BL,[EDX + ECX]
  XOR	BL,[EAX + ECX]
  JE	@LOOP
  TEST	BYTE PTR [EDX + ECX],040h
  JE	@END
  TEST	BYTE PTR [EAX + ECX],040h
  JE	@END
  AND	BL,0DFh
  JE	@LOOP
  POP	EBX
  XOR	EAX,EAX
  RET
  @MATCH:
  POP	EBX
  MOV	EAX,true
  RET
  @END:
  POP	EBX
  @EXIT:
  XOR	EAX,EAX
end;

function Utf8ToUnicodeNum(Source, Dest: PChar; var CharCount: Integer): Integer; register; assembler;
asm	// EAX = @Source, EDX = @Dest, ECX = @Count
  MOV	DWORD PTR [ECX],0
  OR	EAX,EAX
  JE	@END
  OR	EDX,EDX
  JE	@END
  PUSH	EBX
  PUSH	EDX
  MOV   EDX,ECX
  XOR	EBX,EBX
  MOV	BL,[EAX]
  CMP	BL,0FDh
  JA	@ZERO
  CMP	BL,080h
  JB	@ASCII
  XOR	ECX,ECX
  @COUNT:
  INC	CL
  SHL	BL,1
  JC	@COUNT
  CMP	CL,3
  JB	@ZERO
  SHR	BL,CL
  DEC	ECX
  MOV	[EDX],ECX
  ADD	EAX,ECX
  PUSH	EDX
  NEG	ECX
  INC	ECX
  @UTF_LOOP:
  MOV	DL,[EAX + ECX]
  CMP	DL,080h
  JB	@EXIT
  AND	DL,07Fh
  CMP	DL,040h
  JNB	@EXIT
  SHL	EBX,6
  OR	BL,DL
  INC	ECX
  JNE	@UTF_LOOP
  MOV	EAX,EBX
  POP	EDX
  POP	EDX
  JMP	@GET_ENTITY
  @ASCII:
  MOV	DWORD PTR [EDX],1
  MOV	EAX,EBX
  POP	EDX
  // AX = Unicode, EDX = @Dest
  @GET_ENTITY:
  PUSH	EDI
  MOV	DWORD PTR [EDX], '#&'
  MOV	EDI,EDX
  ADD	EDI,2
  MOV	ECX,5
  CMP	AX,10000
  JA	@CONVERT
  DEC	ECX
  CMP	AX,1000
  JA	@CONVERT
  DEC	ECX
  CMP	AX,100
  JA	@CONVERT
  DEC	ECX
  CMP	AX,10
  JA	@CONVERT
  DEC	ECX
  @CONVERT:
  MOV	BYTE PTR [EDI + ECX],';'
  MOV	EBX,10
  PUSH	ECX
  DEC	ECX
  JE	@DONE
  @DEZ_LOOP:
  XOR	DX,DX
  DIV	BX
  ADD	DL,'0'
  MOV	[EDI + ECX],DL
  DEC	ECX
  JNE	@DEZ_LOOP
  @DONE:
  ADD	AL,'0'
  MOV	[EDI + ECX],AL
  POP	EAX
  POP	EDI
  POP	EBX
  ADD	EAX,3
  RET
  @EXIT:
  POP	EDX
  @ZERO:
  XOR	EAX,EAX
  MOV	[EDX],EAX
  POP	EDX
  POP	EBX
  @END:
end;

function Utf8ToUnicode(const Input: String): String;
var
  iInput, iResult, iBeg, iEnd, iLen, iPos: Integer;
  pResult: PChar;
begin
  if Input <> E then begin
    iInput := Length(Input);
    iResult := iInput shl 1;
    SetString(Result, nil, iResult);
    pResult := Pointer(Result);
    iBeg := 1;
    repeat
      iEnd := iBeg;
      while (iEnd <= iInput) and (Input[iEnd] < #128) do inc(iEnd);
      iLen := iEnd - iBeg;
      if iLen > 0 then begin
	Move(Input[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	inc(iBeg, iLen);
      end;
      if iEnd < iInput then begin
	inc(pResult, Utf8ToUnicodeNum(@Input[iEnd], pResult, iLen));
	if iLen > 0 then
	  inc(iBeg, iLen)
	else begin
	  pResult^ := Input[iEnd];
	  inc(pResult);
	  inc(iBeg);
	end;
      end;
      iLen := succ(iInput - iBeg);
      iPos := pResult - Pointer(Result);
      if iPos + iLen > iResult then begin
	inc(iResult, iLen shl 1);
	SetLength(Result, iResult);
	pResult := Pointer(Result);
	inc(pResult, iPos);
      end;
    until iBeg > iInput;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function BeautifyHtmlFile(const FileName: String): Longbool;
var
  tdoc: TTidyDoc;
begin
  Result := false;
  if (HTidy <> 0) and (FileName <> E) then begin
    tdoc := tidyCreate;
    if tdoc <> nil then begin
      try
	if tidyOptResetAllToDefault(tdoc) then begin
	  tidyOptSetInt(tdoc, TidyBodyOnly, ord(TidyNoState));
	  tidyOptSetBool(tdoc, TidyEncloseBodyText, true);
	  tidyOptSetBool(tdoc, TidyEncloseBlockText, true);
	  tidyOptSetBool(tdoc, TidyFixBackslash, true);
	  tidyOptSetBool(tdoc, TidyFixUri, true);
	  tidyOptSetBool(tdoc, TidyForceOutput, true);
	  tidyOptSetBool(tdoc, TidyHideComments, false);
	  tidyOptSetBool(tdoc, TidyHtmlOut, true);
	  tidyOptSetBool(tdoc, TidyIndentCdata, true);
	  tidyOptSetInt(tdoc, TidyIndentContent, ord(TidyAutoState));
	  tidyOptSetBool(tdoc, TidyLowerLiterals, true);
	  tidyOptSetBool(tdoc, TidyQuiet, true);
	  tidyOptSetBool(tdoc, TidyQuoteAmpersand, true);
	  tidyOptSetBool(tdoc, TidyQuoteMarks, true);
	  tidyOptSetBool(tdoc, TidyQuoteNbsp, true);
	  tidyOptSetInt(tdoc, TidyWrapLen, 0);
	  tidyOptSetBool(tdoc, TidyXmlDecl, false);
	  Result := (tidyParseFile(tdoc, PChar(FileName)) >= 0) and
		    (tidyCleanAndRepair(tdoc) >= 0) and
		    (tidySaveFile(tdoc, PChar(FileName)) >= 0);
	end;
      finally
	tidyRelease(tdoc);
      end;
    end;
  end;
end;

function TidyConvertEntities(const Input: String): String;
var
  tdoc: TTidyDoc;
  uSize: UInt;
begin
  if (HTidy <> 0) and (Input <> E) then begin
    tdoc := tidyCreate;
    if tdoc <> nil then begin
      try
	if tidyOptResetAllToDefault(tdoc) then begin
	  tidyOptSetInt(tdoc, TidyBodyOnly, ord(TidyYesState));
	  tidyOptSetInt(tdoc, TidyDoctypeMode, ord(TidyDoctypeOmit));
	  tidyOptSetBool(tdoc, TidyDropFontTags, true);
	  tidyOptSetBool(tdoc, TidyEncloseBodyText, false);
	  tidyOptSetBool(tdoc, TidyEncloseBlockText, false);
	  tidyOptSetBool(tdoc, TidyFixBackslash, true);
	  tidyOptSetBool(tdoc, TidyFixUri, true);
	  tidyOptSetBool(tdoc, TidyForceOutput, true);
	  tidyOptSetBool(tdoc, TidyHideComments, true);
	  tidyOptSetBool(tdoc, TidyHtmlOut, true);
	  tidyOptSetBool(tdoc, TidyLowerLiterals, true);
	  tidyOptSetBool(tdoc, TidyNumEntities, false);
	  tidyOptSetBool(tdoc, TidyPreserveEntities, false);
	  tidyOptSetBool(tdoc, TidyQuiet, true);
	  tidyOptSetBool(tdoc, TidyQuoteAmpersand, false);
	  tidyOptSetBool(tdoc, TidyQuoteMarks, false);
	  tidyOptSetBool(tdoc, TidyQuoteNbsp, false);
	  tidyOptSetInt(tdoc, TidyWrapLen, 0);
	  tidyOptSetBool(tdoc, TidyXmlDecl, false);
	  tidySetCharEncoding(tdoc, 'utf8');
	  if (tidyParseString(tdoc, PChar(Input)) >= 0)
	  and (tidyCleanAndRepair(tdoc) >= 0) then begin
	    uSize := 0;
	    if tidySaveString(tdoc, nil, uSize) = -12 then begin
	      SetString(Result, nil, uSize);
	      tidySaveString(tdoc, Pointer(Result), uSize);
	      while (uSize > 0) and (Result[uSize] <= SP) do dec(uSize);
	      Delete(Result, succ(uSize), MAXINT);
	      Exit;
	    end;
	  end;
	end;
      finally
	tidyRelease(tdoc);
      end;
    end;
  end;
  Result := Input;
end;

function TidyHtml(const Input, Font: String; FromFile, DropFont, Utf8: Longbool): THtmlData;

  function GetUrlBase(tdoc: TTidyDoc): String;
  var
    tnChild: TTidyNode;
    taFound: TTidyAttr;
  begin
    Result := E;
    tnChild := tidyGetChild(tidyGetHead(tdoc));
    while tnChild <> nil do begin
      if tidyNodeIsBASE(tnChild) then begin
	taFound := tidyAttrFirst(tnChild);
	while taFound <> nil do begin
	  if tidyAttrIsHREF(taFound) then begin
	    Result := tidyAttrValue(taFound);
	    Break;
	  end;
	  taFound := tidyAttrNext(tnChild);
	end;
      end;
      tnChild := tidyGetNext(tnChild);
    end;
  end;

var
  tdoc: TTidyDoc;
  uSize, uPos: UInt;
begin
  if (HTidy <> 0) and (Input <> E) then begin
    tdoc := tidyCreate;
    if tdoc <> nil then begin
      try
	if tidyOptResetAllToDefault(tdoc) then begin
	  tidyOptSetInt(tdoc, TidyBodyOnly, ord(TidyYesState));
	  tidyOptSetInt(tdoc, TidyDoctypeMode, ord(TidyDoctypeOmit));
	  tidyOptSetBool(tdoc, TidyDropFontTags, DropFont);
	  tidyOptSetBool(tdoc, TidyEncloseBodyText, true);
	  tidyOptSetBool(tdoc, TidyEncloseBlockText, true);
	  tidyOptSetBool(tdoc, TidyFixBackslash, true);
	  tidyOptSetBool(tdoc, TidyFixUri, true);
	  tidyOptSetBool(tdoc, TidyForceOutput, true);
	  tidyOptSetBool(tdoc, TidyHideComments, true);
	  tidyOptSetBool(tdoc, TidyHtmlOut, true);
	  tidyOptSetBool(tdoc, TidyLowerLiterals, true);
	  tidyOptSetBool(tdoc, TidyNumEntities, false);
	  tidyOptSetBool(tdoc, TidyPreserveEntities, false);
	  tidyOptSetBool(tdoc, TidyQuiet, true);
	  tidyOptSetBool(tdoc, TidyQuoteAmpersand, true);
	  tidyOptSetBool(tdoc, TidyQuoteMarks, true);
	  tidyOptSetBool(tdoc, TidyQuoteNbsp, true);
	  tidyOptSetInt(tdoc, TidyWrapLen, 0);
	  tidyOptSetBool(tdoc, TidyXmlDecl, false);
	  if Utf8 then tidySetCharEncoding(tdoc, 'utf8');
	  if (FromFile and (tidyParseFile(tdoc, PChar(Input)) >= 0))
	  or (not FromFile and (tidyParseString(tdoc, PChar(Input)) >= 0)) then begin
	    if tidyCleanAndRepair(tdoc) >= 0 then with Result do begin
	      Base := GetUrlBase(tdoc);
	      uSize := 0;
	      if tidySaveString(tdoc, nil, uSize) = -12 then begin
		SetString(Body, nil, uSize);
		tidySaveString(tdoc, Pointer(Body), uSize);
	      end;
	      while (uSize > 0) and (Body[uSize] <= SP) do dec(uSize);
	      if (Font = E) and (UInt(Length(Body)) > succ(uSize)) and (Body[uSize + 2] = LF) then
		Delete(Body, uSize + 3, MAXINT)
	      else Delete(Body, succ(uSize), MAXINT);
	    end
	    else Result.Body := Input;
	  end
	  else Result.Body := Input;
	end
	else Result.Body := Input;
      finally
	tidyRelease(tdoc);
      end;
    end
    else Result.Body := Input;
    with Result do begin
      if LoStrAtPos('<?xml', Body, 1) then begin
	uSize := Length(Body);
	uPos := succ(pos('>', Body));
	while (uPos < uSize) and (Body[uPos] < SP) do inc(uPos);
	Delete(Body, 1, pred(uPos));
      end;
      if Font <> E then Body := '<font face="' + Font + '">' + CRLF + Body + CRLF + '</font>';
    end;
  end
  else Result.Body := Input;
end;

function LoadTidy(const Path: String): THandle;
const
  LibTidy = 'LibTidy.dll';
begin
  Result := GetModuleHandle(LibTidy);
  if Result = 0 then Result := LoadLibrary(PChar(Path + LibTidy));
  if Result <> 0 then begin
    tidyAccessWarningCount := GetProcAddress(Result, 'tidyAccessWarningCount');
    tidyAttrFirst := GetProcAddress(Result, 'tidyAttrFirst');
    tidyAttrGetABBR := GetProcAddress(Result, 'tidyAttrGetABBR');
    tidyAttrGetALINK := GetProcAddress(Result, 'tidyAttrGetALINK');
    tidyAttrGetALT := GetProcAddress(Result, 'tidyAttrGetALT');
    tidyAttrGetBGCOLOR := GetProcAddress(Result, 'tidyAttrGetBGCOLOR');
    tidyAttrGetById := GetProcAddress(Result, 'tidyAttrGetById');
    tidyAttrGetCHECKED := GetProcAddress(Result, 'tidyAttrGetCHECKED');
    tidyAttrGetCOLSPAN := GetProcAddress(Result, 'tidyAttrGetCOLSPAN');
    tidyAttrGetCONTENT := GetProcAddress(Result, 'tidyAttrGetCONTENT');
    tidyAttrGetDATAFLD := GetProcAddress(Result, 'tidyAttrGetDATAFLD');
    tidyAttrGetFOR := GetProcAddress(Result, 'tidyAttrGetFOR');
    tidyAttrGetHEIGHT := GetProcAddress(Result, 'tidyAttrGetHEIGHT');
    tidyAttrGetHREF := GetProcAddress(Result, 'tidyAttrGetHREF');
    tidyAttrGetHTTP_EQUIV := GetProcAddress(Result, 'tidyAttrGetHTTP_EQUIV');
    tidyAttrGetID := GetProcAddress(Result, 'tidyAttrGetID');
    tidyAttrGetIdOf := GetProcAddress(Result, 'tidyAttrGetIdOf');
    tidyAttrGetISMAP := GetProcAddress(Result, 'tidyAttrGetISMAP');
    tidyAttrGetLANG := GetProcAddress(Result, 'tidyAttrGetLANG');
    tidyAttrGetLANGUAGE := GetProcAddress(Result, 'tidyAttrGetLANGUAGE');
    tidyAttrGetLINK := GetProcAddress(Result, 'tidyAttrGetLINK');
    tidyAttrGetLONGDESC := GetProcAddress(Result, 'tidyAttrGetLONGDESC');
    tidyAttrGetNAME := GetProcAddress(Result, 'tidyAttrGetNAME');
    tidyAttrGetOnBLUR := GetProcAddress(Result, 'tidyAttrGetOnBLUR');
    tidyAttrGetOnCLICK := GetProcAddress(Result, 'tidyAttrGetOnCLICK');
    tidyAttrGetOnFOCUS := GetProcAddress(Result, 'tidyAttrGetOnFOCUS');
    tidyAttrGetOnKEYDOWN := GetProcAddress(Result, 'tidyAttrGetOnKEYDOWN');
    tidyAttrGetOnKEYPRESS := GetProcAddress(Result, 'tidyAttrGetOnKEYPRESS');
    tidyAttrGetOnKEYUP := GetProcAddress(Result, 'tidyAttrGetOnKEYUP');
    tidyAttrGetOnMOUSEDOWN := GetProcAddress(Result, 'tidyAttrGetOnMOUSEDOWN');
    tidyAttrGetOnMOUSEMOVE := GetProcAddress(Result, 'tidyAttrGetOnMOUSEMOVE');
    tidyAttrGetOnMOUSEOUT := GetProcAddress(Result, 'tidyAttrGetOnMOUSEOUT');
    tidyAttrGetOnMOUSEOVER := GetProcAddress(Result, 'tidyAttrGetOnMOUSEOVER');
    tidyAttrGetOnMOUSEUP := GetProcAddress(Result, 'tidyAttrGetOnMOUSEUP');
    tidyAttrGetREL := GetProcAddress(Result, 'tidyAttrGetREL');
    tidyAttrGetROWSPAN := GetProcAddress(Result, 'tidyAttrGetROWSPAN');
    tidyAttrGetSELECTED := GetProcAddress(Result, 'tidyAttrGetSELECTED');
    tidyAttrGetSRC := GetProcAddress(Result, 'tidyAttrGetSRC');
    tidyAttrGetSTYLE := GetProcAddress(Result, 'tidyAttrGetSTYLE');
    tidyAttrGetSUMMARY := GetProcAddress(Result, 'tidyAttrGetSUMMARY');
    tidyAttrGetTARGET := GetProcAddress(Result, 'tidyAttrGetTARGET');
    tidyAttrGetTEXT := GetProcAddress(Result, 'tidyAttrGetTEXT');
    tidyAttrGetTITLE := GetProcAddress(Result, 'tidyAttrGetTITLE');
    tidyAttrGetTYPE := GetProcAddress(Result, 'tidyAttrGetTYPE');
    tidyAttrGetUSEMAP := GetProcAddress(Result, 'tidyAttrGetUSEMAP');
    tidyAttrGetVALUE := GetProcAddress(Result, 'tidyAttrGetVALUE');
    tidyAttrGetVLINK := GetProcAddress(Result, 'tidyAttrGetVLINK');
    tidyAttrGetWIDTH := GetProcAddress(Result, 'tidyAttrGetWIDTH');
    tidyAttrGetXMLNS := GetProcAddress(Result, 'tidyAttrGetXMLNS');
    tidyAttrIsABBR := GetProcAddress(Result, 'tidyAttrIsABBR');
    tidyAttrIsALINK := GetProcAddress(Result, 'tidyAttrIsALINK');
    tidyAttrIsALT := GetProcAddress(Result, 'tidyAttrIsALT');
    tidyAttrIsBGCOLOR := GetProcAddress(Result, 'tidyAttrIsBGCOLOR');
    tidyAttrIsCHECKED := GetProcAddress(Result, 'tidyAttrIsCHECKED');
    tidyAttrIsCOLSPAN := GetProcAddress(Result, 'tidyAttrIsCOLSPAN');
    tidyAttrIsCONTENT := GetProcAddress(Result, 'tidyAttrIsCONTENT');
    tidyAttrIsDATAFLD := GetProcAddress(Result, 'tidyAttrIsDATAFLD');
    tidyAttrIsEvent := GetProcAddress(Result, 'tidyAttrIsEvent');
    tidyAttrIsFOR := GetProcAddress(Result, 'tidyAttrIsFOR');
    tidyAttrIsHEIGHT := GetProcAddress(Result, 'tidyAttrIsHEIGHT');
    tidyAttrIsHREF := GetProcAddress(Result, 'tidyAttrIsHREF');
    tidyAttrIsHTTP_EQUIV := GetProcAddress(Result, 'tidyAttrIsHTTP_EQUIV');
    tidyAttrIsID := GetProcAddress(Result, 'tidyAttrIsID');
    tidyAttrIsISMAP := GetProcAddress(Result, 'tidyAttrIsISMAP');
    tidyAttrIsLANG := GetProcAddress(Result, 'tidyAttrIsLANG');
    tidyAttrIsLANGUAGE := GetProcAddress(Result, 'tidyAttrIsLANGUAGE');
    tidyAttrIsLINK := GetProcAddress(Result, 'tidyAttrIsLINK');
    tidyAttrIsLONGDESC := GetProcAddress(Result, 'tidyAttrIsLONGDESC');
    tidyAttrIsNAME := GetProcAddress(Result, 'tidyAttrIsNAME');
    tidyAttrIsOnBLUR := GetProcAddress(Result, 'tidyAttrIsOnBLUR');
    tidyAttrIsOnCLICK := GetProcAddress(Result, 'tidyAttrIsOnCLICK');
    tidyAttrIsOnFOCUS := GetProcAddress(Result, 'tidyAttrIsOnFOCUS');
    tidyAttrIsOnKEYDOWN := GetProcAddress(Result, 'tidyAttrIsOnKEYDOWN');
    tidyAttrIsOnKEYPRESS := GetProcAddress(Result, 'tidyAttrIsOnKEYPRESS');
    tidyAttrIsOnKEYUP := GetProcAddress(Result, 'tidyAttrIsOnKEYUP');
    tidyAttrIsOnMOUSEDOWN := GetProcAddress(Result, 'tidyAttrIsOnMOUSEDOWN');
    tidyAttrIsOnMOUSEMOVE := GetProcAddress(Result, 'tidyAttrIsOnMOUSEMOVE');
    tidyAttrIsOnMOUSEOUT := GetProcAddress(Result, 'tidyAttrIsOnMOUSEOUT');
    tidyAttrIsOnMOUSEOVER := GetProcAddress(Result, 'tidyAttrIsOnMOUSEOVER');
    tidyAttrIsOnMOUSEUP := GetProcAddress(Result, 'tidyAttrIsOnMOUSEUP');
    tidyAttrIsProp := GetProcAddress(Result, 'tidyAttrIsProp');
    tidyAttrIsREL := GetProcAddress(Result, 'tidyAttrIsREL');
    tidyAttrIsROWSPAN := GetProcAddress(Result, 'tidyAttrIsROWSPAN');
    tidyAttrIsSELECTED := GetProcAddress(Result, 'tidyAttrIsSELECTED');
    tidyAttrIsSRC := GetProcAddress(Result, 'tidyAttrIsSRC');
    tidyAttrIsSTYLE := GetProcAddress(Result, 'tidyAttrIsSTYLE');
    tidyAttrIsSUMMARY := GetProcAddress(Result, 'tidyAttrIsSUMMARY');
    tidyAttrIsTARGET := GetProcAddress(Result, 'tidyAttrIsTARGET');
    tidyAttrIsTEXT := GetProcAddress(Result, 'tidyAttrIsTEXT');
    tidyAttrIsTITLE := GetProcAddress(Result, 'tidyAttrIsTITLE');
    tidyAttrIsTYPE := GetProcAddress(Result, 'tidyAttrIsTYPE');
    tidyAttrIsUSEMAP := GetProcAddress(Result, 'tidyAttrIsUSEMAP');
    tidyAttrIsVALUE := GetProcAddress(Result, 'tidyAttrIsVALUE');
    tidyAttrIsVLINK := GetProcAddress(Result, 'tidyAttrIsVLINK');
    tidyAttrIsWIDTH := GetProcAddress(Result, 'tidyAttrIsWIDTH');
    tidyAttrIsXMLNS := GetProcAddress(Result, 'tidyAttrIsXMLNS');
    tidyAttrName := GetProcAddress(Result, 'tidyAttrName');
    tidyAttrNext := GetProcAddress(Result, 'tidyAttrNext');
    tidyAttrValue := GetProcAddress(Result, 'tidyAttrValue');

    tidyBufAlloc:= GetProcAddress(Result, 'tidyBufAlloc');
    tidyBufAllocWithAllocator:= GetProcAddress(Result, 'tidyBufAllocWithAllocator');
    tidyBufAppend:= GetProcAddress(Result, 'tidyBufAppend');
    tidyBufAttach:= GetProcAddress(Result, 'tidyBufAttach');
    tidyBufCheckAlloc:= GetProcAddress(Result, 'tidyBufCheckAlloc');
    tidyBufClear:= GetProcAddress(Result, 'tidyBufClear');
    tidyBufDetach:= GetProcAddress(Result, 'tidyBufDetach');
    tidyBufEndOfInput:= GetProcAddress(Result, 'tidyBufEndOfInput');
    tidyBufFree:= GetProcAddress(Result, 'tidyBufFree');
    tidyBufGetByte:= GetProcAddress(Result, 'tidyBufGetByte');
    tidyBufInit:= GetProcAddress(Result, 'tidyBufInit');
    tidyBufInitWithAllocator:= GetProcAddress(Result, 'tidyBufInitWithAllocator');
    tidyBufPopByte:= GetProcAddress(Result, 'tidyBufPopByte');
    tidyBufPutByte:= GetProcAddress(Result, 'tidyBufPutByte');
    tidyBufUngetByte:= GetProcAddress(Result, 'tidyBufUngetByte');

    tidyCleanAndRepair := GetProcAddress(Result, 'tidyCleanAndRepair');
    tidyConfigErrorCount := GetProcAddress(Result, 'tidyConfigErrorCount');
    tidyCreate := GetProcAddress(Result, 'tidyCreate');
    tidyCreateWithAllocator:= GetProcAddress(Result, 'tidyCreateWithAllocator');
    tidyDetectedGenericXml := GetProcAddress(Result, 'tidyDetectedGenericXml');
    tidyDetectedHtmlVersion := GetProcAddress(Result, 'tidyDetectedHtmlVersion');
    tidyDetectedXhtml := GetProcAddress(Result, 'tidyDetectedXhtml');
    tidyErrorCount := GetProcAddress(Result, 'tidyErrorCount');
    tidyErrorSummary := GetProcAddress(Result, 'tidyErrorSummary');
    tidyFileExists := GetProcAddress(Result, 'tidyFileExists');
    tidyGeneralInfo := GetProcAddress(Result, 'tidyGeneralInfo');
    tidyGetAppData := GetProcAddress(Result, 'tidyGetAppData');
    tidyGetBody := GetProcAddress(Result, 'tidyGetBody');
    tidyGetByte := GetProcAddress(Result, 'tidyGetByte');
    tidyGetChild := GetProcAddress(Result, 'tidyGetChild');
    tidyGetHead := GetProcAddress(Result, 'tidyGetHead');
    tidyGetHtml := GetProcAddress(Result, 'tidyGetHtml');
    tidyGetNext := GetProcAddress(Result, 'tidyGetNext');
    tidyGetNextOption := GetProcAddress(Result, 'tidyGetNextOption');
    tidyGetOption := GetProcAddress(Result, 'tidyGetOption');
    tidyGetOptionByName := GetProcAddress(Result, 'tidyGetOptionByName');
    tidyGetOptionList := GetProcAddress(Result, 'tidyGetOptionList');
    tidyGetParent := GetProcAddress(Result, 'tidyGetParent');
    tidyGetPrev := GetProcAddress(Result, 'tidyGetPrev');
    tidyGetRoot := GetProcAddress(Result, 'tidyGetRoot');
    tidyInitInputBuffer:= GetProcAddress(Result, 'tidyInitInputBuffer');
    tidyInitOutputBuffer:= GetProcAddress(Result, 'tidyInitOutputBuffer');
    tidyInitSink := GetProcAddress(Result, 'tidyInitSink');
    tidyInitSource := GetProcAddress(Result, 'tidyInitSource');
    tidyIsEOF := GetProcAddress(Result, 'tidyIsEOF');
    tidyLoadConfig := GetProcAddress(Result, 'tidyLoadConfig');
    tidyLoadConfigEnc := GetProcAddress(Result, 'tidyLoadConfigEnc');

    tidyNodeColumn := GetProcAddress(Result, 'tidyNodeColumn');
    tidyNodeGetId := GetProcAddress(Result, 'tidyNodeGetId');
    tidyNodeGetName := GetProcAddress(Result, 'tidyNodeGetName');
    tidyNodeGetText := GetProcAddress(Result, 'tidyNodeGetText');
    tidyNodeGetType := GetProcAddress(Result, 'tidyNodeGetType');
    tidyNodeGetValue:= GetProcAddress(Result, 'tidyNodeGetValue');
    tidyNodeHasText := GetProcAddress(Result, 'tidyNodeHasText');
    tidyNodeIsA := GetProcAddress(Result, 'tidyNodeIsA');
    tidyNodeIsADDRESS := GetProcAddress(Result, 'tidyNodeIsADDRESS');
    tidyNodeIsAPPLET := GetProcAddress(Result, 'tidyNodeIsAPPLET');
    tidyNodeIsAREA := GetProcAddress(Result, 'tidyNodeIsAREA');
    tidyNodeIsB := GetProcAddress(Result, 'tidyNodeIsB');
    tidyNodeIsBASE := GetProcAddress(Result, 'tidyNodeIsBASE');
    tidyNodeIsBASEFONT := GetProcAddress(Result, 'tidyNodeIsBASEFONT');
    tidyNodeIsBIG := GetProcAddress(Result, 'tidyNodeIsBIG');
    tidyNodeIsBLINK := GetProcAddress(Result, 'tidyNodeIsBLINK');
    tidyNodeIsBLOCKQUOTE := GetProcAddress(Result, 'tidyNodeIsBLOCKQUOTE');
    tidyNodeIsBODY := GetProcAddress(Result, 'tidyNodeIsBODY');
    tidyNodeIsBR := GetProcAddress(Result, 'tidyNodeIsBR');
    tidyNodeIsCAPTION := GetProcAddress(Result, 'tidyNodeIsCAPTION');
    tidyNodeIsCENTER := GetProcAddress(Result, 'tidyNodeIsCENTER');
    tidyNodeIsCOL := GetProcAddress(Result, 'tidyNodeIsCOL');
    tidyNodeIsCOLGROUP := GetProcAddress(Result, 'tidyNodeIsCOLGROUP');
    tidyNodeIsDD := GetProcAddress(Result, 'tidyNodeIsDD');
    tidyNodeIsDIR := GetProcAddress(Result, 'tidyNodeIsDIR');
    tidyNodeIsDIV := GetProcAddress(Result, 'tidyNodeIsDIV');
    tidyNodeIsDL := GetProcAddress(Result, 'tidyNodeIsDL');
    tidyNodeIsDT := GetProcAddress(Result, 'tidyNodeIsDT');
    tidyNodeIsEM := GetProcAddress(Result, 'tidyNodeIsEM');
    tidyNodeIsEMBED := GetProcAddress(Result, 'tidyNodeIsEMBED');
    tidyNodeIsFONT := GetProcAddress(Result, 'tidyNodeIsFONT');
    tidyNodeIsFORM := GetProcAddress(Result, 'tidyNodeIsFORM');
    tidyNodeIsFRAME := GetProcAddress(Result, 'tidyNodeIsFRAME');
    tidyNodeIsFRAMESET := GetProcAddress(Result, 'tidyNodeIsFRAMESET');
    tidyNodeIsH1 := GetProcAddress(Result, 'tidyNodeIsH1');
    tidyNodeIsH2 := GetProcAddress(Result, 'tidyNodeIsH2');
    tidyNodeIsH3 := GetProcAddress(Result, 'tidyNodeIsH3');
    tidyNodeIsH4 := GetProcAddress(Result, 'tidyNodeIsH4');
    tidyNodeIsH5 := GetProcAddress(Result, 'tidyNodeIsH5');
    tidyNodeIsH6 := GetProcAddress(Result, 'tidyNodeIsH6');
    tidyNodeIsHEAD := GetProcAddress(Result, 'tidyNodeIsHEAD');
    tidyNodeIsHeader := GetProcAddress(Result, 'tidyNodeIsHeader');
    tidyNodeIsHR := GetProcAddress(Result, 'tidyNodeIsHR');
    tidyNodeIsHTML := GetProcAddress(Result, 'tidyNodeIsHTML');
    tidyNodeIsI := GetProcAddress(Result, 'tidyNodeIsI');
    tidyNodeIsIFRAME := GetProcAddress(Result, 'tidyNodeIsIFRAME');
    tidyNodeIsIMG := GetProcAddress(Result, 'tidyNodeIsIMG');
    tidyNodeIsINPUT := GetProcAddress(Result, 'tidyNodeIsINPUT');
    tidyNodeIsISINDEX := GetProcAddress(Result, 'tidyNodeIsISINDEX');
    tidyNodeIsLABEL := GetProcAddress(Result, 'tidyNodeIsLABEL');
    tidyNodeIsLAYER := GetProcAddress(Result, 'tidyNodeIsLAYER');
    tidyNodeIsLI := GetProcAddress(Result, 'tidyNodeIsLI');
    tidyNodeIsLINK := GetProcAddress(Result, 'tidyNodeIsLINK');
    tidyNodeIsLISTING := GetProcAddress(Result, 'tidyNodeIsLISTING');
    tidyNodeIsMAP := GetProcAddress(Result, 'tidyNodeIsMAP');
    tidyNodeIsMARQUEE := GetProcAddress(Result, 'tidyNodeIsMARQUEE');
    tidyNodeIsMENU := GetProcAddress(Result, 'tidyNodeIsMENU');
    tidyNodeIsMETA := GetProcAddress(Result, 'tidyNodeIsMETA');
    tidyNodeIsNOBR := GetProcAddress(Result, 'tidyNodeIsNOBR');
    tidyNodeIsNOFRAMES := GetProcAddress(Result, 'tidyNodeIsNOFRAMES');
    tidyNodeIsNOSCRIPT := GetProcAddress(Result, 'tidyNodeIsNOSCRIPT');
    tidyNodeIsOBJECT := GetProcAddress(Result, 'tidyNodeIsOBJECT');
    tidyNodeIsOL := GetProcAddress(Result, 'tidyNodeIsOL');
    tidyNodeIsOPTGROUP := GetProcAddress(Result, 'tidyNodeIsOPTGROUP');
    tidyNodeIsOPTION := GetProcAddress(Result, 'tidyNodeIsOPTION');
    tidyNodeIsP := GetProcAddress(Result, 'tidyNodeIsP');
    tidyNodeIsPARAM := GetProcAddress(Result, 'tidyNodeIsPARAM');
    tidyNodeIsPRE := GetProcAddress(Result, 'tidyNodeIsPRE');
    tidyNodeIsProp := GetProcAddress(Result, 'tidyNodeIsProp');
    tidyNodeIsQ := GetProcAddress(Result, 'tidyNodeIsQ');
    tidyNodeIsS := GetProcAddress(Result, 'tidyNodeIsS');
    tidyNodeIsSCRIPT := GetProcAddress(Result, 'tidyNodeIsSCRIPT');
    tidyNodeIsSELECT := GetProcAddress(Result, 'tidyNodeIsSELECT');
    tidyNodeIsSMALL := GetProcAddress(Result, 'tidyNodeIsSMALL');
    tidyNodeIsSPACER := GetProcAddress(Result, 'tidyNodeIsSPACER');
    tidyNodeIsSPAN := GetProcAddress(Result, 'tidyNodeIsSPAN');
    tidyNodeIsSTRIKE := GetProcAddress(Result, 'tidyNodeIsSTRIKE');
    tidyNodeIsSTRONG := GetProcAddress(Result, 'tidyNodeIsSTRONG');
    tidyNodeIsSTYLE := GetProcAddress(Result, 'tidyNodeIsSTYLE');
    tidyNodeIsTABLE := GetProcAddress(Result, 'tidyNodeIsTABLE');
    tidyNodeIsTD := GetProcAddress(Result, 'tidyNodeIsTD');
    tidyNodeIsText := GetProcAddress(Result, 'tidyNodeIsText');
    tidyNodeIsTEXTAREA := GetProcAddress(Result, 'tidyNodeIsTEXTAREA');
    tidyNodeIsTH := GetProcAddress(Result, 'tidyNodeIsTH');
    tidyNodeIsTITLE := GetProcAddress(Result, 'tidyNodeIsTITLE');
    tidyNodeIsTR := GetProcAddress(Result, 'tidyNodeIsTR');
    tidyNodeIsU := GetProcAddress(Result, 'tidyNodeIsU');
    tidyNodeIsUL := GetProcAddress(Result, 'tidyNodeIsUL');
    tidyNodeIsWBR := GetProcAddress(Result, 'tidyNodeIsWBR');
    tidyNodeIsXMP := GetProcAddress(Result, 'tidyNodeIsXMP');
    tidyNodeLine := GetProcAddress(Result, 'tidyNodeLine');

    tidyOptCopyConfig := GetProcAddress(Result, 'tidyOptCopyConfig');
    tidyOptDiffThanDefault := GetProcAddress(Result, 'tidyOptDiffThanDefault');
    tidyOptDiffThanSnapshot := GetProcAddress(Result, 'tidyOptDiffThanSnapshot');
    tidyOptGetBool := GetProcAddress(Result, 'tidyOptGetBool');
    tidyOptGetCategory := GetProcAddress(Result, 'tidyOptGetCategory');
    tidyOptGetCurrPick := GetProcAddress(Result, 'tidyOptGetCurrPick');
    tidyOptGetDeclTagList := GetProcAddress(Result, 'tidyOptGetDeclTagList');
    tidyOptGetDefault := GetProcAddress(Result, 'tidyOptGetDefault');
    tidyOptGetDefaultBool := GetProcAddress(Result, 'tidyOptGetDefaultBool');
    tidyOptGetDefaultInt := GetProcAddress(Result, 'tidyOptGetDefaultInt');
    tidyOptGetDoc := GetProcAddress(Result, 'tidyOptGetDoc');
    tidyOptGetDocLinksList := GetProcAddress(Result, 'tidyOptGetDocLinksList');
    tidyOptGetEncName := GetProcAddress(Result, 'tidyOptGetEncName');
    tidyOptGetId := GetProcAddress(Result, 'tidyOptGetId');
    tidyOptGetIdForName := GetProcAddress(Result, 'tidyOptGetIdForName');
    tidyOptGetInt := GetProcAddress(Result, 'tidyOptGetInt');
    tidyOptGetName := GetProcAddress(Result, 'tidyOptGetName');
    tidyOptGetNextDeclTag := GetProcAddress(Result, 'tidyOptGetNextDeclTag');
    tidyOptGetNextDocLinks := GetProcAddress(Result, 'tidyOptGetNextDocLinks');
    tidyOptGetNextPick := GetProcAddress(Result, 'tidyOptGetNextPick');
    tidyOptGetPickList := GetProcAddress(Result, 'tidyOptGetPickList');
    tidyOptGetType := GetProcAddress(Result, 'tidyOptGetType');
    tidyOptGetValue := GetProcAddress(Result, 'tidyOptGetValue');
    tidyOptIsReadOnly := GetProcAddress(Result, 'tidyOptIsReadOnly');
    tidyOptParseValue := GetProcAddress(Result, 'tidyOptParseValue');
    tidyOptResetAllToDefault := GetProcAddress(Result, 'tidyOptResetAllToDefault');
    tidyOptResetToDefault := GetProcAddress(Result, 'tidyOptResetToDefault');
    tidyOptResetToSnapshot := GetProcAddress(Result, 'tidyOptResetToSnapshot');
    tidyOptSaveFile := GetProcAddress(Result, 'tidyOptSaveFile');
    tidyOptSaveSink := GetProcAddress(Result, 'tidyOptSaveSink');
    tidyOptSetBool := GetProcAddress(Result, 'tidyOptSetBool');
    tidyOptSetInt := GetProcAddress(Result, 'tidyOptSetInt');
    tidyOptSetValue := GetProcAddress(Result, 'tidyOptSetValue');
    tidyOptSnapshot := GetProcAddress(Result, 'tidyOptSnapshot');

    tidyParseBuffer := GetProcAddress(Result, 'tidyParseBuffer');
    tidyParseFile := GetProcAddress(Result, 'tidyParseFile');
    tidyParseSource := GetProcAddress(Result, 'tidyParseSource');
    tidyParseStdin := GetProcAddress(Result, 'tidyParseStdin');
    tidyParseString := GetProcAddress(Result, 'tidyParseString');

    tidyPutByte := GetProcAddress(Result, 'tidyPutByte');
    tidyRelease := GetProcAddress(Result, 'tidyRelease');
    tidyReleaseDate := GetProcAddress(Result, 'tidyReleaseDate');
    tidyRunDiagnostics := GetProcAddress(Result, 'tidyRunDiagnostics');
    tidySaveBuffer := GetProcAddress(Result, 'tidySaveBuffer');
    tidySaveFile := GetProcAddress(Result, 'tidySaveFile');
    tidySaveSink := GetProcAddress(Result, 'tidySaveSink');
    tidySaveStdout := GetProcAddress(Result, 'tidySaveStdout');
    tidySaveString := GetProcAddress(Result, 'tidySaveString');
    tidySetAppData := GetProcAddress(Result, 'tidySetAppData');
    tidySetCharEncoding := GetProcAddress(Result, 'tidySetCharEncoding');
    tidySetErrorBuffer := GetProcAddress(Result, 'tidySetErrorBuffer');
    tidySetErrorFile := GetProcAddress(Result, 'tidySetErrorFile');
    tidySetErrorSink := GetProcAddress(Result, 'tidySetErrorSink');
    tidySetFreeCall := GetProcAddress(Result, 'tidySetFreeCall');
    tidySetInCharEncoding := GetProcAddress(Result, 'tidySetInCharEncoding');
    tidySetMallocCall := GetProcAddress(Result, 'tidySetMallocCall');
    tidySetOptionCallback := GetProcAddress(Result, 'tidySetOptionCallback');
    tidySetOutCharEncoding := GetProcAddress(Result, 'tidySetOutCharEncoding');
    tidySetPanicCall := GetProcAddress(Result, 'tidySetPanicCall');
    tidySetReallocCall := GetProcAddress(Result, 'tidySetReallocCall');
    tidySetReportFilter := GetProcAddress(Result, 'tidySetReportFilter');
    tidyStatus := GetProcAddress(Result, 'tidyStatus');
    tidyUngetByte := GetProcAddress(Result, 'tidyUngetByte');
    tidyWarningCount := GetProcAddress(Result, 'tidyWarningCount');
    HTidy := Result;
  end;
end;

end.

