{$J+,Z4}
unit ToolTip;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, RichEdit, RichEd20, Messages, CommCtrl, SysUtils, Classes, Graphics,
  Forms, Controls, ComCtrls, StdCtrls, StrUtil, WindowUtils, UxThemes;

{$IFDEF VER90}

type
  PTrackMouseEvent = ^TTrackMouseEvent;
  TTrackMouseEvent = Record
    cbSize: DWORD;
    dwFlags: DWORD;
    hwndTrack: HWND;
    dwHoverTime: DWORD;
  end;

const
  DT_HIDEPREFIX = $00100000;
  DT_PREFIXONLY = $00200000;

// correcting typo in Delphi 2's CommCtrl.pas
const
  TOOLTIPS_CLASS = TOOTIPS_CLASS;

{$ELSE}

type
  PTrackMouseEvent = Windows.PTrackMouseEvent;
  TTrackMouseEvent = Windows.TTrackMouseEvent;

{$ENDIF}

// (partly) coming with ComCtl32.dll v4.70 or later
type
  PToolInfo = ^TToolInfo;
  TToolInfo = Packed Record
    cbSize: UINT;
    uFlags: UINT;
    hwnd: HWND;
    uId: UINT;
    Rect: TRect;
    hInst: HINST;
    lpszText: PChar;
    lParam: LPARAM;
    lpReserved: Pointer;
  end;

const
  // additional ToolTip flags
  TTF_IDISHWND		= $0001;
  TTF_CENTERTIP		= $0002;
  TTF_RTLREADING	= $0004;
  TTF_SUBCLASS		= $0010;
  TTF_TRACK		= $0020;
  TTF_ABSOLUTE		= $0080;
  TTF_TRANSPARENT	= $0100;
  TTF_PARSELINKS	= $1000;
  TTF_DI_SETITEM	= $8000;

  // additional ToolTip messages
  TTM_TRACKACTIVATE	= WM_USER + 17;
  TTM_TRACKPOSITION	= WM_USER + 18;
  TTM_SETTIPBKCOLOR	= WM_USER + 19;
  TTM_SETTIPTEXTCOLOR	= WM_USER + 20;
  TTM_GETDELAYTIME	= WM_USER + 21;
  TTM_GETTIPBKCOLOR	= WM_USER + 22;
  TTM_GETTIPTEXTCOLOR	= WM_USER + 23;
  TTM_SETMAXTIPWIDTH	= WM_USER + 24;
  TTM_GETMAXTIPWIDTH	= WM_USER + 25;
  TTM_SETMARGIN		= WM_USER + 26;
  TTM_GETMARGIN		= WM_USER + 27;
  TTM_POP		= WM_USER + 28;
  TTM_UPDATE		= WM_USER + 29;
  TTM_GETBUBBLESIZE	= WM_USER + 30;
  TTM_ADJUSTRECT	= WM_USER + 31;
  TTM_SETTITLE		= WM_USER + 32;
  TTM_POPUP		= WM_USER + 34;
  TTM_GETTITLE		= WM_USER + 35;

  // additional ToolTip themes
  TTM_SETWINDOWTHEME	= $0000200B;

  // additionl ToolTip notifications
  TTN_GETDISPINFO	= TTN_NEEDTEXT;

  // additional ToolTip styles
  TTS_NOANIMATE		= $10;
  TTS_NOFADE		= $20;
  TTS_BALLOON		= $40;
  TTS_CLOSE		= $80;
  TTS_USEVISUALSTYLE	= $100;

  // mouse tracking messages
  WM_MOUSEHOVER		= $000002A1;
  WM_MOUSELEAVE		= $000002A3;
  WM_NCMOUSEHOVER	= $000002A0;
  WM_NCMOUSELEAVE	= $000002A2;

  // mouse tracking events
  TME_HOVER		= $00000001;
  TME_LEAVE		= $00000002;
  TME_NONCLIENT		= $00000010;
  TME_QUERY		= $40000000;
  TME_CANCEL		= $80000000;

function TrackMouseEvent(var EventTrack: TTrackMouseEvent): BOOL; stdcall;
			 external 'comctl32.dll' name '_TrackMouseEvent';

type
  PNMTTDispInfo = ^TNMTTDispInfo;
  TNMTTDispInfo = Packed Record
    hdr: TNMHdr;
    lpszText: PChar;
    szText: Array[0..79] of Char;
    hinst: HINST;
    uFlags: UINT;
    lParam: LPARAM;
  end;

// TToolTip implementation
const
  MAX_DELAYTIME = $7FFF;

const
  WC_TOOLTIP: PWideChar = 'TOOLTIP';

const
  TOOLRECT_EMPTY: TRect = (Left: 0; Top: 0; Right: 0; Bottom: 0);

type // TToolTip types
  TCenterToolTip = (cttNot, cttTrack, cttSubMouse, cttInParent, cttSubParent, cttInToolRect, cttSubToolRect);
  TToolTipStyle = (ttsLiteralPrefix, ttsCenterText, ttsUnderscore, ttsRichText);
  TOnGetOversizedText = function(const Text: String): String of Object;
  TOnGetToolTipOrigin = procedure(var TopLeft: TPoint) of Object;
  TOnLinkClick = function(const Url: String): String of Object;
  TTipTextInfo = Record
    TipText: String;
    TextHeight: Integer;
  end;
  TToolTip = Class(TWinControl)
  private
    FAppOnMessage: TMessageEvent;
    FCenterToolTip: TCenterToolTip;
    FChildTheme: THandle;
    FChildHandle: THandle;
    FChildProc: Pointer;
    FChildFont: TFont;
    FCursorPosOnShow: TPoint;
    FDrawStyle: Integer;
    FFixedPos: Boolean;
    FFocusHandle: THandle;
    FKeepPosition: Boolean;
    FHMargin: Integer;
    FVMargin: Integer;
    FMaxWidth: Integer;
    FTextHeight: Integer;
    FOversizeOnly: Boolean;
    FParentHandle: THandle;
    FParentProc: Pointer;
    FParentFont: TFont;
    FRichEditRect: TRect;
    FRichEditVersion: Integer;
    FToolInfo: TToolInfo;
    FToolTipOrigin: TPoint;
    FToolTipStyle: TToolTipStyle;
    FToolTipText: String;
    FUnicodeText: String;
    FOnGetOversizedText: TOnGetOversizedText;
    FOnGetToolTipOrigin: TOnGetToolTipOrigin;
    FOnHideToolTip: TNotifyEvent;
    FOnShowToolTip: TNotifyEvent;
    FOnLinkClick: TOnLinkClick;
    function  RichEditKeyDown(Msg: Cardinal; CharCode: Word; KeyData: Longint): Boolean;
    function  RichEditLinkClicked(const CharRange: TCharRange): Boolean;
    function  OversizedText(const Text: String): String;
    function  DefaultCharFormat: TCharFormat;
    function  OpenChildTheme: THandle;
    procedure CloseChildTheme;
    function  GetChildFont: TFont;
    function  ChildColors(ChildDC: THandle): THandle;
    procedure SetActivate(Value: Boolean);
    function  GetHintPause: Integer;
    procedure SetHintPause(Value: Integer);
    function  GetHintHidePause: Integer;
    procedure SetHintHidePause(Value: Integer);
    function  GetHintShortPause: Integer;
    procedure SetHintShortPause(Value: Integer);
    procedure SetMaxWidth(Value: Integer);
    function  GetParentFontHandle: THandle;
    function  GetParentWidth: Integer;
    function  GetShowing: Boolean;
    function  GetTextRect: TRect;
    procedure GetTextRectDimensions(var Width, Height: Integer);
    procedure GetWindowPos(var WindowPos: TWindowPos);
    procedure SetToolRect(const Value: TRect);
    procedure SetToolTipStyle(Value: TToolTipStyle);
    procedure SetToolTipText(const Value: String);
    function  GetVisible: Boolean;
    procedure SetVisible(Value: Boolean);
    procedure SetAnsiText(const Value: String);
    procedure SetUtf8Text(const Value: String);
    procedure SetRichText(const Value: String);
    procedure LoadFromRtfStream(const RichText: String);
    function  ChildPaint(ChildHandle: THandle): Integer;
    function  ChildEraseBkGnd(ChildHandle, ChildDC: THandle): Integer;
    function  Reactivate(ToolTip: THandle): Boolean;
    procedure ShowHandler;
    procedure PopHandler;
    procedure SetAppOnMessage(Activate: Boolean);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMParentColorChanged(var Message: TMessage); message CM_PARENTCOLORCHANGED;
    procedure CMParentFontChanged(var Message: TMessage); message CM_PARENTFONTCHANGED;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure WMCreate;
    procedure WMDestroy;
    procedure WMPaint(var Message: TWMPaint);
    procedure WMShowWindow(var Message: TWMShowWindow);
    procedure WMMouseActivate(var Message: TWMMouseActivate);
    procedure WMSetFont(var Message: TWMSetFont);
    procedure WMWindowPosChanging(var Message: TWMWindowPosChanging);
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged);
    procedure WMNotify(var Message: TWMNotify);
    procedure WMCommand(var Message: TWMCommand);
    procedure WMCtlColorStatic(var Message: TWMCtlColorStatic);
    procedure WMParentNotify(var Message: TWMParentNotify);
    procedure AppOnMessage(var Msg: TMsg; var Handled: Boolean);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WndProc(var Message: TMessage); override;
    function  AddToolTip: Boolean;
    function  DelToolTip: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function  HasText: Boolean;
    procedure Show;
    procedure Hide;
    procedure ResetHintPauseToDefault;
    procedure SetParent(AParent: TWinControl); override;
    property  Activate: Boolean write SetActivate;
    property  CursorPosOnShow: TPoint read FCursorPosOnShow;
    property  ParentHandle: THandle read FParentHandle write FParentHandle;
    property  Showing: Boolean read GetShowing;
    property  ToolRect: TRect read FToolInfo.rect write SetToolRect;
    property  AnsiText: String write SetAnsiText;
    property  Utf8Text: String write SetUtf8Text;
  published
    property  CenterToolTip: TCenterToolTip read FCenterToolTip write FCenterToolTip;
    property  FixedPos: Boolean read FFixedPos write FFixedPos;
    property  HintPause: Integer read GetHintPause write SetHintPause;
    property  HintHidePause: Integer read GetHintHidePause write SetHintHidePause;
    property  HintShortPause: Integer read GetHintShortPause write SetHintShortPause;
    property  HMargin: Integer read FHMargin write FHMargin;
    property  VMargin: Integer read FVMargin write FVMargin;
    property  MaxWidth: Integer read FMaxWidth write SetMaxWidth;
    property  TextHeight: Integer read FTextHeight write FTextHeight;
    property  OversizeOnly: Boolean read FOversizeOnly write FOversizeOnly;
    property  ToolTipOrigin: TPoint read FToolTipOrigin write FToolTipOrigin;
    property  ToolTipStyle: TToolTipStyle read FToolTipStyle write SetToolTipStyle;
    property  Visible: Boolean read GetVisible write SetVisible;
    property  OnGetOversizedText: TOnGetOversizedText read FOnGetOversizedText write FOnGetOversizedText;
    property  OnGetToolTipOrigin: TOnGetToolTipOrigin read FOnGetToolTipOrigin write FOnGetToolTipOrigin;
    property  OnHideToolTip: TNotifyEvent read FOnHideToolTip write FOnHideToolTip;
    property  OnShowToolTip: TNotifyEvent read FOnShowToolTip write FOnShowToolTip;
    property  OnLinkClick: TOnLinkClick read FOnLinkClick write FOnLinkClick;
    property  Color;
    property  Font;
    property  ParentColor;
    property  ParentFont;
  end;


type
  TTitleIconIndex = (TTI_NONE, TTI_INFO, TTI_WARNING, TTI_ERROR, TTI_PREVIOUS);
  TTitleData = Record
    tdIcon: TTitleIconIndex;
    tdTitle: String;
  end;


function CreateTrackingToolTip(ParentHandle, ToolHandle: THandle; const ToolRect: TRect;
			       const TipText: String; MaxTipWidth: Cardinal; const TitleData: TTitleData; UserData: Pointer;
			       AbsolutePos: Boolean; Balloon: Boolean; ParseLinks: Boolean): THandle;
function GetToolTipData(ToolTipHandle, ParentHandle: THandle; ToolHandle: THandle): TToolInfo;
function ShowTrackingToolTip(ToolTipHandle, ParentHandle, ToolHandle: THandle; const AddError: String;
			     const ToolTipPos: TPoint; PTI: PToolInfo): Longbool;
procedure HideTrackingToolTip(ToolTipHandle, ParentHandle, ToolHandle: THandle; PTI: PToolInfo);
function DestroyTrackingToolTip(ToolTipHandle, ParentHandle, ToolHandle: THandle): THandle;


implementation

const
  TTS_DEFAULT = WS_POPUP or TTS_ALWAYSTIP or TTS_USEVISUALSTYLE or TTS_NOANIMATE or TTS_NOFADE or TTS_NOPREFIX;

procedure TrackActivateToolTip(ToolTipHandle, ParentHandle: THandle; Activate: Boolean; ToolHandle: THandle; PTI: PToolInfo);
var
  tiNew: TToolInfo;
begin
  if (ToolTipHandle <> 0) and (ParentHandle <> 0) then begin
    if PTI = NIL then with tiNew do begin
      FillChar(tiNew, SizeOf(TToolInfo), 0);
      cbSize := SizeOf(TToolInfo);
      uFlags := TTF_IDISHWND or TTF_TRACK;
      hwnd := ParentHandle;
      if ToolHandle = 0 then
	uId := ParentHandle
      else uId := ToolHandle;
      hinst := HInstance;
      PTI := @tiNew
    end;
    SendMessage(ToolTipHandle, TTM_TRACKACTIVATE, Ord(Activate), Longint(PTI));
  end;
end;

function CreateTrackingToolTip(ParentHandle, ToolHandle: THandle; const ToolRect: TRect;
			       const TipText: String; MaxTipWidth: Cardinal; const TitleData: TTitleData; UserData: Pointer;
			       AbsolutePos: Boolean; Balloon: Boolean; ParseLinks: Boolean): THandle;
var
  ti: TToolInfo;
begin
  Result := CreateWindowEx(WS_EX_TOOLWINDOW, TOOLTIPS_CLASS, NIL, TTS_DEFAULT or (TTS_BALLOON * Byte(Balloon)),
			   CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 0, 0, HInstance, NIL);
  if Result <> 0 then begin
    FillChar(ti, SizeOf(TToolInfo), 0);
    with ti do begin
      cbSize := SizeOf(TToolInfo);
      uFlags := TTF_ABSOLUTE * Byte(AbsolutePos) or TTF_IDISHWND or TTF_PARSELINKS * Byte(ParseLinks) or TTF_TRACK;
      hwnd := ParentHandle;
      if ToolHandle = 0 then
	uId := ParentHandle
      else uId := ToolHandle;
      rect := ToolRect;
      hinst := HInstance;
      lpszText := PChar(TipText);
      lParam := Integer(UserData);
    end;
    if TitleData.tdIcon > TTI_NONE then with TitleData do SendMessage(Result, TTM_SETTITLE, Ord(tdIcon), Longint(PChar(tdTitle)));
    if MaxTipWidth > 0 then SendMessage(Result, TTM_SETMAXTIPWIDTH, 0, Longint(MaxTipWidth));
    SendMessage(Result, TTM_ADDTOOL, 0, Longint(@ti));
  end
  else Result := 0;
end;

function GetToolTipData(ToolTipHandle, ParentHandle: THandle; ToolHandle: THandle): TToolInfo;
var
  sToolTipText: String;
begin
  if (ToolTipHandle <> 0) and (ParentHandle <> 0) then with Result do begin
    FillChar(Result, SizeOf(TToolInfo), 0);
    SetString(sToolTipText, NIL, $FFFF);
    cbSize := SizeOf(TToolInfo);
    uFlags := TTF_IDISHWND;
    hwnd := ParentHandle;
    if ToolHandle = 0 then
      uId := ParentHandle
    else uId := ToolHandle;
    hinst := HInstance;
    lpszText := Pointer(sToolTipText);
    if (SendMessage(ToolTipHandle, TTM_GETTOOLINFO, 0, Longint(@Result)) <> 0) then begin
      Delete(sToolTipText, succ(StrLen(PChar(sToolTipText))), MAXINT);
      Exit;
    end;
  end;
  FillChar(Result, SizeOf(TToolInfo), 0);
end;

function ShowTrackingToolTip(ToolTipHandle, ParentHandle, ToolHandle: THandle; const AddError: String;
			     const ToolTipPos: TPoint; PTI: PToolInfo): Longbool;
var
  ti: TToolInfo;
begin
  if ToolTipHandle <> 0 then begin
    with ToolTipPos do SendMessage(ToolTipHandle, TTM_TRACKPOSITION, 0, MakeLong(X, Y));
    if AddError <> E then begin
      ti := GetToolTipData(ToolTipHandle, ParentHandle, ToolHandle);
      ti.lpszText := PChar(ti.lpszText + AddError);
      SendMessage(ToolTipHandle, TTM_UPDATETIPTEXT, 0, Longint(@ti));
      TrackActivateToolTip(ToolTipHandle, ParentHandle, TRUE, ToolHandle, @ti);
    end
    else begin
      ti := GetToolTipData(ToolTipHandle, ParentHandle, ToolHandle);
      if ti.lpszText <> E then TrackActivateToolTip(ToolTipHandle, ParentHandle, TRUE, ToolHandle, @ti);
    end;
  end;
  Result := (ToolTipHandle <> 0) and (ti.lpszText <> E);
end;

procedure HideTrackingToolTip(ToolTipHandle, ParentHandle, ToolHandle: THandle; PTI: PToolInfo);
begin
  TrackActivateToolTip(ToolTipHandle, ParentHandle, FALSE, ToolHandle, PTI);
end;

function DestroyTrackingToolTip(ToolTipHandle, ParentHandle, ToolHandle: THandle): THandle;
var
  ti: TToolInfo;
begin
  Result := ToolTipHandle;
  FillChar(ti, SizeOf(TToolInfo), 0);
  with ti do begin
    cbSize := SizeOf(TToolInfo);
    hwnd := ParentHandle;
    if ToolHandle = 0 then
      uId := ParentHandle
    else uId := ToolHandle;
    hinst := HInstance;
  end;
  HideTrackingToolTip(Result, ParentHandle, ToolHandle, @ti);
  SendMessage(Result, TTM_DELTOOL, 0, Longint(@ti));
  if DestroyWindow(Result) then Result := 0;
end;


function ChildProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  ttSelf: TToolTip;
  tme: TTrackMouseEvent;
begin
  ttSelf := TToolTip(GetProp(hWnd, TOOLTIPS_CLASS));
  case Msg of
    WM_PAINT: begin
      Result := ttSelf.ChildPaint(hWnd);
      if Result = 0 then Exit;
    end;
    WM_ERASEBKGND: begin
      Result := ttSelf.ChildEraseBkGnd(hWnd, wParam);
      if Result <> 0 then Exit;
    end;
    WM_MOUSEMOVE: begin
      with tme do begin
	cbSize := SizeOf(TTrackMouseEvent);
	dwFlags := TME_LEAVE;
	hwndTrack := hWnd;
	dwHoverTime := 0;
      end;
      TrackMouseEvent(tme);
    end;
    WM_MOUSELEAVE: ttSelf.Hide;
  end;
  Result := CallWindowProc(ttSelf.FChildProc, hWnd, Msg, wParam, lParam);
  if Msg = WM_PAINT then HideCaret(hWnd);
end;

function ParentProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  ttSelf: TToolTip;
begin
  ttSelf := TToolTip(GetProp(hWnd, TOOLTIPS_CLASS));
  case Msg of
    WM_NOTIFY: begin
      case PNMHdr(lParam)^.code of
	TTN_SHOW: ttSelf.ShowHandler;
	TTN_POP: ttSelf.PopHandler;
      end;
    end;
  end;
  Result := CallWindowProc(ttSelf.FParentProc, hWnd, Msg, wParam, lParam);
end;

function ToolTipProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  ttSelf: TToolTip;
  Message: TMessage;
begin
  Result := 0;
  Message.Msg := Msg;
  Message.wParam := wParam;
  Message.lParam := lParam;
  Message.Result := Result;
  ttSelf := TToolTip(GetWindowLong(hWnd, GWL_USERDATA));
  if ttSelf <> nil then ttSelf.WndProc(Message);
  Result := Message.Result;
end;

function TToolTip.OversizedText(const Text: String): String;
var
  hParentTheme, hParentDC, hFont: THandle;
  rParent, rText: TRect;
begin
  if Assigned(FOnGetOversizedText) then
    Result := FOnGetOversizedText(Text)
  else begin
    hParentTheme := 0;
    hParentDC := GetDC(FParentHandle);
    hFont := SelectObject(hParentDC, GetParentFontHandle);
    try
      rParent := WindowUtils.GetClientRect(FParentHandle);
      rText := rParent;
      if FChildTheme <> 0 then begin
        hParentTheme := OpenThemeData(FParentHandle, 'EDIT');
	GetThemeTextExtent(hParentTheme, hParentDC, EP_EDITTEXT, ETS_NORMAL,
			   Pointer(AnsiToWideString(Text)), -1, FDrawStyle, @rText, rText);
      end
      else DrawText(hParentDC, PChar(Text), -1, rText, FDrawStyle or DT_CALCRECT);
      if rText.Right > rParent.Right then
	Result := Text
      else Result := E;
    finally
      if hParentTheme <> 0 then CloseThemeData(hParentTheme);
      if hFont <> 0 then SelectObject(hParentDC, hFont);
      ReleaseDC(FParentHandle, hParentDC);
    end;
  end;
end;

function TToolTip.DefaultCharFormat: TCharFormat;
begin
  with FChildFont do Result := GetCharFormatFromFont(FChildFont, Name, succ(Size), CFM_FONT);
end;

function TToolTip.OpenChildTheme: THandle;
begin
  if UseThemes then begin
    Result := OpenThemeData(FChildHandle, 'TOOLTIP');
    Exit;
  end;
  Result := 0;
end;

procedure TToolTip.CloseChildTheme;
begin
  if FChildTheme <> 0 then begin
    CloseThemeData(FChildTheme);
    FChildTheme := 0;
  end;
end;

function TToolTip.GetChildFont: TFont;
begin
  if ParentFont then
    Result := FParentFont
  else Result := FChildFont;
end;

function TToolTip.ChildColors(ChildDC: THandle): THandle;
begin
  Result := Brush.Handle;
end;

function TToolTip.RichEditKeyDown(Msg: Cardinal; CharCode: Word; KeyData: Longint): Boolean;

  function GotoNextLink: Boolean;

    procedure CaretMove(Highlight, WordJump, Right: Boolean);
    var
      ks: TKeyboardState;
      bCtrl, bShift: Byte;
      lKey: Longint;
    begin
      GetKeyBoardState(ks);
      bCtrl := ks[VK_CONTROL];
      bShift := ks[VK_SHIFT];
      if Highlight then
	ks[VK_SHIFT] := bShift or $80
      else ks[VK_SHIFT] := bShift and $7F;
      if WordJump then
	ks[VK_CONTROL] := bCtrl or $80
      else ks[VK_CONTROL] := bCtrl and $7F;
      SetKeyboardState(ks);
      if Right then
	lKey := VK_RIGHT
      else lKey := VK_LEFT;
      SendMessage(FChildHandle, WM_KEYDOWN, lKey, 0);
      GetKeyBoardState(ks);
      ks[VK_CONTROL] := bCtrl;
      ks[VK_SHIFT] := bShift;
      SetKeyboardState(ks);
    end;

  var
    fForward: Boolean;
    rChildClient: TRect;
    crSelection: TCharRange;
    cfSelection: TCharFormat;
    lMaxPos, lBegPos, lEndPos, lSelStart: Longint;
  begin
    fForward := (GetKeyState(VK_SHIFT) >= 0);
    SendMessage(FChildHandle, WM_SETREDRAW, 0, 0);
    try
      rChildClient := WindowUtils.GetClientRect(FChildHandle);
      SendMessage(FChildHandle, EM_EXGETSEL, 0, Longint(@crSelection));
      with cfSelection do begin
	cbSize := SizeOf(TCharFormat);
	dwMask := CFM_LINK;
      end;
      lMaxPos := GetRtfTextLength(FChildHandle, FRichEditVersion);
      with crSelection do begin
	lBegPos := cpMin;
	lEndPos := cpMax;
      end;
      Result := (lBegPos = lEndPos);
      if fForward then begin
	if lEndPos > lBegPos then begin
	  lBegPos := succ(lEndPos);
	  SendMessage(FChildHandle, EM_SETSEL, lBegPos, lBegPos);
	  SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	  if lEndPos = crSelection.cpMax then begin
	    SendMessage(FChildHandle, EM_EXSETSEL, 0, Longint(@crSelection));
	    Exit;
	  end;
	end;
	SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	Result := (Result or (lBegPos = 0)) and (cfSelection.dwEffects and CFE_LINK <> 0);
	if not Result then begin
	  while lBegPos <= lMaxPos do begin
	    SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	    Result := (cfSelection.dwEffects and CFE_LINK <> 0);
	    CaretMove(false, false, true);
	    SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	    if Result and (lEndPos > lBegPos) then
	      inc(lBegPos)
	    else Break;
	  end;
	  repeat
	    SendMessage(FChildHandle, EM_GETSEL, Longint(@lBegPos), 0);
	    CaretMove(false, true, true);
	    SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	    CaretMove(true, false, true);
	    SendMessage(FChildHandle, EM_GETSEL, Longint(@lSelStart), 0);
	    if lSelStart = lEndPos then begin
	      SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	      CaretMove(true, false, false);
	      Result := (cfSelection.dwEffects and CFE_LINK <> 0);
	    end
	    else begin
	      SendMessage(FChildHandle, EM_SETSEL, lBegPos, lBegPos);
	      CaretMove(false, true, true);
	    end;
	  until Result or (lBegPos = lEndPos);
	end;
      end
      else begin
	if lEndPos > lBegPos then begin
	  lBegPos := pred(lBegPos);
	  SendMessage(FChildHandle, EM_SETSEL, lBegPos, lBegPos);
	  SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	  if lEndPos = crSelection.cpMin then begin
	    SendMessage(FChildHandle, EM_EXSETSEL, 0, Longint(@crSelection));
	    Exit;
	  end;
	end;
	SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	Result := (Result or (lBegPos = lMaxPos)) and (cfSelection.dwEffects and CFE_LINK <> 0);
	if not Result then begin
	  while lBegPos > 0 do begin
	    SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	    Result := (cfSelection.dwEffects and CFE_LINK <> 0);
	    CaretMove(false, false, false);
	    SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	    if Result and (lEndPos < lBegPos) then
	      dec(lBegPos)
	    else Break;
	  end;
	  repeat
	    SendMessage(FChildHandle, EM_GETSEL, Longint(@lBegPos), 0);
	    CaretMove(false, true, false);
	    SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	    CaretMove(true, false, true);
	    SendMessage(FChildHandle, EM_GETSEL, Longint(@lSelStart), 0);
	    if lSelStart = lEndPos then begin
	      SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	      CaretMove(true, false, false);
	      Result := (cfSelection.dwEffects and CFE_LINK <> 0);
	    end
	    else begin
	      SendMessage(FChildHandle, EM_SETSEL, lBegPos, lBegPos);
	      CaretMove(false, true, false);
	    end;
	  until Result or (lBegPos = lEndPos);
	end;
      end;
      if Result then begin
	repeat
	  SendMessage(FChildHandle, EM_GETSEL, Longint(@lBegPos), 0);
	  CaretMove(false, false, false);
	  SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	  SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	until (cfSelection.dwEffects and CFE_LINK = 0) or (lEndPos = lBegPos);
	if lEndPos <> lBegPos then CaretMove(false, false, true);
	repeat
	  SendMessage(FChildHandle, EM_GETSEL, Longint(@lBegPos), 0);
	  CaretMove(true, false, true);
	  SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
	  SendMessage(FChildHandle, EM_GETSEL, 0, Longint(@lEndPos));
	until (cfSelection.dwEffects and CFE_LINK = 0) or (lEndPos = lBegPos);
	if lEndPos <> lBegPos then CaretMove(true, false, false);
      end;
    finally
      SendMessage(FChildHandle, WM_SETREDRAW, 1, 0);
      InvalidateRect(FChildHandle, @rChildClient, false);
    end;
  end;

  function CheckLink: Boolean;

    function ClientPosFromChar(CharPos: Longint): Longint;
    var
      ptPos: TPoint;
    begin
      if FRichEditVersion <> 2 then begin
	SendMessage(FChildHandle, EM_POSFROMCHAR, Longint(@ptPos), CharPos);
	with ptPos do Result := MakeLong(X, Y);
      end
      else Result := SendMessage(FChildHandle, EM_POSFROMCHAR, CharPos, 0);
    end;

  var
    lBegPos, lEndPos, lClickPos: Longint;
    cfSelection: TCharFormat;
  begin
    SendMessage(FChildHandle, WM_SETREDRAW, 0, 0);
    try
      SendMessage(FChildHandle, EM_GETSEL, Longint(@lBegPos), Longint(@lEndPos));
      if lEndPos > lBegPos then SendMessage(FChildHandle, EM_SETSEL, lBegPos, pred(lEndPos));
      with cfSelection do begin
	cbSize := SizeOf(TCharFormat);
	dwMask := CFM_LINK;
      end;
      SendMessage(FChildHandle, EM_GETCHARFORMAT, SCF_SELECTION, Longint(@cfSelection));
      Result := (cfSelection.dwEffects and CFE_LINK <> 0);
      if Result then begin
	lClickPos := ClientPosFromChar(lBegPos);
	SendMessage(FChildHandle, WM_LBUTTONDOWN, MK_LBUTTON, lClickPos);
	SendMessage(FChildHandle, WM_LBUTTONUP, MK_LBUTTON, lClickPos);
      end;
    finally
      SendMessage(FChildHandle, WM_SETREDRAW, 1, 0);
      ProcessMessages;
    end;
  end;

begin
  Result := false;
  if Showing then begin
    case CharCode of
      VK_TAB: Result := GotoNextLink;
      VK_RETURN: Result := CheckLink;
    end;
  end;
end;

function TToolTip.RichEditLinkClicked(const CharRange: TCharRange): Boolean;
var
  sLink, sNewText: String;
  trLink: TTextRangeW;
  ptMouse: TPoint;
begin
  Result := false;
  if Assigned(FOnLinkClick) then begin
    with CharRange do SetString(sLink, nil, succ((cpMax - cpMin) shl 1));
    with trLink do begin
      chrg := CharRange;
      lpstrText := Pointer(sLink);
    end;
    if SendMessage(FChildHandle, EM_GETTEXTRANGE, 0, Longint(@trLink)) <> 0 then begin
      ptMouse := Point(Left + Width div 2, Top + GetSystemMetrics(SM_CYCURSOR) div 2);
      Hide;
    end
    else Exit;
    Result := true;
    sNewText := FOnLinkClick(WideCharToString(Pointer(sLink)));
    if sNewText <> E then begin
      with ptMouse do SetCursorPos(X, Y);
      SetRichText(sNewText);
      Show;
    end
    else SetRichText(E);
  end;
end;

procedure TToolTip.SetActivate(Value: Boolean);
begin
  SendMessage(WindowHandle, TTM_ACTIVATE, ord(Value = true), 0);
end;

function TToolTip.GetHintPause: Integer;
begin
  Result := SendMessage(WindowHandle, TTM_GETDELAYTIME, TTDT_INITIAL, 0);
end;

procedure TToolTip.SetHintPause(Value: Integer);
begin
  SendMessage(WindowHandle, TTM_SETDELAYTIME, TTDT_INITIAL, Value);
end;

function TToolTip.GetHintHidePause: Integer;
begin
  Result := SendMessage(WindowHandle, TTM_GETDELAYTIME, TTDT_AUTOPOP, 0);
end;

procedure TToolTip.SetHintHidePause(Value: Integer);
begin
  SendMessage(WindowHandle, TTM_SETDELAYTIME, TTDT_AUTOPOP, Value);
end;

function TToolTip.GetHintShortPause: Integer;
begin
  Result := SendMessage(WindowHandle, TTM_GETDELAYTIME, TTDT_RESHOW, 0);
end;

procedure TToolTip.SetHintShortPause(Value: Integer);
begin
  SendMessage(WindowHandle, TTM_SETDELAYTIME, TTDT_RESHOW, Value);
end;

procedure TToolTip.SetMaxWidth(Value: Integer);
begin
  if Value <> FMaxWidth then begin
    SendMessage(WindowHandle, TTM_SETMAXTIPWIDTH, 0, Value);
    FDrawStyle := FDrawStyle or (DT_WORDBREAK * ord(Value <> 0));
    FMaxWidth := Value;
  end;
end;

function TToolTip.GetParentFontHandle: THandle;
begin
  Result := SendMessage(FParentHandle, WM_GETFONT, 0, 0);
end;

function TToolTip.GetParentWidth: Integer;
begin
  Result := WindowUtils.GetClientRect(FParentHandle).Right;
end;

function TToolTip.GetShowing: Boolean;
begin
  Result := GetVisible;
end;

function TToolTip.GetTextRect: TRect;
var
  hDC, hFont: THandle;
  rText, rScreen: TRect;
begin
  SetRectEmpty(Result);
  if FUnicodeText <> E then begin
    hDC := GetDC(FChildHandle);
    hFont := 0;
    try
      if not FOversizeOnly then
	hFont := SelectObject(hDC, GetChildFont.Handle)
      else hFont := SelectObject(hDC, GetParentFontHandle);
      if FMaxWidth <> 0 then begin
	Result := WindowUtils.GetClientRect(FParentHandle);
	if FMaxWidth > 0 then Result.Right := FMaxWidth;
	SystemParametersInfo(SPI_GETWORKAREA, 0, @rScreen, 0);
	with rScreen do Result.Bottom := Bottom - Top;
      end;
      if (FChildTheme <> 0) and not FOversizeOnly then begin
	GetThemeTextExtent(FChildTheme, hDC, TTP_STANDARD, TTSS_NORMAL,
			   Pointer(FUnicodeText), -1, FDrawStyle, @Result, Result);
      end
      else DrawTextW(hDC, Pointer(FUnicodeText), -1, Result, FDrawStyle or DT_CALCRECT);
    finally
      if hFont <> 0 then SelectObject(hDC, hFont);
      ReleaseDC(FChildHandle, hDC);
    end;
  end
  else if FToolTipStyle = ttsRichText then begin
    hDC := GetDC(FChildHandle);
    hFont := 0;
    try
      hFont := SelectObject(hDC, GetChildFont.Handle);
      if FMaxWidth <> 0 then begin
	rText := WindowUtils.GetClientRect(FParentHandle);
	if FMaxWidth > 0 then rText.Right := FMaxWidth;
	Result.Right := rText.Right - FHMargin shl 1;
	SendMessage(FChildHandle, EM_SETTARGETDEVICE, hDC, MulDiv(Result.Right, TWIPS, Screen.PixelsPerInch));
	if FTextHeight = 0 then
	  Result.Bottom := FRichEditRect.Bottom + FVMargin
	else Result.Bottom := FTextHeight + FVMargin shl 2;
      end;
    finally
      if hFont <> 0 then SelectObject(hDC, hFont);
      ReleaseDC(FChildHandle, hDC);
    end;
  end;
end;

procedure TToolTip.GetTextRectDimensions(var Width, Height: Integer);
begin
  inc(Width, FHMargin shl 1);
  inc(Height, FVMargin shl 1);
end;

procedure TToolTip.GetWindowPos(var WindowPos: TWindowPos);
var
  rScreen, rText: TRect;
  iOversize: Integer;
begin
  with WindowPos do begin
    SystemParametersInfo(SPI_GETWORKAREA, 0, @rScreen, 0);
    rText := GetTextRect;
    if not IsAllRectEmpty(rText) then begin
      with rText do begin
	cx := Right;
	cy := Bottom;
      end;
      GetTextRectDimensions(cx, cy);
      with rScreen do begin
	iOversize := x + cx - Right;
	if iOversize > 0 then dec(x, iOversize);
	if x < Left then x := Left;
	iOversize := y + cy - Bottom;
	if iOversize > 0 then dec(y, iOversize);
	if y < Top then y := Top;
      end;
    end;
    if FFixedPos or (FCenterToolTip <> cttNot) then begin
      SetRectEmpty(rText);
      if FKeepPosition then begin
	rText.Left := Left;
	rText.Top := Top;
      end
      else begin
	case FCenterToolTip of
	  cttSubMouse: with rText do begin
	    TopLeft := GetCursorPos;
	    dec(Left, cx div 2);
	    dec(Top, succ(FVMargin));
	  end;
	  cttInParent: with rText do begin
	    rText := GetWindowRect(FParentHandle);
	    inc(Left, (Right - Left - cx) div 2);
	    inc(Top, (Bottom - Top - cy) div 2);
	  end;
	  cttSubParent: with rText do begin
	    rText := GetWindowRect(FParentHandle);
	    inc(Left, (Right - Left - cx) div 2);
	    Top := Bottom + 2;
	  end;
	  cttInToolRect: with rText do begin
	    rText := ToolRect;
	    Windows.ClientToScreen(FParentHandle, TopLeft);
	    dec(Left, (cx - Right) div 2);
	    dec(Top, (cy - Bottom) div 2);
	  end;
	  cttSubToolRect: with rText do begin
	    rText := ToolRect;
	    Windows.ClientToScreen(FParentHandle, TopLeft);
	    dec(Left, (cx - Right) div 2);
	    inc(Top, Bottom + 2);
	  end;
	else
	  if Windows.GetClientRect(FParentHandle, rText) then Windows.ClientToScreen(FParentHandle, rText.TopLeft);
	end;
      end;
      if Assigned(FOnGetToolTipOrigin) then
	FOnGetToolTipOrigin(rText.TopLeft)
      else if not SamePoints(FToolTipOrigin, OffScreenPos) then rText.TopLeft := FToolTipOrigin;
      with rText do begin
	iOversize := Left + cx - rScreen.Right;
	if iOversize > 0 then dec(Left, iOversize);
	iOversize := Top + cy - rScreen.Bottom;
	if iOversize > 0 then dec(Top, iOversize);
	if Left < 0 then Left := 0;
	if Top < 0 then Top := 0;
	x := Left;
	y := Top;
      end;
    end;
  end;
end;

procedure TToolTip.SetToolRect(const Value: TRect);
begin
  with FToolInfo do begin
    cbSize := SizeOf(TToolInfo);
    uFlags := TTF_PARSELINKS or TTF_SUBCLASS or TTF_TRANSPARENT;
    hwnd := FParentHandle;
    uId := 0;
    rect := Value;
  end;
  SendMessage(WindowHandle, TTM_NEWTOOLRECT, 0, Longint(@FToolInfo));
end;

procedure TToolTip.SetToolTipStyle(Value: TToolTipStyle);
const
  RICHED_STYLES = WS_CHILD or WS_VISIBLE or ES_MULTILINE or ES_READONLY;
  RICHED_EX_STYLES = WS_EX_TRANSPARENT;
var
  wcRtf: TWndClass;
begin
  FToolTipStyle := Value;
  if FChildHandle <> 0 then begin
    DestroyWindow(FChildHandle);
    FChildHandle := 0;
  end;
  if FToolTipStyle = ttsRichText then begin
    if GetClassInfo(GetModuleHandle(LIB_MSFTEDIT), RICHEDIT50W, wcRtf) then begin
      FChildHandle := CreateWindowEx(RICHED_EX_STYLES, RICHEDIT50W, nil, RICHED_STYLES,
				     0, 0, 0, 0, WindowHandle, 0, HInstance, nil);
    end
    else if GetClassInfo(GetModuleHandle(LIB_RICHED20), RICHEDIT20W, wcRtf) then begin
      FChildHandle := CreateWindowEx(RICHED_EX_STYLES, RICHEDIT20W, nil, RICHED_STYLES,
				     0, 0, 0, 0, WindowHandle, 0, HInstance, nil);
    end;
    if FChildHandle <> 0 then begin
      if FMaxWidth = 0 then SetMaxWidth(GetParentWidth);
      SendMessage(FChildHandle, EM_EXLIMITTEXT, 0, MAXINT);
      SendMessage(FChildHandle, EM_SETEVENTMASK, 0, ENM_CHANGE or ENM_KEYEVENTS or ENM_LINK or ENM_MOUSEEVENTS or
						    ENM_REQUESTRESIZE or ENM_SCROLL or ENM_SELCHANGE or ENM_UPDATE);
      SendMessage(FChildHandle, EM_SETBKGNDCOLOR, 0, ColorToRGB(Brush.Color));
      SendMessage(FChildHandle, EM_HIDESELECTION, 0, 0);
      SendMessage(FChildHandle, EM_SETREADONLY, 1, 0);
      FHMargin := 4;
      FVMargin := 4;
    end;
  end
  else begin
    FChildHandle := CreateWindow('STATIC', nil, WS_CHILD or WS_VISIBLE, 0, 0, 0, 0, WindowHandle, 0, HInstance, nil);
    FDrawStyle := DT_CENTER or DT_EXTERNALLEADING or DT_HIDEPREFIX or DT_VCENTER or (DT_WORDBREAK * ord(FMaxWidth <> 0));
    if FToolTipStyle = ttsUnderscore then FDrawStyle := FDrawStyle and not DT_HIDEPREFIX;
    FHMargin := 4;
    FVMargin := 2;
  end;
  if FChildHandle <> 0 then begin
    SetProp(FChildHandle, TOOLTIPS_CLASS, THandle(Self));
    if EnableThemes(FChildHandle, true) then FChildTheme := OpenChildTheme;
    FChildProc := Ptr(SetWindowLong(FChildHandle, GWL_WNDPROC, Longint(@ChildProc)));
    if not ParentFont then FChildFont.Name := ShellDlgFont(FChildTheme, TMT_MSGBOXFONT);
  end;
end;

procedure TToolTip.SetToolTipText(const Value: String);
begin
  MoveWindow(FChildHandle, 0, 0, 0, 0, false);
  if Value <> FToolTipText then begin
    FToolTipText := Value;
    if Value <> E then begin
      FToolInfo.lpszText := SP;
      if FCenterToolTip = cttNot then
	AddToolTip
      else if Showing then begin
	FKeepPosition := true;
	try
	  SendMessage(WindowHandle, TTM_UPDATETIPTEXT, 0, Longint(@FToolInfo));
	finally
	  FKeepPosition := false;
	end;
      end;
      Exit;
    end;
  end;
  if Value = E then begin
    FToolInfo.lpszText := nil;
    FToolTipText := E;
    DelToolTip;
  end;
end;

function TToolTip.GetVisible: Boolean;
begin
  Result := IsWindowVisible(WindowHandle);
end;

procedure TToolTip.SetVisible(Value: Boolean);
begin
  SetWindowLong(WindowHandle, GWL_STYLE, ((GetWindowLong(WindowHandle, GWL_STYLE)
					   and not WS_VISIBLE) or (ord(Value) * WS_VISIBLE)));
end;

procedure TToolTip.SetAnsiText(const Value: String);
begin
  if FToolTipStyle = ttsRichText then begin
    SetRichText(Value);
    Exit;
  end
  else if FOversizeOnly then
    FUnicodeText := AnsiToWideString(OversizedText(Value))
  else FUnicodeText := AnsiToWideString(Value);
  SetToolTipText(FUnicodeText);
end;

procedure TToolTip.SetUtf8Text(const Value: String);
begin
  if FToolTipStyle = ttsRichText then begin
    SetRichText(Utf8StringToAnsi(Value));
    Exit;
  end
  else if FOversizeOnly then
    FUnicodeText := Utf8ToWideString(OversizedText(Value))
  else FUnicodeText := Utf8ToWideString(Value);
  SetToolTipText(FUnicodeText);
end;

procedure TToolTip.SetRichText(const Value: String);
var
  steToolTip: TSetTextEx;
  rMargin: TRect;
begin
  with steToolTip do begin
    flags := ST_DEFAULT;
    codepage := CP_ACP;
  end;
  SetToolTipText(Value);
  rMargin := Rect(FHMargin, FVMargin, FHMargin, FVMargin);
  SendMessage(FChildHandle, EM_SETRECTNP, 1, Longint(@rMargin));
  if Value <> E then
    LoadFromRtfStream(Value)
  else SendMessage(FChildHandle, EM_SETTEXTEX, Longint(@steToolTip), Longint(Value));
end;

procedure TToolTip.LoadFromRtfStream(const RichText: String);
var
  ss: TStringStream;
  esRtf: TEditStream;

  function EditStreamLoad(dwCookie: DWord; pbBuff: PByte; cb: Longint; var pcb: Longint): DWord; stdcall;
  begin
    Result := 0;
    try
      pcb := TStringStream(dwCookie).Read(pbBuff^, cb);
    except
      Result := DWord(-1);
    end;
  end;

begin
  ss := TStringStream.Create(RichText);
  try
    with esRtf do begin
      dwCookie := DWord(ss);
      dwError := 0;
      pfnCallback := @EditStreamLoad;
    end;
    ss.Seek(0, soFromBeginning);
    SendMessage(FChildHandle, EM_STREAMIN, SF_RTF, Longint(@esRtf));
  finally
    ss.Free;
  end;
end;

function TToolTip.ChildPaint(ChildHandle: THandle): Integer;
var
  hFont, hBrush: THandle;
  ps: TPaintStruct;
begin
  Result := -1;
  if FToolTipStyle <> ttsRichText then begin
    if BeginPaint(ChildHandle, ps) <> 0 then with ps do begin
      if not FOversizeOnly then
	hFont := SelectObject(hDC, GetChildFont.Handle)
      else hFont := SelectObject(hDC, GetParentFontHandle);
      hBrush := SelectObject(hDC, Brush.Handle);
      try
	if (FChildTheme <> 0) and not FOversizeOnly then begin
	  DrawThemeParentBackground(ChildHandle, hDC, nil);
	  DrawThemeEdge(FChildTheme, hDC, TTP_STANDARD, TTSS_NORMAL, rcPaint, 0, BF_MONO or BF_RECT, @rcPaint);
	  inc(rcPaint.Top, GetSystemMetrics(SM_CYBORDER));
	  DrawThemeText(FChildTheme, hDC, TTP_STANDARD, TTSS_NORMAL, Pointer(FUnicodeText), -1, FDrawStyle, 0, rcPaint);
	end
	else begin
	  SetBkMode(hDC, TRANSPARENT);
	  DrawEdge(hDC, rcPaint, 0, BF_MONO or BF_RECT);
	  inc(rcPaint.Top, GetSystemMetrics(SM_CYBORDER));
	  DrawTextW(hDC, Pointer(FUnicodeText), -1, rcPaint, FDrawStyle);
	end;
      finally
	if hBrush <> 0 then SelectObject(hDC, hBrush);
	if hFont <> 0 then SelectObject(hDC, hFont);
	EndPaint(ChildHandle, ps);
      end;
    end;
  end;
end;

function TToolTip.ChildEraseBkGnd(ChildHandle, ChildDC: THandle): Integer;
begin
  Result := 0;
  if FToolTipStyle = ttsRichText then begin
    if FChildTheme <> 0 then
      DrawThemeParentBackground(ChildHandle, ChildDC, nil)
    else SetBkMode(ChildDC, TRANSPARENT);
    Result := -1;
  end;
end;

function TToolTip.Reactivate(ToolTip: THandle): Boolean;
begin
  SendMessage(ToolTip, TTM_ACTIVATE, 0, 0);
  SendMessage(ToolTip, TTM_ACTIVATE, 1, 0);
  Result := true;
end;

procedure TToolTip.ShowHandler;
begin
  if not Showing then begin
    SetAppOnMessage(true);
    if Assigned(FOnShowTooltip) then FOnShowTooltip(Self);
    if FToolTipStyle <> ttsRichText then begin
      SetCapture(FParentHandle);
      FFocusHandle := 0;
    end
    else begin
      FFocusHandle := GetFocus;
      Windows.SetFocus(FChildHandle);
    end;
  end;
end;

procedure TToolTip.PopHandler;
var
  wc: TWinControl;
  wParam, lParam: Longint;
begin
  SetAppOnMessage(false);
  ShowWindow(WindowHandle, SW_HIDE);
  if GetCapture <> 0 then ReleaseCapture;
  if FToolTipStyle = ttsRichText then begin
    wc := FindControl(FFocusHandle);
    if wc <> nil then begin
      if wc is TCustomEdit then begin
	SendMessage(wc.Handle, EM_GETSEL, Longint(@wParam), Longint(@lParam));
	try
	  wc.SetFocus;
	finally
	  SendMessage(wc.Handle, EM_SETSEL, wParam, lParam);
	end;
      end
      else wc.SetFocus;
    end
    else Windows.SetFocus(FFocusHandle);
  end;
  if Assigned(FOnHideTooltip) then FOnHideTooltip(Self);
end;

procedure TToolTip.WMParentNotify(var Message: TWMParentNotify);
begin
  case Message.Event of
    WM_CREATE: begin
    end;
    WM_DESTROY: begin
      SetWindowLong(FChildHandle, GWL_WNDPROC, Longint(FChildProc));
      RemoveProp(FChildHandle, TOOLTIPS_CLASS);
      CloseChildTheme;
    end;
  end;
end;

procedure TToolTip.SetAppOnMessage(Activate: Boolean);
begin
  with Application do begin
    if Activate then begin
      FAppOnMessage := OnMessage;
      OnMessage := AppOnMessage;
    end
    else begin
      OnMessage := FAppOnMessage;
      FAppOnMessage := nil;
    end;
  end;
end;

procedure TToolTip.CMColorChanged(var Message: TMessage);
begin
  inherited;
  if WindowHandle = 0 then Exit;
  SendMessage(WindowHandle, TTM_SETTIPBKCOLOR, ColorToRGB(Brush.Color), 0);
end;

procedure TToolTip.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if WindowHandle = 0 then Exit;
  SendMessage(WindowHandle, WM_SETFONT, Font.Handle, 0);
  SendMessage(WindowHandle, TTM_SETTIPTEXTCOLOR, ColorToRGB(Font.Color), 0);
end;

procedure TToolTip.CMParentColorChanged(var Message: TMessage);
var
  hParentDC: THandle;
begin
  if not ParentFont or (FParentHandle = 0) then Exit;
  hParentDC := GetDC(FParentHandle);
  try
    Brush.Color := GetBkColor(hParentDC);
  finally
    ReleaseDC(FParentHandle, hParentDC);
  end;
  Perform(CM_COLORCHANGED, 0, 0);
end;

procedure TToolTip.CMParentFontChanged(var Message: TMessage);
var
  hParentDC: THandle;
begin
  if not ParentFont or (FParentHandle = 0) then Exit;
  Font.Handle := SendMessage(FParentHandle, WM_GETFONT, 0, 0);
  hParentDC := GetDC(FParentHandle);
  try
    Font.Color := GetTextColor(hParentDC);
  finally
    ReleaseDC(FParentHandle, hParentDC);
  end;
  Perform(CM_FONTCHANGED, 0, 0);
end;

procedure TToolTip.CNNotify(var Message: TWMNotify);
begin
  case Message.NMHdr^.code of
    TTN_SHOW: ShowHandler;
    TTN_POP: PopHandler;
  end;
end;

procedure TToolTip.WMCreate;
var
  rParent: TRect;
begin
  if Name = E then Name := ClassString(FParentHandle) + ClassString(WindowHandle);
  SendMessage(WindowHandle, TTM_SETWINDOWTHEME, 0, Longint(@WC_TOOLTIP));
  SendMessage(WindowHandle, TTM_SETMAXTIPWIDTH, 0, MAXINT);
  SetWindowLong(WindowHandle, GWL_USERDATA, Longint(Self));
  rParent := WindowUtils.GetClientRect(FParentHandle);
  with Application do begin
    HintPause := 500;
    HintHidePause := 2500;
    HintShortPause := 250;
  end;
  Color := clInfoBk;
  Brush.Color := clInfoBk;
  SetToolRect(rParent);
  SetToolTipStyle(FToolTipStyle);
end;

procedure TToolTip.WMDestroy;
begin
  FChildHandle := 0;
end;

procedure TToolTip.WMPaint(var Message: TWMPaint);
var
  ps: TPaintStruct;
begin
  if BeginPaint(WindowHandle, ps) <> 0 then with WindowUtils.GetClientRect(WindowHandle) do begin
    try
      MoveWindow(FChildHandle, 0, 0, Right, Bottom, TRUE);
    finally
      EndPaint(WindowHandle, ps);
    end;
  end;
end;

procedure TToolTip.WMShowWindow(var Message: TWMShowWindow);
begin
  if not Message.Show then FToolTipText := E;
end;

procedure TToolTip.WMMouseActivate(var Message: TWMMouseActivate);
begin
  with Message do if MouseMsg = WM_MOUSEMOVE then Result := MA_NOACTIVATEANDEAT;
end;

procedure TToolTip.WMSetFont(var Message: TWMSetFont);
var
  cf: TCharFormat;
begin
  if FChildHandle <> 0 then begin
    if ToolTipStyle = ttsRichText then begin
      cf := GetCharFormatFromFont(Font, E, 0, CFM_FONT);
      SendMessage(FChildHandle, EM_SETCHARFORMAT, SCF_DEFAULT, Longint(@cf));
    end
    else SendMessage(FChildHandle, WM_SETFONT, Font.Handle, LoWord(-1));
  end;
end;

procedure TToolTip.WMWindowPosChanging(var Message: TWMWindowPosChanging);
begin
  GetWindowPos(Message.WindowPos^);
end;

procedure TToolTip.WMWindowPosChanged(var Message: TWMWindowPosChanged);
var
  iTopOffset: Integer;
begin
  with Message.WindowPos^ do if (flags and (SWP_HIDEWINDOW or SWP_NOREDRAW)) = 0 then begin
    if FToolTipStyle = ttsRichText then begin
      iTopOffset := Y + GetSystemMetrics(SM_CYCURSOR) div 2;
      if not PtInRect(Rect(X, iTopOffset, X + CX, Y + CY), GetCursorPos) then begin
	SetCursorPos(X + CX div 2, iTopOffset);
	SetCursor(crDefault);
      end;
    end;
    FCursorPosOnShow := GetCursorPos;
  end
  else if (flags and SWP_HIDEWINDOW) <> 0 then FCursorPosOnShow := OffScreenPos;
end;

procedure TToolTip.WMNotify(var Message: TWMNotify);
begin
  with Message do begin
    Result := 0;
    if NMHdr^.hwndFrom = FChildHandle then begin
      HideCaret(FChildHandle);
      case NMHdr^.code of
	EN_MSGFILTER: with PMsgFilter(NMHdr)^ do if msg = WM_KEYDOWN then Result := ord(RichEditKeyDown(Msg, wParam, lParam));
	EN_REQUESTRESIZE: begin
	  if FTextHeight = 0 then begin
	    FRichEditRect := PReqSize(NMHdr)^.rc;
	    FRichEditRect.Bottom := GetRtfTextHeight(FChildHandle, FRichEditVersion,
						     GetValueFromCharFormat(DefaultCharFormat, CFM_SIZE),
						     FVMargin shl 1);
	  end;
	end;
	EN_LINK: with PENLink(NMHdr)^ do if msg = WM_LBUTTONUP then Result := ord(RichEditLinkClicked(chrg));
      end;
    end;
  end;
end;

procedure TToolTip.WMCommand(var Message: TWMCommand);
begin
  with Message do begin
    Result := 0;
    if (Ctl = FChildHandle) and ((NotifyCode = EN_CHANGE) or (NotifyCode = EN_UPDATE)) and (FToolTipText <> E) then begin
      SendMessage(FChildHandle, EM_REQUESTRESIZE, 0, 0);
    end;
  end;
end;

procedure TToolTip.WMCtlColorStatic(var Message: TWMCtlColorStatic);
begin
  with Message do Result := ChildColors(ChildDC);
end;

procedure TToolTip.AppOnMessage(var Msg: TMsg; var Handled: Boolean);

  function EnumThreadWindowsProc(hWnd: HWND; var Handled: Boolean): BOOL; stdcall;
  var
    ctlFound: TControl;
  begin
    Result := true;
    ctlFound := FindControl(hWnd);
    if not (ctlFound is TToolTip) then Exit;
    with TToolTip(ctlFound) do if Showing then begin
      Handled := Reactivate(WindowHandle);
      Result := false;
    end;
  end;

begin
  case Msg.Message of
    WM_KEYDOWN: with Msg do if (hWnd = FChildHandle) or (hWnd = WindowHandle) then begin
      if wParam <> VK_ESCAPE then Handled := true;
      case wParam of
	VK_TAB, VK_RETURN: Handled := (FToolTipStyle = ttsRichText) and RichEditKeyDown(Message, wParam, lParam);
	VK_ESCAPE: EnumThreadWindows(GetCurrentThreadID, @EnumThreadWindowsProc, Longint(@Handled));
      else
	if (FToolTipStyle = ttsRichText) and (GetKeyState(VK_SHIFT) or GetKeyState(VK_CONTROL) < 0) then begin
	  case chr(wParam) of
	    'A': SendMessage(FChildHandle, EM_SETSEL, 0, -1);
	    'C': SendMessage(FChildHandle, WM_COPY, 0, 0);
	  else
	    if wParam > ord(SP) then Beep;
	  end;
	end
	else Handled := false;
      end;
      if not Handled and Showing then Handled := Reactivate(WindowHandle);
    end;
  end;
end;

procedure TToolTip.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  FParentHandle := Parent.Handle;
  with Params do begin
    Style := TTS_DEFAULT and not (WS_CLIPCHILDREN or WS_CLIPSIBLINGS);
    ExStyle := WS_EX_TOOLWINDOW;
    X := Integer(CW_USEDEFAULT);
    Y := Integer(CW_USEDEFAULT);
    Width := Integer(CW_USEDEFAULT);
    Height := Integer(CW_USEDEFAULT);
    WndParent := FParentHandle;
  end;
  CreateSubClass(Params, TOOLTIPS_CLASS);
end;

procedure TToolTip.WndProc(var Message: TMessage);
begin
  with Message do begin
    Result := 0;
    case Msg of
      WM_CREATE: WMCreate;
      WM_DESTROY: WMDestroy;
      WM_PAINT: begin
	WMPaint(TWMPaint(Message));
	if not FOversizeOnly then Exit;
      end;
      WM_SHOWWINDOW: WMShowWindow(TWMShowWindow(Message));
      WM_MOUSEACTIVATE: WMMouseActivate(TWMMouseActivate(Message));
      WM_SETFONT: WMSetFont(TWMSetFont(Message));
      WM_WINDOWPOSCHANGING: WMWindowPosChanging(TWMWindowPosChanging(Message));
      WM_WINDOWPOSCHANGED: WMWindowPosChanged(TWMWindowPosChanged(Message));
      WM_NOTIFY: WMNotify(TWMNotify(Message));
      WM_COMMAND: WMCommand(TWMCommand(Message));
      WM_CTLCOLORSTATIC: WMCtlColorStatic(TWMCtlColorStatic(Message));
      WM_PARENTNOTIFY: WMParentNotify(TWMParentNotify(Message));
    end;
    inherited;
  end;
end;

function TToolTip.AddToolTip: Boolean;
begin
  Result := false;
  if SendMessage(WindowHandle, TTM_GETTOOLCOUNT, 0, 0) = 0 then begin
    if ParentFont then FParentFont.Handle := GetParentFontHandle;
    with Application do begin
      SetHintPause(HintPause);
      SetHintHidePause(HintHidePause);
      SetHintShortPause(HintShortPause);
    end;
    SendMessage(WindowHandle, TTM_ADDTOOL, 0, Longint(@FToolInfo));
    SendMessage(WindowHandle, TTM_ACTIVATE, 1, 0);
    if FCenterToolTip <> cttNot then SendMessage(WindowHandle, TTM_TRACKACTIVATE, 1, Longint(@FToolInfo));
    Result := true;
  end;
end;

function TToolTip.DelToolTip: Boolean;
begin
  Result := false;
  if SendMessage(WindowHandle, TTM_GETTOOLCOUNT, 0, 0) <> 0 then begin
    if FCenterToolTip <> cttNot then SendMessage(WindowHandle, TTM_TRACKACTIVATE, 0, Longint(@FToolInfo));
    SendMessage(WindowHandle, TTM_ACTIVATE, 0, 0);
    SendMessage(WindowHandle, TTM_DELTOOL, 0, Longint(@FToolInfo));
    Result := true;
  end;
end;

constructor TToolTip.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited Visible := false;
  SetParentComponent(AOwner);
  FChildFont := TFont.Create;
  FChildFont.Assign(Self.Font);
  FParentFont := TFont.Create;
  FParentFont.Assign(Self.Font);
  FToolTipOrigin := OffScreenPos;
  FRichEditVersion := GetRichEditVersion;
end;

destructor TToolTip.Destroy;
begin
  if FParentFont <> nil then FParentFont.Free;
  if FChildFont <> nil then FChildFont.Free;
  inherited Destroy;
end;

function TToolTip.HasText: Boolean;
begin
  with FToolInfo do Result := (lpszText <> nil);
end;

procedure TToolTip.Show;
begin
  DelToolTip;
  AddToolTip;
end;

procedure TToolTip.Hide;
begin
  DelToolTip;
end;

procedure TToolTip.ResetHintPauseToDefault;
begin
  SendMessage(WindowHandle, TTM_SETDELAYTIME, TTDT_AUTOMATIC, -1);
end;

procedure TToolTip.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if AParent <> nil then CreateHandle;
end;

begin
  LoadLibrary(LIB_MSFTEDIT);
end.

