{$J+,Z4}
unit TreeViewEx;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2005-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, CommCtrl, SysUtils, Classes, Graphics, Forms, Controls, ComCtrls, ToolTip, UxThemes;

// coming with ComCtl32.dll v4.70 or later
const
  NM_CUSTOMDRAW	= (NM_FIRST - 12);

  // custom draw return flags
  CDRF_DODEFAULT		= $00000000;
  CDRF_NEWFONT			= $00000002;
  CDRF_SKIPDEFAULT		= $00000004;

  CDRF_NOTIFYPOSTPAINT		= $00000010;
  CDRF_NOTIFYITEMDRAW		= $00000020;
  CDRF_NOTIFYPOSTERASE		= $00000040;

  // custom draw drawstage flags
  CDDS_PREPAINT			= $00000001;
  CDDS_POSTPAINT		= $00000002;
  CDDS_PREERASE			= $00000003;
  CDDS_POSTERASE		= $00000004;

  CDDS_ITEM			= $00010000;
  CDDS_ITEMPREPAINT		= (CDDS_ITEM or CDDS_PREPAINT);
  CDDS_ITEMPOSTPAINT		= (CDDS_ITEM or CDDS_POSTPAINT);
  CDDS_ITEMPREERASE		= (CDDS_ITEM or CDDS_PREERASE);
  CDDS_ITEMPOSTERASE		= (CDDS_ITEM or CDDS_POSTERASE);

  // custom draw itemstate flags
  CDIS_SELECTED			= $0001;
  CDIS_GRAYED			= $0002;
  CDIS_DISABLED			= $0004;
  CDIS_CHECKED			= $0008;
  CDIS_FOCUS			= $0010;
  CDIS_DEFAULT			= $0020;
  CDIS_HOT			= $0040;
  CDIS_MARKED			= $0080;
  CDIS_INDETERMINATE		= $0100;

// coming with ComCtl32.dll v4.70 or later
type
  PNMCustomDraw = ^TNMCustomDraw;
  TNMCustomDraw = packed record
    hdr: TNMHdr;
    dwDrawStage: DWORD;
    hdc: HDC;
    rc: TRect;
    dwItemSpec: DWORD;
    uItemState: UINT;
    lItemlParam: LPARAM;
  end;
  PNMTVCustomDraw = ^TNMTVCustomDraw;
  TNMTVCustomDraw = packed record
    nmcd: TNMCustomDraw;
    clrText: COLORREF;
    clrTextBk: COLORREF;
  end;

// coming with ComCtl32.dll v4.71 or later
const
  TVM_SETITEMHEIGHT	= (TV_FIRST + 27);
  TVM_GETITEMHEIGHT	= (TV_FIRST + 28);
  TVM_SETBKCOLOR	= (TV_FIRST + 29);
  TVM_SETTEXTCOLOR	= (TV_FIRST + 30);
  TVM_GETBKCOLOR	= (TV_FIRST + 31);
  TVM_GETTEXTCOLOR	= (TV_FIRST + 32);

// coming with ComCtl32.dll v4.71 or later
const
  TVS_INFOTIP		= $0800;

type
  TOnEnumSelectedNodes = Procedure(Node: TTreeNode; var Cancel: Boolean) of Object;
  TOnMultiSelecting = Procedure(Node: TTreeNode; SelectAllParam: Longbool; var Cancel: Boolean) of Object;
  TRangeItem = Record
    Node: TTreeNode;
    Index: Integer;
  end;
  TSelRange = Record
    FirstItem: TRangeItem;
    LastItem: TRangeItem;
  end;
  TTreeViewEx = class(TTreeView)
  private
    FColorBrush: THandle;
    {$IFNDEF CONDITIONALEXPRESSIONS}
    FMultiSelect: Longbool;
    {$ENDIF}
    FToolTip: TToolTip;
    FToolTips: Longbool;
    FSelecting: Longbool;
    FSelectAllParam: Longbool;
    FShowHint: Longbool;
    FOnEnumSelectedNodes: TOnEnumSelectedNodes;
    FOnMultiSelecting: TOnMultiSelecting;
    function  GetItemHeight: Integer;
    procedure SetItemHeight(Value: Integer);
    function  GetMultiSelect: Longbool;
    procedure SetMultiSelect(Value: Longbool);
    procedure SetToolTips(Value: Longbool);
    function  GetThemed: Longbool;
    procedure CreateColorBrush;
    procedure FreeColorBrush;
    procedure ThemeHandler;
    procedure Changed;
    function  Changing: Longbool;
    function  GetNodeAtCursorPos: TTreeNode;
    function  GetSelectedNode: TTreeNode;
    function  GetSelectionCount: Integer;
    function  GetSelectionRange: TSelRange;
    procedure SelectAll;
    procedure UnselectAll(Focus: TTreeNode);
    function  GetHint: String;
    procedure SetHint(const Value: String);
    procedure MoveFocus(Direction: Integer);
    procedure ShiftSelect(NewTarget: TTreeNode);
    procedure ControlSelect(NewTarget: TTreeNode);
    procedure SelectRange(First, Last: TTreeNode);
    procedure UnselectRange(First, Last: TTreeNode);
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMKeyDown(var Message: TWMKey); message WM_KEYDOWN;
    procedure WMLButtonDown(var Message: TWMMouse); message WM_LBUTTONDOWN;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure EnumSelectedNodes;
    procedure SelectAllItems(const SelectAllParam: Longbool);
    procedure UnselectAllItems(const SelectAllParam: Longbool);
    property NodeAtCursorPos: TTreeNode read GetNodeAtCursorPos;
    property SelectedNode: TTreeNode read GetSelectedNode;
    property SelectionCount: Integer read GetSelectionCount;
    property Themed: Longbool read GetThemed;
    property ToolTip: TToolTip read FToolTip;
  published
    property Hint: String read GetHint write SetHint;
    property ItemHeight: Integer read GetItemHeight write SetItemHeight;
    property MultiSelect: Longbool read GetMultiSelect write SetMultiSelect;
    property ShowHint: Longbool read FShowHint write FShowHint;
    property ToolTips: Longbool read FToolTips write SetToolTips;
    property OnEnumSelectedNodes: TOnEnumSelectedNodes read FOnEnumSelectedNodes write FOnEnumSelectedNodes;
    property OnMultiSelecting: TOnMultiSelecting read FOnMultiSelecting write FOnMultiSelecting;
  end;

procedure Register;


implementation

function ShiftState: Integer;
begin
  Result := (ord(GetKeyState(VK_SHIFT) < 0) * MK_SHIFT) + (ord(GetKeyState(VK_CONTROL) < 0) * MK_CONTROL);
end;

function TTreeViewEx.GetItemHeight: Integer;
begin
  Result := SendMessage(WindowHandle, TVM_GETITEMHEIGHT, 0, 0);
end;

procedure TTreeViewEx.SetItemHeight(Value: Integer);
begin
  SendMessage(WindowHandle, TVM_SETITEMHEIGHT, Value, 0);
end;

function TTreeViewEx.GetMultiSelect: Longbool;
begin
  {$IFDEF CONDITIONALEXPRESSIONS}
  Result := inherited MultiSelect;
  {$ELSE}
  Result := FMultiSelect;
  {$ENDIF}
end;

procedure TTreeViewEx.SetMultiSelect(Value: Longbool);
begin
  {$IFDEF CONDITIONALEXPRESSIONS}
  inherited MultiSelect := Value;
  {$ELSE}
  FMultiSelect := Value;
  {$ENDIF}
end;

procedure TTreeViewEx.SetToolTips(Value: Longbool);
var
  iStyle: Integer;
begin
  if FToolTips <> Value then begin
    FToolTips := Value;
    iStyle := GetWindowLong(WindowHandle, GWL_STYLE);
    if FToolTips then
      iStyle := iStyle or TVS_INFOTIP
    else iStyle := iStyle and not TVS_INFOTIP;
    SetWindowLong(WindowHandle, GWL_STYLE, iStyle);
  end;
end;

function TTreeViewEx.GetThemed: Longbool;
begin
  Result := UseThemes;
end;

procedure TTreeViewEx.CreateColorBrush;
begin
  FreeColorBrush;
  FColorBrush := CreateSolidBrush(ColorToRGB(Color));
end;

procedure TTreeViewEx.FreeColorBrush;
begin
  if FColorBrush <> 0 then begin
    DeleteObject(FColorBrush);
    FColorBrush := 0;
  end;
end;

procedure TTreeViewEx.ThemeHandler;
var
  iBorderStyle: Integer;
begin
  iBorderStyle := DWord(EnableThemes(WindowHandle, (Color = clWindow)) = true) * WS_BORDER;
  SetWindowLong(WindowHandle, GWL_STYLE, GetWindowLong(WindowHandle, GWL_STYLE) and not iBorderStyle);
  SetWindowPos(WindowHandle, 0, 0, 0, 0, 0, SWP_FRAMECHANGED or SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TTreeViewEx.Changed;
begin
  FSelecting := false;
  Items.EndUpdate;
  Change(nil);
end;

function TTreeViewEx.Changing: Longbool;
begin
  Result := CanChange(nil);
  if Result then begin
    Items.BeginUpdate;
    FSelecting := true;
  end;
end;

function TTreeViewEx.GetNodeAtCursorPos: TTreeNode;
var
  ptMouse: TPoint;
begin
  if GetCursorPos(ptMouse) then with Self.ScreenToClient(ptMouse) do
    Result := GetNodeAt(X, Y)
  else Result := nil;
end;

function TTreeViewEx.GetSelectedNode: TTreeNode;
begin
  if Items <> nil then
    Result := Items.GetNode(TreeView_GetSelection(WindowHandle))
  else Result := nil;
end;

function TTreeViewEx.GetSelectionCount: Integer;
var
  tnNext: TTreeNode;
begin
  Result := 0;
  if Items <> nil then begin
    tnNext := Items.GetFirstNode;
    while tnNext <> nil do begin
      if tnNext.Selected then inc(Result);
      tnNext := tnNext.GetNext;
    end;
  end;
end;

function GetRangeItem(ANode: TTreeNode; AIndex: Integer): TRangeItem;
begin
  with Result do begin
    Node := ANode;
    Index := AIndex;
  end;
end;

function TTreeViewEx.GetSelectionRange: TSelRange;

  function InitSelection: TSelRange;
  var
    iIndex: Integer;
  begin
    with Result do begin
      if (Items <> nil) and (Items.Count > 0) then begin
	iIndex := Selected.AbsoluteIndex;
	FirstItem.Node := Selected;
	FirstItem.Index := iIndex;
	LastItem.Node := Selected;
	LastItem.Index := iIndex;
      end
      else begin
	FirstItem.Node := nil;
	FirstItem.Index := -1;
	LastItem.Node := nil;
	LastItem.Index := -1;
      end;
    end;
  end;

var
  tnNext, tnPrev: TTreeNode;
begin
  Result := InitSelection;
  if Items <> nil then begin
    tnNext := Items[0];
    while tnNext <> nil do begin
      if tnNext.Selected then begin
	Result.FirstItem := GetRangeItem(tnNext, tnNext.AbsoluteIndex);
	Break;
      end;
      tnNext := tnNext.GetNext;
    end;
    tnPrev := Items[pred(Items.Count)];
    while tnPrev <> nil do begin
      if tnPrev.Selected then begin
	Result.LastItem := GetRangeItem(tnPrev, tnPrev.AbsoluteIndex);
	Break;
      end;
      tnPrev := tnPrev.GetPrev;
    end;
  end;
end;

function InitItemData(Item: HTreeItem; Selected: Cardinal): TTVItem;
begin
  with Result do begin
    Mask := TVIF_HANDLE or TVIF_STATE;
    StateMask := TVIS_SELECTED;
    State := Selected;
    hItem := Item;
  end;
end;

procedure TTreeViewEx.SelectAll;
begin
  if Changing then begin
    try
      SelectRange(Items.GetFirstNode, nil);
    finally
      Changed;
    end;
  end;
end;

procedure TTreeViewEx.UnselectAll(Focus: TTreeNode);
begin
  if Changing then begin
    try
      UnselectRange(Items.GetFirstNode, nil);
    finally
      if Focus <> nil then begin
	TreeView_SelectItem(WindowHandle, nil);
	FSelecting := false;
	TreeView_SelectItem(WindowHandle, Focus.ItemId);
	Items.EndUpdate;
      end
      else Changed;
    end;
  end;
end;

function TTreeViewEx.GetHint: String;
begin
  Result := inherited Hint;
end;

procedure TTreeViewEx.SetHint(const Value: String);
begin
  inherited Hint := Value;
  if FToolTip <> nil then FToolTip.AnsiText := Value;
end;

procedure TTreeViewEx.MoveFocus(Direction: Integer);
var
  tvOrigin, tvTarget: TTVItem;
begin
  if Selected <> nil then begin
    tvOrigin := InitItemData(Selected.ItemId, TVIS_SELECTED);
    case Direction of
      TVGN_NEXT: begin
	if Selected.AbsoluteIndex = pred(Items.Count) then Exit;
	tvTarget := InitItemData(Selected.GetNext.ItemId, TVIS_SELECTED);
      end;
      TVGN_PREVIOUS: begin
	if Selected.AbsoluteIndex = 0 then Exit;
	tvTarget := InitItemData(Selected.GetPrev.ItemId, TVIS_SELECTED);
      end;
    else
      Exit;
    end;
    if Changing then begin
      try
	TreeView_GetItem(WindowHandle, tvOrigin);
	TreeView_GetItem(WindowHandle, tvTarget);
	TreeView_EnsureVisible(WindowHandle, tvTarget.hItem);
	TreeView_SelectItem(WindowHandle, tvTarget.hItem);
	TreeView_SetItem(WindowHandle, tvTarget);
	TreeView_SetItem(WindowHandle, tvOrigin);
      finally
	Changed;
      end;
    end;
  end;
end;

procedure TTreeViewEx.ShiftSelect(NewTarget: TTreeNode);
var
  iIndex: Integer;
  srItems: TSelRange;
begin
  if NewTarget <> nil then begin
    if Changing then begin
      try
	iIndex := NewTarget.AbsoluteIndex;
	srItems := GetSelectionRange;
	with srItems do begin
	  if iIndex < FirstItem.Index then
	    FirstItem := GetRangeItem(NewTarget, iIndex)
	  else if iIndex > LastItem.Index then
	    LastItem := GetRangeItem(NewTarget, iIndex)
	  else begin
	    if iIndex > Selected.AbsoluteIndex then
	      FirstItem := GetRangeItem(NewTarget, iIndex)
	    else LastItem := GetRangeItem(NewTarget, iIndex);
	  end;
	  if FirstItem.Index > 0 then UnselectRange(Items[0], FirstItem.Node.GetPrev);
	  if LastItem.Index < pred(Items.Count) then UnselectRange(LastItem.Node.GetNext, Items[pred(Items.Count)]);
	  TreeView_EnsureVisible(WindowHandle, NewTarget.ItemId);
	  TreeView_SelectItem(WindowHandle, NewTarget.ItemId);
	  SelectRange(FirstItem.Node, LastItem.Node);
	end;
      finally
	Changed;
      end;
    end;
  end;
end;

procedure TTreeViewEx.ControlSelect(NewTarget: TTreeNode);
var
  tvOrigin, tvTarget: TTVItem;
  fCancel: Boolean;
begin
  if NewTarget <> nil then begin
    if Changing then begin
      try
	tvOrigin := InitItemData(Selected.ItemId, 0);
	tvTarget := InitItemData(NewTarget.ItemId, 0);
	TreeView_GetItem(WindowHandle, tvOrigin);
	TreeView_GetItem(WindowHandle, tvTarget);
	TreeView_EnsureVisible(WindowHandle, NewTarget.ItemId);
	fCancel := false;
	if Assigned(FOnMultiSelecting) then FOnMultiSelecting(NewTarget, FSelectAllParam, fCancel);
	TreeView_SelectItem(WindowHandle, NewTarget.ItemId);
	with tvTarget do if fCancel then
	  State := State and not TVIS_SELECTED
	else State := State xor TVIS_SELECTED;
	TreeView_SetItem(WindowHandle, tvTarget);
	if tvOrigin.hItem <> tvTarget.hItem then TreeView_SetItem(WindowHandle, tvOrigin);
      finally
	Changed;
      end;
    end;
  end;
end;

procedure TTreeViewEx.SelectRange(First, Last: TTreeNode);
var
  tvItem: TTVItem;
  tnNext: TTreeNode;
  fCancel: Boolean;
begin
  if First <> nil then begin
    tvItem := InitItemData(nil, TVIS_SELECTED);
    tnNext := First;
    while tnNext <> nil do begin
      fCancel := false;
      if Assigned(FOnMultiSelecting) then FOnMultiSelecting(tnNext, FSelectAllParam, fCancel);
      with tvItem do begin
	hItem := tnNext.ItemId;
	if fCancel then
	  State := State and not TVIS_SELECTED
	else State := State or TVIS_SELECTED;
      end;
      TreeView_SetItem(WindowHandle, tvItem);
      if tnNext <> Last then
	tnNext := tnNext.GetNext
      else Break;
    end;
  end;
end;

procedure TTreeViewEx.UnselectRange(First, Last: TTreeNode);
var
  tvItem: TTVItem;
  tnNext: TTreeNode;
begin
  if First <> nil then begin
    tvItem := InitItemData(nil, 0);
    tnNext := First;
    while tnNext <> nil do begin
      tvItem.hItem := tnNext.ItemId;
      TreeView_SetItem(WindowHandle, tvItem);
      if tnNext <> Last then
	tnNext := tnNext.GetNext
      else Break;
    end;
  end;
end;

procedure TTreeViewEx.CMColorChanged(var Message: TMessage);
begin
  inherited;
  CreateHandle;
  ThemeHandler;
  CreateColorBrush;
  SendMessage(WindowHandle, TVM_SETBKCOLOR, 0, ColorToRGB(Color));
end;

procedure TTreeViewEx.CMMouseEnter(var Message: TMessage);
begin
  if FToolTip <> nil then FToolTip.Activate := true;
  inherited;
end;

procedure TTreeViewEx.CMMouseLeave(var Message: TMessage);
begin
  if FToolTip <> nil then FToolTip.Activate := false;
  inherited;
end;

procedure TTreeViewEx.CMShowingChanged(var Message: TMessage);
begin
  if Showing then SetItemHeight(GetSystemMetrics(SM_CYSMICON)+ 2);
  inherited;
end;

procedure TTreeViewEx.CNNotify(var Message: TWMNotify);
begin
  with Message do begin
    case NMHdr^.Code of
      NM_CUSTOMDRAW: with PNMCustomDraw(NMHdr)^ do begin
	case dwDrawStage of
	  CDDS_PREPAINT: Result := CDRF_NOTIFYITEMDRAW;
	  CDDS_ITEMPREPAINT: begin
	    if Color <> clWindow then FillRect(hdc, rc, FColorBrush);
	    if (uItemState and CDIS_SELECTED) <> 0 then with PNMTVCustomDraw(NMHdr)^ do begin
	      clrText := ColorToRGB(clHighlightText);
	      clrTextBk := ColorToRGB(clHighlight);
	    end;
	  end;
	end;
      end;
      TVN_SELCHANGING: with PNMTreeView(Message.NMHdr)^ do begin
	if not FSelecting then Message.Result := ord(CanChange(Items.GetNode(itemNew.hItem)) = false);
	Exit;
      end;
      TVN_SELCHANGED: with PNMTreeView(Message.NMHdr)^ do begin
	if not FSelecting then Change(Items.GetNode(itemNew.hItem));
	Exit;
      end;
    end;
  end;
  inherited;
end;

procedure TTreeViewEx.WMCreate(var Message: TWMCreate);
begin
  inherited;
  ThemeHandler;
  if not (csDesigning in ComponentState) then FToolTip := TToolTip.Create(Self);
  if FToolTip <> nil then with FToolTip do begin
    ToolTipStyle := ttsUnderscore;
    AnsiText := Self.Hint;
  end;
end;

procedure TTreeViewEx.WMDestroy(var Message: TWMDestroy);
begin
  FToolTip.Free;
end;

procedure TTreeViewEx.WMKeyDown(var Message: TWMKey);

  procedure RemoveKeyMsg;
  var
    msgKey: TMsg;
  begin
    PeekMessage(msgKey, WindowHandle, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE);
  end;

begin
  if MultiSelect then begin
    case ShiftState of
      0: UnselectAll(Selected);
      MK_SHIFT: begin
	case Message.CharCode of
	  VK_UP: begin
	    ShiftSelect(Selected.GetPrev);
	    Exit;
	  end;
	  VK_DOWN: begin
	    ShiftSelect(Selected.GetNext);
	    Exit;
	  end;
	end;
      end;
      MK_CONTROL: begin
	case Message.CharCode of
	  VK_SPACE: begin
	    ControlSelect(Selected);
	    RemoveKeyMsg;
	    Exit;
	  end;
	  VK_UP: begin
	    MoveFocus(TVGN_PREVIOUS);
	    RemoveKeyMsg;
	    Exit;
	  end;
	  VK_DOWN: begin
	    MoveFocus(TVGN_NEXT);
	    RemoveKeyMsg;
	    Exit;
	  end;
	  ord('A'): begin
	    SelectAllItems(false);
	    RemoveKeyMsg;
	    Exit;
	  end;
	  ord('U'): begin
	    UnselectAllItems(false);
	    RemoveKeyMsg;
	    Exit;
	  end;
	end;
      end;
    end;
  end;
  inherited;
end;

procedure TTreeViewEx.WMLButtonDown(var Message: TWMMouse);
var
  htInfo: TTVHitTestInfo;
begin
  with htInfo do begin
    with Message do pt := Point(XPos, YPos);
    SendMessage(WindowHandle, TVM_HITTEST, 0, Longint(@htInfo));
    case flags of
      TVHT_ONITEMLABEL: begin
	Windows.SetFocus(WindowHandle);
	if MultiSelect then begin
	  case ShiftState of
	    0: UnselectAll(Items.GetNode(hItem));
	    MK_SHIFT: begin
	      ShiftSelect(Items.GetNode(hItem));
	      Exit;
	    end;
	    MK_CONTROL: begin
	      ControlSelect(Items.GetNode(hItem));
	      Exit;
	    end;
	  end;
	end;
      end;
    end;
  end;
  inherited;
end;

procedure TTreeViewEx.WMSize(var Message: TWMSize);
begin
  inherited;
  if FToolTip <> nil then FToolTip.ToolRect := ClientRect;
end;

constructor TTreeViewEx.Create(AOwner: TComponent);
begin
  inherited;
  CreateColorBrush;
  ToolTips := true;
  {$IFDEF CONDITIONALEXPRESSIONS}
  MultiSelectStyle := [msControlSelect, msShiftSelect];
  {$ENDIF}
end;

destructor TTreeViewEx.Destroy;
begin
  FreeColorBrush;
  inherited;
end;

procedure TTreeViewEx.EnumSelectedNodes;
var
  fCancel: Boolean;
  tnNext: TTreeNode;
begin
  if MultiSelect and Assigned(FOnEnumSelectedNodes) then begin
    fCancel := false;
    tnNext := Items.GetFirstNode;
    while tnNext <> nil do begin
      if tnNext.Selected then begin
	FOnEnumSelectedNodes(tnNext, fCancel);
	if fCancel then Break;
      end;
      tnNext := tnNext.GetNext;
    end;
  end;
end;

procedure TTreeViewEx.SelectAllItems(const SelectAllParam: Longbool);
begin
  if MultiSelect then begin
    FSelectAllParam := SelectAllParam;
    SelectAll;
    FSelectAllParam := false;
  end;
end;

procedure TTreeViewEx.UnselectAllItems(const SelectAllParam: Longbool);
begin
  if MultiSelect then begin
    FSelectAllParam := SelectAllParam;
    UnselectAll(nil);
    FSelectAllParam := false;
  end;
end;

procedure Register;
begin
  RegisterComponents('Modified', [TTreeViewEx]);
end;

end.

