{$J+,Z4}
unit UTF8;
{$IFDEF CONDITIONALEXPRESSIONS} {$WARN SYMBOL_PLATFORM OFF} {$ENDIF}

{------------------------------------------------------------------------------}
{                                                                              }
{          UTF8 utilities - especially for versions prior to Delphi 6          }
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2003-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

const // from errno.h
  EPERM		 = 1;
  ENOENT	 = 2;
  ESRCH 	 = 3;
  EINTR		 = 4;
  EIO		 = 5;
  ENXIO		 = 6;
  E2BIG		 = 7;
  ENOEXEC	 = 8;
  EBADF		 = 9;
  ECHILD	= 10;
  EAGAIN	= 11;
  ENOMEM	= 12;
  EACCES	= 13;
  EFAULT	= 14;
  EBUSY		= 16;
  EEXIST	= 17;
  EXDEV		= 18;
  ENODEV	= 19;
  ENOTDIR	= 20;
  EISDIR	= 21;
  EINVAL	= 22;
  ENFILE	= 23;
  EMFILE	= 24;
  ENOTTY	= 25;
  EFBIG		= 27;
  ENOSPC	= 28;
  ESPIPE	= 29;
  EROFS		= 30;
  EMLINK	= 31;
  EPIPE		= 32;
  EDOM		= 33;
  ERANGE	= 34;
  EDEADLK	= 36;
  ENAMETOOLONG	= 38;
  ENOLCK	= 39;
  ENOSYS	= 40;
  ENOTEMPTY	= 41;
  EILSEQ	= 42;

const // from winnls.h
  CP_UTF7		= 65000;	// UTF-7 translation
  CP_UTF8		= 65001;	// UTF-8 translation

  MB_PRECOMPOSED	= $00000001;	// use precomposed chars
  MB_COMPOSITE		= $00000002;	// use composite chars
  MB_USEGLYPHCHARS	= $00000004;	// use glyph chars, not ctrl chars
  MB_ERR_INVALID_CHARS	= $00000008;	// error for invalid chars

  WC_COMPOSITECHECK	= $00000200;	// convert composite to precomposed
  WC_DISCARDNS		= $00000010;	// discard non-spacing chars
  WC_SEPCHARS		= $00000020;	// generate separate chars
  WC_DEFAULTCHAR	= $00000040;	// replace w/ default char

const
  CHAR_			= 'CHAR';
  US_ASCII 		= 'US-ASCII';
  WINDOWS_ 		= 'WINDOWS-';
  ISO_8859_1		= 'ISO-8859-1';
  GB2312		= 'GB2312';
  EUC_CN		= 'EUC-CN';
  BIG5			= 'BIG5';
  KSC_5601		= 'KSC_5601';
  EUC_KR		= 'EUC-KR';
  UTF_8			= 'UTF-8';
  _1252			= '1252';

const
  UTF8Factor		= 4;
  Utf8BomDWord		= $00BFBBEF;
  Utf8BomChars		= #$EF + #$BB + #$BF;
  Utf16BomChars		= #$FE + #$FF;
  UnicodeBomChar	= #$FEFF;

type
  Utf8String = type String;
  WideString = type String;

function IsWideString(const S: String): Longbool;

function CheckWinCharset(const Charset: String): String;

function AnsiTextAtPos(const ASubText, AText: String; Pos: Integer): Longbool;

function CharToAnsi(const Char: Char; const Chars: PChar; MaxBytes: Integer; const Charset: String): Integer;
function CharToUtf8(const Char: Char; const Chars: PChar; MaxBytes: Integer; const Charset: String): Integer;
function CharToWide(const Char: Char; const Chars: PChar; MaxBytes: Integer; const Charset: String): Integer;

function AnsiToAnyString(const Source: AnsiString; const Charset: String): AnsiString;
function AnyStringToAnsi(const Source: AnsiString; const Charset: String): AnsiString;

function Utf8ToAnyString(const Source: Utf8String; const Charset: String): AnsiString;
function AnyStringToUtf8(const Source: AnsiString; const Charset: String): Utf8String;

function WideToAnyString(const Source: WideString; const Charset: String): AnsiString;
function AnyStringToWide(const Source: AnsiString; const Charset: String): WideString;

function AnsiToWideString(const Ansi: AnsiString): WideString;
function WideStringToAnsi(const Wide: WideString): AnsiString;

function Utf8ToWideString(const Utf8: Utf8String): WideString;
function WideStringToUtf8(const Wide: WideString): Utf8String;

function AnsiStringToUtf8(const Ansi: AnsiString): Utf8String;
function Utf8ToAnsiString(const Utf8: Utf8String): AnsiString;

function Utf8OrAnsi(const S: String): String;
function Utf8OrWide(const S: String): String;


implementation

uses
  Windows, SysUtils;

const // Iconv params
  TRANSLIT	= '//TRANSLIT';
  UCS_2_I	= 'UCS-2-INTERNAL';

const // convenience
  STR_LEN_OFF	= 4;
  NULL		= #0;
  E		= '';

type
  size_t = Integer;
  state_t = Cardinal;
  iconv_t = ^iconv_struct;
  iconv_struct = packed record
    iState: state_t;
    oState: state_t;
  end;

var
  (* Allocates descriptor for code conversion from encoding 'fromcode' to encoding `tocode'. *)
  libiconv_open: function(const tocode, fromcode: PChar): iconv_t; cdecl;
  (* Frees resources allocated for conversion descriptor 'cd'. *)
  libiconv_close: function(cd: iconv_t): Integer; cdecl;
  (* Retrieves errno after libiconv fails with a return value of -1. *)
  iconv_getlasterror: function: Integer; cdecl;
  (* Canonicalize an encoding name.
     The result is either a canonical encoding name, or name itself. *)
  iconv_canonicalize: function(const name: PChar): PChar; cdecl;
  (* Determine the current locale's character encoding, and canonicalize it
     into one of the canonical names listed in config.charset.
     The result must not be freed; it is statically allocated.
     If the canonical name cannot be determined, the result is a non-canonical name. *)
  locale_charset: function: PChar; cdecl;
  (* Converts, using conversion descriptor 'cd', at most '*inbytesleft' bytes
     starting at '*inbuf', writing at most '*outbytesleft' bytes starting at '*outbuf'.
     Decrements '*inbytesleft' and increments '*inbuf' by the same amount.
     Decrements '*outbytesleft' and increments '*outbuf' by the same amount. *)
  libiconv: function(cd: iconv_t; var inbuf: PChar; var inbytesleft: size_t;
		     var outbuf: PChar; var outbytesleft: size_t): size_t; cdecl;

function WideToUtf8String(const Wide: String): String; register; assembler;
asm	// EAX = @Wide
  OR	EAX,EAX
  JE	@EXIT
  PUSH	EAX
  PUSH	EDX
  MOV	EAX,[EAX - STR_LEN_OFF]
  MOV	EDX,Utf8Factor / 2
  MUL	EDX
  CALL	System.@NewAnsiString
  POP	EDX
  MOV	[EDX],EAX
  MOV	EDX,EAX
  POP	EAX
  PUSH	EBX
  PUSH	EDX
  XOR	EBX,EBX
  @STR_LOOP:
  MOV	BX,[EAX]
  OR	EBX,EBX
  JE	@DONE
  ADD	EDX,Utf8Factor - 1
  XOR	ECX,ECX
  DEC	EDX
  CMP	EBX,00000800h
  JNB	@3BYTES
  DEC	EDX
  CMP	EBX,00000080h
  JNB	@2BYTES
  DEC	EDX
  JMP	@ASCII
  @3BYTES:
  INC	ECX
  MOV	BYTE PTR [EDX],BL
  AND	BYTE PTR [EDX],03Fh
  OR	BYTE PTR [EDX],080h
  SHR	EBX,6
  DEC	EDX
  @2BYTES:
  INC	ECX
  MOV	BYTE PTR [EDX],BL
  AND	BYTE PTR [EDX],03Fh
  OR	BYTE PTR [EDX],080h
  SHR	EBX,6
  DEC	EDX
  INC	ECX
  MOV	[EDX],BL
  MOV	BX,00FFh
  ROR	BX,CL
  OR	[EDX],BH
  ADD	EDX,ECX
  ADD	EAX,2
  JMP	@STR_LOOP
  @ASCII:
  MOV	[EDX],BL
  ADD	EAX,2
  INC	EDX
  JMP	@STR_LOOP
  @DONE:
  MOV	BYTE PTR [EDX],0
  POP	EAX
  SUB	EDX,EAX
  MOV	DWORD PTR [EAX - STR_LEN_OFF],EDX
  POP	EBX
  RET
  @EXIT:
  MOV	EAX,EDX
  CALL	System.@LStrClr
end;

function Utf8StringToWide(const Utf8: String): String; register; assembler;
asm	// EAX = @Utf8
  OR	EAX,EAX
  JE	@EXIT
  PUSH	ESI
  PUSH	EDI
  PUSH	EBX
  MOV	ESI,EAX
  MOV	EDI,EDX
  MOV	EAX,[EAX - STR_LEN_OFF]
  SHL	EAX,1
  ADD	EAX,2
  CALL	System.@NewAnsiString
  MOV	[EDI],EAX
  MOV	EBX,EDI
  MOV	EDI,EAX
  @STR_LOOP:
  XOR	EAX,EAX
  MOV	AL,[ESI]
  CMP	AL,0FDh
  JA	@ZERO
  CMP	AL,080h
  JB	@ASCII
  XOR	ECX,ECX
  @COUNT:
  INC	CL
  SHL	AL,1
  JC	@COUNT
  CMP	CL,3
  JB	@ZERO
  SHR	AL,CL
  DEC	ECX
  ADD	ESI,ECX
  NEG	ECX
  INC	ECX
  @LOOP:
  MOV	DL,[ESI + ECX]
  CMP	DL,080h
  JB	@ZERO
  AND	DL,07Fh
  CMP	DL,040h
  JNB	@ZERO
  SHL	EAX,6
  OR	AL,DL
  INC	ECX
  JNE	@LOOP
  MOV	[EDI],AX
  ADD	EDI,2
  JMP	@STR_LOOP
  @ASCII:
  MOV	[EDI],AX
  ADD	EDI,2
  INC	ESI
  OR	EAX,EAX
  JNE	@STR_LOOP
  @DONE:
  MOV	EBX,[EBX]
  SUB	EDI,EBX
  MOV	DWORD PTR [EBX - STR_LEN_OFF],EDI
  POP	EBX
  POP	EDI
  POP	ESI
  RET
  @ZERO:
  MOV	EAX,EBX
  CALL	System.@LStrClr
  POP	EBX
  POP	EDI
  POP	ESI
  RET
  @EXIT:
  MOV	EAX,EDX
  CALL	System.@LStrClr
end;

function AnsiToUtf8(const S: String): UTF8String;
var
  WideLen: Cardinal;
  Wide: String;
begin
  if S <> E then begin
    WideLen := succ(Length(S));
    SetString(Wide, nil, WideLen shl 1);
    WideLen := MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED or MB_ERR_INVALID_CHARS, Pointer(S), -1, Pointer(Wide), WideLen);
    if WideLen > 0 then begin
      Result := WideToUtf8String(Wide);
      Exit;
    end;
  end;
  Result := E;
end;

function Utf8ToAnsi(const S: UTF8String): String;
var
  Wide: String;
  WideLen: Cardinal;
begin
  if S <> E then begin
    Wide := Utf8StringToWide(S);
    WideLen := Length(Wide) shr 1;
    if WideLen > 1 then begin
      SetString(Result, nil, pred(WideLen));
      WideLen := WideCharToMultiByte(CP_ACP, 0, Pointer(Wide), -1, Pointer(Result), WideLen, nil, nil);
      if WideLen > 0 then Exit;
    end;
  end;
  Result := E;
end;

function AnsiToUtf8PChar(const Ansi, Utf8: PChar; Len: Cardinal): Longint;
var
  Utf8Str: UTF8String;
begin
  Utf8^ := NULL;
  Utf8Str := AnsiToUtf8(Ansi);
  Result := Length(Utf8Str);
  if (Result <> 0) and (Len >= Cardinal(Result)) then Move(Pointer(Utf8Str)^, Utf8^, succ(Result));
end;

function Utf8ToAnsiPChar(const Utf8, Ansi: PChar; Len: Cardinal): Longint;
var
  AnsiStr: AnsiString;
begin
  Ansi^ := NULL;
  AnsiStr := Utf8ToAnsi(Utf8);
  Result := Length(AnsiStr);
  if (Result <> 0) and (Len >= Cardinal(Result)) then Move(Pointer(AnsiStr)^, Ansi^, succ(Result));
end;

function WideToUtf8PChar(const Wide, Utf8: PChar; Len: Cardinal): Longint;
var
  Utf8Str: UTF8String;
begin
  Utf8^ := NULL;
  Utf8Str := WideToUtf8String(Wide);
  Result := Length(Utf8Str);
  if (Result <> 0) and (Len >= Cardinal(Result))then Move(Pointer(Utf8Str)^, Utf8^, succ(Result));
end;

function Utf8ToWidePChar(const Utf8, Wide: PChar; Len: Cardinal): Longint;
var
  WideStr: WideString;
begin
  Wide^ := NULL;
  WideStr := Utf8StringToWide(Utf8);
  Result := Length(WideStr);
  if (Result <> 0) and (Len >= Cardinal(Result)) then Move(Pointer(WideStr)^, Wide^, Result);
end;

function WideToAnsiPChar(const Wide, Ansi: PChar; Len: Cardinal): Longint;
var
  AnsiStr: AnsiString;
begin
  Ansi^ := NULL;
  AnsiStr := WideCharToString(PWideChar(Wide));
  Result := Length(AnsiStr);
  if (Result <> 0) and (Len >= Cardinal(Result)) then Move(Pointer(AnsiStr)^, Ansi^, succ(Result));
end;

function AnsiToWidePChar(const Ansi, Wide: PChar; Len: Cardinal): Longint;
var
  PCharLen: Cardinal;
  WideStr: WideString;
begin
  Wide^ := NULL;
  PCharLen := StrLen(Ansi);
  Result := succ(PCharLen) shl 1;
  SetString(WideStr, nil, Result);
  StringToWideChar(Ansi, Pointer(WideStr), succ(PCharLen));
  if (Result <> 0) and (Len >= Cardinal(Result)) then Move(Pointer(WideStr)^, Wide^, Result);
end;

function HasUtf8Bom(pUtf8: PChar): Longbool; register; assembler;
asm
  OR	EAX,EAX
  MOV	EDX,[EAX]
  JE	@EXIT
  AND	EDX,00FFFFFFh
  XOR	EAX,EAX
  CMP	EDX,Utf8BomDWord
  JNE	@EXIT
  INC	EAX
  RET
  @EXIT:
  XOR	EAX,EAX
end;

function GB2BIG(const Charset: String): String;
begin
  if (Charset <> E) and Assigned(iconv_canonicalize)
  and (iconv_canonicalize(Pointer(Charset)) = EUC_CN) then
    Result := BIG5
  else Result := E;
end;

function KSC2KR(const Charset: String): String;
begin
  if (Charset <> E) and Assigned(iconv_canonicalize)
  and (iconv_canonicalize(Pointer(Charset)) = KSC_5601) then
    Result := EUC_KR
  else Result := E;
end;

function CheckAsian(const Charset: String): String;
var
  sCS: String;
begin
  sCS := Charset;
  UniqueString(sCS);
  CharUpper(Pointer(sCS));
  if sCS <> BIG5 then Result := GB2BIG(sCS);
  if (Result = E) and (sCS <> EUC_KR) then Result := KSC2KR(sCS);
end;

function WINDOWS2CP(const Charset: String): String;
begin
  Result := Charset;
  if Result <> E then begin
    if Assigned(iconv_canonicalize) then
      Result := iconv_canonicalize(Pointer(Result))
    else if AnsiTextAtPos(WINDOWS_, Result, 1) then Result := 'CP' + Copy(Result, succ(Length(WINDOWS_)), MAXINT);
  end;
end;

function IsWideString(const S: String): Longbool;
begin
  Result := (S <> E) and (S[Length(S)] = NULL);
end;

function CheckWinCharset(const Charset: String): String;
var
  sCS: String;
begin
  sCS := Charset;
  UniqueString(sCS);
  CharUpper(Pointer(sCS));
  if (sCS = E) or (sCS = CHAR_) or (sCS = US_ASCII) then
    Result := WINDOWS_ + IntToStr(GetACP)
  else if sCS = ISO_8859_1 then
    Result := WINDOWS_ + _1252
  else Result := sCS;
end;

function AnsiTextAtPos(const ASubText, AText: String; Pos: Integer): Longbool;
var
  iText, iSubText: Integer;
begin
  Result := false;
  if (ASubText <> E) and (AText <> E) then begin
    iText := Length(AText);
    iSubText := Length(ASubText);
    if (Pos <= 0) or (Pos + iSubText > succ(iText)) then Exit;
    Result := (CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
			     Pointer(ASubText), iSubText,
			     @aText[Pos], iSubText) = 2);
  end;
end;

function CharToAnsi(const Char: Char; const Chars: PChar; MaxBytes: Integer; const Charset: String): Integer;
var
  SourceLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
begin
  Result := 0;
  if Assigned(libiconv_open) and (Chars <> nil) and (MaxBytes > 0) and (Charset <> E) then begin
    SourceLen := 1;
    pInput := @Char;
    pOutput := Chars;
    cd := libiconv_open(Pointer(locale_charset + TRANSLIT), Pointer(WINDOWS2CP(Charset)));
    if Integer(cd) <> -1 then begin
      try
	if libiconv(cd, pInput, SourceLen, pOutput, MaxBytes) <> -1 then
	  Result := pOutput - Chars
	else Result := CharToAnsi(Char, Chars, MaxBytes, CheckAsian(Charset));
      finally
	libiconv_close(cd);
      end;
    end;
  end;
  if Result = 0 then Chars^ := NULL;
end;

function CharToUtf8(const Char: Char; const Chars: PChar; MaxBytes: Integer; const Charset: String): Integer;
var
  SourceLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
  Str: String;
begin
  Result := 0;
  if (Chars <> nil) and (MaxBytes > 0) and (Charset <> E) then begin
    SourceLen := 1;
    pInput := @Char;
    pOutput := Chars;
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UTF_8, Pointer(WINDOWS2CP(Charset)));
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pInput, SourceLen, pOutput, MaxBytes) <> -1 then
	    Result := pOutput - Chars
	  else Result := CharToUtf8(Char, Chars, MaxBytes, CheckAsian(Charset));
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      Str := AnyStringToUtf8(Char, Charset);
      Result := Length(Str);
      if Result <= MaxBytes then Move(Pointer(Str)^, Chars^, Result);
    end;
  end;
  if Result = 0 then Chars^ := NULL;
end;

function CharToWide(const Char: Char; const Chars: PChar; MaxBytes: Integer; const Charset: String): Integer;
var
  SourceLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
  Str: String;
begin
  Result := 0;
  if (Chars <> nil) and (MaxBytes > 0) and (Charset <> E) then begin
    SourceLen := 1;
    pInput := @Char;
    pOutput := Chars;
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UCS_2_I, Pointer(WINDOWS2CP(Charset)));
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pInput, SourceLen, pOutput, MaxBytes) <> -1 then
	    Result := pOutput - Chars
	  else Result := CharToWide(Char, Chars, MaxBytes, CheckAsian(Charset));
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      Str := AnyStringToWide(Char, Charset);
      Result := Length(Str);
      if Result <= MaxBytes then Move(Pointer(Str)^, Chars^, Result);
    end;
  end;
  if Result = 0 then Chars^ := NULL;
end;

function AnsiToAnyString(const Source: AnsiString; const Charset: String): AnsiString;
var
  SourceLen, DestLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Source);
  if Assigned(libiconv_open) and (SourceLen > 0) and (Charset <> E) then begin
    pInput := Pointer(Source);
    DestLen := SourceLen * UTF8Factor;
    SetString(Result, nil, DestLen);
    pOutput := Pointer(Result);
    cd := libiconv_open(Pointer(WINDOWS2CP(Charset) + TRANSLIT), locale_charset);
    if Integer(cd) <> -1 then begin
      try
	if libiconv(cd, pInput, SourceLen, pOutput, DestLen) <> -1 then begin
	  Delete(Result, succ(pOutput - Pointer(Result)), MAXINT);
	  pOutput^ := NULL;
	  Exit;
	end
	else begin
	  Result := AnsiToAnyString(Source, CheckAsian(Charset));
	  Exit;
	end;
      finally
	libiconv_close(cd);
      end;
    end;
  end;
  Result := E;
end;

function AnyStringToAnsi(const Source: AnsiString; const Charset: String): AnsiString;
var
  SourceLen, DestLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Source);
  if Assigned(libiconv_open) and (SourceLen > 0) and (Charset <> E) then begin
    pInput := Pointer(Source);
    DestLen := SourceLen;
    SetString(Result, nil, DestLen);
    pOutput := Pointer(Result);
    cd := libiconv_open(Pointer(locale_charset + TRANSLIT), Pointer(WINDOWS2CP(Charset)));
    if Integer(cd) <> -1 then begin
      try
	if libiconv(cd, pInput, SourceLen, pOutput, DestLen) <> -1 then begin
	  Delete(Result, succ(pOutput - Pointer(Result)), MAXINT);
	  pOutput^ := NULL;
	  Exit;
	end
	else begin
	  Result := AnyStringToAnsi(Source, CheckAsian(Charset));
	  Exit;
	end;
      finally
	libiconv_close(cd);
      end;
    end;
  end;
  Result := E;
end;

function Utf8ToAnyString(const Source: Utf8String; const Charset: String): AnsiString;
var
  SourceLen, DestLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Source);
  if Assigned(libiconv_open) and (SourceLen > 0) and (Charset <> E) then begin
    pInput := Pointer(Source);
    DestLen := SourceLen;
    SetString(Result, nil, DestLen);
    pOutput := Pointer(Result);
    cd := libiconv_open(Pointer(WINDOWS2CP(Charset) + TRANSLIT), UTF_8);
    if Integer(cd) <> -1 then begin
      try
	if libiconv(cd, pInput, SourceLen, pOutput, DestLen) <> -1 then begin
	  Delete(Result, succ(pOutput - Pointer(Result)), MAXINT);
	  pOutput^ := NULL;
	  Exit;
	end
	else begin
	  Result := Utf8ToAnyString(Source, CheckAsian(Charset));
	  Exit;
	end;
      finally
	libiconv_close(cd);
      end;
    end;
  end;
  Result := E;
end;

function AnyStringToUtf8(const Source: AnsiString; const Charset: String): Utf8String;
var
  SourceLen, DestLen: Integer;
  pSource, pUtf8: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Source);
  if (SourceLen > 0) and (Charset <> E) then begin
    pSource := Pointer(Source);
    DestLen := SourceLen shl 2;
    SetString(Result, nil, DestLen);
    pUtf8 := Pointer(Result);
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UTF_8, Pointer(WINDOWS2CP(Charset)));
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pSource, SourceLen, pUtf8, DestLen) <> -1 then begin
	    Delete(Result, succ(pUtf8 - Pointer(Result)), MAXINT);
	    pUtf8^ := NULL;
	    Exit;
	  end
	  else begin
	    Result := AnyStringToUtf8(Source, CheckAsian(Charset));
	    Exit;
	  end;
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      Delete(Result, succ(AnsiToUtf8PChar(pSource, pUtf8, Length(Result))), MAXINT);
      Exit;
    end;
  end;
  Result := E;
end;

function WideToAnyString(const Source: WideString; const Charset: String): AnsiString;
var
  SourceLen, DestLen: Integer;
  pInput, pOutput: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Source);
  if Assigned(libiconv_open) and (SourceLen > 0) and (Charset <> E) then begin
    pInput := Pointer(Source);
    DestLen := SourceLen;
    SetString(Result, nil, DestLen);
    pOutput := Pointer(Result);
    cd := libiconv_open(Pointer(WINDOWS2CP(Charset) + TRANSLIT), UCS_2_I);
    if Integer(cd) <> -1 then begin
      try
	if libiconv(cd, pInput, SourceLen, pOutput, DestLen) <> -1 then begin
	  Delete(Result, succ(pOutput - Pointer(Result)), MAXINT);
	  pOutput^ := NULL;
	  Exit;
	end
	else begin
	  Result := WideToAnyString(Source, CheckAsian(Charset));
	  Exit;
	end;
      finally
	libiconv_close(cd);
      end;
    end;
  end;
  Result := E;
end;

function AnyStringToWide(const Source: AnsiString; const Charset: String): WideString;
var
  SourceLen, DestLen: Integer;
  pSource, pWide: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Source);
  if (SourceLen > 0) and (Charset <> E) then begin
    pSource := Pointer(Source);
    DestLen := SourceLen shl 1;
    SetString(Result, nil, DestLen + 2);
    pWide := Pointer(Result);
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UCS_2_I, Pointer(WINDOWS2CP(Charset)));
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pSource, SourceLen, pWide, DestLen) <> -1 then begin
	    Delete(Result, pWide - Pointer(Result) + 3, MAXINT);
	    PWideChar(pWide)^ := NULL;
	    Exit;
	  end
	  else begin
	    Result := AnyStringToWide(Source, CheckAsian(Charset));
	    Exit;
	  end;
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      AnsiToWidePChar(pSource, pWide, Length(Result));
      Exit;
    end;
  end;
  Result := E;
end;

function AnsiToWideString(const Ansi: String): WideString;
var
  SourceLen, DestLen: Integer;
  pAnsi, pWide: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Ansi);
  if SourceLen > 0 then begin
    pAnsi := Pointer(Ansi);
    DestLen := SourceLen shl 1;
    SetString(Result, nil, DestLen + 2);
    pWide := Pointer(Result);
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UCS_2_I, locale_charset);
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pAnsi, SourceLen, pWide, DestLen) <> -1 then begin
	    Delete(Result, pWide - Pointer(Result) + 3, MAXINT);
	    PWideChar(pWide)^ := NULL;
	    Exit;
	  end
	  else Result := E;
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      AnsiToWidePChar(pAnsi, pWide, Length(Result));
      Exit;
    end;
  end;
  Result := E;
end;

function WideStringToAnsi(const Wide: WideString): String;
var
  SourceLen, DestLen: Integer;
  pWide, pAnsi: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Wide) - 2;
  if SourceLen > 0 then begin
    pWide := Pointer(Wide);
    DestLen := SourceLen * UTF8Factor;
    SetString(Result, nil, DestLen);
    pAnsi := Pointer(Result);
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(Pointer(locale_charset + TRANSLIT), UCS_2_I);
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pWide, SourceLen, pAnsi, DestLen) <> -1 then begin
	    Delete(Result, succ(pAnsi - Pointer(Result)), MAXINT);
	    pAnsi^ := NULL;
	    Exit;
	  end;
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      Delete(Result, succ(WideToAnsiPChar(pWide, pAnsi, Length(Result))), MAXINT);
      Exit;
    end;
  end;
  Result := E;
end;

function Utf8ToWideString(const Utf8: Utf8String): WideString;
var
  SourceLen, DestLen: Integer;
  pUtf8, pWide: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Utf8);
  if SourceLen > 0 then begin
    pUtf8 := Pointer(Utf8);
    if HasUtf8Bom(pUtf8) then begin
      dec(SourceLen, 3);
      inc(pUtf8, 3);
    end;
    if SourceLen > 0 then begin
      DestLen := SourceLen shl 1;
      SetString(Result, nil, DestLen + 2);
      pWide := Pointer(Result);
      if Assigned(libiconv_open) then begin
	cd := libiconv_open(UCS_2_I, UTF_8);
	if Integer(cd) <> -1 then begin
	  try
	    if libiconv(cd, pUtf8, SourceLen, pWide, DestLen) <> -1 then begin
	      Delete(Result, pWide - Pointer(Result) + 3, MAXINT);
	      PWideChar(pWide)^ := NULL;
	      Exit;
	    end;
	  finally
	    libiconv_close(cd);
	  end;
	end;
      end
      else begin
	Delete(Result, succ(Utf8ToWidePChar(pUtf8, pWide, Length(Result))), MAXINT);
	Exit;
      end;
    end;
  end;
  Result := E;
end;

function WideStringToUtf8(const Wide: WideString): Utf8String;
var
  SourceLen, DestLen: Integer;
  pWide, pUtf8: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Wide) - 2;
  if SourceLen > 0 then begin
    pWide := Pointer(Wide);
    DestLen := SourceLen * UTF8Factor;
    SetString(Result, nil, DestLen);
    pUtf8 := Pointer(Result);
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UTF_8, UCS_2_I);
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pWide, SourceLen, pUtf8, DestLen) <> -1 then begin
	    Delete(Result, succ(pUtf8 - Pointer(Result)), MAXINT);
	    pUtf8^ := NULL;
	    Exit;
	  end
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      Delete(Result, succ(WideToUtf8PChar(pWide, pUtf8, Length(Result))), MAXINT);
      Exit;
    end;
  end;
  Result := E;
end;

function AnsiStringToUtf8(const Ansi: String): Utf8String;
var
  SourceLen, DestLen: Integer;
  pAnsi, pUtf8: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Ansi);
  if SourceLen > 0 then begin
    pAnsi := Pointer(Ansi);
    DestLen := SourceLen shl 2;
    SetString(Result, nil, DestLen);
    pUtf8 := Pointer(Result);
    if Assigned(libiconv_open) then begin
      cd := libiconv_open(UTF_8, locale_charset);
      if Integer(cd) <> -1 then begin
	try
	  if libiconv(cd, pAnsi, SourceLen, pUtf8, DestLen) <> -1 then begin
	    Delete(Result, succ(pUtf8 - Pointer(Result)), MAXINT);
	    pUtf8^ := NULL;
	    Exit;
	  end
	finally
	  libiconv_close(cd);
	end;
      end;
    end
    else begin
      Delete(Result, succ(AnsiToUtf8PChar(pAnsi, pUtf8, Length(Result))), MAXINT);
      Exit;
    end;
  end;
  Result := E;
end;

function Utf8ToAnsiString(const Utf8: Utf8String): String;
var
  SourceLen, DestLen: Integer;
  pUtf8, pAnsi: PChar;
  cd: iconv_t;
begin
  SourceLen := Length(Utf8);
  if SourceLen > 0 then begin
    pUtf8 := Pointer(Utf8);
    if HasUtf8Bom(Pointer(Utf8)) then begin
      dec(SourceLen, 3);
      inc(pUtf8, 3);
    end;
    if SourceLen > 0 then begin
      DestLen := SourceLen;
      SetString(Result, nil, DestLen);
      pAnsi := Pointer(Result);
      if Assigned(libiconv_open) then begin
	cd := libiconv_open(Pointer(locale_charset + TRANSLIT), UTF_8);
	if Integer(cd) <> -1 then begin
	  try
	    if libiconv(cd, pUtf8, SourceLen, pAnsi, DestLen) <> -1 then begin
	      Delete(Result, succ(pAnsi - Pointer(Result)), MAXINT);
	      pAnsi^ := NULL;
	      Exit;
	    end;
	  finally
	    libiconv_close(cd);
	  end;
	end;
      end
      else begin
	Delete(Result, succ(Utf8ToAnsiPChar(pUtf8, pAnsi, Length(Result))), MAXINT);
	Exit;
      end;
    end;
  end;
  Result := E;
end;

function Utf8OrAnsi(const S: String): String;
begin
  if S <> E then begin
    if S[Length(S)] <> NULL then begin
      Result := Utf8ToAnsiString(S);
      if Result = E then Result := S;
    end
    else Result := WideStringToAnsi(S);
  end
  else Result := E;
end;

function Utf8OrWide(const S: String): String;
begin
  if S <> E then begin
    if S[Length(S)] <> NULL then begin
      Result := Utf8ToWideString(S);
      if Result = E then Result := AnsiToWideString(S);
    end
    else Result := S;
  end
  else Result := E;
end;

function IConvInit: Longbool;
const
  ICONV = 'IConv.dll';
var
  hIConv: THandle;
begin
  hIConv := GetModuleHandle(ICONV);
  if hIConv = 0 then hIConv := LoadLibrary(ICONV);
  libiconv_open := GetProcAddress(hIConv, 'libiconv_open');
  libiconv_close := GetProcAddress(hIConv, 'libiconv_close');
  iconv_getlasterror := GetProcAddress(hIConv, 'iconv_getlasterror');
  iconv_canonicalize := GetProcAddress(hIConv, 'iconv_canonicalize');
  locale_charset := GetProcAddress(hIConv, 'locale_charset');
  libiconv := GetProcAddress(hIConv, 'libiconv');
  Result := Assigned(iconv_canonicalize);
end;


initialization

  IConvInit;


end.

