{$J+,Z4}
unit AboutDialog;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Classes, SysUtils, StrUtil, Messages, Graphics, Forms, Controls, ExtCtrls, StdCtrls,
  Buttons, FlatBtn, GroupBoxEx, HotLabel, FileUtils, MailUtils, WebInterface,
  PMMessages, PMPGPForm, PMPGPCommon, PMPGPResources;

const // Update & key checking
  LOOKUP_COMMAND 	= 'lookup?op=get&search=0xFC55A062C45D831B';
  KEYSERVER1_URL 	= 'http://pgp.surfnet.nl:11371/pks/' + LOOKUP_COMMAND;
  KEYSERVER2_URL 	= 'http://subkeys.pgp.net:11371/pks/' + LOOKUP_COMMAND;
  KEYSERVER3_URL 	= 'http://wwwkeys.pgp.net:11371/pks/' + LOOKUP_COMMAND;
  KEY_SERVERS		= KEYSERVER1_URL + TAB + KEYSERVER2_URL + TAB + KEYSERVER3_URL;
  HOMEPAGE_PATH		= 'https://www.pmpgp.de/pmpgp/';
  KEY_DATA_FILE		= 'idw.asc';
  DOWNLOAD_PAGE		= 'download.htm';
  HISTORY_DE		= 'historyde.htm';
  HISTORY_EN 		= 'historyen.htm';
  VERSION_TAG		= 'meta name="Version" content=';

type
  TAboutDialog = class(TPMPGPForm)
    pnLeft: TPanel;
    pnImage: TPanel;
    imgThorn: TImage;
    pnCenter: TPanel;
    gbeInfo: TGroupBoxEx;
    hlVersion: THotLabel;
    hlCopyright: THotLabel;
    hlName: THotLabel;
    hlUpdate: THotLabel;
    hlTopSpace: THotLabel;
    hlPGPKey: THotLabel;
    hlBottomSpace: THotLabel;
    hlAddress: THotLabel;
    fbAbout: TFlatBtn;
    pnRight: TPanel;
    tmWait: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure hlUpdateClick(Sender: TObject);
    procedure hlPGPKeyClick(Sender: TObject);
    procedure hlAddressClick(Sender: TObject);
    procedure tmWaitTimer(Sender: TObject);
    procedure fbAboutClick(Sender: TObject);
  private
    DomainName: String;
    PGPKeyData: String;
    PMPGPProductInfo: TPMPGPProductInfo;
    SecondsRemaining: DWord;
    ReloadWinInet: Longbool;
    WinInetInitialized: Longbool;
    WinInetThread: TWinInetThread;
    procedure SetAboutButtonGlyph(ModalResult: TModalResult);
    procedure StartDownload(const URLs: String; OnDownloaded: TNotifyEvent);
    function  GetWinInetError(Error: TWinInetError): String;
    procedure CleanupOnDownloadCancelled;
    procedure SetLabelTexts;
    procedure SetStatusText;
    procedure OnConnect(Sender: TObject);
    procedure OnKeyServerChecked(Sender: TObject);
    procedure OnUpdateChecked(Sender: TObject);
  protected
  public
  end;

function About(MainClass: TObject; const VersionInfo: TPMPGPProductInfo;
	       PMReloadsWinsock: Longbool; var SupportMessage: Longbool): String;


implementation

{$R *.DFM}

uses
  MessageDialogs;


function About(MainClass: TObject; const VersionInfo: TPMPGPProductInfo;
	       PMReloadsWinsock: Longbool; var SupportMessage: Longbool): String;
var
  frmAboutDlg: TAboutDialog;
begin
  frmAboutDlg := TAboutDialog.Create(MainClass);
  if frmAboutDlg <> nil then with frmAboutDlg do begin
    try
      PMPGPProductInfo := VersionInfo;
      ReloadWinInet := PMReloadsWinsock;
      SupportMessage := (ShowModal = mrYes);
      Result := PGPKeyData;
    finally
      Release;
    end;
  end;
end;


// TAboutDialog ------------------------------------------------------------------------------------

procedure TAboutDialog.FormCreate(Sender: TObject);
begin
  SetLabelTexts;
  DisableSystemMenu;
  SetAboutButtonGlyph(mrOk);
end;

procedure TAboutDialog.FormShow(Sender: TObject);
begin
  with PMPGPProductInfo do begin
    hlVersion.Caption := 'Version ' + Version;
    hlCopyright.Caption := Copyright;
    hlName.Caption := Name;
  end;
end;

procedure TAboutDialog.hlUpdateClick(Sender: TObject);
begin
  if hlUpdate.Hovering then StartDownload(HOMEPAGE_PATH + DOWNLOAD_PAGE, OnUpdateChecked);
end;

procedure TAboutDialog.hlPGPKeyClick(Sender: TObject);
begin
  if hlPGPKey.Hovering then StartDownload(KEY_SERVERS + TAB + HOMEPAGE_PATH + TAB + KEY_DATA_FILE, OnKeyServerChecked);
end;

procedure TAboutDialog.hlAddressClick(Sender: TObject);
begin
  if hlAddress.Hovering then begin
    ProcessMessages(0);
    WaitForSingleObject(GetCurrentThread, 100);
    try
      Self.Enabled := false;
      ModalResult := mrYes;
    finally
      SetLabelTexts;
    end;
  end;
end;

procedure TAboutDialog.SetAboutButtonGlyph(ModalResult: TModalResult);
begin
  case ModalResult of
    mrOk: fbAbout.Glyph.LoadFromResourceID(hInstance, IDW_BUTTON_OK);
    mrCancel: fbAbout.Glyph.LoadFromResourceID(hInstance, IDW_BUTTON_CANCEL);
  end;
end;

procedure TAboutDialog.StartDownload(const URLs: String; OnDownloaded: TNotifyEvent);
begin
  if (URLs <> E) and Assigned(OnDownloaded) then begin
    WinInetInitialized := LoadWinInet(ReloadWinInet);
    WinInetThread := TWinInetThread.Create(hlName.Caption);
    if WinInetThread <> nil then with WinInetThread do begin
      SetAboutButtonGlyph(mrCancel);
      gbeInfo.Enabled := false;
      ProcessMessages(0);
      WaitForSingleObject(GetCurrentThread, 100);
      WebTargets := URLs;
      OnConnect := Self.OnConnect;
      OnTerminate := OnDownloaded;
      Resume;
    end;
  end;
end;

function TAboutDialog.GetWinInetError(Error: TWinInetError): String;
begin
  case Error of
    wieWinInet: Result := PMPGPStrings[WEB_LIBRARY_ERROR];
    wieConnect: Result := PMPGPStrings[WEB_CONNECT_ERROR];
    wieLoadPage: Result := PMPGPStrings[WEB_LOAD_ERROR];
    wieReadPage: Result := PMPGPStrings[WEB_READ_ERROR];
  else
    Result := E;
  end;
end;

procedure TAboutDialog.CleanupOnDownloadCancelled;
begin
  SetAboutButtonGlyph(mrOk);
  gbeInfo.Enabled := true;
  tmWait.Enabled := false;
  SetLabelTexts;
end;

procedure TAboutDialog.SetLabelTexts;
begin
  Caption := PMPGPStrings[ABOUT_DIALOG] + DOTS;
  hlUpdate.Caption := GET_UPDATE;
  hlPGPKey.Caption := IDWS_PGP_KEY;
  hlAddress.Caption := IDWS_ADDRESS;
  ProcessMessages(0);
end;

procedure TAboutDialog.SetStatusText;
begin
  Caption := PMPGPStrings[SEARCHING_] + SP + DomainName + SP + LB + IntToStr(SecondsRemaining) + RB + DOTS;
end;

procedure TAboutDialog.OnConnect(Sender: TObject);
begin
  SecondsRemaining := WinInetThread.ConnectTimeout;
  DomainName := WinInetThread.WebDomain;
  tmWait.Enabled := false;
  SetStatusText;
  tmWait.Enabled := true;
end;

procedure TAboutDialog.OnKeyServerChecked(Sender: TObject);
var
  iPos: Integer;
begin
  SetLabelTexts;
  case WinInetThread.WinInetError of
    wieNone: begin
      Self.Enabled := false;
      PGPKeyData := WinInetThread.WebData;
      iPos := LastStrPos(PGP_END + KEY_END, PGPKeyData) + Length(PGP_END) + Length(PGP_END);
      Delete(PGPKeyData, iPos, MAXINT);
      iPos := pred(FirstStrPos(PGP_BEG + KEY_BEG, PGPKeyData));
      Delete(PGPKeyData, 1, iPos);
    end;
    wieCancelled: begin
      CleanupOnDownloadCancelled;
      Exit;
    end;
  else
    Error(GetWinInetError(WinInetThread.WinInetError));
  end;
  Close;
end;

procedure TAboutDialog.OnUpdateChecked(Sender: TObject);
var
  sWebVersion, sAppVersion, sError: String;
  iPos: Integer;
begin
  SetLabelTexts;
  case WinInetThread.WinInetError of
    wieNone: begin
      Self.Enabled := false;
      sWebVersion := WinInetThread.WebData;
      iPos := FirstStrPos(VERSION_TAG, sWebVersion);
      if iPos > 0 then begin
	Delete(sWebVersion, 1, iPos + Length(VERSION_TAG));
	Delete(sWebVersion, FirstCharPos(QU, sWebVersion), MAXINT);
	sAppVersion := hlVersion.Caption;
	Delete(sAppVersion, 1, FirstCharPos(SP, sAppVersion));
	Delete(sAppVersion, FirstCharPos(SP, sAppVersion), MAXINT);
	while Length(sWebVersion) < 4 do sWebVersion := sWebVersion + '0';
	iPos := 0;
	repeat
	  iPos := NextCharPos(DOT, sAppVersion, succ(iPos));
	  Delete(sAppVersion, iPos, 1);
	until iPos = 0;
	while Length(sAppVersion) < 4 do sAppVersion := sAppVersion + '0';
	try
	  if StrToInt(sWebVersion) > StrToInt(sAppVersion) then begin
	    if Confirm(PMPGPStrings[NEW_VERSION]) then begin
	      if PMPGPProductInfo.MainLanguageID = LANG_GERMAN then
		MainClass.LaunchBrowser(HOMEPAGE_PATH + HISTORY_DE + '#new')
	      else MainClass.LaunchBrowser(HOMEPAGE_PATH + HISTORY_EN + '#new');
	    end;
	  end
	  else Info(PMPGPStrings[NO_NEW_VERSION]);
	except
	  sError := PMPGPStrings[VERSION_COMPARE_ERROR];
	end;
      end
      else sError := PMPGPStrings[VERSION_FIND_ERROR];
    end;
    wieCancelled: begin
      CleanupOnDownloadCancelled;
      Exit;
    end;
  else
    sError := GetWinInetError(WinInetThread.WinInetError);
  end;
  if sError <> E then Error(sError);
  Close;
end;

procedure TAboutDialog.tmWaitTimer(Sender: TObject);
begin
  if StrStart(PMPGPStrings[SEARCHING_], Caption) then begin
    if SecondsRemaining > 0 then dec(SecondsRemaining);
    SetStatusText;
  end;
end;

procedure TAboutDialog.fbAboutClick(Sender: TObject);
begin
  if tmWait.Enabled then begin
    ModalResult := mrNone;
    WinInetThread.Cancel;
  end;
end;


end.

