{$J+,Z4}
unit AutoListHandler;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Classes, SysUtils, StrUtil, Forms, FileUtils, MailUtils, PMMessages, PMOptions,
  KeyPropTypes, PMPGPCommon, PMPGPResources, PMHandler, PGPHandler;

const // Flag constants
  FLAG_SIZE		=   2;
  FLAG_ENABLED		= '+';
  FLAG_DISABLED		= '-';
  FLAG_USE_SMIME	= 'S';

const // AutoList option flags
  AUTO_FLAG_NOTHING	=  0;
  AUTO_FLAG_ENCRYPT	=  1;
  AUTO_FLAG_SIGN	=  2;
  AUTO_FLAG_ASK		=  4;
  AUTO_FLAG_PMIME	=  8;
  AUTO_FLAG_SMIME	= 16;
  AUTO_FLAG_DISABLED	= 32;
  AUTO_FLAG_CODE_FILTER = AUTO_FLAG_ENCRYPT or AUTO_FLAG_SIGN;
  AUTO_FLAG_BASE_FILTER	= AUTO_FLAG_CODE_FILTER or AUTO_FLAG_ASK;
  AUTO_FLAG_MIME_FILTER	= AUTO_FLAG_PMIME or AUTO_FLAG_SMIME;

const // AutoList constants
  AUTO_LIST		= 'AutoList.txt';
  HEX_ID_NULL		= HEX_MARKER + '0000000000000000';
  MAX_LIST_LINE_LEN	= MaxUserIDLen + Length(BS) + DefHexIDLen + FLAG_SIZE + Length(CRLF);

type
  TAddressBookList = Class(TStringList)
  private
    FAddBookIndex: Integer;
    FAddressBookFlags: String;
    FAddressCache: String;
    FMaxCacheSize: Integer;
    FOmitNullHexIDs: Longbool;
    function	AddAddressBook(const FileName: String): Integer;
    function	GetAddressList(BookIndex: Integer): String;
    procedure	ReadAddressBooks(const Directory: String);
    procedure	CloseAddressBooks;
  protected
  public
    constructor	Create(const PathList: String);
    destructor	Destroy; override;
    function	AllAddresses: String;
    procedure	CreateAddressCache(const AddressBookFlags: String);
    function	ResolveNames(const NameCommaList: String): String;
    property	AddBookIndex: Integer read FAddBookIndex write FAddBookIndex;
    property	AddressList[BookIndex: Integer]: String read GetAddressList; default;
  end;

const
  BEG_COMMENT_BLOCK	= '#++++';
  END_COMMENT_BLOCK	= '++++#';
  UPDATE_HEADER         = '# Update ';

const
  AL_FLAG_ENCRYPT	= CIFF_ENCRYPT_AS_WELL;
  AL_FLAG_SIGN		= CIFF_SIGN_FILE;

type
  PALItem = ^TALItem;
  TALItem = Record
    UserID: String;
    HexID: String;
    ALFlags: Integer;
    ALObject: TObject;
  end;

type
  PALItemList = ^TALItemList;
  TALItemList = Array[0..pred(MaxListSize)] of TALItem;

type
  TSortOrder = (soNone, soByHexID, soByUserID);

type
  TAutoList = Class(TStrings)
  private
    FAddressBookList: TAddressBookList;
    FAutoList: PALItemList;
    FCount: Integer;
    FCapacity: Integer;
    FCommentMarker: Char;
    FDefaultFlags: Integer;
    FDuplicates: Longbool;
    FFileName: String;
    FListHeader: String;
    FLastChecked: Integer;
    FSortOrder: TSortOrder;
    function	GetListStr: String;
    function	GetListVersion: Integer;
    procedure	SetListVersion(Value: Integer);
    procedure	SetDuplicates(Value: Longbool);
    function	ExtractAutoListData(const AutoListEntry: String; var ListItem: TALItem): Longbool;
    function	RestoreAutoListEntry(const ListItem: TALItem): String;
    function	InsertALItem(Index: Integer; const ListItem: TALItem): Integer;
    function	InsertItem(Index: Integer; const Value: String): Integer;
    procedure	ExchangeItems(Index1, Index2: Integer);
    procedure	Grow;
  protected
    function	HexIDFind(const Item: String; var Index: Integer): Longbool;
    procedure	HexIDSort(FirstItem, LastItem: Integer);
    function	UserIDFind(const Item: String; var Index: Integer): Longbool;
    procedure	UserIDSort(FirstItem, LastItem: Integer);
    function	Get(Index: Integer): String; override;
    procedure	Put(Index: Integer; const Value: String); override;
    function	GetObject(Index: Integer): TObject; override;
    procedure	PutObject(Index: Integer; AObject: TObject); override;
    function	GetCount: Integer; override;
    function	GetItem(Index: Integer): TALItem;
    procedure	SetItem(Index: Integer; Value: TALItem);
    function	GetFlags(Index: Integer): Integer;
    procedure	SetFlags(Index: Integer; Value: Integer);
    function	GetFlagsEncrypt(Index: Integer): Longbool;
    procedure	SetFlagsEncrypt(Index: Integer; Value: Longbool);
    function	GetFlagsSign(Index: Integer): Longbool;
    procedure	SetFlagsSign(Index: Integer; Value: Longbool);
    function	GetFlagsAsk(Index: Integer): Longbool;
    procedure	SetFlagsAsk(Index: Integer; Value: Longbool);
    function	GetFlagsPMime(Index: Integer): Longbool;
    procedure	SetFlagsPMime(Index: Integer; Value: Longbool);
    function	GetFlagsSMime(Index: Integer): Longbool;
    procedure	SetFlagsSMime(Index: Integer; Value: Longbool);
    function	GetFlagsDisabled(Index: Integer): Longbool;
    procedure	SetFlagsDisabled(Index: Integer; Value: Longbool);
    function	GetHexID(Index: Integer): String;
    procedure	SetHexID(Index: Integer; const Value: String);
    function	GetUserID(Index: Integer): String;
    procedure	SetUserID(Index: Integer; const Value: String);
    function	GetHexIDCommaText: String;
    function	GetUserIDText: String;
    {$IFDEF VER90}
    procedure	SetCapacity(NewCapacity: Integer);
    {$ELSE}
    procedure	SetCapacity(NewCapacity: Integer); override;
    {$ENDIF}
    function	GetTextStr: String; override;
    procedure	SetTextStr(const Value: String); override;
    procedure	Sort(Value: TSortOrder);
  public
    constructor	Create(const FileName: String);
    destructor	Destroy; override;
    function	EditAutoListFile: Longbool;
    procedure	SetDefaultFlags(AutoEncrypt, AutoSign: Longbool);
    procedure	AddressBookListCreate(const PathList: String);
    procedure	AddressBookListFree;
    function	EnabledItemListCreate: TAutoList;
    function	BalanceKeys(KeyPropsList: TKeyPropsList): Integer;
    procedure	AddKeyIDDataToNewList(const KIDD: TKeyIDData; KPL: TKeyPropsList; NewList: TAutoList);
    procedure	AddOccurencesToNewList(const Recipient: String; KPL: TKeyPropsList; NewList: TAutoList);
    function	ListItemExists(const ListItem: TALItem; var Index: Integer): Longbool;
    function	EntryExists(const TestUserID, TestHexID: String): Longbool;
    function	AddListItem(Value: TALItem): Integer;
    procedure	AddListItems(ItemList: TAutoList);
    procedure	Clear; override;
    procedure	Delete(Index: Integer); override;
    procedure	Exchange(Index1, Index2: Integer); override;
    procedure	Move(CurIndex, NewIndex: Integer); override;
    function	Add(const Value: String): Integer; override;
    function	AddObject(const Value: String; AObject: TObject): Integer; override;
    procedure	Insert(Index: Integer; const Value: String); override;
    {$IFDEF CONDITIONALEXPRESSIONS}
    procedure	InsertObject(Index: Integer; const Value: String; AObject: TObject); override;
    {$ELSE}
    procedure	InsertObject(Index: Integer; const Value: String; AObject: TObject);
    {$ENDIF}
    function 	RemoveByHexID(const HexIDCommaList: String): Integer;
    procedure	LoadFromFile(const FileName: String); override;
    procedure	SaveToFile(const FileName: String); override;
    procedure	LoadList;
    procedure	SaveList;
    property	AddressBookList: TAddressBookList read FAddressBookList;
    property	Capacity: Integer read FCapacity write SetCapacity;
    property	CommentMarker: Char read FCommentMarker write FCommentMarker;
    property	DefaultFlags: Integer read FDefaultFlags write FDefaultFlags;
    property	Duplicates: Longbool read FDuplicates write SetDuplicates;
    property	FileName: String read FFileName;
    property	SortOrder: TSortOrder read FSortOrder;
    property	Items[Index: Integer]: TALItem read GetItem write SetItem;
    property	Flags[Index: Integer]: Integer read GetFlags write SetFlags;
    property	FlagsEncrypt[Index: Integer]: Longbool read GetFlagsEncrypt write SetFlagsEncrypt;
    property	FlagsSign[Index: Integer]: Longbool read GetFlagsSign write SetFlagsSign;
    property	FlagsAsk[Index: Integer]: Longbool read GetFlagsAsk write SetFlagsAsk;
    property	FlagsPMime[Index: Integer]: Longbool read GetFlagsPMime write SetFlagsPMime;
    property	FlagsSMime[Index: Integer]: Longbool read GetFlagsSMime write SetFlagsSMime;
    property	FlagsDisabled[Index: Integer]: Longbool read GetFlagsDisabled write SetFlagsDisabled;
    property	HexIDs[Index: Integer]: String read GetHexID write SetHexID;
    property	UserIDs[Index: Integer]: String read GetUserID write SetUserID;
    property	EntryList: String read GetListStr write SetTextStr;
    property	HexIDCommaText: String read GetHexIDCommaText;
    property	UserIDText: String read GetUserIDText;
  end;

type
  TRecipientsData = Class(TAutoList)
  private
    FMixOptions: Longbool;
    FUnknownKeys: Integer;
  protected
  public
    AddressCounts: TAddressCounts;
    procedure	Clear; override;
    procedure	SetOptionFlags(Value: Integer);
    property	MixOptions: Longbool read FMixOptions write FMixOptions;
    property	UnknownKeys: Integer read FUnknownKeys write FUnknownKeys;
  end;

type
  TALInterface = Class(TAutoList)
  private
  protected
  public
    constructor	Create(const FileName: String);
    destructor	Destroy; override;
    procedure	UpdateList(PGPInterface: TPGPInterface);
  end;


implementation


// TAddressBookList --------------------------------------------------------------------------------

const
  BookNameSize = 128;
  BookEntrySize = 458;

type
  TBookName = Array[0..pred(BookNameSize)] of Char;
  TBookEntry = Array[0.. pred(BookEntrySize)] of Char;

constructor TAddressBookList.Create(const PathList: String);
var
  iLen, iBeg, iEnd: Integer;
begin
  inherited Create;
  iLen := Length(PathList);
  iBeg := 1;
  repeat
    iEnd := NextCharPos(LF, PathList, iBeg);
    if iEnd = 0 then iEnd := succ(iLen);
    if iEnd > iBeg then begin
      ReadAddressBooks(Trim(Copy(PathList, iBeg, iEnd - iBeg)));
      iBeg := succ(iEnd);
    end
    else Break;
  until iEnd > iLen;
end;

destructor TAddressBookList.Destroy;
begin
  CloseAddressBooks;
  inherited;
end;

function TAddressBookList.AddAddressBook(const FileName: String): Integer;
var
  fsBook: TFileStream;
  aBookName: TBookName;
  iLen: Integer;
begin
  Result := -1;
  try
    fsBook := TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    if fsBook <> nil then begin
      try
	with fsBook do if (Size > BookNameSize) and (Read(aBookName, BookNameSize) = BookNameSize) then begin
	  iLen := StrLen(aBookName);
	  if (iLen > 0) and (iLen < BookNameSize) then begin
	    aBookName[iLen] := EQ;
	    aBookName[succ(iLen)] := NULL;
	    Result := AddObject(aBookName, fsBook);
	  end;
	end;
      finally
	if Result = -1 then fsBook.Free;
      end;
    end;
  except
  end;
end;

function TAddressBookList.AllAddresses: String;
var
  pResult: PChar;
  iIndex, iPos, iLen: Integer;
  sAddressBook, sAddresses: String;
begin
  if Count > 0 then begin
    SetString(Result, nil, FMaxCacheSize);
    pResult := Pointer(Result);
    for iIndex := 0 to pred(Count) do begin
      if FAddressBookFlags <> E then begin
	sAddressBook := Copy(Strings[iIndex], 1, pred(Length(Strings[iIndex]))) + EXT_ADDR_BOOK + EQ;
	iPos := FirstLoStrPos(sAddressBook, FAddressBookFlags);
	if iPos > 0 then begin
	  inc(iPos, Length(sAddressBook) + 4);
	  if (iPos <= Length(FAddressBookFlags)) and (FAddressBookFlags[iPos] = '1') then Continue;
	end;
      end;
      sAddresses := GetAddressList(iIndex);
      if sAddresses <> E then begin
	iLen := Length(sAddresses);
	System.Move(Pointer(sAddresses)^, pResult^, iLen);
	inc(pResult, iLen);
      end;
    end;
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

procedure TAddressBookList.CreateAddressCache(const AddressBookFlags: String);
begin
  FOmitNullHexIDs := true;
  try
    FAddressBookFlags := AddressBookFlags;
    FAddressCache := AllAddresses;
  finally
    FOmitNullHexIDs := false;
    CloseAddressBooks;
  end;
end;

function TAddressBookList.ResolveNames(const NameCommaList: String): String;
const
  CMSP = CM + SP;
var
  iListSize, iMaxSize, iStart, iLen, iBeg, iEnd, iCopy, iContent: Integer;
  pResult: PChar;
  sAddress, sEmail: String;
begin
  if (FAddressCache <> E) and (NameCommaList <> E) then begin
    iListSize := Length(NameCommaList);
    iMaxSize := iListSize + Length(FAddressCache);
    SetString(Result, nil, iMaxSize);
    pResult := Pointer(Result);
    iStart := 1;
    while iStart <= iListSize do begin
      iLen := GetAddressLength(NameCommaList, iStart);
      if iLen > 0 then begin
	sAddress := Copy(NameCommaList, iStart, iLen);
	sEmail := ExtractEmail(sAddress);
	if sEmail = E then begin
	  sAddress := sAddress + TAB + LT;
	  if LoStrStart(sAddress, FAddressCache) then
	    iBeg := 1
	  else iBeg := FirstLoStrPos(CRLF + sAddress, FAddressCache);
	  if iBeg > 0 then begin
	    iEnd := NextStrPos(CRLF, FAddressCache, iBeg + Length(sAddress));
	    if iEnd = 0 then iEnd := succ(iListSize);
	    if iEnd > iBeg then sEmail := ExtractEmail(Trim(Copy(FAddressCache, iBeg, iEnd - iBeg)));
	  end
	  else begin
	    System.Delete(sAddress, succ(iLen), MAXINT);
	    sEmail := sAddress;
	  end;
	end;
	if sEmail <> E then begin
	  iCopy := Length(sEmail);
	  iContent := pResult - Pointer(Result);
	  if iContent + iCopy >= iMaxSize then begin
	    iMaxSize := iMaxSize shl 1;
	    SetLength(Result, iMaxSize);
	    pResult := Pointer(Result);
	    inc(pResult, iContent);
	  end;
	  System.Move(Pointer(sEmail)^, pResult^, iCopy);
	  inc(pResult, iCopy);
	  PDChar(pResult)^ := CMSP;
	  inc(pResult, Length(CMSP));
	end;
      end;
      inc(iStart, iLen);
      while (iStart <= iListSize) and (NameCommaList[iStart] <> CM) do inc(iStart);
      inc(iStart);
    end;
    System.Delete(Result, succ(pResult - Pointer(Result)) - Length(CMSP), MAXINT);
  end
  else Result := E;
end;

function TAddressBookList.GetAddressList(BookIndex: Integer): String;

  function GetName(const BookEntry: TBookEntry): String;
  var
    pName: PChar;
  begin
    pName := @BookEntry[2];
    OEMToChar(pName, pName);
    Result := DelQuotes(pName);
  end;

  function GetEmail(const aBookEntry: TBookEntry): String;
  var
    pEmail: PChar;
  begin
    pEmail := @aBookEntry[SizeOf(TBookEntry) - 116];
    OEMToChar(pEmail, pEmail);
    Result := DelQuotes(pEmail);
  end;

var
  aBookEntry: TBookEntry;
  sName, sEmail: String;
  pCurrent: PChar;
  iSize, iLen, iPos: Integer;
begin
  if (BookIndex >= 0) and (BookIndex < Count) then begin
    if Objects[BookIndex] <> nil then with TFileStream(Objects[BookIndex]) do begin
      try
	iSize := Size;
	SetString(Result, nil, iSize - Position);
	pCurrent := Pointer(Result);
	while (iSize - Position) >= BookEntrySize do begin
	  Read(aBookEntry, BookEntrySize);
	  sName := GetName(aBookEntry);
	  sEmail := GetEmail(aBookEntry);
	  iLen := Length(sName);
	  if iLen > 0 then begin
	    System.Move(Pointer(sName)^, pCurrent^, iLen);
	    inc(pCurrent, iLen);
	    if FOmitNullHexIDs then begin
	      pCurrent^ := TAB;
	      inc(pCurrent, Length(TAB));
	    end
	    else begin
	      pCurrent^ := SP;
	      inc(pCurrent, Length(SP));
	    end;
	  end;
	  iLen := Length(sEmail);
	  if iLen > 0 then begin
	    pCurrent^ := LT;
	    inc(pCurrent, Length(LT));
	    System.Move(Pointer(sEmail)^, pCurrent^, iLen);
	    inc(pCurrent, iLen);
	    if FOmitNullHexIDs then begin
	      pCurrent^ := GT;
	      inc(pCurrent, Length(GT));
	    end
	    else begin
	      PQChar(pCurrent)^ := GT + BS + HEX_MARKER;
	      inc(pCurrent, SizeOf(QChar));
	      for iPos := 0 to pred((LongHexIDLen - Length(HEX_MARKER)) div SizeOf(QChar)) do begin
		PQChar(pCurrent)^ := '0000';
		inc(pCurrent, SizeOf(QChar));
	      end;
	    end;
	    PDChar(pCurrent)^ := CRLF;
	    inc(pCurrent, Length(CRLF));
	  end;
	end;
	System.Delete(Result, succ(pCurrent - Pointer(Result)), MAXINT);
	Strings[BookIndex] := Strings[BookIndex] + Result;
      finally
	Free;
	Objects[BookIndex] := nil;
      end;
    end
    else Result := Copy(Strings[BookIndex], succ(FirstCharPos(EQ, Strings[BookIndex])), MAXINT);
  end;
end;

procedure TAddressBookList.ReadAddressBooks(const Directory: String);
var
  hSearch: THandle;
  fdFound: TWin32FindData;
begin
  if DirExists(Directory) then begin
    hSearch := FindFirst(Directory + AST + EXT_ADDR_BOOK, fdFound);
    if hSearch <> INVALID_HANDLE_VALUE then begin
      try
	repeat
	  inc(FMaxCacheSize, fdFound.nFileSizeLow);
	  AddAddressBook(Directory + fdFound.cFileName);
	until not FindNext(hSearch, fdFound);
      finally
	FindClose(hSearch);
      end;
    end;
  end;
end;

procedure TAddressBookList.CloseAddressBooks;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do begin
    if Objects[iIndex] <> nil then begin
      TFileStream(Objects[iIndex]).Free;
      Objects[iIndex] := nil;
    end;
    Strings[iIndex] := E;
  end;
  Clear;
end;


// TAutoList ---------------------------------------------------------------------------------------

constructor TAutoList.Create(const FileName: String);
begin
  inherited Create;
  FCommentMarker := DC;
  FFileName := FileName;
end;

destructor TAutoList.Destroy;
begin
  Clear;
end;

function TAutoList.EditAutoListFile: Longbool;
begin
  Result := FileExecute(0, E, PChar(FFileName), E, true);
end;

procedure TAutoList.SetDefaultFlags(AutoEncrypt, AutoSign: Longbool);
begin
  FDefaultFlags := ord(AutoEncrypt = true) * AUTO_FLAG_ENCRYPT + ord(AutoSign = true) * AUTO_FLAG_SIGN;
end;

procedure TAutoList.AddressBookListCreate(const PathList: String);
begin
  FAddressBookList := TAddressBookList.Create(PathList);
end;

procedure TAutoList.AddressBookListFree;
begin
  FAddressBookList.Free;
  FAddressBookList := nil;
end;

function TAutoList.EnabledItemListCreate: TAutoList;

  procedure ResultInsertALItem(const ALItem: TALItem; NewList: TAutoList);
  var
    iIndex: Integer;
  begin
    with NewList do begin
      if ALItem.ALFlags and AUTO_FLAG_DISABLED = 0 then begin
	UserIDFind(ALItem.UserID, iIndex);
	if iIndex < FCount then System.Move(FAutoList^[iIndex], FAutoList^[iIndex + 1], (FCount - iIndex) * SizeOf(TALItem));
	Initialize(FAutoList^[iIndex]);
	FAutoList^[iIndex] := ALItem;
	inc(FCount);
      end;
    end;
  end;

var
  iIndex: Integer;
begin
  Result := TAutoList.Create(E);
  if Result <> nil then begin
    Result.SetCapacity(FCount);
    Result.FSortOrder := soByUserID;
    for iIndex := 0 to pred(FCount) do ResultInsertALItem(FAutoList^[iIndex], Result);
  end;
end;

function TAutoList.BalanceKeys(KeyPropsList: TKeyPropsList): Integer;
var
  iIndex: Integer;
begin
  LoadList;
  iIndex := 0;
  Result := Count;
  while (iIndex < Count) do begin
    if KeyPropsList.IndexOf(HexIDs[iIndex]) = -1 then
      Delete(iIndex)
    else inc(iIndex);
  end;
  dec(Result, Count);
  if Result > 0 then SaveList;
end;

procedure TAutoList.AddKeyIDDataToNewList(const KIDD: TKeyIDData; KPL: TKeyPropsList; NewList: TAutoList);

  procedure AddItemsByUserID(const AnyID: String; var ALI: TALItem);
  begin
    ALI.UserID := PMPGPCommon.ExtractName(AnyID);
    if ALI.UserID <> E then NewList.AddListItem(ALI);
    ALI.UserID := PMPGPCommon.ExtractEmail(AnyID);
    if ALI.UserID <> E then NewList.AddListItem(ALI);
  end;

var
  aliNew: TALItem;
  iPos, iIndex, iItem, iUser, iLen, iBeg, iEnd: Integer;
begin
  if NewList <> nil then begin
    with aliNew do begin
      UserID := E;
      HexID := HEX_ID_NULL;
      ALFlags := DefaultFlags;
      ALObject := nil;
    end;
    with KIDD do begin
      if KeyCount > 0 then begin
	iPos := 0;
	repeat
	  iPos := NextStrPos(HEX_MARKER, HexIDCommaList, succ(iPos));
	  if iPos > 0 then begin
	    iIndex := KPL.IndexOf(Copy(HexIDCommaList, iPos, DefHexIDLen));
	    if iIndex > -1 then begin
	      with aliNew do begin
		HexID := KPL[iIndex];
		UserID := KPL.KeyProps[iIndex].kUserID;
	      end;
	      if ListItemExists(aliNew, iItem) then with KPL.KeyProps[iIndex].kUserIDList do begin
		for iUser := 0 to pred(Count) do begin
		  aliNew.UserID := Strings[iUser];
		  if not ListItemExists(aliNew, iItem) then NewList.AddListItem(aliNew);
		end;
	      end
	      else NewList.AddListItem(aliNew);
	    end;
	    inc(iPos, DefHexIDLen);
	  end;
	until iPos = 0;
      end
      else if HexIDCommaList <> E then begin
	iLen := Length(UserIDLineList);
	iPos := 0;
	iBeg := 1;
	repeat
	  iPos := NextStrPos(HEX_MARKER, HexIDCommaList, succ(iPos));
	  if iPos > 0 then begin
	    iIndex := KPL.IndexOf(Copy(HexIDCommaList, iPos, DefHexIDLen));
	    if iIndex > -1 then begin
	      with aliNew do begin
		HexID := KPL[iIndex];
		iEnd := NextCharPos(CR, UserIDLineList, succ(iBeg));
		if iEnd = 0 then iEnd := succ(iLen);
		if iEnd > iBeg then begin
		  AddItemsByUserID(Copy(UserIDLineList, iBeg, iEnd - iBeg), aliNew);
		  while (iEnd <= iLen) and (UserIDLineList[iEnd] < SP) do inc(iEnd);
		  iBeg := iEnd;
		end
		else Break;
	      end;
	      if (aliNew.UserID <> E) and not ListItemExists(aliNew, iItem) then NewList.AddListItem(aliNew);
	    end;
	    inc(iPos, DefHexIDLen);
	  end;
	until (aliNew.UserID = E) or (iPos = 0);
      end
      else begin
	iLen := Length(UserIDLineList);
	iBeg := 1;
	repeat
	  iEnd := NextCharPos(CR, UserIDLineList, succ(iBeg));
	  if iEnd = 0 then iEnd := succ(iLen);
	  if iEnd > iBeg then begin
	    AddItemsByUserID(Copy(UserIDLineList, iBeg, iEnd - iBeg), aliNew);
	    while (iEnd <= iLen) and (UserIDLineList[iEnd] < SP) do inc(iEnd);
	    iBeg := iEnd;
	  end
	  else Break;
	until iBeg > iLen;
      end;
    end;
  end;
end;

procedure TAutoList.AddOccurencesToNewList(const Recipient: String; KPL: TKeyPropsList; NewList: TAutoList);

  procedure NewListAddItemAndIndex(const ALItem: TALItem);
  var
    iIndex, iPos: Integer;
  begin
    iIndex := KPL.IndexOf(ALItem.HexID);
    if iIndex > -1 then with NewList do begin
      iPos := AddListItem(ALItem);
      if iPos > -1 then Objects[iPos] := Ptr(iIndex);
    end;
  end;

var
  iIndex: Integer;
  sAddress, sName: String;
begin
  if NewList <> nil then begin
    NewList.Sort(soByUserID);
    Sort(soByUserID);
    if UserIDFind(Recipient, iIndex) then begin
      repeat
	NewListAddItemAndIndex(FAutoList^[iIndex]);
	inc(iIndex);
      until (iIndex = FCount) or (AnsiCompareText(Recipient, FAutoList^[iIndex].UserID) <> 0);
    end
    else begin
      sAddress := ExtractEmail(Recipient);
      if sAddress <> E then for iIndex := 0 to pred(FCount) do begin
	if FirstLoStrPos(sAddress, FAutoList^[iIndex].UserID) > 0 then NewListAddItemAndIndex(FAutoList^[iIndex]);
      end;
      if NewList.Count > 0 then Exit;
      sName := PMPGPCommon.ExtractName(Recipient);
      if sName <> E then for iIndex := 0 to pred(FCount) do begin
	if AnsiCompareText(sName, PMPGPCommon.ExtractName(FAutoList^[iIndex].UserID)) = 0 then begin
	  NewListAddItemAndIndex(FAutoList^[iIndex]);
	end;
      end;
    end;
  end;
end;

function TAutoList.ListItemExists(const ListItem: TALItem; var Index: Integer): Longbool;
var
  sAddress: String;
  iIndex: Integer;
begin
  Result := false;
  if not FDuplicates then begin
    if FSortOrder <> soByHexID then Sort(soByHexID);
    if not HexIDFind(ListItem.HexID, Index) then Exit;
    iIndex := Index;
    if IsEmailAddress(ListItem.UserID) then sAddress := ListItem.UserID;
    while iIndex < FCount do with FAutoList^[iIndex] do begin
      if CompareText(ListItem.HexID, HexID) <> 0 then Exit;
      if (AnsiCompareText(ListItem.UserID, UserID) = 0)
      or ((sAddress <> E) and (CompareText(sAddress, ExtractEmail(UserID)) = 0)) then begin
	Index := iIndex;
	Result := true;
	Exit;
      end;
      inc(iIndex);
    end;
  end
  else Index := FCount;
end;

function TAutoList.EntryExists(const TestUserID, TestHexID: String): Longbool;
var
  aliTest: TALItem;
  iIndex: Integer;
begin
  Result := false;
  if (TestUserID <> E) and (TestHexID <> E) then begin
    with aliTest do begin
      UserID := TestUserID;
      HexID := TestHexID;
      ALFlags := 0;
      ALObject := nil;
    end;
    Result := ListItemExists(aliTest, iIndex);
  end;
end;

function TAutoList.AddListItem(Value: TALItem): Integer;
begin
  Result := InsertALItem(FCount, Value);
end;

procedure TAutoList.AddListItems(ItemList: TAutoList);
var
  iIndex: Integer;
begin
  with ItemList do for iIndex := 0 to pred(Count) do Self.AddListItem(Items[iIndex]);
end;

procedure TAutoList.Clear;
begin
  if FCount > 0 then begin
    Finalize(FAutoList^[0], FCount);
    SetCapacity(0);
    FCount := 0;
  end;
end;

procedure TAutoList.Delete(Index: Integer);
begin
  if (Index >= 0) and (Index < FCount) then begin
    Finalize(FAutoList^[Index]);
    dec(FCount);
    if Index < FCount then System.Move(FAutoList^[Index + 1], FAutoList^[Index], (FCount - Index) * SizeOf(TALItem));
  end;
end;

procedure TAutoList.Exchange(Index1, Index2: Integer);
begin
  if (Index1 < 0) or (Index1 >= FCount) or (Index2 < 0) or (Index2 >= FCount) then Exit;
  ExchangeItems(Index1, Index2);
end;

procedure TAutoList.Move(CurIndex, NewIndex: Integer);
var
  TempString: String;
begin
  if CurIndex <> NewIndex then begin
    TempString := Get(CurIndex);
    Delete(CurIndex);
    InsertItem(NewIndex, TempString);
  end;
end;

function TAutoList.Add(const Value: String): Integer;
begin
  Result := InsertItem(FCount, Value);
end;

function TAutoList.AddObject(const Value: String; AObject: TObject): Integer;
begin
  Result := InsertItem(FCount, Value);
  if Result > -1 then PutObject(Result, AObject);
end;

procedure TAutoList.Insert(Index: Integer; const Value: String);
begin
  InsertItem(Index, Value);
end;

procedure TAutoList.InsertObject(Index: Integer; const Value: String; AObject: TObject);
begin
  if InsertItem(Index, Value) > -1 then PutObject(Index, AObject);
end;

function TAutoList.RemoveByHexID(const HexIDCommaList: String): Integer;
var
  iPos, iIndex: Integer;
begin
  iPos := 0;
  Result := Count;
  try
    Sort(soByHexID);
    repeat
      iPos := NextStrPos(HEX_MARKER, HexIDCommaList, succ(iPos));
      if iPos > 0 then begin
	if HexIDFind(Copy(HexIDCommaList, iPos, DefHexIDLen), iIndex) then Delete(iIndex);
	inc(iPos, DefHexIDLen);
      end;
    until iPos = 0;
  finally
    dec(Result, Count);
  end;
  SaveToFile(FileName);
end;

procedure TAutoList.LoadFromFile(const FileName: String);
var
  iFileAge: Integer;
begin
  iFileAge := FileAge(FFileName);
  if iFileAge > FLastChecked then
    inherited LoadFromFile(FFileName)
  else if iFileAge = -1 then Clear;
  FLastChecked := iFileAge;
end;

procedure TAutoList.SaveToFile(const FileName: String);
begin
  inherited SaveToFile(FFileName);
  FLastChecked := FileAge(FFileName);
end;

procedure TAutoList.LoadList;
begin
  LoadFromFile(FFileName);
end;

procedure TAutoList.SaveList;
begin
  SaveToFile(FFileName);
end;

function TAutoList.HexIDFind(const Item: String; var Index: Integer): Longbool;
var
  iLow, iHigh, iIndex, iCompare: Integer;
begin
  Result := false;
  iLow := 0;
  iHigh := pred(FCount);
  while iLow <= iHigh do begin
    iIndex := (iLow + iHigh) shr 1;
    iCompare := SysUtils.CompareText(FAutoList^[iIndex].HexID, Item);
    if iCompare < 0 then
      iLow := succ(iIndex)
    else begin
      iHigh := pred(iIndex);
      if iCompare = 0 then Result := true;
    end;
  end;
  Index := iLow;
end;

procedure TAutoList.HexIDSort(FirstItem, LastItem: Integer);
var
  FirstIndex, LastIndex, MidIndex: Integer;
begin
  BeginUpdate;
  repeat
    FirstIndex := FirstItem;
    LastIndex := LastItem;
    MidIndex := (FirstItem + LastItem) shr 1;
    repeat
      while SysUtils.CompareText(FAutoList^[FirstIndex].HexID, FAutoList^[MidIndex].HexID) < 0 do inc(FirstIndex);
      while SysUtils.CompareText(FAutoList^[LastIndex].HexID, FAutoList^[MidIndex].HexID) > 0 do dec(LastIndex);
      if FirstIndex <= LastIndex then begin
	Exchange(FirstIndex, LastIndex);
	if MidIndex = FirstIndex then
	  MidIndex := LastIndex
	else if MidIndex = LastIndex then MidIndex := FirstIndex;
	inc(FirstIndex);
	dec(LastIndex);
      end;
    until FirstIndex > LastIndex;
    if FirstItem < LastIndex then HexIDSort(FirstItem, LastIndex);
    FirstItem := FirstIndex;
  until FirstIndex >= LastItem;
end;

function TAutoList.UserIDFind(const Item: String; var Index: Integer): Longbool;
var
  iLow, iHigh, iIndex, iCompare: Integer;
begin
  Result := false;
  iLow := 0;
  iHigh := pred(FCount);
  while iLow <= iHigh do begin
    iIndex := (iLow + iHigh) shr 1;
    iCompare := SysUtils.AnsiCompareText(FAutoList^[iIndex].UserID, Item);
    if iCompare < 0 then
      iLow := succ(iIndex)
    else begin
      iHigh := pred(iIndex);
      if iCompare = 0 then Result := true;
    end;
  end;
  Index := iLow;
end;

procedure TAutoList.UserIDSort(FirstItem, LastItem: Integer);
var
  FirstIndex, LastIndex, MidIndex: Integer;
begin
  BeginUpdate;
  repeat
    FirstIndex := FirstItem;
    LastIndex := LastItem;
    MidIndex := (FirstItem + LastItem) shr 1;
    repeat
      while SysUtils.AnsiCompareText(FAutoList^[FirstIndex].UserID, FAutoList^[MidIndex].UserID) < 0 do inc(FirstIndex);
      while SysUtils.AnsiCompareText(FAutoList^[LastIndex].UserID, FAutoList^[MidIndex].UserID) > 0 do dec(LastIndex);
      if FirstIndex <= LastIndex then begin
	Exchange(FirstIndex, LastIndex);
	if MidIndex = FirstIndex then
	  MidIndex := LastIndex
	else if MidIndex = LastIndex then MidIndex := FirstIndex;
	inc(FirstIndex);
	dec(LastIndex);
      end;
    until FirstIndex > LastIndex;
    if FirstItem < LastIndex then UserIDSort(FirstItem, LastIndex);
    FirstItem := FirstIndex;
  until FirstIndex >= LastItem;
end;

function TAutoList.Get(Index: Integer): String;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := RestoreAutoListEntry(FAutoList^[Index])
  else Result := E;
end;

procedure TAutoList.Put(Index: Integer; const Value: String);
var
  aliNew: TALItem;
begin
  if (Index >= 0) and (Index < FCount) and ExtractAutoListData(Value, aliNew) then FAutoList^[Index] := aliNew;
end;

function TAutoList.GetObject(Index: Integer): TObject;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := FAutoList^[Index].ALObject
  else Result := nil;
end;

procedure TAutoList.PutObject(Index: Integer; AObject: TObject);
begin
  if (Index >= 0) and (Index < FCount) then FAutoList^[Index].ALObject := AObject;
end;

function TAutoList.GetCount: Integer;
begin
  Result := FCount;
end;

function TAutoList.GetItem(Index: Integer): TALItem;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := FAutoList^[Index]
  else FillByte(Result, SizeOf(TALItem), 0);
end;

procedure TAutoList.SetItem(Index: Integer; Value: TALItem);
begin
  if (Index >= 0) and (Index < FCount) then FAutoList^[Index] := Value;
end;

function TAutoList.GetFlags(Index: Integer): Integer;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := FAutoList^[Index].ALFlags
  else Result := 0;
end;

procedure TAutoList.SetFlags(Index: Integer; Value: Integer);
begin
  if (Index >= 0) and (Index < FCount) then FAutoList^[Index].ALFlags := Value;
end;

function TAutoList.GetFlagsEncrypt(Index: Integer): Longbool;
begin
  Result := (GetFlags(Index) and AUTO_FLAG_ENCRYPT) <> 0;
end;

procedure TAutoList.SetFlagsEncrypt(Index: Integer; Value: Longbool);
begin
  SetFlags(Index, GetFlags(Index) and not AUTO_FLAG_ENCRYPT or AUTO_FLAG_ENCRYPT);
end;

procedure TAutoList.SetFlagsSign(Index: Integer; Value: Longbool);
begin
  SetFlags(Index, GetFlags(Index) and not AUTO_FLAG_SIGN or AUTO_FLAG_SIGN);
end;

procedure TAutoList.SetFlagsAsk(Index: Integer; Value: Longbool);
begin
  SetFlags(Index, GetFlags(Index) and not AUTO_FLAG_ASK or AUTO_FLAG_ASK);
end;

procedure TAutoList.SetFlagsPMime(Index: Integer; Value: Longbool);
begin
  SetFlags(Index, GetFlags(Index) and not AUTO_FLAG_PMIME or AUTO_FLAG_PMIME);
end;

procedure TAutoList.SetFlagsSMime(Index: Integer; Value: Longbool);
begin
  SetFlags(Index, GetFlags(Index) and not AUTO_FLAG_SMIME or AUTO_FLAG_SMIME);
end;

procedure TAutoList.SetFlagsDisabled(Index: Integer; Value: Longbool);
begin
  SetFlags(Index, GetFlags(Index) and not AUTO_FLAG_DISABLED or AUTO_FLAG_DISABLED);
end;

function TAutoList.GetFlagsSign(Index: Integer): Longbool;
begin
  Result := (GetFlags(Index) and AUTO_FLAG_SIGN) <> 0;
end;

function TAutoList.GetFlagsAsk(Index: Integer): Longbool;
begin
  Result := (GetFlags(Index) and AUTO_FLAG_ASK) <> 0;
end;

function TAutoList.GetFlagsPMime(Index: Integer): Longbool;
begin
  Result := (GetFlags(Index) and AUTO_FLAG_PMIME) <> 0;
end;

function TAutoList.GetFlagsSMime(Index: Integer): Longbool;
begin
  Result := (GetFlags(Index) and AUTO_FLAG_SMIME) <> 0;
end;

function TAutoList.GetFlagsDisabled(Index: Integer): Longbool;
begin
  Result := (GetFlags(Index) and AUTO_FLAG_DISABLED) <> 0;
end;

function TAutoList.GetHexID(Index: Integer): String;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := FAutoList^[Index].HexID
  else Result := E;
end;

procedure TAutoList.SetHexID(Index: Integer; const Value: String);
begin
  if (Index >= 0) and (Index < FCount) then FAutoList^[Index].HexID := Value;
end;

function TAutoList.GetUserID(Index: Integer): String;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := FAutoList^[Index].UserID
  else Result := E;
end;

procedure TAutoList.SetUserID(Index: Integer; const Value: String);
begin
  if (Index >= 0) and (Index < FCount) then FAutoList^[Index].UserID := Value;
end;

function TAutoList.GetHexIDCommaText: String;
var
  iIndex, iLen: Integer;
  pResult: PChar;
  sHexID: String;
begin
  SetString(Result, nil, FCount * (DefHexIDLen + Length(CM)));
  pResult := Pointer(Result);
  for iIndex := 0 to pred(FCount) do begin
    sHexID := FAutoList^[iIndex].HexID;
    iLen := Length(sHexID);
    if iLen > 0 then begin
      System.Move(Pointer(sHexID)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    PDChar(pResult)^ := CM;
    inc(pResult);
  end;
  System.Delete(Result, succ(pResult - Pointer(Result) - Length(CM)), MAXINT);
end;

function TAutoList.GetUserIDText: String;
var
  iIndex, iLen: Integer;
  pResult: PChar;
  sUserID: String;
begin
  SetString(Result, nil, FCount * (MaxUserIDLen + Length(CRLF)));
  pResult := Pointer(Result);
  for iIndex := 0 to pred(FCount) do begin
    sUserID := FAutoList^[iIndex].UserID;
    iLen := Length(sUserID);
    if iLen > 0 then begin
      System.Move(Pointer(sUserID)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    PDChar(pResult)^ := CRLF;
    inc(pResult, Length(CRLF));
  end;
  System.Delete(Result, succ(pResult - Pointer(Result)) - Length(CRLF), MAXINT);
end;

procedure TAutoList.SetCapacity(NewCapacity: Integer);
begin
  ReallocMem(FAutoList, NewCapacity * SizeOf(TALItem));
  FCapacity := NewCapacity;
end;

function TAutoList.GetTextStr: String;
begin
  Result := FListHeader + GetListStr;
end;

procedure TAutoList.SetTextStr(const Value: String);
var
  pPosition, pStart: PChar;
  iPos, iLen: Integer;
  sLine: String;
begin
  BeginUpdate;
  try
    Clear;
    pPosition := Pointer(Value);
    if pPosition <> nil then begin
      if StrStart(BEG_COMMENT_BLOCK, Value) then begin
	iPos := NextStrPos(END_COMMENT_BLOCK, Value, succ(FirstCharPos(LF, Value)));
	if iPos > 0 then begin
	  iLen := Length(Value);
	  inc(iPos, Length(END_COMMENT_BLOCK));
	  while (iPos <= iLen) and (Value[iPos] <= SP) do inc(iPos);
	  dec(iPos);
	  inc(pPosition, iPos);
	  FListHeader := Copy(Value, 1, iPos);
	end;
      end;
      while pPosition^ <> NULL do begin
	while pPosition^ in [#1..#32] do inc(pPosition);
	if (FCommentMarker <> NULL) and (pPosition^ = FCommentMarker) then begin
	  while not (pPosition^ in [NULL, LF, CR]) do inc(pPosition);
	  Continue;
	end;
	pStart := pPosition;
	while not (pPosition^ in [NULL, LF, CR]) do inc(pPosition);
	iLen := pPosition - pStart;
	if iLen > 0 then begin
	  SetString(sLine, pStart, iLen);
	  Add(sLine);
	end;
      end;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TAutoList.Sort(Value: TSortOrder);
begin
  if not FDuplicates then begin
    if (Value <> FSortOrder) and (FCount > 0) then begin
      case Value of
	soByHexID: HexIDSort(0, pred(FCount));
	soByUserID: UserIDSort(0, pred(FCount));
      end;
    end;
    FSortOrder := Value;
  end;
end;

function TAutoList.GetListStr: String;
var
  iIndex, iLen: Integer;
  pResult: PChar;
  sLine: String;
begin
  SetString(Result, nil, FCount * MAX_LIST_LINE_LEN);
  pResult := Pointer(Result);
  for iIndex := 0 to pred(FCount) do begin
    sLine := Get(iIndex);
    iLen := Length(sLine);
    if iLen > 0 then begin
      System.Move(Pointer(sLine)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    PDChar(pResult)^ := CRLF;
    inc(pResult, Length(CRLF));
  end;
  System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
end;

function TAutoList.GetListVersion: Integer;
var
  iPos: Integer;
  sVersionLine: String;
begin
  iPos := FirstLoStrPos(CRLF + UPDATE_HEADER, FListHeader);
  if iPos > 0 then inc(iPos, Length(CRLF) + Length(UPDATE_HEADER));
  Result := VersionNumberFromString(Copy(FListHeader, iPos, NextCharPos(SP, FListHeader, succ(iPos)) - iPos));
  if iPos = 0 then begin
    iPos := FirstStrPos(LF, FListHeader);
    sVersionLine := Copy(FListHeader, 1, iPos);
    sVersionLine := XChgStrAtPos(sVersionLine, UPDATE_HEADER + '0', 1, succ(Length(UPDATE_HEADER)));
    XChgChars(sVersionLine, '+', SP);
    System.Insert(sVersionLine, FListHeader, succ(iPos));
  end;
end;

procedure TAutoList.SetListVersion(Value: Integer);
var
  iPos: Integer;
  sVersion: String;
begin
  iPos := FirstLoStrPos(CRLF + UPDATE_HEADER, FListHeader);
  if iPos > 0 then inc(iPos, Length(CRLF) + Length(UPDATE_HEADER));
  sVersion := VersionStringFromNumber(Value);
  FListHeader := XChgStrAtPos(FListHeader, sVersion, iPos, Length(sVersion));
end;

procedure TAutoList.SetDuplicates(Value: Longbool);
begin
  FDuplicates := Value;
  if FDuplicates then FSortOrder := soNone;
end;

function TAutoList.ExtractAutoListData(const AutoListEntry: String; var ListItem: TALItem): Longbool;
var
  iPos: Integer;
begin
  Result := false;
  FillByte(ListItem, SizeOf(TALItem), 0);
  iPos := Length(AutoListEntry) - (DefHexIDLen + FLAG_SIZE);
  if (iPos <= 0) or (AutoListEntry[iPos] <> BS) then iPos := LastCharPos(BS, AutoListEntry);
  if iPos > 0 then with ListItem do begin
    UserID := Trim(Copy(AutoListEntry, 1, pred(iPos)));
    HexID := Trim(Copy(AutoListEntry, succ(iPos), DefHexIDLen));
    if (UserID <> E) and (Length(HexID) >= ShortHexIDLen) then begin
      iPos := Length(AutoListEntry);
      if not (AutoListEntry[iPos] in [FLAG_ENABLED, FLAG_DISABLED]) then begin
	ALFlags := ord(AutoListEntry[iPos] = FLAG_USE_SMIME) * AUTO_FLAG_SMIME;
	dec(iPos);
      end;
      inc(ALFlags, HexToDigit(HexID[1]) + ord(AutoListEntry[iPos] = FLAG_DISABLED) * AUTO_FLAG_DISABLED);
      PChar(Pointer(ListItem.HexID))[0] := '0';
      Result := true;
    end;
  end;
end;

function TAutoList.RestoreAutoListEntry(const ListItem: TALItem): String;
var
  cSMimeFlag: Char;
  iPos: Integer;
begin
  with ListItem do begin
    Result := UserID + BS + HexID;
    iPos := Length(Result) - Length(HexID);
    if iPos > 0 then begin
      PChar(Pointer(Result))[iPos] := DigitToHex(ALFlags and $F);
      cSMimeFlag := chr(ord(FLAG_USE_SMIME) or ($20 * ord(ALFlags and AUTO_FLAG_SMIME = 0)));
      if (ALFlags and AUTO_FLAG_DISABLED) = 0 then
	Result := Result + FLAG_ENABLED + cSMimeFlag
      else Result := Result + FLAG_DISABLED + cSMimeFlag;
    end;
  end;
end;

function TAutoList.InsertALItem(Index: Integer; const ListItem: TALItem): Integer;
begin
  if not ListItemExists(ListItem, Result) then begin
    if FCount = FCapacity then Grow;
    if Result < FCount then System.Move(FAutoList^[Result], FAutoList^[Result + 1], (FCount - Result) * SizeOf(TALItem));
    Initialize(FAutoList^[Result]);
    FAutoList^[Result] := ListItem;
    inc(FCount);
  end
  else Result := -1;
end;

function TAutoList.InsertItem(Index: Integer; const Value: String): Integer;
var
  aliNew: TALItem;
begin
  if ExtractAutoListData(Value, aliNew) then
    Result := InsertALItem(Index, aliNew)
  else Result := -1;
end;

procedure TAutoList.ExchangeItems(Index1, Index2: Integer);
var
  aliBuffer: TALItem;
begin
  aliBuffer := FAutoList^[Index1];
  FAutoList^[Index1] := FAutoList^[Index2];
  FAutoList^[Index2] := aliBuffer;
end;

procedure TAutoList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 8 then
    Delta := 16
  else if FCapacity > 4 then
    Delta := 8
  else Delta := 4;
  SetCapacity(FCapacity + Delta);
end;


// TRecipientsData -------------------------------------------------------------------------------

procedure TRecipientsData.Clear;
begin
  FDefaultFlags := AUTO_FLAG_NOTHING;
  with AddressCounts do begin
    ToRecipients := 0;
    CcRecipients := 0;
    BccRecipients := 0;
  end;
  FMixOptions := false;
  FUnknownKeys := 0;
  inherited Clear;
end;

procedure TRecipientsData.SetOptionFlags(Value: Integer);
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do if HexIDs[iIndex] <> HEX_ID_NULL then begin
    Flags[iIndex] := (Flags[iIndex] and AUTO_FLAG_MIME_FILTER) or (Value and AUTO_FLAG_BASE_FILTER);
  end;
end;


// TALInterface ------------------------------------------------------------------------------

constructor TALInterface.Create(const FileName: String);
var
  hAutoList: THandle;
begin
  hAutoList := FileOpen(FileName, fmOpenReadWrite);
  if hAutoList = INVALID_HANDLE_VALUE then hAutoList := FileCreate(FileName);
  if hAutoList <> INVALID_HANDLE_VALUE then begin
    FileClose(hAutoList);
    inherited Create(FileName);
    SetCapacity(FileSize(FileName) div (MAX_LIST_LINE_LEN shr 2));
    FSortOrder := soByHexID;
    LoadFromFile(FileName);
  end;
end;

destructor TALInterface.Destroy;
begin
  inherited;
end;

procedure TALInterface.UpdateList(PGPInterface: TPGPInterface);
var
  sShortHexIDs: String;
  pShortHexIDs: PChar;
  iVersion, iIndex: Integer;

  procedure ShortHexIDToLongHexID;
  var
    iPos: Integer;
  begin
    iPos := FirstStrPos(HexIDs[iIndex], sShortHexIDs);
    if iPos > 0 then HexIDs[iIndex] := PGPInterface.KeyPropsList[iPos div ShortHexIDLen];
  end;

  procedure Update4550;
  begin
    if Flags[iIndex] = 0 then Flags[iIndex] := DefaultFlags;
  end;

  procedure Update4960;
  begin
    if FlagsPMime[iIndex] then Flags[iIndex] := (Flags[iIndex] and not AUTO_FLAG_PMIME) or AUTO_FLAG_DISABLED;
  end;

begin
  iVersion := GetListVersion;
  if iVersion < 4960 then begin
    try
      if iVersion < 4850 then begin
	PGPInterface.GetAutoListIDs;
	with PGPInterface.KeyPropsList do begin
	  SetString(sShortHexIDs, nil, Count * ShortHexIDLen);
	  pShortHexIDs := Pointer(sShortHexIDs);
	  for iIndex := 0 to pred(Count) do begin
	    System.Move(Pointer(KeyProps[iIndex].kHexID)^, pShortHexIDs^, ShortHexIDLen);
	    inc(pShortHexIDs, ShortHexIDLen);
	  end;
	end;
      end;
      case iVersion of
	0000..4549: for iIndex := 0 to pred(Count) do begin
	  ShortHexIDToLongHexID;
	  Update4550;
	  Update4960;
	end;
	4550..4849: for iIndex := 0 to pred(Count) do begin
	  ShortHexIDToLongHexID;
	  Update4960;
	end;
	4850..4959: for iIndex := 0 to pred(Count) do begin
	  Update4960;
	end;
      else
	Exit;
      end;
      SetListVersion(4960);
      SaveToFile(FileName);
    finally
      PGPInterface.KeyPropsList.Clear;
    end;
  end;
end;


end.

