{$J+,Z4}
unit ImageViewer;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, Classes, SysUtils, StrUtil, Graphics, Forms, Controls, ComCtrls, ExtCtrls, Menus,
  ExtCtrlsEx, ColGrip, GdiPlus, UxThemes, FileUtils, MailUtils, WindowUtils, PMPGPForm, PMPGPCommon, PMPGPResources;

type
  TImageViewer = class(TPMPGPForm)
    pnTop: TPanel;
    pnLeft: TPanel;
    pnImage: TColoredFramePanel;
    imgViewer: TImage;
    pnRight: TPanel;
    cgImageViewer: TColoredGrip;
    pmImage: TPopupMenu;
    miDecrease: TMenuItem;
    miIncrease: TMenuItem;
    miRotateLeft: TMenuItem;
    miRotateRight: TMenuItem;
    miSeparator1: TMenuItem;
    miCopy: TMenuItem;
    miSave: TMenuItem;
    miPrint: TMenuItem;
    miSeparator2: TMenuItem;
    miClose: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormDestroy(Sender: TObject);
    procedure cgImageViewerResize(Sender: TObject);
    procedure miOpenClick(Sender: TObject);
    procedure miClick(Sender: TObject);
    procedure miCopyClick(Sender: TObject);
    procedure miSaveClick(Sender: TObject);
    procedure miPrintClick(Sender: TObject);
    procedure miCloseClick(Sender: TObject);
    procedure miDecreaseClick(Sender: TObject);
    procedure miIncreaseClick(Sender: TObject);
    procedure miRotateLeftClick(Sender: TObject);
    procedure miRotateRightClick(Sender: TObject);
    procedure AllMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  private
    FileName: String;
    GDIPI: TGdipImage;
    Frames: TSize;
    MinSize: TSize;
    Scalings: TSize;
    procedure DrawGdipImage;
    procedure FormatScreen;
    procedure ResizeForm(Decrease: Longbool);
    procedure RotateForm(ToLeft: Longbool);
    function  TempFileSave(const ImageData: String): String;
    procedure ShowPopupMenu(const TrackPos: TPoint; Alignment: TPopupAlignment);
    procedure OnAdvancedCommand(const FileName: String; var fCancelCommand: Longbool);
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMGetMinMaxInfo(var Message: TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
  protected
    procedure SetColors(Default: Longbool); override;
  public
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  end;

function  IsGdipImage(const FileExt: String): Longbool;
procedure CopyGdipImage(const ImageData: String);
procedure ShowGdipImage(MainClass: TObject; ImageData: TSectionNode; SectionIcon: TIcon);


implementation

{$R *.DFM}

uses
  MessageDialogs;

function IsGdipImage(const FileExt: String): Longbool;
begin
  Result := (Length(FileExt) > 3) and (FirstLoStrPos(FileExt, ImagePreviewFileExtensions) <> 0);
end;

procedure CopyGdipImage(const ImageData: String);
var
  gdipi: TGdipImage;
begin
  gdipi := TGdipImage.Create;
  if gdipi <> nil then with gdipi do begin
    try
      if LoadFromStrBuffer(ImageData) then CopyImage;
    finally
      Free;
    end;
  end;
end;

procedure ShowGdipImage(MainClass: TObject; ImageData: TSectionNode; SectionIcon: TIcon);
var
  frmImageViewer: TImageViewer;
begin
  frmImageViewer := TImageViewer.Create(MainClass);
  if frmImageViewer <> nil then with frmImageViewer do begin
    try
      GDIPI.OriginalFileName := ImageData.FileName;
      GDIPI.LoadFromStrBuffer(ImageData.SectionData);
      FileName := ImageData.FileName;
      Caption := Utf8Check(ExtractFileName(FileName));
      Icon := SectionIcon;
      ShowModal;
    finally
      Release;
    end;
  end;
end;

procedure TImageViewer.FormCreate(Sender: TObject);
begin
  GDIPI := TGdipImage.Create;
  Caption := PMPGPStrings[IMAGE_VIEWER];
  pnTop.Caption := PMPGPStrings[IMAGE_MENU_HINT];
  miCopy.Caption := PMPGPStrings[COPY_ITEM];
  miSave.Caption := PMPGPStrings[SAVE_ITEM];
  miPrint.Caption := PMPGPStrings[PRINT_ITEM];
  miClose.Caption := PMPGPStrings[CLOSE_ITEM];
  miDecrease.Caption := PMPGPStrings[DECREASE_ITEM] + TAB + '-';
  miIncrease.Caption := PMPGPStrings[INCREASE_ITEM] + TAB + '+';
  miRotateLeft.Caption := PMPGPStrings[ROTATE_LEFT_ITEM]  + TAB + '<-';
  miRotateRight.Caption := PMPGPStrings[ROTATE_RIGHT_ITEM] + TAB + '->';
end;

procedure TImageViewer.FormShow(Sender: TObject);
begin
  DrawGdipImage;
  FormatScreen;
end;

procedure TImageViewer.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE: Close;
    VK_LEFT: miRotateLeft.Click;
    VK_UP: miIncrease.Click;
    VK_RIGHT: miRotateRight.Click;
    VK_DOWN: miDecrease.Click;
    ord('C'): if Shift = [ssCtrl] then miCopy.Click;
    ord('P'): if Shift = [ssCtrl] then miPrint.Click;
    ord('S'): if Shift = [ssCtrl] then miSave.Click;
  else
    if (Key = VK_APPS) or ((Shift = [ssShift]) and (Key = VK_F10)) then begin
      ShowPopupMenu(Point(Left + Width div 2, Top + Height div 2), paCenter);
    end;
  end;
end;

procedure TImageViewer.FormKeyPress(Sender: TObject; var Key: Char);
begin
  case Key of
    '+': miIncrease.Click;
    '-': miDecrease.Click;
  end;
end;

procedure TImageViewer.FormDestroy(Sender: TObject);
begin
  GDIPI.Free;
end;

procedure TImageViewer.cgImageViewerResize(Sender: TObject);
begin
  pnTop.Height := cgImageViewer.Height;
  pnLeft.Width := cgImageViewer.Height;
  pnRight.Width := cgImageViewer.Height;
end;

procedure TImageViewer.miOpenClick(Sender: TObject);
begin
  FileExecute(0, E, TempFileSave(GDIPI.RawData), E, true);
end;

procedure TImageViewer.miClick(Sender: TObject);
begin
  FileExecute(0, TMenuItem(Sender).Hint, TempFileSave(GDIPI.RawData), E, true);
end;

procedure TImageViewer.miCopyClick(Sender: TObject);
begin
  WriteClipboardData(Longint(imgViewer.Picture.Bitmap), CF_BITMAP, Self.Handle);
end;

procedure TImageViewer.miSaveClick(Sender: TObject);
var
  sFileName: String;
begin
  sFileName := ImageFileSave(FileName, PMPGPStrings[SAVE_FILE_DIALOG], ImageConvertFileExtensions, Self.Handle);
  if sFileName <> E then begin
    if GDIPI.SaveToFile(sFileName) then begin
      Caption := ExtractFileName(sFileName);
      GDIPI.LoadFromFile(sFileName);
      FileName := sFileName;
      DrawGdipImage;
    end
    else Error(PMPGPStrings[INVALID_FILE_FORMAT]);
  end;
end;

procedure TImageViewer.miPrintClick(Sender: TObject);
var
  pdBitmap: TBitmapPrinterData;
begin
  FillByte(pdBitmap, SizeOf(TBitmapPrinterData), 0);
  with pdBitmap, PD do begin
    OwnerWindow := Self.Handle;
    PrintJobName := Caption;
    Bitmap := imgViewer.Picture.Bitmap;
    PrinterPrefs := MainClass.PrinterPrefs;
    PrefsCaptions := MainClass.PrefsCaptions;
  end;
  with pdBitmap.PD.PrefsCaptions do begin
    HeadFootGroup := E;
    CustomOption := E;
  end;
  GDIPI.PrintImage(pdBitmap);
  MainClass.PrinterPrefs := pdBitmap.PD.PrinterPrefs;
end;

procedure TImageViewer.miCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TImageViewer.miDecreaseClick(Sender: TObject);
begin
  ResizeForm(true);
end;

procedure TImageViewer.miIncreaseClick(Sender: TObject);
begin
  ResizeForm(false);
end;

procedure TImageViewer.miRotateLeftClick(Sender: TObject);
begin
  RotateForm(true);
end;

procedure TImageViewer.miRotateRightClick(Sender: TObject);
begin
  RotateForm(false);
end;

procedure TImageViewer.AllMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbRight then ShowPopupMenu(Point(-1, -1), paLeft);
end;

procedure TImageViewer.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  with MinSize do if (AWidth >= cx) and (AHeight >= cy) then
    inherited
  else Beep;
end;

procedure TImageViewer.SetColors(Default: Longbool);
begin
  inherited;
  imgViewer.Canvas.Brush.Color := Color;
end;

procedure TImageViewer.DrawGdipImage;
begin
  with imgViewer do GDIPI.DrawImage(Picture.Bitmap);
end;

procedure TImageViewer.FormatScreen;
var
  iWidth, iHeight: Integer;
begin
  Application.ProcessMessages;
  if GDIPI <> nil then with GDIPI do begin
    with Frames do begin
      cx := Width - imgViewer.Width;
      cy := Height - imgViewer.Height;
      iWidth := ImageWidth + cx;
      iHeight := ImageHeight + cy;
    end;
    with MinSize do begin
      if iWidth < cx then iWidth := cx;
      if iHeight < cy then iHeight := cy;
    end;
    SetBounds(Left + (Width - iWidth) div 2, Top + (Height - iHeight) div 2, iWidth, iHeight);
    Scalings := TSize(Point(Width div 4, Height div 4));
    with imgViewer do begin
      Picture.Bitmap.Width := ClientWidth;
      Picture.Bitmap.Height := ClientHeight;
    end;
  end;
end;

procedure TImageViewer.ResizeForm(Decrease: Longbool);
var
  rOffsets, rNewBounds: TRect;
begin
  with Scalings do begin
    if Decrease then
      rOffsets := Rect(cx div 2, cy div 2, -cx, -cy)
    else rOffsets := Rect(-cx div 2, -cy div 2, cx, cy);
    with rOffsets do begin
      if (cx > cy) and (Width > Height) then
	rNewBounds := Rect(Self.Left + Left, Self.Top + Top, Width + Right, Height + Bottom)
      else rNewBounds := Rect(Self.Left + Top, Self.Top + Left, Width + Bottom, Height + Right);
    end;
  end;
  with rNewBounds do begin
    if Decrease then begin
      if (Right < Frames.cx) or (Bottom < Frames.cy) or (Right < MinSize.cx) or (Bottom < MinSize.cy) then begin
	Beep;
	Exit;
      end;
    end
    else begin
      if (Left <= 0) or (Top <= 0) or (Right >= Screen.Width) or (Bottom >= Screen.Height) then begin
	Beep;
	Exit;
      end;
    end;
    SetBounds(Left, Top, Right, Bottom)
  end;
end;

procedure TImageViewer.RotateForm(ToLeft: Longbool);
var
  iWidth, iHeight: Integer;
begin
  with Frames, imgViewer do begin
    iWidth := Height + cx;
    iHeight := Width + cy;
  end;
  with MinSize do if (iWidth >= cx) and (iHeight >= cy) then begin
    GDIPI.RotateImage(imgViewer.Picture.Bitmap, ToLeft);
    SetBounds(Left + (Width - iWidth) div 2, Top + (Height - iHeight) div 2, iWidth, iHeight);
  end
  else Beep;
end;

function TImageViewer.TempFileSave(const ImageData: String): String;
begin
  Result := UniqueFile(GetTempDir + FileName);
  if ImageData = E then
    FileClose(FileCreate(Result))
  else if not WriteTextToFile(Result, ImageData) then begin
    Error(PMPGPStrings[DATA_WRITE_ERROR]);
    Result := E;
  end;
end;

procedure TImageViewer.ShowPopupMenu(const TrackPos: TPoint; Alignment: TPopupAlignment);
begin
  CreateFileContextMenu(TempFileSave(E), PMPGPStrings[ADVANCED_ITEM], Handle, pmImage.Handle,
			miCopy.MenuIndex, MIN_CONTEXT_MENU_ID, @TrackPos, Alignment, OnAdvancedCommand);
end;

procedure TImageViewer.OnAdvancedCommand(const FileName: String; var fCancelCommand: Longbool);
begin
  fCancelCommand := (TempFileSave(GDIPI.RawData) = E)
end;

procedure TImageViewer.WMCommand(var Message: TWMCommand);
begin
  with Message do if (Ctl = 0) and (NotifyCode = 0) then pmImage.DispatchCommand(ItemID);
end;

procedure TImageViewer.WMGetMinMaxInfo(var Message: TWMGetMinMaxInfo);
begin
  inherited;
  with Message.MinMaxInfo.ptMinTrackSize, Frames do begin
    if cx > X then X := cx;
    if cy > Y then Y := cy;
  end;
  MinSize := TSize(Message.MinMaxInfo.ptMinTrackSize);
end;

end.

