{$J+,Z4}
unit PGPHandler;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Classes, SysUtils, StrUtil, Messages, Forms, HelpCtrl, FileUtils, MailUtils, PMPGPCommon, PMPGPResources,
  pgpBase, pgpErrors, pgpPubTypes, KeyPropTypes, X509Types, PrefFuncs, KeyFuncs, X509Funcs, SecureFuncs,
  PGPEncode, PGPDecode, PGPDialogs, CertStore;

type
  TPGPVersion = (v65X, v7XX, v8XX, v9XX, v96X, v97X);

type
  TKeyManager = Record
    ExeName: String;
    Parameters: String;
  end;
  TKeyManagers = Array[TPGPVersion] of TKeyManager;

const
  PGPkeys = 'PGPkeys';
  PGPdesk = 'PGPdesk';
  S = '/s';
  E = '';
  KeyManagers: TKeyManagers = (
    // v65X
    (ExeName: PGPkeys; Parameters: S),
    // v7XX
    (ExeName: PGPkeys; Parameters: S),
    // v8XX
    (ExeName: PGPkeys; Parameters: S),
    // v9XX
    (ExeName: PGPdesk; Parameters: E),
    // v96X
    (ExeName: PGPdesk; Parameters: E),
    // v97X
    (ExeName: PGPdesk; Parameters: E)
  );

type
  TPGPHelpID = (idADK, idGROUPS, idWIPE_FILES);
  THelpTopics = Array[TPGPVersion, TPGPHelpID] of String;

const
  HelpTopics: THelpTopics = (
    // v65X
    ('Additional Decryption Keys',
     'Groups',
     'wipe\,using to wipe free space'),
    // v7XX
    ('/pgp/About_ADKs.htm',
     '/pgp/To_create_a_distribution_list.htm',
     '/pgp/Recommended_number_of_passes.htm'),
    // v8XX
    ('/65keys/keys-aboutadk.htm',
     '/65keys/keys-createlist.htm',
     '/45wipe/wipe-passes.htm'),
    // v9XX
    ('/Additional_Decryption_Keys.htm',
     '',
     '/Free_Space_Wiping.htm'),
    // v96X
    ('2047.htm',
     '',
     '2063.htm'),
    // v97X
    ('3443.htm',
     '',
     '2063.htm')
  );

const
  ID_ADK	= 'ADK';
  ID_GROUPS	= 'Groups';
  ID_WIPE_FILES	= 'WipeFiles';

type
  TOnSetCacheTimer = Procedure(Seconds: Longint; Wiped: Longbool) of Object;
  TSMimeCapability = (smDisabled, smReadOnly, smReadWrite);
  TKeyServer = Record
    Protocol: Integer;
    Name: String;
    Port: String;
    Flags: Integer;
  end;
  TKeyIDData = Record
    UserIDLineList: String;
    HexIDCommaList: String;
    KeyCount: Integer;
  end;
  TSignatureData = Record
    SignerID: String;
    Text: String;
    Flags: Integer;
  end;
  TKeyRingPreferences = TPreferenceRec;
  TPGPInterface = Class
  private
    FDecode: TPGPDecode;
    FEncode: TPGPEncode;
    FHelpFile: String;
    FPath: String;
    FHelpHandler: THelpHandler;
    FPMPGPStrings: TPMPGPStrings;
    FPGPErrorStrings: TPGPErrorStrings;
    FSecure: TSecure;
    FPGPVersion: String;
    FPGPVersionNumber: Integer;
    FPGPVersionSelector: TPGPVersion;
    FPGPVersionString: String;
    FComment: String;
    FPassphraseCacheSeconds: Longint;
    FWipeFilePasses: Longint;
    FSHA2Enabled: Longbool;
    FErrorCode: Longint;
    FKeyPropsList: TKeyPropsList;
    FKeyRingPrefs: TKeyRingPreferences;
    FOnSetCacheTimer: TOnSetCacheTimer;
    procedure	InitDecode;
    procedure	InitEncode;
    function	GetNoError: Longbool;
    function	GetUserAbort: Longbool;
    function	GetCacheEmpty: Longbool;
    procedure	SetCache(CacheSeconds: Longint);
    procedure	SetOnSetCacheTimer(Value: TOnSetCacheTimer);
    function	GetRingCacheUpdate: Longbool;
    procedure	SetRingCacheUpdate(Value: Longbool);
    function	GetMarginalIsInvalid: Longbool;
    function	GetSystemHasPrivateCerts: Longbool;
    function	GetKeyIDData(KPL: TKeyPropsList; var KIDD: TKeyIDData): Longint;
    function	GetCertIDData(CPL: TX509CertPropsList; var KIDD: TKeyIDData): Longint;
    function	SelectSingleKey(const Prompt: String; Owner: THandle; var KPL: TKeyPropsList): Longbool;
    procedure	SetPGPVersionString(const Value: String);
  public
    constructor	Create(HelpHandler: THelpHandler; PMPGPStrings: TPMPGPStrings; PGPErrorStrings: TPGPErrorStrings);
    destructor	Destroy; override;
    procedure	GetKeyringPreferences;
    function	Initialize: TSMimeCapability;
    function	ShowError(Code: Longint): Longbool;
    function	IsEncryptKeyError(Code: Longint): Longbool;
    procedure	LaunchHelp(const Command: String);
    procedure	LaunchKeyManager;
    procedure	WipeFile(const FileName: String);
    procedure	ClearBuffer(const Buffer: String);
    procedure	EnableCaching;
    procedure	ResetCaching;
    procedure	WipeCache;
    procedure	GetDefaultKeyServer(var KeyServer: TKeyServer);
    procedure	SetDefaultKeyServer(const KeyServer: TKeyServer);
    procedure	SetSignAlgorithm(Value: Integer; UseRipeMD: Longbool);
    function	AddBasicKeyPropsToStringList(const DefaultIDs: String; BasicKeyProps: TStrings): Longint;
    function	FormatVerificationInfo(const Info: String): String;
    function	GetShortHexID(const HexID: String): String;
    function	GetAutoListIDs: Longint;
    function	GetEncryptKeys: Longint;
    function	GetPrivateKeys(const KeyID: String): Longint;
    function	GetKeyCount(const HexIDs: String; CountUserIDs: Longbool): Longint;
    function	GetKeyHasADKProperty(const HexID: String): Longbool;
    function	GetInvalidKeyProperties(const HexID: String): String;
    function	GetSignKeyData(var KIDD: TKeyIDData; SMime: Longbool): Longbool;
    function	GetFingerprintData(Index: Integer; Owner: THandle; IncludeDetails: Longbool): String;
    function	GetSignatureData(Analysis: TAnalysis; SigProps: TSigPropsRec; var Status: Integer): String;
    function	GetKeyFromServer(const KeyID: String; Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
    function	ImportKeys(const KeyData: String; Owner: THandle; ReturnKeyIDs, Quiet: Longbool): TKeyIDData;
    function	DeleteKeys(Owner: THandle; const HexID: String; ReturnKeyIDs: Longbool): TKeyIDData;
    function	ExtractKeys(Owner: THandle; var KeyIDs: TKeyIDData): String;
    function	ImportCerts(Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
    function	DeleteCert(Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
    function	ExtractCert(Owner: THandle; var KeyIDs: TKeyIDData): String;
    function	CheckArmor(const Data: String; var MD: TMessageData): String;
    function	FindKeys(const Input: String; Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
    function	InsertExtraDataInfo(var Signature: TSignatureData; const MD: TMessageData): Longbool;
    function	VerifySignature(const Input, KeyData: String; Owner: THandle;
    				KeyServerLookup, Display: Longbool): TSignatureData;
    function	DecodeMessage(const Input: String; Owner: THandle;
			      KeyServerLookup, ReturnSignature, Recursively, Mute: Longbool;
			      var Signature: TSignatureData; var ImportedKeys: TKeyIDData;
			      var MimeCharset: String): String;
    function	DecodeFolderData(const Source: String; Owner: THandle; Recursively: Longbool;
				 var MimeCharset, SignatureText: String; var SignatureFlags: Integer): String;
    function	DetachedSignData(const Source: String; Owner: THandle; var SigningKey: TKeyIDData): String;
    function	EncodeFolderData(const Source, HexID: String; Owner: THandle): String;
    function	EncodePSMime(const Source: String; Encrypt, Sign, ToSelf, SMime, HideKeyDialog, AllInOne: Longbool;
			     Owner: THandle; var Recipient, SigningKey, ImportedKeys: TKeyIDData): String;
    function	EncodeInline(const Source, MimeCharset: String;
			     Encrypt, Sign, ToSelf, IsText, HideKeyDialog, AllInOne: Longbool;
			     Owner: THandle; var Recipient, SigningKey, ImportedKeys: TKeyIDData;
			     var ToGroup: Longbool): String;
    property	Path: String read FPath;
    property	HelpFile: String read FHelpFile;
    property	PGPVersion: String read FPGPVersion;
    property	PGPVersionNumber: Integer read FPGPVersionNumber;
    property	PGPVersionString: String read FPGPVersionString write SetPGPVersionString;
    property	Comment: String read FComment write FComment;
    property	PassphraseCacheSeconds: Longint read FPassphraseCacheSeconds write FPassphraseCacheSeconds;
    property	WipeFilePasses: Longint read FWipeFilePasses write FWipeFilePasses;
    property	CacheEmpty: Longbool read GetCacheEmpty;
    property	SHA2Enabled: Longbool read FSHA2Enabled;
    property	KeyPropsList: TKeyPropsList read FKeyPropsList;
    property	KeyRingPrefs: TKeyRingPreferences read FKeyRingPrefs;
    property	SystemHasPrivateCerts: Longbool read GetSystemHasPrivateCerts;
    property	NoError: Longbool read GetNoError;
    property	UserAbort: Longbool read GetUserAbort;
    property	ErrorCode: Longint read FErrorCode write FErrorCode;
    property	RingCacheUpdate: Longbool read GetRingCacheUpdate write SetRingCacheUpdate;
    property	OnSetCacheTimer: TOnSetCacheTimer read FOnSetCacheTimer write SetOnSetCacheTimer;
  end;

procedure ClearKeyIDData(var KeyIDData: TKeyIDData);
procedure ClearSignatureData(var SignatureData: TSignatureData);


implementation

uses
  MessageDialogs;

const
  MINUS = ' - ';
  PLUS = ' + ';


procedure ClearKeyIDData(var KeyIDData: TKeyIDData);
begin
  with KeyIDData do begin
    UserIDLineList := E;
    HexIDCommaList := E;
    KeyCount := 0;
  end;
end;

procedure ClearSignatureData(var SignatureData: TSignatureData);
begin
  with SignatureData do begin
    SignerID := E;
    Text := E;
    Flags := SF_NOTHING;
  end;
end;


// TPGPInterface -----------------------------------------------------------------------------------

constructor TPGPInterface.Create(HelpHandler: THelpHandler; PMPGPStrings: TPMPGPStrings; PGPErrorStrings: TPGPErrorStrings);
begin
  if PGPInitErrorCode = ieNone then begin
    inherited Create;
    FHelpHandler := HelpHandler;
    FSecure := TSecure.Create;
    FDecode := TPGPDecode.Create;
    FEncode := TPGPEncode.Create;
    FKeyPropsList := TKeyPropsList.Create(spgpKeyPropFlag_IDComplete or
					  spgpKeyPropFlag_X509Cert or
					  spgpKeyPropFlag_IsRevoked or
					  spgpKeyPropFlag_IsDisabled or
					  spgpKeyPropFlag_IsExpired or
					  spgpKeyPropFlag_HasADK or
					  spgpKeyPropFlag_IncludeUserIDs);
    if SDKVersion >= PGP7XX then
      FHelpFile := PGPPath + PGP_NAME + CHM
    else FHelpFile := PGPPath + PGP_NAME + HLP;
    if not FileExists(FHelpFile) then FHelpFile := GetSysDir + ExtractFileName(FHelpFile);
    FPath := PGPPath;
    FPMPGPStrings := PMPGPStrings;
    FPGPErrorStrings := PGPErrorStrings;
    FPGPVersion := PGP_NAME + SP + pgpBase.PGPVersion;
    FPGPVersionNumber := VersionNumberFromString(PGPVersion);
    case GetMainVersion(FPGPVersionNumber) of
      6: FPGPVersionSelector := v65X;
      7: FPGPVersionSelector := v7XX;
      8: FPGPVersionSelector := v8XX;
      9: begin
	FPGPVersionSelector := v9XX;
	case SDKVersion of
	  PGP96X: FPGPVersionSelector := v96X;
	  PGP97X: FPGPVersionSelector := v97X;
	end;
      end;
      10: FPGPVersionSelector := v97X;
    end;
    GetPrefComment(FComment);
    GetPrefCacheTime(FPassphraseCacheSeconds);
    GetPrefWipePasses(FWipeFilePasses);
    FSHA2Enabled := (SDKVersion >= PGP95X);
    InitDecode;
    InitEncode;
  end
  else Raise EAbort.Create(E);
end;

destructor TPGPInterface.Destroy;
begin
  FSecure.EnableCaching(-1);
  FKeyPropsList.Free;
  FEncode.Free;
  FDecode.Free;
  FSecure.Free;
  inherited;
end;

procedure TPGPInterface.GetKeyringPreferences;
var
  iPreferences: Integer;
begin
  iPreferences := PrefsFlag_PublicKeyring or PrefsFlag_PrivateKeyring or PrefsFlag_GroupsFile or PrefsFlag_DefaultKeyID;
  FErrorCode := GetPreferences(FKeyRingPrefs, iPreferences);
end;

function TPGPInterface.Initialize: TSMimeCapability;
var
  fSecringExists, fPubringExists: Longbool;
  sError, sDefUserID, sDefHexID: String;
  iHexIDLen, iIndex: Integer;
begin
  Result := smDisabled;
  with FKeyRingPrefs do begin
    KeyRings.GroupsFile := GroupsFile;
    repeat
      fSecringExists := FileExists(PrivateKeyring);
      fPubringExists := FileExists(PublicKeyring);
      if not fSecringExists and not fPubringExists then begin
	sError := LowerFirstChar(DelAmp(FPMPGPStrings[SECRING_LABEL])) + PLUS +
		  LowerFirstChar(DelAmp(FPMPGPStrings[PUBRING_LABEL]));
      end
      else if not fSecringExists then sError := LowerFirstChar(DelAmp(FPMPGPStrings[SECRING_LABEL]))
      else if not fPubringExists then sError := LowerFirstChar(DelAmp(FPMPGPStrings[PUBRING_LABEL]));
    until (fPubringExists and fSecringExists)
    or not Confirm(Format(FPMPGPStrings[PGP_INIT_RETRY_CONFIRMATION], [sError]));
    KeyRings.SetKeyRings(PublicKeyring, PrivateKeyring);
  end;
  if GetPrivateKeys(E) > 0 then with FKeyPropsList do begin
    try
      ExtractKeyIDs(FKeyRingPrefs.DefaultKeyID, sDefUserID, sDefHexID);
      iHexIDLen := Length(sDefHexID);
      case iHexIDLen of
	LongHexIDLen: iIndex := IndexOf(sDefHexID);
	ShortHexIDLen: for iIndex := 0 to pred(Count) do begin
	  if KeyProps[iIndex].kUserID = sDefHexID then Break;
	end;
      else
	iIndex := -1;
      end;
      if iIndex = -1 then iIndex := 0;
      FKeyRingPrefs.DefaultKeyID := KeyProps[iIndex].kUserID + BS + Strings[iIndex];
      if SDKVersion >= PGP81X then begin
	Result := smReadOnly;
	for iIndex := 0 to pred(Count) do if KeyProps[iIndex].kX509Cert > BadCert then begin
	  Result := smReadWrite;
	  Break;
	end;
      end;
    finally
      FKeyPropsList.Clear;
    end;
  end;
end;

function TPGPInterface.ShowError(Code: Longint): Longbool;
begin
  Result := false;
  if (Code < 0) and (Code <> kPGPError_UserAbort) then begin
    Error(FPGPErrorStrings[Code]);
    Result := true;
  end;
end;

function TPGPInterface.IsEncryptKeyError(Code: Longint): Longbool;
begin
  case Code of
    kPGPError_KeyRevoked..kPGPError_KeyUnusableForEncryption: Result := true;
  else
    Result := false;
  end;
end;

procedure TPGPInterface.LaunchHelp(const Command: String);
var
  idPGPHelp: TPGPHelpID;
  sPGPHelpTopic: String;
begin
  idPGPHelp := idADK;
  if LoStrStart(ID_GROUPS, Command) then
    idPGPHelp := idGROUPS
  else if LoStrStart(ID_WIPE_FILES, Command) then idPGPhelp := idWIPE_FILES;
  sPGPHelpTopic := HelpTopics[FPGPVersionSelector, idPGPHelp];
  if sPGPHelpTopic = E then begin
    Error(FPGPErrorStrings[kPGPError_FeatureNotAvailable] + SP + LB + FPGPVersion + RB);
    Exit;
  end;
  if SDKVersion >= PGP7XX then
    FileExecute(0, E, 'HH.exe', 'ms-its:' + FHelpFile + '::' + sPGPHelpTopic, true)
  else WinHelp(0, PChar(FHelpFile), HELP_PARTIALKEY, DWord(PChar(sPGPHelpTopic)));
end;

procedure TPGPInterface.LaunchKeyManager;
begin
  with KeyManagers[FPGPVersionSelector] do FileExecute(0, E, PGPPath + ExeName + EXE, Parameters, true);
end;

procedure TPGPInterface.WipeFile(const FileName: String);
begin
  with FSecure do begin
    WipeCount := FWipeFilePasses;
    WipeFile(FileName);
  end;
end;

procedure TPGPInterface.ClearBuffer(const Buffer: String);
begin
  FillByte(Pointer(Buffer)^, Length(Buffer), 255);
end;

procedure TPGPInterface.EnableCaching;
begin
  SetCache(0);
end;

procedure TPGPInterface.ResetCaching;
begin
  SetCache(FPassphraseCacheSeconds);
end;

procedure TPGPInterface.WipeCache;
begin
  SetCache(-1);
end;

procedure TPGPInterface.GetDefaultKeyServer(var KeyServer: TKeyServer);
var
  seDefault: TServerEntry;
begin
  FErrorCode := GetDefaultServer(seDefault);
  if FErrorCode = kPGPError_NoErr then with KeyServer, seDefault do begin
    Protocol := pred(ServerProtocol);
    Name := ServerDNS;
    if ServerPort > 0 then Port := IntToStr(ServerPort);
    Flags := ServerFlags;
  end
  else FillByte(KeyServer, SizeOf(TKeyServer), 0);
end;

procedure TPGPInterface.SetDefaultKeyServer(const KeyServer: TKeyServer);
var
  seDefault: TServerEntry;
begin
  if KeyServer.Protocol > -1 then with seDefault, KeyServer do begin
    ServerProtocol := succ(Protocol);
    ServerDNS := Name;
    ServerPort := StrToIntDef(Port, 0);
    ServerFlags := Flags;
    FErrorCode := SetDefaultServer(seDefault);
  end
  else FErrorCode := kPGPError_BadParams;
end;

procedure TPGPInterface.SetSignAlgorithm(Value: Integer; UseRipeMD: Longbool);
begin
  with FEncode do begin
    SignAlgorithm := HashAlgorithm_Default;
    if (Value = 0) and UseRipeMD then
      SignAlgorithm := HashAlgorithm_RIPEMD160
    else if FSHA2Enabled then begin
      case Value of
	1: SignAlgorithm := HashAlgorithm_SHA256;
	2: SignAlgorithm := HashAlgorithm_SHA384;
	3: SignAlgorithm := HashAlgorithm_SHA512;
      end;
    end;
  end;
end;

function TPGPInterface.AddBasicKeyPropsToStringList(const DefaultIDs: String; BasicKeyProps: TStrings): Longint;
var
  iIndex, iHexIDLen: Integer;
  sDefUserID, sDefHexID: String;
begin
  Result := -1;
  if FKeyPropsList <> nil then begin
    if BasicKeyProps <> nil then with FKeyPropsList do begin
      if DefaultIDs <> E then
	ExtractKeyIDs(DefaultIDs, sDefUserID, sDefHexID)
      else ExtractKeyIDs(FKeyRingPrefs.DefaultKeyID, sDefUserID, sDefHexID);
      iHexIDLen := Length(sDefHexID);
      for iIndex := 0 to pred(Count) do begin
	case iHexIDLen of
	  LongHexIDLen: if (Strings[iIndex] = sDefHexID) then Result := iIndex;
	  ShortHexIDLen: if (KeyProps[iIndex].kHexID = sDefHexID) then Result := iIndex;
	end;
	with KeyProps[iIndex] do BasicKeyProps.AddObject(kUserID + SP + LB + kSize + RB, Ptr(iIndex));
      end;
    end;
  end;
end;

function TPGPInterface.FormatVerificationInfo(const Info: String): String;
begin
  Result := Info;
  if not StrStart(FPMPGPStrings[NO_KEY], Result) then begin
    if not StrStart(HEX_MARKER, Result) then begin
      Result := TrimLeft(ExtractStr(Result, 1, FirstCharPos(RB, Result)) + CRLF +
				    TrimLeft(ExtractStr(Result, 1, LastCharPos(CM, Result)) + CRLF +
				    TrimLeft(Copy(Result, 2, MAXINT))));
    end
    else Delete(Result, 1, FirstCharPos(SP, Result));
  end;
end;

function TPGPInterface.GetShortHexID(const HexID: String): String;
begin
  Result := ShortHexID(HexID);
end;

function TPGPInterface.GetAutoListIDs: Longint;
begin
  FKeyPropsList.Clear;
  FErrorCode := kPGPError_NoErr;
  Result := FindKeyProps(E, FKeyPropsList,
			 spgpKeyPropFlag_KeyID,
			 KeyFilterFlag_CanEncrypt, KeyID_Ordering);
  if Result < 0 then FErrorCode := Result;
end;

function TPGPInterface.GetEncryptKeys: Longint;
begin
  FKeyPropsList.Clear;
  FErrorCode := kPGPError_NoErr;
  Result := FindKeyProps(E, FKeyPropsList,
			 spgpKeyPropFlag_IDComplete or
			 spgpKeyPropFlag_Fingerprint or
			 spgpKeyPropFlag_CreationTimeStr or
			 spgpKeyPropFlag_Trust or
			 spgpKeyPropFlag_Validity or
			 spgpKeyPropFlag_HasADK or
			 spgpKeyPropFlag_X509Cert or
			 spgpKeyPropFlag_IncludeUserIDs,
			 KeyFilterFlag_CanEncrypt, UserID_Ordering);
  if Result < 0 then FErrorCode := Result;
end;

function TPGPInterface.GetPrivateKeys(const KeyID: String): Longint;
begin
  FKeyPropsList.Clear;
  FErrorCode := kPGPError_NoErr;
  Result := FindKeyProps(KeyID, FKeyPropsList,
			 spgpKeyPropFlag_IDComplete or
			 spgpKeyPropFlag_X509Cert or
			 spgpKeyPropFlag_IncludeUserIDs,
			 KeyFilterFlag_CanSign, UserID_Ordering);
  if Result < 0 then FErrorCode := Result;
end;

function TPGPInterface.GetKeyCount(const HexIDs: String; CountUserIDs: Longbool): Longint;
var
  KPL: TKeyPropsList;
begin
  KPL := nil;
  try
    if CountUserIDs then
      Result := FindKeyProps(HexIDs, KPL, spgpKeyPropFlag_None, KeyFilterFlag_CountUserIDs, Any_Ordering)
    else Result := FindKeyProps(HexIDs, KPL, spgpKeyPropFlag_None, KeyFilterFlag_AllKeys, Any_Ordering);
  finally
    KPL.Free;
  end;
end;

function TPGPInterface.GetKeyHasADKProperty(const HexID: String): Longbool;
var
  KPL: TKeyPropsList;
begin
  KPL := nil;
  try
    FErrorCode := FindKeyProps(HexID, KPL, spgpKeyPropFlag_HasADK, KeyFilterFlag_CanEncrypt, Any_Ordering);
    Result := (FErrorCode = 1) and (KPL.KeyProps[0].kHasADK <> NoADK);
  finally
    KPL.Free;
  end;
end;

function TPGPInterface.GetInvalidKeyProperties(const HexID: String): String;
var
  KPL: TKeyPropsList;
begin
  Result := E;
  KPL := nil;
  try
    FErrorCode := FindKeyProps(HexID, KPL,
			       spgpKeyPropFlag_IDFlags or
			       spgpKeyPropFlag_IsCorrupt or
			       spgpKeyPropFlag_IsRevoked or
			       spgpKeyPropFlag_IsDisabled or
			       spgpKeyPropFlag_IsExpired,
			       KeyFilterFlag_AllKeys,
			       KeyID_Ordering);
    case FErrorCode of
      0: Result := FPMPGPStrings[MISSING_KEY];
      1: with KPL.KeyProps[0] do begin
	if kCorrupt then
	  Result := FPMPGPStrings[CORRUPT_KEY]
	else if kRevoked then
	  Result := FPMPGPStrings[REVOKED_KEY]
	else if kExpired then
	  Result := FPMPGPStrings[EXPIRED_KEY]
	else if kDisabled then
	  Result := FPMPGPStrings[DISABLED_KEY];
      end;
    else
      if FErrorCode > 1 then
	Result := FPMPGPStrings[MULTIPLE_KEYS]
      else Result := FPGPErrorStrings[FErrorCode];
    end;
  finally
    KPL.Free;
  end;
end;

function TPGPInterface.GetSignKeyData(var KIDD: TKeyIDData; SMime: Longbool): Longbool;

  function FindUserID(const UserID: String; KeyIndex: Integer; SMime: Longbool): String;
  var
    iUserIDIndex: Integer;
    sEmailAddress: String;
  begin
    if (KeyIndex > -1) and (UserID <> E) then with FKeyPropsList.KeyProps[KeyIndex].kUserIDList do begin
      iUserIDIndex := IndexOf(UserID);
      if (iUserIDIndex = -1) or (SMime and (X509Types[iUserIDIndex] < StdCert)) then begin
	sEmailAddress := ExtractEmail(UserID);
	if sEmailAddress <> E then begin
	  for iUserIDIndex := 0 to pred(Count) do begin
	    if (FirstLoStrPos(sEmailAddress, Strings[iUserIDIndex]) > 0)
	    and (not SMime or (X509Types[iUserIDIndex] > BadCert)) then begin
	      Result := Strings[iUserIDIndex];
	      Exit;
	    end;
	  end;
	end;
      end
      else begin
	Result := Strings[iUserIDIndex];
	Exit;
      end;
    end;
    Result := FKeyPropsList.KeyProps[0].kUserID;
  end;

var
  rPreferences: TPreferenceRec;
  iKeyIndex: Integer;
begin
  Result := false;
  with KIDD do begin
    KeyCount := 0;
    try
      if HexIDCommaList <> E then begin
	if GetPrivateKeys(HexIDCommaList) > 0 then begin
	  UserIDLineList := FindUserID(UserIDLineList, 0, SMime);
	  HexIDCommaList := FKeyPropsList[0];
	  KeyCount := 1;
	  Result := true;
	  Exit;
	end;
      end;
      if UserIDLineList <> E then begin
	if GetPrivateKeys(UserIDLineList) > 0 then begin
	  if GetPreferences(rPreferences, PrefsFlag_DefaultKeyID) = kPGPError_NoErr then begin
	    iKeyIndex := FKeyPropsList.IndexOf(rPreferences.DefaultKeyID);
	  end
	  else iKeyIndex := -1;
	  if iKeyIndex = -1 then iKeyIndex := 0;
	  UserIDLineList := FindUserID(UserIDLineList, iKeyIndex, SMime);
	  HexIDCommaList := FKeyPropsList.Strings[iKeyIndex];
	  KeyCount := 1;
	  Result := true;
	end;
      end;
    finally
      FKeyPropsList.Clear;
    end;
  end;
end;

function TPGPInterface.GetFingerprintData(Index: Integer; Owner: THandle; IncludeDetails: Longbool): String;

  function CertificateInfo(const HexID: String): String;
  var
    CPL: TX509CertPropsList;
    iIndex: Integer;
  begin
    Result := E;
    CPL := nil;
    try
      if FindX509CertProps(E, HexID, E, KeyFilterFlag_X509Cert, false, CPL) > kPGPError_NoErr then begin
	for iIndex := 0 to pred(CPL.Count) do begin
	  with CPL.X509CertProps[iIndex] do begin
	    Result := Result + x509OwnerCertUserID + CRLF +
		      FPMPGPStrings[CERTIFIED_BY] + CRLF +
		      GetUserFromLongName(x509IssuerLongName, true) + PARA;
	  end;
	end;
      end;
    finally
      CPL.Free;
    end;
  end;

  function FingerprintInfo: String;
  const
    ADK_HINT = '>>>>>>>>>> A D K <<<<<<<<<<';
  var
    sUserIDs, sVTLevels, sCreationDate: String;
    iIndex: Integer;
  begin
    Result := E;
    with FKeyPropsList.KeyProps[Index] do begin
      if IncludeDetails then begin
	with kUserIDList do begin
	  for iIndex := pred(Count) downto 1 do if X509Types[iIndex] > BadCert then Delete(iIndex);
	  if Count > 0 then begin
	    if kHasADK = NoADK then
	      sUserIDs := Text + CRLF
	    else sUserIDs := Text + CRLF + ADK_HINT + CRLF;
	  end;
	end;
	if kX509Cert > NoCert then System.Insert(CertificateInfo(FKeyPropsList[Index]), sUserIDs, 1);
	with FPMPGPStrings do sVTLevels := KeyValidities[kValidity] + CRLF + KeyTrusts[kTrust] + PARA;
      end
      else begin
	sUserIDs := kUserID + CRLF;
	sVTLevels := E;
      end;
      sCreationDate := kCreaTimeStr;
      Delete(sCreationDate, 1, FirstCharPos(SP, sCreationDate));
      Delete(sCreationDate, LastCharPos(CM, sCreationDate), MAXINT);
      Result := sUserIDs + FKeyPropsList[Index] + CRLF + LB + kHexID + RB + PARA +
		FPMPGPStrings.KeyAlgorithms[kKeyAlgorithm] + MINUS + kSize  + MINUS + sCreationDate + PARA +
		sVTLevels + kFingerprint + CRLF;
    end;
  end;

begin
  FErrorCode := kPGPError_NoErr;
  if Index = -1 then FKeyPropsList.Clear;
  if (Index > -1) or SelectSingleKey(FPMPGPStrings[SELECT_FINGERPRINT_KEY], Owner, FKeyPropsList) then begin
    if Index = -1 then Index := 0;
    try
      Result := FingerprintInfo;
    except
      FErrorCode := kPGPError_UnknownError;
    end;
  end;
  if ShowError(FErrorCode) then Result := E;
end;

function TPGPInterface.GetSignatureData(Analysis: TAnalysis; SigProps: TSigPropsRec; var Status: Integer): String;
const
  LBR = ' [ ';
  RBR = ' ] ';
var
  sKeyStatus: String;
  sShortSubKeyID: String;
begin
  Result := HEX_MARKER + SP + FPMPGPStrings[NO_SIGNATURE];
  Status := SF_NOTHING;
  with SigProps do begin
    if sStatus <> SIGNED_NOT then begin
      if sStatus <> SIGNED_NO_KEY then begin
	case sStatus of
	  SIGNED_GOOD, SIGNED_SIG_EXPIRED: begin
	    if Analysis > SMIME_UNRECOGNIZED then begin
	      if sValidCertChain > CertChain_Invalid then begin
		if (sValidCertChain = CertChain_Root) and (sValidity <> KeyPropTypes.Validity_Complete) then
		  sValidity := KeyPropTypes.Validity_Invalid
		else sValidity := KeyPropTypes.Validity_Complete;
	      end
	      else sValidity := KeyPropTypes.Validity_Invalid;
	    end;
	    Status := SF_VALID_UNTRUSTED;
	    case sValidity of
	      KeyPropTypes.Validity_Marginal: if not GetMarginalIsInvalid then Status := SF_VALID_TRUSTED;
	      KeyPropTypes.Validity_Complete: Status := SF_VALID_TRUSTED;
	    end;
	    Result := FPMPGPStrings[GOOD_SIG_INFO] + FPMPGPStrings[SIGNED_BY];
	  end;
	  SIGNED_BAD: begin
	    Result := FPMPGPStrings[BAD_SIG_INFO] + FPMPGPStrings[SIGNED_BY];
	    Status := SF_INVALID;
	  end;
	  SIGNED_UNKNOWN_ALG: begin
	    Result := FPMPGPStrings[UNKNOWN_ALGORITHM] + SP + FPMPGPStrings[FOR_SIGNATURE];
	    Status := SF_NO_KEY;
	  end;
	end;
	if sStatus <> SIGNED_UNKNOWN_ALG then begin
	  if sStatus <> SIGNED_BAD then begin
	    if sStatus <> SIGNED_SIG_EXPIRED then
	      Insert(SP + LB + sCreaTimeStr + RB, Result, LastCharPos(SP, Result))
	    else begin
	      Insert(SP + FPMPGPStrings[EXPIRED_SIG_INFO], Result, LastCharPos(SP, Result));
	      Status := SF_EXPIRED;
	    end;
	  end;
	  if Analysis > SMIME_UNRECOGNIZED then begin
	    if sRevoked then
	      sKeyStatus := FPMPGPStrings[CERT_REVOKED]
	    else if sExpired then
	      sKeyStatus := FPMPGPStrings[CERT_EXPIRED]
	    else if sDisabled then
	      sKeyStatus := FPMPGPStrings[CERT_CORRUPT]
	    else begin
	      case sValidCertChain of
		CertChain_Invalid: sKeyStatus := FPMPGPStrings[CHAIN_INCOMPLETE];
		CertChain_Valid: sKeyStatus := FPMPGPStrings[CHAIN_COMPLETE];
		CertChain_Root: sKeyStatus := FPMPGPStrings[CHAIN_SELF_SIGNED];
	      end;
	    end;
	  end
	  else begin
	    if sSubKeyHexID <> E then sShortSubKeyID := SP + LB + ShortHexID(sSubKeyHexID) + RB;
	    if sRevoked then
	      sKeyStatus := FPMPGPStrings[KEY_REVOKED] + sShortSubKeyID
	    else if sExpired then
	      sKeyStatus := FPMPGPStrings[KEY_EXPIRED] + sShortSubKeyID
	    else if sValidity = KeyPropTypes.Validity_Unknown then
	      sKeyStatus := FPMPGPStrings.KeyValidities[sValidity]
	    else sKeyStatus := FPMPGPStrings[KEY_NAME] + SP + FPMPGPStrings.KeyValidities[sValidity];
	  end;
	  if (sStatus = SIGNED_GOOD) and (sRevoked or sExpired or sDisabled) then Status := SF_NO_KEY;
	  Result := Result + SP + sUserID + CM + SP +
		    ShortHexID(sHexID) + CL + SP +
		    sKeyStatus;
	end;
      end
      else begin
	Result := FPMPGPStrings[NO_KEY] + LBR + ShortHexID(sHexID) + RBR +
		  FPMPGPStrings[FOR_SIGNATURE] + CRLF + FPMPGPStrings[KEYSERVER_HINT];
	Status := SF_NO_KEY;
      end;
    end
    else Status := SF_NO_SIGNATURE;
  end;
  Status := Status or (ord(Analysis in [ENCRYPTED, SMIME_ENCRYPTED]) * SF_ENCRYPTED);
end;

function TPGPInterface.GetKeyFromServer(const KeyID: String; Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
var
  KPL: TKeyPropsList;
begin
  Result.KeyCount := 0;
  if KeyID <> E then begin
    KPL := nil;
    try
      FErrorCode := GetKeyFromServerDialog(FPMPGPStrings[IMPORT_KEYS_PROMPT], KeyID, KPL, spgpKeyPropFlag_IDFlags, Owner);
      if not ShowError(FErrorCode) and (GetKeyIDData(KPL, Result) > 0) and not ReturnKeyIDs then with Result do begin
	Info(IntToStr(KeyCount) + FPMPGPStrings[KEYS_CERTS_ADDED]);
	KeyCount := 0;
      end;
    finally
      KPL.Free;
    end;
  end;
end;

function TPGPInterface.ImportKeys(const KeyData: String; Owner: THandle; ReturnKeyIDs, Quiet: Longbool): TKeyIDData;
var
  KPL: TKeyPropsList;
  pKS: pPGPKeySet;
begin
  Result.KeyCount := 0;
  if KeyData <> E then begin
    KPL := nil;
    pKS := nil;
    try
      if Quiet then begin
	FErrorCode := KeyImportDialog(E, E, KeyData, nil, KPL, false, spgpKeyPropFlag_IDFlags, Owner, pKS);
      end
      else begin
	FErrorCode := KeyImportDialog(FPMPGPStrings[IMPORT_KEYS_PROMPT],
				      FPMPGPStrings[ENTER_PASSPHRASE_PROMPT] + CL,
				      KeyData, nil, KPL, false, spgpKeyPropFlag_IDFlags, Owner, pKS);
      end;
      if (GetKeyIDData(KPL, Result) > 0) and not ReturnKeyIDs then with Result do begin
	if not Quiet then Info(IntToStr(KeyCount) + FPMPGPStrings[KEYS_CERTS_ADDED]);
	KeyCount := 0;
      end;
    finally
      KPL.Free;
    end;
  end
  else FErrorCode := kPGPError_PublicKeyNotFound;
end;

function TPGPInterface.DeleteKeys(Owner: THandle; const HexID: String; ReturnKeyIDs: Longbool): TKeyIDData;
var
  KPL: TKeyPropsList;
begin
  Result.KeyCount := 0;
  KPL := nil;
  try
    if IsHexID(HexID) = 0 then begin
      FErrorCode := SelectKeysDialog(FPMPGPStrings[DELETE_KEYS_PROMPT], KPL, false,
				     spgpKeyPropFlag_IDFlags, KeyFilterFlag_AllKeys, Owner);
      if not ShowError(FErrorCode) and (GetKeyIDData(KPL, Result) > 0) then with Result do begin
	if not Cancel(UserIDLineList + PARA + FPMPGPStrings[DELETE_KEYS_PROMPT] + QM) then begin
	  if ShowError(KeyRemove(PChar(HexIDCommaList))) or not ReturnKeyIDs then begin
	    Info(IntToStr(KeyCount) + FPMPGPStrings[KEYS_CERTS_REMOVED]);
	    KeyCount := 0;
	  end
	end
	else KeyCount := 0;
      end;
    end
    else KeyRemove(PChar(HexID));
  finally
    KPL.Free;
  end;
end;

function TPGPInterface.ExtractKeys(Owner: THandle; var KeyIDs: TKeyIDData): String;
var
  KPL: TKeyPropsList;
  sKeyData: String;
  iIndex, iLen: Integer;
  pResult: PChar;
begin
  with KeyIDs do begin
    UserIDLineList := E;
    HexIDCommaList := E;
    KeyCount := 0;
  end;
  KPL := nil;
  try
    FErrorCode := KeyExportDialog(FPMPGPStrings[EXPORT_KEYS_PROMPT], sKeyData, KPL,
				  false, true, false, spgpKeyPropFlag_IDComplete, Owner);
    if not ShowError(FErrorCode) and (GetKeyIDData(KPL, KeyIDs) > 0) then with KeyIDs do begin
      SetString(Result, nil, (MaxUserIDLen * 2) + Length(sKeyData));
      pResult := Pointer(Result);
      for iIndex := 0 to pred(KPL.Count) do with KPL.KeyProps[iIndex] do begin
	iLen := Length(kUserID);
	System.Move(Pointer(kUserID)^, pResult^, iLen);
	inc(pResult, iLen);
	PDChar(pResult)^ := CM + SP;
	inc(pResult, Length(CRLF));
	iLen := Length(FPMPGPStrings.KeyAlgorithms[kKeyAlgorithm]);
	System.Move(Pointer(FPMPGPStrings.KeyAlgorithms[kKeyAlgorithm])^, pResult^, iLen);
	inc(pResult, iLen);
	PDChar(pResult)^ := CM + SP;
	inc(pResult, Length(CRLF));
	iLen := Length(kSize);
	System.Move(Pointer(kSize)^, pResult^, iLen);
	inc(pResult, iLen);
	PDChar(pResult)^ := CRLF;
	inc(pResult, Length(CRLF));
      end;
      PDChar(pResult)^ := CRLF;
      inc(pResult, Length(CRLF));
      iLen := Length(sKeyData);
      System.Move(Pointer(sKeyData)^, pResult^, iLen);
      inc(pResult, iLen);
      System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    end;
  finally
    KPL.Free;
  end;
end;

function TPGPInterface.ImportCerts(Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
var
  CPL: TX509CertPropsList;
begin
  Result.KeyCount := 0;
  CPL := nil;
  try
    FErrorCode := ImportPrivateSystemCerts(FPMPGPStrings[IMPORT_CERTS_PROMPT],
					   FPMPGPStrings[ENTER_PASSPHRASE_PROMPT] + CL,
					   PChar(IntToHex(FSecure.Isaac.Val, 8)), CPL, Owner);
    if not ShowError(FErrorCode) and (CPL <> nil) and (CPL.Count > 0) then begin
      if (GetCertIDData(CPL, Result) > 0) and not ReturnKeyIDs then with Result do begin
	Info(IntToStr(KeyCount) + FPMPGPStrings[KEYS_CERTS_ADDED]);
	KeyCount := 0;
      end;
    end;
  finally
    CPL.Free;
  end;
end;

function TPGPInterface.DeleteCert(Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
var
  CPL: TX509CertPropsList;
begin
  Result.KeyCount := 0;
  CPL := nil;
  try
    FErrorCode := SelectX509CertificateDialog(FPMPGPStrings[DELETE_CERT_PROMPT], CPL, Owner);
    if not ShowError(FErrorCode) and (CPL <> nil) and (CPL.Count > 0) then with Result do begin
      UserIDLineList := GetUserFromLongName(CPL.X509CertProps[0].x509OwnerLongName, true);
      if not Cancel(UserIDLineList + PARA + FPMPGPStrings[DELETE_CERT_PROMPT] + QM) then begin
	HexIDCommaList := CPL[0];
	CPL.Clear;
	if not ShowError(RemoveX509Certificate(E, HexIDCommaList, true, CPL, Owner)) and (CPL <> nil) then begin
	  if ReturnKeyIDs then begin
	    HexIDCommaList := CPL.X509CertProps[0].x509OwnerKeyHexID;
	    KeyCount := 1;
	  end
	  else Info(FPMPGPStrings[KEYS_CERTS_REMOVED]);
	end;
      end;
    end;
  finally
    CPL.Free;
  end;
end;

function TPGPInterface.ExtractCert(Owner: THandle; var KeyIDs: TKeyIDData): String;
var
  CPL: TX509CertPropsList;
  sCertData, sOf: String;
begin
  with KeyIDs do begin
    UserIDLineList := E;
    HexIDCommaList := E;
    KeyCount := 0;
  end;
  CPL := nil;
  try
    FErrorCode := ExportX509Certificate(FPMPGPStrings[EXPORT_CERT_PROMPT], E, nil,
				        true, true, false, false, sCertData, CPL, Owner);
    if not ShowError(FErrorCode) and (CPL <> nil) and (CPL.Count > 0) then with CPL, X509CertProps[0] do begin
      sOf := FPMPGPStrings[CERTIFICATES_OF];
      System.Delete(sOf, 1, pred(PrevCharPos(SP, sOf, pred(Length(sOF)))));
      with KeyIDs do begin
	if Count > 1 then
	  UserIDLineList := FPMPGPStrings[CERTIFICATE_CHAIN] + sOf
	else if Count = 1 then begin
	  if x509IsRoot then
	    UserIDLineList := FPMPGPStrings[CHAIN_SELF_SIGNED] + sOf
	  else UserIDLineList := FPMPGPStrings[CERTIFICATES_OF];
	end;
	UserIDLineList := UserIDLineList + GetUserFromLongName(x509OwnerLongName, true) + CRLF;
	UpperChars(UserIDLineList, 1);
	if not X509IsRoot then begin
	  sOf := FPMPGPStrings[CERTIFIED_BY];
	  System.Delete(sOf, 1, FirstCharPos(SP, sOf));
	  HexIDCommaList := sOf + SP + GetUserFromLongName(x509IssuerLongName, false) + CRLF;
	  UpperChars(HexIDCommaList, 1);
	end;
	Result := UserIDLineList + HexIDCommaList + CRLF + sCertData;
	KeyCount := 1;
      end;
    end;
  finally
    CPL.Free;
  end;
end;

function TPGPInterface.CheckArmor(const Data: String; var MD: TMessageData): String;
begin
  Result := GetArmoredData(Data, MD);
  if MD.Encoding <= eCerts then Result := UnwrapArmoredKeys(Result);
  if (MD.Encoding = eNone) or (FDecode.AnalyseBuffer(Result) <> 0) then begin
    Result := UnwrapArmoredKeys(GetArmoredData(QPConvert(Data, false), MD));
  end;
end;

function TPGPInterface.FindKeys(const Input: String; Owner: THandle; ReturnKeyIDs: Longbool): TKeyIDData;
begin
  Result.KeyCount := 0;
  EnableCaching;
  try
    with FDecode do begin
      KeyData := E;
      CheckSig := false;
      FileOutput := false;
      OutputFileName := E;
      GetKeyFromServer := false;
      ParentHandle := Owner;
      PassThroughClearSigned := false;
      PassThroughIfUnrecognized := false;
      PassThroughKeys := false;
      QueryAddKeys := true;
      RecursivelyDecode := true;
      ShowSigLog := false;
      KeyDlgPrompt := FPMPGPStrings[IMPORT_KEYS_PROMPT];
      PassDlgPrompt := FPMPGPStrings[ENTER_PASSPHRASE_PROMPT] + CL;
      FErrorCode := DecodeMimeBuffer(Input);
      if FErrorCode = kPGPError_NoErr then begin
	if (KeyAddResult <> kPGPError_NoErr) and (KeyPropsList.Count = 0) then FErrorCode := KeyAddResult;
	if (FErrorCode = kPGPError_NoErr) and (GetKeyIDData(KeyPropsList, Result) > 0) then begin
	  if not ReturnKeyIDs then with Result do begin
	    Info(IntToStr(KeyCount) + FPMPGPStrings[KEYS_CERTS_ADDED]);
	    KeyCount := 0;
	  end;
	end
	else if FErrorCode <> kPGPError_UserAbort then begin
	  FErrorCode := kPGPError_PublicKeyNotFound;
	  Result.UserIDLineList := OutputBuffer;
	end
	else FErrorCode := kPGPError_NoErr;
	if FErrorCode = kPGPError_NoErr then ClearBuffer(OutputBuffer);
      end
      else if FErrorCode = kPGPError_ItemNotFound then FErrorCode := kPGPError_PublicKeyNotFound;
    end;
  finally
    ResetCaching;
  end;
end;

function TPGPInterface.InsertExtraDataInfo(var Signature: TSignatureData; const MD: TMessageData): Longbool;
begin
  Result := false;
  if (MD.Encoding < eCerts) and MD.ExtraData and (LoStrStart(FPMPGPStrings[GOOD_SIG_INFO], Signature.Text)) then begin
    Insert(FPMPGPStrings[EXTRA_DATA_INFO] + SP, Signature.Text, 1);
    Result := true;
  end;
end;

function TPGPInterface.VerifySignature(const Input, KeyData: String; Owner: THandle;
				       KeyServerLookup, Display: Longbool): TSignatureData;
var
  sData: String;
  MD: TMessageData;
  iStatusIndex: Integer;
  hStatusIcon: THandle;
begin
  EnableCaching;
  try
    ClearSignatureData(Result);
    FDecode.KeyData := KeyData;
    sData := CheckArmor(Input, MD);
    with FDecode do begin
      CheckSig := true;
      FileOutput := false;
      OutputFileName := E;
      GetKeyFromServer := KeyServerLookup;
      ParentHandle := Owner;
      PassThroughClearSigned := true;
      PassThroughIfUnrecognized := false;
      PassThroughKeys := false;
      QueryAddKeys := false;
      RecursivelyDecode := true;
      ShowSigLog := false;
      KeyDlgPrompt := E;
      PassDlgPrompt := FPMPGPStrings[ENTER_PASSPHRASE_PROMPT] + CL;
      FErrorCode := DecodeMimeBuffer(sData);
      ClearBuffer(OutputBuffer);
      if FErrorCode = kPGPError_NoErr then begin
	with Result do begin
	  SignerID := SigPropsRec.sHexID;
	  Text := GetSignatureData(Analysis, SigPropsRec, Flags);
	  if Display then begin
	    InsertExtraDataInfo(Result, MD);
	    if (Analysis > SMIME_UNRECOGNIZED) and (SigPropsRec.sStatus in [SIGNED_GOOD, SIGNED_SIG_EXPIRED]) then
	      ShowCertificateDialog(SMimeCertBag, E, false, Owner)
	    else begin
	      iStatusIndex := LoWord(Flags);
	      hStatusIcon := LoadIcon(PChar(iStatusIndex));
	      if iStatusIndex in [SF_VALID_TRUSTED, SF_VALID_UNTRUSTED] then
		CustomIconInfo(hStatusIcon, FormatVerificationInfo(Text))
	      else CustomIconWarn(hStatusIcon, FormatVerificationInfo(Text));
	    end;
	  end;
	end;
      end
      else ShowError(FErrorCode);
    end;
  finally
    ResetCaching;
  end;
end;

function TPGPInterface.DecodeMessage(const Input: String; Owner: THandle;
				     KeyServerLookup, ReturnSignature, Recursively, Mute: Longbool;
				     var Signature: TSignatureData; var ImportedKeys: TKeyIDData;
				     var MimeCharset: String): String;
begin
  Result := E;
  ImportedKeys.KeyCount := 0;
  ClearSignatureData(Signature);
  with FDecode do begin
    KeyData := E;
    CheckSig := true;
    FileOutput := false;
    OutputFileName := E;
    GetKeyFromServer := KeyServerLookup;
    ParentHandle := Owner;
    PassThroughClearSigned := ReturnSignature;
    PassThroughIfUnrecognized := false;
    PassThroughKeys := true;
    QueryAddKeys := false;
    RecursivelyDecode := Recursively;
    ShowSigLog := false;
    KeyDlgPrompt := E;
    PassDlgPrompt := FPMPGPStrings[ENTER_PASSPHRASE_PROMPT] + FPMPGPStrings[FOR_DECRYPTION];
    FErrorCode := DecodeMimeBuffer(Input);
    if Mute or not ShowError(FErrorCode) then begin
      MimeCharset := Charset;
      Result := OutputBuffer;
      with Signature do begin
	SignerID := SigPropsRec.sHexID;
	Text := GetSignatureData(Analysis, SigPropsRec, Flags);
      end;
      GetKeyIDData(KeyPropsList, ImportedKeys);
    end;
  end;
end;

function TPGPInterface.DecodeFolderData(const Source: String; Owner: THandle; Recursively: Longbool;
					var MimeCharset, SignatureText: String; var SignatureFlags: Integer): String;
var
  Signature: TSignatureData;
begin
  Result := E;
  ClearSignatureData(Signature);
  KeyRings.CacheUpdate := false;
  try
    with FDecode do begin
      KeyData := E;
      CheckSig := Recursively;
      FileOutput := false;
      OutputFileName := E;
      GetKeyFromServer := false;
      ParentHandle := Owner;
      PassThroughClearSigned := false;
      PassThroughIfUnrecognized := true;
      PassThroughKeys := true;
      QueryAddKeys := false;
      RecursivelyDecode := Recursively;
      ShowSigLog := false;
      KeyDlgPrompt := E;
      PassDlgPrompt := FPMPGPStrings[ENTER_PASSPHRASE_PROMPT] + CL;
      FErrorCode := DecodeMimeBuffer(Source);
      if not ShowError(FErrorCode) then with Signature do begin
	SignerID := SigPropsRec.sHexID;
	if Recursively then begin
	  Text := GetSignatureData(Analysis, SigPropsRec, Flags);
	  SignatureText := IntToStr(LoWord(Flags)) + SP + FormatVerificationInfo(Text);
	end
	else SignatureText := Text;
	SignatureFlags := Flags;
	MimeCharset := Charset;
	Result := OutputBuffer;
      end;
    end;
  finally
    KeyRings.CacheUpdate := true;
  end;
end;

function TPGPInterface.DetachedSignData(const Source: String; Owner: THandle; var SigningKey: TKeyIDData): String;
begin
  Result := E;
  with FEncode do begin
    FileOutput := false;
    OutputFileName := E;
    Armor := true;
    MIME := false;
    Textmode := false;
    Comment := E;
    ParentHandle := Owner;
    PassDlgPrompt := Format(FPMPGPStrings[SELECT_KEY_AND_ENTER_PASSPHRASE_PROMPT], [E]);
    with SigningKey do begin
      SignKeyID := HexIDCommaList;
      SignUserID := UserIDLineList;
    end;
    FErrorCode := DetachedSignBuffer(Source);
    if FErrorCode = kPGPError_NoErr then begin
      Result := OutputBuffer;
      SigningKey.HexIDCommaList := SignKeyID;
    end;
  end;
end;

function TPGPInterface.EncodeFolderData(const Source, HexID: String; Owner: THandle): String;
begin
  Result := E;
  with FEncode do begin
    FileOutput := false;
    OutputFileName := E;
    Armor := false;
    MIME := false;
    Textmode := false;
    Comment := FComment;
    EncryptKeyID := HexID;
    GroupSelected := false;
    MultipleRecipients := false;
    AlwaysShowKeyDlg := false;
    KeySelDlgPrompt := E;
    ParentHandle := Owner;
    PassDlgPrompt := E;
    Recipients := E;
    FErrorCode := KeyEncryptBuffer(Source, false, false);
    if not ShowError(FErrorCode) then Result := OutputBuffer;
  end;
end;

function TPGPInterface.EncodePSMime(const Source: String; Encrypt, Sign, ToSelf, SMime, HideKeyDialog, AllInOne: Longbool;
				    Owner: THandle; var Recipient, SigningKey, ImportedKeys: TKeyIDData): String;
var
  iLen: Integer;
begin
  Result := E;
  with FEncode do begin
    FileOutput := false;
    OutputFileName := E;
    Armor := true;
    MIME := true;
    Textmode := true;
    Comment := FComment;
    EncryptToSelf := ToSelf;
    with Recipient do begin
      if KeyCount > 0 then
	EncryptKeyID := HexIDCommaList
      else EncryptKeyID := UserIDLineList;
      Recipients := UserIDLineList;
    end;
    GroupSelected := false;
    AlwaysShowKeyDlg := not HideKeyDialog;
    MultipleRecipients := AllInOne;
    CertSelDlgPrompt := FPMPGPStrings[SELECT_CERT_OF_PROMPT];
    KeySelDlgPrompt := FPMPGPStrings[SELECT_KEY_FOR_PROMPT];
    ShowKeyDetails := FPMPGPStrings[SHOW_KEY_DETAILS];
    ParentHandle := Owner;
    PassDlgPrompt := Format(FPMPGPStrings[SELECT_KEY_AND_ENTER_PASSPHRASE_PROMPT], [E]);
    with SigningKey do begin
      SignKeyID := HexIDCommaList;
      SignUserID := UserIDLineList;
    end;
    if Encrypt then
      FErrorCode := KeyEncryptBuffer(Source, Sign, SMime)
    else if Sign then FErrorCode := ClearSignBuffer(Source, SMime);
    case FErrorCode of
      kPGPError_KeyRevoked..kPGPError_KeyUnusableForEncryption: ShowError(FErrorCode);
      kPGPError_NoErr: begin
	Result := OutputBuffer;
	if Encrypt then with Recipient do begin
	  HexIDCommaList := EncryptKeyID;
	  KeyCount := 1;
	end;
	SigningKey.HexIDCommaList := SignKeyID;
	with Recipient do begin
	  iLen := Length(HexIDCommaList);
	  if (iLen > LongHexIDLen) and StrAtPos(SignKeyID, HexIDCommaList, iLen - pred(LongHexIDLen)) then begin
	    Delete(HexIDCommaList, iLen - LongHexIDLen, MAXINT);
	  end;
	end;
	GetKeyIDData(KeyPropsList, ImportedKeys);
	if KeysFromServer = 0 then ImportedKeys.KeyCount := 0;
      end;
    end;
  end;
end;

function TPGPInterface.EncodeInline(const Source, MimeCharset: String;
				    Encrypt, Sign, ToSelf, IsText, HideKeyDialog, AllInOne: Longbool;
				    Owner: THandle; var Recipient, SigningKey, ImportedKeys: TKeyIDData;
				    var ToGroup: Longbool): String;
var
  iLen: Integer;
begin
  Result := E;
  with FEncode do begin
    FileOutput := false;
    OutputFileName := E;
    Armor := true;
    MIME := false;
    Textmode := IsText;
    Charset := MimeCharset;
    Comment := FComment;
    EncryptToSelf := ToSelf;
    with Recipient do begin
      if KeyCount > 0 then
	EncryptKeyID := HexIDCommaList
      else EncryptKeyID := UserIDLineList;
      Recipients := UserIDLineList;
    end;
    GroupSelected := ToGroup;
    AlwaysShowKeyDlg := not HideKeyDialog;
    MultipleRecipients := AllInOne;
    CertSelDlgPrompt := FPMPGPStrings[SELECT_CERT_OF_PROMPT];
    KeySelDlgPrompt := FPMPGPStrings[SELECT_KEY_FOR_PROMPT];
    ShowKeyDetails := FPMPGPStrings[SHOW_KEY_DETAILS];
    ParentHandle := Owner;
    PassDlgPrompt := Format(FPMPGPStrings[SELECT_KEY_AND_ENTER_PASSPHRASE_PROMPT], [E]);
    with SigningKey do begin
      SignKeyID := HexIDCommaList;
      SignUserID := UserIDLineList;
    end;
    if Encrypt then
      FErrorCode := KeyEncryptBuffer(Source, Sign, false)
    else if Sign then FErrorCode := ClearSignBuffer(Source, false);
    case FErrorCode of
      kPGPError_KeyRevoked..kPGPError_KeyUnusableForEncryption: ShowError(FErrorCode);
      kPGPError_NoErr: begin
	Result := OutputBuffer;
	if Encrypt then with Recipient do begin
	  HexIDCommaList := EncryptKeyID;
	  KeyCount := 1;
	end;
	SigningKey.HexIDCommaList := SignKeyID;
	ToGroup := GroupSelected;
	with Recipient do begin
	  iLen := Length(HexIDCommaList);
	  if (iLen > LongHexIDLen) and StrAtPos(SignKeyID, HexIDCommaList, iLen - pred(LongHexIDLen)) then begin
	    Delete(HexIDCommaList, iLen - LongHexIDLen, MAXINT);
	  end;
	end;
	GetKeyIDData(KeyPropsList, ImportedKeys);
	if KeysFromServer = 0 then ImportedKeys.KeyCount := 0;
      end;
    end;
  end;
end;

procedure TPGPInterface.InitDecode;
begin
  with FDecode do begin
    LogBadSignature := FPMPGPStrings[BAD_SIGNATURE];
    LogExpSignature := FPMPGPStrings[EXPIRED_SIGNATURE];
    LogKeyDisabled := FPMPGPStrings[KEY_DISABLED];
    LogKeyExpired := FPMPGPStrings[KEY_EXPIRED];
    LogKeyRevoked := FPMPGPStrings[KEY_REVOKED];
    LogKeyUnknown := FPMPGPStrings[KEY_UNKNOWN];
    HelpHandler := FHelpHandler;
    Secure := Self.FSecure;
  end;
end;

procedure TPGPInterface.InitEncode;
begin
  with FEncode do begin
    GroupText := FPMPGPStrings[GROUP_NAME];
    KeyServerBtn := FPMPGPStrings[FROM_KEYSERVER_BUTTON];
    KeyUseAddBtn := FPMPGPStrings[SELECT_AND_ADD_BUTTON];
    KeyCancelBtn := FPMPGPStrings[CANCEL_BUTTON];
    KeyOkBtn := '&Ok';
    SelectedKeysText := FPMPGPStrings[SELECTED_KEYS_TEXT];
    SelectAllBtn := FPMPGPStrings[SELECT_ALL_BUTTON];
    UnselectBtn := FPMPGPStrings[UNSELECT_BUTTON];
    CertPropsViewerBtn := FPMPGPStrings[SHOW_CERT_PROPS_BUTTON];
    CertificateText := FPMPGPStrings[CERTIFICATE_NAME];
    CertifiedText := FPMPGPStrings[CERTIFIED_BY];
    KeyTrusts := FPMPGPStrings.KeyTrustLevels;
    KeyValidities := FPMPGPStrings.KeyValidityLevels;
    KeyAlgorithms := FPMPGPStrings.KeyAlgorithmNames;
    HashAlgorithms := FPMPGPStrings.HashAlgorithmNames;
    CertValidities := FPMPGPStrings.CertValidityLevels;
    KeyDialogHelpIndex := ord(IDH_ENCRYPT_KEY_DLG);
    OmitMimeVersion := true;
    HelpHandler := FHelpHandler;
    Secure := FSecure;
  end;
end;

function TPGPInterface.GetNoError: Longbool;
begin
  Result := (FErrorCode = kPGPError_NoErr);
end;

function TPGPInterface.GetUserAbort: Longbool;
begin
  Result := (FErrorCode = kPGPError_UserAbort);
end;

function TPGPInterface.GetCacheEmpty: Longbool;
begin
  Result := FSecure.CacheEmpty;
end;

procedure TPGPInterface.SetCache(CacheSeconds: Longint);
var
  fCached: Longbool;
begin
  if (FSecure <> nil) and Assigned(FOnSetCacheTimer) then begin
    fCached := not CacheEmpty;
    FSecure.EnableCaching(CacheSeconds);
    FOnSetCacheTimer(CacheSeconds, fCached and CacheEmpty);
  end;
end;

procedure TPGPInterface.SetOnSetCacheTimer(Value: TOnSetCacheTimer);
begin
  FOnSetCacheTimer := Value;
  ResetCaching;
end;

function TPGPInterface.GetRingCacheUpdate: Longbool;
begin
  Result := KeyRings.CacheUpdate;
end;

procedure TPGPInterface.SetRingCacheUpdate(Value: Longbool);
begin
  KeyRings.CacheUpdate := Value;
end;

function TPGPInterface.GetMarginalIsInvalid: Longbool;
begin
  GetPrefMarginalIsInvalid(Result);
end;

function TPGPInterface.GetSystemHasPrivateCerts: Longbool;
begin
  Result := CertStore.SystemHasPrivateCerts;
end;

function TPGPInterface.GetKeyIDData(KPL: TKeyPropsList; var KIDD: TKeyIDData): Longint;
var
  pUserIDs, pHexIDs: PChar;
  iIndex, iLen: Integer;
  sID: String;
begin
  Result := 0;
  FillByte(KIDD, SizeOf(TKeyIDData), 0);
  if KPL <> nil then with KPL, KIDD do begin
    if Count > 0 then begin
      SetString(UserIDLineList, nil, Count * (kPGPMaxUserIDSize + Length(CRLF)));
      SetString(HexIDCommaList, nil, Count * (LongHexIDLen + Length(CM)));
      pUserIDs := Pointer(UserIDLineList);
      pHexIDs := Pointer(HexIDCommaList);
      for iIndex := 0 to pred(Count) do begin
	sID := KeyProps[iIndex].kUserID;
	iLen := Length(sID);
	System.Move(Pointer(sID)^, pUserIDs^, iLen);
	inc(pUserIDs, iLen);
	PDChar(pUserIDs)^ := CRLF;
	inc(pUserIDs, Length(CRLF));
	sID := Strings[iIndex];
	iLen := Length(sID);
	System.Move(Pointer(sID)^, pHexIDs^, iLen);
	inc(pHexIDs, iLen);
	pHexIDs^ := CM;
	inc(pHexIDs);
	inc(KeyCount);
      end;
      Result := KeyCount;
      System.Delete(UserIDLineList, succ(pUserIDs - Pointer(UserIDLineList) - Length(CRLF)), MAXINT);
      System.Delete(HexIDCommaList, succ(pHexIDs - Pointer(HexIDCommaList) - Length(CM)), MAXINT);
    end;
  end;
end;

function TPGPInterface.GetCertIDData(CPL: TX509CertPropsList; var KIDD: TKeyIDData): Longint;
var
  pUserIDs, pHexIDs: PChar;
  iIndex, iLen: Integer;
  sID: String;
begin
  Result := 0;
  FillByte(KIDD, SizeOf(TKeyIDData), 0);
  if CPL <> nil then with CPL, KIDD do begin
    if Count > 0 then begin
      SetString(UserIDLineList, nil, Count * (kPGPMaxUserIDSize + Length(CRLF)));
      SetString(HexIDCommaList, nil, Count * (LongHexIDLen + Length(CM)));
      pUserIDs := Pointer(UserIDLineList);
      pHexIDs := Pointer(HexIDCommaList);
      for iIndex := 0 to pred(Count) do begin
	sID := X509CertProps[iIndex].x509OwnerCertUserID;
	iLen := Length(sID);
	System.Move(Pointer(sID)^, pUserIDs^, iLen);
	inc(pUserIDs, iLen);
	PDChar(pUserIDs)^ := CRLF;
	inc(pUserIDs, Length(CRLF));
	sID := X509CertProps[iIndex].x509OwnerKeyHexID;
	iLen := Length(sID);
	System.Move(Pointer(sID)^, pHexIDs^, iLen);
	inc(pHexIDs, iLen);
	pHexIDs^ := CM;
	inc(pHexIDs);
	inc(KeyCount);
      end;
      Result := KeyCount;
      System.Delete(UserIDLineList, succ(pUserIDs - Pointer(UserIDLineList) - Length(CRLF)), MAXINT);
      System.Delete(HexIDCommaList, succ(pHexIDs - Pointer(HexIDCommaList) - Length(CM)), MAXINT);
    end;
  end;
end;

function TPGPInterface.SelectSingleKey(const Prompt: String; Owner: THandle; var KPL: TKeyPropsList): Longbool;
begin
  FErrorCode := SelectKeysDialog(Prompt, KPL, true,
				 spgpKeyPropFlag_IDComplete or
				 spgpKeyPropFlag_Fingerprint or
				 spgpKeyPropFlag_CreationTimeStr or
				 spgpKeyPropFlag_Trust or
				 spgpKeyPropFlag_Validity or
				 spgpKeyPropFlag_HasADK or
				 spgpKeyPropFlag_X509Cert or
				 spgpKeyPropFlag_IncludeUserIDs,
				 KeyFilterFlag_AllKeys, Owner);
  Result := ((FErrorCode = kPGPError_NoErr) and (KPL <> nil) and (KPL.Count = 1));
end;

procedure TPGPInterface.SetPGPVersionString(const Value: String);
begin
  FPGPVersionString := Value;
  StrPLCopy(MyVersion, FPGPVersionString, pred(SizeOf(TVersionString)));
end;


end.

