{$J+,Z4}
unit WebLoader;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2007-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Classes, StrUtil, WebInterface;

type
  TWinInetError = (wieNone, wieNoUrl, wieConnect, wieLoadPage, wieReadPage, wieWriteFile, wieCancelled);
  TInetHandles = Record
    AccessType: DWord;
    AccessFlags: DWord;
    Connection: HInternet;
    Page: HInternet;
  end;
  PReadWebData = ^TReadWebData;
  TReadWebData = Record
    StatusWindow: THandle;
    UserAgent: String;
    TargetUrls: String;
    StatusText: String;
    FileName: String;
    InetHandles: TInetHandles;
    Domain: String;
    Seconds: DWord;
    Status: String;
  end;

function  ReadWebData(var RWD: TReadWebData): TWinInetError; stdcall;
function  IsBrowserProtocol(Scheme: Integer): Longbool;
function  IsBrowserUrl(const Url: String): Longbool;
function  IsAboutUrl(const Url: String): Longbool;
function  IsCidUrl(const Url: String): Longbool;
function  IsDataUrl(const Url: String): Longbool;
function  IsFileUrl(const Url: String): Longbool;
function  IsMailToUrl(const Url: String): Longbool;
function  IsResUrl(const Url: String): Longbool;
function  IsScriptUrl(const Url: String): Longbool;
procedure CloseConnection(var InetHandles: TInetHandles);


implementation

const
  PROT_ABOUT = 'about:';
  PROT_CID   = 'cid:';
  PROT_DATA  = 'data:';
  PROT_FILE  = 'file:';
  PROT_MAIL  = 'mailto:';
  PROT_RES   = 'res:';

const
  PROT_A_DS  = PROT_ABOUT + DS;
  PROT_C_DS  = PROT_CID + DS;
  PROT_D_DS  = PROT_DATA + DS;
  PROT_M_DS  = PROT_MAIL + DS;
  PROT_F_DS  = PROT_FILE + DS;
  PROT_R_DS  = PROT_RES + DS;

const
  INTERNET_SCHEME_PARTIAL				= -2;
  INTERNET_SCHEME_UNKNOWN				= -1;
  INTERNET_SCHEME_DEFAULT				= 00;
  INTERNET_SCHEME_FTP					= 01;
  INTERNET_SCHEME_GOPHER				= 02;
  INTERNET_SCHEME_HTTP					= 03;
  INTERNET_SCHEME_HTTPS					= 04;
  INTERNET_SCHEME_FILE					= 05;
  INTERNET_SCHEME_NEWS					= 06;
  INTERNET_SCHEME_MAILTO				= 07;
  INTERNET_SCHEME_SOCKS					= 08;
  INTERNET_SCHEME_JAVASCRIPT				= 09;
  INTERNET_SCHEME_VBSCRIPT				= 10;
  INTERNET_SCHEME_RES					= 11;
  INTERNET_SCHEME_FIRST					= INTERNET_SCHEME_FTP;
  INTERNET_SCHEME_LAST					= INTERNET_SCHEME_RES;

function TrimUrl(const Url: String): String;
const
  WSP = [#1..#32];
var
  pEnd, pBeg: PChar;
begin
  if Url <> E then begin
    pBeg := Pointer(Url);
    pEnd := pBeg + Length(Url) - 1;
    while (pEnd > pBeg) and (pEnd^ in WSP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ in WSP) do inc(pBeg);
    while (pBeg + 1 < pEnd) and (pBeg^ = PC) and (HexToChar(pBeg + 1) in WSP) do inc(pBeg, 3);
    while (pEnd - 1 > pBeg) and ((pEnd - 2)^ = PC) and (HexToChar(pEnd - 1) in WSP) do dec(pEnd, 3);
    SetString(Result, pBeg, succ(pEnd - pBeg));
    Exit;
  end;
  Result := Url;
end;

function TabExtract(var Text: String): String;
var
  iCharPos: Integer;
begin
  Result := E;
  iCharPos := pos(TAB, Text);
  if iCharPos > 0 then begin
    Result := Copy(Text, 1, pred(iCharPos));
    Delete(Text, 1, iCharPos);
  end
  else begin
    Result := Text;
    Text := E;
  end;
end;

function CheckConnectedState: DWord;
var
  dwSize: DWord;
  iciState: TInternetConnectedInfo;
begin
  Result := 0;
  dwSize := SizeOf(Result);
  if InternetQueryOption(nil, INTERNET_OPTION_CONNECTED_STATE, @Result, dwSize)
  and ((Result and INTERNET_STATE_DISCONNECTED_BY_USER) <> 0) then begin
    with iciState do begin
      dwConnectedState := INTERNET_STATE_CONNECTED;
      dwFlags := 0;
    end;
    InternetSetOption(nil, INTERNET_OPTION_CONNECTED_STATE, @iciState, SizeOf(iciState));
  end;
end;

procedure ResetConnectedState(PreviousState: DWord);
var
  iciState: TInternetConnectedInfo;
begin
  if (PreviousState and INTERNET_STATE_DISCONNECTED_BY_USER) <> 0 then begin
    with iciState do begin
      dwConnectedState := INTERNET_STATE_DISCONNECTED_BY_USER;
      dwFlags := ISO_FORCE_DISCONNECTED;
    end;
    InternetSetOption(nil, INTERNET_OPTION_CONNECTED_STATE, @iciState, SizeOf(iciState));
  end;
end;

function GetConnectTimeout(Connection: HInternet): DWord; // seconds
var
  dwSize: DWord;
begin
  Result := 0;
  dwSize := SizeOf(Result);
  InternetQueryOption(Connection, INTERNET_OPTION_CONNECT_TIMEOUT, @Result, dwSize);
  Result := Result div 1000;
end;

function GetUsesProxy(Connection: HInternet): Longbool;
var
  dwSize: DWord;
  ipiConnection: TInternetProxyInfo;
begin
  Result := false;
  dwSize := SizeOf(TInternetProxyInfo);
  if InternetQueryOption(Connection, INTERNET_OPTION_PROXY, @ipiConnection, dwSize) then with ipiConnection do begin
    Result := (dwAccessType = INTERNET_OPEN_TYPE_PROXY) and (lpszProxy <> nil) and (lpszProxy[0] <> NULL);
  end;
end;

function GetStatusCode(WebPage: HInternet): DWord;
var
  dwSize: DWord;
begin
  Result := 0;
  dwSize := SizeOf(Result);
  InternetQueryDataAvailable(WebPage, nil, 0, 0);
  HttpQueryInfo(WebPage, HTTP_QUERY_STATUS_CODE or HTTP_QUERY_FLAG_NUMBER, @Result, dwSize, nil);
end;

function GetContentLength(WebPage: HInternet; MaxSize: DWord): DWord;
var
  dwSize: DWord;
begin
  Result := MaxSize;
  dwSize := SizeOf(Result);
  InternetQueryDataAvailable(WebPage, nil, 0, 0);
  HttpQueryInfo(WebPage, HTTP_QUERY_CONTENT_LENGTH or HTTP_QUERY_FLAG_NUMBER, @Result, dwSize, nil);
end;

function CanonicalizeUrl(const Url: String; Decode: Longbool): String;
var
  dwLen, dwFlags: DWord;
  sUrl: String;
begin
  sUrl := Unquote(Url, LT, GT);
  if not IsDataUrl(sUrl) then begin
    dwLen := Length(NULL);
    if Decode then begin
      sUrl := Unescape(Url, false, PC);
      dwFlags := ICU_DECODE or ICU_NO_ENCODE;
    end
    else dwFlags := ICU_BROWSER_MODE or ICU_DECODE;
    if DWord(Length(sUrl)) > dwLen then begin
      if LoStrAtPos(PROT_ABOUT, sUrl, 1) then
	Delete(sUrl, 1, Length(PROT_ABOUT))
      else if FirstLoStrPos(TS, sUrl) > 0 then
	sUrl := XChgStrings(sUrl, TS, DS)
      else UniqueString(sUrl);
      InternetCanonicalizeUrl(Pointer(sUrl), E, dwLen, dwFlags);
      if GetLastError = ERROR_INSUFFICIENT_BUFFER then begin
	SetString(Result, NIL, pred(dwLen));
	if InternetCanonicalizeUrl(Pointer(sUrl), Pointer(Result), dwLen, dwFlags) then begin
	  Result := TrimUrl(Result);
	  Exit;
	end;
      end;
      Result := TrimUrl(sUrl);
    end
    else Result := E;
  end
  else Result := Url;
end;

function GetUrlProtocol(const Url: String): Integer;
var
  sDummy: String;
  uc: TUrlComponents;
begin
  if Url <> E then begin
    sDummy := CanonicalizeUrl(Url, false);
    FillChar(uc, SizeOf(TUrlComponents), 0);
    uc.dwStructSize := SizeOf(TUrlComponents);
    if InternetCrackUrl(PChar(sDummy), 0, 0, @uc)
    or InternetCrackUrl(PChar(Unescape(sDummy, false, PC)), 0, 0, @uc) then begin
      Result := uc.nScheme;
      Exit;
    end;
  end;
  Result := INTERNET_SCHEME_UNKNOWN;
end;

function IsBrowserProtocol(Scheme: Integer): Longbool;
begin
  Result := (Scheme in [INTERNET_SCHEME_FTP..INTERNET_SCHEME_HTTPS,
			INTERNET_SCHEME_SOCKS..INTERNET_SCHEME_VBSCRIPT]);
end;

function IsBrowserUrl(const Url: String): Longbool;
begin
  Result := IsBrowserProtocol(GetUrlProtocol(Url));
end;

function IsAboutUrl(const Url: String): Longbool;
begin
  Result := LoStrAtPos(PROT_ABOUT, Url, 1);
end;

function IsCidUrl(const Url: String): Longbool;
begin
  Result := LoStrAtPos(PROT_CID, Url, 1);
end;

function IsDataUrl(const Url: String): Longbool;
begin
  Result := LoStrAtPos(PROT_DATA, Url, 1);
end;

function IsFileUrl(const Url: String): Longbool;
begin
  Result := GetUrlProtocol(Url) = INTERNET_SCHEME_FILE;
end;

function IsMailToUrl(const Url: String): Longbool;
begin
  Result := LoStrAtPos(PROT_MAIL, Url, 1);
end;

function IsResUrl(const Url: String): Longbool;
begin
  Result := LoStrAtPos(PROT_RES, Url, 1);
end;

function IsScriptUrl(const Url: String): Longbool;
begin
  Result := (GetUrlProtocol(Url) in [INTERNET_SCHEME_JAVASCRIPT, INTERNET_SCHEME_VBSCRIPT]);
end;

function GetWebDomain(const TargetUrl: String): String;
var
  ucTarget: TUrlComponents;
begin
  Result := CanonicalizeUrl(TargetUrl, false);
  if Result <> E then begin
    FillChar(ucTarget, SizeOf(TUrlComponents), 0);
    with ucTarget do begin
      dwStructSize := SizeOf(TUrlComponents);
      dwHostNameLength := 1;
    end;
    if InternetCrackUrl(Pointer(Result), 0, 0, @ucTarget) then begin
      with ucTarget do Result := Copy(lpszHostName, 1, dwHostNameLength);
    end;
  end;
end;

procedure SetStatusText(var RWD: TReadWebData);
begin
  with RWD do begin
    Status := Format(StatusText, [Domain, Seconds]);
    SetWindowText(StatusWindow, PChar(Status));
    InvalidateRect(StatusWindow, nil, false);
    UpdateWindow(StatusWindow);
    if Seconds = 0 then
      Seconds := 60
    else dec(Seconds);
  end;
end;

procedure TimerProc(hwnd: HWND; uMsg: UINT; idEvent: UINT; dwTime: DWORD); stdcall;
begin
  SetStatusText(PReadWebData(idEvent)^);
end;

procedure ResetTimer(var RWD: TReadWebData; Connection: HInternet);
begin
  RWD.Seconds := GetConnectTimeout(Connection);
  SetStatusText(RWD);
  SetTimer(RWD.StatusWindow, UINT(@RWD), 1000, @TimerProc);
end;

procedure StopTimer(var RWD: TReadWebData);
begin
  KillTimer(RWD.StatusWindow, UINT(@RWD));
  RWD.Seconds := 0;
  SetStatusText(RWD);
end;

procedure InternetStatusCallback(hInet: HInternet; dwContext: DWord; dwInternetStatus: DWord;
				 lpvStatusInformation: Pointer; dwStatusInformationLength: DWord); stdcall;
begin
  if dwContext <> 0 then begin
    case dwInternetStatus of
      INTERNET_STATUS_RESOLVING_NAME: ;
      INTERNET_STATUS_NAME_RESOLVED: ;
      INTERNET_STATUS_CONNECTING_TO_SERVER: ;
      INTERNET_STATUS_CONNECTED_TO_SERVER: ;
      INTERNET_STATUS_SENDING_REQUEST: ;
      INTERNET_STATUS_REQUEST_SENT: ;
      INTERNET_STATUS_RECEIVING_RESPONSE: ;
      INTERNET_STATUS_RESPONSE_RECEIVED: ;
      INTERNET_STATUS_CLOSING_CONNECTION: ;
      INTERNET_STATUS_CONNECTION_CLOSED: ;
      INTERNET_STATUS_HANDLE_CREATED: ResetTimer(PReadWebData(dwContext)^, hInet);
      INTERNET_STATUS_HANDLE_CLOSING: StopTimer(PReadWebData(dwContext)^);
      INTERNET_STATUS_DETECTING_PROXY: ;
      INTERNET_STATUS_REQUEST_COMPLETE: ;
      INTERNET_STATUS_REDIRECT: ;
      INTERNET_STATUS_INTERMEDIATE_RESPONSE: ;
      INTERNET_STATUS_USER_INPUT_REQUIRED: ;
      INTERNET_STATUS_STATE_CHANGE: ;
      INTERNET_STATUS_COOKIE_SENT: ;
      INTERNET_STATUS_COOKIE_RECEIVED: ;
      INTERNET_STATUS_PRIVACY_IMPACTED: ;
    else
      Exit;
    end;
  end;
end;

function Access(var RWD: TReadWebData; const Target: String): TWinInetError;
begin
  Result := wieNone;
  with RWD.InetHandles do begin
    Connection := InternetOpen(Pointer(RWD.UserAgent), AccessType, nil, 0, 0);
    if Connection <> nil then begin
      InternetSetStatusCallback(Connection, @InternetStatusCallback);
      Page := InternetOpenUrl(Connection, Pointer(Target), nil, 0, AccessFlags, DWord(@RWD));
      if Page = nil then begin
	if GetLastError = ERROR_INTERNET_OPERATION_CANCELLED then
	  Result := wieCancelled
	else Result := wieLoadPage;
	InternetCloseHandle(Connection);
	Connection := nil;
      end;
    end
    else Result := wieConnect;
  end;
end;

function Connect(var RWD: TReadWebData; const Target: String): TWinInetError;
begin
  with RWD.InetHandles do begin
    AccessFlags := INTERNET_FLAG_RELOAD or INTERNET_FLAG_DONT_CACHE;
    Connection := nil;
    Page := nil;
    if GetUsesProxy(nil) then
      AccessType := PRE_CONFIG_INTERNET_ACCESS
    else AccessType := LOCAL_INTERNET_ACCESS;
    Result := Access(RWD, Target);
    if (Result in [wieConnect, wieLoadPage]) and (AccessType = PRE_CONFIG_INTERNET_ACCESS) then begin
      AccessType := LOCAL_INTERNET_ACCESS;
      Result := Access(RWD, Target);
    end;
  end;
end;

procedure CloseConnection(var InetHandles: TInetHandles);
begin
  with InetHandles do begin
    InternetCloseHandle(Page);
    InternetCloseHandle(Connection);
    Page := nil;
    Connection := nil;
  end;
end;

function ReadDataToString(WebPage: HInternet; var Data: String): TWinInetError;
var
  dwLen, dwMax, dwSize: DWord;
  pData: PChar;
begin
  Result := wieReadPage;
  dwLen := GetContentLength(WebPage, $10000);
  if GetLastError <> ERROR_HTTP_HEADER_NOT_FOUND then
    dwMax := dwLen
  else dwMax := 0;
  SetString(Data, nil, dwLen);
  pData := Pointer(Data);
  if pData <> nil then begin
    repeat
      dwSize := dwLen;
      if InternetQueryDataAvailable(WebPage, @dwSize, 0, 0) then begin
	if dwSize > dwLen then dwSize := dwLen;
	if (dwSize > 0) and InternetReadFile(WebPage, pData, dwSize, dwSize) then begin
	  inc(pData, dwSize);
	  dec(dwLen, dwSize);
	end;
      end;
    until dwSize = 0;
    dwLen := pData - Pointer(Data);
    Delete(Data, succ(dwLen), MAXINT);
    if (dwLen = dwMax) or ((dwLen > 0) and (dwMax = 0)) then Result := wieNone;
  end;
end;

function ReadDataToFile(WebPage: HInternet; const FileName: String): TWinInetError;
var
  sData: String;
  fsData: TFileStream;
  iDataSize: Integer;
begin
  Result := ReadDataToString(WebPage, sData);
  if Result = wieNone then begin
    Result := wieWriteFile;
    try
      fsData := TFileStream.Create(FileName, fmCreate);
      if fsData <> nil then with fsData do begin
	try
	  iDataSize := Length(sData);
	  if Write(Pointer(sData)^, iDataSize) = iDataSize then Result := wieNone;
	finally
	  fsData.Free;
	  if Result <> wieNone then DeleteFile(FileName);
	end;
      end;
    except
    end;
  end;
end;

function ReadWebData(var RWD: TReadWebData): TWinInetError; stdcall;
var
  sUrlList, sTarget: String;
  dwConnectedState: DWord;
begin
  Result := wieNoUrl;
  sUrlList := RWD.TargetUrls;
  dwConnectedState := CheckConnectedState;
  try
    if sUrlList <> E then begin
      repeat
	sTarget := TabExtract(sUrlList);
	RWD.Domain := GetWebDomain(sTarget);
	Result := Connect(RWD, sTarget);
	case Result of
	  wieNone: begin
	    try
	      Result := ReadDataToFile(RWD.InetHandles.Page, RWD.FileName);
	    finally
	      CloseConnection(RWD.InetHandles);
	    end;
	  end;
	  wieCancelled: Exit;
	end;
      until (Result = wieNone) or (sUrlList = E);
    end;
  finally
    ResetConnectedState(dwConnectedState);
  end;
end;

function LoadWinInet: Longbool;
var
  hWinInet: HModule;
begin
  Result := false;
  hWinInet := LoadLibrary(WININET);
  if hWinInet <> 0 then begin
    InternetGetConnectedState := GetProcAddress(hWinInet, 'InternetGetConnectedState');
    InternetQueryOption := GetProcAddress(hWinInet, 'InternetQueryOptionA');
    if not Assigned(InternetQueryOption) then InternetQueryOption := GetProcAddress(hWinInet, 'InternetQueryOption');
    InternetSetOption := GetProcAddress(hWinInet, 'InternetSetOptionA');
    if not Assigned(InternetSetOption) then InternetSetOption := GetProcAddress(hWinInet, 'InternetSetOption');
    InternetOpen := GetProcAddress(hWinInet, 'InternetOpenA');
    InternetOpenUrl := GetProcAddress(hWinInet, 'InternetOpenUrlA');
    InternetSetStatusCallback := GetProcAddress(hWinInet, 'InternetSetStatusCallback');
    InternetQueryDataAvailable := GetProcAddress(hWinInet, 'InternetQueryDataAvailable');
    InternetReadFile := GetProcAddress(hWinInet, 'InternetReadFile');
    InternetCloseHandle := GetProcAddress(hWinInet, 'InternetCloseHandle');
    HttpQueryInfo := GetProcAddress(hWinInet, 'HttpQueryInfoA');
    HttpSendRequest := GetProcAddress(hWinInet, 'HttpSendRequestA');
    InternetCanonicalizeUrl := GetProcAddress(hWinInet, 'InternetCanonicalizeUrlA');
    InternetCrackUrl := GetProcAddress(hWinInet, 'InternetCrackUrlA');
    DeleteUrlCacheEntry := GetProcAddress(hWinInet, 'DeleteUrlCacheEntryA');
    Result := true;
  end;
end;

function FreeWinInet: Longbool;
begin
  Result := FreeLibrary(GetModuleHandle(WININET));
end;

initialization
  LoadWinInet;

finalization
  FreeWinInet;

end.

