{$J+,Z4}
unit KeyPropTypes;
{$IFDEF CONDITIONALEXPRESSIONS} {$DEFINE FIXED} {$ENDIF} // Delphi 6 and later

{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2001-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  Windows,
  Classes,
  SysUtils,
  StrTools,
  pgpBase,
  pgpPubTypes;

const
  ShortHexIDLen				= 10;
  LongHexIDLen				= 18;
  DefHexIDLen				= LongHexIDLen;
  MaxUserIDLen				= kPGPMaxUserIDSize;

{ My tribute to Steve Heller's great work with creating SPGP & DPGP }
const
  spgpKeyPropFlag_None			= $00000000;
  spgpKeyPropFlag_All			= $FFFFFFFF;

  // "string" properties
  spgpKeyPropFlag_KeyID			= $00000001;
  spgpKeyPropFlag_UserID		= $00000002;
  spgpKeyPropFlag_Fingerprint		= $00000004;
  spgpKeyPropFlag_CreationTimeStr	= $00000008;
  spgpKeyPropFlag_ExpirationTimeStr	= $00000010;

  // "numeric" properties
  spgpKeyPropFlag_KeyBits		= $00000040;
  spgpKeyPropFlag_X509Cert		= $00000080;
  spgpKeyPropFlag_KeyAlg		= $00000100;
  spgpKeyPropFlag_HashAlg		= $00000200;
  spgpKeyPropFlag_Trust			= $00000400;
  spgpKeyPropFlag_Validity		= $00000800;
  spgpKeyPropFlag_CreationTime		= $00001000;
  spgpKeyPropFlag_ExpirationTime	= $00002000;

  // "boolean" properties
  spgpKeyPropFlag_IsCorrupt		= $00004000;
  spgpKeyPropFlag_IsSecret		= $00008000;
  spgpKeyPropFlag_IsAxiomatic		= $00010000;
  spgpKeyPropFlag_IsRevoked		= $00020000;
  spgpKeyPropFlag_IsDisabled		= $00040000;
  spgpKeyPropFlag_IsExpired		= $00080000;
  spgpKeyPropFlag_IsSecretShared	= $00100000;

  spgpKeyPropFlag_CanEncrypt		= $00200000;
  spgpKeyPropFlag_CanDecrypt		= $00400000;
  spgpKeyPropFlag_CanSign		= $00800000;
  spgpKeyPropFlag_CanVerify		= $01000000;

  spgpKeyPropFlag_HasRevoker		= $02000000;
  spgpKeyPropFlag_HasADK		= $04000000;
  spgpKeyPropFlag_HasSubKey		= $08000000;
  spgpKeyPropFlag_LegacyKey		= $10000000;

  // "list" flags
  spgpKeyPropFlag_IncludeUserIDs	= $20000000;
  spgpKeyPropFlag_IncludeSignerIDs	= $40000000;
  spgpKeyPropFlag_IncludeGroupsList	= $80000000;

  // "convenience" flags
  spgpKeyPropFlag_IDFlags		= spgpKeyPropFlag_KeyID or spgpKeyPropFlag_UserID;
  spgpKeyPropFlag_IDComplete		= spgpKeyPropFlag_IDFlags or spgpKeyPropFlag_KeyBits or spgpKeyPropFlag_KeyAlg;

const
  // include all keyring keys
  KeyFilterFlag_AllKeys			= $00000;

  // capabilities
  KeyFilterFlag_CanEncrypt		= $00001;
  KeyFilterFlag_CanDecrypt		= $00002;
  KeyFilterFlag_CanSign			= $00004;
  KeyFilterFlag_CanVerify		= $00008;

  // key status
  KeyFilterFlag_Enabled			= $00010;
  KeyFilterFlag_Disabled		= $00020;

  // algorithms
  KeyFilterFlag_DHDSS			= $00100;
  KeyFilterFlag_RSA			= $00200;

  // versions
  KeyFilterFlag_V4			= $00400;
  KeyFilterFlag_V3			= $00800;

  // X509 types
  KeyFilterFlag_X509Cert		= $01000;
  KeyFilterFlag_X509Root		= $02000;

  // X509 capabilities
  KeyFilterFlag_X509CanEncrypt		= $04000;
  KeyFilterFlag_X509CanSign		= $08000;

  // the groups can be combined
  KeyFilterMask_Boolean			= $000FF;
  KeyFilterMask_Algorithm		= $00F00;
  KeyFilterMask_X509Props		= $0F000;

  // count user IDs of all keys
  KeyFilterFlag_CountUserIDs		= $10000;

const
  // pre-selection flag for key import
  IgnoreFlag_None			= 0;
  IgnoreFlag_ByHexID			= 1;
  IgnoreFlag_ByUserID			= 2;

type
  TPGPKeyOrdering = (
    Invalid_Ordering,
    Any_Ordering,
    UserID_Ordering,
    ReverseUserID_Ordering,
    KeyID_Ordering,
    ReverseKeyID_Ordering,
    Validity_Ordering,
    ReverseValidity_Ordering,
    Trust_Ordering,
    ReverseTrust_Ordering,
    EncryptKeySize_Ordering,
    ReverseEncryptKeySize_Ordering,
    SigKeySize_Ordering,
    ReverseSigKeySize_Ordering,
    Creation_Ordering,
    ReverseCreation_Ordering,
    Expiration_Ordering,
    ReverseExpiration_Ordering
  );
  TTrustLevel = (
    KeyTrust_Undefined,
    KeyTrust_Unknown,
    KeyTrust_Never,
    KeyTrust_Reserved1,
    KeyTrust_Reserved2,
    KeyTrust_Marginal,
    KeyTrust_Complete,
    KeyTrust_Ultimate
  );
  TValidityLevel = (
    Validity_Unknown,
    Validity_Invalid,
    Validity_Marginal,
    Validity_Complete
  );
  TKeyAlgorithm = (
    KeyAlgorithm_Invalid,
    KeyAlgorithm_RSA,
    KeyAlgorithm_RSAEncryptOnly,
    KeyAlgorithm_RSASignOnly,
    KeyAlgorithm_Reserved01,
    KeyAlgorithm_Reserved02,
    KeyAlgorithm_Reserved03,
    KeyAlgorithm_Reserved04,
    KeyAlgorithm_Reserved05,
    KeyAlgorithm_Reserved06,
    KeyAlgorithm_Reserved07,
    KeyAlgorithm_Reserved08,
    KeyAlgorithm_Reserved09,
    KeyAlgorithm_Reserved10,
    KeyAlgorithm_Reserved11,
    KeyAlgorithm_Reserved12,
    KeyAlgorithm_DH,
    KeyAlgorithm_DSS,
    KeyAlgorithm_DHDSS
  );
  TCipherAlgorithm = (
    CipherAlgorithm_None,
    CipherAlgorithm_IDEA,
    CipherAlgorithm_3DES,
    CipherAlgorithm_CAST5,
    CipherAlgorithm_AES128,
    CipherAlgorithm_AES192,
    CipherAlgorithm_AES256,
    CipherAlgorithm_Twofish256
  );
  THashAlgorithm = (
    HashAlgorithm_Invalid,
    HashAlgorithm_MD2,
    HashAlgorithm_MD5,
    HashAlgorithm_SHA,
    HashAlgorithm_RIPEMD160,
    HashAlgorithm_SHA256,
    HashAlgorithm_SHA384,
    HashAlgorithm_SHA512,
    HashAlgorithm_SHA224
  );
  TADKType = (
    NoADK,
    SimpleADK,
    EnforcedADK
  );
  TX509Type = (
    NoCert,
    BadCert,
    StdCert,
    RootCert
  );
  TX509ChainValidity = (
    CertChain_Invalid,
    CertChain_Valid,
    CertChain_Root
  );

  TTypes = (PGPPublic, PGPPrivate);
  TKeyTypes = Array[TTypes] of String;
  TKeyTrusts = Array[TTrustLevel] of String;
  TKeyValidities = Array[TValidityLevel] of String;
  TKeyAlgorithms = Array[TKeyAlgorithm] of String;
  THashAlgorithms = Array[THashAlgorithm] of String;
  TCertValidities = Array[TX509ChainValidity] of String;

  TUserIDs = class(TStringList)
  private
    function  GetChainValidity(Index: Longint): TX509ChainValidity;
    procedure SetChainValidity(Index: Longint; Value: TX509ChainValidity);
    function  GetKeyValidity(Index: Longint): TValidityLevel;
    procedure SetKeyValidity(Index: Longint; Value: TValidityLevel);
    function  GetX509Type(Index: Longint): TX509Type;
    procedure SetX509Type(Index: Longint; Value: TX509Type);
    function  GetX509KeyUsage(Index: Longint): Word;
    procedure SetX509KeyUsage(Index: Longint; Value: Word);
    function  GetX509CanEncrypt(Index: Longint): Longbool;
    function  GetX509CanSign(Index: Longint): Longbool;
  protected
    function  Get(Index: Integer): String; override;
    function  GetCertOwners(Index: Longint): String;
    function  GetCertUserID(Index: Longint): String;
  public
    property  CertOwners[Index: Longint]: String read GetCertOwners;
    property  CertUserIDs[Index: Longint]: String read GetCertUserID;
    property  ChainValidities[Index: Longint]: TX509ChainValidity read GetChainValidity write SetChainValidity;
    property  KeyValidities[Index: Longint]: TValidityLevel read GetKeyValidity write SetKeyValidity;
    property  X509Types[Index: Longint]: TX509Type read GetX509Type write SetX509Type;
    property  X509KeyUsage[Index: Longint]: Word read GetX509KeyUsage write SetX509KeyUsage;
    property  X509CanEncrypt[Index: Longint]: Longbool read GetX509CanEncrypt;
    property  X509CanSign[Index: Longint]: Longbool read GetX509CanSign;
  end;

  TSignerIDs = TStringList;
  TADKeyIDs = TStringList;
  TRevKeyIDs = TStringList;
  TGroupsList = TStringList;

  pKeyPropsRec = ^TKeyPropsRec;
  TKeyPropsRec = Record
    kHexID: String;
    kUserID: String;
    kFingerprint: String;
    kCreaTimeStr: String;
    kExpTimeStr: String;
    kSize: String;
    kX509Cert: TX509Type;
    kKeyAlgorithm: TKeyAlgorithm;
    kHashAlgorithm: THashAlgorithm;
    kTrust: TTrustLevel;
    kValidity: TValidityLevel;
    kCreaTimeNum: Longint;
    kExpTimeNum: Longint;
    kCorrupt: Longbool;
    kPrivate: Longbool;
    kImplicitTrust: Longbool;
    kRevoked: Longbool;
    kDisabled: Longbool;
    kExpired: Longbool;
    kSecShared: Longbool;
    kCanEncrypt: Longbool;
    kCanDecrypt: Longbool;
    kCanSign: Longbool;
    kCanVerify: Longbool;
    kHasRevoker: Longbool;
    kHasADK: TADKType;
    kHasSubKey: Longbool;
    kLegacyKey: Longbool;
    kX509Key: Longbool;
    kUserIDList: TUserIDs;
    kSignerIDList: TSignerIDs;
    kADKeyIDList: TADKeyIDs;
    kRevKeyIDList: TRevKeyIDs;
  end;

  pKeyPropsList = ^TKeyPropsList;
  TKeyPropsList = class(TStringList)
  private
    FDuplicates: Longbool;
    IncludeUserIDs: Longbool;
    IncludeSignerIDs: Longbool;
    IncludeADKeyIDs: Longbool;
    IncludeRevKeyIDs: Longbool;
    function	AllocKeyPropsRec: pKeyPropsRec;
    function	GetKeyProps(Index: Longint): TKeyPropsRec;
    function	FreeKeyPropsRec(Index: Longint): pKeyPropsRec;
  public
    GroupsList:	TGroupsList;
    constructor	Create(PropertyFlags: DWord);
    destructor	Destroy; override;
    procedure	Clear; override;
    procedure 	Changed; override;
    procedure 	Changing; override;
    procedure	Delete(Index: Integer); override;
    procedure	Move(CurIndex, NewIndex: Integer); override;
    function	Add(const S: String): Integer; override;
    function	AddObject(const S: String; AObject: TObject): Integer; override;
    procedure	Insert(Index: Integer; const S: String); override;
    {$IFDEF FIXED}
    procedure	InsertObject(Index: Integer; const S: String; AObject: TObject); override;
    {$ELSE}
    procedure	InsertObject(Index: Integer; const S: String; AObject: TObject);
    {$ENDIF}
    // access indexed KeyProps record like Strings[] or Objects[]
    property	KeyProps[Index: Longint]: TKeyPropsRec read GetKeyProps;
    // just a dummy to prevent using the inherited property
    property	Duplicates: Longbool read FDuplicates;
  end;

function IsEmailAddress(const Text: String): Longbool;
function ExtractEmail(const Text: String): String;
function ExtractName(const Text: String): String;

// extracts standard UserID from the same values as below
function GetUserIDFromLongName(const LongName: String): String;
// extracts CommonName and Email or - if not available - OrganizationName and OrganizationalUnitName
function GetUserFromLongName(const LongName: String; IncludeEmail: Longbool): String;

// Algorithm type conversion utility
function GetTHashAlgorithm(Algorithm: PGPHashAlgorithm): THashAlgorithm;


implementation

uses
  X509Types;


// Utilities -------------------------------------------------------------------

const
  RFC_CHARS = [#33..#90, #94..#126];

function IsEmailAddress(const Text: String): Longbool;
var
  iLen, iATPos, iBegPos, iEndPos: Integer;
begin
  Result := false;
  iLen := Length(Text);
  if iLen > 0 then begin
    iATPos := FirstCharPos(AT, Text);
    if iATPos > 0 then begin
      iBegPos := pred(iATPos);
      while (iBegPos > 0) and (Text[iBegPos] in RFC_CHARS) do dec(iBegPos);
      iEndPos := succ(iATPos);
      while (iEndPos <= iLen) and (Text[iEndPos] in RFC_CHARS) do inc(iEndPos);
      Result := (pred(iEndPos - iBegPos) = iLen)
    end;
  end;
end;

function ExtractEmail(const Text: String): String;
var
  iLen, iATPos, iBegPos, iEndPos: Integer;
begin
  Result := E;
  iLen := Length(Text);
  iATPos := iLen;
  while (iATPos > 0) and (Text[iATPos] <> AT) do dec(iATPos);
  if iATPos > 0 then begin
    iBegPos := pred(iATPos);
    while (iBegPos > 0) and (Text[iBegPos] <> LT) do dec(iBegPos);
    if iBegPos > 0 then begin
      iEndPos := succ(iATPos);
      while (iEndPos <= iLen) and (Text[iEndPos] <> GT) do inc(iEndPos);
      if iEndPos <= iLen then begin
	repeat
	  inc(iBegPos)
	until Text[iBegPos] in RFC_CHARS;
	repeat
	  dec(iEndPos)
	until Text[iEndPos] in RFC_CHARS;
	Result := Copy(Text, iBegPos, succ(iEndPos - iBegPos));
      end;
    end;
    if Result = E then begin
      iBegPos := pred(iATPos);
      while (iBegPos > 0) and (Text[iBegPos] in RFC_CHARS) do dec(iBegPos);
      iEndPos := succ(iATPos);
      while (iEndPos <= iLen) and (Text[iEndPos] in RFC_CHARS) do inc(iEndPos);
      Result := Copy(Text, succ(iBegPos), pred(iEndPos - iBegPos));
    end;
  end;
end;

function ExtractName(const Text: String): String;
var
  iLen, iBegPos, iEndPos, iQUBeg, iQUEnd, iBSPos: Integer;
begin
  Result := E;
  iLen := Length(Text);
  iEndPos := FirstCharPos(LT, Text);
  if iEndPos = 0 then begin
    iEndPos := FirstCharPos(AT, Text);
    while (iEndPos > 0) and (Text[iEndPos] in RFC_CHARS) do dec(iEndPos);
  end
  else dec(iEndPos);
  if iEndPos = 0 then begin
    iBegPos := LastCharPos(LB, Text);
    while (iBegPos > 1) and (Text[pred(iBegPos)] = BS) do iBegPos := PrevCharPos(LB, Text, iBegPos - 2);
    if iBegPos > 0 then begin
      iEndPos := LastCharPos(RB, Text);
      while (iEndPos > iBegPos) and (Text[pred(iEndPos)] = BS) do iEndPos := PrevCharPos(RB, Text, iEndPos - 2);
      inc(iBegPos);
      dec(iEndPos);
    end
    else begin
      iBegPos := LastCharPos(GT, Text);
      if iBegPos = 0 then begin
	iBegPos := FirstCharPos(AT, Text);
	if iBegPos > 0 then while (iBegPos <= iLen) and (Text[iBegPos] in RFC_CHARS) do inc(iBegPos);
      end
      else inc(iBegPos);
      if iBegPos > 0 then iEndPos := iLen;
    end;
  end
  else iBegPos := 1;
  if (iBegPos = 0) and (iEndPos = 0) then begin
    iBegPos := 1;
    iEndPos := iLen;
  end;
  iQUBeg := FirstCharPos(QU, Text);
  while (iQUBeg > 1) and (Text[pred(iQUBeg)] = BS) do iQUBeg := NextCharPos(QU, Text, succ(iQUBeg));
  if iQUBeg > 0 then begin
    iQUEnd := NextCharPos(QU, Text, succ(iQUBeg));
    if iQUEnd > 1 then begin
      while (iQUEnd > iQUBeg) and (Text[pred(iQUEnd)] = BS) do iQUEnd := NextCharPos(QU, Text, succ(iQUEnd));
      iBegPos := succ(iQUBeg);
      iEndPos := pred(iQUEnd);
    end;
  end;
  while (iBegPos <= iEndPos) and (Text[iBegPos] <= SP) do inc(iBegPos);
  while (iEndPos > 0) and (Text[iEndPos] <= SP) do dec(iEndPos);
  iLen := succ(iEndPos - iBegPos);
  if iLen > 0 then begin
    Result := Copy(Text, iBegPos, iLen);
    iBSPos := succ(iLen);
    repeat
      iBSPos := PrevCharPos(BS, Result, pred(iBSPos));
      if iBSPos > 0 then Delete(Result, iBSPos, 1);
    until iBSPos = 0;
  end;
end;

function ExtractValues(const CertIDs, LongName: String; IncludeEmail, UserIDFormat: Longbool): String;
var
  IDBeg, ValBeg, IDsLen, ValLen, IDEnd, ValEnd: Integer;
  ID: String;
begin
  Result := E;
  IDBeg := 1;
  ValBeg := 1;
  IDsLen := Length(CertIDs);
  ValLen := Length(LongName);
  repeat
    IDEnd := NextCharPos(CM, CertIDs, IDBeg);
    if IDEnd = 0 then IDEnd := succ(IDsLen);
    ID := Copy(CertIDs, IDBeg, IDEnd - IDBeg);
    ValBeg := NextStrPos(ID, LongName, ValBeg);
    if ValBeg > 0 then begin
      if UserIDFormat then inc(ValBeg, Length(ID));
      ValEnd := NextCharPos(CM, LongName, ValBeg + Length(ID));
      if ValEnd = 0 then ValEnd := succ(ValLen);
      if ValEnd > ValBeg then begin
	if UserIDFormat then begin
	  if Result = E then
	    Result := Copy(LongName, ValBeg, ValEnd - ValBeg)
	  else Result := Result + SP + LT + Copy(LongName, ValBeg, ValEnd - ValBeg) + GT;
	end
	else begin
	  if Result = E then
	    Result := Copy(LongName, ValBeg, ValEnd - ValBeg)
	  else Result := Result + CM + SP + Copy(LongName, ValBeg, ValEnd - ValBeg);
	end;
      end;
      ValBeg := succ(ValEnd);
    end;
    if IncludeEmail then
      IDBeg := succ(IDEnd)
    else Break;
  until IDBeg > IDsLen;
end;

function GetUserIDFromLongName(const LongName: String): String;
begin
  Result := ExtractValues('CN=,EMAIL=', LongName, true, true);
  if Result = E then Result := ExtractValues('O=,OU=', LongName, true, true);
end;

function GetUserFromLongName(const LongName: String; IncludeEmail: Longbool): String;
begin
  Result := ExtractValues('CN=,EMAIL=', LongName, IncludeEmail, false);
  if Result = E then Result := ExtractValues('O=,OU=', LongName, IncludeEmail, false);
end;

function GetTHashAlgorithm(Algorithm: PGPHashAlgorithm): THashAlgorithm;
begin
  Result := THashAlgorithm(
    Algorithm -
    ord(Algorithm> kPGPHashAlgorithm_RIPEMD160) * pred(kPGPHashAlgorithm_SHA256 - kPGPHashAlgorithm_RIPEMD160) + 1
  );
end;


// TUserIDs --------------------------------------------------------------------

function TUserIDs.Get(Index: Integer): String;
begin
  if X509Types[Index] > BadCert then
    Result := GetUserIDFromLongName(GetCertOwners(Index))
  else Result := inherited Get(Index);
end;

function TUserIDs.GetCertOwners(Index: Longint): String;
begin
  Result := inherited Get(Index);
  System.Delete(Result, 1, FirstCharPos(#9, Result));
end;

function TUserIDs.GetCertUserID(Index: Longint): String;
begin
  Result := inherited Get(Index);
  System.Delete(Result, FirstCharPos(#9, Result), MAXINT);
end;

function TUserIDs.GetChainValidity(Index: Longint): TX509ChainValidity;
begin
  Result := TX509ChainValidity(DWord(Objects[Index]) and $000000FF);
end;

procedure TUserIDs.SetChainValidity(Index: Longint; Value: TX509ChainValidity);
begin
  Objects[Index] := TObject((DWord(Objects[Index]) and $FFFFFF00) or DWord(Value));
end;

function TUserIDs.GetKeyValidity(Index: Longint): TValidityLevel;
begin
  Result := TValidityLevel((DWord(Objects[Index]) and $0000FF00) shr 8);
end;

procedure TUserIDs.SetKeyValidity(Index: Longint; Value: TValidityLevel);
begin
  Objects[Index] := TObject((DWord(Objects[Index]) and $FFFF00FF) or (DWord(Value) shl 8));
end;

function TUserIDs.GetX509Type(Index: Longint): TX509Type;
begin
  Result := TX509Type((DWord(Objects[Index]) and $00FF0000) shr 16);
end;

procedure TUserIDs.SetX509Type(Index: Longint; Value: TX509Type);
begin
  Objects[Index] := TObject((DWord(Objects[Index]) and $FF00FFFF) or (DWord(Value) shl 16));
end;

function TUserIDs.GetX509KeyUsage(Index: Longint): Word;
begin
  Result := (DWord(Objects[Index]) and $FF000000) shr 24;
end;

procedure TUserIDs.SetX509KeyUsage(Index: Longint; Value: Word);
begin
  Objects[Index] := TObject((DWord(Objects[Index]) and $00FFFFFF) or (DWord(Value) shl 24));
end;

function TUserIDs.GetX509CanEncrypt(Index: Longint): Longbool;
begin
  Result := (GetX509KeyUsage(Index) and x509KeyUsage_dataEncipherment <> 0);
end;

function TUserIDs.GetX509CanSign(Index: Longint): Longbool;
begin
  Result := (GetX509KeyUsage(Index) and x509KeyUsage_digitalSignature <> 0);
end;


// TKeyPropsList ---------------------------------------------------------------

function TKeyPropsList.AllocKeyPropsRec: pKeyPropsRec;
begin
  New(Result);
  if Result <> nil then begin
    FillChar(Result^, SizeOf(TKeyPropsRec), 0);
    with Result^ do begin
      if IncludeRevKeyIDs then begin
	kRevKeyIDList := TRevKeyIDs.Create;
	kRevKeyIDList.BeginUpdate;
      end;
      if IncludeADKeyIDs then begin
	kADKeyIDList := TADKeyIDs.Create;
	kADKeyIDList.BeginUpdate;
      end;
      if IncludeSignerIDs then begin
	kSignerIDList := TSignerIDs.Create;
	kSignerIDList.BeginUpdate;
      end;
      if IncludeSignerIDs or IncludeUserIDs then begin
	kUserIDList := TUserIDs.Create;
	kUserIDList.BeginUpdate;
      end;
    end;
  end;
end;

function TKeyPropsList.FreeKeyPropsRec(Index: Longint): pKeyPropsRec;
begin
  Result := pKeyPropsRec(Objects[Index]);
  if Result <> nil then begin
    Result^.kUserIDList.Free;
    Result^.kSignerIDList.Free;
    Result^.kADKeyIDList.Free;
    Result^.kRevKeyIDList.Free;
    Dispose(Result);
    Result := nil;
  end;
end;

function TKeyPropsList.GetKeyProps(Index: Longint): TKeyPropsRec;
begin
  Result := pKeyPropsRec(Objects[Index])^;
end;

constructor TKeyPropsList.Create(PropertyFlags: DWord);
begin
  inherited Create;
  BeginUpdate;
  GroupsList := TGroupsList.Create;
  GroupsList.BeginUpdate;
  IncludeUserIDs := ((PropertyFlags and spgpKeyPropFlag_IncludeUserIDs) <> 0);
  IncludeSignerIDs := ((PropertyFlags and spgpKeyPropFlag_IncludeSignerIDs) <> 0);
  IncludeADKeyIDs := ((PropertyFlags and spgpKeyPropFlag_HasADK) <> 0);
  IncludeRevKeyIDs := ((PropertyFlags and spgpKeyPropFlag_HasRevoker) <> 0);
  if IncludeSignerIDs then IncludeUserIDs := true;
end;

destructor TKeyPropsList.Destroy;
var
  Index: Integer;
begin
  try
    for Index := 0 to pred(Count) do FreeKeyPropsRec(Index);
  finally
    try
      GroupsList.Free;
    finally
      inherited Destroy;
    end;
  end;
end;

procedure TKeyPropsList.Clear;
var
  Index: Integer;
begin
  try
    for Index := 0 to pred(Count) do FreeKeyPropsRec(Index);
  finally
    try
      GroupsList.Clear;
    finally
      inherited Clear;
    end;
  end;
end;

procedure TKeyPropsList.Changed;
begin
  // save clock cycles
end;

procedure TKeyPropsList.Changing;
begin
  // save clock cycles
end;

procedure TKeyPropsList.Delete(Index: Integer);
begin
  FreeKeyPropsRec(Index);
  inherited Delete(Index);
end;

procedure TKeyPropsList.Move(CurIndex, NewIndex: Integer);
var
  TempStr: String;
  TempObj: TObject;
begin
  if CurIndex <> NewIndex then begin
    TempStr := Strings[CurIndex];
    TempObj := Objects[CurIndex];
    inherited Delete(CurIndex);
    inherited Insert(NewIndex, TempStr);
    Objects[NewIndex] := TempObj;
  end;
end;

function TKeyPropsList.Add(const S: String): Integer;
begin
  Result := inherited Add(S);
  if Objects[Result] = nil then Objects[Result] := TObject(AllocKeyPropsRec);
end;

function TKeyPropsList.AddObject(const S: String; AObject: TObject): Integer;
var
  CurrentCount: Integer;
begin
  CurrentCount := Count;
  {$IFNDEF FIXED}
  Result := inherited Add(S);
  {$ELSE}
  Result := inherited AddObject(S, TObject(AllocKeyPropsRec));
  {$ENDIF}
  if Count = CurrentCount then Exit;
  if AObject <> nil then pKeyPropsRec(Objects[Result])^ := pKeyPropsRec(AObject)^;
end;

procedure TKeyPropsList.Insert(Index: Integer; const S: String);
begin
  InsertObject(Index, S, nil);
end;

procedure TKeyPropsList.InsertObject(Index: Integer; const S: String; AObject: TObject);
begin
  {$IFNDEF FIXED}
  inherited Insert(Index, S);
  Objects[Index] := TObject(AllocKeyPropsRec);
  {$ELSE}
  inherited InsertObject(Index, S, TObject(AllocKeyPropsRec));
  {$ENDIF}
  if AObject <> nil then pKeyPropsRec(Objects[Index])^ := pKeyPropsRec(AObject)^;
end;


end.

