{$J+,Z4}
unit PGPDecode;

{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  Windows,
  Classes,
  StrTools,
  SysUtils,
  SecureFuncs,
  PSMimeTools,
  KeyPropTypes,
  X509Types,
  UTF8,
  pgpBase,
  pgpErrors,
  pgpPubTypes,
  pgpUtilities,
  pgpOptionList,
  pgpMemoryMgr,
  pgpEvents,
  pgpKeys,
  pgpTLS,
  pgpSC,
  KeyFuncs,
  PrefFuncs,
  X509Funcs,
  PGPDialogs,
  KeyDialog,
  HelpCtrl;

type
  TSigStatus = (
    SIGNED_NOT,
    SIGNED_BAD,
    SIGNED_GOOD,
    SIGNED_NO_KEY,
    SIGNED_UNKNOWN_ALG,
    SIGNED_SIG_EXPIRED
  );
  TSigPropsRec = Record
    sStatus: TSigStatus;
    sHexID: String;
    sUserID: String;
    sSubKeyHexID: String;
    sCreaTimeStr: String;
    sCreaTimeNum: DWord;
    sChecked: Longbool;
    sVerified: Longbool;
    sValidity: TValidityLevel;
    sRevoked: Longbool;
    sDisabled: Longbool;
    sExpired: Longbool;
    sDetached: Longbool;
    sAxiomatic: Longbool;
    sExpTimeNum: Longint;
    sExpTimeStr: String;
    sHashAlgorithm: THashAlgorithm;
    sValidCertChain: TX509ChainValidity;
  end;
  TAnalysis = (
    ENCRYPTED,
    SIGNED,
    DETACHED_SIGNATURE,
    KEY,
    UNKNOWN,
    X509_CERTIFICATE,
    X509_SECRETKEY,
    SMIME_UNRECOGNIZED,
    SMIME_ENCRYPTED,
    SMIME_SIGNED
  );
  TPassOption = (
    AnalyseOnly,
    AnalysePass,
    DecodePass
  );
  TX509Mode = (
    X509Off,
    X509Secret,
    X509Binary,
    X509Armored,
    X509IsMime,
    X509Signed,
    X509Encrypted,
    X509Processed
  );
  TOnShowProgress = procedure(BytesProcessed, BytesTotal: Longint) of Object;
  TPGPDecode = Class
  private
    // internal
    FContext: pPGPContext;
    FKeySetMain: pPGPKeySet;
    FAllocatedOutputBuffer: PChar;
    FActualOutputSize: PGPSize;
    FOutputOptions: pPGPOptionList;
    FInputSize: PGPSize;
    FInputBuffer: PChar;
    FInputFileName: String;
    FFileNameBuffer: String;
    FClearSigned: Longbool;
    FDetachedSigVerify: Longbool;
    FEyesOnly: Longbool;
    FFileToFile: Longbool;
    FOutputExists: Longbool;
    FIsPGPMime: Longbool;
    FX509Mode: TX509Mode;
    FRecognized: Longbool;
    FRecurring: Longbool;
    FIsConventional: Longbool;
    FPassOption: TPassOption;
    FPassphrase: PChar;
    FPassSection: PGPUInt32;
    FRecipientsData: TPGPEventRecipientsData;
    FSectionNumber: PGPUInt32;
    FSignedDataBuffer: String;
    FDecryptionKey: pPGPKey;
    FSecureBuffer: PChar;
    FSecureBufferSize: PGPSize;
    FSignCertProps: TX509CertPropsRec;
    FValidCertChain: TX509ChainValidity;
    FWait: HCursor;
    FArrow: HCursor;
    // public properties
    FAnalysis: TAnalysis;
    FCharset: String;
    FKeyAddResult: Longint;
    FKeyData: String;
    FMultiDecode: Longbool;
    FSMimeCertBag: String;
    FParentHandle: THandle;
    FOutputBuffer: String;
    FSigPropsRec: TSigPropsRec;
    FKeyPropsList: TKeyPropsList;
    // published properties
    FCheckSig: Longbool;
    FEnforceFileStreaming: Longbool;
    FFileOutput: Longbool;
    FGetKeyFromServer: Longbool;
    FLogBadSignature: String;
    FLogExpSignature: String;
    FLogKeyDisabled: String;
    FLogKeyExpired: String;
    FLogKeyRevoked: String;
    FLogKeyUnknown: String;
    FLogUnknownAlg: String;
    FPassThroughClearSigned: Longbool;
    FPassThroughIfUnrecognized: Longbool;
    FPassThroughKeys: Longbool;
    {FProgressInterval: Cardinal;}
    FQueryAddKeys: Longbool;
    FRecursivelyDecode: Longbool;
    FShowSigLog: Longbool;
    FKeyDlgPrompt: String;
    FPassDlgPrompt: String;
    FOutputFileName: String;
    FKeyDialogHelpIndex: Longint;
    FHelpHandler: THelpHandler;
    FSecure: TSecure;
    // events
    {FOnShowProgress: TOnShowProgress;}
    procedure SetMultiDecode(Value: Longbool);
    procedure ClearSigPropsRec;
    procedure ClearSignCertProps;
    function  InitDecode: PGPError;
    procedure FinitDecode;
    function  HandleFileInput: PGPError;
    function  X509Test: TX509Mode;
    function  SecureMove: PGPError;
    function  ShowSigInfo: PGPError;
    function  MaxFileSize: DWord;
    function  FileSize(const FileName: String; var Bytes: DWord): PGPError;
    function  ReadInputFile: PGPError;
    function  GetOutputFileName: String;
    function  Analyze(InFileSpec: pPGPFileSpec): PGPError;
    function  FixNonAsciiBug(const Data: PChar): PGPError;
    function  GetOptionList(var OptionList: pPGPOptionList): PGPError;
    function  SetOutputOption(var OptionList: pPGPOptionList): PGPError;
    function  PrepareOutput(const Input: PChar; InputSize: DWord; const MimeHeaders, MimeBody: String): PGPError;
    function  DoRecursion(const Input: PChar; InputSize: DWord): PGPError;
    function  SecureView(const Output: PChar; OutputSize: DWord): PGPError;
    function  WriteOutputFile(Output: PChar; OutputSize: DWord): PGPError;
    function  Decode(IsFile: Longbool): Longint;
    function  AnalyzeHandler(Event: pPGPEvent): PGPError;
    function  RecipientsHandler(Event: pPGPEvent): PGPError;
    function  PassphraseHandler(Event: pPGPEvent): PGPError;
    function  KeyFoundHandler(Event: pPGPEvent): PGPError;
    function  OutputHandler(Event: pPGPEvent): PGPError;
    function  SignatureHandler(Event: pPGPEvent): PGPError;
    function  EndLexHandler(Event: pPGPEvent): PGPError;
    function  DecryptionHandler(Event: pPGPEvent): PGPError;
  public
    constructor Create;
    destructor Destroy; override;
    function AnalyseBuffer(const DataBuffer: String): Longint;
    function AnalyseFile(const FileName: String): Longint;
    function AnalyseMimeBuffer(const DataBuffer: String): Longint;
    function AnalyseMimeFile(const FileName: String): Longint;
    function DecodeBuffer(const DataBuffer: String): Longint;
    function DecodeFile(const FileName: String): Longint;
    function DecodeMimeBuffer(const DataBuffer: String): Longint;
    function DecodeMimeFile(const FileName: String): Longint;
    property Analysis: TAnalysis
      read FAnalysis;
    property Charset: String
      read FCharset;
    property KeyAddResult: Longint
      read FKeyAddResult;
    property KeyData: String
      read FKeyData
      write FKeyData;
    property MultiDecode: Longbool
      read FMultiDecode
      write SetMultiDecode;
    property SMimeCertBag: String
      read FSMimeCertBag;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property OutputBuffer: String
      read FOutputBuffer;
    property SigPropsRec: TSigPropsRec
      read FSigPropsRec;
    property KeyPropsList: TKeyPropsList
      read FKeyPropsList;
  published
    property CheckSig: Longbool
      read FCheckSig
      write FCheckSig;
    property EnforceFileStreaming: Longbool
      read FEnforceFileStreaming
      write FEnforceFileStreaming;
    property FileOutput: Longbool
      read FFileOutput
      write FFileOutput;
    property GetKeyFromServer: Longbool
      read FGetKeyFromServer
      write FGetKeyFromServer;
    property LogBadSignature: String
      read FLogBadSignature
      write FLogBadSignature;
    property LogExpSignature: String
      read FLogExpSignature
      write FLogExpSignature;
    property LogKeyDisabled: String
      read FLogKeyDisabled
      write FLogKeyDisabled;
    property LogKeyExpired: String
      read FLogKeyExpired
      write FLogKeyExpired;
    property LogKeyRevoked: String
      read FLogKeyRevoked
      write FLogKeyRevoked;
    property LogKeyUnknown: String
      read FLogKeyUnknown
      write FLogKeyUnknown;
    property LogUnknownAlg: String
      read FLogUnknownAlg
      write FLogUnknownAlg;
    property PassThroughClearSigned: Longbool
      read FPassThroughClearSigned
      write FPassThroughClearSigned;
    property PassThroughIfUnrecognized: Longbool
      read FPassThroughIfUnrecognized
      write FPassThroughIfUnrecognized;
    property PassThroughKeys: Longbool
      read FPassThroughKeys
      write FPassThroughKeys;
    {property ProgressInterval: Cardinal
      read FProgressInterval
      write FProgressInterval;}
    property QueryAddKeys: Longbool
      read FQueryAddKeys
      write FQueryAddKeys;
    property RecursivelyDecode: Longbool
      read FRecursivelyDecode
      write FRecursivelyDecode;
    property SignedDataBuffer: String
      read FSignedDataBuffer
      write FSignedDataBuffer;
    property ShowSigLog: Longbool
      read FShowSigLog
      write FShowSigLog;
    property KeyDlgPrompt: String
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property PassDlgPrompt: String
      read FPassDlgPrompt
      write FPassDlgPrompt;
    property OutputFileName: String
      read FOutputFileName
      write FOutputFileName;
    property KeyDialogHelpIndex: Longint
      read FKeyDialogHelpIndex
      write FKeyDialogHelpIndex;
    property HelpHandler: THelpHandler
      read FHelpHandler
      write FHelpHandler;
    property Secure: TSecure
      read FSecure
      write FSecure;
    {property OnShowProgress: TOnShowProgress
      read FOnShowProgress
      write FOnShowProgress;}
  end;


implementation

function DataCopy(var Output: PChar; const Input: PChar; InputSize: DWord): PGPError;
begin
  Result := kPGPError_OutOfMemory;
  if InputSize > 0 then begin
    Output := PGPNewData(PGPGetDefaultMemoryMgr, InputSize, kPGPMemoryMgrFlags_None);
    if Output <> nil then begin
      Move(Input[0], Output[0], InputSize);
      Result := kPGPError_NoErr;
    end;
  end;
end;

function EventHandler(Context: pPGPContext; Event: pPGPEvent; UserValue: PGPUserValue): PGPError; cdecl;
begin
  Result := 0;
  with TPGPDecode(UserValue) do begin
    case Event^.EType of
      kPGPEvent_NullEvent:		{if (FPassOption = DecodePass)
					and not (FDetachedSigVerify or FClearSigned) then begin
					  if TMethod(FOnShowProgress).Code <> nil then begin
					    // BytesTotal always stays 0 => use FInputSize (at least in 6.5.X)
					    with Event^.EData.NullData do FOnShowProgress(BytesWritten, FInputSize);
					  end;
					  ProcessMessages;
					end};
      kPGPEvent_InitialEvent:		;
      kPGPEvent_FinalEvent:		;
      kPGPEvent_ErrorEvent:	  	Result := Event^.EData.ErrorData.Error;
      kPGPEvent_WarningEvent:		;
      kPGPEvent_PassphraseEvent:        Result := PassphraseHandler(Event);
      kPGPEvent_AnalyzeEvent:		Result := AnalyzeHandler(Event);
      kPGPEvent_RecipientsEvent:  	if FPassOption = DecodePass then Result := RecipientsHandler(Event);
      kPGPEvent_KeyFoundEvent:          if FPassOption = DecodePass then Result := KeyFoundHandler(Event);
      kPGPEvent_OutputEvent:		if FPassOption = DecodePass then OutputHandler(Event);
      kPGPEvent_SignatureEvent:         if FPassOption = DecodePass then Result := SignatureHandler(Event);
      kPGPEvent_BeginLexEvent:	  	FSectionNumber := succ(Event^.EData.BeginLexData.SectionNumber);
      kPGPEvent_EndLexEvent:      	Result := EndLexHandler(Event);
      kPGPEvent_DetachedSignatureEvent:	;
      kPGPEvent_DecryptionEvent:	Result := DecryptionHandler(Event);
    end;
  end;
end;

constructor TPGPDecode.Create;
begin
  inherited;
  {FProgressInterval := 1000;}
  FWait := LoadCursor(0, IDC_WAIT);
  FArrow := LoadCursor(0, IDC_ARROW);
  FKeyPropsList := TKeyPropsList.Create(spgpKeyPropFlag_IDFlags);
end;

destructor TPGPDecode.Destroy;
begin
  FKeyPropsList.Free;
  inherited Destroy;
end;

procedure TPGPDecode.SetMultiDecode(Value: Longbool);
begin
  if Value <> FMultiDecode then begin
    if Value = true then
      FMultiDecode := (KeyRings.InitKeyRings(FContext, FKeySetMain) = 0)
    else begin
      FMultiDecode := false;
      PGPFreeData(FPassphrase);
      FPassphrase := nil;
      KeyRings.FreeKeyRings;
    end;
  end;
end;

procedure TPGPDecode.ClearSigPropsRec;
begin
  with FSigPropsRec do begin
    sStatus := SIGNED_NOT;
    sHexID := E;
    sUserID := E;
    sSubKeyHexID := E;
    sCreaTimeStr := E;
    sCreaTimeNum := 0;
    sChecked := false;
    sVerified := false;
    sValidity := Validity_Unknown;
    sRevoked := false;
    sDisabled := false;
    sExpired := false;
    sDetached := false;
    sAxiomatic := false;
    sExpTimeNum := 0;
    sExpTimeStr := E;
    sHashAlgorithm := HashAlgorithm_Invalid;
    sValidCertChain := CertChain_Invalid;
  end;
end;

procedure TPGPDecode.ClearSignCertProps;
begin
  with FSignCertProps do begin
    x509CertificateID := E;
    x509IssuerKeyAlgorithm := KeyAlgorithm_Invalid;
    x509IssuerHashAlgorithm := HashAlgorithm_Invalid;
    x509IssuerLongName := E;
    x509IssuerSerialNumber := E;
    x509OwnerKeyHexID := E;
    x509OwnerCertUserID := E;
    x509OwnerLongName := E;
    x509OwnerSerialNumber := E;
    x509CreaTimeStr := E;
    x509ExpTimeStr := E;
    x509CreaTimeNum := 0;
    x509ExpTimeNum := 0;
    x509IsVerified := false;
    x509IsCorrupt := false;
    x509IsRevoked := false;
    x509IsExpired := false;
    x509IsRoot := false;
  end;
end;

function TPGPDecode.InitDecode: PGPError;
begin
  Result := kPGPError_NoErr;
  FAnalysis := UNKNOWN;
  FSectionNumber := 0;
  FOutputBuffer := E;
  FActualOutputSize := 0;
  FAllocatedOutputBuffer := nil;
  FSecureBuffer := nil;
  FSecureBufferSize := 0;
  FSignedDataBuffer := E;
  FClearSigned := false;
  FDetachedSigVerify := false;
  FIsConventional := false;
  FIsPGPMime := false;
  FRecognized := false;
  FDecryptionKey := nil;
  FPassSection := 0;
  if not FRecurring then begin
    FCharset := E;
    FEyesOnly := false;
    FKeyAddResult := 0;
    FSMimeCertBag := E;
    FKeyPropsList.Clear;
    ClearSignCertProps;
    ClearSigPropsRec;
    FValidCertChain := CertChain_Invalid;
  end;
  FillChar(FRecipientsData, SizeOf(TPGPEventRecipientsData), 0);
  if not FMultiDecode then Result := KeyRings.InitKeyRings(FContext, FKeySetMain);
end;

procedure TPGPDecode.FinitDecode;
var
  FoundKey	: pPGPKey;
  KeyIndex	: Integer;
begin
  FX509Mode := X509Off;
  with FKeyPropsList do begin
    if (Count <> 0) or (FAnalysis = X509_CERTIFICATE) then KeyRings.UpdateKeyRings;
    if (Count > 0) and (IsHexID(FSigPropsRec.sHexID) <> 0) then begin
      GetKeyByHexID(FKeySetMain, FSigPropsRec.sHexID, FoundKey);
      if FoundKey = nil then begin
	KeyIndex := IndexOf(FSigPropsRec.sHexID);
	if KeyIndex > -1 then Delete(KeyIndex);
      end;
    end;
  end;
  if not FMultiDecode then SetCursor(FArrow);
  with FRecipientsData do begin
    PGPFreeData(KeyIDArray);
    PGPFreeKeySet(RecipientSet);
  end;
  if not FMultiDecode then KeyRings.FreeKeyRings;
end;

function TPGPDecode.HandleFileInput: PGPError;
begin
  Result := 0;
  if not FileExists(FInputBuffer) then
    Result := kPGPError_FileNotFound
  else FInputFileName := FInputBuffer;
  if Result <> 0 then Exit;
  Result := FileSize(FInputFileName, FInputSize);
  if (SDKVersion >= PGP7XX) and ((FFileOutput and FEnforceFileStreaming) or (FInputSize > MaxFileSize)) then begin
    if FX509Mode >= X509IsMime then
      Result := kPGPError_FeatureNotAvailable
    else FFileToFile := FFileOutput;
  end
  else begin
    if FFileOutput and FEnforceFileStreaming then
      FFileToFile := true
    else FFileToFile := false;
  end;
end;

function TPGPDecode.X509Test: TX509Mode;
var
  Pos		: DWord;
  Error		: PGPError;
  OptionList	: pPGPOptionList;
  TestDB	: pPGPKeyDB;
begin
  Result := X509Off;
  if SDKVersion >= PGP81X then begin
    Pos := DWord(StrPos(FInputBuffer, X509Cert_Beg));
    if Pos > 0 then begin
      dec(Pos, DWord(FInputBuffer));
      if ((Pos = 0) or (FInputBuffer[pred(Pos)] = #10)) then Result := X509Armored;
    end;
    if Result = X509Off then begin
      TestDB := nil;
      try
	// test for PKCS7 format
	OptionList := nil;
	try
	  Error := PGPBuildOptionList(FContext, OptionList,
	    [
	      PGPOInputFormat(FContext, kPGPInputFormat_X509DataInPKCS7),
	      PGPOInputBuffer(FContext, FInputBuffer, FInputSize)
	    ]);
	  if (Error = 0) and (PGPImport(FContext, TestDB, OptionList, PGPOLastOption(FContext)) = 0) then begin
	    Result := X509Binary;
	  end;
	finally
	  PGPFreeOptionList(OptionList);
	end;
	if Result = X509Off then begin
	  // test for PKCS12 format
	  OptionList := nil;
	  try
	    Error := PGPBuildOptionList(FContext, OptionList,
	      [
		PGPOInputFormat(FContext, kPGPInputFormat_PKCS12),
		PGPOInputBuffer(FContext, FInputBuffer, FInputSize)
	      ]);
	    if (Error = 0)
	    and (PGPImport(FContext, TestDB, OptionList, PGPOLastOption(FContext)) = kPGPError_BadPassphrase) then begin
	      Result := X509Secret;
	    end;
	  finally
	    PGPFreeOptionList(OptionList);
	  end;
	end;
      finally
	PGPFreeKeyDB(TestDB);
      end;
    end;
  end;
end;

function TPGPDecode.SecureMove: PGPError;
begin
  Result := kPGPError_NoErr;
  if FActualOutputSize <> 0 then begin
    try
      if FEyesOnly and (FSecureBuffer = nil) then begin
	Result := kPGPError_OutOfMemory;
	FSecureBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr,
					  FSecureBufferSize + FActualOutputSize,
					  kPGPMemoryMgrFlags_None);
	if FSecureBuffer <> nil then Result := 0;
      end
      else begin
    	if (FSecureBuffer = nil) and (FAllocatedOutputBuffer <> nil) then begin
	  FSecureBuffer := FAllocatedOutputBuffer;
	  FSecureBufferSize := FActualOutputSize;
	  Exit;
	end;
	Result := PGPReallocData(PGPGetDefaultMemoryMgr,
				 Pointer(FSecureBuffer),
				 FSecureBufferSize + FActualOutputSize,
				 kPGPMemoryMgrFlags_None);
      end;
      if Result = 0 then begin
	Move(FAllocatedOutputBuffer[0], FSecureBuffer[FSecureBufferSize], FActualOutputSize);
	if FEyesOnly then FillChar(FAllocatedOutputBuffer^, FActualOutputSize, 0);
	inc(FSecureBufferSize, FActualOutputSize);
      end;
    finally
      if FSecureBuffer <> FAllocatedOutputBuffer then PGPFreeData(FAllocatedOutputBuffer);
      FAllocatedOutputBuffer := nil;
      FActualOutputSize := 0;
    end;
  end;
end;

function TPGPDecode.ShowSigInfo: PGPError;
var
  Status	: String;
  Source	: String;
  Dest		: String;
  ID		: String;
  Validity	: String;
  Info		: String;
begin
  Result := 0;
  with FSigPropsRec do begin
    if sStatus <> SIGNED_NOT then begin
      // don't change these (constant) PGP strings, see SigEvent.c
      if sStatus = SIGNED_GOOD then
	Status := 'Good Signature' + #10
      else Status := 'Bad Signature' + #10;
      // these strings may be customized
      if FInputFileName <> E then begin
	if FDetachedSigVerify or (FX509Mode = X509Signed) then
	  Source := ExtractFileName(FOutputFileName) + #10
	else if FRecurring then
	  Source := ExtractFileName(FFileNameBuffer) + #10
	else Source := ExtractFileName(FInputFileName) + #10;
	if FFileOutput and (FOutputFileName <> E) then begin
	  Dest := ExtractFileName(FOutputFileName);
	  if Dest <> E then Insert(' => ' + Dest, Source, Length(Source));
	end;
      end
      else Source := '(...)' + #10;
      if (sStatus = SIGNED_NO_KEY) or (sUserID = E) then begin
	if SDKVersion >= PGP8XX then
	  ID := '(' + sHexID + ')' + #10 + ShortHexID(sHexID) + #10
	else ID := '(' + ShortHexID(sHexID) + ')' + #10;
      end
      else begin
	if SDKVersion >= PGP8XX then
	  ID := sUserID + #10 + ShortHexID(sHexID) + #10
	else ID := sUserID + #10;
      end;
      // don't change these (constant) PGP strings, see SigEvent.c
      if sAxiomatic then
	Validity := 'Implicit Trust' + #10 + sCreaTimeStr
      else begin
	case sValidity of
	  Validity_Unknown, Validity_Invalid: Validity := 'Invalid Key' + #10 + sCreaTimeStr;
	  Validity_Marginal: Validity := 'Marginal Key' + #10 + sCreaTimeStr;
	  Validity_Complete: Validity := 'Valid Key' + #10 + sCreaTimeStr;
	end;
      end;
      // these strings may be customized
      if sUserID = E then
	Info := FLogKeyUnknown
      else begin
	if sStatus <> SIGNED_UNKNOWN_ALG then begin
	  if sVerified then begin
	    if sStatus = SIGNED_SIG_EXPIRED then
	      Info := FLogExpSignature
	    else begin
	      if sRevoked then Info := FLogKeyRevoked;
	      if sDisabled then Info := FLogKeyDisabled;
	      if sExpired then Info := FLogKeyExpired;
	    end;
	  end
	  else Info := FLogBadSignature;
	end
	else Info := FLogUnknownAlg;
      end;
      if Info <> E then Info := ' (' + Info + ')';
      if not SendPGPlogMsg(FParentHandle, Status + Source + ID + Validity + Info) then begin
	Result := kPGPError_FeatureNotAvailable;
      end;
    end;
  end;
end;

function TPGPDecode.MaxFileSize: DWord;
var
  MemoryStatus	: TMemoryStatus;
begin
  MemoryStatus.dwLength := SizeOf(TMemoryStatus);
  GlobalMemoryStatus(MemoryStatus);
  Result := MemoryStatus.dwAvailPhys shr 2;
end;

function TPGPDecode.FileSize(const FileName: String; var Bytes: DWord): PGPError;
var
  FileHandle	: THandle;
  FileSizeHigh	: DWord;
begin
  Result := 0;
  FileHandle := FileOpen(FileName, fmOpenRead or fmShareDenyNone);
  if FileHandle <> INVALID_HANDLE_VALUE then begin
    Bytes := GetFileSize(FileHandle, @FileSizeHigh);
    FileClose(FileHandle);
  end
  else Bytes := $FFFFFFFF;
  if (Bytes = 0) or (Bytes = $FFFFFFFF) or (FileSizeHigh <> 0) then Result := kPGPError_ReadFailed;
end;

function TPGPDecode.ReadInputFile: PGPError;
var
  InputFile	: THandle;
  BytesRead	: DWord;
begin
  FInputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(FInputSize), kPGPMemoryMgrFlags_None);
  if FInputBuffer <> nil then begin
    FInputBuffer[FInputSize] := #0;
    InputFile := CreateFile(PChar(FInputFileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING,
			    FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if InputFile <> INVALID_HANDLE_VALUE then begin
      try
	if not ReadFile(InputFile, FInputBuffer[0], FInputSize, BytesRead, nil) or (BytesRead <> FInputSize) then
	  Result := kPGPError_ReadFailed
	else Result := kPGPError_NoErr;
      finally
	CloseHandle(InputFile);
      end;
    end
    else Result := kPGPError_CantOpenFile;
  end
  else begin
    Result := kPGPError_OutOfMemory;
    Exit;
  end;
end;

function TPGPDecode.GetOutputFileName: String;
var
  FileExtension	: String;
begin
  if FDetachedSigVerify then begin
    if FOutputFileName = E then begin
      FOutputFileName := Trim(FInputFileName);
      FileExtension := LowerCase(ExtractFileExt(FOutputFileName));
      if (FileExtension = '.sig') or (FileExtension = '.asc') then begin
	Delete(FOutputFileName, Length(FOutputFileName) - 3, 4);
      end;
    end;
  end
  else begin
    if FOutputFileName = E then begin
      FOutputFileName := Trim(FInputFileName);
      FileExtension := LowerCase(ExtractFileExt(FOutputFileName));
      if (FileExtension = '.pgp') or (FileExtension = '.gpg') or (FileExtension = '.asc') then begin
	Delete(FOutputFileName, Length(FOutputFileName) - 3, 4);
      end;
    end;
  end;
  Result := FOutputFileName;
end;

function TPGPDecode.Analyze(InFileSpec: pPGPFileSpec): PGPError;
var
  OptionList	: pPGPOptionList;
begin
  OptionList := nil;
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPODiscardOutput(FContext, PGPTrue),
      PGPOEventHandler(FContext, EventHandler, Self)
    ]);
  if Result <> 0 then Exit;
  if InFileSpec <> nil then begin
    Result := PGPAppendOptionList(OptionList, [PGPOInputFile(FContext, InFileSpec)]);
  end
  else if FInputSize <> 0 then begin
    Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(FContext, FInputBuffer, FInputSize)]);
  end;
  if Result <> 0 then Exit;
  try
    Result := pgpEvents.PGPDecode(FContext, OptionList, PGPOLastOption(FContext));
    if Result = kPGPError_UserAbort then Result := 0;
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function TPGPDecode.FixNonAsciiBug(const Data: PChar): PGPError;
begin
  // try to circumvent errors with some non-ASCII data
  inc(FInputSize);
  if FEyesOnly then
    FInputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, FInputSize + 2, kPGPMemoryMgrFlags_None)
  else FInputBuffer := PGPNewData(PGPGetDefaultMemoryMgr, FInputSize + 2, kPGPMemoryMgrFlags_None);
  if FInputBuffer <> nil then begin
    Move(Data[0], FInputBuffer[1], FInputSize);
    FInputBuffer[succ(FInputSize)] := #0;
    FInputBuffer[0] := '@';
    Result := 0;
  end
  else Result := kPGPError_OutOfMemory;
end;

function TPGPDecode.GetOptionList(var OptionList: pPGPOptionList): PGPError;
var
  X509Option	: pPGPOptionList;
begin
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPOKeySetRef(FContext, FKeySetMain),
      PGPOPassThroughKeys(FContext, PGPBoolean(FPassThroughKeys = true)),
      // PGPOPassThroughClearSigned(FContext, PGPBoolean(FPassThroughClearSigned = true)), // buggy
      PGPOPassThroughIfUnrecognized(FContext, PGPBoolean(FPassThroughIfUnrecognized = true)),
      PGPORecursivelyDecode(FContext, PGPBoolean(FRecursivelyDecode = true)),
      PGPOSendEventIfKeyFound(FContext, PGPTrue),
      PGPOEventHandler(FContext, EventHandler, Self)
    ]);
  if Result <> 0 then Exit;
  case FX509Mode of
    X509Secret: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_PKCS12);
    X509Binary: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_X509DataInPKCS7);
    X509Armored: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_PEMEncodedX509Cert);
    X509IsMime..X509Encrypted: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_SMIMEBody);
  else
    // doesn't work properly with S/MIME
    // Result := PGPAppendOptionList(OptionList, [PGPOSendNullEvents(FContext, FProgressInterval)]);
    Exit;
  end;
  if FX509Mode in [X509IsMime..X509Encrypted] then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOSMIMEMatchCriterion(FContext, kPGPSMIMEMatchCriterion_Any),
	X509Option
      ]);
  end;
end;

function TPGPDecode.SetOutputOption(var OptionList: pPGPOptionList): PGPError;
var
  FileSpec	: pPGPFileSpec;
begin
  if FDetachedSigVerify then begin
    FFileToFile := false;
    if GetOutputFileName <> E then begin
      FileSpec := nil;
      Result := PGPNewFileSpecFromFullPath(FContext, PChar(FOutputFileName), FileSpec);
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList,
	  [
	    PGPODetachedSig(FContext, PGPOInputFile(FContext, FileSpec), PGPOLastOption(FContext))
	  ]);
      finally
	PGPFreeFileSpec(FileSpec);
      end;
    end
    else Result := kPGPError_DetachedSignatureFound;
  end
  else if FX509Mode = X509IsMime then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPODetachedSig(FContext,
			PGPOInputBuffer(FContext, PChar(FSignedDataBuffer), Length(FSignedDataBuffer)),
			PGPOLastOption(FContext))
      ]);
  end
  else begin
    if FFileOutput then begin
      if GetOutputFileName = E then begin
	Result := kPGPError_CantOpenFile;
	Exit;
      end;
    end;
    if FFileToFile then begin
      FileSpec := nil;
      Result := PGPNewFileSpecFromFullPath(FContext, PChar(FOutputFileName), FileSpec);
      try
	if Result = 0 then begin
	  Result := PGPAppendOptionList(OptionList,
	    [
	      PGPOOutputFile(FContext, FileSpec),
	      PGPOAppendOutput(FContext, PGPTrue)
	    ]);
	end;
      finally
	PGPFreeFileSpec(FileSpec);
      end;
    end
    else begin
      if FClearSigned and FPassThroughClearSigned then
	Result := PGPAppendOptionList(OptionList, [PGPODiscardOutput(FContext, PGPTrue)]) // for bugfix
      else begin
	Result := PGPAppendOptionList(OptionList,
	  [
	    PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
	  ]);
      end;
    end;
  end;
end;

function TPGPDecode.PrepareOutput(const Input: PChar; InputSize: DWord; const MimeHeaders, MimeBody: String): PGPError;
var
  HeaderSize	: DWord;
begin
  Result := 0;
  if not FFileToFile then begin
    if not (FDetachedSigVerify or FClearSigned)
    or (FIsPGPMime and not FPassThroughClearSigned) then begin
      Result := SecureMove;
      if Result <> 0 then Exit;
      HeaderSize := Length(MimeHeaders);
      if (FIsPGPMime or (FX509Mode in [X509Signed, X509Encrypted])) and (HeaderSize > 0) then begin
	FActualOutputSize := HeaderSize + FSecureBufferSize;
	FAllocatedOutputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, FActualOutputSize, kPGPMemoryMgrFlags_None);
	if FAllocatedOutputBuffer <> nil then begin
	  Move(MimeHeaders[1], FAllocatedOutputBuffer[0], HeaderSize);
	  Move(FSecureBuffer[0], FAllocatedOutputBuffer[HeaderSize], FSecureBufferSize);
	  PGPFreeData(FSecureBuffer);
	  FSecureBuffer := nil;
	  FSecureBufferSize := 0;
	end
	else Result := kPGPError_OutOfMemory;
      end
      else begin
	FAllocatedOutputBuffer := FSecureBuffer;
	FActualOutputSize := FSecureBufferSize;
      end;
    end;
    // fixes for signed data
    if (FClearSigned or (FX509Mode = X509IsMime))
    and FPassThroughClearSigned then begin
      FActualOutputSize := InputSize;
      Result := DataCopy(FAllocatedOutputBuffer, Input, FActualOutputSize);
    end
    else if (FX509Mode in [X509IsMime..X509Encrypted]) and (FActualOutputSize = 0) then begin
      FActualOutputSize := Length(MimeHeaders) + Length(MimeBody);
      Result := DataCopy(FAllocatedOutputBuffer, PChar(MimeHeaders + MimeBody), FActualOutputSize);
    end;
  end;
end;

function TPGPDecode.DoRecursion(const Input: PChar; InputSize: DWord): PGPError;
begin
  FFileNameBuffer := FInputFileName;
  if FFileToFile then begin
    SetString(FInputFileName, nil, MAX_PATH);
    if GetTempFileName(PChar(ExtractFilePath(FOutputFileName)), 'PGP', 0, PChar(FInputFileName)) <> 0 then begin
      Delete(FInputFileName, succ(StrLen(Pointer(FInputFileName))), MAXINT);
      if MoveFileEx(PChar(FOutputFileName), PChar(FInputFileName), MOVEFILE_REPLACE_EXISTING) then begin
	FInputBuffer := PChar(FInputFileName);
	FInputSize := Length(FInputFileName);
	Result := Decode(true);
	if Result <> 0 then DeleteFile(FInputFileName);
      end
      else begin
	Result := kPGPError_FileOpFailed;
	DeleteFile(FOutputFileName);
	DeleteFile(FInputFileName);
	Exit;
      end;
    end
    else begin
      Result := kPGPError_FileOpFailed;
      DeleteFile(FOutputFileName);
      Exit;
    end;
  end
  else begin
    FInputBuffer := Input;
    FInputSize := InputSize;
    Result := Decode(false);
  end;
end;

function TPGPDecode.SecureView(const Output: PChar; OutputSize: DWord): PGPError;
var
  PGPsc		: pPGPsc;
  PGPtls	: pPGPtls;
  Utf8OutputSize: DWord;
  Utf8Output	: PChar;
begin
  PGPsc := nil;
  PGPtls := nil;
  if InitPGPsc(FParentHandle, PGPsc, PGPtls, 0) then begin
    try
      if SDKVersion >= PGP9XX then begin
	Utf8OutputSize := OutputSize * UTF8Factor;
	Utf8Output := PGPNewSecureData(PGPGetDefaultMemoryMgr, Utf8OutputSize, kPGPMemoryMgrFlags_None);
	try
	  Utf8OutputSize := SecureAnsiCharToUtf8(Output, Utf8Output, Utf8OutputSize);
	  Result := TempestViewer(PGPsc, FParentHandle, Utf8Output, Utf8OutputSize, true);
	finally
	  FillChar(Utf8Output^, Utf8OutputSize, 0);
	  PGPFreeData(Utf8Output);
	end;
      end
      else Result := TempestViewer(PGPsc, FParentHandle, Output, OutputSize, true);
      ProcessMessages;
    finally
      UninitPGPsc(FParentHandle, PGPsc, PGPtls);
    end;
  end
  else Result := kPGPError_FeatureNotAvailable;
end;

function TPGPDecode.WriteOutputFile(Output: PChar; OutputSize: DWord): PGPError;
var
  OutputFile	: THandle;
  BytesWritten	: DWord;
begin
  OutputFile := CreateFile(PChar(FOutputFileName), GENERIC_WRITE, 0, nil, CREATE_ALWAYS,
			   FILE_ATTRIBUTE_ARCHIVE or FILE_FLAG_WRITE_THROUGH, 0);
  if OutputFile <> INVALID_HANDLE_VALUE then begin
    try
      if not WriteFile(OutputFile, Output^, OutputSize, BytesWritten, nil) or (BytesWritten <> OutputSize) then
	Result := kPGPError_WriteFailed
      else Result := kPGPError_NoErr;
    finally
      CloseHandle(OutputFile);
    end;
  end
  else Result := kPGPError_CantOpenFile;
end;

function TPGPDecode.Decode(IsFile: Longbool): Longint;
var
  BugFix	: Longbool;
  Buffered	: Longbool;
  Recursively	: Longbool;
  OptionList	: pPGPOptionList;
  InFileSpec	: pPGPFileSpec;
  Output	: PChar;
  OutputSize	: DWord;
  Input		: PChar;
  InputSize	: DWord;
  MimeHeaders	: String;
  MimeBody	: String;
  MimeSig	: String;
  OutputBuffer	: PChar;
  BufferSize	: DWord;
begin
  Result := 0;
  BugFix := false;
  Buffered := false;
  Output := nil;
  OutputSize := 0;
  if IsFile then
    Result := HandleFileInput
  else if FInputSize = 0 then Result := kPGPError_ItemNotFound;
  if Result <> 0 then Exit;
  Input := FInputBuffer;
  InputSize := FInputSize;
  try
    InFileSpec := nil;
    OptionList := nil;
    Result := InitDecode;
    if Result <> 0 then Exit;
    if IsFile then begin
      if (SDKVersion >= PGP7XX) and not FFileToFile then begin
	Result := ReadInputFile;
	if Result <> 0 then Exit;
	Input := FInputBuffer;
	InputSize := FInputSize;
	Buffered := true;
	IsFile := false;
      end
      else begin
	Result := PGPNewFileSpecFromFullPath(FContext, PChar(FInputFileName), InFileSpec);
	if Result <> 0 then Exit;
      end;
    end;
    try
      Result := Analyze(InFileSpec);
      if FPassOption in [AnalyseOnly, AnalysePass] then begin
	if Result <> kPGPError_AsciiParseIncomplete then Result := 0; // assume FAnalysis = UNKNOWN on other errors
	if (Result = 0) and not FRecognized then begin
	  if (FX509Mode = X509IsMime) and not IsFile then begin
	    case GetPSMimeType(FInputBuffer, MimeHeaders, MimeBody, MimeSig, FCharset, true) of
	      psClearSigned, psSigned: FAnalysis := SMIME_SIGNED;
	      psEncrypted: FAnalysis := SMIME_ENCRYPTED;
	    else
	      FX509Mode := X509Test;
	      case FX509Mode of
		X509Secret: FAnalysis := X509_SECRETKEY;
		X509Binary..X509Armored: FAnalysis := X509_CERTIFICATE;
	      end;
	    end;
	  end;
	end;
	FRecognized := (FAnalysis <> UNKNOWN);
      end;
      if FPassOption <> AnalyseOnly then begin
	if (FAnalysis = UNKNOWN) and not IsFile and (FInputSize > 0) and (FInputBuffer[0] > #127) then begin
	  Result := FixNonAsciiBug(Input);
	  if Result <> 0 then Exit;
	  BugFix := true;
	end;
	// X509 and S/MIME handling
	if not IsFile then begin
	  if SDKVersion >= PGP81X then begin
	    case FX509Mode of
	      X509Off: FX509Mode := X509Processed;
	      X509Secret..X509Armored: Exit;
	      X509IsMime..X509Encrypted: begin
		case GetPSMimeType(FInputBuffer, MimeHeaders, MimeBody, MimeSig, FCharset, false) of
		  psClearSigned: begin
		    FAnalysis := SMIME_SIGNED;
		    FX509Mode := X509IsMime;
		    FSignedDataBuffer := MimeBody;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FSMimeCertBag := MimeSig;
		    FInputSize := Length(MimeSig);
		    DataCopy(FInputBuffer, PChar(MimeSig), FInputSize);
		  end;
		  psEncrypted: begin
		    FAnalysis := SMIME_ENCRYPTED;
		    FX509Mode := X509Encrypted;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FInputSize := Length(MimeBody);
		    DataCopy(FInputBuffer, PChar(MimeBody), FInputSize);
		  end;
		  psSigned: begin
		    FAnalysis := SMIME_SIGNED;
		    FX509Mode := X509Signed;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FSMimeCertBag := MimeBody;
		    FInputSize := Length(MimeBody);
		    DataCopy(FInputBuffer, PChar(MimeBody), FInputSize);
		  end;
		  psPGP: begin
		    FIsPGPMime := true;
		    FX509Mode := X509Processed;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FInputSize := Length(MimeBody);
		    DataCopy(FInputBuffer, PChar(MimeBody), FInputSize);
		  end;
		else
		  FX509Mode := X509Processed;
		end;
	      end;
	    end;
	  end
	  else begin
	    FX509Mode := X509Processed;
	    if (GetPSMimeType(FInputBuffer, MimeHeaders, MimeBody, MimeSig, FCharset, false) = psPGP) then begin
	      FIsPGPMime := true;
	      Input := FInputBuffer;
	      InputSize := FInputSize;
	      FInputSize := Length(MimeBody);
	      DataCopy(FInputBuffer, PChar(MimeBody), FInputSize);
	    end;
	  end;
	end;
	// FPassThroughClearSigned may truncate signed data,
	// RecursivelyDecode enables FPassThroughClearSigned,
	// if we already have clear signed data just disable it
	Recursively := FRecursivelyDecode;
	if FClearSigned then FRecursivelyDecode := false;
	Result := GetOptionList(OptionList);
	if Result <> 0 then Exit;
	if IsFile then
	  Result := PGPAppendOptionList(OptionList, [PGPOInputFile(FContext, InFileSpec)])
	else Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(FContext, FInputBuffer, FInputSize)]);
	try
	  if Result <> 0 then Exit;
	  if FDetachedSigVerify or FClearSigned or (FX509Mode = X509IsMime) then begin
	    Result := SetOutputOption(OptionList);
	    if Result <> 0 then Exit;
	  end
	  else FOutputExists := false;
	  if not FRecurring or FClearSigned or (FX509Mode in [X509IsMime..X509Signed]) then begin
	    FPassOption := DecodePass;
	    Result := pgpEvents.PGPDecode(FContext, OptionList, PGPOLastOption(FContext));
	    if Result <> 0 then Exit;
	    if FPassOption <> AnalyseOnly then begin
	      Result := PrepareOutput(Input, InputSize, MimeHeaders, MimeBody);
	      if Result <> 0 then Exit;
	      Output := FAllocatedOutputBuffer;
	      OutputSize := FActualOutputSize;
	      if BugFix and FPassThroughIfUnrecognized then begin // remove leading AT
		dec(OutputSize);
		inc(Output);
	      end;
	    end;
	  end;
	finally
	  try
	    if not FMultiDecode and (FSecure <> nil) and not FIsConventional then begin
	      FSecure.DoCachePassphrase(FDecryptionKey, FPassphrase, FAllocatedOutputBuffer, FActualOutputSize, false);
	    end;
	  finally
	    if not FMultiDecode then begin
	      PGPFreeData(FPassphrase);
	      FPassphrase := nil;
	    end;
	  end;
	  // reset after bugfixing
	  FRecursivelyDecode := Recursively;
	end;
      end;
    finally
      PGPFreeFileSpec(InFileSpec);
      PGPFreeOptionList(OptionList);
      if BugFix then begin
	PGPFreeData(FInputBuffer);
	FInputBuffer := Input;
	FInputSize := InputSize;
      end;
    end;
    // workaround for verifying encrypted PGP/MIME & S/MIME signatures like PGP does internally with encrypted clear signed data
    if FRecursivelyDecode and not FRecurring and (FPassOption <> AnalyseOnly)
    and (FAnalysis in [ENCRYPTED, SMIME_ENCRYPTED]) and (FSigPropsRec.sStatus = SIGNED_NOT) then begin
      if IsFile then
	FX509Mode := X509Off
      else if FIsPGPMime then FX509Mode := X509IsMime;
      OutputBuffer := FAllocatedOutputBuffer;
      BufferSize := FActualOutputSize;
      FPassOption := AnalysePass;
      FRecurring := true;
      try
	Result := DoRecursion(Output, OutputSize);
      finally
	if FEyesOnly and (OutputBuffer <> nil) then FillChar(OutputBuffer^, BufferSize, 0);
	PGPFreeData(OutputBuffer);
	// reset to start value
	if FAnalysis in [SMIME_UNRECOGNIZED, SMIME_SIGNED] then
	  FAnalysis := SMIME_ENCRYPTED
	else FAnalysis := ENCRYPTED;
	FRecurring := false;
      end;
    end
    else if FFileToFile then begin
      if FRecurring then begin
	if FAnalysis = SIGNED then
	  DeleteFile(FInputFileName)
	else if not RenameFile(FInputFileName, FOutputFileName) then begin
	  Result := kPGPError_FileOpFailed;
	  DeleteFile(FInputFileName);
	end;
      end;
    end
    else begin
      if FRecurring then begin
	if FSigPropsRec.sStatus = SIGNED_NOT then begin
	  Output := Input;
	  OutputSize := InputSize;
	end
	else if FAnalysis = SMIME_UNRECOGNIZED then begin
	  Output := PChar(MimeHeaders + MimeBody);
	  OutputSize := Length(MimeHeaders) + Length(MimeBody);
	end;
      end;
      if FEyesOnly then
	SecureView(Output, OutputSize)
      else if not FFileOutput then begin
	try
	  SetString(FOutputBuffer, nil, OutputSize);
	  Move(Output^, Pointer(FOutputBuffer)^, OutputSize);
	except
	  on EOutOfMemory do Result := kPGPError_OutOfMemory;
	end;
      end
      else WriteOutputFile(Output, OutputSize);
    end;
  finally
    if Buffered then PGPFreeData(Input);
    try
      if FEyesOnly and (FAllocatedOutputBuffer <> nil) then FillChar(FAllocatedOutputBuffer^, FActualOutputSize, 0);
    finally
      try
	PGPFreeData(FAllocatedOutputBuffer);
	FAllocatedOutputBuffer := nil;
      finally
	FinitDecode;
      end;
    end;
  end;
end;

function TPGPDecode.AnalyseBuffer(const DataBuffer: String): Longint;
begin
  FPassOption := AnalyseOnly;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := nil;
  try
    Result := Decode(false);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.AnalyseFile(const FileName: String): Longint;
begin
  FPassOption := AnalyseOnly;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := nil;
  try
    Result := Decode(true);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.AnalyseMimeBuffer(const DataBuffer: String): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalyseOnly;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := nil;
  try
    Result := Decode(false);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.AnalyseMimeFile(const FileName: String): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalyseOnly;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := nil;
  try
    Result := Decode(true);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.DecodeBuffer(const DataBuffer: String): Longint;
begin
  FPassOption := AnalysePass;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := nil;
  try
    Result := Decode(false);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.DecodeFile(const FileName: String): Longint;
begin
  FPassOption := AnalysePass;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := nil;
  try
    Result := Decode(true);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.DecodeMimeBuffer(const DataBuffer: String): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalysePass;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := nil;
  try
    Result := Decode(false);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.DecodeMimeFile(const FileName: String): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalysePass;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := nil;
  try
    Result := Decode(true);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecode.AnalyzeHandler(Event: pPGPEvent): PGPError;
begin
  Result := 0;
  with Event^.EData.AnalyzeData do begin
    case FPassOption of
      AnalyseOnly: begin
	FRecognized := true;
	if SectionType = kPGPAnalyze_Encrypted then FAnalysis := ENCRYPTED;
	if SectionType <> kPGPAnalyze_Key then Result := kPGPError_SkipSection; // avoid key data bug
      end;
      AnalysePass: begin
	FRecognized := (SectionType <> kPGPAnalyze_Unknown);
	case SectionType of
	  kPGPAnalyze_Signed: FClearSigned := true;
	  kPGPAnalyze_DetachedSignature: FDetachedSigVerify := true;
	end;
	if FRecognized then FAnalysis := TAnalysis(SectionType);
	if FDetachedSigVerify or FClearSigned or (FX509Mode >= X509IsMime) then
	  Result := kPGPError_UserAbort
	else if SectionType <> kPGPAnalyze_Key then Result := kPGPError_SkipSection; // avoid key data bug
      end;
      DecodePass: if (SectionType <> kPGPAnalyze_Unknown) and (FX509Mode < X509IsMime) then begin
	FAnalysis := TAnalysis(SectionType);
      end;
    end;
  end;
end;

function TPGPDecode.RecipientsHandler(Event: pPGPEvent): PGPError;
var
  KeyIDArraySize: PGPSize;
begin
  Result := 0;
  if FAnalysis = SMIME_UNRECOGNIZED then FAnalysis := SMIME_ENCRYPTED;
  with Event^.EData.RecipientsData do begin
    PGPIncKeySetRefCount(RecipientSet);
    FRecipientsData.RecipientSet := RecipientSet;
    FRecipientsData.KeyCount := KeyCount;
    if SDKVersion >= PGP7XX then
      KeyIDArraySize := SizeOf(TPGPKeyID7) * KeyCount
    else KeyIDArraySize := SizeOf(TPGPKeyID6) * KeyCount;
    FRecipientsData.KeyIDArray := PGPNewData(PGPGetDefaultMemoryMgr, KeyIDArraySize, kPGPMemoryMgrFlags_None);
    if FRecipientsData.KeyIDArray <> nil then Move(KeyIDArray^, FRecipientsData.KeyIDArray^, KeyIDArraySize);
  end;
end;

function TPGPDecode.PassphraseHandler(Event: pPGPEvent): PGPError;
begin
  Result := 0;
  with Event^.EData.PassphraseData do begin
    FIsConventional := Boolean(IsConventional);
    if FSectionNumber = FPassSection then begin
      PGPFreeData(FPassphrase);
      FPassphrase := nil;
    end;
    if FIsConventional then begin
      if FPassphrase = nil then Result := ConvDecPassphraseDialog(FContext, FPassphrase, FPassDlgPrompt, FParentHandle);
    end
    else begin
      if (not FMultiDecode and ((FSecure = nil) or not FSecure.GetCachedPassphrase(FDecryptionKey, FPassphrase, false)))
      or (FSectionNumber = FPassSection) or (FPassphrase = nil) then with FRecipientsData do begin
	Result := DecryptionPassphraseDialog(FContext, RecipientSet, KeyCount, KeyIDArray,
					     FDecryptionKey, FPassphrase, true, FPassDlgPrompt, FParentHandle);
      end;
    end;
    FPassSection := FSectionNumber;
  end;
  if Result <> 0 then Exit;
  Result := PGPAddJobOptions(Event^.Job, PGPOPassphrase(FContext, FPassphrase), PGPOLastOption(FContext));
  if not FMultiDecode then SetCursor(FWait);
end;

function TPGPDecode.KeyFoundHandler(Event: pPGPEvent): PGPError;
var
  IsSMime	: Longbool;
  CertsToImport	: pPGPKeySet;
begin
  Result := 0; // don't cancel decoding whatever happens here
  with Event^.EData.KeyFoundData do begin
    ClearSignCertProps;
    IsSMime := (FX509Mode in [X509IsMime..X509Signed]);
    if IsSMime then FValidCertChain := VerifyX509CertificateChainDB(KeysFound, FSignCertProps);
    if FQueryAddKeys then begin
      if SDKVersion >= PGP7XX then begin
	if not IsSMime then
	  CertsToImport := PGPPeekKeyDBRootKeySet(KeysFound)
	else begin
	  FKeyAddResult := GetCertKeySetFromKeySet(PGPPeekKeyDBRootKeySet(KeysFound), false, CertsToImport);
	  if FKeyAddResult <> 0 then Exit;
	end;
	try
	  FKeyAddResult := SelectAddKeysToKeyRing(FContext, CertsToImport, FKeySetMain,
						  FKeyPropsList, PChar(FKeyDlgPrompt),
						  false, spgpKeyPropFlag_IDFlags,
						  FParentHandle);
	finally
	  if IsSMime then PGPFreeKeySet(CertsToImport);
	end;
      end
      else begin
	FKeyAddResult := SelectAddKeysToKeyRing(FContext, KeysFound, FKeySetMain,
						FKeyPropsList, PChar(FKeyDlgPrompt),
						false, spgpKeyPropFlag_IDFlags,
						FParentHandle);
      end;
      if FKeyAddResult = kPGPError_UserAbort then FQueryAddKeys := false;
    end;
  end;
end;

// causes trouble when used with FPassThroughClearSigned on clear signed input
// causes trouble with FPassThroughUnrecognized and local output options
function TPGPDecode.OutputHandler(Event: pPGPEvent): PGPError;
var
  FileSpec	: pPGPFileSpec;
begin
  Result := 0;
  with Event^.EData.OutputData do begin
    if (ForYourEyesOnly = PGPTrue) and not FFileToFile then FEyesOnly := true;
    if FOutputExists then begin
      if not FFileToFile then Result := SecureMove;
    end
    else if FFileOutput and not FEyesOnly then begin
      if (FOutputFileName = E) and (SuggestedName <> E) then begin
	FOutputFileName := ExtractFilePath(FInputFileName) + SuggestedName;
      end;
      if GetOutputFileName = E then begin
	Result := kPGPError_CantOpenFile;
	Exit;
      end;
      if FFileToFile then begin
	FileSpec := nil;
	Result := PGPNewFileSpecFromFullPath(FContext, PChar(FOutputFileName), FileSpec);
	try
	  if Result = 0 then begin
	    Result := PGPBuildOptionList(FContext, FOutputOptions,
	      [
		PGPOOutputFile(FContext, FileSpec),
		PGPOAppendOutput(FContext, PGPTrue)
	      ]);
	  end;
	finally
	  PGPFreeFileSpec(FileSpec);
	end;
      end
      else begin
	Result := PGPBuildOptionList(FContext, FOutputOptions,
	  [
	    PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
	  ]);
      end;
    end
    else begin
      FFileToFile := false;
      Result := PGPBuildOptionList(FContext, FOutputOptions,
	[
	  PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
	]);
    end;
  end;
  if Result = 0 then begin
    Result := PGPAddJobOptions(Event^.Job, FOutputOptions, PGPOLastOption(FContext));
    FOutputExists := true;
  end;
end;

function TPGPDecode.SignatureHandler(Event: pPGPEvent): PGPError;
var
  SignKeyType	: PGPInt32;
  SignKeyOnRing	: pPGPKeyDBObj;
  PGPSignKeyID	: TPGPKeyID7;
  SignKeyID	: TKeyID;
  KeysFound	: Pointer;
begin
  Result := 0; // don't cancel decoding whatever happens here
  if FCheckSig then begin
    ClearSigPropsRec;
    if FAnalysis = SMIME_UNRECOGNIZED then FAnalysis := SMIME_SIGNED;
    if SDKVersion >= PGP7XX then with Event^.EData.SignatureData.PGPEventSignatureData7, FSigPropsRec do begin
      if (SDKVersion >= PGP7XX)
      and (PGPGetKeyDBObjNumericProperty(SigningKey, kPGPKeyDBObjProperty_ObjectType, SignKeyType) = 0) then begin
	if SignKeyType = kPGPKeyDBObjType_SubKey then begin
	  sSubKeyHexID := GetKeyPropKeyID(SigningKey);
	  SigningKey := PGPPeekKeyDBObjKey(SigningKey);
	end;
      end;
      if Boolean(Checked) then begin
	sHexID := GetKeyPropKeyID(SigningKey);
	sUserID := GetKeyPropUserID(SigningKey);
	sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
	sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
	sChecked := Boolean(Checked);
	sVerified := Boolean(Verified);
	// workaround for S/MIME signatures
	if PGPKeySetIsMember(SigningKey, FKeySetMain) = PGPFalse then begin
	  GetKeyFromNewSet(SigningKey, FKeySetMain, SignKeyOnRing);
	  sValidity := GetKeyPropValidity(SignKeyOnRing);
	end
	else sValidity := TValidityLevel(KeyValidity);
	if FAnalysis = SMIME_SIGNED then begin
	  sRevoked := Boolean(KeyRevoked) or FSignCertProps.x509IsRevoked;
	  sDisabled := Boolean(KeyDisabled) or FSignCertProps.x509IsCorrupt;
	  sExpired := Boolean(KeyExpired) or FSignCertProps.x509IsExpired;
	end
	else begin
	  sRevoked := Boolean(KeyRevoked);
	  sDisabled := Boolean(KeyDisabled);
	  sExpired := Boolean(KeyExpired);
	end;
	sAxiomatic := GetKeyPropIsAxiomatic(SigningKey);
	if (SDKVersion >= PGP8XX) and (ExpirationPeriod <> 0) then begin
	  sExpTimeNum := sCreaTimeNum + ExpirationPeriod;
	  sExpTimeStr := UnixTimeToLocalTimeStr(sExpTimeNum);
	end;
	if SDKVersion >= PGP9XX then sHashAlgorithm := GetTHashAlgorithm(HashAlgorithm);
	if GetKeyPropIsSigningKey(SigningKey) then begin
	  sStatus := TSigStatus(Checked + Verified);
	  if (SDKVersion >= PGP8XX) and (sStatus = SIGNED_GOOD) and (ExpirationPeriod <> 0)
	  and (sExpTimeNum < SystemTimeToUnixTimeNum) then begin
	    sStatus := SIGNED_SIG_EXPIRED;
	  end;
	end
	else sStatus := SIGNED_UNKNOWN_ALG;
	sValidCertChain := FValidCertChain;
      end
      else begin
	Move(SigningKeyID, PGPSignKeyID, SizeOf(TPGPKeyID7));
	Result := PGPGetKeyIDString(PGPSignKeyID, kPGPKeyIDString_Full, SignKeyID);
	if Result = 0 then begin
	  sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
	  sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
	  sHexID := SignKeyID;
	end;
	sStatus := SIGNED_NO_KEY;
	if Result = 0 then begin
	  if FKeyData <> E then begin
	    KeysFound := Ptr(-1);
	    try
	      Result:= KeyImportDialog(E, E, FKeyData, E, FKeyPropsList, false,
				       spgpKeyPropFlag_IDFlags, FParentHandle, KeysFound);
	      if (Result > 0) and (FKeyPropsList.IndexOf(SignKeyID) > -1) then begin
		PGPAddJobOptions(Event^.Job,
				 PGPOKeySetRef(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound))),
				 PGPOLastOption(FContext));
	      end;
	    finally
	      PGPFreeKeyDB(pPGPKeyDB(KeysFound));
	    end;
	  end;
	  if FGetKeyFromServer then begin
	    KeysFound := nil;
	    Result := KeyServerDialog(FContext, FKeySetMain, SignKeyID, KeyFilterFlag_CanVerify,
				      pPGPKeyDB(KeysFound), FParentHandle);
	    if (Result = 0) and (KeysFound <> nil) then begin
	      try
		Result := SelectAddKeysToKeyRing(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound)), FKeySetMain,
						 FKeyPropsList, PChar(FKeyDlgPrompt),
						 false, spgpKeyPropFlag_IDFlags,
						 FParentHandle);
	      if Result = 0 then begin
		  PGPAddJobOptions(Event^.Job,
				   PGPOKeySetRef(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound))),
				   PGPOLastOption(FContext));
		end;
	      finally
		PGPFreeKeyDB(pPGPKeyDB(KeysFound));
	      end;
	    end;
	  end;
	  Result := 0;
	end;
      end;
      sDetached := FDetachedSigVerify;
      // GetPSMimeType() does this on buffered data for all versions
      // if PGP95 then FCharset := Charset;
    end
    else with Event^.EData.SignatureData.PGPEventSignatureData6, FSigPropsRec do begin
      if Boolean(Checked) then begin
	sHexID := GetKeyPropKeyID(SigningKey);
	sUserID := GetKeyPropUserID(SigningKey);
	sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
	sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
	sChecked := Boolean(Checked);
	sVerified := Boolean(Verified);
	sValidity := TValidityLevel(KeyValidity);
	sRevoked := Boolean(KeyRevoked);
	sDisabled := Boolean(KeyDisabled);
	sExpired := Boolean(KeyExpired);
	sAxiomatic := GetKeyPropIsAxiomatic(SigningKey);
	if GetKeyPropIsSigningKey(SigningKey) then
	  sStatus := TSigStatus(Checked + Verified)
	else sStatus := SIGNED_UNKNOWN_ALG;
      end
      else begin
	Move(SigningKeyID, PGPSignKeyID, SizeOf(TPGPKeyID6));
	Result := PGPGetKeyIDString(PGPSignKeyID, kPGPKeyIDString_Full, SignKeyID);
	if Result = 0 then begin
	  sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
	  sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
	  sHexID := SignKeyID;
	end;
	sStatus := SIGNED_NO_KEY;
	if Result = 0 then begin
	  if FKeyData <> E then begin
	    KeysFound := Ptr(-1);
	    try
	      Result:= KeyImportDialog(E, E, FKeyData, E, FKeyPropsList, false,
				       spgpKeyPropFlag_IDFlags, FParentHandle, KeysFound);
	      if (Result > 0) and (FKeyPropsList.IndexOf(SignKeyID) > -1) then begin
		PGPAddJobOptions(Event^.Job,
				 PGPOKeySetRef(FContext, pPGPKeySet(KeysFound)),
				 PGPOLastOption(FContext));
	      end;
	    finally
	      PGPFreeKeySet(pPGPKeySet(KeysFound));
	    end;
	  end;
	  if FGetKeyFromServer then begin
	    KeysFound := nil;
	    Result := KeyServerDialog(FContext, FKeySetMain, SignKeyID, KeyFilterFlag_CanVerify,
				      pPGPKeySet(KeysFound), FParentHandle);
	    if (Result = 0) and (KeysFound <> nil) then begin
	      try
		SelectAddKeysToKeyRing(FContext, pPGPKeySet(KeysFound), FKeySetMain,
				       FKeyPropsList, PChar(FKeyDlgPrompt),
				       false, spgpKeyPropFlag_IDFlags,
				       FParentHandle);
		if Result = 0 then begin
		  PGPAddJobOptions(Event^.Job,
				   PGPOKeySetRef(FContext, pPGPKeySet(KeysFound)),
				   PGPOLastOption(FContext));
		end;
	      finally
		PGPFreeKeySet(pPGPKeySet(KeysFound));
	      end;
	    end;
	  end;
	end;
	Result := 0;
      end;
      sDetached := FDetachedSigVerify;
    end;
    if FEyesOnly or FShowSigLog then ShowSigInfo;
  end;
end;

function TPGPDecode.EndLexHandler(Event: pPGPEvent): PGPError;
begin
  Result := 0;
  with FRecipientsData do begin
    PGPFreeData(KeyIDArray);
    PGPFreeKeySet(RecipientSet);
  end;
  FillChar(FRecipientsData, SizeOf(TPGPEventRecipientsData), 0);
end;

function TPGPDecode.DecryptionHandler(Event: pPGPEvent): PGPError;
begin
  Result := 0;
  // GetPSMimeType() does this on buffered data for all versions
  // if PGP95 then FCharset := Event^.EData.DecryptionData.Charset;
end;


end.

