{$J+,Z4}
unit PGPDialogs;

{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  Windows,
  Classes,
  SysUtils,
  StrTools,
  KeyPropTypes,
  X509Types,
  UTF8,
  pgpBase,
  pgpErrors,
  pgpPubTypes,
  pgpUtilities,
  pgpOptionList,
  pgpMemoryMgr,
  pgpKeyServer,
  pgpGroups,
  pgpKeys,
  pgpTLS,
  pgpUI,
  pgpCL,
  KeyFuncs,
  PrefFuncs;

type
  TPassphraseCallback = procedure(const Passphrase: PChar; const UserValue);

{ Tools }
procedure RedrawWindow(Window: THandle);
function IsDialog(Window: THandle): Longbool;
procedure CenterDialog(Dialog, Owner: THandle);
function GetCenterDialogRect(Dialog, Owner: THandle): TRect;

{ "public" functions }
function PreferencesDialog(PrefsPage, WinHandle: Integer): Integer;
function SelectKeysDialog(const Prompt: String; var KeyPropsList: TKeyPropsList;
			  SingleKey: Longbool; PropertyFlags, FilterFlags, WinHandle: Integer): Integer;
function SendKeyToServerDialog(const Prompt: String; var KeyPropsList: TKeyPropsList;
			       KeyPropFlags, WinHandle: Integer): Integer;
function GetKeyFromServerDialog(const Prompt, KeyData: String; var KeyPropsList: TKeyPropsList;
				KeyPropFlags, WinHandle: Integer): Integer;
function KeyImportDialog(const KeyPrompt, CertPassPrompt, KeyData: String; CertPassphrase: PChar;
			 var KeyPropsList: TKeyPropsList; FromFile: Longbool;
			 KeyPropFlags, WinHandle: Integer; var KeySetFound: Pointer): Integer;
function KeyExportDialog(Prompt: String; var KeyData: String; var KeyPropsList: TKeyPropsList;
			 ToFile, ExportCompatible, ExportPrivate: Longbool; KeyPropFlags, WinHandle: Integer): Integer;
function KeyRevokeDialog(const KeyHexID, PassPrompt: String; WinHandle: Integer): Integer;
function KeyPassChangeDialog(const KeyHexID, OldPrompt, NewPrompt: String;
			     MinPassLength, MinPassQuality, WinHandle: Integer): Integer;
function SimplePassphraseDialog(const Title, Prompt: String; WinHandle: Integer;
				PassphraseCallback: TPassphraseCallback; const CallbackParam): Integer;

{ "private" functions }
function SelectX509CertifiedKeysDialog(Context: pPGPContext; KeySetX509CertifiedKeys: pPGPKeySet;
				       var KeySetSelected: pPGPKeySet; var CertSelected: pPGPKeyDBObj;
				       const Prompt: String; CertsOnly, SingleKey: Longbool;
				       WinHandle: Integer): Integer;
function SelectAddKeysToKeyRing(Context: pPGPContext; KeysToAdd: pPGPKeySet; KeySetMain: pPGPKeySet;
				var KeyPropsList: TKeyPropsList; const Prompt: String; SingleKey: Longbool;
				KeyPropFlags, WinHandle: Integer): PGPError;
function RecipientsDialog(Context: pPGPContext; AllKeys: pPGPKeySet;
			  AlwaysDisplayDialog: PGPBoolean; SMime: Longbool;
			  const DefaultKey: String; var RecipientKeys: Pointer;
			  const Prompt: String; WinHandle: Integer): PGPError;
function KeyServerDialog(Context: pPGPContext; KeySetMain: pPGPKeySet; const KeyData: String;
			 FilterFlags: Integer; var KeysFound: Pointer; WinHandle: Integer): PGPError;
function CollectRandomDataDialog(Context: pPGPContext; NeededEntropyBits: PGPUInt32; WinHandle: Integer): PGPError;
function PassphraseDialog(Context: pPGPContext; const Title, Prompt: String;
			  var Passphrase: PChar; WinHandle: Integer): Integer;
function ConvEncPassphraseDialog(Context: pPGPContext; var Passphrase: PChar;
				 const Prompt: String; WinHandle: Integer): PGPError;
function ConvDecPassphraseDialog(Context: pPGPContext; var Passphrase: PChar;
				 const Prompt: String; WinHandle: Integer): PGPError;
function KeyPassphraseDialog(Context: pPGPContext; KeySet: pPGPKeySet; var Passphrase: PChar;
			     const Prompt: String; WinHandle: Integer): PGPError;
function SigningPassphraseDialog(Context: pPGPContext; KeySetMain: pPGPKeySet;
				 var SigningKey: pPGPKey; var SignKeyPass: PChar;
				 const DefaultKeyID: String; FindMatchingKey: Longbool;
				 const Prompt: String; WinHandle: Integer): PGPError;
function DecryptionPassphraseDialog(Context: pPGPContext; RecipientKeys: pPGPKeySet;
				    KeyIDCount: PGPUInt32; const KeyIDList: pPGPKeyID7;
				    var DecryptionKey: pPGPKey; var DecryptKeyPass: PChar;
				    FindMatchingKey: Longbool; const Prompt: String; WinHandle: Integer): PGPError;
function ConfirmationPassphraseDialog(Context: pPGPContext; var Passphrase: PChar;
				      MinPassLength, MinPassQuality: Integer;
				      ShowPassphraseQuality: Longbool;
				      const Prompt: String; WinHandle: Integer): PGPError;


implementation

uses
  Messages;

const
  SYS_DLG = '#32770';
  PGP_DLG = 'PGP';

procedure RedrawWindow(Window: THandle);
begin
  Windows.RedrawWindow(Window, nil, 0, RDW_UPDATENOW or RDW_ALLCHILDREN);
end;

function IsDialog(Window: THandle): Longbool;
var
  aDialog: Array[0..MAX_PATH] of Char;
  iLen: Integer;
begin
  Result := true;
  iLen := GetClassName(Window, @aDialog, succ(MAX_PATH));
  if iLen > 0 then begin
    if lstrcmpi(SYS_DLG, aDialog) = 0 then Exit;
    aDialog[Length(PGP_DLG)] := #0;
    if lstrcmpi(PGP_DLG, aDialog) = 0 then Exit;
  end;
  Result := false;
end;

procedure CenterDialog(Dialog, Owner: THandle);
begin
  with GetCenterDialogRect(Dialog, Owner) do MoveWindow(Dialog, Left, Top, Right, Bottom, false);
end;

function GetCenterDialogRect(Dialog, Owner: THandle): TRect;
var
  rOwner: TRect;
begin
  GetWindowRect(Owner, rOwner);
  with rOwner do begin
    dec(Right, pred(Left));
    dec(Bottom, pred(Top));
  end;
  GetWindowRect(Dialog, Result);
  with Result do begin
    dec(Right, pred(Left));
    dec(Bottom, pred(Top));
    Left := rOwner.Left + (rOwner.Right - Right) div 2;
    Top := rOwner.Top + (rOwner.Bottom - Bottom) div 2;
  end;
end;


const
  OwnerWnd: HWND = 0;
  CallWndHook: HHOOK = 0;

function CallWndProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  if nCode = HC_ACTION then with PCWPStruct(lParam)^ do begin
    Result := 0;
    if GetWindowLong(hWnd, GWL_STYLE) and WS_CHILD = 0 then begin
      case Message of
	WM_CREATE: if IsDialog(hWnd) then PCreateStruct(lParam)^.hwndParent := OwnerWnd;
	WM_DESTROY: if IsDialog(hWnd) then OwnerWnd := GetWindow(hWnd, GW_OWNER);
	WM_SHOWWINDOW: if wParam <> 0 then begin
	  if IsDialog(hWnd) then begin
	    SetWindowPos(hWnd, HWND_TOP or HWND_NOTOPMOST, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOMOVE or SWP_NOSIZE);
	    if OwnerWnd = 0 then OwnerWnd := GetDesktopWindow;
	    CenterDialog(hWnd, OwnerWnd);
	    OwnerWnd := hWnd;
	  end;
	end;
      end;
    end;
  end
  else Result := CallNextHookEx(CallWndHook, nCode, wParam, lParam);
end;

procedure InstallHook(WinHandle: THandle);
begin
  OwnerWnd := WinHandle;
  CallWndHook := SetWindowsHookEx(WH_CALLWNDPROC, @CallWndProc, 0, GetCurrentThreadID);
end;

procedure RemoveHook;
begin
  UnhookWindowsHookEx(CallWndHook);
  CallWndHook := 0;
  OwnerWnd := 0;
end;

function PreferencesDialog(PrefsPage, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := PGPclPreferences(Context, WinHandle, PrefsPage, KeySetMain);
    RedrawWindow(WinHandle);
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function SelectKeysDialog(const Prompt: String; var KeyPropsList: TKeyPropsList;
			  SingleKey: Longbool; PropertyFlags, FilterFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeyFilter	 : pPGPFilter;
  KeySetFiltered : pPGPKeySet;
  KeySetSelected : pPGPKeySet;
  KeyCount	 : PGPUInt32;
begin
  KeyFilter := nil;
  KeySetFiltered := nil;
  KeySetSelected := nil;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    if FilterFlags <> KeyFilterFlag_AllKeys then begin
      Result := GetKeyFilter(Context, FilterFlags, KeyFilter);
      if Result <> 0 then Exit;
      try
	Result := PGPFilterKeySet(KeySetMain, KeyFilter, KeySetFiltered);
      finally
	PGPFreeFilter(KeyFilter);
      end;
      if Result <> 0 then Exit;
    end
    else KeySetFiltered := KeySetMain;
    try
      Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetFiltered, KeySetMain,
				ord(SingleKey = true) * PGPCL_SINGLESELECTION, KeySetSelected);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
    finally
      if KeySetFiltered <> KeySetMain then PGPFreeKeySet(KeySetFiltered);
    end;
    try
      Result := PGPCountKeys(KeySetSelected, KeyCount);
      if Result <> 0 then Exit;
      if KeyCount <> 0 then begin
	Result := GetKeySetProps(Context, KeySetSelected, KeyPropsList, PropertyFlags,
				 KeyFilterFlag_AllKeys, UserID_Ordering);
	if Result > 0 then Result := 0;
      end;
    finally
      PGPFreeKeySet(KeySetSelected);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function InitKeyServer(Context: pPGPContext; var ServerThreadStorage: pPGPKeyServerThreadStorage;
		       var KeyServerSpecs: pPGPKeyServerSpec; var KeyServerCount: PGPUInt32): PGPError;
var
  KeyServerList	 : TStringList;
  KeyServerIndex : Integer;
begin
  ServerThreadStorage := nil;
  KeyServerSpecs := nil;
  KeyServerCount := 0;
  KeyServerList := nil;
  Result := PGPKeyServerInit;
  if Result <> 0 then Exit;
  Result := PGPKeyServerCreateThreadStorage(ServerThreadStorage);
  if Result <> 0 then Exit;
  try
    Result := GetServerList(KeyServerList);
    if (Result = 0) and (KeyServerList <> nil) then with KeyServerList do begin
      KeyServerSpecs := AllocMem(SizeOf(TPGPKeyServerSpec) * Count);
      if KeyServerSpecs <> nil then begin
	for KeyServerIndex := 0 to pred(Count) do begin
	  if SDKVersion >= PGP7XX then begin
	    Result := PGPNewKeyServer(Context, kPGPKeyServerClass_PGP, KeyServerSpecs.Spec.PGP7[KeyServerIndex].Server,
				      PGPONetURL(Context, PChar(Strings[KeyServerIndex])), PGPOLastOption(Context));
	  end
	  else begin
	    Result := PGPNewKeyServer(Context, kPGPKeyServerClass_PGP, KeyServerSpecs.Spec.PGP6[KeyServerIndex].Server,
				      PGPONetURL(Context, PChar(Strings[KeyServerIndex])), PGPOLastOption(Context));
	  end;
	  if Result <> 0 then Exit;
	  inc(KeyServerCount);
	end;
      end
      else Result := kPGPError_OutOfMemory;
    end;
  finally
    KeyServerList.Free;
  end;
end;

procedure FreeKeyServer(ServerThreadStorage: pPGPKeyServerThreadStorage;
			KeyServerSpecs: pPGPKeyServerSpec; KeyServerCount: PGPUInt32);
begin
  try
    for KeyServerCount := pred(KeyServerCount) downto 0 do begin
      if SDKVersion >= PGP7XX then
	PGPFreeKeyServer(KeyServerSpecs.Spec.PGP7[KeyServerCount].Server)
      else PGPFreeKeyServer(KeyServerSpecs.Spec.PGP6[KeyServerCount].Server);
    end;
  finally
    if KeyServerSpecs <> nil then FreeMem(KeyServerSpecs);
    PGPKeyServerDisposeThreadStorage(ServerThreadStorage);
    PGPKeyServerCleanup;
  end;
end;

function SendKeyToServerDialog(const Prompt: String; var KeyPropsList: TKeyPropsList;
			       KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  tlsContext	 : pPGPtlsContext;
  KeySetToSend	 : pPGPKeySet;
  ThreadStorage	 : pPGPKeyServerThreadStorage;
  KeyServerSpecs : pPGPKeyServerSpec;
  KeyServerCount : PGPUInt32;
  KeySetFailed	 : pPGPKeySet;
begin
  KeySetToSend := nil;
  KeySetFailed := nil;
  tlsContext := nil;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetMain, KeySetMain, 0, KeySetToSend);
    RedrawWindow(WinHandle);
    if Result <> 0 then Exit;
    try
      Result := PGPNewTLSContext(Context, tlsContext);
      if Result <> 0 then Exit;
      try
	Result := InitKeyServer(Context, ThreadStorage, KeyServerSpecs, KeyServerCount);
	if Result <> 0 then Exit;
	try
	  InstallHook(WinHandle);
	  Result := PGPSendToKeyServerDialog(Context, KeyServerSpecs^, tlsContext, KeySetToSend, KeySetFailed,
					     PGPOUIParentWindowHandle(Context, WinHandle), PGPOLastOption(Context));
	  RedrawWindow(WinHandle);
	finally
	  FreeKeyServer(ThreadStorage, KeyServerSpecs, KeyServerCount);
	  RemoveHook;
	end;
	if (KeySetFailed <> nil) and (KeySetToSend <> nil) then PGPRemoveKeys(KeySetFailed, KeySetToSend);
	GetKeySetProps(Context, KeySetToSend, KeyPropsList, KeyPropFlags, KeyFilterFlag_AllKeys, UserID_Ordering);
      finally
	PGPFreeTLSContext(tlsContext);
      end;
    finally
      PGPFreeKeySet(KeySetToSend);
      PGPFreeKeySet(KeySetFailed);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function SearchKeyServerOptions(Context: pPGPContext; const KeyData: String; FilterFlags: Integer;
				IncludeSubKeys, IncludeUTF8: Longbool; WinHandle: Integer;
				var OptionList: pPGPOptionList): Integer;
var
  KeyFilter	 : pPGPFilter;
  AddFilter	 : pPGPFilter;
  SearchFilter	 : pPGPFilter;
begin
  KeyFilter := nil;
  AddFilter := nil;
  SearchFilter := nil;
  OptionList := nil;
  Result := GetKeyFilterByAnyID(Context, KeyData, IncludeSubKeys, IncludeUTF8, KeyFilter);
  if Result <> 0 then Exit;
  try
    if FilterFlags <> KeyFilterFlag_AllKeys then begin
      Result := GetKeyFilter(Context, FilterFlags, AddFilter);
      if Result <> 0 then Exit;
    end;
    try
      if AddFilter <> nil then
	Result := PGPIntersectFilters(KeyFilter, AddFilter, SearchFilter)
      else SearchFilter := KeyFilter;
      KeyFilter := nil;
      AddFilter := nil;
      if Result <> 0 then Exit;
      try
	Result := PGPBuildOptionList(Context, OptionList,
	  [
	    PGPOKeyServerSearchFilter(Context, SearchFilter),
	    PGPOUIParentWindowHandle(Context, WinHandle)
	  ]);
      finally
	PGPFreeFilter(SearchFilter);
      end;
    finally
      if AddFilter <> nil then PGPFreeFilter(AddFilter);
    end;
  finally
    if KeyFilter <> nil then PGPFreeFilter(KeyFilter);
  end;
end;

function GetKeyFromServerDialog(const Prompt, KeyData: String; var KeyPropsList: TKeyPropsList;
				KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  tlsContext	 : pPGPtlsContext;
  ThreadStorage	 : pPGPKeyServerThreadStorage;
  KeyServerSpecs : pPGPKeyServerSpec;
  KeyServerCount : PGPUInt32;
  OptionList	 : pPGPOptionList;
  KeysFound	 : Pointer;
begin
  tlsContext := nil;
  OptionList := nil;
  KeysFound := nil;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := PGPNewTLSContext(Context, tlsContext);
    if Result <> 0 then Exit;
    try
      Result := InitKeyServer(Context, ThreadStorage, KeyServerSpecs, KeyServerCount);
      if Result <> 0 then Exit;
      try
	Result := SearchKeyServerOptions(Context, KeyData, KeyFilterFlag_AllKeys, false, false, WinHandle, OptionList);
	if Result <> 0 then Exit;
	try
	  InstallHook(WinHandle);
	  Result := PGPSearchKeyServerDialog(Context, KeyServerCount, KeyServerSpecs, tlsContext, PGPFalse,
					     KeysFound, OptionList, PGPOLastOption(Context));
	  RedrawWindow(WinHandle);
	finally
	  PGPFreeOptionList(OptionList);
	  RemoveHook;
	end;
	if Result <> 0 then Exit;
	if SDKVersion >= PGP7XX then begin
	  try
	    Result := SelectAddKeysToKeyRing(Context, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound)), KeySetMain,
					     KeyPropsList, Prompt, false, KeyPropFlags, WinHandle);
	  finally
	    PGPFreeKeyDB(pPGPKeyDB(KeysFound));
	  end;
	end
	else begin
	  try
	    Result := SelectAddKeysToKeyRing(Context, pPGPKeySet(KeysFound), KeySetMain,
					     KeyPropsList, Prompt, false, KeyPropFlags, WinHandle);
	  finally
	    PGPFreeKeySet(pPGPKeySet(KeysFound));
	  end;
	end;
      finally
	FreeKeyServer(ThreadStorage, KeyServerSpecs, KeyServerCount);
      end;
    finally
      PGPFreeTLSContext(tlsContext);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyImportDialog(const KeyPrompt, CertPassPrompt, KeyData: String; CertPassphrase: PChar;
			 var KeyPropsList: TKeyPropsList; FromFile: Longbool;
			 KeyPropFlags, WinHandle: Integer; var KeySetFound: Pointer): Integer;
type
  TInputFormat = (ifNone, ifPKCS_12, ifPKCS_7, ifPEM, ifKEY);
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  FileHandle	 : THandle;
  InputBuffer	 : String;
  InputSize	 : Longint;
  InputFormat	 : TInputFormat;
  OptionList	 : pPGPOptionList;
  PosKeyData	 : Longint;
  LenKeyData	 : Longint;
  KeysToImport	 : Pointer;
  KeySetSelected : pPGPKeySet;
  OptionListCopy : pPGPOptionList;
  KeysToAdd	 : pPGPKeyDB;
  CertsToImport	 : pPGPKeySet;
begin
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  OptionList := nil;
  KeysToImport := nil;
  KeySetSelected := nil;
  try
    if FromFile then begin
      FileHandle := FileOpen(KeyData, fmOpenReadWrite);
      if FileHandle <> THandle(-1) then begin
	InputSize := FileSeek(FileHandle, 0, 2);
	SetString(InputBuffer, nil, InputSize);
	FileSeek(FileHandle, 0, 0);
	FileRead(FileHandle, Pointer(InputBuffer)^, InputSize);
	FileClose(FileHandle);
      end
      else InputSize := 0;
    end
    else begin
      InputSize := Length(KeyData);
      InputBuffer := KeyData;
    end;
    if SDKVersion >= PGP7XX then
      InputFormat := ifNone
    else InputFormat := ifKEY;
    repeat
      PosKeyData := 1;
      LenKeyData := InputSize;
      case InputFormat of
	ifNone: begin
	  InputFormat := ifPKCS_12;
	  Result := PGPBuildOptionList(Context, OptionList, [PGPOInputFormat(Context, kPGPInputFormat_PKCS12)]);
	end;
	ifPKCS_12: begin
	  InputFormat := ifPKCS_7;
	  Result := PGPBuildOptionList(Context, OptionList, [PGPOInputFormat(Context, kPGPInputFormat_X509DataInPKCS7)]);
	end;
	ifPKCS_7 : begin
	  InputFormat := ifPEM;
	  PosKeyData := FirstStrPos(X509Cert_Beg, InputBuffer);
	  if PosKeyData > 0 then
	    LenKeyData := InputSize - pred(PosKeyData)
	  else Continue;
	  Result := PGPBuildOptionList(Context, OptionList, [PGPOInputFormat(Context, kPGPInputFormat_PEMEncodedX509Cert)]);
	end
      else
	InputFormat := ifKEY;
	Result := PGPBuildOptionList(Context, OptionList, [PGPONullOption(Context)]);
      end;
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(Context, @InputBuffer[PosKeyData], LenKeyData)]);
	if Result <> 0 then Exit;
	if SDKVersion >= PGP7XX then begin
	  Result := PGPImport(Context, KeysToImport, OptionList, PGPOLastOption(Context));
	  if Result = kPGPError_BadPassphrase then begin
	    repeat
	      Result := PGPCopyOptionList(OptionList, OptionListCopy);
	      if Result <> 0 then Exit;
	      try
		if CertPassPrompt <> '' then begin
		  CertPassphrase := nil;
		  Result := PassphraseDialog(Context, PGP_DLG, CertPassPrompt, CertPassphrase, WinHandle);
		  if Result <> 0 then Exit;
		end;
		try
		  Result := PGPAppendOptionList(OptionListCopy, [PGPOPassphrase(Context, CertPassphrase)]);
		  if Result <> 0 then Exit;
		  Result := PGPImport(Context, KeysToImport, OptionListCopy, PGPOLastOption(Context));
		finally
		  if CertPassPrompt <> '' then PGPFreeData(CertPassphrase);
		end;
	      finally
		PGPFreeOptionList(OptionListCopy);
	      end;
	    until (Result <> kPGPError_BadPassphrase) or (CertPassPrompt = '');
	  end
	  else if InputFormat = ifPEM then begin
	    KeySetSelected := PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysToImport));
	    repeat
	      PosKeyData := NextStrPos(X509Cert_Beg, InputBuffer, PosKeyData + Length(X509Cert_Beg) + Length(CRLF));
	      if PosKeyData > 0 then begin
		LenKeyData := InputSize - pred(PosKeyData);
		PGPFreeOptionList(OptionList);
		OptionList := nil;
		Result := PGPBuildOptionList(Context, OptionList,
		  [
		    PGPOInputFormat(Context, kPGPInputFormat_PEMEncodedX509Cert),
		    PGPOInputBuffer(Context, @InputBuffer[PosKeyData], LenKeyData)
		  ]);
		if Result <> 0 then Exit;
		KeysToAdd := nil;
		Result := PGPImport(Context, KeysToAdd, OptionList, PGPOLastOption(Context));
		if Result <> 0 then Exit;
		try
		  Result := PGPCopyKeys(PGPPeekKeyDBRootKeySet(KeysToAdd), KeysToImport, KeySetSelected);
		  if Result <> 0 then Exit;
		finally
		  PGPFreeKeyDB(KeysToAdd);
		end;
	      end;
	    until PosKeyData = 0;
	  end;
	end
	else Result := PGPImportKeySet(Context, KeysToImport, OptionList, PGPOLastOption(Context));
	if Result = 0 then begin
	  if KeySetFound = Ptr(-1) then begin
	    KeySetFound := KeysToImport;
	    if SDKVersion >= PGP7XX then begin
	      Result := GetKeySetProps(Context, PGPPeekKeyDBRootKeySet(KeySetFound), KeyPropsList, KeyPropFlags,
				       KeyFilterFlag_AllKeys, UserID_Ordering);
	    end
	    else begin
	      Result := GetKeySetProps(Context, KeySetFound, KeyPropsList, KeyPropFlags,
				       KeyFilterFlag_AllKeys, UserID_Ordering);
	    end;
	  end
	  else begin
	    try
	      if SDKVersion >= PGP7XX then begin
		CertsToImport := nil;
		if InputFormat = ifKEY then
		  CertsToImport := PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysToImport))
		else begin
		  Result := GetCertKeySetFromKeySet(PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysToImport)), false, CertsToImport);
		  if Result <> 0 then Exit;
		end;
		try
		  Result := SelectAddKeysToKeyRing(Context, CertsToImport, KeySetMain, KeyPropsList,
						   KeyPrompt, false, KeyPropFlags, WinHandle);
		finally
		  if InputFormat <> ifKEY then PGPFreeKeySet(CertsToImport);
		end;
	      end
	      else begin
		Result := SelectAddKeysToKeyRing(Context, pPGPKeySet(KeysToImport), KeySetMain,
						 KeyPropsList, KeyPrompt, false, KeyPropFlags, WinHandle);
	      end;
	    finally
	      if SDKVersion >= PGP7XX then
		PGPFreeKeyDB(pPGPKeyDB(KeysToImport))
	      else PGPFreeKeySet(pPGPKeySet(KeysToImport));
	      KeysToImport := nil;
	    end;
	  end;
	end
	else if (InputFormat < ifKEY) and (Result <> kPGPError_BadPassphrase) then Result := kPGPError_InvalidInputFormat;
      finally
	PGPFreeOptionList(OptionList);
	OptionList := nil;
      end;
    until (Result <> kPGPError_InvalidInputFormat) or (InputFormat = ifKEY);
    if (Result < 0) and (Result >= -40) then Result := kPGPError_InvalidInputFormat;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyExportDialog(Prompt: String; var KeyData: String; var KeyPropsList: TKeyPropsList;
			 ToFile, ExportCompatible, ExportPrivate: Longbool; KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeySetSelected : pPGPKeySet;
  OptionList	 : pPGPOptionList;
  FileSpec	 : pPGPFileSpec;
  KeyCount	 : PGPUInt32;
  ExportFormat	 : PGPExportFormat;
  KeyOutput	 : PChar;
  KeyLength	 : PGPUInt32;
begin
  KeySetSelected := nil;
  OptionList := nil;
  FileSpec := nil;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetMain, KeySetMain, 0, KeySetSelected);
    RedrawWindow(WinHandle);
    if Result <> 0 then Exit;
    try
      Result := PGPCountKeys(KeySetSelected, KeyCount);
      if Result <> 0 then Exit;
      if KeyCount <> 0 then begin
	if ExportCompatible then
	  ExportFormat := kPGPExportFormat_Basic
	else ExportFormat := kPGPExportFormat_Complete;
	Result := PGPBuildOptionList(Context, OptionList,
	  [
	    PGPOVersionString(Context, MyVersion),
	    PGPOExportFormat(Context, ExportFormat),
	    PGPOExportPrivateKeys(Context, PGPBoolean(ExportPrivate = true)),
	    PGPOExportPrivateSubKeys(Context, PGPBoolean(ExportPrivate = true))
	  ]);
	if Result <> 0 then Exit;
	try
	  if ToFile then begin
	    Result := PGPNewFileSpecFromFullPath(Context, PChar(KeyData), FileSpec);
	    if Result <> 0 then Exit;
	    try
	      if Result <> 0 then Exit;
	      Result := PGPAppendOptionList(OptionList, [PGPOOutputFile(Context, FileSpec)]);
	      if Result <> 0 then Exit;
	      if SDKVersion >= PGP7XX then begin
		Result := PGPAppendOptionList(OptionList, [PGPOExportKeySet(Context, KeySetSelected)]);
		if Result = 0 then Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	      end
	      else Result := PGPExportKeySet(KeySetSelected, OptionList, PGPOLastOption(Context));
	      if Result = 0 then Result := GetKeySetProps(Context, KeySetSelected, KeyPropsList, KeyPropFlags,
							  KeyFilterFlag_AllKeys, UserID_Ordering);
	      if Result > 0 then Result := 0;
	    finally
	      PGPFreeFileSpec(FileSpec);
	    end;
	  end
	  else begin
	    Result := PGPAppendOptionList(OptionList, [PGPOAllocatedOutputBuffer(Context, KeyOutput, $FFFFFFFF, KeyLength)]);
	    if Result <> 0 then Exit;
	    if SDKVersion >= PGP7XX then begin
	      Result := PGPAppendOptionList(OptionList, [PGPOExportKeySet(Context, KeySetSelected)]);
	      if Result = 0 then Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	    end
	    else Result := PGPExportKeySet(KeySetSelected, OptionList, PGPOLastOption(Context));
	    if (Result = 0) and (KeyLength <> 0) then begin
	      try
		SetString(KeyData, nil, KeyLength);
		StrPLCopy(PChar(KeyData), KeyOutput, KeyLength);
	      finally
		PGPFreeData(KeyOutput);
	      end;
	    end
	    else KeyData := '';
	    if Result = 0 then Result := GetKeySetProps(Context, KeySetSelected, KeyPropsList, KeyPropFlags,
							KeyFilterFlag_AllKeys, UserID_Ordering);
	    if Result > 0 then Result := 0;
	  end;
	finally
	  PGPFreeOptionList(OptionList);
	end;
      end;
    finally
      PGPFreeKeySet(KeySetSelected);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyRevokeDialog(const KeyHexID, PassPrompt: String; WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeySetFound	 : pPGPKeySet;
  Passphrase	 : PChar;
begin
  KeySetFound := nil;
  Passphrase := nil;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyHexID, KeySetFound);
    if Result <> 0 then Exit;
    try
      Result := KeyPassphraseDialog(Context, KeySetFound, Passphrase, PassPrompt, WinHandle);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
    finally
      PGPFreeKeySet(KeySetFound);
    end;
    try
      KeyRevoke(PChar(KeyHexID), Passphrase);
    finally
      PGPFreeData(Passphrase);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function KeyPassChangeDialog(const KeyHexID, OldPrompt, NewPrompt: String;
			     MinPassLength, MinPassQuality, WinHandle: Integer): Integer;
var
  Context	 : pPGPContext;
  KeySetMain	 : pPGPKeySet;
  KeySetFound	 : pPGPKeySet;
  OldPassphrase	 : PChar;
  NewPassphrase  : PChar;
begin
  KeySetFound := nil;
  OldPassphrase := nil;
  NewPassphrase := nil;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyHexID, KeySetFound);
    if Result <> 0 then Exit;
    try
      Result := KeyPassphraseDialog(Context, KeySetFound, OldPassphrase, OldPrompt, WinHandle);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
    finally
      PGPFreeKeySet(KeySetFound);
    end;
    try
      Result := ConfirmationPassphraseDialog(Context, NewPassphrase,
					     MinPassLength, MinPassQuality, true,
					     NewPrompt, WinHandle);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
      try
	ChangePassphrase(PChar(KeyHexID), OldPassphrase, NewPassphrase);
      finally
	PGPFreeData(NewPassphrase);
      end;
    finally
      PGPFreeData(OldPassphrase);
    end;
  finally
    KeyRings.FreeKeyRings;
  end;
end;

function SimplePassphraseDialog(const Title, Prompt: String; WinHandle: Integer;
				PassphraseCallback: TPassphraseCallback; const CallbackParam): Integer;
var
  Context	 : pPGPContext;
  Passphrase	 : PChar;
begin
  Result := kPGPError_BadParams;
  if not Assigned(PassphraseCallback) then Exit;
  Result := PGPNewContext(kPGPsdkAPIVersion, Context);
  if Result <> 0 then Exit;
  Passphrase := nil;
  try
    Result := PassphraseDialog(Context, Title, Prompt, Passphrase, WinHandle);
    if Result = 0 then PassphraseCallback(Passphrase, CallbackParam);
  finally
    PGPFreeData(Passphrase);
    PGPFreeContext(Context);
  end;
end;

function SelectX509CertifiedKeysDialog(Context: pPGPContext; KeySetX509CertifiedKeys: pPGPKeySet;
				       var KeySetSelected: pPGPKeySet; var CertSelected: pPGPKeyDBObj;
				       const Prompt: String; CertsOnly, SingleKey: Longbool;
				       WinHandle: Integer): Integer;
begin
  KeySetSelected := nil;
  if (SDKVersion >= PGP81X) and CertsOnly and SingleKey then begin
    Result := PGPclSelectX509Cert(Context, WinHandle, Prompt, KeySetX509CertifiedKeys,
				  kPGPclDefaultX509SelectionFlags, CertSelected);
    if Result = 0 then Result := PGPNewSingletonKeySet(CertSelected, KeySetSelected);
  end
  else begin
    Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetX509CertifiedKeys, nil,
			      ord(SingleKey = true) * PGPCL_SINGLESELECTION, KeySetSelected);
  end;
  RedrawWindow(WinHandle);
end;

function SelectAddKeysToKeyRing(Context: pPGPContext; KeysToAdd: pPGPKeySet; KeySetMain: pPGPKeySet;
				var KeyPropsList: TKeyPropsList; const Prompt: String; SingleKey: Longbool;
				KeyPropFlags, WinHandle: Integer): PGPError;
var
  KeyCount	 : PGPUInt32;
  KeyDBTemp	 : pPGPKeyDB;
  KeySetTemp	 : pPGPKeySet;
  KeySetSelected : pPGPKeySet;
  KeySetSelToAdd : pPGPKeySet;
begin
  KeySetSelected := nil;
  KeySetSelToAdd := nil;
  Result := PGPCountKeys(KeysToAdd, KeyCount);
  if Result <> 0 then Exit;
  if KeyCount <> 0 then begin
    if SDKVersion >= PGP7XX then begin
      KeyDBTemp := nil;
      Result := PGPNewKeyDB(Context, KeyDBTemp);
      if Result <> 0 then Exit;
      try
	KeySetTemp := nil;
	Result := PGPCopyKeys(KeySetMain, KeyDBTemp, KeySetTemp);
	if Result <> 0 then Exit;
	PGPFreeKeySet(KeySetTemp);
	Result := PGPCopyKeys(KeysToAdd, KeyDBTemp, KeySetTemp);
	if Result <> 0 then Exit;
	try
	  // KeyDBTemp and KeySetTemp are needed for displaying "web of trust" validities
	  Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeySetTemp, nil,
				    ord(SingleKey = true) * PGPCL_SINGLESELECTION, KeySetSelected);
	  RedrawWindow(WinHandle);
	  if Result <> 0 then Exit;
	  try
	    Result := PGPCopyKeys(KeySetSelected, PGPPeekKeySetKeyDB(KeySetMain), KeySetSelToAdd);
	  finally
	    PGPFreeKeySet(KeySetSelected);
	  end;
	finally
	  PGPFreeKeySet(KeySetTemp);
	end;
      finally
	PGPFreeKeyDB(KeyDBTemp);
      end;
    end
    else begin
      Result := PGPclSelectKeys(Context, WinHandle, Prompt, KeysToAdd, KeySetMain,
				ord(SingleKey = true) * PGPCL_SINGLESELECTION, KeySetSelected);
      RedrawWindow(WinHandle);
      if Result <> 0 then Exit;
      KeySetSelToAdd := KeySetSelected;
      Result := PGPAddKeys(KeySetSelToAdd, KeySetMain);
    end;
    if Result <> 0 then Exit;
    try
      Result := KeyRings.UpdateKeyRings;
      if Result <> 0 then Exit;
      Result := GetKeySetProps(Context, KeySetSelToAdd, KeyPropsList, KeyPropFlags,
			       KeyFilterFlag_AllKeys, UserID_Ordering);
      if Result > 0 then Result := 0;
    finally
      PGPFreeKeySet(KeySetSelToAdd);
    end;
  end
  else Result := kPGPError_PublicKeyNotFound;
end;

function RecipientsDialog(Context: pPGPContext; AllKeys: pPGPKeySet;
			  AlwaysDisplayDialog: PGPBoolean; SMime: Longbool;
			  const DefaultKey: String; var RecipientKeys: Pointer;
			  const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  GroupFileSpec	 : pPGPFileSpec;
  GroupSet	 : pPGPGroupSet;
  GroupOption	 : pPGPOptionList;
  OptionList	 : pPGPOptionList;
  RecipientSpec	 : TPGPRecipientSpec;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  RecipientKeys := nil;
  GroupFileSpec := nil;
  GroupSet := nil;
  OptionList := nil;
  if SMime then
    Result := kPGPError_FeatureNotAvailable
  else Result := PGPNewFileSpecFromFullPath(Context, PChar(KeyRings.GroupsFile), GroupFileSpec);
  try
    if Result = 0 then Result := PGPNewGroupSetFromFile(Context, GroupFileSpec, GroupSet);
    try
      if Result = 0 then
	GroupOption := PGPOUIRecipientGroups(Context, GroupSet)
      else GroupOption := PGPONullOption(Context);
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOUIDisplayMarginalValidity(Context, PGPTrue),
	  GroupOption,
	  PGPOUIWindowTitle(Context, PGP_DLG),
	  PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
	  PGPOUIParentWindowHandle(Context, WinHandle)
	]);
      if Result <> 0 then Exit;
      if DefaultKey <> '' then begin
	FillChar(RecipientSpec, SizeOf(TPGPRecipientSpec), 0);
	RecipientSpec.AType := kPGPRecipientSpecType_Key;
	Result := GetKeyByHexID(AllKeys, DefaultKey, RecipientSpec.ID.Key);
	if Result = 0 then PGPAppendOptionList(OptionList, [PGPOUIDefaultRecipients(Context, 1, @RecipientSpec)]);
      end;
      try
	if SDKVersion >= PGP7XX then begin
	  Result := PGPRecipientDialog(Context, PGPPeekKeySetKeyDB(AllKeys), AlwaysDisplayDialog,
				       pPGPKeyDB(RecipientKeys), OptionList, PGPOLastOption(Context));
	end
	else begin
	  Result := PGPRecipientDialog(Context, AllKeys, AlwaysDisplayDialog,
				       pPGPKeySet(RecipientKeys), OptionList, PGPOLastOption(Context));
	end;
	RedrawWindow(WinHandle);
      finally
	PGPFreeOptionList(OptionList);
      end;
    finally
      if GroupSet <> nil then PGPFreeGroupSet(GroupSet);
    end;
  finally
    PGPFreeFileSpec(GroupFileSpec);
  end;
end;

function KeyServerDialog(Context: pPGPContext; KeySetMain: pPGPKeySet; const KeyData: String;
			 FilterFlags: Integer; var KeysFound: Pointer; WinHandle: Integer): PGPError;
var
  tlsContext	 : pPGPtlsContext;
  ThreadStorage	 : pPGPKeyServerThreadStorage;
  KeyServerSpecs : pPGPKeyServerSpec;
  KeyServerCount : PGPUInt32;
  OptionList	 : pPGPOptionList;
begin
  KeysFound := nil;
  tlsContext := nil;
  OptionList := nil;
  Result := PGPNewTLSContext(Context, tlsContext);
  if Result <> 0 then Exit;
  try
    Result := InitKeyServer(Context, ThreadStorage, KeyServerSpecs, KeyServerCount);
    if Result <> 0 then Exit;
    try
      Result := SearchKeyServerOptions(Context, KeyData, FilterFlags, false, false, WinHandle, OptionList);
      if Result <> 0 then Exit;
      try
	InstallHook(WinHandle);
	Result := PGPSearchKeyServerDialog(Context, KeyServerCount, KeyServerSpecs, tlsContext, PGPFalse,
					   KeysFound, OptionList, PGPOLastOption(Context));
	RedrawWindow(WinHandle);
      finally
	PGPFreeOptionList(OptionList);
	RemoveHook;
      end;
    finally
      FreeKeyServer(ThreadStorage, KeyServerSpecs, KeyServerCount);
    end;
  finally
    PGPFreeTLSContext(tlsContext);
  end;
end;

function CollectRandomDataDialog(Context: pPGPContext; NeededEntropyBits: PGPUInt32; WinHandle: Integer): PGPError;
begin
  Result := PGPCollectRandomDataDialog(Context,
				       NeededEntropyBits,
				       PGPOUIParentWindowHandle(Context, WinHandle),
				       PGPOLastOption(Context));
  RedrawWindow(WinHandle);
end;

function PassphraseDialog(Context: pPGPContext; const Title, Prompt: String;
			  var Passphrase: PChar; WinHandle: Integer): Integer;
var
  Utf8Title	 : String;
  Utf8Prompt	 : String;
  OptionList	 : pPGPOptionList;
begin
  if Utf8Title = '' then Utf8Title := PGP_DLG;
  if SDKVersion >= PGP9XX then begin
    Utf8Title := AnsiStringToUtf8(Title);
    Utf8Prompt := AnsiStringToUtf8(Prompt);
  end
  else begin
    Utf8Title := Title;
    Utf8Prompt := Prompt;
  end;
  Passphrase := nil;
  OptionList := nil;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PChar(Utf8Title)),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function ConvEncPassphraseDialog(Context: pPGPContext; var Passphrase: PChar;
				 const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  OptionList  : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := nil;
  OptionList := nil;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPConventionalEncryptionPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function ConvDecPassphraseDialog(Context: pPGPContext; var Passphrase: PChar;
				 const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := nil;
  OptionList := nil;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPConventionalDecryptionPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function KeyPassphraseDialog(Context: pPGPContext; KeySet: pPGPKeySet; var Passphrase: PChar;
			     const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  Key		 : pPGPKey;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := nil;
  OptionList := nil;
  Result := GetKeyFromKeySet(KeySet, Key);
  if Result <> 0 then Exit;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPKeyPassphraseDialog(Context, Key, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function SigningPassphraseDialog(Context: pPGPContext; KeySetMain: pPGPKeySet;
				 var SigningKey: pPGPKey; var SignKeyPass: PChar;
				 const DefaultKeyID: String; FindMatchingKey: Longbool;
				 const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  DefaultKey	 : pPGPKey;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  SigningKey := nil;
  SignKeyPass := nil;
  OptionList := nil;
  if (DefaultKeyID = '') or (GetKeyByHexID(KeySetMain, PChar(DefaultKeyID), DefaultKey) <> 0) then DefaultKey := nil;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIDefaultKey(Context, DefaultKey),
      PGPOUIFindMatchingKey(Context, PGPBoolean(FindMatchingKey = true)),
      PGPOUIVerifyPassphrase(Context, PGPTrue),
      PGPOUIOutputPassphrase(Context, SignKeyPass),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    if SDKVersion >= PGP7XX then begin
      Result := PGPSigningPassphraseDialog(Context, PGPPeekKeySetKeyDB(KeySetMain), SigningKey,
					   OptionList, PGPOLastOption(Context));
    end
    else Result := PGPSigningPassphraseDialog(Context, KeySetMain, SigningKey, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function DecryptionPassphraseDialog(Context: pPGPContext; RecipientKeys: pPGPKeySet;
				    KeyIDCount: PGPUInt32; const KeyIDList: pPGPKeyID7;
				    var DecryptionKey: pPGPKey; var DecryptKeyPass: PChar;
				    FindMatchingKey: Longbool; const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  DecryptionKey := nil;
  DecryptKeyPass := nil;
  OptionList := nil;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIFindMatchingKey(Context, PGPBoolean(FindMatchingKey = true)),
      PGPOUIVerifyPassphrase(Context, PGPTrue),
      PGPOUIOutputPassphrase(Context, DecryptKeyPass),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPDecryptionPassphraseDialog(Context, RecipientKeys, KeyIDCount, KeyIDList,
					    DecryptionKey, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function ConfirmationPassphraseDialog(Context: pPGPContext; var Passphrase: PChar;
				      MinPassLength, MinPassQuality: Integer;
				      ShowPassphraseQuality: Longbool;
				      const Prompt: String; WinHandle: Integer): PGPError;
var
  Utf8Prompt	 : String;
  OptionList	 : pPGPOptionList;
begin
  if SDKVersion >= PGP9XX then
    Utf8Prompt := AnsiStringToUtf8(Prompt)
  else Utf8Prompt := Prompt;
  Passphrase := nil;
  OptionList := nil;
  Result := PGPBuildOptionList(Context, OptionList,
    [
      PGPOUIWindowTitle(Context, PGP_DLG),
      PGPOUIDialogPrompt(Context, PChar(Utf8Prompt)),
      PGPOUIMinimumPassphraseLength(Context, MinPassLength),
      PGPOUIMinimumPassphraseQuality(Context, MinPassQuality),
      PGPOUIShowPassphraseQuality(Context, PGPBoolean(ShowPassphraseQuality = true)),
      PGPOUIOutputPassphrase(Context, Passphrase),
      PGPOUIParentWindowHandle(Context, WinHandle)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPConfirmationPassphraseDialog(Context, OptionList, PGPOLastOption(Context));
    RedrawWindow(WinHandle);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;


end.

