{$J+,Z4}
unit PGPEncode;

{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  Windows,
  Classes,
  StrTools,
  SysUtils,
  SecureFuncs,
  PSMimeTools,
  KeyPropTypes,
  X509Types,
  pgpBase,
  pgpErrors,
  pgpPubTypes,
  pgpUtilities,
  pgpOptionList,
  pgpMemoryMgr,
  pgpEvents,
  pgpKeys,
  pgpTLS,
  KeyFuncs,
  PrefFuncs,
  X509Funcs,
  PGPDialogs,
  KeyDialog,
  HelpCtrl;

type
  TSignAlgorithm = (
    HashAlgorithm_Default,
    HashAlgorithm_MD5,
    HashAlgorithm_SHA,
    HashAlgorithm_RIPEMD160,
    HashAlgorithm_SHA256,
    HashAlgorithm_SHA384,
    HashAlgorithm_SHA512,
    HashAlgorithm_SHA224
  );
  TConventionalAlgorithm = (
    CipherAlgorithm_IDEA,
    CipherAlgorithm_3DES,
    CipherAlgorithm_CAST5,
    CipherAlgorithm_AES128,
    CipherAlgorithm_AES192,
    CipherAlgorithm_AES256,
    CipherAlgorithm_Twofish256
  );
  TOnShowProgress = procedure(BytesProcessed, BytesTotal: Longint) of Object;
  pPGPEncode = ^TPGPEncode;
  TPGPEncode = Class
  private
    // internal
    FContext: pPGPContext;
    FKeySetMain: pPGPKeySet;
    FPreferences: TPreferenceRec;
    FAllocatedOutputBuffer: PChar;
    FActualOutputSize: PGPSize;
    FInputSize: PGPSize;
    FInputBuffer: PChar;
    FInputFileName: String;
    FPassphrase: PChar;
    FSigningKey: pPGPKey;
    FEncryptKeySet: pPGPKeySet;
    FClear: Longbool;
    FCompress: Longbool;
    FConventional: Longbool;
    FDetachedSign: Longbool;
    FEncrypt: Longbool;
    FFileToFile: Longbool;
    FInsertCharset: Longbool;
    FSMime: Longbool;
    FSign: Longbool;
    FWait: HCursor;
    FArrow: HCursor;
    // public properties
    FKeysFromServer: Longint;
    FMimeBodyOffset: Longint;
    FMimeSeparator: String;
    FParentHandle: THandle;
    FOutputBuffer: String;
    FKeyPropsList: TKeyPropsList;
    // published properties
    FAlwaysShowKeyDlg: Longbool;
    FArmor: Longbool;
    FCharset: String;
    FComment: String;
    FConventionalAlgorithm: TConventionalAlgorithm;
    FEncryptToSelf: Longbool;
    FEyesOnly: Longbool;
    FFileOutput: Longbool;
    FGroupSelected: Longbool;
    FMime: Longbool;
    FMultipleRecipients: Longbool;
    FOmitMimeVersion: Longbool;
    FParanoiaSwitch: Longbool;
    {FProgressInterval: Cardinal;}
    FSignAlgorithm: TSignAlgorithm;
    FTextmode: Longbool;
    FEncryptKeyID: String;
    FSignKeyID: String;
    FSignUserID: String;
    FGroupText: String;
    FKeyServerBtn: String;
    FKeyUseAddBtn: String;
    FKeyCancelBtn: String;
    FKeyOkBtn: String;
    FSelectedKeysText: String;
    FSelectAllBtn: String;
    FUnselectBtn: String;
    FCertificateText: String;
    FCertifiedText: String;
    FCertPropsViewerBtn: String;
    FCertSelDlgPrompt: String;
    FKeySelDlgPrompt: String;
    FShowKeyDetails: String;
    FPassDlgPrompt: String;
    FRecipients: String;
    FOutputFileName: String;
    FKeyDialogHelpIndex: Longint;
    FHelpHandler: THelpHandler;
    FSecure: TSecure;
    // events
    {FOnShowProgress: TOnShowProgress;}
    procedure SetClear(Value: Longbool);
    procedure SetConventional(Value: Longbool);
    procedure SetEncrypt(Value: Longbool);
    procedure SetSign(Value: Longbool);
    function  InitEncode: PGPError;
    procedure FinitEncode;
    function  AddSignOptions(const OptionList: pPGPOptionList): PGPError;
    function  GetOptionList(var OptionList: pPGPOptionList): PGPError;
    function  ReadInputFile: PGPError;
    function  SetInputOption(IsFile: Longbool; var OptionList: pPGPOptionList): PGPError;
    function  SetOutputOption(var OptionList: pPGPOptionList): PGPError;
    function  WriteOutputFile: PGPError;
    function  Encode(IsFile: Longbool): Longint;
  public
    KeyTrusts: TKeyTrusts;
    KeyValidities: TKeyValidities;
    KeyAlgorithms: TKeyAlgorithms;
    HashAlgorithms: THashAlgorithms;
    CertValidities: TCertValidities;
    constructor Create;
    destructor Destroy; override;
    function KeyEncryptBuffer(const DataBuffer: String; Sign, SMime: Longbool): Longint;
    function KeyEncryptFile(const FileName: String; Sign, SMime: Longbool): Longint;
    function ConventionalEncryptBuffer(const DataBuffer: String): Longint;
    function ConventionalEncryptFile(const FileName: String): Longint;
    function ClearSignBuffer(const DataBuffer: String; SMime: Longbool): Longint;
    function ClearSignFile(const FileName: String; SMime: Longbool): Longint;
    function ArmorBuffer(const DataBuffer: String): Longint;
    function ArmorFile(const FileName: String): Longint;
    function DetachedSignBuffer(const DataBuffer: String): Longint;
    function DetachedSignFile(const FileName: String): Longint;
    function GetEncryptKeyID(var EncryptKeyID, SignKeyID: String): Longint;
    property KeysFromServer: Longint
      read FKeysFromServer;
    property MimeBodyOffset: Longint
      read FMimeBodyOffset;
    property MimeSeparator: String
      read FMimeSeparator;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
    property OutputBuffer: String
      read FOutputBuffer;
    property KeyPropsList: TKeyPropsList
      read FKeyPropsList;
  published
    property AlwaysShowKeyDlg: Longbool
      read FAlwaysShowKeyDlg
      write FAlwaysShowKeyDlg;
    property Armor: Longbool
      read FArmor
      write FArmor;
    property Charset: String
      read FCharset
      write FCharset;
    property Comment: String
      read FComment
      write FComment;
    property ConventionalAlgorithm: TConventionalAlgorithm
      read FConventionalAlgorithm
      write FConventionalAlgorithm;
    property EncryptToSelf: Longbool
      read FEncryptToSelf
      write FEncryptToSelf;
    property EyesOnly: Longbool
      read FEyesOnly
      write FEyesOnly;
    property FileOutput: Longbool
      read FFileOutput
      write FFileOutput;
    property GroupSelected: Longbool
      read FGroupSelected
      write FGroupSelected;
    property Mime: Longbool
      read FMime
      write FMime;
    property MultipleRecipients: Longbool
      read FMultipleRecipients
      write FMultipleRecipients;
    property OmitMimeVersion: Longbool
      read FOmitMimeVersion
      write FOmitMimeVersion;
    property ParanoiaSwitch: Longbool
      read FParanoiaSwitch
      write FParanoiaSwitch;
    {property ProgressInterval: Cardinal
      read FProgressInterval
      write FProgressInterval;}
    property SignAlgorithm: TSignAlgorithm
      read FSignAlgorithm
      write FSignAlgorithm;
    property Textmode: Longbool
      read FTextmode
      write FTextmode;
    property EncryptKeyID: String
      read FEncryptKeyID
      write FEncryptKeyID;
    property SignKeyID: String
      read FSignKeyID
      write FSignKeyID;
    property SignUserID: String
      read FSignUserID
      write FSignUserID;
    property GroupText: String
      read FGroupText
      write FGroupText;
    property KeyServerBtn: String
      read FKeyServerBtn
      write FKeyServerBtn;
    property KeyUseAddBtn: String
      read FKeyUseAddBtn
      write FKeyUseAddBtn;
    property KeyCancelBtn: String
      read FKeyCancelBtn
      write FKeyCancelBtn;
    property KeyOkBtn: String
      read FKeyOkBtn
      write FKeyOkBtn;
    property SelectedKeysText: String
      read FSelectedKeysText
      write FSelectedKeysText;
    property SelectAllBtn: String
      read FSelectAllBtn
      write FSelectAllBtn;
    property UnselectBtn: String
      read FUnselectBtn
      write FUnselectBtn;
    property CertificateText: String
      read FCertificateText
      write FCertificateText;
    property CertifiedText: String
      read FCertifiedText
      write FCertifiedText;
    property CertPropsViewerBtn: String
      read FCertPropsViewerBtn
      write FCertPropsViewerBtn;
    property CertSelDlgPrompt: String
      read FCertSelDlgPrompt
      write FCertSelDlgPrompt;
    property KeySelDlgPrompt: String
      read FKeySelDlgPrompt
      write FKeySelDlgPrompt;
    property ShowKeyDetails: String
      read FShowKeyDetails
      write FShowKeyDetails;
    property PassDlgPrompt: String
      read FPassDlgPrompt
      write FPassDlgPrompt;
    property Recipients: String
      read FRecipients
      write FRecipients;
    property OutputFileName: String
      read FOutputFileName
      write FOutputFileName;
    property KeyDialogHelpIndex: Longint
      read FKeyDialogHelpIndex
      write FKeyDialogHelpIndex;
    property HelpHandler: THelpHandler
      read FHelpHandler
      write FHelpHandler;
    property Secure: TSecure
      read FSecure
      write FSecure;
    {property OnShowProgress: TOnShowProgress
      read FOnShowProgress
      write FOnShowProgress;}
  end;


implementation

function EventHandler(Context: pPGPContext; Event: pPGPEvent; UserValue: PGPUserValue): PGPError; cdecl;
begin
  Result := 0;
  with TPGPEncode(UserValue) do begin
    case Event^.EType of
      kPGPEvent_NullEvent:	{if not FClear then begin
				  if TMethod(FOnShowProgress).Code <> nil then begin
				    // BytesTotal always stays 0 => use FInputSize (at least in 6.5.X)
				    with Event^.EData.NullData do FOnShowProgress(BytesWritten, FInputSize);
				  end;
				  ProcessMessages;
				end};
      kPGPEvent_InitialEvent:	;
      kPGPEvent_FinalEvent:	;
      kPGPEvent_ErrorEvent:	Result := Event^.EData.ErrorData.Error;
      kPGPEvent_WarningEvent:	;
    end;
  end;
end;

constructor TPGPEncode.Create;
begin
  inherited;
  {FProgressInterval := 1000;}
  FWait := LoadCursor(0, IDC_WAIT);
  FArrow := LoadCursor(0, IDC_ARROW);
  FCharset := 'Windows-' + IntToStr(GetACP);
  FKeyPropsList := TKeyPropsList.Create(spgpKeyPropFlag_IDFlags);
end;

destructor TPGPEncode.Destroy;
begin
  FKeyPropsList.Free;
  inherited Destroy;
end;

procedure TPGPEncode.SetClear(Value: Longbool);
begin
  if Value <> FClear then FClear := Value;
  if FClear then begin
    FSign := true;
    FArmor := true;
    FTextmode := true;
    FEncrypt := false;
    FCompress := false;
    FConventional := false;
    FDetachedSign := false;
  end;
end;

procedure TPGPEncode.SetConventional(Value: Longbool);
begin
  if Value <> FConventional then FConventional := Value;
  if FConventional then begin
    FDetachedSign := false;
    FEncrypt := false;
    FClear := false;
    FSign := false;
  end;
end;

procedure TPGPEncode.SetEncrypt(Value: Longbool);
begin
  if Value <> FEncrypt then FEncrypt := Value;
  if FEncrypt then begin
    FDetachedSign := false;
    FConventional := false;
    FClear := false;
  end;
end;

procedure TPGPEncode.SetSign(Value: Longbool);
begin
  if Value <> FSign then FSign := Value;
  if FSign then FConventional := false;
end;

function TPGPEncode.InitEncode: PGPError;
begin
  FPassphrase := nil;
  FOutputBuffer := '';
  FActualOutputSize := 0;
  FAllocatedOutputBuffer := nil;
  FKeyPropsList.Clear;
  FKeysFromServer := 0;
  FMimeBodyOffset := 0;
  SetString(FMimeSeparator, nil, SizeOf(TMimeSeparator));
  PChar(Pointer(FMimeSeparator))^ := #0;
  Result := KeyRings.InitKeyRings(FContext, FKeySetMain);
  GetPreferences(FPreferences, PrefsFlag_GroupsFile or PrefsFlag_DefaultKeyID);
end;

procedure TPGPEncode.FinitEncode;
begin
  SetCursor(FArrow);
  KeyRings.FreeKeyRings;
end;

function SignToHashAlgorithm(SignAlgorithm: TSignAlgorithm): THashAlgorithm;
begin
  Result := THashAlgorithm(ord(SignAlgorithm) + ord(SignAlgorithm > HashAlgorithm_Default));
end;

function GetPGPHashAlgorithm(Algorithm: TSignAlgorithm): PGPHashAlgorithm;
begin
  Result := PGPHashAlgorithm(
    ord(Algorithm) +
    ord(Algorithm > HashAlgorithm_RIPEMD160) * ord(pred(kPGPHashAlgorithm_SHA256 - kPGPHashAlgorithm_RIPEMD160))
  );
end;

function TPGPEncode.AddSignOptions(const OptionList: pPGPOptionList): PGPError;
var
  SignKeySet	: pPGPKeySet;
  SigningCert	: pPGPKeyDBObj;
  KeyPropsList	: TKeyPropsList;
  SignKeyHexID	: String;
  SMimeOptions	: pPGPOptionList;
begin
  Result := 0;
  FSigningKey := nil;
  if FSign or FDetachedSign then begin
    if FSMime then begin
      SignKeySet := nil;
      Result := GetX509CertifierSet(FKeySetMain, SignKeySet, false, true, true);
      if Result <> 0 then Exit;
      SigningCert := nil;
    end
    else SignKeySet := FKeySetMain;
    try
      KeyPropsList := nil;
      try
	if FindKeyProps(FSignKeyID, KeyPropsList,
			spgpKeyPropFlag_KeyID,
			KeyFilterFlag_CanSign,
			Creation_Ordering) <> 1 then begin
	  FSignKeyID := FPreferences.DefaultKeyID;
	  if (FSignKeyID = '') and (KeyPropsList.Count > 0) then FSignKeyID := KeyPropsList[0];
	end
	else FSignKeyID := KeyPropsList[0];
      finally
	KeyPropsList.Free;
      end;
      if GetHexIDByAnyID(FContext, SignKeySet, FSignKeyID, SignKeyHexID) = 1 then begin
	FSignKeyID := SignKeyHexID;
	Result := GetKeyByHexID(SignKeySet, PChar(FSignKeyID), FSigningKey);
      end
      else Result := kPGPError_KeyUnusableForSignature;
      if Result <> 0 then Exit;
      if (FSecure = nil) or not FSecure.GetCachedPassphrase(FSigningKey, FPassphrase, true) then begin
	if FSMime then begin
	  Result := X509GetCertAndPassphraseDialog(FContext, SignKeySet, FSigningKey, SigningCert, FPassphrase,
						   FSignUserID, FSignKeyID, FPassDlgPrompt, FParentHandle);
	  if Result <> 0 then Exit;
	end
	else begin
	  Result := SigningPassphraseDialog(FContext, SignKeySet, FSigningKey, FPassphrase, FSignKeyID,
					    true, FPassDlgPrompt, FParentHandle);
	end;
      end;
      if Result = 0 then begin
	if FSMime and not (FSignAlgorithm in [HashAlgorithm_MD5, HashAlgorithm_SHA]) then begin
	  FSignAlgorithm := HashAlgorithm_Default;
	end;
	if FSignAlgorithm = HashAlgorithm_Default then begin
	  case GetKeyPropKeyAlg(FSigningKey) of
	    KeyAlgorithm_RSA..KeyAlgorithm_RSASignOnly: if GetKeyPropLegacy(FSigningKey) then
	      FSignAlgorithm := HashAlgorithm_MD5
	    else FSignAlgorithm := HashAlgorithm_SHA;
	    KeyAlgorithm_DH..KeyAlgorithm_DHDSS: FSignAlgorithm := HashAlgorithm_SHA;
	  else
	    Result := kPGPError_UnknownPublicKeyAlgorithm;
	    Exit;
	  end;
	end;
	if FSMime then begin
	  if SigningCert = nil then SigningCert := GetValidX509CertFromKey(FSignUserID, FSigningKey, nil, SignKeySet);
	  Result := PGPBuildOptionList(FContext, SMimeOptions,
	    [
	      PGPOPassphrase(FContext, FPassphrase),
	      PGPOSMIMESigner(FContext, SigningCert)
	    ]);
	  if Result <> 0 then Exit;
	  try
	    Result := PGPAppendOptionList(OptionList,
	      [
		PGPOSignWithKey(FContext, FSigningKey, SMimeOptions, PGPOLastOption(FContext)),
		PGPOHashAlgorithm(FContext, GetPGPHashAlgorithm(FSignAlgorithm)),
		PGPODetachedSig(FContext, PGPOLastOption(FContext), nil)
	      ]);
	  finally
	    PGPFreeOptionList(SMimeOptions);
	  end;
	end
	else begin
	  Result := PGPAppendOptionList(OptionList,
	    [
	      PGPOSignWithKey(FContext, FSigningKey, PGPOPassphrase(FContext, FPassphrase), PGPOLastOption(FContext)),
	      PGPOHashAlgorithm(FContext, GetPGPHashAlgorithm(FSignAlgorithm))
	    ]);
	end;
      end;
    finally
      if FSMime then PGPFreeKeySet(SignKeySet);
    end;
  end;
  if (Result = 0) and (FSigningKey <> nil) then FSignKeyID := GetKeyPropKeyID(FSigningKey);
end;

function TPGPEncode.GetOptionList(var OptionList: pPGPOptionList): PGPError;
var
  AlgorithmList	: TPGPCipherAlgorithms;
  AlgCount	: Longint;
begin
  if FSMime then begin
    if FEncrypt then begin
      Result := PGPBuildOptionList(FContext, OptionList,
	[
	  {PGPOSendNullEvents(FContext, FProgressInterval),}
	  PGPOEventHandler(FContext, EventHandler, Self),
	  PGPODataIsASCII(FContext, PGPTrue),
	  PGPOOutputFormat(FContext, kPGPOutputFormat_SMIMEBodyEncryptedData)
	]);
    end
    else begin
      Result := PGPBuildOptionList(FContext, OptionList,
	[
	  PGPOEventHandler(FContext, EventHandler, Self),
	  PGPODataIsASCII(FContext, PGPTrue),
	  PGPOOutputFormat(FContext, kPGPOutputFormat_SMIMEBodySignedData)
	]);
    end;
  end
  else if FDetachedSign then begin
    Result := PGPBuildOptionList(FContext, OptionList,
      [
	PGPOEventHandler(FContext, EventHandler, Self),
	PGPOArmorOutput(FContext, PGPBoolean(FArmor = true)),
	PGPODetachedSig(FContext, PGPOLastOption(FContext), nil),
	PGPOCommentString(FContext, PChar(FComment)),
	PGPOVersionString(FContext, MyVersion)
      ]);
  end
  else begin
    Result := PGPBuildOptionList(FContext, OptionList,
      [
	{PGPOSendNullEvents(FContext, FProgressInterval),}
	PGPOEventHandler(FContext, EventHandler, Self),
	PGPOArmorOutput(FContext, PGPBoolean(FArmor = true)),
	PGPOClearSign(FContext, PGPBoolean(FClear = true)),
	PGPOCompression(FContext, PGPBoolean(FCompress = true)),
	PGPOForYourEyesOnly(FContext, PGPBoolean(FEyesOnly = true)),
	PGPODataIsASCII(FContext, PGPBoolean(FTextMode = true)),
	PGPOCommentString(FContext, PChar(FComment)),
	PGPOVersionString(FContext, MyVersion)
      ]);
    if FMime and (Result = 0) then begin
      Result := PGPAppendOptionList(OptionList,
	[
	  PGPOPGPMIMEEncoding(FContext, PGPTrue, PGPSize(FMimeBodyOffset), Pointer(FMimeSeparator)),
	  PGPOOmitMIMEVersion(fContext, PGPBoolean(FOmitMimeVersion = true))
	]);
    end;
  end;
  if Result <> 0 then Exit;
  if FConventional then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOCipherAlgorithm(FContext, PGPCipherAlgorithm(succ(FConventionalAlgorithm)))
      ]);
    if Result <> 0 then Exit;
    Result := ConvEncPassphraseDialog(FContext, FPassphrase, FPassDlgPrompt, FParentHandle);
    if Result <> 0 then Exit;
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOConventionalEncrypt(FContext, PGPOPassphrase(FContext, FPassphrase), PGPOLastOption(FContext))
      ]);
    if Result <> 0 then Exit;
  end
  else begin
    Result := AddSignOptions(OptionList);
    if Result <> 0 then Exit;
    if FEncrypt and not (FSMime and FSign) then begin
      Result := GetEncryptKeyID(FEncryptKeyID, FSignKeyID);
      if Result <> 0 then Exit;
      Result := GetKeySetByAnyIDs(FContext, FKeySetMain, FEncryptKeyID, FEncryptKeySet);
      if Result <> 0 then Exit;
      if FSMime then begin
	Result := CheckValidX509CertifierSet(FEncryptKeySet);
	if Result <> 0 then Exit;
      end;
      Result := PGPAppendOptionList(OptionList, [PGPOEncryptToKeySet(FContext, FEncryptKeySet)]);
      if Result <> 0 then Exit;
      if GetAllowedCipherAlgorithms(AlgorithmList, AlgCount) = 0 then begin
	PGPAppendOptionList(OptionList, [PGPOPreferredAlgorithms(FContext, AlgorithmList, AlgCount)]);
      end;
    end;
  end;
  // InsertCharset() does this on buffered data for all versions
  // if PGP95 then Result := PGPAppendOptionList(OptionList, [PGPOCharsetString(FContext, PChar(FCharset))]);
end;

function TPGPEncode.ReadInputFile: PGPError;
var
  InputFile	: THandle;
  BytesRead	: DWord;
begin
  FInputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(FInputSize), kPGPMemoryMgrFlags_None);
  if FInputBuffer <> nil then begin
    FInputBuffer[FInputSize] := #0;
    InputFile := CreateFile(PChar(FInputFileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING,
			    FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if InputFile <> INVALID_HANDLE_VALUE then begin
      try
	if not ReadFile(InputFile, FInputBuffer[0], FInputSize, BytesRead, nil) or (BytesRead <> FInputSize) then
	  Result := kPGPError_ReadFailed
	else Result := kPGPError_NoErr;
      finally
	CloseHandle(InputFile);
      end;
    end
    else Result := kPGPError_CantOpenFile;
  end
  else begin
    Result := kPGPError_OutOfMemory;
    Exit;
  end;
end;

function TPGPEncode.SetInputOption(IsFile: Longbool; var OptionList: pPGPOptionList): PGPError;
var
  FileHandle	: THandle;
  FileSizeHigh	: DWord;
  InFileSpec	: pPGPFileSpec;
begin
  Result := 0;
  if IsFile then begin
    FileHandle := FileOpen(FInputFileName, fmOpenRead or fmShareDenyNone);
    if FileHandle <> INVALID_HANDLE_VALUE then begin
      FInputSize := GetFileSize(FileHandle, @FileSizeHigh);
      FileClose(FileHandle);
    end
    else FInputSize := $FFFFFFFF;
    if (not FParanoiaSwitch and (FInputSize = 0)) or (FInputSize = $FFFFFFFF) or (FileSizeHigh <> 0) then begin
      Result := kPGPError_ReadFailed;
      Exit;
    end;
    if not FSMime then begin
      InFileSpec := nil;
      Result := PGPNewFileSpecFromFullPath(FContext, PChar(FInputFileName), InFileSpec);
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList, [PGPOInputFile(FContext, InFileSpec)]);
      finally
	PGPFreeFileSpec(InFileSpec);
      end;
    end
    else Result := ReadInputFile;
    if Result <> 0 then Exit;
  end;
  if not IsFile or FSMime then begin
    Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(FContext, FInputBuffer, FInputSize)]);
  end;
end;

function TPGPEncode.SetOutputOption(var OptionList: pPGPOptionList): PGPError;
var
  FileSpec	: pPGPFileSpec;
begin
  if FFileOutput and (FOutputFileName = '') then begin
    Result := kPGPError_CantOpenFile;
    Exit;
  end;
  if FFileToFile then begin
    FileSpec := nil;
    Result := PGPNewFileSpecFromFullPath(FContext, PChar(FOutputFileName), FileSpec);
    try
      if Result = 0 then Result := PGPAppendOptionList(OptionList, [PGPOOutputFile(FContext, FileSpec)]);
    finally
      PGPFreeFileSpec(FileSpec);
    end;
  end
  else begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
      ]);
  end;
end;

function TPGPEncode.WriteOutputFile: PGPError;
var
  OutputFile	: THandle;
  BytesToWrite	: DWord;
  BytesWritten	: DWord;
begin
  OutputFile := CreateFile(PChar(FOutputFileName), GENERIC_WRITE, 0, nil, CREATE_ALWAYS,
			   FILE_ATTRIBUTE_ARCHIVE or FILE_FLAG_WRITE_THROUGH, 0);
  if OutputFile <> INVALID_HANDLE_VALUE then begin
    try
      BytesToWrite := Length(FOutputBuffer);
      if not WriteFile(OutputFile, Pointer(FOutputBuffer)^, BytesToWrite, BytesWritten, nil)
      or (BytesWritten <> BytesToWrite) then
	Result := kPGPError_WriteFailed
      else Result := kPGPError_NoErr;
    finally
      CloseHandle(OutputFile);
    end;
  end
  else Result := kPGPError_CantOpenFile;
end;

function TPGPEncode.Encode(IsFile: Longbool): Longint;
var
  OptionList	: pPGPOptionList;
begin
  Result := 0;
  FOutputBuffer := '';
  if IsFile then begin
    if not FileExists(FInputBuffer) then
      Result := kPGPError_FileNotFound
    else FInputFileName := FInputBuffer;
    if (SDKVersion >= PGP7XX) and not (FSMime or FMime) then
      FFileToFile := FFileOutput
    else FFileToFile := false;
  end
  else if not FParanoiaSwitch and (FInputSize = 0) then Result := kPGPError_ItemNotFound;
  if Result = 0 then begin
    OptionList := nil;
    FEncryptKeySet := nil;
    Result := InitEncode;
    if Result <> 0 then Exit;
    try
      Result := GetOptionList(OptionList);
      try
	if Result <> 0 then Exit;
	Result := SetInputOption(IsFile, OptionList);
	if Result <> 0 then Exit;
	Result := SetOutputOption(OptionList);
	if Result <> 0 then Exit;
	SetCursor(FWait);
	FInsertCharset := FInsertCharset and FArmor and FTextmode;
	Result := pgpEvents.PGPEncode(FContext, OptionList, PGPOLastOption(FContext));
	if not FFileToFile and (Result = 0) and (FActualOutputSize <> 0) then begin
	  try
	    SetString(FOutputBuffer, nil, FActualOutputSize);
	    Move(FAllocatedOutputBuffer^, Pointer(FOutputBuffer)^, FActualOutputSize);
	  except
	    on EOutOfMemory do Result := kPGPError_OutOfMemory;
	  end;
	end;
      finally
	try
	  Delete(FMimeSeparator, succ(StrLen(Pointer(FMimeSeparator))), MAXINT);
	  PGPFreeOptionList(OptionList);
	  PGPFreeKeySet(FEncryptKeySet);
	finally
	  try
	    if (FSecure <> nil) and (FSign or FDetachedSign) then begin
	      FSecure.DoCachePassphrase(FSigningKey, FPassphrase, FAllocatedOutputBuffer, FActualOutputSize, true);
	    end;
	  finally
	    PGPFreeData(FAllocatedOutputBuffer);
	    PGPFreeData(FPassphrase);
	    FPassphrase := nil;
	  end;
	end;
      end;
    finally
      FinitEncode;
    end;
  end;
end;

function TPGPEncode.KeyEncryptBuffer(const DataBuffer: String; Sign, SMime: Longbool): Longint;
var
  SMimeSigned: String;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetSign(Sign);
  SetEncrypt(true);
  FCompress := true;
  FSMime := SMime;
  if SMime and Sign then FEncrypt := false;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInsertCharset := (FCharset <> '');
  Result := Encode(false);
  if Result = 0 then begin
    if FSMime then begin
      if not FEncrypt then begin
	SMimeSigned := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), false, true);
	FInputBuffer := PChar(SMimeSigned);
	FInputSize := Length(SMimeSigned);
	FEncrypt := true;
	FSign := false;
	Result := Encode(false);
	if Result <> 0 then Exit;
      end;
      FOutputBuffer := SMimeBuild(FOutputBuffer, '', FMimeSeparator, FMimeBodyOffset,
				  HashAlgorithm_Invalid, true, FOmitMimeVersion);
    end
    else if FMime then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    HashAlgorithm_Invalid, true, FOmitMimeVersion);
    end
    else if FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, false)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => string based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncode.KeyEncryptFile(const FileName: String; Sign, SMime: Longbool): Longint;
var
  SMimeSigned: String;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetSign(Sign);
  SetEncrypt(true);
  FCompress := true;
  FSMime := SMime;
  if SMime and Sign then FEncrypt := false;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  FInsertCharset := (FCharset <> '');
  Result := Encode(true);
  if Result = 0 then begin
    if FSMime then begin
      if not FEncrypt then begin
	SMimeSigned := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), false, true);
	FInputBuffer := PChar(SMimeSigned);
	FInputSize := Length(SMimeSigned);
	FEncrypt := true;
	FSign := false;
	Result := Encode(false);
	if Result <> 0 then Exit;
      end;
      FOutputBuffer := SMimeBuild(FOutputBuffer, '', FMimeSeparator, FMimeBodyOffset,
				  HashAlgorithm_Invalid, true, FOmitMimeVersion);
    end
    else if FMime then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    HashAlgorithm_Invalid, true, FOmitMimeVersion);
    end
    else if not FFileToFile and FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, false)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => string based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncode.ConventionalEncryptBuffer(const DataBuffer: String): Longint;
begin
  FCompress := true;
  SetConventional(true);
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  Result := Encode(false);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncode.ConventionalEncryptFile(const FileName: String): Longint;
begin
  FCompress := true;
  SetConventional(true);
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  Result := Encode(true);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncode.ClearSignBuffer(const DataBuffer: String; SMime: Longbool): Longint;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetClear(true);
  FSMime := SMime;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInsertCharset := (FCharset <> '');
  Result := Encode(false);
  if Result = 0 then begin
    if FSMime then begin
      FOutputBuffer := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), false, FOmitMimeVersion);
    end
    else if FMime then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    SignToHashAlgorithm(FSignAlgorithm), false, FOmitMimeVersion);
    end
    else if FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, true)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => string based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncode.ClearSignFile(const FileName: String; SMime: Longbool): Longint;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetClear(true);
  FSMime := SMime;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  FInsertCharset := (FCharset <> '');
  Result := Encode(true);
  if Result = 0 then begin
    if FSMime then begin
      FOutputBuffer := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), false, FOmitMimeVersion);
    end
    else if FMime then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    SignToHashAlgorithm(FSignAlgorithm), false, FOmitMimeVersion);
    end
    else if not FFileToFile and FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, true)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => string based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncode.ArmorBuffer(const DataBuffer: String): Longint;
begin
  FArmor := true;
  FClear := false;
  FCompress := true;
  FDetachedSign := false;
  FEncrypt := false;
  FSign := false;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  Result := Encode(false);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncode.ArmorFile(const FileName: String): Longint;
begin
  FArmor := true;
  FClear := false;
  FCompress := true;
  FDetachedSign := false;
  FEncrypt := false;
  FSign := false;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  Result := Encode(true);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncode.DetachedSignBuffer(const DataBuffer: String): Longint;
begin
  FConventional := false;
  FDetachedSign := true;
  FEncrypt := false;
  FSign := false;
  FInputBuffer := PChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  Result := Encode(false);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncode.DetachedSignFile(const FileName: String): Longint;
begin
  FConventional := false;
  FDetachedSign := true;
  FEncrypt := false;
  FSign := false;
  FInputBuffer := PChar(FileName);
  FInputSize := Length(FileName);
  Result := Encode(true);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncode.GetEncryptKeyID(var EncryptKeyID, SignKeyID: String): Longint;
var
  KeyPropsList	: TKeyPropsList;
  CertPropsList	: TX509CertPropsList;
  HexKeyID	: String;
  {
  RecipientKeys	: Pointer;
  RecipientDB	: pPGPKeyDB;
  }
  SelKeyDialog	: TSelKeyDialog;
begin
  Result := 0;
  with KeyRings do if GroupsFile = '' then GroupsFile := FPreferences.GroupsFile;
  if not FSign then begin
    KeyPropsList := nil;
    try
      if FindKeyProps(SignKeyID, KeyPropsList,
		      spgpKeyPropFlag_KeyID or
		      spgpKeyPropFlag_CanEncrypt,
		      KeyFilterFlag_CanSign,
		      Creation_Ordering) <> 1 then begin
	SignKeyID := FPreferences.DefaultKeyID;
	if (SignKeyID = '') and (KeyPropsList.Count > 0) then SignKeyID := KeyPropsList[0];
      end;
    finally
      KeyPropsList.Free;
    end;
  end;
  if not (FAlwaysShowKeyDlg or FGroupSelected or FMultipleRecipients) then begin
    if FSMime then begin
      CertPropsList := nil;
      try
	if IsHexID(EncryptKeyID) > 0 then
	  Result := FindX509CertProps(E, EncryptKeyID, E, KeyFilterFlag_X509CanEncrypt, false, CertPropsList)
	else Result := FindX509CertProps(EncryptKeyID, E, E, KeyFilterFlag_X509CanEncrypt, false, CertPropsList);
	if Result = 1 then EncryptKeyID := CertPropsList.X509CertProps[0].x509OwnerKeyHexID;
      finally
	CertPropsList.Free;
      end;
    end
    else begin
      Result := GetHexIDByAnyID(FContext, FKeySetMain, EncryptKeyID, HexKeyID);
      if Result = 1 then EncryptKeyID := HexKeyID;
    end;
  end;
  if (Result <> 1) and not FGroupSelected then begin
    KeyPropsList := nil;
    try
      {
      if FMultipleRecipients then begin
	RecipientKeys := nil;
	if PGP7X then begin
	  if FSMime then begin
	    Result := GetX509EncryptCertDB(FContext, FKeySetMain, RecipientDB);
	    if Result < 0 then Exit;
	  end
	  else RecipientDB := PGPPeekKeySetKeyDB(FKeySetMain);
	  try
	    Result := RecipientsDialog(FContext, PGPPeekKeyDBRootKeySet(RecipientDB), PGPTrue, FSMime, SignKeyID,
				       pPGPKeyDB(RecipientKeys), FKeySelDlgPrompt, FParentHandle);
	    if Result = 0 then begin
	      try
		Result := GetKeySetProps(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(RecipientKeys)),
					 KeyPropsList, spgpKeyPropFlag_KeyID,
					 KeyFilterFlag_CanEncrypt, UserID_Ordering);
	      finally
		PGPFreeKeyDB(pPGPKeyDB(RecipientKeys));
	      end;
	    end;
	  finally
	    if FSMime then PGPFreeKeyDB(RecipientDB);
	  end;
	end
	else begin
	  Result := RecipientsDialog(FContext, FKeySetMain, PGPTrue, false, SignKeyID,
				     pPGPKeySet(RecipientKeys), FKeySelDlgPrompt, FParentHandle);
	  if Result = 0 then begin
	    try
	      Result := GetKeySetProps(FContext, pPGPKeySet(RecipientKeys),
				       KeyPropsList, spgpKeyPropFlag_KeyID,
				       KeyFilterFlag_CanEncrypt, UserID_Ordering);
	    finally
	      PGPFreeKeySet(pPGPKeySet(RecipientKeys));
	    end;
	  end;
	end;
	if Result > 0 then EncryptKeyID := KeyPropsList.CommaText;
      end
      else
      }
      begin
	Result := GetKeySetProps(FContext,
				 FKeySetMain,
				 KeyPropsList,
				 spgpKeyPropFlag_IDComplete or
				 spgpKeyPropFlag_Fingerprint or
				 spgpKeyPropFlag_CreationTimeStr or
				 spgpKeyPropFlag_Trust or
				 spgpKeyPropFlag_Validity or
				 spgpKeyPropFlag_HasADK or
				 spgpKeyPropFlag_X509Cert or
				 spgpKeyPropFlag_LegacyKey or
				 spgpKeyPropFlag_IncludeUserIDs or
				 (DWord(not FSMime) * spgpKeyPropFlag_IncludeGroupsList),
				 KeyFilterFlag_CanEncrypt or (ord(FSMime = true) * KeyFilterFlag_X509CanEncrypt),
				 UserID_Ordering);
	if Result < 0 then Exit;
	SelKeyDialog := TSelKeyDialog.CreateDialog(FParentHandle);
	if SelKeyDialog <> nil then with SelKeyDialog do begin
	  try
	    Result := 0;
	    SMime := FSMime;
	    if SMime then
	      Prompt := FCertSelDlgPrompt
	    else Prompt := FKeySelDlgPrompt;
	    ShowDetails := FShowKeyDetails;
	    EditText := FRecipients;
	    GroupText := FGroupText;
	    ServerBtnText := FKeyServerBtn;
	    if SMime then
	      UseAddBtnText := FCertPropsViewerBtn
	    else UseAddBtnText := FKeyUseAddBtn;
	    CancelBtnText := FKeyCancelBtn;
	    OkBtnText := FKeyOkBtn;
	    SelectedKeyText := FSelectedKeysText;
	    SelectAllBtnText := FSelectAllBtn;
	    UnselectBtnText := FUnselectBtn;
	    CertificateText := FCertificateText;
	    CertifiedText := FCertifiedText;
	    KeyTrustsText := KeyTrusts;
	    KeyValiditiesText := KeyValidities;
	    KeyAlgorithmsText := KeyAlgorithms;
	    HashAlgorithmsText := HashAlgorithms;
	    CertValitidiesText := CertValidities;
	    HelpHandler := FHelpHandler;
	    HelpIndex := FKeyDialogHelpIndex;
	    MultiSelect := FMultipleRecipients;
	    case SelectEncryptKey(EncryptKeyID, KeyPropsList, FGroupSelected) of
	      skCancel: Result := kPGPError_UserAbort;
	      skServer: begin
		if (GetKeyFromServerDialog(FKeySelDlgPrompt, EncryptKeyID, FKeyPropsList,
					   spgpKeyPropFlag_IDFlags, FParentHandle) = 0)
		and (FKeyPropsList <> nil) then with FKeyPropsList do begin
		  if Count = 1 then EncryptKeyID := Strings[0];
		  FKeysFromServer := Count;
		end;
	      end;
	      skUseAdd: begin
		if (FKeyPropsList <> nil) then with FKeyPropsList do begin
		  Add(ExtractStr(EncryptKeyID, 1, DefHexIDLen));
		  with pKeyPropsRec(Objects[pred(Count)])^ do begin
		    kHexID := Strings[pred(Count)];
		    kUserID := EncryptKeyID;
		    EncryptKeyID := kHexID;
		  end;
		end;
	      end;
	    end;
	  finally
	    SelKeyDialog.Free;
	  end;
	end
	else Result := kPGPError_PublicKeyNotFound;
      end;
    finally
      KeyPropsList.Free;
    end;
  end;
  if Result >= 0 then
    Result := 0
  else Exit;
  if (EncryptKeyID <> '') and FEncryptToSelf and (pos(SignKeyID, EncryptKeyID) = 0) then begin
    EncryptKeyID := EncryptKeyID + ',' + SignKeyID;
  end;
end;


end.

