{$J+,Z4}
unit SecureFuncs;
{$IFDEF CONDITIONALEXPRESSIONS} {$WARN SYMBOL_PLATFORM OFF} {$ENDIF}

{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2001-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

// for details about the Isaac algorithm see <http://burtleburtle.net/bob>
// for details about the wipe patterns see <http://www.cs.auckland.ac.nz/~pgut001/pubs/secure_del.html>

interface

uses
  UTF8,
  Isaac,
  Windows,
  SysUtils,
  pgpBase,
  pgpErrors,
  pgpPubTypes,
  pgpUtilities,
  pgpOptionList,
  pgpMemoryMgr,
  pgpEvents,
  PrefFuncs,
  KeyFuncs,
  pgpKeys,
  pgpSC;

type
  TPhraseData = Record
    Size: Longint;
    Cache: PChar;
    Key: PChar;
  end;

  TSecure = Class
  private
    FIsaac: TIsaac;
    FWait: HCursor;
    FArrow: HCursor;
    FContext: pPGPContext;
    FCacheEnabled: Longbool;
    FCacheInterval: Longint;
    FWipeCount: PGPUInt32;
    pdDecrypt: TPhraseData;
    pdSigning: TPhraseData;
    function SecAlloc(Size: Cardinal): PChar;
    function SecFree(var Alloc: PChar): Longint;
    function EncodePhrase(const Passphrase: PChar; var pdData: TPhraseData): PGPError;
    function DecodePhrase(var Passphrase: PChar; const pdData: TPhraseData): PGPError;
  public
    constructor Create;
    destructor Destroy; override;
    function  GetCachedPassphrase(Key: pPGPKey; var Passphrase: PChar; Signing: Longbool): Longbool;
    function  DoCachePassphrase(Key: pPGPKey; var Passphrase: PChar;
    				const Data: PChar; DataSize: PGPSize; Signing: Longbool): Longbool;
    procedure EnableCaching(TimeOut: Longint);
    function  CacheEmpty: Longbool;
    function  PGP8XXCacheEmpty: Longbool;
    function  PGP8XXPurgeCache: Longbool;
    function  DecryptCacheEmpty: Longbool;
    function  SigningCacheEmpty: Longbool;
    procedure PurgeSigningPhrase;
    procedure PurgeDecryptPhrase;
    procedure PurgeCachedPhrases;
    procedure WipeFile(const FileName: String);
    property Isaac: TIsaac
      read FIsaac
      write FIsaac;
    property CacheInterval: Longint
      read FCacheInterval
      write FCacheInterval;
    property WipeCount: PGPUInt32
      read FWipeCount
      write FWipeCount;
  end;

function SecureAnsiCharToUtf8(const pAnsi, pUtf8: PChar; MaxBytes: Cardinal): Cardinal;
function SecureUtf8ToAnsiChar(const pUtf8, pAnsi: PChar; MaxBytes: Cardinal): Cardinal;


implementation


function TSecure.SecAlloc(Size: Cardinal): PChar;
begin
  Result := PGPNewSecureData(PGPGetDefaultMemoryMgr, Size, kPGPMemoryMgrFlags_Clear);
end;

function TSecure.SecFree(var Alloc: PChar): Longint;
begin
  Result := kPGPError_BadMemAddress;
  if Alloc <> nil then begin
    Result := PGPFreeData(Alloc);
    if Result = kPGPError_NoErr then Alloc := nil;
  end;
end;

function TSecure.EncodePhrase(const Passphrase: PChar; var pdData: TPhraseData): PGPError;
var
  OptionList: pPGPOptionList;
  OutputSize: PGPSize;
begin
  OptionList := nil;
  try
    Result := PGPBuildOptionList(FContext, OptionList,
      [
       PGPOCompression(FContext, PGPTrue),
       PGPOCipherAlgorithm(FContext, kPGPCipherAlgorithm_IDEA),
       PGPOInputBuffer(FContext, Passphrase, StrLen(Passphrase)),
       PGPOOutputBuffer(FContext, pdData.Cache, 512, OutputSize),
       PGPOConventionalEncrypt(FContext,
			       PGPOPassphraseBuffer(FContext, pdData.Key, 32), PGPOLastOption(FContext)),
			       PGPOLastOption(FContext)
      ]);
    if Result <> 0 then Exit;
    if pgpEvents.PGPEncode(FContext, OptionList, PGPOLastOption(FContext)) <> 0 then with pdData do begin
      SecFree(Cache);
      SecFree(Key);
      Size := 0;
    end
    else pdData.Size := OutputSize;
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function TSecure.DecodePhrase(var Passphrase: PChar; const pdData: TPhraseData): PGPError;
var
  OptionList: pPGPOptionList;
  OutputSize: PGPSize;
begin
  if pdData.Size <> 0 then begin
    OptionList := nil;
    try
      Result := PGPBuildOptionList(FContext, OptionList,
	[
	 PGPOInputBuffer(FContext, pdData.Cache, pdData.Size),
	 PGPOOutputBuffer(FContext, Passphrase, 1024, OutputSize),
	 PGPOPassphraseBuffer(FContext, pdData.Key, 32)
	]);
      if Result <> 0 then Exit;
      Result := pgpEvents.PGPDecode(FContext, OptionList, PGPOLastOption(FContext));
    finally
      PGPFreeOptionList(OptionList);
    end;
  end
  else Result := kPGPError_MissingPassphrase;
end;

function TSecure.GetCachedPassphrase(Key: pPGPKey; var Passphrase: PChar; Signing: Longbool): Longbool;
begin
  Result := false;
  if FCacheEnabled then begin
    Passphrase := SecAlloc(1024);
    if Passphrase <> nil then begin
      try
	if Signing then begin
	  if SDKVersion >= PGP8XX then
	    Result := (PGPPassphraseIsValid(Key, PGPOLastOption(FContext), PGPOLastOption(FContext)) = PGPTrue)
	  else begin
	    Result := (DecodePhrase(Passphrase, pdSigning) = 0) and
		      (PGPPassphraseIsValid(Key, PGPOPassphrase(FContext, Passphrase), PGPOLastOption(FContext)) = PGPTrue);
	  end;
	end
	else begin
	  if SDKVersion >= PGP8XX then begin
	    SecFree(Passphrase);
	    Result := true;
	  end
	  else Result := (DecodePhrase(Passphrase, pdDecrypt) = 0);
	end;
      finally
	if not Result then SecFree(Passphrase);
      end;
    end;
  end;
end;

function TSecure.DoCachePassphrase(Key: pPGPKey; var Passphrase: PChar;
				   const Data: PChar; DataSize: PGPSize; Signing: Longbool): Longbool;
var
  OptionList: pPGPOptionList;
  ShareCache: Longbool;
  RandomStr: String;
  Position: Longint;
begin
  Result := false;
  if FCacheEnabled then begin
    if SDKVersion >= PGP8XX then begin
      OptionList := nil;
      GetPrefShareCache(ShareCache);
      if PGPBuildOptionList(FContext, OptionList,
	[
	 PGPOPassphrase(FContext, Passphrase),
	 PGPOCachePassphrase(FContext, kPGPMaxTimeInterval, PGPBoolean(ShareCache = true))
	]
      ) = 0 then begin
	try
	  Result := (PGPPassphraseIsValid(Key, OptionList, PGPOLastOption(FContext)) = PGPTrue) and
		    (PGPCacheKeyDB(PGPPeekKeyDBObjKeyDB(Key), kPGPMaxTimeInterval) = 0);
	finally
	  SecFree(Passphrase);
	  PGPFreeOptionList(OptionList);
	end;
      end;
    end
    else if (Passphrase <> nil) and (Data <> nil) and (DataSize <> 0) then begin
      try
	RandomStr := IntToStr(abs(FIsaac.Val));
	if Signing then with pdSigning do begin
	  PurgeSigningPhrase;
	  Cache := SecAlloc(512);
	  Key := SecAlloc(32);
	  if (Cache <> nil) and (Key <> nil) then begin
	    for Position := 0 to 31 do Key[Position] := Data[abs(DWord(FIsaac.Val) mod DataSize)];
	    for Position := 1 to Length(RandomStr) do byte(Key[abs(FIsaac.Val mod 32)]) := StrToInt(RandomStr[Position]);
	    Result := (EncodePhrase(Passphrase, pdSigning) = 0);
	  end
	  else PurgeSigningPhrase;
	end
	else with pdDecrypt do begin
	  PurgeDecryptPhrase;
	  Cache := SecAlloc(512);
	  Key := SecAlloc(32);
	  if (Cache <> nil) and (Key <> nil) then begin
	    for Position := 0 to 31 do Key[Position] := Data[abs(DWord(FIsaac.Val) mod DataSize)];
	    for Position := 1 to Length(RandomStr) do byte(Key[abs(FIsaac.Val mod 32)]) := StrToInt(RandomStr[Position]);
	    Result := (EncodePhrase(Passphrase, pdDecrypt) = 0);
	  end
	  else PurgeDecryptPhrase;
	end;
      finally
	SecFree(Passphrase);
      end;
    end;
  end;
end;

procedure TSecure.EnableCaching(TimeOut: Longint);
begin
  FCacheInterval := TimeOut;
  FCacheEnabled := (FCacheInterval <> -1);
  if not FCacheEnabled then PurgeCachedPhrases;
end;

function TSecure.CacheEmpty: Longbool;
begin
  if SDKVersion >= PGP8XX then
    Result := PGP8XXCacheEmpty
  else Result := DecryptCacheEmpty and SigningCacheEmpty;
end;

function TSecure.PGP8XXCacheEmpty: Longbool;
var
  Global, Local, Other, Persistent: PGPUInt32;
begin
  Result := (PGPCountCachedPassphrases(FContext, Global, Local, Other, Persistent) = 0) and
	    (Global + Local + Other + Persistent = 0);
end;

function TSecure.PGP8XXPurgeCache: Longbool;
begin
  Result := (PGPPurgePassphraseCache(FContext) = 0) and (PGPPurgeKeyDBCache(FContext) = 0);
end;

function TSecure.DecryptCacheEmpty: Longbool;
begin
  Result := (pdDecrypt.Size = 0);
end;

function TSecure.SigningCacheEmpty: Longbool;
begin
  Result := (pdSigning.Size = 0);
end;

procedure TSecure.PurgeDecryptPhrase;
begin
  with pdDecrypt do begin
    SecFree(Cache);
    SecFree(Key);
    Size := 0;
  end;
end;

procedure TSecure.PurgeSigningPhrase;
begin
  with pdSigning do begin
    SecFree(Cache);
    SecFree(Key);
    Size := 0;
  end;
end;

procedure TSecure.PurgeCachedPhrases;
begin
  if SDKVersion >= PGP8XX then PGP8XXPurgeCache;
  PurgeDecryptPhrase;
  PurgeSigningPhrase;
end;

procedure TSecure.WipeFile(const FileName: String);
type
  TPattern = Array[0..2] of Word;
const
  aPatterns: Array[0..21] of TPattern = (
    ($55, $55, $55),
    ($AA, $AA, $AA),
    ($92, $49, $24),
    ($49, $24, $92),
    ($24, $92, $49),
    ($00, $00, $00),
    ($FF, $FF, $FF),
    ($11, $11, $11),
    ($22, $22, $22),
    ($33, $33, $33),
    ($44, $44, $44),
    ($66, $66, $66),
    ($77, $77, $77),
    ($88, $88, $88),
    ($99, $99, $99),
    ($BB, $BB, $BB),
    ($CC, $CC, $CC),
    ($DD, $DD, $DD),
    ($EE, $EE, $EE),
    ($6D, $B6, $DB),
    ($B6, $DB, $6D),
    ($DB, $6D, $B6)
  );
var
  // PGP 7.X
  sFileName: String;
  // PGP 6.5.X
  hFile: THandle;
  aData: Array[0..511] of TPattern;
  iCount, iIndex, iSizeLo, iSizeHi, iLength, iResult: Longint;
  dwClusterSecs, dwSecBytes, dwFreeClusts, dwTotalClusts: DWord;
begin
  if FileExists(FileName) then begin
    if FWipeCount = 0 then
      DeleteFile(FileName)
    else begin
      if SDKVersion >= PGP7XX then begin
	SetCursor(FWait);
	try
	  SetString(sFileName, PChar(FileName), Length(FileName));
	  WipeFileList(0, FContext, CmdLineToFileList(PChar(sFileName)),
		       WIPE_DONOTWARN or WIPE_DONOTPROGRESS or WIPE_DONOTERROR, FWipeCount);
	finally
	  SetCursor(FArrow);
	end;
      end
      else begin
	try
	  FileSetAttr(FileName, 0);
	  GetDiskFreeSpace(PChar(ExtractFileDrive(FileName) + '\'), dwClusterSecs, dwSecBytes, dwFreeClusts, dwTotalClusts);
	  hFile := CreateFile(PChar(FileName), GENERIC_READ or GENERIC_WRITE, 0, nil, OPEN_EXISTING,
			      FILE_ATTRIBUTE_NORMAL or FILE_FLAG_NO_BUFFERING or FILE_FLAG_DELETE_ON_CLOSE, 0);
	  if hFile <> INVALID_HANDLE_VALUE then begin
	    SetCursor(FWait);
	    try
	      iSizeLo := GetFileSize(hFile, @iSizeHi);
	      if dwSecBytes > SizeOf(aData) then dwSecBytes := SizeOf(aData);
	      if not ((iSizeLo = -1) and (GetLastError <> NO_ERROR)) then begin
		iSizeHi := dwSecBytes div 6;
		FillChar(aData, SizeOf(aData), 0);
		for iCount := 0 to pred(FWipeCount) do begin
		  for iIndex := 0 to iSizeHi do aData[iIndex] := aPatterns[abs(FIsaac.Val mod (succ(High(aPatterns))))];
		  FileSeek(hFile, 0, 0);
		  iLength := iSizeLo;
		  iResult := 0;
		  while (iResult <> -1) and (iLength > 0) do begin
		    iResult := FileWrite(hFile, aData, dwSecBytes);
		    iLength := iLength - iResult;
		  end;
		  FlushFileBuffers(hFile);
		end;
	      end;
	    finally
	      SetCursor(FArrow);
	      FileClose(hFile);
	    end;
	  end;
	except
	end;
	DeleteFile(FileName);
      end;
    end;
  end;
end;

constructor TSecure.Create;
var
  stLocal: TSystemTime;
  ptCursor: TPoint;
begin
  inherited;
  FWipeCount := 3;
  GetLocalTime(stLocal);
  GetCursorPos(ptCursor);
  GetPrefCacheTime(FCacheInterval);
  FWait := LoadCursor(0, IDC_WAIT);
  FArrow := LoadCursor(0, IDC_ARROW);
  PGPNewContext(kPGPsdkAPIVersion, FContext);
  with stLocal, ptCursor do begin
    FIsaac := TIsaac.Create;
    FIsaac.Seed([MakeLong(wMonth, wSecond), MakeLong(wHour, wDay),
		 MakeLong(wMilliseconds, wYear), MakeLong(wDayOfWeek, wMinute),
		 MakeLong(X, Y), GetMessageTime, GetMessagePos, GetTickCount], true);
  end;
end;

destructor TSecure.Destroy;
begin
  PurgeCachedPhrases;
  PGPFreeContext(FContext);
  FIsaac.Free;
  inherited;
end;


function SecureAnsiCharToUtf8(const pAnsi, pUtf8: PChar; MaxBytes: Cardinal): Cardinal;
var
  Wide: PWideChar;
  WideLen: Cardinal;
begin
  Result := 0;
  if (pAnsi <> nil) and (pUtf8 <> nil) and (MaxBytes > 0) then begin
    WideLen := succ(MaxBytes div Utf8Factor);
    Wide := PGPNewSecureData(PGPGetDefaultMemoryMgr, WideLen shl 1, kPGPMemoryMgrFlags_Clear);
    try
      WideLen := MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED or MB_ERR_INVALID_CHARS, pAnsi, -1, Wide, WideLen);
      if WideLen > 0 then Result := WideCharToMultiByte(CP_UTF8, 0, Wide, -1, pUtf8, MaxBytes, nil, nil);
    finally
      PGPFreeData(Wide);
    end;
  end;
end;

function SecureUtf8ToAnsiChar(const pUtf8, pAnsi: PChar; MaxBytes: Cardinal): Cardinal;
var
  Wide: PWideChar;
  WideLen: Cardinal;
begin
  Result := 0;
  if (pUtf8 <> nil) and (pAnsi <> nil) and (MaxBytes > 0) then begin
    WideLen := succ(MaxBytes);
    Wide := PGPNewSecureData(PGPGetDefaultMemoryMgr, WideLen shl 1, kPGPMemoryMgrFlags_Clear);
    try
      WideLen := MultiByteToWideChar(CP_UTF8, 0, pUtf8, -1, Wide, WideLen);
      if WideLen > 0 then Result := WideCharToMultiByte(CP_ACP, 0, Wide, -1, pAnsi, MaxBytes, nil, nil);
    finally
      PGPFreeData(Wide);
    end;
  end;
end;


end.

