{$J+,Z4}
unit X509Types;
{$IFDEF CONDITIONALEXPRESSIONS} {$DEFINE FIXED} {$ENDIF} // Delphi 6 and later

{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2004-2023 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  Windows,
  Classes,
  KeyPropTypes,
  pgpOptionList;

const // from pgpX509Keys.c
  MAX_509_CERT_SIZE			= $10000;
  // returned by VerifyX509CertificateChain
  k509Error_NeededCertNotAvailable	= -10;
  k509Error_SelfSignedCert		= -20;
  k509Error_InvalidCertificateSignature	= -30;
  k509Error_InvalidCertificateFormat	= -40;
  // key usage flags (RFC 3280)
  x509KeyUsage_digitalSignature	= $001; // used for entity authentication and data origin authentication with integrity
  x509KeyUsage_nonRepudiation	= $002; // used to verify digital signatures used to provide a non-repudiation service
  x509KeyUsage_keyEncipherment	= $004; // used for key transport / used for key management
  x509KeyUsage_dataEncipherment	= $008; // used for enciphering user data
  x509KeyUsage_keyAgreement	= $010; // used for key agreement
  x509KeyUsage_keyCertSign	= $020; // used for verifying a signature on certificates
  x509KeyUsage_cRLSign		= $040; // used for verifying a signature on revocation information
  x509KeyUsage_encipherOnly	= $080; // used only for enciphering data while performing key agreement
  x509KeyUsage_decipherOnly	= $100; // used only for deciphering data while performing key agreement
  // certificate delimiters
  X509Cert_Beg = '-----BEGIN CERTIFICATE-----';
  X509Cert_End = '-----END CERTIFICATE-----';

type
  // Verisign's EmbedEmail excluded
  TX509CertAttrArray = Array[kPGPAVAttributeFirstPointer..kPGPAVAttributeLastPointer] of TPGPAttributeValue;
  // Verisign attributes excluded
  TX509CertAttributes = Record
    CommonName				: String;
    Email				: String;
    OrganizationName			: String;
    OrganizationalUnitName		: String;
    SurName				: String;
    SerialNumber			: String;
    Country				: String;
    Locality				: String;
    State				: String;
    StreetAddress			: String;
    Title				: String;
    Description				: String;
    PostalCode				: String;
    POBOX				: String;
    PhysicalDeliveryOfficeName		: String;
    TelephoneNumber			: String;
    X121Address				: String;
    ISDN				: String;
    DestinationIndicator		: String;
    Name				: String;
    GivenName				: String;
    Initials				: String;
    HouseIdentifier			: String;
    DirectoryManagementDomain		: String;
    DomainComponent			: String;
    UnstructuredName			: String;
    UnstructuredAddress			: String;
    RFC822Name				: String;
    DNSName				: String;
    AnotherName				: String;
    IPAddress				: String;
  end;

  TX509CertData = Array[0..pred(MAX_509_CERT_SIZE)] of Char;

  TX509KeyUsage = (
    KeyUsage_digitalSignature,
    KeyUsage_nonRepudiation,
    KeyUsage_keyEncipherment,
    KeyUsage_dataEncipherment,
    KeyUsage_keyAgreement,
    KeyUsage_keyCertSign,
    KeyUsage_cRLSign,
    KeyUsage_encipherOnly,
    KeyUsage_decipherOnly
  );
  TX509KeyUsages = Set of TX509KeyUsage;

  TX509CertProp = (
    CertProp_CertSerialNumber,
    CertProp_IssuerKeyAlgorithm,
    CertProp_IssuerHashAlgorithm,
    CertProp_IssuerLongName,
    CertProp_IssuerSerialNumber,
    CertProp_OwnerHexID,
    CertProp_OwnerUserID,
    CertProp_OwnerLongName,
    CertProp_OwnerSerialNumber,
    CertProp_CreaTimeStr,
    CertProp_ExpTimeStr,
    CertProp_CreaTimeNum,
    CertProp_ExpTimeNum,
    CertProp_KeyUsages,
    CertProp_CanEncryptData,
    CertProp_CanSignData,
    CertProp_CanSignCerts,
    CertProp_CanSignCRLs,
    CertProp_Verified,
    CertProp_Corrupt,
    CertProp_Revoked,
    CertProp_Expired,
    CertProp_Root
  );
  TX509CertProps = Set of TX509CertProp;

  pX509CertPropsRec = ^TX509CertPropsRec;
  TX509CertPropsRec = Record
    x509CertificateID			: String;
    x509IssuerKeyAlgorithm		: TKeyAlgorithm;
    x509IssuerHashAlgorithm		: THashAlgorithm;
    x509IssuerLongName			: String;
    x509IssuerSerialNumber		: String;
    x509OwnerKeyHexID			: String;
    x509OwnerCertUserID			: String;
    x509OwnerLongName			: String;
    x509OwnerSerialNumber		: String;
    x509CreaTimeStr			: String;
    x509ExpTimeStr			: String;
    x509CreaTimeNum			: DWord;
    x509ExpTimeNum			: DWord;
    x509KeyUsages			: TX509KeyUsages;
    x509CanEncryptData			: Longbool;
    x509CanSignData			: Longbool;
    x509CanSignCerts			: Longbool;
    x509CanSignCRLs			: Longbool;
    x509IsVerified			: Longbool;
    x509IsCorrupt			: Longbool;
    x509IsRevoked			: Longbool;
    x509IsExpired			: Longbool;
    x509IsRoot				: Longbool;
  end;

  pX509CertPropsList = ^TX509CertPropsList;
  TX509CertPropsList = class(TStringList)
  private
    FDuplicates: Longbool;
    function	AllocX509CertPropsRec: pX509CertPropsRec;
    function	GetX509CertProps(Index: Longint): TX509CertPropsRec;
    function	FreeX509CertPropsRec(Index: Longint): pX509CertPropsRec;
  public
    constructor	Create;
    destructor	Destroy; override;
    procedure	Clear; override;
    procedure 	Changed; override;
    procedure 	Changing; override;
    procedure	Delete(Index: Integer); override;
    procedure	Move(CurIndex, NewIndex: Integer); override;
    function	Add(const S: String): Integer; override;
    function	AddObject(const S: String; AObject: TObject): Integer; override;
    procedure	Insert(Index: Integer; const S: String); override;
    {$IFDEF FIXED}
    procedure	InsertObject(Index: Integer; const S: String; AObject: TObject); override;
    {$ELSE}
    procedure	InsertObject(Index: Integer; const S: String; AObject: TObject);
    {$ENDIF}
    // returns true if requested record item exists
    function	GetCertPropsRec(var CertPropsRec: TX509CertPropsRec; Index: Integer): Longbool;
    // access indexed CertProps record like Strings[] or Objects[]
    property	X509CertProps[Index: Longint]: TX509CertPropsRec read GetX509CertProps;
    // just a dummy to prevent using the inherited property
    property	Duplicates: Longbool read FDuplicates;
  end;

function GetKeyUsageSet(KeyUsages: Word): TX509KeyUsages;
function GetKeyUsageFlags(KeyUsages: TX509KeyUsages): Word;


implementation

function GetKeyUsageSet(KeyUsages: Word): TX509KeyUsages;
begin
  Result := [];
  if (KeyUsages and x509KeyUsage_digitalSignature) <> 0 then Include(Result, KeyUsage_digitalSignature);
  if (KeyUsages and x509KeyUsage_nonRepudiation) <> 0 then Include(Result, KeyUsage_nonRepudiation);
  if (KeyUsages and x509KeyUsage_keyEncipherment) <> 0 then Include(Result, KeyUsage_keyEncipherment);
  if (KeyUsages and x509KeyUsage_dataEncipherment) <> 0 then Include(Result, KeyUsage_dataEncipherment);
  if (KeyUsages and x509KeyUsage_keyAgreement) <> 0 then Include(Result, KeyUsage_keyAgreement);
  if (KeyUsages and x509KeyUsage_keyCertSign) <> 0 then Include(Result, KeyUsage_keyCertSign);
  if (KeyUsages and x509KeyUsage_cRLSign) <> 0 then Include(Result, KeyUsage_cRLSign);
  if (KeyUsages and x509KeyUsage_encipherOnly) <> 0 then Include(Result, KeyUsage_encipherOnly);
  if (KeyUsages and x509KeyUsage_decipherOnly) <> 0 then Include(Result, KeyUsage_decipherOnly);
end;

function GetKeyUsageFlags(KeyUsages: TX509KeyUsages): Word;
begin
  Result := ord(KeyUsage_digitalSignature in KeyUsages) * x509KeyUsage_digitalSignature or
	    ord(KeyUsage_nonRepudiation in KeyUsages) * x509KeyUsage_nonRepudiation or
	    ord(KeyUsage_keyEncipherment in KeyUsages) * x509KeyUsage_keyEncipherment or
	    ord(KeyUsage_dataEncipherment in KeyUsages) * x509KeyUsage_dataEncipherment or
	    ord(KeyUsage_keyAgreement in KeyUsages) * x509KeyUsage_keyAgreement or
	    ord(KeyUsage_keyCertSign in KeyUsages) * x509KeyUsage_keyCertSign or
	    ord(KeyUsage_cRLSign in KeyUsages) * x509KeyUsage_cRLSign or
	    ord(KeyUsage_encipherOnly in KeyUsages) * x509KeyUsage_encipherOnly or
	    ord(KeyUsage_decipherOnly in KeyUsages) * x509KeyUsage_decipherOnly;
end;

function TX509CertPropsList.AllocX509CertPropsRec: pX509CertPropsRec;
begin
  New(Result);
  if Result <> nil then FillChar(Result^, SizeOf(TX509CertPropsRec), 0);
end;

function TX509CertPropsList.FreeX509CertPropsRec(Index: Longint): pX509CertPropsRec;
begin
  Result := pX509CertPropsRec(Objects[Index]);
  if Result <> nil then Result := nil;
end;

function TX509CertPropsList.GetX509CertProps(Index: Longint): TX509CertPropsRec;
begin
  Result := pX509CertPropsRec(Objects[Index])^;
end;

constructor TX509CertPropsList.Create;
begin
  inherited Create;
  BeginUpdate;
  inherited Duplicates := dupAccept;
end;

destructor TX509CertPropsList.Destroy;
var
  Index: Integer;
begin
  try
    for Index := 0 to pred(Count) do FreeX509CertPropsRec(Index);
  finally
    inherited Destroy;
  end;
end;

procedure TX509CertPropsList.Clear;
var
  Index: Integer;
begin
  try
    for Index := 0 to pred(Count) do FreeX509CertPropsRec(Index);
  finally
    inherited Clear;
  end;
end;

procedure TX509CertPropsList.Changed;
begin
  // save clock cycles
end;

procedure TX509CertPropsList.Changing;
begin
  // save clock cycles
end;

procedure TX509CertPropsList.Delete(Index: Integer);
begin
  FreeX509CertPropsRec(Index);
  inherited Delete(Index);
end;

procedure TX509CertPropsList.Move(CurIndex, NewIndex: Integer);
var
  TempStr: String;
  TempObj: TObject;
begin
  if CurIndex <> NewIndex then begin
    TempStr := Strings[CurIndex];
    TempObj := Objects[CurIndex];
    inherited Delete(CurIndex);
    inherited InsertObject(NewIndex, TempStr, TempObj);
  end;
end;

function TX509CertPropsList.Add(const S: String): Integer;
begin
  Result := inherited Add(S);
  if Objects[Result] = nil then Objects[Result] := TObject(AllocX509CertPropsRec);
end;

function TX509CertPropsList.AddObject(const S: String; AObject: TObject): Integer;
var
  CurrentCount: Integer;
begin
  CurrentCount := Count;
  {$IFNDEF FIXED}
  Result := inherited Add(S);
  {$ELSE}
  Result := inherited AddObject(S, TObject(AllocX509CertPropsRec));
  {$ENDIF}
  if Count = CurrentCount then Exit;
  if AObject <> nil then pX509CertPropsRec(Objects[Result])^ := pX509CertPropsRec(AObject)^;
end;

procedure TX509CertPropsList.Insert(Index: Integer; const S: String);
begin
  InsertObject(Index, S, nil);
end;

procedure TX509CertPropsList.InsertObject(Index: Integer; const S: String; AObject: TObject);
begin
  {$IFNDEF FIXED}
  inherited Insert(Index, S);
  Objects[Index] := TObject(AllocX509CertPropsRec);
  {$ELSE}
  inherited InsertObject(Index, S, TObject(AllocX509CertPropsRec));
  {$ENDIF}
  if AObject <> nil then pX509CertPropsRec(Objects[Index])^ := pX509CertPropsRec(AObject)^;
end;

function TX509CertPropsList.GetCertPropsRec(var CertPropsRec: TX509CertPropsRec; Index: Integer): Longbool;
begin
  Result := false;
  if (Index < Count) and (Objects[Index] <> nil) then begin
    CertPropsRec := pX509CertPropsRec(Objects[Index])^;
    Result := true;
  end;
end;


end.

