{$J+,Z4}
unit HtsTer;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2010-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, RichEdit, RichEd20, Graphics, GdiPlus, Classes,
  SysUtils, StrUtil, FileUtils, WindowUtils, PMOptions;


// TImageList --------------------------------------------------------------------------------------

type
  TImageList = class(TStringList)
  private
    function GetImageID(Index: Integer): Integer;
    procedure SetImageID(Index: Integer; Value: Integer);
    function GetImageName(Index: Integer): String;
    procedure SetImageName(Index: Integer; const Str: String);
  public
    function Add(const Str: String): Integer; override;
    function AddObject(const Str: String; AObject: TObject): Integer; override;
    property ImageIDs[Index: Integer]: Integer read GetImageID write SetImageID;
    property ImageNames[Index: Integer]: String read GetImageName write SetImageName;
  end;


// THtsTerConverter --------------------------------------------------------------------------------

type
  THtsTerConverter = Class
  private
    FDefaultBmp: String;
    FHostWindow: THandle;
    FHostWndProc: TFarProc;
    FImageList: TImageList;
    FPropName: String;
    FTerVersion: String;
    FTerWindow: THandle;
    function HostWindowCreate(MaxWidth: Integer): Longbool;
    procedure HostWindowDestroy;
    function TerWindowCreate(ParentHandle: THandle): Longbool;
    procedure TerWindowResize(MaxWidth: Integer);
    procedure TerWindowClose;
  public
    constructor Create(const ModulePath: String; CheckVersions: Longbool; MaxWidth: Integer);
    destructor Destroy; override;
    function GetRtf(var TextHeight: Integer): String;
    function GetHtml(var TextHeight: Integer): String;
    function GetTextA(var TextHeight: Integer): String;
    function GetTextW(var TextHeight: Integer): String;
    function SetPictureLink(Index: Integer): Longbool;
    function HtmlToRtf(const Html: String; MaxWidth: Integer; var TextHeight: Integer): String;
    function RtfToHtml(const Rtf: String; MaxWidth: Integer; var TextHeight: Integer): String;
    property ImageList: TImageList read FImageList;
  end;


function CreateHtmlToRtfConverter(const Html: String; MaxWidth: Integer; var Rtf: String): THtsTerConverter;
function GetRtfWithPicturesAndFree(HtsTerConverter: THtsTerConverter; var TextHeight: Integer): String;
function HtmlToRtf(const Html: String; MaxWidth: Integer; var TextHeight: Integer): String;
function RtfToHtml(const Rtf: String; MaxWidth: Integer; var TextHeight: Integer): String;


{$I HtsTerConsts.inc}


implementation


function CreateHtmlToRtfConverter(const Html: String; MaxWidth: Integer; var Rtf: String): THtsTerConverter;
var
  iTextHeight: Integer;
begin
  iTextHeight := 0;
  Result := THtsTerConverter.Create(GetModuleFilePath(0), true, MaxWidth);
  if Result <> nil then Rtf := Result.HtmlToRtf(Html, MaxWidth, iTextHeight);
end;

function GetRtfWithPicturesAndFree(HtsTerConverter: THtsTerConverter; var TextHeight: Integer): String;
var
  iIndex: Integer;
begin
  try
    if HtsTerConverter <> nil then with HtsTerConverter do begin
      with ImageList do if Count > 0 then for iIndex := 0 to pred(Count) do SetPictureLink(iIndex);
      Result := GetRtf(TextHeight);
    end;
  finally
    HtsTerConverter.Free;
  end;
end;


// Converter functions -----------------------------------------------------------------------------

function HtmlToRtf(const Html: String; MaxWidth: Integer; var TextHeight: Integer): String;
var
  htc: THtsTerConverter;
begin
  htc := THtsTerConverter.Create(GetModuleFilePath(0), true, MaxWidth);
  if htc <> nil then begin
    try
      Result := htc.HtmlToRtf(Html, MaxWidth, TextHeight);
    finally
      htc.Free;
    end;
  end;
end;

function RtfToHtml(const Rtf: String; MaxWidth: Integer; var TextHeight: Integer): String;
var
  htc: THtsTerConverter;
begin
  htc := THtsTerConverter.Create(GetModuleFilePath(0), true, MaxWidth);
  if htc <> nil then begin
    try
      Result := htc.RtfToHtml(Rtf, MaxWidth, TextHeight);
    finally
      htc.Free;
    end;
  end;
end;


// TImageList --------------------------------------------------------------------------------------

function TImageList.GetImageID(Index: Integer): Integer;
begin
  if (Index > -1) and (Index < Count) then begin
    Result := Integer(Objects[Index]);
    Exit;
  end;
  Result := -1;
end;

procedure TImageList.SetImageID(Index: Integer; Value: Integer);
begin
  if (Index > -1) and (Index < Count) then begin
    Objects[Index] := TObject(Value);
  end;
end;

procedure TImageList.SetImageName(Index: Integer; const Str: String);
begin
  if (Index > -1) and (Index < Count) then begin
    Strings[Index] := Str;
  end;
end;

function TImageList.GetImageName(Index: Integer): String;
begin
  if (Index > -1) and (Index < Count) then begin
    Result := Strings[Index];
    Exit;
  end;
  Result := E;
end;

function TImageList.Add(const Str: String): Integer;
begin
  Result := inherited Add(Str);
end;

function TImageList.AddObject(const Str: String; AObject: TObject): Integer;
begin
  Result := inherited AddObject(Str, AObject);
end;


// HtsTerTools -------------------------------------------------------------------------------------

const

  HTML_FILE = 0;
  HTML_BUF  = 1;

  MAX_WIDTH = 2000;

  HTS_LIB = 'HTS??.DLL';
  HTS_32  = 'HTS32.DLL';

  TER_LIB = 'TER??.DLL';
  TER_32  = 'TER32.DLL';

  DEFAULT_BMP   = 'DEFAULT.BMP';

  TER_VER_CLASS = 'TER%sCLASS';

  TER_XX_CLASS  = 'TER??CLASS';

  TER_DH_CLASS  = 'TERDHCLASS';

  HTSTER_CLASS  = 'HTSTER';

var
  HtsInitialize: function(hWnd, hParentWnd: THandle): Bool; stdcall;
  HtsClearWindow: function(hWnd: THandle): Bool; stdcall;
  HtsReformat: function(hWnd: THandle): Bool; stdcall;
  HtsClose: function(hWnd: THandle; OutputTo: Integer; FileName: PAnsiChar;
		     var hBuf: THandle; var BufLen: Integer): Bool; stdcall;
  HtsRead: function(hWnd: THandle; InputType: Integer; FileName: PAnsiChar;
		    hBuf: THandle; BufLen: Integer; TagName: PAnsiChar): Bool; stdcall;
  HtsSave: function(hWnd: THandle; OutputTo: Integer; FileName: PAnsiChar;
		    var hBuf: THandle; var BufLen: Integer): Bool; stdcall;
  TerCreateWindowAlt: function(Left, Top, Width, Height: Integer; IsContained: Bool;
			       ParentWnd: THandle; Styles: Cardinal): THandle; stdcall;
  CloseTer: function(hWnd: THandle; ForcedClose: Bool): Bool; stdcall;
  TerSetOutputFormat: function(hWnd: THandle; Format: Integer): Bool; stdcall;
  //TerRegisterMsgCallback: function(hWnd: THandle; TerMsgCallBack: TTerMsgCallback): Bool; stdcall;
  //GetTerFields: function(hWnd: THandle; var Field: TStrTerField): Bool; stdcall;
  //SetTerFields: function(hWnd: THandle; const Field: TStrTerField): Bool; stdcall;
  GetTerBuffer: function(hWnd: THandle; var Size: Longint): THandle; stdcall;
  SetTerBuffer: function(hWnd: THandle; Buffer: THandle; Size: Longint; Title: PAnsiChar; Release: Bool): Bool; stdcall;
  ParaNormal: function(hWnd: THandle; Repaint: Bool): Bool; stdcall;
  ParaHangingIndent: function(hWnd: THandle; Indent, Repaint: Bool): Bool; stdcall;
  TerCommand2: function(hWnd: THandle; CommandId: Integer; Send: Bool): Bool; stdcall;
  TerEnableScrollBar: function(hWnd: THandle; HScrollBar, Enable: Bool): Bool; stdcall;
  TerGetParam: function(hWnd: THandle; ParamType: Integer): Integer; stdcall;
  TerGetLineParam: function(hWnd: THandle; LineNo, ParamType: Integer): Integer; stdcall;
  TerGetListInfo: function(hWnd: THandle; ListId: Integer; ListName: PAnsiChar; var LevelCount, Flags: Integer): Bool; stdcall;
  TerRepaint: function(hWnd: THandle; ClearBackground: Bool): Bool; stdcall;
  TerRepaginate: function(hWnd: THandle; Repaint: Bool): Bool; stdcall;
  TerAbsToRowCol: procedure(hWnd: THandle; Pos: Longint; var Line, Column: Integer); stdcall;
  TerRowColToAbs: function(hWnd: THandle; Line, Column: Integer): Longint; stdcall;
  GetTerCursorPos: function(hWnd: THandle; var CurLine, CurCol: Integer): Bool; stdcall;
  SetTerCursorPos: function(hWnd: THandle; Line, Column: Integer; Repaint: Bool): Bool; stdcall;
  TerGetCaretPos: function(hWnd: THandle): Longint; stdcall;
  TerSetCaretPos: function(hWnd: THandle; Pos: Integer): Bool; stdcall;
  TerEngageCaret: function(hWnd: THandle; AtCursor: Bool): Bool; stdcall;
  DeselectTerText: function(hWnd: THandle; Repaint: Bool): Bool; stdcall;
  SelectTerText: function(hWnd: THandle; FirstLine, FirstCol, LastLine, LastCol: Integer; Repaint: Bool): Bool; stdcall;
  TerSelectParaText: function(hWnd: THandle; Repaint: Bool): Bool; stdcall;
  TerGetSelection: function(hWnd: THandle; var FirstLine, FirstCol, EndLine, EndCol: Integer): Bool; stdcall;
  TerDeleteBlock: function(hWnd: THandle; Repaint: Bool): Bool; stdcall;
  TerDeleteCells: function(hWnd: THandle; Selection: Integer; Repaint: Bool): Bool; stdcall;
  TerDeleteCellText: function(hWnd: THandle; Selection: Integer; Repaint: Bool): Bool; stdcall;
  TerGetReadOnly: function(hWnd: THandle): Bool; stdcall;
  TerSetReadOnly: function(hWnd: THandle; ReadOnly: Bool): Bool; stdcall;
  TerGetRtfSel: function(hWnd: THandle; var Size: Integer): THandle; stdcall;
  TerGetTextSelA: function(hWnd: THandle; var Size: Integer): THandle; stdcall;
  TerGetTextSelU: function(hWnd: THandle; var Size: Integer): THandle; stdcall;
  TerGetLineA: function(hWnd: THandle; Line: Integer; Text: PAnsiChar; Font: PWord): Integer; stdcall;
  TerGetLineU: function(hWnd: THandle; Line: Integer; Text: PWideChar; Font: PWord): Integer; stdcall;
  InsertRtfBufA: function(hWnd: THandle; Rtf: PAnsiChar; Length, Line, Column: Integer; Repaint: Bool): Bool; stdcall;
  InsertRtfBufU: function(hWnd: THandle; Rtf: PWideChar; Length, Line, Column: Integer; Repaint: Bool): Bool; stdcall;
  TerInsertLine: function(hWnd : THandle; Text: PAnsiChar; FontId, ParaId, CellId, ParaFID: Integer; Repaint: Bool): Bool; stdcall;
  TerInsertTextA: function(hWnd: THandle; Text: PAnsiChar; Font, Para: Integer; Repaint: Bool): Bool; stdcall;
  TerInsertTextU: function(hWnd: THandle; Text: PWideChar; Font, Para: Integer; Repaint: Bool): Bool; stdcall;
  TerGetTextHeight: function(hWnd: THandle): Integer; stdcall;
  TerGetPictOffset: function(hWnd: THandle; ID: Integer): Integer; stdcall;
  TerPastePicture: function(hWnd: THandle; Format: Cardinal; hData: THandle; ParaFrameId, Align: Integer; Insert: Bool): Integer; stdcall;
  TerPictAltInfo: function(hWnd: THandle; ID: Integer; GetInfo: Bool; AltInfo: PAnsiChar): Bool; stdcall;
  TerPictLinkName: function(hWnd: THandle; ID: Integer; GetName: Bool; FileName: PAnsiChar): Bool; stdcall;
  TerSetLinkPictDir: function(hWnd: THandle; Dir: PAnsiChar): Bool; stdcall;
  TerSectBreak: function(hWnd: THandle; Repaint: Bool): Bool; stdcall;
  TerSetFocus: function(hWnd: THandle): Bool; stdcall;
  TerSetFlags: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags2: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags3: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags4: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags5: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags6: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags7: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags8: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags9: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags10: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetFlags11: function(hWnd: THandle; SetFlag: Bool; FlagValue: Cardinal): Cardinal; stdcall;
  TerSetPflags: function(hWnd : THandle; PFlags: Integer; OnOff: Bool; Repaint: Bool): Bool; stdcall;
  TerCreateParaId: function(hWnd: THandle; ReuseId: Integer; Shared: Bool;
			    LeftIndent, RightIndent, FirstIndent: Integer;
			    TabId, StyleId, AuxId, Shading: Integer; PFlags: Cardinal;
			    SpaceBefore, SpaceAfter, SpaceBetween: Integer;
			    Flags: Integer): Integer; stdcall;
  TerGetParaInfo3: function(hWnd: THandle; LineNo: Integer; IsStyleItem: Bool;
			    var LeftIndent, RightIndent, FirstIndent: Integer;
			    var TabId, StyleId, AuxId, Shading: Integer; var PFlags: Cardinal;
			    var SpaceBefore, SpaceAfter, SpaceBetween: Integer;
			    var Flags, Aux1Id: Cardinal; var BkColor: TColorRef): Bool; stdcall;
  TerSetParaIndent: function(hWnd: THandle; Left, Right, First: Integer; Repaint: Bool): Bool; stdcall;
  TerNormalizeBlock: function(hWnd: THandle): Bool; stdcall;
  TerIgnoreCommand: function(hWnd: THandle): Bool; stdcall;
  TerDisableDragDrop: procedure(hWnd: THandle); stdcall;
  TerEnableDragDrop: procedure(hWnd: THandle); stdcall;


function GetTerFlags(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags(Window, FALSE, FlagValue);
end;

function SetTerFlags(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags(Window, TRUE, FlagValue);
end;

function GetTerFlags2(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags2(Window, FALSE, FlagValue);
end;

function SetTerFlags2(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags2(Window, TRUE, FlagValue);
end;

function GetTerFlags3(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags3(Window, FALSE, FlagValue);
end;

function SetTerFlags3(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags3(Window, TRUE, FlagValue);
end;

function GetTerFlags4(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags4(Window, FALSE, FlagValue);
end;

function SetTerFlags4(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags4(Window, TRUE, FlagValue);
end;

function GetTerFlags5(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags5(Window, FALSE, FlagValue);
end;

function SetTerFlags5(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags5(Window, TRUE, FlagValue);
end;

function GetTerFlags6(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags6(Window, FALSE, FlagValue);
end;

function SetTerFlags6(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags6(Window, TRUE, FlagValue);
end;

function GetTerFlags7(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags7(Window, FALSE, FlagValue);
end;

function SetTerFlags7(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags7(Window, TRUE, FlagValue);
end;

function GetTerFlags8(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags8(Window, FALSE, FlagValue);
end;

function SetTerFlags8(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags8(Window, TRUE, FlagValue);
end;

function GetTerFlags9(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags9(Window, FALSE, FlagValue);
end;

function SetTerFlags9(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  Result := TerSetFlags9(Window, TRUE, FlagValue);
end;

function GetTerFlags10(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  if Assigned(TerSetFlags10) then
    Result := TerSetFlags10(Window, FALSE, FlagValue)
  else Result := 0;
end;

function SetTerFlags10(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  if Assigned(TerSetFlags10) then
    Result := TerSetFlags10(Window, TRUE, FlagValue)
  else Result := 0;
end;

function GetTerFlags11(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  if Assigned(TerSetFlags11) then
    Result := TerSetFlags11(Window, FALSE, FlagValue)
  else Result := 0;
end;

function SetTerFlags11(Window: THandle; FlagValue: Cardinal): Cardinal;
begin
  if Assigned(TerSetFlags11) then
    Result := TerSetFlags11(Window, TRUE, FlagValue)
  else Result := 0;
end;

function FindMostRecentVersion(const ModulePath, PatternName, File32Name: String; CheckVersions: Longbool): String;
var
  hSearch: THandle;
  fdFile: TWin32FindData;
  s32File, sPatternFile: String;
begin
  Result := E;
  if not CheckVersions then begin
    Result := File32Name;
    Exit;
  end;
  hSearch := FindFirstFile(PAnsiChar(ModulePath + PatternName), fdFile);
  if hSearch <> INVALID_HANDLE_VALUE then begin
    try
      repeat
	if (fdFile.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
	  if CompareText(fdFile.cFileName, File32Name) = 0 then
	    s32File := ModulePath + File32Name
	  else sPatternFile := ModulePath + fdFile.cFileName;
	end;
      until not FindNextFile(hSearch, fdFile);
    finally
      Windows.FindClose(hSearch);
    end;
  end;
  if GetModuleHandle(PAnsiChar(s32File)) <> 0 then
    Result := ExtractFileName(s32File)
  else if GetModuleHandle(PAnsiChar(sPatternFile)) <> 0 then
    Result := ExtractFileName(sPatternFile)
  else Result := File32Name;
end;

function LoadHtsTer(const ModulePath: String; CheckVersions: Longbool; var TerVersion: String): Longbool;
var
  hTER, hHTS: THandle;
begin
  Result := false;
  hTER := LoadLibrary(PAnsiChar(FindMostRecentVersion(ModulePath, TER_LIB, TER_32, CheckVersions)));
  if hTER = 0 then Exit;
  hHTS := LoadLibrary(PAnsiChar(FindMostRecentVersion(ModulePath, HTS_LIB, HTS_32, CheckVersions)));
  if hHTS = 0 then Exit;
  HtsInitialize := GetProcAddress(hHTS, 'HtsInitialize');
  HtsClearWindow := GetProcAddress(hHTS, 'HtsClearWindow');
  HtsReformat := GetProcAddress(hHTS, 'HtsReformat');
  HtsClose := GetProcAddress(hHTS, 'HtsClose');
  HtsRead := GetProcAddress(hHTS, 'HtsRead');
  HtsSave := GetProcAddress(hHTS, 'HtsSave');
  TerCreateWindowAlt := GetProcAddress(hTER, 'TerCreateWindowAlt');
  CloseTer := GetProcAddress(hTER, 'CloseTer');
  TerSetOutputFormat := GetProcAddress(hTER, 'TerSetOutputFormat');
  //TerRegisterMsgCallback := GetProcAddress(hTER, 'TerRegisterMsgCallback');
  //GetTerFields := GetProcAddress(hTER, 'GetTerFields');
  //SetTerFields := GetProcAddress(hTER, 'SetTerFields');
  GetTerBuffer := GetProcAddress(hTER, 'GetTerBuffer');
  SetTerBuffer := GetProcAddress(hTER, 'SetTerBuffer');
  ParaNormal := GetProcAddress(hTER, 'ParaNormal');
  ParaHangingIndent := GetProcAddress(hTER, 'ParaHangingIndent');
  TerCommand2 := GetProcAddress(hTER, 'TerCommand2');
  TerEnableScrollBar := GetProcAddress(hTER, 'TerEnableScrollBar');
  TerGetParam := GetProcAddress(hTER, 'TerGetParam');
  TerGetLineParam := GetProcAddress(hTER, 'TerGetLineParam');
  TerGetListInfo := GetProcAddress(hTER, 'TerGetListInfo');
  TerRepaint := GetProcAddress(hTER, 'TerRepaint');
  TerRepaginate := GetProcAddress(hTER, 'TerRepaginate');
  TerAbsToRowCol := GetProcAddress(hTER, 'TerAbsToRowCol');
  TerRowColToAbs := GetProcAddress(hTER, 'TerRowColToAbs');
  GetTerCursorPos := GetProcAddress(hTER, 'GetTerCursorPos');
  SetTerCursorPos := GetProcAddress(hTER, 'SetTerCursorPos');
  TerGetCaretPos := GetProcAddress(hTER, 'TerGetCaretPos');
  TerSetCaretPos := GetProcAddress(hTER, 'TerSetCaretPos');
  TerEngageCaret := GetProcAddress(hTER, 'TerEngageCaret');
  DeselectTerText := GetProcAddress(hTER, 'DeselectTerText');
  SelectTerText := GetProcAddress(hTER, 'SelectTerText');
  TerSelectParaText := GetProcAddress(hTER, 'TerSelectParaText');
  TerGetSelection := GetProcAddress(hTER, 'TerGetSelection');
  TerDeleteBlock := GetProcAddress(hTER, 'TerDeleteBlock');
  TerDeleteCells := GetProcAddress(hTER, 'TerDeleteCells');
  TerDeleteCellText := GetProcAddress(hTER, 'TerDeleteCellText');
  TerGetReadOnly := GetProcAddress(hTER, 'TerGetReadOnly');
  TerSetReadOnly := GetProcAddress(hTER, 'TerSetReadOnly');
  TerGetRtfSel := GetProcAddress(hTER, 'TerGetRtfSel');
  TerGetTextSelA := GetProcAddress(hTER, 'TerGetTextSel');
  TerGetTextSelU := GetProcAddress(hTER, 'TerGetTextSelU');
  TerGetLineA := GetProcAddress(hTER, 'TerGetLine');
  TerGetLineU := GetProcAddress(hTER, 'TerGetLineU');
  InsertRtfBufA := GetProcAddress(hTER, 'InsertRtfBuf');
  InsertRtfBufU := GetProcAddress(hTER, 'InsertRtfBufU');
  TerInsertLine := GetProcAddress(hTER, 'TerInsertLine');
  TerInsertTextA := GetProcAddress(hTER, 'TerInsertText');
  TerInsertTextU := GetProcAddress(hTER, 'TerInsertTextU');
  TerGetTextHeight := GetProcAddress(hTER, 'TerGetTextHeight');
  TerGetPictOffset := GetProcAddress(hTER, 'TerGetPictOffset');
  TerPastePicture := GetProcAddress(hTER, 'TerPastePicture');
  TerPictAltInfo := GetProcAddress(hTER, 'TerPictAltInfo');
  TerPictLinkName := GetProcAddress(hTER, 'TerPictLinkName');
  TerSetLinkPictDir := GetProcAddress(hTER, 'TerSetLinkPictDir');
  TerSectBreak := GetProcAddress(hTER, 'TerSectBreak');
  TerSetFocus := GetProcAddress(hTER, 'TerSetFocus');
  TerSetFlags := GetProcAddress(hTER, 'TerSetFlags');
  TerSetFlags2 := GetProcAddress(hTER, 'TerSetFlags2');
  TerSetFlags3 := GetProcAddress(hTER, 'TerSetFlags3');
  TerSetFlags4 := GetProcAddress(hTER, 'TerSetFlags4');
  TerSetFlags5 := GetProcAddress(hTER, 'TerSetFlags5');
  TerSetFlags6 := GetProcAddress(hTER, 'TerSetFlags6');
  TerSetFlags7 := GetProcAddress(hTER, 'TerSetFlags7');
  TerSetFlags8 := GetProcAddress(hTER, 'TerSetFlags8');
  TerSetFlags9 := GetProcAddress(hTER, 'TerSetFlags9');
  TerSetFlags10 := GetProcAddress(hTER, 'TerSetFlags10');
  TerSetFlags11 := GetProcAddress(hTER, 'TerSetFlags11');
  TerSetPflags := GetProcAddress(hTER, 'TerSetPflags');
  TerCreateParaId := GetProcAddress(hTER, 'TerCreateParaId');
  TerGetParaInfo3 := GetProcAddress(hTER, 'TerGetParaInfo3');
  TerSetParaIndent := GetProcAddress(hTER, 'TerSetParaIndent');
  TerNormalizeBlock := GetProcAddress(hTER, 'TerNormalizeBlock');
  TerIgnoreCommand := GetProcAddress(hTER, 'TerIgnoreCommand');
  TerDisableDragDrop := GetProcAddress(hTER, 'TerDisableDragDrop');
  TerEnableDragDrop := GetProcAddress(hTER, 'TerEnableDragDrop');
  TerVersion := CardToStr(GetSimpleFileVersion(GetModuleFileName(hTER), true));
  Result := true;
end;


// THtsTerConverter --------------------------------------------------------------------------------

function SubClassWndProc(hWindow: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
type
  TPictArray = Array[TImageType] of 0..7;
const
  HTS_PICTURE = $CCC;
  HTS_PICT_ID = $CCE;
const
  PICT_UNKNOWN = 0;
  PICT_BMP     = 1;
  PICT_GIF     = 2;
  PICT_WMF     = 3;
  PICT_TIFF    = 4;
  PICT_JPG     = 5;
  PICT_PNG     = 6;
  PICT_NONE    = 7;
const
  PICT_TYPES: TPictArray = (PICT_NONE, PICT_BMP, PICT_GIF, PICT_JPG, PICT_PNG, PICT_TIFF, PICT_WMF);
var
  HtsTerConverter: THtsTerConverter;
  sFileName: String;
  GdipImage: TGdipImage;
begin
  HtsTerConverter := THtsTerConverter(GetProp(hWindow, HTSTER_CLASS));
  if HtsTerConverter <> nil then begin
    case uMsg of
      HTS_PICTURE: begin
	Result := PICT_UNKNOWN;
	with HtsTerConverter do if (THandle(wParam) = FTerWindow) and (FImageList <> nil) then begin
	  sFileName := PAnsiChar(lParam);
	  if sFileName <> E then begin
	    if IsFileUrl(sFileName) then
	      sFileName := GetFilePathFromUrl(sFileName)
	    else if IsResUrl(sFileName) then
	      sFileName := FDefaultBmp;
	    GdipImage := TGdipImage.Create;
	    if GdipImage <> nil then with GdipImage do begin
	      try
		if IsCIDUrl(sFileName) then
		  FImageList.Add(Copy(sFileName, succ(Length(PROT_CID)), MAXINT))
		else if FileExists(sFileName) and LoadFromFile(sFileName) then begin
		  Result := PICT_TYPES[ImageType];
		  if Result <> PICT_NONE then begin
		    StrPLCopy(PAnsiChar(lParam), sFileName, MAX_PATH);
		    Exit;
		  end;
		end;
	      finally
		GdipImage.Free;
	      end;
	    end;
	  end;
	end;
	if Result in [PICT_UNKNOWN, PICT_NONE] then with HtsTerConverter do begin
	  StrPLCopy(PAnsiChar(lParam), PAnsiChar(FDefaultBmp), MAX_PATH);
	  Result := PICT_BMP;
	  Exit;
	end;
      end;
      HTS_PICT_ID: begin
	with HtsTerConverter do if THandle(wParam) = FTerWindow then begin
	  with FImageList do if Count > 0 then begin
	    Objects[pred(Count)] := TObject(lParam);
	    TerPictAltInfo(FTerWindow, lParam, false, PAnsiChar(AnsiString(ImageNames[pred(Count)])));
	  end;
	end;
      end;
    end;
  end;
  Result := DefWindowProc(hWindow, uMsg, wParam, lParam);
end;

function THtsTerConverter.HostWindowCreate(MaxWidth: Integer): Longbool;
var
  rWindow: TRect;
begin
  rWindow := GetSysScreenRect;
  with rWindow do begin
    if MaxWidth > 0 then
      Right := Left + MaxWidth
    else Right := MAXINT - Left;
    FHostWindow := CreateWindow(CLASS_DIALOG, nil, WS_POPUP or WS_DISABLED,
				Left, Top, Right - Left, Bottom - Top, 0, 0, HInstance, nil);
  end;
  if FHostWindow <> 0 then begin
    FHostWndProc := TFarProc(SetWindowLong(FHostWindow, GWL_WNDPROC, Longint(@SubClassWndProc)));
    SetWindowPos(FHostWindow, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
    SetProp(FHostWindow, PChar(FPropName), THandle(Self));
    Result := true;
    Exit;
  end;
  Result := false;
end;

procedure THtsTerConverter.HostWindowDestroy;
begin
  SetWindowLong(FHostWindow, GWL_WNDPROC, Longint(FHostWndProc));
  RemoveProp(FHostWindow, PChar(FPropName));
  if FHostWindow <> 0 then begin
    DestroyWindow(FHostWindow);
    FHostWindow := 0;
  end;
end;

function THtsTerConverter.TerWindowCreate(ParentHandle: THandle): Longbool;
const
  TER_DEF_STYLES = WS_CHILD or TER_WORD_WRAP or TER_BORDER_MARGIN or TER_READ_ONLY or TER_FITTED_VIEW or TER_INVISIBLE;
begin
  with GetClientRect(FHostWindow) do begin
    FTerWindow := TerCreateWindowAlt(0, 0, Right, Bottom, TRUE, ParentHandle, TER_DEF_STYLES or TER_OUTPUT_RTF);
    Result := (FTerWindow <> 0);
  end;
end;

procedure THtsTerConverter.TerWindowResize(MaxWidth: Integer);
begin
  with GetClientRect(FHostWindow) do begin
    MoveWindow(FTerWindow, 0, 0, MaxWidth, Bottom, FALSE);
    SetTerFlags4(FTerWindow, TFLAG4_TEXT_HIDDEN);
    SetTerFlags5(FTerWindow, TFLAG5_FULL_REPAGINATE);
  end;
end;

procedure THtsTerConverter.TerWindowClose;
var
  iLen: Integer;
  hDummy: THandle;
begin
  iLen := 0;
  hDummy := 0;
  if FTerWindow <> 0 then begin
    HtsClearWindow(FTerWindow);
    if HtsClose(FTerWindow, HTML_BUF, nil, hDummy, iLen) and (hDummy <> 0) then
      GlobalFree(hDummy)
    else CloseTer(FTerWindow, true);
    DestroyWindow(FTerWindow);
    FTerWindow := 0;
  end;
end;

constructor THtsTerConverter.Create(const ModulePath: String; CheckVersions: Longbool; MaxWidth: Integer);
begin
  inherited Create;
  FImageList := TImageList.Create;
  if FImageList <> nil then with FImageList do begin
    Duplicates := dupAccept;
    Sorted := false;
    BeginUpdate;
  end;
  FPropName := HTSTER_CLASS;
  FDefaultBmp := ModulePath + DEFAULT_BMP;
  if not FileExists(FDefaultBmp) then FDefaultBmp := E;
  if LoadHtsTer(ModulePath, CheckVersions, FTerVersion) and HostWindowCreate(MaxWidth)
  and TerWindowCreate(FHostWindow) then begin
    SetTerFlags6(FTerWindow, GetTerFlags6(FTerWindow, 0) and not TFLAG6_DEL_CELL_TEXT);
    HtsInitialize(FTerWindow, FHostWindow);
  end;
end;

destructor THtsTerConverter.Destroy;
begin
  TerWindowClose;
  HostWindowDestroy;
  FImageList.Free;
  inherited;
end;

function THtsTerConverter.GetRtf(var TextHeight: Integer): String;
var
  hRtf: THandle;
  lLen: Longint;
begin
  TextHeight := PixelsFromTwips(TerGetTextHeight(FTerWindow));
  TerSetOutputFormat(FTerWindow, SAVE_RTF);
  hRtf := GetTerBuffer(FTerWindow, lLen);
  Result := GetHandleStringA(hRtf, lLen, TRUE);
  HtsClearWindow(FTerWindow);
end;

function THtsTerConverter.GetHtml(var TextHeight: Integer): String;
var
  hHtml: THandle;
  lLen: Longint;
begin
  TextHeight := PixelsFromTwips(TerGetTextHeight(FTerWindow));
  TerSetOutputFormat(FTerWindow, SAVE_HTML);
  hHtml := GetTerBuffer(FTerWindow, lLen);
  Result := GetHandleStringA(hHtml, lLen, TRUE);
  HtsClearWindow(FTerWindow);
end;

function THtsTerConverter.GetTextA(var TextHeight: Integer): String;
var
  hText: THandle;
  lLen: Longint;
begin
  TextHeight := PixelsFromTwips(TerGetTextHeight(FTerWindow));
  TerSetOutputFormat(FTerWindow, SAVE_TEXT);
  hText := GetTerBuffer(FTerWindow, lLen);
  Result := GetHandleStringA(hText, lLen, TRUE);
  HtsClearWindow(FTerWindow);
end;

function THtsTerConverter.GetTextW(var TextHeight: Integer): String;
var
  hText: THandle;
  lLen: Longint;
begin
  TextHeight := PixelsFromTwips(TerGetTextHeight(FTerWindow));
  TerSetOutputFormat(FTerWindow, SAVE_UTEXT);
  hText := GetTerBuffer(FTerWindow, lLen);
  Result := GetHandleStringW(hText, lLen, TRUE);
  HtsClearWindow(FTerWindow);
end;

function THtsTerConverter.SetPictureLink(Index: Integer): Longbool;
begin
  with ImageList do Result := TerPictLinkName(FTerWindow, ImageIDs[Index], FALSE, PAnsiChar(AnsiString(ImageNames[Index])));
end;

function THtsTerConverter.HtmlToRtf(const Html: String; MaxWidth: Integer; var TextHeight: Integer): String;
var
  iIndex: Integer;
begin
  HtsClearWindow(FTerWindow);
  if (FTerWindow = 0) or (Html = E) then begin
    Result := E;
    Exit;
  end;
  if MaxWidth > -1 then TerWindowResize(MaxWidth);
  if HtsRead(FTerWindow, HTML_BUF, NIL, GetStringHandleA(Html, 0, false), Length(Html), NIL)
  and TerCommand2(FTerWindow, ID_SELECT_ALL, TRUE) then begin
    if TerGetParam(FTerWindow, TP_TOTAL_LISTS) > 0 then begin
      DeselectTerText(FTerWindow, FALSE);
      if MaxWidth < GetSysScreenDimensions.cx div 4 then begin
	for iIndex := 0 to pred(TerGetParam(FTerWindow, TP_TOTAL_LINES)) do begin
	  SetTerCursorPos(FTerWindow, iIndex, 0, FALSE);
	  if TerGetLineParam(FTerWindow, iIndex, LP_LIST_ID) <> 0 then
	    TerSetParaIndent(FTerWindow, TWIPS shr 1, -1, -1, FALSE)
	  else TerSetParaIndent(FTerWindow, -1, -1, -1, FALSE);
	end;
      end;
    end;
    with FImageList do for iIndex := 0 to pred(Count) do begin
      TerPictAltInfo(FTerWindow, ImageIDs[iIndex], false, PChar(ImageNames[iIndex]));
    end;
    Result := GetRtf(TextHeight);
  end;
end;

function THtsTerConverter.RtfToHtml(const Rtf: String; MaxWidth: Integer; var TextHeight: Integer): String;
begin
  HtsClearWindow(FTerWindow);
  if (FTerWindow = 0) or (Rtf = E) then begin
    Result := E;
    Exit;
  end;
  if MaxWidth > -1 then TerWindowResize(MaxWidth);
  if InsertRtfBufA(FTerWindow, PAnsiChar(Rtf), Length(Rtf), -1, -1, FALSE) then Result := GetHtml(TextHeight);
end;


end.

