{$J+,Z4}
unit MailUtils;
{------------------------------------------------------------------------------}
{        This code is copyright (C) 2006-2023 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, CommCtrl, Classes, Controls, ComCtrls, Graphics, SysUtils, Tidy,
  StrUtil, FileUtils, TimeUtils, TreeViewEx, ImgList, ImgListEx, RichEd20, HtsTer,
  UTF8, DatedStrings, Base64, PMMessages, PMPGPCommon;

const // MIME message headers
  DD			= DA + DA;
  UTC			= '(UTC)';
  MIME_VERSION		= 'Mime-Version:';
  MIME_V10		= '1.0';
  CT_HEADER		= 'Content-Type:';
  CT_MESSAGE		= 'message/RFC822';
  CT_MULTIPART		= 'multipart/';
  CT_BOUNDARY		= 'boundary';
  CT_MIXED		= 'mixed';
  CT_DIGEST		= 'digest';
  CT_RELATED		= 'related';
  CT_SIGNED		= 'signed';
  CT_ENCRYPTED		= 'encrypted';
  CT_ALTERNATIVE	= 'alternative';
  CT_APPLICATION	= 'application/';
  CT_TEXT		= 'text/';
  _PLAIN		= 'plain';
  _HTML			= 'html';
  _RTF			= 'rtf';
  CT_CHARSET		= 'charset';
  CT_OCTET_STREAM	= 'octet-stream';
  CT_NAME		= 'name';
  CTE_HEADER		= 'Content-Transfer-Encoding:';
  CTE_7BIT		= '7BIT';
  CTE_8BIT		= '8BIT';
  CTE_B64		= 'base64';
  CTE_QP		= 'quoted-printable';
  CDP_HEADER		= 'Content-Disposition:';
  CDP_ATTACHMENT	= 'attachment';
  CDP_FILENAME 		= 'filename';
  CDP_INLINE		= 'inline';
  CTD_HEADER		= 'Content-Description:';
  CTD_BODY		= 'Mail message body';
  CTD_AINFO		= 'Attachment information';
  CLC_HEADER		= 'Content-Location:';
  CID_HEADER		= 'Content-ID:';
  CT_TEXT_ASCII		= CT_HEADER + SP + CT_TEXT + _PLAIN + SC + SP + CT_CHARSET + EQ + US_ASCII;
  CT_TEXT_UTF8		= CT_HEADER + SP + CT_TEXT + _PLAIN + SC + SP + CT_CHARSET + EQ + UTF_8;
  CTE_TEXT_7BIT		= CTE_HEADER + SP + CTE_7BIT;
  CTE_TEXT_8BIT		= CTE_HEADER + SP + CTE_8BIT;
  CTE_QUOTED_PRINTABLE	= CTE_HEADER + SP + CTE_QP;
  CTE_BASE_64		= CTE_HEADER + SP + CTE_B64;
  MIME_TEXT_HEADERS	= MIME_VERSION + SP + MIME_V10 + CRLF +
			  CT_TEXT_ASCII + CRLF +
			  CTE_TEXT_7BIT + CRLF +
			  CTD_HEADER + SP + CTD_BODY;

const // Date/Time abbreviations
  MONTHS = 'jan/feb/mar/apr/may/jun/jul/aug/sep/oct/nov/dec';
  TZONES = ' ut/gmt/edt/est/cdt/cst/mdt/mst/pdt/pst';

const // Additional mail headers
  X_CONFIRM_TO		= 'X-Confirm-reading-to:';
  X_PRIORITY		= 'X-Priority:';
  H_IMPORTANCE		= 'Importance:';
  H_PRIORITY		= 'Priority:';
  H_DATE		= 'Date:';
  H_FROM		= 'From:';
  H_TO			= 'To:';
  H_CC			= 'Cc:';
  H_BCC			= 'Bcc:';
  H_SUBJECT		= 'Subject:';
  H_REPLY_TO		= 'Reply-to:';

const // Pegasus Mail headers
  XPM_PLACEHOLDER	= 'X-PM-PLACEHOLDER:';
  XPM_FLAGS		= 'X-PMFLAGS:';

const // PM Encryption flags
  XPM_FLAG_ENCRYPTED	= 1;
  XPM_FLAG_SIGNED	= 4;

const // Encryption headers
  XPM_ENCRYPTOR		= 'X-PM-Encryptor:';
  X_ENCRYPTION		= 'x-encryption';
  X_ENCRYPTION_FLAGS	= X_ENCRYPTION + '-flags';

const // PMPGP encryption headers
  IDWPGP_TAG		= 'IDWPGP-PM32';
  XPM_IDWPGP		= XPM_ENCRYPTOR + SP + IDWPGP_TAG + CM + SP + '%d' + CRLF;
  X_IDWPGP		= X_ENCRYPTION + EQ + IDWPGP_TAG + SC + SP + X_ENCRYPTION_FLAGS + EQ;

const // PGP headers and markers
  // S/MIME
  PKCS7			= 'PKCS7';
  SM_ENC		= 'enveloped';
  SM_SIG		= 'signature';
  SM_SIGNED		= 'signed';
  // PGP/MIME
  APP_PGP		= CT_APPLICATION + 'pgp-';
  PGPMIME		= 'protocol="' + APP_PGP;
  PM_ENC		= 'encrypted';
  PM_SIG		= 'signature';
  PM_KEYS		= 'keys';
  // Any encoding
  ANY_BEG		= '-----BEGIN ';
  ANY_END		= '-----END ';
  // Certs
  IS_CERT		= 'CERTIFICATE-----';
  CERT_BEG		= ANY_BEG + IS_CERT;
  CERT_END		= ANY_END + IS_CERT;
  // Plain
  IS_PGP		= 'PGP';
  PGP_BEG		= ANY_BEG + IS_PGP;
  PGP_END		= ANY_END + IS_PGP;
  KEY_BEG		= ' PUBLIC KEY BLOCK-----';
  KEY_END		= KEY_BEG;
  SIGN_BEG		= ' SIGNED MESSAGE-----';
  SIG_BEG		= ' SIGNATURE-----';
  SIG_END		= SIG_BEG;
  MSG_BEG		= ' MESSAGE-----';
  MSG_END		= MSG_BEG;

const // Header transfer encoding
  CODE_BEG		= '=?';
  CODE_QP		= '?Q?';
  CODE_B64		= '?B?';
  CODE_END		= '?=';

const // Char ranges
  CHARS_BINARY		= [#0..#31] - [TAB, LF, CR];
  CHARS_HIGH_BIT	= [#127..#255];
  CHARS_EXTENDED	= CHARS_BINARY + CHARS_HIGH_BIT;
  CHARS_TO_ENCODE	= CHARS_BINARY + [EQ, QM, UL] + CHARS_HIGH_BIT;

const // QP stuff
  MAX_QP_LINE_LEN	= 76;
  CRLFSP		= CRLF + SP;

const // Separators
  UNIX_MAIL_START	= 'From ???@???';
  PM_MAIL_END		= #26;

type // Find multipart types
  TFindMultipartOption = (
    fmoFirstSubHeader,
    fmoFirstNonMixed,
    fmoLastMultipart,
    fmoFirstContent
  );

type // Find content headers
  TFindHeaderOption = (
    fhoContentType,
    fhoContentTransferEncoding,
    fhoContentDisposition,
    fhoContentDescription,
    fhoContentLocation,
    fhoContentID
  );
  TFindHeaderOptions = Set of TFindHeaderOption;

const
  FHO_ALL = [
    fhoContentType,
    fhoContentTransferEncoding,
    fhoContentDisposition,
    fhoContentDescription,
    fhoContentLocation,
    fhoContentID
  ];

type // Mime decode options
  TMimeDecodeOption = (
    mdoAnalyseEncoding,
    mdoReturnUnicode,
    mdoReturnRtf,
    mdoFancyText,
    mdoFullParse,
    mdoDecode
  );
  TMimeDecodeOptions = Set of TMimeDecodeOption;

const
  MDO_ALL = [
    mdoReturnUnicode,
    mdoReturnRtf,
    mdoFancyText,
    mdoFullParse,
    mdoDecode
  ];

type // Encoding
  TEncoding = (
    eNone,
    ePlainEncrypted,
    ePlainSigned,
    ePlainKeys,
    eCerts,
    ePMEncrypted,
    ePMSigned,
    ePMKeys,
    eSMEncrypted,
    eSMSigned,
    eSMInline
  );

type // Data types
  TMsgDataType = (
    mdtUnknownData,
    mdtMultipartUnknown,
    mdtMultipartMixed,
    mdtMultipartRelated,
    mdtMultipartAlternative,
    mdtMultipartEncrypted,
    mdtMultipartSigned,
    mdtMultipartDigest,
    mdtMainHeaders,
    mdtSubHeaders,
    mdtSignature,
    mdtPlainText,
    mdtHtmlText,
    mdtRtfText,
    mdtAltText,
    mdtFile
  );

type // Header data
  PHeaderData = ^THeaderData;
  THeaderData = Record
    Text: String;
    Pos: Integer;
    Len: Integer;
    MaxPos: Integer;
    Recursion: Integer;
    Encoding: TEncoding;
    CTHeader: String;
    CTEHeader: String;
    CDPHeader: String;
    CTDHeader: String;
    CLCHeader: String;
    CIDHeader: String;
  end;

type // Multipart data
  THeaderDataEx = Record
    HeaderData: THeaderData;
    MultipartType: TMsgDataType;
    Boundary: String;
    Charset: String;
  end;

type // Mail message
  TMessageData = Record
    Encoding: TEncoding;
    ExtraData: Longbool;
    Prefix: String;
    Suffix: String;
  end;

type // Address counts
  TAddressCounts = Record
    ToRecipients: Integer;
    CcRecipients: Integer;
    BccRecipients: Integer;
  end;

type // Header addresses
  TAddressType = (
    atTo,
    atCc,
    atBcc
  );
  TAddressStrings = Array[TAddressType] of String;

type // Bounds
  TBounds = Record
    BegPos: Integer;
    EndPos: Integer;
  end;

type // Classes
  TMimeMsgTree = Class;


// TSectionNode ------------------------------------------------------------------------------------

  PSectionNode = ^TSectionNode;
  TSectionNode = class(TTreeNode)
  private
    FHeaderDataEx: THeaderDataEx;
    FIncludeImages: Longbool;
    FSectionCID: String;
    FSectionData: String;
    FSectionSig: String;
    FSectionFlags: Integer;
    FSectionType: TMsgDataType;
    FUTCDateTime: TDateTime;
    function  GetUtf8Date: String;
    function  GetUtf8UTCDate: String;
    function  GetShortUtf8Date: String;
    function  GetUTCDateTime: TDateTime;
    function  GetUtf8Sender: String;
    function  GetUtf8Recipient: String;
    function  GetUtf8Subject: String;
    function  GetUtf8ReplyTo: String;
    function  GetFileExt: String;
    function  GetFileName: String;
    function  GetFileNameFromHeaders: String;
    function  GetHasFormattedText: Longbool;
    function  GetIsPSMEncoded: Longbool;
    function  GetAnsiMsgHeaders: String;
    function  GetUtf8MainHeaders: String;
    function  GetMimeMsgTree: TMimeMsgTree;
    function  GetSectionIcon: TIcon;
    function  GetSelectedIcon: TIcon;
    function  GetSectionText: String;
    function  GetSectionCID: String;
    procedure SetSectionCID(const Value: String);
    function  GetSectionName: String;
    procedure SetSectionName(const Value: String);
    function  GetParent: TSectionNode;
    function  GetChild(Index: Integer): TSectionNode;
  protected
    procedure CopyHeaderPosLen(const HeaderData: THeaderData);
    function  GetHeader(const Name: String; DropComments: Longbool): String;
    property  HeaderDataEx: THeaderDataEx read FHeaderDataEx write FHeaderDataEx;
    property  HeaderData: THeaderData read FHeaderDataEx.HeaderData write FHeaderDataEx.HeaderData;
  public
    function  GetFirstChild: TSectionNode;
    function  GetFullMultipartSection: String;
    function  GetRelatedSectionsAsRtf: String;
    function  QPFileName(const Extension: String): String;
    function  SaveRelatedSectionsAsHtml(const HtmlFileName: String): Longbool;
    function  SetEncodedData(const Data, MimeCode: String; XPMHeaderFlags: Integer;
			     AppendPGPExtension, InsertSignature: Longbool): Integer;
    function  SetPMFlags(IsMultipart, IsEncrypted, IsMime, IsRead: Boolean; const FileName: String): String;
    property  Child[Index: Integer]: TSectionNode read GetChild; default;
    property  Utf8Date: String read GetUtf8Date;
    property  Utf8UTCDate: String read GetUtf8UTCDate;
    property  ShortUtf8Date: String read GetShortUtf8Date;
    property  UTCDateTime: TDateTime read GetUTCDateTime;
    property  Utf8Sender: String read GetUtf8Sender;
    property  Utf8Recipient: String read GetUtf8Recipient;
    property  Utf8Subject: String read GetUtf8Subject;
    property  Utf8ReplyTo: String read GetUtf8ReplyTo;
    property  FileExt: String read GetFileExt;
    property  FileName: String read GetFileName;
    property  FileNameFromHeaders: String read GetFileNameFromHeaders;
    property  HasFormattedText: Longbool read GetHasFormattedText;
    property  HeaderText: String read FHeaderDataEx.HeaderData.Text;
    property  IsPSMEncoded: Longbool read GetIsPSMEncoded;
    property  AnsiMsgHeaders: String read GetAnsiMsgHeaders;
    property  Utf8MainHeaders: String read GetUtf8MainHeaders;
    property  MimeMsgTree: TMimeMsgTree read GetMimeMsgTree;
    property  Parent: TSectionNode read GetParent;
    property  SectionIcon: TIcon read GetSectionIcon;
    property  SelectedIcon: TIcon read GetSelectedIcon;
    property  SectionText: String read GetSectionText;
    property  SectionData: String read FSectionData write FSectionData;
    property  SectionCID: String read GetSectionCID write SetSectionCID;
    property  SectionSig: String read FSectionSig write FSectionSig;
    property  SectionFlags: Integer read FSectionFlags write FSectionFlags;
    property  SectionName: String read GetSectionName write SetSectionName;
    property  SectionType: TMsgDataType read FSectionType write FSectionType;
    property  IncludeImages: Longbool read FIncludeImages write FIncludeImages;
    property  Charset: String read FHeaderDataEx.Charset write FHeaderDataEx.Charset;
    property  Boundary: String read FHeaderDataEx.Boundary write FHeaderDataEx.Boundary;
    property  Encoding: TEncoding read FHeaderDataEx.HeaderData.Encoding write FHeaderDataEx.HeaderData.Encoding;
    property  MultipartType: TMsgDataType read FHeaderDataEx.MultipartType write FHeaderDataEx.MultipartType;
  end;


// TTextList ---------------------------------------------------------------------------------------

  TTextList = class(TStringList)
  private
    FMimeMsgTree: TMimeMsgTree;
    function  GetAnsiText: String;
    function  GetUnicodeText: String;
    function  GetSectionNode(Index: Integer): TSectionNode;
  protected
  public
    property  MimeMsgTree: TMimeMsgTree read FMimeMsgTree;
    property  AnsiText: String read GetAnsiText;
    property  UnicodeText: String read GetUnicodeText;
    property  SectionNodes[Index: Integer]: TSectionNode read GetSectionNode;
  end;


// TFileSections ----------------------------------------------------------------------------------

  PFSItem = ^TFSItem;
  TFSItem = Record
    Data: String;
    Name: String;
    Used: Longbool;
  end;

  TFileSections = class(TStringList)
  private
    FMaxSingleEntrySize: Integer;
    FModifiedMainSection: String;
    function  GetCID(Index: Integer): String;
    function  GetFileData(Index: Integer): String;
    function  GetFileName(Index: Integer): String;
    function  GetUsed(Index: Integer): Longbool;
    procedure SetUsed(Index: Integer; Value: Longbool);
  protected
  public
    constructor Create(SectionNode: TSectionNode; ReturnRtf: Longbool);
    destructor Destroy; override;
    function  AddItem(const ItemName, ItemData, ItemCID: String): Integer;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    function  WriteToFiles(const TextFileName: String): Longbool;
    property  FileNames[Index: Integer]: String read GetFileName;
    property  FileData[Index: Integer]: String read GetFileData;
    property  CIDs[Index: Integer]: String read GetCID;
    property  Used[Index: Integer]: Longbool read GetUsed write SetUsed;
    property  MaxSingleEntrySize: Integer read FMaxSingleEntrySize write FMaxSingleEntrySize;
    property  ModifiedMainSection: String read FModifiedMainSection write FModifiedMainSection;
  end;


// TCIDList ----------------------------------------------------------------------------------------

  PCIDItem = ^TCIDItem;
  TCIDItem = Record
    CIDStr: String;
    SectionBeg: Integer;
    SectionEnd: Integer;
  end;

  TCIDList = Class(TList)
  private
    FFileSections: TFileSections;
    FReturnRtf: Longbool;
    FSource: String;
    procedure CreateHtmlVersion;
    procedure CreateRtfVersion;
    function  GetCIDString(Index: Integer): String;
    function  GetSectionBeg(Index: Integer): Integer;
    function  GetSectionEnd(Index: Integer): Integer;
    function  GetHtmlVersion: String;
  protected
  public
    constructor Create(HtmlNode: TSectionNode; ReturnRtf: Longbool);
    destructor Destroy; override;
    function  AddItem(const ItemCID: String; SecBeg, SecEnd: Integer): Integer;
    {$IFDEF VER90}
    procedure Clear;
    {$ELSE}
    procedure Clear; override;
    {$ENDIF}
    procedure Delete(Index: Integer);
    property  CIDString[Index: Integer]: String read GetCIDString;
    property  SectionBeg[Index: Integer]: Integer read GetSectionBeg;
    property  SectionEnd[Index: Integer]: Integer read GetSectionEnd;
  end;


// TMimeMsgTree ----------------------------------------------------------------------------------------

  TOnEncodedData = function(const Data: String; var Charset, SignatureText: String;
			    var SignatureFlags: Integer): String of Object;

  TPSMimeSignatureResult = Record
    Text: String;
    Flags: Integer;
  end;

  TMsgTreeData = Record
    DefaultFontName: String;
    DefaultMailPath: String;
    DateHeaderName: String;
    SenderHeaderName: String;
    RecipientHeaderName: String;
    SubjectHeaderName: String;
    ReplyToHeaderName: String;
    MissingHeaderText: String;
    MailMessageName: String;
    HeaderSectionName: String;
    TextSectionName: String;
    UnknownSectionName: String;
    SingleCharExtensions: String;
    DoubleCharExtensions: String;
    TripleCharExtensions: String;
    OnEncodedData: TOnEncodedData;
  end;

  TMimeMsgTree = class(TTreeViewEx)
  private
    FDefaultCharset: String;
    FBodyEncoding: TEncoding;
    FFileAttachments: Integer;
    FFirstTextSection: TSectionNode;
    FGetRawData: Longbool;
    FHasAttachments: Longbool;
    FIsMessageDigest: Longbool;
    FSubMimeDecoding: Longbool;
    FUnencodedMimeData: Longbool;
    FPSMimeBounds: TBounds;
    FMaxLineLength: Integer;
    FMessageText: TTextList;
    FMsgTreeData: TMsgTreeData;
    FFindHeaderOptions: TFindHeaderOptions;
    FMimeDecodeOptions: TMimeDecodeOptions;
    procedure CreateTreeView;
    function  Get32BitImages: T32BitImages;
    procedure Set32BitImages(Value: T32BitImages);
    function  GetContainsData: Longbool;
    function  GetCount: Integer;
    function  GetAnsiHeaders: String;
    function  GetMainHeaders: String;
    function  GetMessageData: String;
    function  GetUTCDateTime: TDateTime;
    function  GetCharset(Index: Integer): String;
    function  GetFileName(Index: Integer): String;
    function  GetFirstTextIsFormatted: Longbool;
    function  GetSectionNode(Index: Integer): TSectionNode;
    function  GetSectionData(Index: Integer): String;
    procedure SetSectionData(Index: Integer; Value: String);
    function  GetSectionName(Index: Integer): String;
    procedure SetSectionName(Index: Integer; Value: String);
    function  GetSectionType(Index: Integer): TMsgDataType;
    procedure SetSectionType(Index: Integer; Value: TMsgDataType);
    function  GetSelection: TSectionNode;
    procedure SetSelection(const Value: TSectionNode);
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    property  FDefaultFontName: String read FMsgTreeData.DefaultFontName;
    property  FDateHeaderName: String read FMsgTreeData.DateHeaderName;
    property  FSenderHeaderName: String read FMsgTreeData.SenderHeaderName;
    property  FRecipientHeaderName: String read FMsgTreeData.RecipientHeaderName;
    property  FSubjectHeaderName: String read FMsgTreeData.SubjectHeaderName;
    property  FReplyToHeaderName: String read FMsgTreeData.ReplyToHeaderName;
    property  FMissingHeaderText: String read FMsgTreeData.MissingHeaderText;
    property  FMailMessageName: String read FMsgTreeData.MailMessageName;
    property  FHeaderSectionName: String read FMsgTreeData.HeaderSectionName;
    property  FTextSectionName: String read FMsgTreeData.TextSectionName;
    property  FUnknownSectionName: String read FMsgTreeData.UnknownSectionName;
    property  FSingleCharExtensions: String read FMsgTreeData.SingleCharExtensions;
    property  FDoubleCharExtensions: String read FMsgTreeData.DoubleCharExtensions;
    property  FTripleCharExtensions: String read FMsgTreeData.TripleCharExtensions;
    property  FOnEncodedData: TOnEncodedData read FMsgTreeData.OnEncodedData;
  protected
    function  CreateNode: TTreeNode; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure AddIcons;
    procedure Clear;
    function  AddSection(Parent: TSectionNode; const SecName, SecData: String; MDT: TMsgDataType): TSectionNode;
    function  InsertSectionAfter(Section: TSectionNode; const Name, Data: String; MDT: TMsgDataType): TSectionNode;
    function  FindPSEncodedSection(Encoding: TEncoding): TSectionNode;
    function  GetAttachmentList: String;
    function  GetMessageText(IncludeImages: Longbool): TTextList;
    function  GetPSMimeSignatureResult: TPSMimeSignatureResult;
    function  IsPSMimeSignedSection(Section: TSectionNode): Longbool;
    function  IsUnknownSectionText(Index: Integer): Longbool;
    function  RecreateMimeHeaders(const Headers: String; DateTime: TDateTime): String;
    procedure MimeParse(const MailMessage: String; BegPos, EndPos: Integer; MDO: TMimeDecodeOptions);
    function  SingleCharExtensionMatch(const FileExt: String): Longbool;
    function  DoubleCharExtensionMatch(const FileExt: String): Longbool;
    function  TripleCharExtensionMatch(const FileExt: String): Longbool;
    property  ContainsData: Longbool read GetContainsData;
    property  Count: Integer read GetCount;
    property  DefaultMailPath: String read FMsgTreeData.DefaultMailPath;
    property  Images: T32BitImages read Get32BitImages write Set32BitImages;
    property  DefaultCharset: String read FDefaultCharset write FDefaultCharset;
    property  BodyEncoding: TEncoding read FBodyEncoding write FBodyEncoding;
    property  MimeDecodeOptions: TMimeDecodeOptions read FMimeDecodeOptions write FMimeDecodeOptions;
    property  FirstTextIsFormatted: Longbool read GetFirstTextIsFormatted;
    property  FirstTextSection: TSectionNode read FFirstTextSection;
    property  HasAttachments: Longbool read FHasAttachments;
    property  IsMessageDigest: Longbool read FIsMessageDigest;
    property  UnencodedMimeData: Longbool read FUnencodedMimeData;
    property  PSMimeBounds: TBounds read FPSMimeBounds;
    property  AnsiHeaders: String read GetAnsiHeaders;
    property  MainHeaders: String read GetMainHeaders;
    property  MessageData: String read GetMessageData;
    property  UTCDateTime: TDateTime read GetUTCDateTime;
    property  MaxLineLength: Integer read FMaxLineLength write FMaxLineLength;
    property  MsgTreeData: TMsgTreeData read FMsgTreeData write FMsgTreeData;
    property  Selected: TSectionNode read GetSelection write SetSelection;
    property  Charsets[Index: Integer]: String read GetCharset;
    property  FileNames[Index: Integer]: String read GetFileName;
    property  SectionNodes[Index: Integer]: TSectionNode read GetSectionNode; default;
    property  SectionData[Index: Integer]: String read GetSectionData write SetSectionData;
    property  SectionNames[Index: Integer]: String read GetSectionName write SetSectionName;
    property  SectionTypes[Index: Integer]: TMsgDataType read GetSectionType write SetSectionType;
  end;


// TDatedMessages ----------------------------------------------------------------------------------

TOnShowProgress = function(Value: Integer): Longbool of Object;

TDatedMessages = Class
private
  FAttachmentHint: String;
  FDigestHint: String;
  FDisplayEndMarker: String;
  FMDO: TMimeDecodeOptions;
  FMsgList: TDatedStrings;
  FMsgTree: TMimeMsgTree;
  FTotalBytes: Integer;
  FMsgTreeData: TMsgTreeData;
  FOnShowProgress: TOnShowProgress;
  function  CreateMimeHeaders(const MailData: String; DateTime: TDateTime): String;
  procedure SetAttachmentHint(const Value: String);
  procedure SetDigestHint(const Value: String);
protected
public
  constructor Create(MsgList: TDatedStrings);
  destructor Destroy; override;
  procedure ProcessFolder(const Data: String; Unix: Longbool);
  function  SaveAsUnixMailbox(const FileName: String): Longbool;
  property  AttachmentHint: String read FAttachmentHint write SetAttachmentHint;
  property  DigestHint: String read FDigestHint write SetDigestHint;
  property  TotalBytes: Integer read FTotalBytes;
  property  MsgTreeData: TMsgTreeData read FMsgTreeData write FMsgTreeData;
  property  OnShowProgress: TOnShowProgress read FOnShowProgress write FOnShowProgress;
end;


// Routines ----------------------------------------------------------------------------------------

function  IsBinary(const Data: String): Longbool;
function  IsExtended(const Data: String): Longbool;
function  Utf8Check(const Text: String): String;
function  Utf8ToAnsi(const Text: String): String;
function  AnsiToUtf8(const Text: String): String;
function  Utf8ToWide(const Text: String): String;
function  WideToUtf8(const Text: String): String;
function  AnsiToWide(const Text: String): String;
function  WideToAnsi(const Text: String): String;
function  StringToAnsi(const Text, Charset: String): String;
function  AnsiToString(const Text, Charset: String): String;
function  StringToUtf8(const Text, Charset: String): String;
function  Utf8ToString(const Text, Charset: String): String;
function  ContentCheck(const Text, RawData, Charset: String; const Font: TPMPGPFont): String;
function  B64Convert(const Text: String): String;
function  B64Encode(const Text: String): String;
function  QPConvert(const Text: String; IsHeader: Longbool): String;
function  QPEncode(const Text, Charset: String): String;
function  StripExcessCRs(const Text: String): String;
function  Concat(const Strings: Array of String): String;
function  Abbreviate(const Text: String; MaxChars: Integer): String;
function  GetAddressLength(const Addresses: String; Pos: Integer): Integer;
function  GetArmoredData(const MailMessage: String; var MD: TMessageData): String;
function  GetPSMimeBoundary(const MimeHeaders: THeaderData; GetBoundaryStr: Boolean; var Boundary: String): TEncoding;
function  GetEncoding(const MailMessage: String; var EncodedData: THeaderData): TEncoding;
function  IsEncoded(const Encoding: TEncoding; KeysOnly: Longbool): Longbool;
function  StripSignatureData(const Text: String): String;
function  UnwrapArmoredKeys(const Text: String): String;
function  MimeDecodeHeaders(const Headers: String): String;
function  IsPMMessage(const FileName, PMMsgExtensions: String): Longbool;
function  IsPMAttachment(const FileName, MailMessage, PMTempExt: String): String;
function  GetTabbedPMAddressList(const Addresses: TAddressStrings; var AddressCounts: TAddressCounts;
				 var TabbedAddresses: String): Integer;
function  AddPMFlags(const CNMFile: String; PMScanLines: Longint; SMimeCapable, PSMimeDecode: Longbool;
		     var Subject: String; var SaveDecoded: Longbool): String;
function  RemovePMCryptoHeaders(var MailMessage: String; const Charset: String; QPEncodeText: Longbool): String;
function  PreparePSMimeEncoding(const MailMessage, Charset: String; EncodeFlags: Integer): THeaderData;
function  ExtractHeaderAddresses(const Headers, BCCRecipients: String; var RecipientCommaList: String): Integer;


implementation

const
  CID_MARK = 'cid:';

const
  WSP = [TAB, SP];
  LFCR = [LF, CR];
  WSPEQ = [TAB, SP, EQ];
  NLFCR = [NULL, LF, CR];
  NLFCRSC = [NULL, LF, CR, SC];


// Routines ----------------------------------------------------------------------------------------

function IsBinary(const Data: String): Longbool;
var
  iPos: Integer;
begin
  Result := false;
  if Data <> E then begin
    for iPos := 1 to Length(Data) do begin
      Result := (Data[iPos] in CHARS_BINARY);
      if Result then Break;
    end;
  end;
end;

function IsExtended(const Data: String): Longbool;
var
  iPos: Integer;
begin
  Result := false;
  if Data <> E then begin
    for iPos := 1 to Length(Data) do begin
      Result := (Data[iPos] in CHARS_EXTENDED);
      if Result then Break;
    end;
  end;
end;

function Utf8Check(const Text: String): String;
begin
  Result := Utf8OrAnsi(Text);
end;

function Utf8ToAnsi(const Text: String): String;
begin
  Result := Utf8ToAnsiString(Text);
  if Result = E then Result := Text;
end;

function AnsiToUtf8(const Text: String): String;
begin
  Result := AnsiStringToUtf8(Text);
end;

function Utf8ToWide(const Text: String): String;
begin
  Result := Utf8ToWideString(Text);
end;

function WideToUtf8(const Text: String): String;
begin
  Result := WideStringToUtf8(Text);
end;

function AnsiToWide(const Text: String): String;
begin
  Result := AnsiToWideString(Text);
end;

function WideToAnsi(const Text: String): String;
begin
  Result := WideStringToAnsi(Text);
end;

function StringToAnsi(const Text, Charset: String): String;
begin
  Result := AnyStringToAnsi(Text, Charset);
  if Result = E then Result := Text;
end;

function AnsiToString(const Text, Charset: String): String;
begin
  Result := AnsiToAnyString(Text, Charset);
  if Result = E then Result := Text;
end;

function StringToUtf8(const Text, Charset: String): String;
begin
  Result := AnyStringToUtf8(Text, Charset);
  if Result = E then Result := AnsiStringToUtf8(Text);
end;

function Utf8ToString(const Text, Charset: String): String;
begin
  Result := Utf8ToAnyString(Text, Charset);
  if Result = E then Result := AnsiToAnyString(Text, Charset);
end;

function DateTimeToMimeUTC(DateTime: TDateTime): String;
const
  LCID_US = SORT_DEFAULT shl 16 or $0409;
  DATE_FORMAT = 'ddd, d MMM yyyy';
  TIME_FORMAT = 'HH:mm:ss';
var
  stDateTime: TSystemTime;
  iStrLen: Integer;
  pResult: PChar;
begin
  stDateTime := DateTimeToSystemTime(DateTime);
  iStrLen := pred(GetDateFormat(LCID_US, 0, @stDateTime, DATE_FORMAT + SP, nil, 0) + Length(TIME_FORMAT));
  SetLength(Result, iStrLen);
  pResult := Pointer(Result);
  iStrLen := pred(GetDateFormat(LCID_US, 0, @stDateTime, DATE_FORMAT + SP, pResult, iStrLen));
  GetTimeFormat(LCID_US, TIME_FORCE24HOURFORMAT, @stDateTime, TIME_FORMAT, @pResult[iStrLen], Length(TIME_FORMAT));
end;

function ContentCheck(const Text, RawData, Charset: String; const Font: TPMPGPFont): String;
var
  RtfFont: TRtfFont;
  iWidth: Integer;
  sHtml: String;
begin
  if (Text <> E) and (RawData <> E) and (RawData[1] = LT) then begin
    if Charset <> E then
      Result := Utf8ToString(Text, Charset)
    else Result := Utf8Check(Text);
    with RtfFont do begin
      Name := Font.Name;
      Size := Font.Size;
      Styles := Font.Styles;
    end;
    Result := RichEd20.TextToRtf(Result, RtfFont, false);
    if Result <> E then begin
      iWidth := 0;
      sHtml := HtsTer.RtfToHtml(Result, 0, iWidth);
      if sHtml <> E then begin
	Result := TidyHtml(sHtml, E, false, true, false).Body;
	if Result <> E then Exit;
      end
      else Result := E;
    end;
  end;
  if Charset <> E then
    Result := Utf8ToString(Text, Charset)
  else Result := Utf8Check(Text);
end;

function B64Convert(const Text: String): String;
begin
  Result := MimeDecodeString(Trim(Text));
end;

function B64Encode(const Text: String): String;
begin
  Result := MimeEncodeString(Text);
end;

function NoQuotedPair(const Text: String; Pos: Integer): Longbool;
begin
  Result := (Pos = 1) or ((Pos > 1) and (Text[Pos - 1] <> BS)) or ((Pos > 2) and (Text[Pos - 2] = BS));
end;

function QPConvert(const Text: String; IsHeader: Longbool): String;
var
  iText, iBeg, iEnd, iLen, iPos: Integer;
  pResult: PChar;
begin
  iText := Length(Text);
  if iText > 0 then begin
    SetString(Result, nil, iText);
    pResult := Pointer(Result);
    iBeg := 1;
    iEnd := 1;
    while iEnd <= iText do begin
      if Text[iEnd] = EQ then begin
	iLen := iEnd - iBeg;
	Move(Text[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	iPos := succ(iEnd);
	if iPos < iText then begin
	  if PDChar(@Text[iPos])^ <> CRLF then begin
	    if not (Text[iPos] in LFCR) then begin
	      pResult^ := HexToChar(@Text[iPos]);
	      if pResult^ = NULL then begin
		iBeg := iEnd;
		inc(iEnd);
		Continue;
	      end
	      else iLen := 1;
	      inc(pResult, iLen);
	    end
	    else dec(iEnd);
	  end;
	  iBeg := iEnd + 3;
	  iEnd := iBeg;
	end
	else begin
	  iBeg := succ(iText);
	  Break;
	end;
      end
      else if IsHeader and (Text[iEnd] in [TAB, UL]) then begin
	iLen := iEnd - iBeg;
	Move(Text[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	pResult^ := SP;
	inc(pResult);
	iBeg := succ(iEnd);
	iEnd := iBeg;
      end
      else inc(iEnd);
    end;
    iLen := succ(iText - iBeg);
    if iLen > 0 then begin
      Move(Text[iBeg], pResult^, iLen);
      inc(pResult, iLen);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function CharsetCodeLen(const Charset: String; QP: Longbool): Integer;
begin
  if QP then
    Result := Length(CODE_BEG) + Length(Charset) + Length(CODE_QP)
  else Result := Length(CODE_BEG) + Length(Charset) + Length(CODE_B64);
end;

function MaxQPEncodeLen(TextLen: Integer; const Charset: String): Integer;
var
  iCodeLen: Integer;
begin
  if Charset <> E then begin
    iCodeLen := CharsetCodeLen(Charset, true) + Length(CODE_END) + Length(SP);
    Result := TextLen + (succ(TextLen div (MAX_QP_LINE_LEN - iCodeLen)) * iCodeLen) * 2;
  end
  else Result := TextLen + succ(TextLen div MAX_QP_LINE_LEN) * 2;
end;

function QPEncode(const Text, Charset: String): String;

  function NewEncodedWord(var pResult: PChar; const Charset: String): Integer;
  var
    iLen: Integer;
    cLast: Char;
  begin
    Result := Integer(pResult);
    PDChar(pResult)^ := CODE_BEG;
    inc(pResult, Length(CODE_BEG));
    iLen := Length(Charset);
    Move(Pointer(Charset)^, pResult^, iLen);
    inc(pResult, iLen);
    cLast := pResult[Length(CODE_QP)];
    PQChar(pResult)^ := CODE_QP;
    inc(pResult, Length(CODE_QP));
    pResult^ := cLast;
    Result := Integer(pResult) - Result;
  end;

  procedure FoldEncodedWords(const Charset: String; pLastSpace: PChar; var pResult: PChar; var iLinePos: Integer);
  const
    CE_CRLF = CODE_END + CRLF;
  var
    iLen: Integer;
  begin
    inc(pLastSpace);
    iLen := pResult - pLastSpace;
    pResult := Pointer(Longint(pLastSpace) + Length(CE_CRLF) + Length(SP) + CharsetCodeLen(Charset, true));
    if iLen > 0 then begin
      Move(pLastSpace^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    PQChar(pLastSpace)^ := CE_CRLF;
    inc(pLastSpace, Length(CE_CRLF));
    pLastSpace^ := SP;
    inc(pLastSpace);
    iLinePos := succ(NewEncodedWord(pLastSpace, Charset));
  end;

  procedure FoldAfterLastSpace(pLastSpace: PChar; var pResult: PChar; var iLinePos: Integer);
  const
    EQCRLF = EQ + CRLF;
  var
    iLen: Integer;
  begin
    inc(pLastSpace);
    iLen := pResult - pLastSpace;
    if iLen < iLinePos then begin
      pResult := Pointer(Longint(pLastSpace) + Length(EQCRLF));
      Move(pLastSpace^, pResult^, iLen);
      inc(pResult, iLen);
      pLastSpace^ := EQ;
      inc(pLastSpace);
      PDChar(pLastSpace)^ := CRLF;
    end
    else begin
      PQChar(pResult)^ := EQCRLF;
      inc(pResult, Length(EQCRLF));
    end;
    iLinePos := 0;
  end;

var
  iTextLen, iLinePos, iCharPos: Integer;
  pResult, pLastSpace: PChar;
  cPos: Char;
begin
  iTextLen := Length(Text);
  if iTextLen <> 0 then begin
    iTextLen := iTextLen * 3;
    SetString(Result, nil, MaxQPEncodeLen(iTextLen, Charset));
    pResult := Pointer(Result);
    if Charset <> E then
      iLinePos := succ(NewEncodedWord(pResult, Charset))
    else iLinePos := 0;
    pLastSpace := nil;
    for iCharPos := 1 to (iTextLen div 3) do begin
      cPos := Text[iCharPos];
      if (Charset <> E) and (iLinePos >= MAX_QP_LINE_LEN - 14) then begin
	FoldEncodedWords(Charset, pLastSpace, pResult, iLinePos);
      end;
      if (cPos in CHARS_TO_ENCODE)
      or ((Charset <> E) and (cPos in [LB, RB, CM]))
      or ((Text[succ(iCharPos)] in [CR, NULL]) and (cPos in WSP)) then begin
	if (iLinePos >= MAX_QP_LINE_LEN - 3) and (Text[succ(iCharPos)] <> CR) then begin
	  FoldAfterLastSpace(pLastSpace, pResult, iLinePos);
	end;
	if (Charset <> E) and (cPos = SP) then begin
	  pResult^ := UL;
	  inc(pResult);
	end
	else begin
	  pResult^ := EQ;
	  inc(pResult);
	  ByteToHex(ord(cPos), pResult);
	  inc(pResult, 2);
	  inc(iLinePos, 3);
	end;
      end
      else begin
	if (iLinePos >= MAX_QP_LINE_LEN - 1) and (Text[succ(iCharPos)] <> CR) then begin
	  FoldAfterLastSpace(pLastSpace, pResult, iLinePos);
	end;
	if cPos in [TAB, SP, SL, BS] then begin
	  if (Charset <> E) and (cPos = SP) then
	    pResult^ := UL
	  else pResult^ := cPos;
	  pLastSpace := pResult;
	end
	else pResult^ := cPos;
	inc(pResult);
	if cPos in LFCR then
	  iLinePos := 0
	else inc(iLinePos);
      end;
    end;
    if Charset <> E then begin
      PDChar(pResult)^ := CODE_END;
      inc(pResult, Length(CODE_END));
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function QPEncodeAddressHeader(const Text, Charset: String): String;
var
  iPos, iLen, iEnd, iBeg, iCopy, iQPos: Integer;
  sToEncode, sEncoded: String;
  pToEncode, pResult: PChar;
begin
  if Text <> E then begin
    iPos := 1;
    iLen := Length(Text);
    SetString(sToEncode, nil, MaxQPEncodeLen(iLen, Charset));
    SetString(Result, nil, Length(sToEncode));
    pToEncode := Pointer(sToEncode);
    pResult := Pointer(Result);
    iEnd := 1;
    repeat
      while (iPos <= iLen) and (Text[iPos] < #128) do inc(iPos);
      if iPos <= iLen then begin
	iBeg := iPos;
	repeat
	  dec(iBeg);
	  while (iBeg > 0) and not (Text[iBeg] in [LB, GT]) do dec(iBeg);
	until (iBeg = 0) or (Text[iBeg] = GT) or NoQuotedPair(Text, iBeg);
	if (Text[iBeg] = GT) and (Text[succ(iBeg)] = SP) then inc(iBeg);
	inc(iBeg);
	iCopy := iBeg - iEnd;
	if iCopy > 0 then begin
	  Move(Text[iEnd], pResult^, iCopy);
	  inc(pResult, iCopy);
	end;
	iEnd := iPos;
	repeat
	  inc(iEnd);
	  while (iEnd <= iLen) and not (Text[iEnd] in [RB, LT]) do inc(iEnd);
	until (iEnd > iLen) or (Text[iEnd] = LT) or NoQuotedPair(Text, iEnd);
	if (Text[iEnd] = LT) and (Text[pred(iEnd)] = SP) then dec(iEnd);
	if iEnd > iBeg then begin
	  repeat
	    iQPos := pred(iBeg);
	    repeat
	      inc(iQPos);
	      while (iQPos < iEnd) and (Text[iQPos] <> QU) do inc(iQPos);
	    until (iQPos = iEnd) or NoQuotedPair(Text, iQPos);
	    if iQPos < iEnd then
	      iCopy := iQPos - iBeg
	    else iCopy := iEnd - iBeg;
	    Move(Text[iBeg], pToEncode^, iCopy);
	    inc(pToEncode, iCopy);
	    iBeg := succ(iQPos);
	  until iQPos = iEnd;
	  iCopy := pToEncode - Pointer(sToEncode);
	  if iCopy > 0 then begin
	    sEncoded := QPEncode(Copy(sToEncode, 1, iCopy), Charset);
	    pToEncode := Pointer(sToEncode);
	    iCopy := Length(sEncoded);
	    Move(Pointer(sEncoded)^, pResult^, iCopy);
	    inc(pResult, iCopy);
	  end;
	end;
	iPos := iEnd;
      end
      else begin
	iCopy := succ(iLen - iEnd);
	if iCopy > 0 then begin
	  Move(Text[iEnd], pResult^, iCopy);
	  inc(pResult, iCopy);
	  Break;
	end;
      end;
    until iPos > iLen;
    iLen := pResult - Pointer(Result);
    if (iLen > MAX_QP_LINE_LEN - 10) and (FirstStrPos(CRLF, Result) = 0) then begin
      iPos := FirstCharPos(SP, Result);
      if Length(Result) >= iLen + Length(CRLF) then begin
	dec(iPos);
	pResult := Pointer(Result);
	inc(pResult, iPos);
	Move(pResult[0], pResult[2], iLen - iPos);
	PDChar(pResult)^ := CRLF;
	iPos := iLen + succ(Length(CRLF));
	Delete(Result, iPos, MAXINT);
      end
      else begin
        Delete(Result, succ(iLen), MAXINT);
	Insert(CRLF, Result, iPos);
      end;
    end
    else Delete(Result, succ(iLen), MAXINT);
  end
  else Result := E;
end;

function StripExcessCRs(const Text: String): String;
var
  iText, iBeg, iEnd, iLen: Integer;
  pResult: PChar;
begin
  iText := Length(Text);
  if iText > 0 then begin
    SetString(Result, nil, iText);
    pResult := Pointer(Result);
    iBeg := 1;
    repeat
      iEnd := NextStrPos(CR + CRLF, Text, iBeg);
      if iEnd >= iBeg then begin
	iLen := iEnd - iBeg;
	while (iLen > 1) and (Text[pred(iBeg + iLen)] = CR) do dec(iLen);
	Move(Text[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	PDChar(pResult)^ := CRLF;
	inc(pResult, Length(CRLF));
	iBeg := iEnd + 3;
      end
      else if iBeg <= iText then begin
	if iBeg > 1 then begin
	  iLen := succ(iText - iBeg);
	  Move(Text[iBeg], pResult^, iLen);
	  inc(pResult, iLen);
	  Break;
	end
	else begin
	  Result := Text;
	  Exit;
	end;
      end;
    until iEnd = 0;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function Concat(const Strings: Array of String): String;
var
  iResult, iLength, iIndex: Integer;
  fUnicode: Longbool;
  pResult: PChar;
begin
  if High(Strings) > 0 then begin
    iResult := 0;
    fUnicode := false;
    for iIndex := 0 to High(Strings) do begin
      iLength := Length(Strings[iIndex]);
      if iLength > 0 then begin
	fUnicode := (Strings[iIndex][iLength] = NULL);
	inc(iResult, iLength);
      end;
    end;
    SetString(Result, nil, iResult);
    pResult := Pointer(Result);
    if fUnicode then begin
      for iIndex := 0 to High(Strings) do begin
	iResult := Length(Strings[iIndex]) - Length(wNUL);
	if iResult > 0 then begin
	  Move(Pointer(Strings[iIndex])^, pResult^, iResult);
	  inc(pResult, iResult);
	end;
      end;
      iResult := succ(pResult - Pointer(Result));
      if iResult > 0 then begin
	Delete(Result, iResult + Length(wNUL), MAXINT);
	PDChar(pResult)^ := wNUL;
      end
      else Result := E;
    end
    else for iIndex := 0 to High(Strings) do begin
      iResult := Length(Strings[iIndex]);
      if iResult > 0 then begin
	Move(Pointer(Strings[iIndex])^, pResult^, iResult);
	inc(pResult, iResult);
      end;
    end;
  end
  else Result := Strings[0];
end;

function Abbreviate(const Text: String; MaxChars: Integer): String;
const
  ELLIPSIS_W = #$20#$00#$26#$20;
  ELLIPSIS_A = #$20#$85;
var
  iTextChars: Integer;
  pResult: PChar;
begin
  iTextChars := Length(Text);
  if (iTextChars > 0) and (MaxChars > 3) then begin
    if Text[iTextChars] = NULL then begin
      iTextChars := pred(iTextChars shr 1);
      if iTextChars > MaxChars then begin
	MaxChars := succ(MaxChars) shl 1;
	SetString(Result, nil, MaxChars);
	pResult := Pointer(Result);
	dec(MaxChars, Length(ELLIPSIS_W) + 2);
	Move(Pointer(Text)^, pResult^, MaxChars);
	inc(pResult, MaxChars);
	PQChar(pResult)^ := ELLIPSIS_W;
	inc(pResult, Length(ELLIPSIS_W));
	PDChar(pResult)^ := NULL;
      end
      else Result := Text;
    end
    else begin
      if iTextChars > MaxChars then begin
	SetString(Result, nil, MaxChars);
	pResult := Pointer(Result);
	dec(MaxChars, Length(ELLIPSIS_A));
	Move(Pointer(Text)^, pResult^, MaxChars);
	inc(pResult, MaxChars);
	PDChar(pResult)^ := ELLIPSIS_A;
      end
      else Result := Text;
    end;
  end
  else Result := E;
end;

const
  A_TAG = 'a';
  IMG_TAG = 'img';
  A_HREF_TAG = 'a href=';

function ExtractImageName(const Input, Base, Font: String): String;

  function GetDimValue(const Dimension: String): Integer;
  const
    DIGITS = ['0'..'9'];
  var
    sDigits: String;
    pDigits: PChar;
    iPos, iLen: Integer;
  begin
    iLen := Length(Dimension);
    if iLen > 0 then begin
      SetString(sDigits, nil, iLen);
      pDigits := Pointer(sDigits);
      iPos := 1;
      while (iPos <= iLen) and not (Dimension[iPos] in DIGITS) do inc(iPos);
      while (iPos <= iLen) and (Dimension[iPos] in DIGITS) do begin
	pDigits^ := Dimension[iPos];
	inc(pDigits);
	inc(iPos);
      end;
      Delete(sDigits, succ(pDigits - Pointer(sDigits)), MAXINT);
    end;
    Result := StrToIntDef(sDigits, 0);
  end;

const
  SRC_ATTR = 'src=';
  WIDTH_ATTR = 'width=';
  HEIGHT_ATTR = 'height=';
var
  iLen, iBegPos, iEndPos, iPos: Integer;
  iWidthBeg, iWidthEnd, iWidthLen: Integer;
  iHeightBeg, iHeightEnd, iHeightLen: Integer;
  sName, sFontBeg, sFontEnd, sWidth, sHeight, sDimensions: String;
begin
  iBegPos := FirstStrPos(SRC_ATTR, Input);
  if iBegPos > 0 then begin
    iBegPos := NextCharPos(QU, Input, iBegPos + Length(SRC_ATTR));
    inc(iBegPos);
    iEndPos := NextCharPos(QU, Input, iBegPos);
    iLen := iEndPos - iBegPos;
    if (iLen > 0) and not LoStrAtPos('cid:', Input, iBegPos) then begin
      iPos := PrevCharPos(SL, Input, pred(iEndPos));
      if iPos > 0 then begin
	iBegPos := succ(iPos);
	iLen := iEndPos - iBegPos;
      end;
    end;
  end
  else iLen := 0;
  if iLen > 0 then begin
    iWidthBeg :=  FirstStrPos(WIDTH_ATTR, Input);
    if iWidthBeg > 0 then begin
      iWidthEnd := NextCharPos(QU, Input, succ(iWidthBeg + Length(WIDTH_ATTR)));
      iWidthLen := succ(iWidthEnd - iWidthBeg);
    end
    else iWidthLen := 0;
    iHeightBeg := FirstStrPos(HEIGHT_ATTR, Input);
    if iHeightBeg > 0 then begin
      iHeightEnd := NextCharPos(QU, Input, succ(iHeightBeg + Length(HEIGHT_ATTR)));
      iHeightLen := succ(iHeightEnd - iHeightBeg);
    end
    else iHeightLen := 0;
    sName := Copy(Input, iBegPos, iLen);
    if (sName <> E) and (sName[1] = SL) then sName := Base + sName;
    if Font <> E then begin
      sFontBeg := '<font face="' + Font + '">';
      sFontEnd := '</font>';
    end;
    sWidth := Copy(Input, iWidthBeg, iWidthLen);
    sHeight := Copy(Input, iHeightBeg, iHeightLen);
    if (sWidth <> E ) and (sHeight <> E) then sDimensions := SP + sWidth + SP + sHeight;
    if (GetDimValue(sWidth) <> 1) and (GetDimValue(sHeight) <> 1) then begin
      Result := '<table border="0" bgcolor="#DDDDDD"><tr><td' + sDimensions + ' align="center" valign="middle">' +
		sFontBeg + '[' + Unescape(sName, false, PC) + ']' + sFontEnd +
		'</td></tr></table>';
      Exit;
    end;
  end;
  Result := E;
end;

function ExtractHyperlink(const Input, Base: String; BegPos, EndPos: Integer): String;

  function GetTrailingTags(const Input: String; BegPos, EndPos: Integer): String;
  var
    iTagBeg, iTagEnd: Integer;
  begin
    iTagBeg := 0;
    iTagEnd := EndPos;
    repeat
      while (iTagEnd > BegPos) and (Input[iTagEnd] <= SP) do dec(iTagEnd);
      if (iTagEnd > BegPos) and (Input[iTagEnd] = GT) then begin
	iTagBeg := PrevCharPos(LT, Input, iTagEnd);
	iTagEnd := pred(iTagBeg);
      end
      else Break;
    until iTagBeg <= BegPos;
    if iTagBeg > BegPos then
      Result := Copy(Input, iTagBeg, succ(EndPos - iTagBeg))
    else Result := E;
  end;

var
  iBase, iBegPos, iEndPos, iLen: Integer;
  pResult: PChar;
  sTags: String;
begin
  iBase := Length(Base);
  SetString(Result, nil, succ(EndPos - BegPos) + iBase);
  pResult := Pointer(Result);
  BegPos := succ(NextCharPos(QU, Input, BegPos + Length(A_HREF_TAG)));
  if Input[BegPos] <> '#' then begin
    iBegPos := succ(NextCharPos(GT, Input, BegPos));
    iEndPos := EndPos - Length(A_TAG) - 3;
    sTags := GetTrailingTags(Input, iBegPos, iEndPos);
    if sTags <> E then dec(iEndPos, Length(sTags));
    iLen := succ(iEndPos - iBegPos);
    Move(Input[iBegPos], pResult^, iLen);
    inc(pResult, iLen);
    pResult^ := SP;
    inc(pResult);
    PQChar(pResult)^ := '&lt;';
    inc(pResult, 4);
    EndPos := NextCharPos(QU, Input, BegPos);
    iLen := EndPos - BegPos;
    if iLen > 0 then begin
      if (Base <> E) and (Input[BegPos] = SL) then begin
	Move(Pointer(Base)^, pResult^, iBase);
	inc(pResult, iBase);
      end;
      Move(Input[BegPos], pResult^, iLen);
      inc(pResult, iLen);
    end;
    PQChar(pResult)^ := '&gt;';
    inc(pResult, 4);
    if sTags <> E then begin
      iLen := Length(sTags);
      Move(Pointer(sTags)^, pResult^, iLen);
      inc(pResult, iLen);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function AdjustLinks(const Input, Base, Font: String): String;
var
  iInput, iResult, iBeg, iEnd, iLen, iPos: Integer;
  pResult: PChar;
  sImage, sLink: String;
begin
  if Input <> E then begin
    iInput := Length(Input);
    iResult := iInput shl 1;
    SetString(Result, nil, iResult);
    pResult := Pointer(Result);
    iBeg := 1;
    iEnd := NextCharPos(LT, Input, iBeg);
    while (iEnd >= iBeg) and (iEnd < iInput) do begin
      // change "img" tags to "[image name]" entities
      if LoStrAtPos(IMG_TAG, Input, succ(iEnd)) then begin
	iLen := iEnd - iBeg;
	Move(Input[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	iBeg := succ(NextCharPos(GT, Input, iEnd + Length(IMG_TAG)));
	iLen := iBeg - iEnd;
	SetString(sImage, nil, iLen);
	Move(Input[iEnd], Pointer(sImage)^, iLen);
	sImage := ExtractImageName(sImage, Base, Font);
	iLen := succ(iInput - iBeg) + Length(sImage);
	iPos := pResult - Pointer(Result);
	if iPos + iLen > iResult then begin
	  inc(iResult, iLen);
	  SetLength(Result, iResult);
	  pResult := Pointer(Result);
	  inc(pResult, iPos);
	end;
	iLen := Length(sImage);
	Move(Pointer(sImage)^, pResult^, iLen);
	inc(pResult, iLen);
      end
      // change "a href ... a" to "name <link>" sections
      else if LoStrAtPos(A_HREF_TAG, Input, succ(iEnd)) then begin
	iLen := iEnd - iBeg;
	Move(Input[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	iBeg := NextLoStrPos(LT + SL + A_TAG + GT, Input, iEnd + Length(A_HREF_TAG));
	if iBeg > 0 then begin
	  inc(iBeg, Length(A_TAG) + 3);
	  sLink := AdjustLinks(ExtractHyperlink(Input, Base, iEnd, pred(iBeg)), Base, Font);
	  iLen := succ(iInput - iBeg) + Length(sLink);
	  iPos := pResult - Pointer(Result);
	  if iPos + iLen > iResult then begin
	    inc(iResult, iLen);
	    SetLength(Result, iResult);
	    pResult := Pointer(Result);
	    inc(pResult, iPos);
	  end;
	  iLen := Length(sLink);
	  Move(Pointer(sLink)^, pResult^, iLen);
	  inc(pResult, iLen);
	end
	else iBeg := iEnd;
      end;
      if iBeg > iEnd then
	iEnd := NextCharPos(LT, Input, iBeg)
      else iEnd := NextCharPos(LT, Input, succ(iEnd));
    end;
    iLen := succ(iInput - iBeg);
    if iLen > 0 then begin
      Move(Input[iBeg], pResult^, iLen);
      inc(pResult, iLen);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := Input;
end;

function GetValidHtml(const Text, Font: String; RemoveLinksAndImages, ConvertUtf8: Longbool): String;
begin
  with TidyHtml(Text, Font, false, false, ConvertUtf8) do begin
    if not RemoveLinksAndImages then
      Result := Body
    else Result := AdjustLinks(Body, Base, Font);
    if ConvertUtf8 then Result := Utf8ToUnicode(Result);
  end;
end;

function ConvertFormat(const Text, DefaultFontName: String; MDO: TMimeDecodeOptions; MaxLineLength: Integer): String;
var
  iHeight: Integer;
  fReturnUnicode: Longbool;
begin
  Result := E;
  if Text <> E then begin
    iHeight := 0;
    fReturnUnicode := (mdoReturnUnicode in MDO);
    if not LoStrAtPos(RTF_MARKER, Text, 1) then begin
      Result := HtsTer.HtmlToRtf(Text, MaxLineLength, iHeight);
      if mdoReturnRtf in MDO then Exit;
    end
    else Result := Text;
    if not (mdoReturnRtf in MDO) then Result := RichEd20.RtfToText(Result, fReturnUnicode);
  end;
end;

function ReturnPlainAnsiText(const Input: String; TextSection: TSectionNode): String;
var
  mdo: TMimeDecodeOptions;
  mdt: TMsgDataType;
begin
  if Input <> E then begin
    mdo := [];
    if Input[Length(Input)] <> NULL then begin
      mdt := TextSection.SectionType;
      if (mdt = mdtHtmlText) and StrAtPos(RTF_MARKER, Input, 1) then mdt := mdtRtfText;
      case mdt of
	mdtHtmlText: with TextSection do begin
	  Result := ConvertFormat(GetValidHtml(Input, E, true, true), E, mdo, MimeMsgTree.MaxLineLength);
	  if Result = E then Result := Utf8ToAnsi(Input);
	end;
	mdtRtfText: begin
	  Result := ConvertFormat(Input, E, mdo, TextSection.MimeMsgTree.MaxLineLength);
	  if Result = E then Result := Input;
	end;
      else
	Result := Utf8orAnsi(Input);
      end;
    end
    else Result := WideToAnsi(Input);
  end
  else Result := E;
end;

function ReturnUnicodeText(const Input: String; TextSection: TSectionNode): String;
var
  mdo: TMimeDecodeOptions;
  mdt: TMsgDataType;
begin
  if Input <> E then begin
    mdo := [mdoReturnUnicode];
    if Input[Length(Input)] <> NULL then begin
      mdt := TextSection.SectionType;
      if (mdt = mdtHtmlText) and StrAtPos(RTF_MARKER, Input, 1) then mdt := mdtRtfText;
      case mdt of
	mdtHtmlText: with TextSection do begin
	  Result := ConvertFormat(GetValidHtml(Input, E, true, true), E, mdo, MimeMsgTree.MaxLineLength);
	  if Result = E then Result := Utf8ToWide(Input);
	end;
	mdtRtfText: begin
	  Result := ConvertFormat(Input, E, mdo, TextSection.MimeMsgTree.MaxLineLength);
	  if Result = E then Result := Input;
	end;
      else
	Result := Utf8OrWide(Input);
      end;
    end
    else Result := Input;
  end
  else Result := E;
end;

function ReturnFormattedText(const Input: String; TextSection: TSectionNode): String;
var
  mdo: TMimeDecodeOptions;
  mdt: TMsgDataType;
begin
  if Input <> E then begin
    mdo := [mdoReturnUnicode, mdoReturnRtf];
    if Input[Length(Input)] <> NULL then begin
      mdt := TextSection.SectionType;
      if (mdt = mdtHtmlText) and StrAtPos(RTF_MARKER, Input, 1) then mdt := mdtRtfText;
      case mdt of
	mdtHtmlText: with TextSection, MimeMsgTree do begin
	  Result := ConvertFormat(GetValidHtml(Input, FDefaultFontName, true, true), FDefaultFontName, mdo, MaxLineLength);
	  if Result = E then Result := Utf8ToWide(Input);
	end;
	mdtRtfText: with TextSection, MimeMsgTree do begin
	  Result := ConvertFormat(Input, FDefaultFontName, mdo, MaxLineLength);
	  if Result = E then Result := AnsiToWide(Input);
	end;
      else
	Result := Utf8OrWide(Input);
      end;
    end
    else Result := Input;
  end
  else Result := E;
end;

function ClearBlankLineStr(const Text: String): String;
var
  iPos: Integer;
begin
  Result := Text;
  if Result <> E then begin
    iPos := 1;
    while (Result[iPos] > NULL) and (Result[iPos] in LFCR) do inc(iPos);
    if Result[iPos] = NULL then Delete(Result, 1, MAXINT);
  end;
end;

function AppendBlankLine(const Text: String): String;
const
  uPARA = wPARA + NULL + NULL;
  uCRLF = wCRLF + NULL + NULL;
var
  iLen: Integer;
begin
  if Text <> E then begin
    iLen := Length(Text);
    if Text[iLen] = NULL then begin
      if StrAtPos(uPARA, Text, iLen - pred(Length(uPARA))) then
	Result := E
      else if StrAtPos(uCRLF, Text, iLen - pred(Length(uCRLF))) then
	Result := uCRLF
      else Result := uPARA;
    end
    else begin
      if StrAtPos(aPARA, Text, iLen - pred(Length(aPARA))) then
	Result := E
      else if StrAtPos(aCRLF, Text, iLen - pred(Length(aCRLF))) then
	Result := aCRLF
      else Result := aPARA;
    end;
  end
  else Result := E;
end;

procedure RemoveBackslashes(var HeaderText: String);
var
  pHeaderText: PChar;
  iLen, iIndex: Integer;
begin
  if HeaderText <> E then begin
    pHeaderText := Pointer(HeaderText);
    iLen := Length(HeaderText);
    iIndex := 1;
    while iIndex <= iLen do begin
      if HeaderText[iIndex] = BS then inc(iIndex);
      pHeaderText^ := HeaderText[iIndex];
      inc(pHeaderText);
      inc(iIndex);
    end;
    Delete(HeaderText, succ(pHeaderText - Pointer(HeaderText)), MAXINT);
  end;
end;

function GetAddressLength(const Addresses: String; Pos: Integer): Integer;
var
  iLen, iBeg, iEnd: Integer;
begin
  Result := 0;
  if Addresses = E then Exit;
  iLen := Length(Addresses);
  if Pos < 1 then Pos := 1;
  Result := Pos;
  repeat
    iBeg := 0;
    iEnd := 0;
    repeat
      while not (Addresses[Result] in [NULL, QU, LB, CM]) do inc(Result);
      if (Addresses[Result] in [NULL, CM]) then begin
	dec(Result, Pos);
	Exit;
      end
      else begin
	if NoQuotedPair(Addresses, Result) then iBeg := Result;
	if iBeg > 0 then begin
	  Result := succ(iBeg);
	  Break;
	end
	else inc(Result);
      end;
    until Result > iLen;
    repeat
      if Addresses[iBeg] = QU then
	while not (Addresses[Result] in [NULL, QU]) do inc(Result)
      else while not (Addresses[Result] in [NULL, RB]) do inc(Result);
      if (Addresses[Result] in [NULL]) then begin
	dec(Result, Pos);
	Exit;
      end
      else begin
	if NoQuotedPair(Addresses, Result) then iEnd := Result;
	if iEnd > 0 then begin
	  Result := succ(iEnd);
	  Break;
	end
	else inc(Result);
      end;
    until Result > iLen;
  until Result > iLen;
  dec(Result, Pos);
end;

function GetFirstAddress(const Header: String): String;
var
  iStartPos, iAddressLen, iResultLen: Integer;
  pResult: PChar;
begin
  Result := Header;
  if Result <> E then begin
    iStartPos := 1;
    iAddressLen := GetAddressLength(Result, iStartPos);
    if (iAddressLen > 0) and (iAddressLen < Length(Result)) then begin
      iResultLen := iAddressLen + Length(ET_AL);
      if iResultLen <= Length(Result) then
	System.Delete(Result, succ(iResultLen), MAXINT)
      else SetLength(Result, iResultLen);
      pResult := Pointer(Result);
      inc(pResult, iAddressLen);
      Move(ET_AL, pResult^, Length(ET_AL));
    end;
  end;
end;

function DateHeaderToUTC(const DateHeader: String; var DateTime: TDateTime): String;

  function AdjustShortDateTime(const DateTimeStr: String; var DateTime: TDateTime): String;
  var
    iBeg, iEnd, iBuffer, iLen: Integer;
    sTimeZone, sFormat: String;
    sDay, sMonth, sYear: String;
    sHour, sMin, sSec: String;
    dtBuffer: TDateTime;
  begin
    DateTime := 0;
    sTimeZone := DateTimeStr;
    sFormat := ShortDateFormat;
    if DateTimeStr <> E then begin
      try
	iBeg := 1;
	iEnd := FirstCharPos(SP, sTimeZone);
	sDay := Format('%2.2d', [StrToIntDef(Copy(sTimeZone, iBeg, iEnd - iBeg), 0)]);
	iBeg := succ(iEnd);
	iEnd := NextCharPos(SP, sTimeZone, iBeg);
	sMonth := Format('%2.2d', [succ(FirstLoStrPos(Copy(sTimeZone, iBeg, iEnd - iBeg), MONTHS) div 4)]);
	iBeg := succ(iEnd);
	iEnd := NextCharPos(SP, sTimeZone, iBeg);
	sYear := Copy(sTimeZone, iBeg, iEnd - iBeg);
	case Length(sYear) of
	  4: iBuffer := StrToIntDef(sYear, 0);
	  3: iBuffer := 1900 + StrToIntDef(sYear, 0);
	  2: begin
	    iBuffer := StrToIntDef(sYear, 0);
	    if iBuffer > 49 then
	      inc(iBuffer, 1900)
	    else inc(iBuffer, 2000);
	  end;
	else
	  iBuffer := 0;
	  EConvertError.Create(E);
	end;
	sYear := IntToStr(iBuffer);
	iBeg := succ(iEnd);
	iEnd := NextCharPos(CL, sTimeZone, iBeg);
	sHour := Copy(sTimeZone, iBeg, iEnd - iBeg);
	iBeg := succ(iEnd);
	iEnd := NextCharPos(CL, sTimeZone, iBeg);
	if iEnd = 0 then iEnd := NextCharPos(SP, sTimeZone, iBeg);
	sMin := Copy(sTimeZone, iBeg, iEnd - iBeg);
	iBeg := succ(iEnd);
	iEnd := NextCharPos(SP, sTimeZone, iBeg);
	sSec := Copy(sTimeZone, iBeg, iEnd - iBeg);
	if sSec = E then sSec := '00';
	if iEnd = 0 then iEnd := pred(iBeg);
	Delete(sTimeZone, 1, iEnd);
	iLen := Length(sTimeZone);
	if iLen > 3 then begin
	  dtBuffer := EncodeTime(StrToIntDef(Copy(sTimeZone, iLen - 3, 2), 0),
				 StrToIntDef(Copy(sTimeZone, pred(iLen), 2), 0), 0, 0);
	  if sTimeZone[1] = DA then dtBuffer := -dtBuffer;
	end
	else begin
	  iBuffer := FirstLoStrPos(sTimeZone, TZONES) div 4;
	  case iBuffer of
	    2..9: dtBuffer := -EncodeTime(4 + (pred(iBuffer) div 2), 0, 0, 0);
	  else
	    dtBuffer := 0;
	  end;
	end;
	try
	  ShortDateFormat := 'dd' + DateSeparator + 'mm' + DateSeparator + 'yyyy';
	  DateTime := StrToDateTime(sDay + DateSeparator + sMonth + DateSeparator + sYear + SP +
				    sHour + TimeSeparator + sMin + TimeSeparator + sSec + TimeSeparator) - dtBuffer;
	  ShortDateFormat := 'ddd' + CM + SP + 'd' + SP + 'mmm' + SP + 'yyyy';
	  DateTimeToString(Result, ShortDateFormat + CM + SP + LongTimeFormat, DateTime);
	finally
	  ShortDateFormat := sFormat;
	end;
	Result := Result + SP + UTC;
      except
	Result := E;
      end;
    end
    else Result := E;
  end;

var
 iLen, iBeg, iEnd: Integer;
begin
  Result := DateHeader;
  iLen := Length(Result);
  if iLen > 0 then begin
    if not (Result[1] in ['0'..'9']) then
      iEnd := succ(FirstCharPos(CM, Result))
    else iEnd := 1;
    if iEnd > 0 then begin
      while (iEnd < iLen) and (Result[iEnd] <= SP) do inc(iEnd);
      System.Delete(Result, 1, pred(iEnd));
      iEnd := 0;
    end;
    repeat
      iBeg := NextCharPos(LB, Result, succ(iEnd));
      if iBeg > 0 then begin
	iEnd := NextCharPos(RB, Result, succ(iBeg));
	if iEnd > 0 then System.Delete(Result, iBeg, succ(iEnd - iBeg))
      end;
    until iBeg = 0;
    Result := AdjustShortDateTime(Trim(Result), DateTime);
  end
  else Result := E;
end;

procedure RemoveComments(var HeaderText: String);
var
  iLPos, iRPos, iDelLen: Integer;
begin
  iLPos := 1;
  while HeaderText <> E do begin
    while not (HeaderText[iLPos] in [NULL, QU, LB]) do inc(iLPos);
    iRPos := succ(iLPos);
    case HeaderText[iLPos] of
      NULL: Break;
      QU: begin
	repeat inc(iLPos) until (HeaderText[iLPos] in [NULL, QU]);
	if HeaderText[iLPos] = QU then begin
	  inc(iLPos);
	  Continue;
	end
	else Break;
      end;
      LB: begin
	while not (HeaderText[iRPos] in [NULL, RB]) do inc(iRPos);
	if HeaderText[iRPos] = NULL then Break;
      end;
    end;
    iDelLen := succ(iRPos - iLPos);
    if iDelLen > 0 then Delete(HeaderText, iLPos, iDelLen);
  end;
end;

// Name may include the trailing colon
// Name may end with an asterisk as wildcard
// HeaderInfo may either contain data as returned by
// GetHeaderInfo with Pos & Len set to valid values
// or a simple text string with Pos or Len set to zero;
// the result returns the header starting with it's value
// (no name) with Pos and Len from inside the orginal headers
function FindHeaderLine(const Name: String; const HeaderInfo: THeaderData; DropComments: Longbool): THeaderData;
var
  iName, iText, iBeg, iEnd, iCopy: Integer;
  pHeader: PChar;
begin
  FillChar(Result, SizeOf(THeaderData), 0);
  Result.Recursion := HeaderInfo.Recursion;
  with HeaderInfo do begin
    if (Pos = 0) or (Len = 0) then begin
      iBeg := 1;
      iText := MaxPos;
      if iText = 0 then iText := Length(Text);
    end
    else begin
      iBeg := Pos;
      iText := pred(Pos + Len);
    end;
    iName := pred(Length(Name));
    repeat
      iBeg := FirstLoStrPartInRange(Name, Text, iName, iBeg, iText);
      if (iBeg <= 1) or (Text[pred(iBeg)] in LFCR) then
	Break
      else inc(iBeg, iName);
    until iBeg > iText;
    if iBeg > 0 then begin
      Result.Pos := iBeg;
      inc(iBeg, iName);
      if Name[succ(iName)] <> AST then begin
	while (iBeg <= iText) and (Text[iBeg] in WSP) do inc(iBeg);
	if Text[iBeg] <> CL then begin
	  Result.Pos := 0;
	  Exit;
	end
	else inc(iBeg);
      end
      else begin
	while (iBeg <= iText) and (Text[iBeg] <> CL) do inc(iBeg);
	if Text[iBeg] <> CL then begin
	  Result.Pos := 0;
	  Exit;
	end
	else inc(iBeg);
      end;
      while (iBeg <= iText) and (Text[iBeg] in WSP) do inc(iBeg);
      if (iBeg <= iText) and (Text[iBeg] in LFCR) then begin
	while (iBeg <= iText) and (Text[iBeg] in LFCR) do inc(iBeg);
	if (iBeg <= iText) and not (Text[iBeg] in WSP) then begin
	  Result.Pos := 0;
	  Exit;
	end
	else while (iBeg <= iText) and (Text[iBeg] in WSP) do inc(iBeg);
      end;
      SetString(Result.Text, nil, succ(iText - iBeg));
      pHeader := Pointer(Result.Text);
      repeat
	iEnd := succ(iBeg);
	while (iEnd <= iText) and not (Text[iEnd] in NLFCR) do inc(iEnd);
	if iEnd = iText then inc(iEnd);
	iCopy := iEnd - iBeg;
	if iCopy > 0 then begin
	  Move(Text[iBeg], pHeader^, iCopy);
	  inc(pHeader, iCopy);
	  iBeg := succ(iEnd);
	  while (iBeg <= iText) and (Text[iBeg] <= SP) do inc(iBeg);
	  if Text[pred(iBeg)] in WSP then begin
	    pHeader^ := SP;
	    inc(pHeader);
	  end
	  else Break;
	end;
      until iBeg >= iText;
      with Result do begin
	Len := iEnd - Pos;
	MaxPos := pHeader - Pointer(Text);
	Delete(Text, succ(MaxPos), MAXINT);
	if DropComments then RemoveComments(Text);
      end;
    end;
  end;
end;

// HeaderInfo may either contain data as returned by
// GetHeaderInfo with Pos & Len set to valid values
// or a simple text string with Pos or Len set to zero
function GetBoundary(const HeaderInfo: THeaderData): String;
var
  iLen, iPos: Integer;
begin
  if HeaderInfo.CTHeader = E then begin
    if HeaderInfo.Pos > 0 then
      Result := FindHeaderLine(CT_HEADER, HeaderInfo, true).Text
    else Result := HeaderInfo.Text;
  end
  else Result := HeaderInfo.CTHeader;
  iLen := Length(Result);
  if LoStrAtPos(CT_MULTIPART, Result, 1) then begin
    iPos := NextNonSpaceChar(Result, NextLoStrPos(CT_BOUNDARY, Result, succ(Length(CT_MULTIPART))) + Length(CT_BOUNDARY));
    if iPos > 0 then begin
      while (iPos < iLen) and (Result[iPos] in WSPEQ) do inc(iPos);
      if Result[iPos] = QU then begin
	Delete(Result, 1, iPos);
	Delete(Result, FirstCharPos(QU, Result), MAXINT);
      end
      else begin
	Delete(Result, 1, pred(iPos));
	Delete(Result, FirstCharPos(SC, Result), MAXINT);
      end;
      Exit;
    end;
  end;
  Result := E;
end;

// GetContentHeaders adds the selected Content-... headers to HeaderInfo
procedure GetContentHeaders(var HeaderInfo: THeaderData; FHO: TFindHeaderOptions; Add: Longbool);
begin
  with HeaderInfo do begin
    if Len > 0 then begin
      if not Add or (CTHeader = E) then begin
	if fhoContentType in FHO then
	  CTHeader := FindHeaderLine(CT_HEADER, HeaderInfo, true).Text
	else CTHeader := E;
      end;
      if not Add or (CTEHeader = E) then begin
	if fhoContentTransferEncoding in FHO then
	  CTEHeader := FindHeaderLine(CTE_HEADER, HeaderInfo, true).Text
	else CTEHeader := E;
      end;
      if not Add or (CDPHeader = E) then begin
	if fhoContentDisposition in FHO then
	  CDPHeader := FindHeaderLine(CDP_HEADER, HeaderInfo, true).Text
	else CDPHeader := E;
      end;
      if not Add or (CTDHeader = E) then begin
	if fhoContentDescription in FHO then
	  CTDHeader := FindHeaderLine(CTD_HEADER, HeaderInfo, true).Text
	else CTDHeader := E;
      end;
      if not Add or (CLCHeader = E) then begin
	if fhoContentLocation in FHO then
	  CLCHeader := FindHeaderLine(CLC_HEADER, HeaderInfo, true).Text
	else CLCHeader := E;
      end;
      if not Add or (CIDHeader = E) then begin
	if fhoContentID in FHO then
	  CIDHeader := FindHeaderLine(CID_HEADER, HeaderInfo, true).Text
	else CIDHeader := E;
      end;
    end
    else begin
      CTHeader := E;
      CTEHeader := E;
      CDPHeader := E;
      CTDHeader := E;
      CLCHeader := E;
      CIDHeader := E;
    end;
  end;
end;

// GetHeaderInfo returns the complete MailMessage with Pos & Len of the headers
procedure GetHeaderInfo(var HeaderData: THeaderData; FHO: TFindHeaderOptions);
var
  iPos: Integer;
begin
  with HeaderData do begin
    iPos := Pos;
    Pos := FirstRangeCharPos(LF, Text, iPos, MaxPos);
    if Pos > 0 then begin
      if (Pos = 1) or (Text[pred(Pos)] <> CR) then
	Len := FirstRangeStrPos(LF + LF, Text, Pos, MaxPos) - iPos
      else Len := FirstRangeStrPos(PARA, Text, pred(Pos), MaxPos) - iPos;
      if Len > 0 then
	Pos := iPos
      else Pos := 0;
    end
    else Len := 0;
  end;
  GetContentHeaders(HeaderData, FHO, false);
end;

// PrevHeader contains the MailMessage and Pos and Len of the previous (sub)headers
// GetNextHeaderInfo returns the MailMessage and Pos and Len of the next subheader section
function GetNextHeaderInfo(const PrevHeader: THeaderData; const Boundary: String; FHO: TFindHeaderOptions): THeaderData;
var
  iBeg, iPos, iEnd: Integer;
begin
  FillChar(Result, SizeOf(THeaderData), 0);
  if Boundary <> E then with PrevHeader do begin
    iBeg := FirstRangeLoStrPos(DD + Boundary, Text, Pos + Len, MaxPos);
    if (iBeg = 1) or ((iBeg > 1) and (Text[pred(iBeg)] in LFCR)) then begin
      iPos := iBeg + Length(Boundary) + Length(DD);
      iEnd := pred(FirstRangeCharPos(LF, Text, iPos, MaxPos));
      if (iEnd > iBeg) and not StrAtPos(DD, Text, iPos) then begin
	if Text[iEnd] <> CR then
	  iEnd := FirstRangeStrPos(LF + LF, Text, succ(iEnd), MaxPos)
	else iEnd := FirstRangeStrPos(PARA, Text, iEnd, MaxPos);
	Result.Text := Text;
	Result.Pos := iBeg;
	Result.Len := iEnd - iBeg;
	Result.MaxPos := MaxPos;
	Result.Recursion := PrevHeader.Recursion;
      end;
    end;
  end;
  GetContentHeaders(Result, FHO, false);
end;

// PrevHeader contains the MailMessage and Pos and Len of the previous (sub)headers
// GetSubHeaderInfo returns the MailMessage and Pos and Len of the respective subheaders
function GetSubHeaderInfo(const PrevHeader: THeaderData; FHO: TFindHeaderOptions; FMO: TFindMultipartOption): THeaderData;
var
  iBeg, iEnd: Integer;
  sBoundary, sSection: String;
begin
  FillChar(Result, SizeOf(THeaderData), 0);
  Result.Recursion := PrevHeader.Recursion;
  sBoundary := GetBoundary(PrevHeader);
  if sBoundary <> E then with PrevHeader do begin
    Result := GetNextHeaderInfo(PrevHeader, sBoundary, FHO);
    with Result do begin
      iBeg := Pos;
      iEnd := Pos + Len;
    end;
    if (FMO > fmoFirstSubHeader) and (Result.Len > 0) then begin
      if FMO = fmoFirstNonMixed then
	sSection := CT_MULTIPART + CT_MIXED
      else sSection := CT_MULTIPART;
      if FirstRangeLoStrPos(sSection, Result.Text, iBeg, iEnd) > 0 then
	Result := GetSubHeaderInfo(Result, FHO, FMO)
      else if FMO = fmoLastMultipart then Result := PrevHeader;
    end;
  end;
  GetContentHeaders(Result, FHO, false);
end;

// GetPSMimeBoundary returns Encoding and (optionally) Boundary from Content Type Header
function GetPSMimeBoundary(const MimeHeaders: THeaderData; GetBoundaryStr: Boolean; var Boundary: String): TEncoding;
begin
  Boundary := E;
  Result := eNone;
  with MimeHeaders do begin
    if FirstLoStrPos(PKCS7, CTHeader) > 0 then begin
      if LoStrAtPos(CT_MULTIPART, CTHeader, 1)
      and (NextLoStrPos(SM_SIGNED, CTHeader, succ(Length(CT_MULTIPART))) > 0) then
	Result := eSMSigned
      else if LoStrAtPos(CT_APPLICATION, CTHeader, 1)
      and (NextLoStrPos(SM_SIG, CTHeader, succ(Length(CT_APPLICATION))) = 0)
      and (NextLoStrPos(SM_SIGNED, CTHeader, succ(Length(CT_APPLICATION))) > 0) then
	Result := eSMInline
      else if FirstLoStrPos(SM_ENC, CTHeader) > 0 then
	Result := eSMEncrypted
      else Result := eNone;
    end
    else if FirstLoStrPos(PGPMIME, CTHeader) > 0 then begin
      if FirstLoStrPos(PM_ENC, CTHeader) > 0 then
	Result := ePMEncrypted
      else if FirstLoStrPos(PM_SIG, CTHeader) > 0 then
	Result := ePMSigned
      else if FirstLoStrPos(PM_KEYS, CTHeader) > 0 then
	Result := ePMKeys
      else Result := eNone;
    end;
  end;
  if GetBoundaryStr and (Result > eNone) then Boundary := GetBoundary(MimeHeaders);
end;

// PSMimeData returns Pos and Len of any included PGP/MIME or S/MIME message in MailMessage
function ContainsPSMimeMsg(const HeaderData: THeaderData; var PSMimeData: THeaderData): TEncoding;
var
  sParentBoundary, sPSBoundary: String;
  iLen: Integer;
begin
  Result := eNone;
  PSMimeData := HeaderData;
  try
    if PSMimeData.Len = 0 then
      GetHeaderInfo(PSMimeData, [fhoContentType])
    else if PSMimeData.CTHeader = E then GetContentHeaders(PSMimeData, [fhoContentType], false);
    sParentBoundary := GetBoundary(PSMimeData);
    Result := GetPSMimeBoundary(PSMimeData, true, sPSBoundary);
    if Result = eNone then begin
      PSMimeData := GetSubHeaderInfo(PSMimeData, [fhoContentType], fmoFirstNonMixed);
      Result := GetPSMimeBoundary(PSMimeData, true, sPSBoundary);
    end;
    with PSMimeData do begin
      if Result <= eCerts then begin
	Pos := 0;
	Len := 0;
	Exit;
      end;
      if (sPSBoundary = E) and (sParentBoundary = E) then begin
	Len := succ(MaxPos - Pos);
	Exit;
      end;
      if Pos > 0 then begin
	if sPSBoundary <> E then begin
	  iLen := Length(sPSBoundary);
	  Len := FirstRangeStrPos(DD + sPSBoundary + DD, Text, Pos + Len + iLen, MaxPos);
	  if Len > Pos then begin
	    inc(Len, iLen + (Length(DD) shl 1));
	    if (Length(Text) > Len) and (Text[succ(Len)] in LFCR) then inc(Len, Length(CRLF));
	    dec(Len, Pos);
	  end
	  else Pos := 0;
	end
	else begin
	  iLen := Length(sParentBoundary);
	  Len := FirstRangeStrPos(DD + sParentBoundary, Text, Pos + iLen, MaxPos);
	  if Len > Pos then
	    dec(Len, Pos)
	  else Pos := 0;
	end;
      end
      else Len := 0;
    end;
  finally
    with PSMimeData do begin
      if (Pos = 0) or (Len = 0) then begin
	Text := E;
	MaxPos := 0;
      end;
      Encoding := Result;
    end;
  end;
end;

// InlineData returns Pos and Len of any included inline PGP data in MailMessage
function GetInlineEncoding(const HeaderData: THeaderData; var InlineData: THeaderData): TEncoding;
var
  cPrefix: Char;
  iBeg, iEnd, iPos: Integer;
  sMarker: String;
begin
  Result := eNone;
  try
    InlineData := HeaderData;
    with InlineData do begin
      cPrefix := LF;
      Pos := FirstRangeStrPos(ANY_BEG, Text, Pos, MaxPos);
      if Pos = 0 then begin
	Len := 0;
	Exit;
      end
      else if Pos > 1 then cPrefix := Text[pred(Pos)];
      iBeg := Pos;
      inc(iBeg, Length(ANY_BEG));
      if not StrAtPos(IS_PGP, Text, iBeg) then begin
	if StrAtPos(IS_CERT, Text, iBeg) then begin
	  Len := FirstRangeStrPos(cPrefix + CERT_END, Text, Pos + Length(CERT_BEG), MaxPos);
	  if Len > Pos then begin
	    while (Pos > 1) and (Text[pred(Pos)] >= SP) do dec(Pos);
	    inc(Len, succ(Length(CERT_END)));
	    while Text[Len] >= SP do inc(Len);
	    if Text[pred(Len)] = EQ then begin
	      while (Len <= MaxPos) and (Text[Len] < SP) do inc(Len);
	      while Text[Len] >= SP do inc(Len);
	    end;
	    if (MaxPos > Len) and (Text[Len] in LFCR) then inc(Len, Length(CRLF));
	    dec(Len, Pos);
	    if Len > Pos then
	      Result := eCerts
	    else Pos := 0;
	  end
	  else Pos := 0;
	end
	else Len := 0;
	Exit;
      end;
      iBeg := Pos;
      iEnd := succ(iBeg);
      while Text[iEnd] >= SP do inc(iEnd);
      iPos := iEnd - iBeg;
      if iPos > 0 then begin
	sMarker := Copy(Text, iBeg, iPos);
	if sMarker[iPos] = EQ then begin
	  Delete(sMarker, iPos, 1);
	  while (iEnd <= MaxPos) and (Text[iEnd] < SP) do inc(iEnd);
	  iBeg := iEnd;
	  while Text[iEnd] >= SP do inc(iEnd);
	  iPos := iEnd - iBeg;
	  sMarker := sMarker + Copy(Text, iBeg, iPos);
	end;
      end;
      iBeg := succ(Length(PGP_BEG));
      iPos := Pos + Length(sMarker);
      Len := FirstRangeStrPos(cPrefix + PGP_END, Text, iPos, MaxPos);
      if Len > iPos then begin
	if StrAtPos(MSG_BEG, sMarker, iBeg) then
	  Result := ePlainEncrypted
	else if StrAtPos(SIGN_BEG, sMarker, iBeg) then
	  Result := ePlainSigned
	else if StrAtPos(KEY_BEG, sMarker, iBeg) then begin
	  Result := ePlainKeys;
	end;
      end;
      if Result > eNone then begin
	while (Pos > 1) and (Text[pred(Pos)] >= SP) do dec(Pos);
	inc(Len, succ(Length(PGP_END)));
	while Text[Len] >= SP do inc(Len);
	if Text[pred(Len)] = EQ then begin
	  while (Len <= MaxPos) and (Text[Len] < SP) do inc(Len);
	  while Text[Len] >= SP do inc(Len);
	end;
	if (MaxPos > Len) and (Text[Len] in LFCR) then inc(Len, Length(CRLF));
	dec(Len, Pos);
      end
      else begin
	Pos := 0;
	Len := 0;
      end;
    end;
  finally
    with InlineData do begin
      if (Pos = 0) or (Len = 0) then begin
	Text := E;
	MaxPos := 0;
      end;
      Encoding := Result;
    end;
  end;
end;

// GetArmoredData prepares message data for PGP decoding by returning the data to decode
function GetArmoredData(const MailMessage: String; var MD: TMessageData): String;

  function AdjustBreaks(const Input: String): String;
  const
    P_TAG = '<p>';
    BR_TAG = '<br>';
  var
    pResult: PChar;
    iLen, iBeg, iEnd, iCopy: Integer;
  begin
    if Input <> E then begin
      iLen := Length(Input);
      SetString(Result, nil, iLen + (GetCharCount(LF, Input) * Length(BR_TAG)));
      pResult := Pointer(Result);
      iBeg := 1;
      repeat
	iEnd := NextLoStrPos(P_TAG, Input, iBeg);
	if iEnd >= iBeg then begin
	  inc(iEnd, Length(P_TAG));
	  iCopy := iEnd - iBeg;
	  Move(Input[iBeg], pResult^, iCopy);
	  inc(pResult, iCopy);
	  PQChar(pResult)^ := BR_TAG;
	  inc(pResult, Length(BR_TAG));
	  iBeg := iEnd;
	end
	else Break;
      until iEnd = 0;
      if iBeg <= iLen then begin
	iCopy := succ(iLen - iBeg);
	Move(Input[iBeg], pResult^, iCopy);
	inc(pResult, iCopy);
      end;
      Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    end;
  end;

var
  hdEncoding: THeaderData;
  iLen, iPos: Integer;
begin
  Result := E;
  MD.Encoding := GetEncoding(MailMessage, hdEncoding);
  if MD.Encoding > eNone then with hdEncoding do begin
    Result := Copy(Text, Pos, Len);
    iPos := Pos + Len;
    MD.Suffix := Copy(Text, iPos, succ(MaxPos - iPos));
    if MD.Encoding > eCerts then begin
      Pos := 1;
      Len := 0;
      GetHeaderInfo(hdEncoding, []);
      if Len > 0 then MD.Prefix := Copy(Text, Pos, Len) + PARA;
      with MD do ExtraData := (Trim(Suffix) <> E) or (Length(Prefix) > Len + Length(PARA));
    end
    else begin
      iLen := pred(Pos);
      while (iLen > 0) and (Text[iLen] <= SP) do dec(iLen);
      if iLen > 0 then MD.Prefix := Copy(Text, 1, iLen) + PARA;
      with MD do ExtraData := (Trim(Suffix) <> E) or (Trim(Prefix) <> E);
    end;
    if (Result <> E) and (Result[1] = LT) and (MD.Encoding in [ePlainKeys, eCerts]) then begin
      Result := Trim(ConvertFormat(AdjustBreaks(TidyHtml(Result, E, false, false, false).Body), E, [], 0));
    end;
  end;
end;

// EncodedData returns Pos & Len of inline or MIME encoded data inside MailMessage
function GetEncoding(const MailMessage: String; var EncodedData: THeaderData): TEncoding;
var
  hdMessage: THeaderData;
begin
  Result := eNone;
  FillChar(hdMessage, SizeOf(THeaderData), 0);
  with hdMessage do begin
    Text := MailMessage;
    MaxPos := Length(MailMessage);
  end;
  EncodedData := hdMessage;
  if MailMessage = E then Exit;
  inc(hdMessage.Pos);
  Result := ContainsPSMimeMsg(hdMessage, EncodedData);
  if Result <= eCerts then Result := GetInlineEncoding(hdMessage, EncodedData);
end;

function IsEncoded(const Encoding: TEncoding; KeysOnly: Longbool): Longbool;
begin
  if KeysOnly then
    Result := (Encoding in [ePlainKeys, eCerts, eSMEncrypted..eSMInline])
  else Result := (Encoding in [ePlainEncrypted..ePlainSigned, ePMEncrypted..ePMSigned, eSMEncrypted..eSMInline]);
end;

// StripSignatureData removes PGP's signature formatting
function StripSignatureData(const Text: String): String;
var
  pResult: PChar;
  iLen, iBeg, iEnd, iPos: Integer;
begin
  iLen := Length(Text);
  if iLen > 0 then begin
    SetString(Result, nil, iLen);
    pResult := Pointer(Result);
    iBeg := 1;
    while (iBeg <= iLen) and (Text[iBeg] <= SP) do inc(iBeg);
    if StrAtPos(PGP_BEG + SIGN_BEG, Text, iBeg) then begin
      inc(iBeg, Length(PGP_BEG) + Length(SIGN_BEG) + Length(CRLF));
      repeat
	iEnd := iBeg;
	while not (Text[iEnd] in [NULL, LF]) do inc(iEnd);
	while (iBeg <= iEnd) and (Text[iBeg] <= SP) do inc(iBeg);
	if iBeg > iEnd then Break;
	iBeg := succ(iEnd);
      until iEnd = 0;
      iPos := NextStrPos(LF + PGP_BEG + SIG_BEG, Text, iBeg);
      if iPos >= iBeg then begin
	iEnd := iBeg - 2;
	repeat
	  iEnd := NextStrPos('- ', Text, iEnd + 2);
	  if (iEnd < iBeg) or (iEnd > iPos) then Break;
	  iLen := iEnd - iBeg;
	  if (Text[pred(iEnd)] <> LF) and (iLen > 0) then inc(iLen, Length(CRLF));
	  if iLen > 0 then begin
	    Move(Text[iBeg], pResult^, iLen);
	    inc(pResult, iLen);
	    iBeg := iEnd + 2;
	  end;
	until iBeg > iPos;
	iLen := succ(iPos - iBeg);
	if iLen > 0 then begin
	  Move(Text[iBeg], pResult^, iLen);
	  inc(pResult, iLen);
	end;
      end;
      Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    end
    else Result := E;
  end
  else Result := E;
end;

// UnwrapArmoredKeys restores proper ARMOR line length
function UnwrapArmoredKeys(const Text: String): String;
const
  B64_LINE_LEN = 64;
var
  hdText: THeaderData;
  sKeyData: String;
  iKeysLength, iCharIndex, iLastLine, iLinePos, iCount, iCopy: Integer;
  pResult: PChar;
begin
  Result := Text;
  FillChar(hdText, SizeOf(THeaderData), 0);
  with hdText do begin
    Text := Result;
    Pos := 1;
    Len := Length(Text);
    MaxPos := Len;
  end;
  while GetInlineEncoding(hdText, hdText) in [ePlainKeys, eCerts] do begin
    Result := E;
    with hdText do sKeyData := Copy(Text, Pos, Len);
    iKeysLength := Length(sKeyData);
    if hdText.Encoding = ePlainKeys then begin
      iCharIndex := succ(Length(PGP_BEG) + Length(KEY_BEG) + Length(CRLF));
      iLastLine := iKeysLength - Length(CRLF) - Length(PGP_END) - Length(KEY_END);
    end
    else begin
      iCharIndex := succ(Length(CERT_BEG) + Length(CRLF));
      iLastLine := iKeysLength - Length(CRLF) - Length(CERT_END);
    end;
    if (iLastLine > 0) and (sKeyData[iLastLine] = LF) then dec(iLastLine, Length(CRLF));
    while (iLastLine > 0) and (sKeyData[iLastLine] <> CR) do dec(iLastLine);
    if (iKeysLength > iCharIndex) and (iLastLine > iCharIndex) then begin
      if hdText.Encoding = ePlainKeys then begin
	repeat
	  iLinePos := iCharIndex;
	  while not (sKeyData[iLinePos] in [NULL, LF]) do inc(iLinePos);
	  while (iCharIndex <= iLinePos) and (sKeyData[iCharIndex] <= SP) do inc(iCharIndex);
	  if iCharIndex > iLinePos then Break;
	  iCharIndex := succ(iLinePos);
	until iLinePos = 0;
      end
      else iLinePos := pred(iCharIndex);
      SetString(Result, nil, Length(sKeyData));
      pResult := Pointer(Result);
      Move(Pointer(sKeyData)^, pResult^, iLinePos);
      inc(pResult, iLinePos);
      iCount := 0;
      repeat
	iLinePos := iCharIndex;
	while not (sKeyData[iCharIndex] in [NULL, CR]) do begin
	  inc(iCharIndex);
	  inc(iCount);
	end;
	if iCharIndex < iLastLine then begin
	  if iCount < B64_LINE_LEN then begin
	    iCopy := iCharIndex - iLinePos;
	    Move(sKeyData[iLinePos], pResult^, iCopy);
	    inc(iCharIndex, Length(CRLF));
	    inc(pResult, iCopy);
	  end
	  else begin
	    iCopy := iCharIndex - iLinePos + Length(CRLF);
	    Move(sKeyData[iLinePos], pResult^, iCopy);
	    inc(pResult, iCopy);
	    iCount := 0;
	  end;
	end
	else begin
	  if (hdText.Encoding = eCerts) and (iCount < B64_LINE_LEN) then begin
	    iCopy := iCharIndex - iLinePos;
	    Move(sKeyData[iLinePos], pResult^, iCopy);
	    inc(iCharIndex, Length(CRLF));
	    iLinePos := iCharIndex;
	    inc(pResult, iCopy);
	  end;
	  iCount := succ(iKeysLength - iLinePos);
	  Move(sKeyData[iLinePos], pResult^, iCount);
	  inc(pResult, succ(iCount));
	  Delete(Result, pResult - Pointer(Result), MAXINT);
	  with hdText do begin
	    Text := XChgStrAtPos(Text, Result, Pos, Len);
	    inc(Pos, Length(Result));
	    Len := Length(Text);
	    MaxPos := Len;
	    Break;
	  end;
	end;
      until iCharIndex = iLastLine;
      Result := hdText.Text;
    end;
  end;
end;

// MimeDecodeHeaders decodes header contents using
// B64Decode or QPDecode returning UTF-8 encoded ones
function MimeDecodeHeaders(const Headers: String): String;

  procedure StripExcessSpaces(const StartPos: PChar; var Current: PChar);
  begin
    while (Current > StartPos) and (PChar(pred(Longint(Current)))^ in WSP + LFCR) do dec(Current);
  end;

var
  pResult: PChar;
  iHeaders, iBeg, iEnd, iStart, iStop, iPos, iLen, iCode: Integer;
  sConverted, sCharset: String;
begin
  if Headers <> E then begin
    iHeaders := Length(Headers);
    SetString(Result, nil, iHeaders * UTF8Factor);
    pResult := Pointer(Result);
    sConverted := E;
    iBeg := 1;
    repeat
      iEnd := NextStrPos(CODE_BEG, Headers, iBeg);
      if iEnd > 0 then begin
	iLen := iEnd - iBeg;
	if iLen > 0 then begin
	  Move(Headers[iBeg], pResult^, iLen);
	  inc(pResult, iLen);
	end;
	iStart := iEnd + Length(CODE_BEG);
	iStop := 0;
	iPos := NextCharPos(QM, Headers, iStart);
	if iPos > 0 then iStop := NextStrPos(CODE_END, Headers, iPos + 3);
	if (iStop > iPos) and (Headers[iPos + 2] = QM) then begin
	  if (sConverted <> E) and (sConverted[Length(sConverted)] <> SP) then StripExcessSpaces(Pointer(Result), pResult);
	  iCode := ord(Headers[succ(iPos)]) and $DF;
	  sCharset := UpperCase(Copy(Headers, iStart, iPos - iStart));
	  Delete(sCharset, FirstCharPos(AST, sCharset), MAXINT);
	  inc(iPos, 3);
	  case iCode of
	    ord('B'): sConverted := StringToUtf8(B64Convert(Copy(Headers, iPos, iStop - iPos)), sCharset);
	    ord('Q'): sConverted := StringToUtf8(QPConvert(Copy(Headers, iPos, iStop - iPos), true), sCharset);
	  else
	    sConverted := AnsiStringToUtf8(Copy(Headers, iPos, iStop - iPos));
	  end;
	  iLen := Length(sConverted);
	  if iLen > 0 then begin
	    iPos := 1;
	    if iLen > 0 then begin
	      if sConverted[1] = SP then StripExcessSpaces(Pointer(Result), pResult);
	      Move(sConverted[iPos], pResult^, iLen);
	      inc(pResult, iLen);
	    end;
	  end;
	  iBeg := iStop + Length(CODE_END);
	  Continue;
	end;
	sConverted := E;
      end;
      iLen := succ(iHeaders - iBeg);
      if iLen > 0 then begin
	Move(Headers[iBeg], pResult^, iLen);
	inc(pResult, iLen);
      end;
      StripExcessSpaces(Pointer(Result), pResult);
      Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
      Break;
    until iEnd = 0;
  end
  else Result := E;
end;

// MimeGetParameterValue returns header parameter values
// of a header content line if PHeaderInfo = nil
// or extracted by name from PHeaderInfo
function MimeGetParameterValue(const HeaderData, Attribute: String; PHeaderInfo: PHeaderData; Raw: Longbool): String;

  function GetValuePositions(const Header: String; var BegPos, EndPos: Integer): Integer;
  begin
    if Header[BegPos] = QU then begin
      EndPos := BegPos;
      inc(BegPos);
      repeat
	EndPos := NextCharPos(QU, Header, succ(EndPos));
      until (EndPos = 0) or (Header[pred(EndPos)] <> BS);
    end
    else begin
      EndPos := BegPos;
      while not (Header[EndPos] in NLFCRSC) do inc(EndPos);
    end;
    Result := EndPos - BegPos;
  end;

var
  hdFound: THeaderData;
  sHeader, sCharset, sLanguage: String;
  iHeader, iBeg, iEnd, iPos, iLen: Integer;
  fParametersChecked: Longbool;
  pResult: PChar;
begin
  Result := E;
  if PHeaderInfo = nil then begin
    hdFound.Text := HeaderData;
    sHeader := HeaderData;
  end
  else begin
    hdFound := FindHeaderLine(HeaderData, PHeaderInfo^, true);
    with hdFound do sHeader := Trim(Copy(PHeaderInfo^.Text, Pos + Length(HeaderData), Len));
  end;
  iHeader := Length(sHeader);
  if iHeader = 0 then Exit;
  iBeg := FirstLoStrPos(Attribute, sHeader);
  if iBeg = 0 then Exit;
  inc(iBeg, Length(Attribute));
  while (iBeg < iHeader) and (sHeader[iBeg] in WSP) do inc(iBeg);
  iEnd := 0;
  SetString(Result, nil, iHeader);
  pResult := Pointer(Result);
  fParametersChecked:= false;
  case sHeader[iBeg] of
    AST: if (iBeg + 3) < iHeader then begin
      if sHeader[succ(iBeg)] = '0' then
	inc(iBeg, 3)
      else inc(iBeg);
      while iBeg < iHeader do begin
	case sHeader[pred(iBeg)] of
	  AST: if sHeader[iBeg] = EQ then begin
	    inc(iBeg);
	    iEnd := iBeg;
	    while not (sHeader[iEnd] in NLFCR) do inc(iEnd);
	    if not fParametersChecked then begin
	      iPos := FirstRangeCharPos(SQ, sHeader, iBeg, pred(iEnd));
	      if iPos > 0 then begin
		sCharset := UpperCase(Copy(sHeader, iBeg, iPos - iBeg));
		inc(iBeg, succ(Length(sCharset)));
		iPos := FirstRangeCharPos(SQ, sHeader, iBeg, pred(iEnd));
		if iPos > 0 then begin
		  sLanguage := UpperCase(Copy(sHeader, iBeg, iPos - iBeg));
		  inc(iBeg, succ(Length(sLanguage)));
		end;
	      end;
	      fParametersChecked := true;
	    end;
	    iPos := iBeg;
	    repeat
	      while (iPos < iEnd) and (sHeader[iPos] <> PC) do inc(iPos);
	      iLen := iPos - iBeg;
	      if iLen > 0 then begin
		Move(sHeader[iBeg], pResult^, iLen);
		inc(pResult, iLen);
		iBeg := iPos;
	      end;
	      if sHeader[iPos] = PC then begin
		pResult^ := HexToChar(@sHeader[succ(iPos)]);
		inc(pResult);
		inc(iPos, 3);
		iBeg := iPos;
	      end;
	    until iPos >= iEnd;
	  end
	  else Break;
	  EQ: begin
	    iLen := GetValuePositions(sHeader, iBeg, iEnd);
	    if iLen > 0 then begin
	      Move(sHeader[iBeg], pResult^, iLen);
	      inc(pResult, iLen);
	    end;
	    if sHeader[iEnd] = QU then inc(iEnd);
	    while not (sHeader[iEnd] in NLFCRSC) do inc(iEnd);
	  end;
	else
	  Break;
	end;
	iBeg := iEnd;
	while (iBeg <= iHeader) and (sHeader[iBeg] <= SP) do inc(iBeg);
	if LoStrAtPos(Attribute, sHeader, iBeg) then begin
	  inc(iBeg, Length(Attribute));
	  if (iBeg < iHeader) and (sHeader[iBeg] = AST) then begin
	    inc(iBeg);
	    while sHeader[iBeg] in ['0'..'9'] do inc(iBeg);
	    if sHeader[iBeg] > SP then
	      inc(iBeg)
	    else Break;
	  end
	  else Break;
	end
	else Break;
      end;
      Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
      Result := StringToUtf8(Result, CheckWinCharset(sCharset));
    end
    else Exit;
    EQ: begin
      if Raw then
	sHeader := hdFound.Text
      else sHeader := MimeDecodeHeaders(hdFound.Text);
      iBeg := FirstLoStrPos(Attribute, sHeader);
      if iBeg > 0 then begin
	inc(iBeg, Length(Attribute));
	while (iBeg < iHeader) and (sHeader[iBeg] in WSPEQ) do inc(iBeg);
	if iBeg < iHeader then begin
	  iLen := GetValuePositions(sHeader, iBeg, iEnd);
	  if iLen > 0 then begin
	    Move(sHeader[iBeg], pResult^, iLen);
	    inc(pResult, iLen);
	    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
	  end;
	end;
      end;
    end;
  end;
end;

// MimeFoldNameParameters folds the Content-type "name" value
// and the Content-disposition "filename" parameters
// and QP encodes them if required
procedure MimeFoldNameParameters(var MailMessage: String; var HD: THeaderData;
				 const Charset: String; RemoveCryptoParameter: Longbool);

  procedure FoldHeader(const FoldParameter: String; var HDC: THeaderData; RemoveCryptoParameter: Longbool);

    function ParameterEnd(const Header: String; Pos: Integer): Integer;
    begin
      while not (Header[Pos] in [NULL, QU, SC]) do inc(Pos);
      if Header[Pos] = QU then begin
	inc(Pos);
	while not (Header[Pos] in [NULL, QU]) do inc(Pos);
	if Header[Pos] = QU then begin
	  inc(Pos);
	  while not (Header[Pos] in [NULL, SC]) do inc(Pos);
	end;
      end;
      if Header[Pos] <> NULL then begin
	inc(Pos);
	while (Header[Pos] in WSP) do inc(Pos);
      end;
      dec(Pos);
      Result := Pos;
    end;

  var
    iBeg, iEnd, iLen: Integer;
  begin
    with HDC do begin
      if FoldParameter = CDP_FILENAME then begin
	if RemoveCryptoParameter then begin
	  repeat
	    iBeg := FirstLoStrPos(X_ENCRYPTION, Text);
	    if iBeg > 0 then begin
	      iEnd := ParameterEnd(Text, iBeg + succ(Length(X_ENCRYPTION)));
	      Delete(Text, iBeg, succ(iEnd - iBeg));
	    end
	    else Break;
	  until iBeg = 0;
	  iEnd := Length(Text);
	  while (iEnd > 0) and (Text[iEnd] in [TAB, SP, SC]) do dec(iEnd);
	  Delete(Text, succ(iEnd), MAXINT);
	end;
	iBeg := FirstLoStrPos(CDP_FILENAME, Text);
	iEnd := succ(iBeg + Length(CDP_FILENAME));
      end
      else begin
	iBeg := FirstLoStrPos(CT_NAME, Text);
	iEnd := succ(iBeg + Length(CT_NAME));
      end;
      if iBeg = 0 then Exit;
      iLen := Length(Text);
      while (iEnd > 0) and (iEnd < iLen) and not (Text[iEnd] in LFCR) do begin
	if Text[iEnd] in WSP then begin
	  Text := XChgStrAtPos(Text, CRLF, iEnd, 0);
	  inc(iEnd, Length(CRLF));
	  inc(iLen, Length(CRLF));
	end
	else begin
	  Text := XChgStrAtPos(Text, CRLFSP, iEnd, 0);
	  inc(iEnd, Length(CRLFSP));
	  inc(iLen, Length(CRLFSP));
	end;
	iEnd := ParameterEnd(Text, iEnd);
      end;
    end;
  end;

var
  hdCDP, hdCT: THeaderData;
  sBeg, sMid, sEnd, sSubHeaders: String;
  iPos: Integer;
begin
  with HD do begin
    Text := MailMessage;
    hdCDP := FindHeaderLine(CDP_HEADER, HD, true);
    if hdCDP.Pos > 0 then begin
      hdCT := FindHeaderLine(CT_HEADER, HD, true);
      if hdCDP.Pos > hdCT.Pos then begin
	sBeg := Copy(Text, Pos, hdCT.Pos - Pos);
	with hdCT do iPos := Pos + Len;
	sMid := Copy(Text, iPos, hdCDP.Pos - iPos);
	with hdCDP do iPos := Pos + Len;
	sEnd := Copy(Text, iPos, Pos + Len - iPos);
      end
      else begin
	sBeg := Copy(Text, Pos, hdCDP.Pos - Pos);
	with hdCDP do iPos := Pos + Len;
	sMid := Copy(Text, iPos, hdCT.Pos - iPos);
	with hdCT do iPos := Pos + Len;
	sEnd := Copy(Text, iPos, Pos + Len - iPos);
      end;
      Text := E;
      with hdCT do begin
	CTHeader := MimeGetParameterValue(hdCT.Text, CT_NAME, nil, true);
	FoldHeader(CT_NAME, hdCT, true);
	if IsExtended(CTHeader) then begin
	  Len := Length(CTHeader);
	  Pos :=  FirstStrPos(CTHeader, Text);
	  CTHeader := QPEncode(CTHeader, Charset);
	  if Pos > 0 then begin
	    if Text[pred(Pos)] = QU then begin
	      dec(Pos);
	      inc(Len, 2);
	    end;
	    Text := XChgStrAtPos(Text, CTHeader, Pos, Len);
	  end;
	end;
      end;
      with hdCDP do begin
	CDPHeader := MimeGetParameterValue(hdCDP.Text, CDP_FILENAME, nil, true);
	FoldHeader(CDP_FILENAME, hdCDP, true);
	if IsExtended(CDPHeader) then begin
	  Len := Length(CDPHeader);
	  Pos :=  FirstStrPos(CDPHeader, Text);
	  CDPHeader := QPEncode(CDPHeader, Charset);
	  if Pos > 0 then begin
	    if Text[pred(Pos)] = QU then begin
	      dec(Pos);
	      inc(Len, 2);
	    end;
	    Text := XChgStrAtPos(Text, CDPHeader, Pos, Len);
	  end;
	end;
      end;
      sSubHeaders := sBeg + CT_HEADER + SP + hdCT.Text + sMid + CDP_HEADER + SP + hdCDP.Text + sEnd;
      MailMessage := XChgStrAtPos(MailMessage, sSubHeaders, Pos, Len);
      Len := Length(sSubHeaders);
      Text := E;
    end;
  end;
end;

// HeaderData must only contain the headers, Len & MaxPos will be adjusted
// Pos + Len will point to the start position of the main Content-... headers block
// HeaderData may contain headers for replacing existing ones, NewHeaders are added if provided
function ReplaceContentHeaders(var HeaderData: THeaderData; DeleteThem: Longbool;
			       const NewHeaders, Charset: String): Integer;
var
  hdContent: THeaderData;
  sHeaders: String;
begin
  with HeaderData do begin
    Len := 0;
    hdContent := FindHeaderLine(CTD_HEADER, HeaderData, false);
    if hdContent.Len > 0 then begin
      if not DeleteThem and (CTDHeader = E) then CTDHeader := CTD_HEADER + SP + hdContent.Text + CRLF;
      Delete(Text, hdContent.Pos, hdContent.Len + Length(CRLF));
      MaxPos := Length(Text);
    end;
    hdContent := FindHeaderLine(CTE_HEADER, HeaderData, false);
    if hdContent.Len > 0 then begin
      if not DeleteThem and (CTEHeader = E) then CTEHeader := CTE_HEADER + SP + hdContent.Text + CRLF;
      Delete(Text, hdContent.Pos, hdContent.Len + Length(CRLF));
      MaxPos := Length(Text);
    end;
    hdContent := FindHeaderLine(CDP_HEADER, HeaderData, false);
    if hdContent.Len > 0 then begin
      if not DeleteThem and (CDPHeader = E) then CDPHeader := CDP_HEADER + SP + hdContent.Text + CRLF;
      Delete(Text, hdContent.Pos, hdContent.Len + Length(CRLF));
      MaxPos := Length(Text);
    end;
    hdContent := FindHeaderLine(CLC_HEADER, HeaderData, false);
    if hdContent.Len > 0 then begin
      if not DeleteThem and (CLCHeader = E) then CLCHeader := CLC_HEADER + SP + hdContent.Text + CRLF;
      Delete(Text, hdContent.Pos, hdContent.Len + Length(CRLF));
      MaxPos := Length(Text);
    end;
    hdContent := FindHeaderLine(CID_HEADER, HeaderData, false);
    if hdContent.Len > 0 then begin
      if not DeleteThem and (CIDHeader = E) then CIDHeader := CID_HEADER + SP + hdContent.Text + CRLF;
      Delete(Text, hdContent.Pos, hdContent.Len + Length(CRLF));
      MaxPos := Length(Text);
    end;
    hdContent := FindHeaderLine(CT_HEADER, HeaderData, false);
    if not DeleteThem and (CTHeader = E) then CTHeader := CT_HEADER + SP + hdContent.Text + CRLF;
    if DeleteThem or (NewHeaders = E) then
      Delete(Text, hdContent.Pos, hdContent.Len + Length(CRLF))
    else Text := XChgStrAtPos(Text, NewHeaders, hdContent.Pos, hdContent.Len);
    Result := Length(Text);
    while (Result > 0) and (Text[Result] in LFCR) do dec(Result);
    Delete(Text, succ(Result), MAXINT);
    inc(Result, Length(CRLF));
    if not DeleteThem then begin
      if LoStrAtPos(CT_MULTIPART, hdContent.Text, 1) then
	Text := Text + CRLF + TrimRight(CTHeader)
      else Text := Text + CRLF + TrimRight(CTHeader + CLCHeader + CIDHeader + CDPHeader + CTEHeader + CTDHeader);
      sHeaders := Text;
      Len := Length(Text);
      MaxPos := Len;
      Text := E;
      MimeFoldNameParameters(sHeaders, HeaderData, Charset, true);
      Text := sHeaders;
    end;
    Len := Length(Text);
    MaxPos := Len;
  end;
end;

// HeaderData must contain the initial message with Pos & Len of the main header block
// Result contains the new (sub)header block with Pos & Len of the initial headers to replace
function SetLocalCharset(const NewHeaders: String; const HeaderData: THeaderData): THeaderData;

  function ModifyHeaders(const HeaderInfo: THeaderData; MainHeaders: Longbool): THeaderData;
  var
    hdMime: THeaderData;
    fChanged: Longbool;
    sOldHeader, sNewHeader: String;
    iStartPos, iSetPos: Integer;
  begin
    Result := HeaderInfo;
    if Result.Text <> E then begin
      fChanged := false;
      FillChar(hdMime, SizeOf(THeaderData), 0);
      if MainHeaders and (FindHeaderLine(MIME_VERSION, HeaderInfo, true).Text = E) then begin
	with FindHeaderLine('X-PM*', HeaderInfo, true) do begin
	  if Pos = 0 then
	    Result.Text := HeaderInfo.Text + CRLF + MIME_TEXT_HEADERS
	  else Result.Text := Copy(HeaderInfo.Text, 1, pred(Pos)) + MIME_TEXT_HEADERS + CRLF +
			      Copy(HeaderInfo.Text, Pos, succ(HeaderInfo.MaxPos - Pos));
	  with Result do MaxPos := Length(Text);
	end;
	hdMime.Text := Result.Text;
	hdMime.MaxPos := Result.MaxPos;
	fChanged := true;
      end
      else with HeaderInfo do begin
	hdMime.Text := Copy(Text, Pos, Len);
	hdMime.MaxPos := Len;
      end;
      with FindHeaderLine(CT_HEADER, hdMime, true) do begin
	if LoStrAtPos(CT_MULTIPART, Text, 1) then begin
	  FillChar(Result, SizeOf(THeaderData), 0);
	  Exit;
	end;
	if Pos = 0 then begin
	  sOldHeader := CRLF + CT_TEXT_ASCII;
	  Pos := MAXINT;
	end
	else sOldHeader := CT_HEADER + SP + Text;
	iStartPos := LastLoStrPos(CT_CHARSET, sOldHeader) + Length(CT_CHARSET);
	iSetPos := NextLoStrPos(US_ASCII, sOldHeader, iStartPos);
	if iSetPos = 0 then begin
	  iSetPos := NextLoStrPos(ISO_8859_1, sOldHeader, iStartPos);
	  if (iSetPos > 0) and not (sOldHeader[iSetPos + Length(ISO_8859_1)] in ['0'..'9']) then begin
	    sNewHeader := XChgStrAtPos(sOldHeader, WINDOWS_ + _1252, iSetPos, Length(ISO_8859_1));
	  end;
	end
	else sNewHeader := XChgStrAtPos(sOldHeader, WINDOWS_ + IntToStr(GetACP), iSetPos, Length(US_ASCII));
	if sNewHeader <> E then begin
	  hdMime.Text := XChgStrAtPos(hdMime.Text, sNewHeader, Pos, Len);
	  with hdMime do MaxPos := Length(Text);
	  fChanged := true;
	end;
      end;
      with FindHeaderLine(CTE_HEADER, hdMime, true) do begin
	if Pos = 0 then begin
	  sOldHeader := CRLF + CTE_TEXT_7BIT;
	  Pos := MAXINT;
	end
	else sOldHeader := CTE_HEADER + SP + Text;
	iStartPos := NextLoStrPos(CTE_7BIT, sOldHeader, succ(Length(CTE_HEADER)));
	if iStartPos > 0 then begin
	  sNewHeader := XChgStrAtPos(sOldHeader, CTE_8BIT, iStartPos, Length(CTE_7BIT));
	  hdMime.Text := XChgStrAtPos(hdMime.Text, sNewHeader, Pos, Len);
	  with hdMime do MaxPos := Length(Text);
	  fChanged := true;
	end;
      end;
      if fChanged then begin
	Result.Text := hdMime.Text;
	Result.MaxPos := hdMime.MaxPos;
      end
      else FillChar(Result, SizeOf(THeaderData), 0);
    end;
  end;

begin
  Result := HeaderData;
  with Result do begin
    Text := NewHeaders;
    Pos := 1;
    Len := Length(NewHeaders);
    MaxPos := Len;
  end;
  Result := ModifyHeaders(Result, true);
  if Result.Text = E then Result := ModifyHeaders(GetSubHeaderInfo(HeaderData, [], fmoFirstContent), false);
end;

function IsPMMessage(const FileName, PMMsgExtensions: String): Longbool;
begin
  Result := (FirstStrPos(UpperCase(ExtractFileExt(FileName)), PMMsgExtensions) > 0);
end;

function IsPMAttachment(const FileName, MailMessage, PMTempExt: String): String;
var
  hdMessage: THeaderData;
  iPos: Integer;
begin
  Result := LowerCase(ExtractFileExt(FileName));
  if (Result <> E) and not LoStrAtPos(Result, PMTempExt, 1) then Exit;
  with hdMessage do begin
    Text := MailMessage;
    Pos := 1;
    Len := 0;
    MaxPos := Length(MailMessage);
    Recursion := 0;
    Encoding := eNone;
    GetHeaderInfo(hdMessage, [fhoContentType, fhoContentDisposition]);
    iPos := FirstLoStrPos(CDP_FILENAME, CDPHeader);
    if iPos > 0 then begin
      inc(iPos, Length(CDP_FILENAME));
      if (iPos < Length(CDPHeader)) and (CDPHeader[iPos] in [AST, EQ]) then Exit;
    end;
    iPos := FirstLoStrPos(CT_NAME, CTHeader);
    if iPos > 0 then begin
      inc(iPos, Length(CT_NAME));
      if (iPos < Length(CTHeader)) and (CTHeader[iPos] in [AST, EQ]) then Exit;
    end;
  end;
  Result := E;
end;

function GetTabbedPMAddressList(const Addresses: TAddressStrings; var AddressCounts: TAddressCounts;
				var TabbedAddresses: String): Integer;
var
  iMaxLen, iStart, iAddress, iCopy: Integer;
  atIndex: TAddressType;
  pTabbedAddresses: PChar;
  sIndexed, sAddress, sEmail: String;
begin
  Result := 0;
  iMaxLen := 0;
  for atIndex := atTo to atBcc do inc(iMaxLen, succ(Length(Addresses[atIndex])));
  if iMaxLen > 1 then begin
    SetString(TabbedAddresses, nil, succ(iMaxLen));
    pTabbedAddresses := Pointer(TabbedAddresses);
    for atIndex := atTo to atBcc do if Addresses[atIndex] <> E then begin
      sIndexed := Utf8OrAnsi(Addresses[atIndex]);
      iMaxLen := Length(sIndexed);
      iStart := 1;
      while iStart <= iMaxLen do begin
	iAddress := GetAddressLength(sIndexed, iStart);
	if iAddress > 0 then begin
	  sAddress := Copy(sIndexed, iStart, iAddress);
	  sEmail := ExtractEmail(sAddress);
	  if sEmail <> E then begin
	    iCopy := Length(sEmail);
	    Move(Pointer(sEmail)^, pTabbedAddresses^, iCopy);
	  end
	  else begin
	    sAddress := ExtractName(sAddress);
	    iCopy := Length(sAddress);
	    Move(Pointer(sAddress)^, pTabbedAddresses^, iCopy);
	  end;
	  inc(pTabbedAddresses, iCopy);
	  pTabbedAddresses^ := TAB;
	  inc(pTabbedAddresses);
	  inc(iStart, iAddress);
	  case atIndex of
	    atTo: inc(AddressCounts.ToRecipients);
	    atCc: inc(AddressCounts.CcRecipients);
	    atBcc: inc(AddressCounts.BccRecipients);
	  end;
	  inc(Result);
	end;
	while (iStart <= iMaxLen) and (sIndexed[iStart] <> CM) do inc(iStart);
	inc(iStart);
      end;
    end;
    Delete(TabbedAddresses, pTabbedAddresses - Pointer(TabbedAddresses), MAXINT);
  end
  else TabbedAddresses := E;
end;

function AddPMFlags(const CNMFile: String; PMScanLines: Longint; SMimeCapable, PSMimeDecode: Longbool;
		    var Subject: String; var SaveDecoded: Longbool): String;
const
  MAX_HEADER_LEN	= 80;
  CHUNK_SIZE 		= $4000;
var
  fsMessage: TFileStream;
  iLen, iBuf, iPos, iFlags: Integer;
  sMailMessage, sMailHeaders, sCryptHeader, sPMFlags, sEOF: String;
  fRemoveCryptHeader, fModifyCharset, fSetMimeFlag, fReturnMsg, fMultipart: Longbool;
  hdMime, hdTemp, hdEncryptor, hdPMFlags: THeaderData;
  eMessage: TEncoding;
  pNewHeaders: PChar;
begin
  Result := E;
  Subject := E;
  SaveDecoded := false;
  try
    fsMessage := TFileStream.Create(CNMFile, fmOpenReadWrite);
    if (fsMessage <> nil) and (fsMessage.Size > 0) then begin
      try
	with fsMessage do begin
	  iLen := Size;
	  if PMScanLines < 100 then
	    iBuf := 100 * MAX_HEADER_LEN
	  else iBuf := PMScanLines * MAX_HEADER_LEN;
	  iBuf := succ(iBuf div CHUNK_SIZE) * CHUNK_SIZE;
	  if iLen > iBuf then iLen := iBuf;
	  SetString(sMailMessage, nil, iLen);
	  Read(Pointer(sMailMessage)^, iLen);
	end;
	fRemoveCryptHeader := false;
	fModifyCharset := false;
	fSetMimeFlag := true;
	fReturnMsg := false;
	with hdMime do begin
	  Text := sMailMessage;
	  Pos := 1;
	  Len := 0;
	  MaxPos := iLen;
	  Recursion := 0;
	  Encoding := eNone;
	  GetHeaderInfo(hdMime, [fhoContentType]);
	end;
	if hdMime.Len > 0 then begin
	  hdTemp := hdMime;
	  eMessage := GetEncoding(sMailMessage, hdTemp);
	  with hdMime do sMailHeaders := Copy(Text, Pos, Len);
	  hdEncryptor := FindHeaderLine(XPM_ENCRYPTOR, hdMime, true);
	  case eMessage of
	    ePlainEncrypted: begin
	      if hdEncryptor.Text = E then sCryptHeader := Format(XPM_IDWPGP, [XPM_FLAG_ENCRYPTED]);
	      fModifyCharset := true;
	    end;
	    ePlainSigned: begin
	      if hdEncryptor.Text = E then sCryptHeader := Format(XPM_IDWPGP, [XPM_FLAG_SIGNED]);
	      fModifyCharset := true;
	    end;
	    ePMEncrypted: begin
	      if hdEncryptor.Text = E then sCryptHeader := Format(XPM_IDWPGP, [XPM_FLAG_ENCRYPTED]);
	      fSetMimeFlag := false;
	    end;
	    ePMSigned: begin
	      if hdEncryptor.Text = E then sCryptHeader := Format(XPM_IDWPGP, [XPM_FLAG_SIGNED]);
	      if SMimeCapable then fSetMimeFlag := false;
	    end;
	    eSMEncrypted: begin
	      if hdEncryptor.Text = E then sCryptHeader := Format(XPM_IDWPGP, [XPM_FLAG_ENCRYPTED]);
	      fSetMimeFlag := false;
	    end;
	    eSMSigned, eSMInline: begin
	      if hdEncryptor.Text = E then sCryptHeader := Format(XPM_IDWPGP, [XPM_FLAG_SIGNED]);
	      if SMimeCapable then fSetMimeFlag := false;
	    end;
	  else
	    if hdEncryptor.Text <> E then
	      fRemoveCryptHeader := true
	    else Exit;
	  end;
	  if not fRemoveCryptHeader then begin
	    fReturnMsg := PSMimeDecode and not fSetMimeFlag;
	    SaveDecoded := (eMessage in [ePMEncrypted, eSMEncrypted, eSMInline]);
	    if fReturnMsg then Subject := MimeDecodeHeaders(FindHeaderLine(H_SUBJECT, hdMime, false).Text);
	    hdPMFlags := FindHeaderLine(XPM_FLAGS, hdMime, true);
	    if hdPMFlags.Pos = 0 then begin
	      if fReturnMsg then fSetMimeFlag := true;
	      fMultipart := LoStrAtPos(CT_MULTIPART, hdMime.CTHeader, 1);
	      iFlags := (ord(fSetMimeFlag = true) * MIME) or ENCRYPTED or IS_SMTP or (ord(fMultipart = true) * MULTIPART);
	      if FindHeaderLine(X_CONFIRM_TO, hdMime, true).Len > 0 then iFlags := iFlags or CONFIRMATION;
	      if StrAtPos('1', FindHeaderLine(X_PRIORITY, hdMime, true).Text, 1)
	      or LoStrAtPos('high', FindHeaderLine(H_IMPORTANCE, hdMime, true).Text, 1)
	      or LoStrAtPos('urgent', FindHeaderLine(H_PRIORITY, hdMime, true).Text, 1) then begin
		iFlags := iFlags or URGENT;
	      end;
	      sPMFlags := XPM_FLAGS + SP + IntToStr(iFlags) + ' 0 0 ' + ExtractFileName(CNMFile);
	      iPos := FindHeaderLine(XPM_PLACEHOLDER, hdMime, true).Pos;
	      if iPos > 0 then begin
		if sCryptHeader <> E then begin
		  iLen := Length(sCryptHeader);
		  SetLength(sMailHeaders, Length(sMailHeaders) + iLen);
		  pNewHeaders := Pointer(sMailHeaders);
		  inc(pNewHeaders, pred(iPos + iLen));
		  Move(Pointer(sPMFlags)^, pNewHeaders^, Length(sPMFlags));
		  dec(pNewHeaders, iLen);
		  Move(Pointer(sCryptHeader)^, pNewHeaders^, Length(sCryptHeader));
		  inc(pNewHeaders, Length(sPMFlags) + iLen);
		  iLen := pNewHeaders - Pointer(sMailHeaders);
		  FillByte(pNewHeaders^, Length(sMailHeaders) - iLen, ord(SP));
		end
		else begin
		  iLen := Length(sPMFlags);
		  pNewHeaders := Pointer(sMailHeaders);
		  Move(Pointer(sPMFlags)^, pNewHeaders[pred(iPos)], iLen);
		  iPos := NextCharPos(DOT, sMailHeaders, iPos + iLen);
		  pNewHeaders[pred(iPos)] := SP;
		end;
	      end
	      else if sCryptHeader = E then
		sMailHeaders := sMailHeaders + CRLF + sPMFlags
	      else sMailHeaders := sMailHeaders + CRLF + sCryptHeader + sPMFlags;
	    end
	    else begin
	      if fReturnMsg then begin
		with hdPMFlags do sPMFlags := Copy(Text, 1, pred(FirstCharPos(SP, Text)));
		if sPMFlags <> E then begin
		  iFlags := StrToIntDef(sPMFlags, 0) and not ENCRYPTED or MIME;
		  sPMFlags := XPM_FLAGS + SP + IntToStr(iFlags) + ' 0 0 ' + ExtractFileName(CNMFile);
		  iLen := Length(sPMFlags);
		  with hdPMFlags do begin
		    if iLen <= Len then begin
		      pNewHeaders := Pointer(sMailHeaders);
		      inc(pNewHeaders, pred(Pos));
		      Move(Pointer(sPMFlags)^, pNewHeaders^, iLen);
		      inc(pNewHeaders, iLen);
		      FillByte(pNewHeaders^, Len - iLen, ord(SP));
		    end
		    else sMailHeaders := XChgStrAtPos(sMailHeaders, sPMFlags, Pos, Len);
		  end;
		end;
	      end;
	      if sCryptHeader <> E then Insert(sCryptHeader, sMailHeaders, hdPMFlags.Pos);
	    end;
	    if fModifyCharset then begin
	      hdTemp := SetLocalCharset(sMailHeaders, hdMime);
	      if hdTemp.Text <> E then begin
		if hdTemp.Pos > 1 then begin
		  with fsMessage do begin
		    iPos := Position;
		    iLen := Size;
		    if iPos < iLen then begin
		      dec(iLen, iPos);
		      SetString(sEOF, nil, iLen);
		      Read(Pointer(sEOF)^, iLen);
		    end
		    else iLen := 0;
		    Position := 0;
		    Write(Pointer(sMailHeaders)^, Length(sMailHeaders));
		    with hdMime do Write(Text[Pos + Len], pred(hdTemp.Pos - Len));
		    with hdTemp do begin
		      inc(Pos, Len);
		      Write(Pointer(Text)^, MaxPos);
		      Write(hdMime.Text[Pos], succ(hdMime.MaxPos) - Pos);
		    end;
		    if iLen > 0 then Write(Pointer(sEOF)^, iLen);
		  end;
		  Exit;
		end
		else sMailHeaders := hdTemp.Text;
	      end;
	    end;
	    if not fReturnMsg and (hdTemp.Text = E) and (sCryptHeader = E) and (sPMFlags = E) then Exit;
	  end
	  else with hdEncryptor do Delete(sMailHeaders, Pos, Len + Length(CRLF));
	  with fsMessage do begin
	    iPos := Position;
	    iLen := Size;
	    if iPos < iLen then begin
	      dec(iLen, iPos);
	      SetString(sEOF, nil, iLen);
	      Read(Pointer(sEOF)^, iLen);
	    end
	    else begin
	      sEOF := E;
	      iLen := 0;
	    end;
	    if fReturnMsg then begin
	      with hdEncryptor do begin
		iPos := Pos;
		iLen := Len + Length(CRLF);
	      end;
	      if (iPos > 0) and LoStrAtPos(XPM_ENCRYPTOR, sMailHeaders, iPos) then Delete(sMailHeaders, iPos, iLen);
	      with hdMime do Result := sMailHeaders + Copy(Text, Pos + Len, MaxPos - Len) + sEOF;
	    end
	    else begin
	      Position := 0;
	      Write(Pointer(sMailHeaders)^, Length(sMailHeaders));
	      with hdMime do Write(Text[Pos + Len], MaxPos - Len);
	      if iLen > 0 then Write(Pointer(sEOF)^, iLen);
	      Truncate;
	    end;
	  end;
	end;
      finally
	fsMessage.Free;
      end;
    end;
  except
  end;
end;

function RemovePMCryptoHeaders(var MailMessage: String; const Charset: String; QPEncodeText: Longbool): String;

  procedure ApplyQPEncoding(var MailMessage: String; var HD: THeaderData; const Boundary: String);
  var
    sHeaders, sBody, sText: String;
    iPos, iLen, iBeg, iEnd: Integer;
    hdCTE: THeaderData;
    pText: PChar;
  begin
    with HD do begin
      if LoStrAtPos(CT_TEXT, CTHeader, 1)
      and not (LoStrAtPos(CTE_B64, CTEHeader, 1) or LoStrAtPos(CTE_QP, CTEHeader, 1)) then begin
	sHeaders := Copy(MailMessage, Pos, Len);
	iPos := Pos + Len;
	if Boundary <> E then
	  iLen := NextLoStrPos(DD + Boundary, MailMessage, iPos) - iPos
	else iLen := succ(MaxPos - iPos);
	sBody := QPEncode(UnwrapArmoredKeys(Copy(MailMessage, iPos, iLen)), E);
	with hdCTE do begin
	  Text := sHeaders;
	  Pos := 1;
	  Len := HD.Len;
	  MaxPos := Len;
	  Recursion := 0;
	  Encoding := eNone;
	end;
	hdCTE := FindHeaderLine(CTE_HEADER, hdCTE, true);
	with hdCTE do sHeaders := XChgStrAtPos(sHeaders, CTE_QUOTED_PRINTABLE, Pos, Len);
	Text := E;
	if Boundary <> E then
	  MailMessage := Copy(MailMessage, 1, pred(HD.Pos)) + sHeaders + sBody + Copy(MailMessage, iPos + iLen, MAXINT)
	else MailMessage := sHeaders + sBody;
	Text := MailMessage;
	Len := Length(sHeaders);
	MaxPos := Length(Text);
      end
      else if LoStrAtPos(CTE_QP, CTEHeader, 1)
      and LoStrAtPos(CTD_AINFO, FindHeaderLine(CTD_HEADER, HD, false).Text, 1) then begin
	iPos := Pos + Len;
	if Boundary <> E then
	  iLen := NextLoStrPos(DD + Boundary, MailMessage, iPos) - iPos
	else iLen := succ(MaxPos - iPos);
	sBody := Copy(MailMessage, iPos, iLen);
	SetString(sText, nil, iLen);
	pText := Pointer(sText);
	iBeg := 1;
	repeat
	  iEnd := NextStrPos(SP + CRLF, sBody, iBeg);
	  iPos := iEnd + Length(SP) + Length(CRLF);
	  while (iEnd >= iBeg) and (sBody[iEnd] <= SP) do dec(iEnd);
	  iLen := succ(iEnd - iBeg);
	  if iLen > 0 then begin
	    Move(sBody[iBeg], pText^, iLen);
	    inc(pText, iLen);
	    PDChar(pText)^ := CRLF;
	    inc(pText, Length(CRLF));
	    iBeg := iPos;
	  end
	  else begin
	    iLen := succ(Length(sBody) - iBeg);
	    if iLen > 0 then begin
	      Move(sBody[iBeg], pText^, iLen);
	      inc(pText, iLen);
	    end;
	    Break;
	  end;
	until iLen = 0;
	iLen := Length(sBody);
	iPos := pText - Pointer(sText);
	Delete(sText, succ(iPos), MAXINT);
	if iPos < iLen then begin
	  MailMessage := XChgStrAtPos(MailMessage, sText, Pos + Len, iLen);
	  Text := MailMessage;
	  MaxPos := Length(Text);
	end;
      end;
    end;
  end;

  procedure RemoveMainCryptoHeader(var MailMessage: String; var HD: THeaderData;
				   const Charset: String; QPEncodeText: Longbool);
  begin
    with FindHeaderLine(XPM_ENCRYPTOR, HD, true) do if Len > 0 then begin
      inc(Len, Length(CRLF));
      HD.Text := E;
      Delete(MailMessage, Pos, Len);
      HD.Text := MailMessage;
      dec(HD.Len, Len);
      dec(HD.MaxPos, Len);
    end;
    HD.Text := E;
    MimeFoldNameParameters(MailMessage, HD, Charset, true);
    HD.Text := MailMessage;
    HD.MaxPos := Length(MailMessage);
    if QPEncodeText then ApplyQPEncoding(MailMessage, HD, E);
  end;

  procedure RemoveSubCryptoHeaders(var MailMessage: String; var HD: THeaderData;
				   const Charset: String; QPEncodeText: Longbool);
  var
    sBoundary, sSubMessage: String;
    hdSub, hdC: THeaderData;
    fIsDigest: Longbool;
    iPos, iLen: Integer;
  begin
    sBoundary := GetBoundary(HD);
    if sBoundary <> E then begin
      with HD do fIsDigest := LoStrAtPos(CT_DIGEST, CTHeader, NextNonSpaceChar(CTHeader, succ(Length(CT_MULTIPART))));
      repeat
	hdSub := GetNextHeaderInfo(HD, sBoundary, [fhoContentType, fhoContentTransferEncoding]);
	if hdSub.Len > 0 then begin
	  hdC := FindHeaderLine(CDP_HEADER, hdSub, true);
	  hdSub.Text := E;
	  hdSub.MaxPos := 0;
	  if LoStrAtPos(CT_MULTIPART, hdSub.CTHeader, 1) then begin
	    HD.Pos := hdSub.Pos;
	    HD.Len := hdSub.Len;
	    HD.CTHeader := hdSub.CTHeader;
	    RemoveSubCryptoHeaders(MailMessage, HD, Charset, QPEncodeText);
	  end
	  else if LoStrAtPos(CT_MESSAGE, hdSub.CTHeader, 1) or fIsDigest then begin
	    with hdSub do iPos := Pos + Len;
	    while (iPos <= HD.MaxPos) and (MailMessage[iPos] <= SP) do inc(iPos);
	    iLen := NextLoStrPos(DD + sBoundary, MailMessage, iPos) - iPos;
	    sSubMessage := Copy(MailMessage, iPos, iLen);
	    RemovePMCryptoHeaders(sSubMessage, Charset, QPEncodeText);
	    HD.Text := E;
	    MailMessage := XChgStrAtPos(MailMessage, sSubMessage, iPos, iLen);
	    HD.Text := MailMessage;
	    HD.MaxPos := Length(MailMessage);
	  end
	  else if hdC.Pos > 0 then begin
	    HD.Text := E;
	    MimeFoldNameParameters(MailMessage, hdSub, Charset, true);
	    HD.Text := MailMessage;
	    HD.MaxPos := Length(MailMessage);
	  end;
	  HD.Pos := hdSub.Pos;
	  HD.Len := hdSub.Len;
	  HD.CTHeader := hdSub.CTHeader;
	  HD.CTEHeader := hdSub.CTEHeader;
	  if QPEncodeText then ApplyQPEncoding(MailMessage, HD, sBoundary);
	end
	else Break;
      until hdSub.Len = 0;
    end;
  end;

var
  hdResult: THeaderData;
begin
  with hdResult do begin
    Text := MailMessage;
    Pos := 1;
    Len := 0;
    MaxPos := Length(MailMessage);
    Recursion := 0;
    Encoding := eNone;
  end;
  try
    GetHeaderInfo(hdResult, [fhoContentType, fhoContentTransferEncoding]);
    RemoveMainCryptoHeader(MailMessage, hdResult, Charset, QPEncodeText);
    RemoveSubCryptoHeaders(MailMessage, hdResult, Charset, QPEncodeText);
  finally
    Result := MailMessage;
  end;
end;

// All header folding and quoted-printable encoding required will be done on return with
// Result.Pos + Result.Len pointing to the start position of the main Content-... headers block
function PreparePSMimeEncoding(const MailMessage, Charset: String; EncodeFlags: Integer): THeaderData;
var
  sBody: String;
begin
  Result.Text := MailMessage;
  RemovePMCryptoHeaders(Result.Text, Charset, true);
  with Result do begin
    Pos := 1;
    Len := 0;
    MaxPos := Length(Text);
    Recursion := 0;
    Encoding := eNone;
    GetHeaderInfo(Result, []);
  end;
  with Result do begin
    MaxPos := pred(Pos + Len);
    sBody := ExtractStr(Text, succ(MaxPos), MAXINT);
    Len := ReplaceContentHeaders(Result, false, TrimRight(Format(XPM_IDWPGP, [EncodeFlags])), Charset);
    Text := Text + sBody;
    MaxPos := Length(Text);
  end;
end;


// TSectionNode ------------------------------------------------------------------------------------

function TSectionNode.GetFirstChild: TSectionNode;
begin
  Result := TSectionNode(inherited GetFirstChild);
end;

function TSectionNode.GetFullMultipartSection: String;
begin
  with HeaderData do Result := Copy(Text, Pos, succ(MaxPos - Pos));
end;

function TSectionNode.GetRelatedSectionsAsRtf: String;
var
  fsRtf: TFileSections;
  iCount: Integer;
begin
  if FSectionType in [mdtHtmlText..mdtRtfText] then begin
    fsRtf := TFileSections.Create(Self, true);
    if fsRtf <> nil then with fsRtf do begin
      try
	iCount := Count;
	if iCount > 0 then Result := ModifiedMainSection;
	if iCount > 1 then begin
	  if Parent <> nil then with Parent do begin
	    if (SectionType <> mdtMultipartMixed) and (Parent = nil) then Exit;
	    if (SectionType = mdtMultipartAlternative) and (Parent.SectionType <> mdtMultipartMixed) then Exit;
	  end;
	  with MimeMsgTree do FHasAttachments := (FFileAttachments > 0);
	end
	else FIncludeImages := false;
      finally
	Free;
      end;
    end;
  end
  else Result := E;
end;

function TSectionNode.QPFileName(const Extension: String): String;
begin
  if FSectionType in [mdtUnknownData, mdtFile] then
    Result := QPEncode(GetSectionName + Extension, Charset)
  else Result := E;
end;

function TSectionNode.SaveRelatedSectionsAsHtml(const HtmlFileName: String): Longbool;
var
  fsHtml: TFileSections;
  iCount: Integer;
begin
  Result := false;
  if FSectionType = mdtHtmlText then begin
    fsHtml := TFileSections.Create(Self, false);
    if fsHtml <> nil then with fsHtml do begin
      try
	iCount := Count;
	if iCount <= 1 then FIncludeImages := false;
	Result := (iCount > 0) and WriteToFiles(HtmlFileName);
      finally
	Free;
      end;
    end;
  end;
end;

function TSectionNode.SetEncodedData(const Data, MimeCode: String; XPMHeaderFlags: Integer;
				     AppendPGPExtension, InsertSignature: Longbool): Integer;

  function SignatureHeaders: String;
  var
    sFileName, sCTDHeader: String;
  begin
    sFileName := QPFileName(SIG);
    if FSectionCID = E then
      sCTDHeader := CTD_HEADER + SP + FileTypeInfo(SIG, E)
    else sCTDHeader := CID_HEADER + SP + LT + FSectionCID + GT;
    Result := DD + Boundary + CRLF +
	      CT_HEADER + SP + APP_PGP + PM_SIG + SC + SP + CT_NAME + EQ + CRLF +
	      SP + sFileName + CRLF +
	      CTE_HEADER + SP + CTE_7BIT + CRLF +
	      sCTDHeader + CRLF +
	      CDP_HEADER + SP + CDP_ATTACHMENT + SC + SP + CDP_FILENAME + EQ + CRLF +
	      SP + sFileName;
  end;

  function EncryptionHeaders(MimeCode: String; XPMHeaderFlags: Integer; AppendPGPExtension: Longbool): String;
  var
    sFileName, sCTDHeader, sXPMHeader, sCTType: String;
  begin
    if AppendPGPExtension then
      sFileName := QPFileName(PGP)
    else sFileName := QPFileName(E);
    if FSectionCID = E then
      sCTDHeader := CTD_HEADER + SP + FileTypeInfo(MimeDecodeHeaders(sFileName), E)
    else sCTDHeader := CID_HEADER + SP + LT + FSectionCID + GT;
    if XPMHeaderFlags = 0 then
      sXPMHeader := E
    else sXPMHeader := SC + CRLF + SP + X_IDWPGP + DigitToHex(XPMHeaderFlags);
    if AppendPGPExtension then
      sCTType := APP_PGP + PM_ENC
    else sCTType := CT_APPLICATION + CT_OCTET_STREAM;
    Result := CT_HEADER + SP + sCTType + SC + SP + CT_NAME + EQ + CRLF +
	      SP + sFileName + CRLF +
	      CTE_HEADER + SP + MimeCode + CRLF +
	      sCTDHeader + CRLF +
	      CDP_HEADER + SP + CDP_ATTACHMENT + SC + SP + CDP_FILENAME + EQ + CRLF +
	      SP + sFileName +
	      sXPMHeader;
  end;

  procedure SetCTEHeader(const MimeCode, Charset: String; XPMHeaderFlags: Integer; AppendPGPExtension: Longbool);
  var
    snHeader: TSectionNode;
    hd: THeaderData;
  begin
    snHeader := TSectionNode(GetPrev);
    if (snHeader <> nil) then with snHeader do begin
      FillChar(hd, SizeOf(THeaderData), 0);
      with hd do begin
	Text := snHeader.FSectionData;
	MaxPos := Length(Text);
      end;
      if FSectionType in [mdtMainHeaders, mdtSubHeaders] then begin
	if Self.FSectionType = mdtFile then begin
	  if FSectionType = mdtMainHeaders then begin
	    ReplaceContentHeaders(hd, true, E, E);
	    FSectionData := hd.Text + CRLF + EncryptionHeaders(MimeCode, XPMHeaderFlags, AppendPGPExtension);
	  end
	  else begin
	    FSectionData := DD + Self.Boundary + CRLF + EncryptionHeaders(MimeCode, XPMHeaderFlags, AppendPGPExtension);
	  end;
	end
	else begin
	  hd.CTEHeader := CTE_HEADER + SP + MimeCode + CRLF;
	  ReplaceContentHeaders(hd, false, E, Charset);
	  with FindHeaderLine(XPM_ENCRYPTOR, hd, true) do begin
	    if Len > 0 then hd.Text[pred(Pos + Len)] := DigitToHex(XPMHeaderFlags);
	  end;
	  FSectionData := hd.Text;
	end;
      end;
    end;
  end;

var
  snSignature: TSectionNode;
begin
  Result := AbsoluteIndex;
  if InsertSignature then with MimeMsgTree do begin
    if (InsertSectionAfter(SectionNodes[Result], E, SignatureHeaders, mdtSubHeaders) <> nil) then begin
      inc(Result);
      snSignature := InsertSectionAfter(SectionNodes[Result], E, Data, mdtFile);
      if (snSignature = nil) or (snSignature.SectionData = E) then
	Items.Delete(SectionNodes[Result])
      else inc(Result);
      FSectionData := B64Encode(FSectionData);
    end;
  end
  else begin
    if MimeCode = CTE_B64 then
      FSectionData := B64Encode(Data)
    else if MimeCode = CTE_QP then
      FSectionData := QPEncode(Data, E)
    else FSectionData := Data;
  end;
  SetCTEHeader(MimeCode, Charset, XPMHeaderFlags, AppendPGPExtension);
end;

function TSectionNode.SetPMFlags(IsMultipart, IsEncrypted, IsMime, IsRead: Boolean; const FileName: String): String;
const
  FILTER = not (MULTIPART or ENCRYPTED or MIME or HAS_BEEN_READ);
var
  pResult: PChar;
  hdPMFlags: THeaderData;
  sPMFlagsLine: String;
  iFlags: Integer;
begin
  Result := GetFullMultipartSection;
  pResult := PChar(Result);
  hdPMFlags := FindHeaderLine(XPM_FLAGS, HeaderData, true);
  if hdPMFlags.Pos > 0 then begin
    inc(pResult, pred(hdPMFlags.Pos));
    sPMFlagsLine := hdPMFlags.Text;
      iFlags := (ord(IsMultipart) * MULTIPART) or (ord(IsEncrypted) * ENCRYPTED) or
		(ord(IsMime) * MIME) or (ord(IsRead) * HAS_BEEN_READ) or IS_SMTP;
    sPMFlagsLine := XPM_FLAGS + SP + IntToStr(iFlags) + ' 0 0 ' + ExtractFileName(FileName);
    Move(Pointer(sPMFlagsLine)^, pResult^, Length(sPMFlagsLine));
  end
  else begin
    hdPMFlags := FindHeaderLine(XPM_PLACEHOLDER, HeaderData, true);
    if hdPMFlags.Pos > 0 then begin
      inc(pResult, pred(hdPMFlags.Pos));
      iFlags := (ord(IsMultipart) * MULTIPART) or (ord(IsEncrypted) * ENCRYPTED) or
		(ord(IsMime) * MIME) or (ord(IsRead) * HAS_BEEN_READ) or IS_SMTP;
      sPMFlagsLine := XPM_FLAGS + SP + IntToStr(iFlags) + ' 0 0 ' + ExtractFileName(FileName);
      Move(Pointer(sPMFlagsLine)^, pResult^, Length(sPMFlagsLine));
    end
    else begin
      iFlags := (ord(IsMultipart) * MULTIPART) or (ord(IsEncrypted) * ENCRYPTED) or
		(ord(IsMime) * MIME) or (ord(IsRead) * HAS_BEEN_READ) or IS_SMTP;
      sPMFlagsLine := XPM_FLAGS + SP + IntToStr(iFlags) + ' 0 0 ' + ExtractFileName(FileName) + CRLF;
      Insert(sPMFlagsLine, Result, FirstStrPos(PARA, Result) + Length(CRLF));
    end;
  end;
end;

procedure TSectionNode.CopyHeaderPosLen(const HeaderData: THeaderData);
begin
  with FHeaderDataEx.HeaderData do begin
    Pos := HeaderData.Pos;
    Len := HeaderData.Len;
  end;
end;

function TSectionNode.GetHeader(const Name: String; DropComments: Longbool): String;
var
  hd: THeaderData;
begin
  if FSectionType = mdtMainHeaders then begin
    FillChar(hd, SizeOf(THeaderData), 0);
    with hd do begin
      Text := FSectionData;
      MaxPos := Length(FSectionData);
    end;
    Result := FindHeaderLine(Name, hd, DropComments).Text;
  end
  else Result := E;
end;

function TSectionNode.GetUtf8Date: String;
begin
  Result := AnsiStringToUtf8(GetHeader(H_DATE, true));
end;

function TSectionNode.GetUtf8UTCDate: String;
begin
  Result := AnsiStringToUtf8(DateHeaderToUTC(GetHeader(H_DATE, true), FUTCDateTime));
end;

function TSectionNode.GetShortUtf8Date: String;
begin
  Result := AnsiStringToUtf8(FormatDateTime('ddddd, t', UTCToLocalTime(GetUTCDateTime)));
end;

function TSectionNode.GetUtf8Sender: String;
begin
  Result := DecodeIdns(MimeDecodeHeaders(GetHeader(H_FROM, false)));
end;

function TSectionNode.GetUtf8Recipient: String;
begin
  Result := DecodeIdns(MimeDecodeHeaders(GetFirstAddress(GetHeader(H_TO, false))));
end;

function TSectionNode.GetUtf8Subject: String;
begin
  Result := MimeDecodeHeaders(GetHeader(H_SUBJECT, false));
end;

function TSectionNode.GetUtf8ReplyTo: String;
begin
  Result := DecodeIdns(MimeDecodeHeaders(GetHeader(H_REPLY_TO, false)));
end;

function TSectionNode.GetFileExt: String;
var
  iPos: Integer;
begin
  Result := LowerCase(ExtractFileExt(GetFileName));
  if (Result = E) and (FSectionType = mdtFile) then begin
    Result := Utf8ToAnsi(GetSectionName);
    iPos := LastCharPos(SL, Result);
    if iPos > 0 then
      Result := E
    else Result := DOT + LowerCase(Copy(Result, succ(iPos), MAXINT));
  end;
end;

function TSectionNode.GetFileName: String;
var
  sSectionName: String;
  iPos: Integer;
begin
  sSectionName := Utf8ToAnsi(GetSectionName);
  iPos := FirstStrPos(CL + SP, sSectionName);
  if iPos > 0 then
    Result := LegalFileName(Copy(sSectionName, iPos + Length(CL) + Length(SP), MAXINT))
  else Result := LegalFileName(sSectionName);
  case FSectionType of
    mdtUnknownData, mdtSignature, mdtFile: Exit;
    mdtMainHeaders, mdtPlainText, mdtAltText: if ExtractFileExt(Result) = E then Result := Result + TXT;
    mdtHtmlText: Result := Copy(Result, 1, pred(LastCharPos(SP, Result))) + HTM;
    mdtRtfText: Result := Copy(Result, 1, pred(LastCharPos(SP, Result))) + RTF;
  else
    Result := E;
  end;
end;

function TSectionNode.GetFileNameFromHeaders: String;
begin
  if FSectionType = mdtMainHeaders then
    Result := LegalFileName(GetUtf8Subject) + SP + LB + FormatDateTime('yyyy-mm-dd-hh-nn-ss', GetUTCDateTime) + RB
  else Result := E;
end;

function TSectionNode.GetUTCDateTime: TDateTime;
begin
  if (FUTCDateTime = 0) and (FSectionType = mdtMainHeaders) then DateHeaderToUTC(GetUtf8Date, FUTCDateTime);
  Result := FUTCDateTime;
end;

function TSectionNode.GetHasFormattedText: Longbool;
begin
  Result := (FSectionType in [mdtHtmlText..mdtRtfText]);
end;

function TSectionNode.GetIsPSMEncoded: Longbool;
var
  snParent: TSectionNode;
begin
  Result := false;
  snParent := GetParent;
  while snParent <> nil do begin
    if snParent.SectionType in [mdtMultipartEncrypted..mdtMultipartSigned] then begin
      Result := true;
      Exit;
    end
    else snParent := snParent.GetParent;
  end;
end;

function TSectionNode.GetAnsiMsgHeaders: String;
var
  hd, hdResult: THeaderData;
  pText, pResult: PAnsiChar;
begin
  if FSectionType = mdtMainHeaders then begin
    FillChar(hd, SizeOf(THeaderData), 0);
    with hd do begin
      Text := FSectionData;
      pText := Pointer(FSectionData);
      MaxPos := Length(FSectionData);
      SetString(Result, nil, MaxPos);
      pResult := Pointer(Result);
      dec(pText);
    end;
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_DATE, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_FROM, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_TO, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_CC, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_BCC, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_SUBJECT, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    hd.Pos := 0;
    hdResult := FindHeaderLine(H_REPLY_TO, hd, true);
    with hdResult do pResult := StrLAppend(pResult, pText + Pos, Len, true);
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
  Result := MimeDecodeHeaders(Result);
end;

function TSectionNode.GetUtf8MainHeaders: String;
var
  sHeaders, sDate, sFrom, sTo, sSubject, sReplyTo: String;
  hdSection: THeaderData;
begin
  if FSectionType = mdtMainHeaders then begin
    sHeaders := FSectionData;
    FillChar(hdSection, SizeOf(THeaderData), 0);
    with hdSection do begin
      Text := sHeaders;
      MaxPos := Length(sHeaders);
    end;
    sDate := GetUtf8UTCDate;
    if sDate = E then sDate := MimeMsgTree.FMissingHeaderText;
    sFrom := GetUtf8Sender;
    if sFrom = E then
      sFrom := MimeMsgTree.FMissingHeaderText
    else RemoveBackslashes(sFrom);
    sTo := GetUtf8Recipient;
    if sTo = E then
      sTo := MimeMsgTree.FMissingHeaderText
    else RemoveBackslashes(sTo);
    sSubject := GetUtf8Subject;
    if sSubject = E then
      sSubject := MimeMsgTree.FMissingHeaderText
    else RemoveBackslashes(sSubject);
    sReplyTo := GetUtf8ReplyTo;
    RemoveBackslashes(sReplyTo);
    if sReplyTo <> E then
      sReplyTo := CRLF + MimeMsgTree.FReplyToHeaderName + SP + sReplyTo + PARA
    else sReplyTo := PARA;
    with MimeMsgTree do begin
      Result := FDateHeaderName + TAB + sDate + CRLF +
		FSenderHeaderName + TAB + sFrom + CRLF +
		FRecipientHeaderName + TAB + sTo + CRLF +
		FSubjectHeaderName + TAB + sSubject + sReplyTo;
    end;
  end
  else Result := E;
end;

function TSectionNode.GetMimeMsgTree: TMimeMsgTree;
begin
  Result := TMimeMsgTree(TreeView);
end;

function TSectionNode.GetSectionIcon: TIcon;
begin
  Result := TIcon.Create;
  if MimeMsgTree.Images <> nil then MimeMsgTree.Images.GetIcon(ImageIndex, Result);
end;

function TSectionNode.GetSelectedIcon: TIcon;
begin
  Result := TIcon.Create;
  if MimeMsgTree.Images <> nil then MimeMsgTree.Images.GetIcon(SelectedIndex, Result);
end;

function TSectionNode.GetSectionText: String;
var
  mdtBuffer: TMsgDataType;
begin
  case FSectionType of
    mdtMainHeaders: begin
      FIncludeImages := false;
      Result := MimeDecodeHeaders(FSectionData);
      if mdoReturnUnicode in MimeMsgTree.MimeDecodeOptions then
	Result := Utf8ToWide(Result)
      else Result := Utf8ToAnsi(Result);
    end;
  else
    if (FSectionType in [mdtPlainText..mdtAltText])
    or ((FSectionType = mdtFile) and StrAtPos(HTM, GetFileExt, 1)) then begin
      if mdoReturnUnicode in MimeMsgTree.MimeDecodeOptions then begin
	mdtBuffer := FSectionType;
	try
	  if mdoReturnRtf in MimeMsgTree.MimeDecodeOptions then begin
	    if FSectionType in [mdtHtmlText, mdtFile] then begin
	      FSectionType := mdtHtmlText;
	      if FIncludeImages then begin
		Result := GetRelatedSectionsAsRtf;
		Exit;
	      end;
	    end;
	    Result := ReturnFormattedText(FSectionData, Self);
	  end
	  else Result := ReturnUnicodeText(FSectionData, Self);
	finally
	  FSectionType := mdtBuffer;
	end;
      end
      else Result := ReturnPlainAnsiText(FSectionData, Self);
      FIncludeImages := false;
    end
    else Result := FSectionData;
  end;
end;

function TSectionNode.GetSectionCID: String;
begin
  Result := FSectionCID;
  if Result = E then Result := GetFileName;
end;

procedure TSectionNode.SetSectionCID(const Value: String);
begin
  FSectionCID := DelAngles(Unescape(Value, false, PC));
end;

function TSectionNode.GetSectionName: String;
begin
  Result := Text;
end;

procedure TSectionNode.SetSectionName(const Value: String);
begin
  Text := Value;
end;

function TSectionNode.GetParent: TSectionNode;
begin
  Result := TSectionNode(inherited Parent);
end;

function TSectionNode.GetChild(Index: Integer): TSectionNode;
begin
  Result := TSectionNode(inherited Item[Index]);
end;


// TTextList ---------------------------------------------------------------------------------------

function TTextList.GetAnsiText: String;
var
  iLength, iIndex: Integer;
  sAnsiText: String;
  pResult: PChar;
begin
  if Count > 0 then begin
    iLength := 0;
    for iIndex := 0 to pred(Count) do inc(iLength, Length(Strings[iIndex]) + Length(aCRLF));
    SetString(Result, nil, iLength);
    pResult := Pointer(Result);
    for iIndex := 0 to pred(Count) do begin
      sAnsiText := ReturnPlainAnsiText(Strings[iIndex], SectionNodes[iIndex]);
      iLength := Length(sAnsiText);
      if iLength > 0 then begin
	System.Move(Pointer(sAnsiText)^, pResult^, iLength);
	inc(pResult, iLength);
	if (iLength < 2) or not (sAnsiText[iLength - 1] in LFCR) then begin
	  PDChar(pResult)^ := aCRLF;
	  inc(pResult, Length(aCRLF));
	end;
      end;
    end;
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function TTextList.GetUnicodeText: String;
var
  iLength, iIndex: Integer;
  sUnicodeText: String;
  pResult: PChar;
begin
  if Count > 0 then begin
    iLength := 0;
    for iIndex := 0 to pred(Count) do inc(iLength, (Length(Strings[iIndex]) + Length(aCRLF)) shl 1 + Length(wNUL));
    SetString(Result, nil, iLength);
    pResult := Pointer(Result);
    for iIndex := 0 to pred(Count) do begin
      sUnicodeText := ReturnUnicodeText(Strings[iIndex], SectionNodes[iIndex]);
      iLength := Length(sUnicodeText) - Length(wNUL);
      if iLength > 0 then begin
	System.Move(Pointer(sUnicodeText)^, pResult^, iLength);
	inc(pResult, iLength);
	if (iLength < 4) or not (sUnicodeText[iLength - 3] in LFCR) then begin
	  PQChar(pResult)^ := wCRLF;
	  inc(pResult, Length(wCRLF));
	end;
      end;
    end;
    PDChar(pResult)^ := wNUL;
    inc(pResult, Length(wNUL));
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function TTextList.GetSectionNode(Index: Integer): TSectionNode;
begin
  Result := TSectionNode(Objects[Index]);
end;


// TFileSections ----------------------------------------------------------------------------------

constructor TFileSections.Create(SectionNode: TSectionNode; ReturnRtf: Longbool);
var
  snParent, snFile: TSectionNode;
  iIndex, iLen, iNew: Integer;
  sDefaultBmp: String;
  cidl: TCIDList;
begin
  inherited Create;
  snParent := nil;
  iIndex := -1;
  with SectionNode do begin
    if (AddItem(GetTempDir + FileName, SectionData, SectionCID) > -1)
    and (SectionType in [mdtHtmlText..mdtAltText]) and (Parent <> nil) then begin
      case Parent.SectionType of
	mdtMultipartMixed, mdtMultipartRelated: begin
	  snParent := Parent;
	  iIndex := succ(snParent.IndexOf(SectionNode));
	end;
	mdtMultipartAlternative: begin
	  snParent := Parent.Parent;
	  if (snParent <> nil) and (snParent.SectionType in [mdtMultipartMixed, mdtMultipartRelated]) then begin
	    iIndex := succ(snParent.IndexOf(Parent));
	  end;
	end;
      end;
    end;
    if SectionType = mdtHtmlText then begin
      sDefaultBmp := GetDefaultBmp;
      cidl := TCIDList.Create(SectionNode, ReturnRtf);
      try
	if (snParent <> nil) and (iIndex > -1) then begin
	  for iIndex := iIndex to pred(snParent.Count) do begin
	    snFile := snParent.Child[iIndex];
	    with snFile do begin
	      if (SectionType = mdtFile) and (SectionData <> E) then begin
		iLen := 0;
		if ReturnRtf then begin
		  iNew := AddItem(TempFileName(FileExt), GetImageDataAsRtf(SectionData, true), SectionCID);
		  if iNew > -1 then iLen := Length(FileData[iNew]);
		end
		else begin
		  iNew := AddItem(TempFileName(FileExt), SectionData, SectionCID);
		  if iNew > -1 then iLen := Length(FileNames[iNew]);
		end;
		if iLen > FMaxSingleEntrySize then FMaxSingleEntrySize := iLen;
	      end;
	    end;
	  end;
	end;
	if cidl <> nil then begin
	  if not ReturnRtf then begin
	    cidl.FFileSections := Self;
	    FModifiedMainSection := Utf8OrAnsi(cidl.GetHtmlVersion);
	  end
	  else begin
	    FModifiedMainSection := CreateEmbeddedRtfVersion(cidl.FSource, sDefaultBmp, Self);
	    for iIndex := 0 to pred(Self.Count) do Self.Used[iIndex] := true;
	  end;
	end;
      finally
	cidl.Free;
      end;
    end;
  end;
  if (Count > 0) and (FModifiedMainSection = E) then FModifiedMainSection := FileData[0];
  for iIndex := pred(Count) downto 1 do if not Used[iIndex] then Delete(iIndex);
end;

procedure TFileSections.Clear;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do if Objects[iIndex] <> nil then begin
    Dispose(PFSItem(Objects[iIndex]));
    Objects[iIndex] := nil;
  end;
  inherited;
end;

function TFileSections.AddItem(const ItemName, ItemData, ItemCID: String): Integer;
var
  pfsi: PFSItem;
begin
  New(pfsi);
  Result := inherited AddObject(ItemCID, TObject(pfsi));
  if pfsi <> nil then with pfsi^ do begin
    Name := ItemName;
    Data := ItemData;
    Used := false;
  end;
end;

procedure TFileSections.Delete(Index: Integer);
begin
  if (Index > -1) and (Index < Count) and (Objects[Index] <> nil) then begin
    Dispose(PFSItem(Objects[Index]));
    Objects[Index] := nil;
  end;
  inherited;
end;

function TFileSections.WriteToFiles(const TextFileName: String): Longbool;
var
  iIndex: Integer;
begin
  Result := false;
  if FModifiedMainSection <> E then begin
    if WriteTextToFile(TextFileName, FModifiedMainSection) then begin
      for iIndex := 1 to pred(Count) do WriteTextToFile(FileNames[iIndex], FileData[iIndex]);
      Result := true;
    end;
  end;
end;

destructor TFileSections.Destroy;
begin
  Clear;
  inherited;
end;

function TFileSections.GetFileName(Index: Integer): String;
begin
  Result := PFSItem(Objects[Index])^.Name;
end;

function TFileSections.GetFileData(Index: Integer): String;
begin
  Result := PFSItem(Objects[Index])^.Data;
end;

function TFileSections.GetCID(Index: Integer): String;
begin
  Result := Strings[Index];
end;

function TFileSections.GetUsed(Index: Integer): Longbool;
begin
  Result := PFSItem(Objects[Index])^.Used;
end;

procedure TFileSections.SetUsed(Index: Integer; Value: Longbool);
begin
  PFSItem(Objects[Index])^.Used := Value;
end;


// TCIDList ----------------------------------------------------------------------------------------

constructor TCIDList.Create(HtmlNode: TSectionNode; ReturnRtf: Longbool);
var
  mdo: TMimeDecodeOptions;
begin
  inherited Create;
  if HtmlNode <> nil then with HtmlNode do begin
    FReturnRtf := ReturnRtf;
    if FReturnRtf then begin
      mdo := [mdoReturnRtf];
      FSource := GetValidHtml(SectionData, MimeMsgTree.FDefaultFontName, false, true);
      if FSource <> E then FSource := ConvertFormat(FSource, MimeMsgTree.FDefaultFontName, mdo, MimeMsgTree.MaxLineLength);
      if FSource <> E then CreateRtfVersion;
    end
    else begin
      FSource := SectionData;
      CreateHtmlVersion;
    end;
  end;
end;

destructor TCIDList.Destroy;
begin
  Clear;
  inherited;
end;

function TCIDList.AddItem(const ItemCID: String; SecBeg, SecEnd: Integer): Integer;
var
  pcidi: PCIDItem;
begin
  New(pcidi);
  if pcidi <> nil then with pcidi^ do begin
    CIDStr := ItemCID;
    SectionBeg := SecBeg;
    SectionEnd := SecEnd;
  end;
  Result := inherited Add(pcidi);
end;

procedure TCIDList.Clear;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do if Items[iIndex] <> nil then begin
    Dispose(PCIDItem(Items[iIndex]));
    Items[iIndex] := nil;
  end;
  inherited;
end;

procedure TCIDList.Delete(Index: Integer);
begin
  if (Index > -1) and (Index < Count) and (Items[Index] <> nil) then begin
    Dispose(PCIDItem(Items[Index]));
    Items[Index] := nil;
  end;
  inherited;
end;

function TCIDList.GetHtmlVersion: String;
var
  iLen, iPos, iIndex, iFound: Integer;
  sFileName: String;
  pResult: PChar;
begin
  if Count > 0 then begin
    iLen := Length(FSource) + (Count * FFileSections.MaxSingleEntrySize);
    SetString(Result, nil, iLen);
    pResult := Pointer(Result);
    iPos := 1;
    for iIndex := 0 to pred(Count) do begin
      iFound := FFileSections.IndexOf(CIDString[iIndex]);
      if iFound > -1 then begin
	FFileSections.Used[iFound] := true;
	iLen := SectionBeg[iIndex] - Length(CID_MARK) - iPos;
	System.Move(FSource[iPos], pResult^, iLen);
	iPos := succ(SectionEnd[iIndex]);
	inc(pResult, iLen);
	sFileName := ExtractFileName(FFileSections.FileNames[iFound]);
	iLen := Length(sFileName);
	System.Move(Pointer(sFileName)^, pResult^, iLen);
	inc(pResult, iLen);
      end;
    end;
    iLen := succ(Length(FSource) - iPos);
    if iLen > 0 then begin
      System.Move(FSource[iPos], pResult^, iLen);
      inc(pResult, iLen);
    end;
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := FSource;
end;

procedure TCIDList.CreateHtmlVersion;

  function ExtractCIDName(const Text: String; var iBeg, iEnd: Integer): String;
  begin
    Result := E;
    if (Text[iBeg] = QU) and LoStrAtPos(CID_MARK, Text, succ(iBeg)) then begin
      inc(iBeg, succ(Length(CID_MARK)));
      iEnd := NextCharPos(QU, Text, iBeg);
    end
    else if LoStrAtPos(CID_MARK, Text, iBeg) then begin
      inc(iBeg, Length(CID_MARK));
      iEnd := succ(iBeg);
      while not (Text[iEnd] in [NULL, SP, GT]) do inc(iEnd);
      if Text[iEnd] = GT then begin
	repeat
	  dec(iEnd);
	until (iEnd = iBeg) or (Text[pred(iEnd)] <> SP);
      end;
    end
    else Exit;
    if iEnd > iBeg then Result := Copy(Text, iBeg, iEnd - iBeg);
  end;

const
  IMG_TAG = '<img ';
  SRC_ATTR = 'src=';
var
  iBeg, iEnd: Integer;
  sCIDName: String;
begin
  iBeg := 1;
  repeat
    iBeg := NextLoStrPos(IMG_TAG, FSource, iBeg);
    if iBeg > 0 then begin
      inc(iBeg, Length(IMG_TAG));
      repeat
	while not (AsciiLoCase(FSource[iBeg]) in [NULL, GT, 's']) do inc(iBeg);
	case FSource[iBeg] of
	  NULL: Exit;
	  GT: Break;
	  's': if LoStrAtPos(SRC_ATTR, FSource, iBeg) then begin
	    inc(iBeg, Length(SRC_ATTR));
	    sCIDName := ExtractCIDName(FSource, iBeg, iEnd);
	    if sCIDName <> E then begin
	      AddItem(Unescape(sCIDName, false, PC), iBeg, pred(iBeg + Length(sCIDName)));
	      iBeg := succ(iEnd);
	    end
	    else inc(iBeg, Length(CID_MARK));
	    Break;
	  end;
	end;
	inc(iBeg);
      until FSource[iBeg] in [NULL, GT];
    end;
  until iBeg = 0;
end;

procedure TCIDList.CreateRtfVersion;
const
  RTF_LB = '{';
  RTF_RB = '}';
  RTF_FIELD = '\field';
  RTF_FLDINST = '\fldinst';
  RTF_FLDRSLT = '\fldrslt';
  RTF_IMGLINK = 'INCLUDEPICTURE';
var
  iLen, iBeg, iEnd, iIndex, iNameBeg, iNameEnd: Integer;
  sFileName: String;
begin
  iLen := Length(FSource);
  iBeg := 1;
  repeat
    iBeg := NextStrPos(RTF_FLDINST, FSource, iBeg);
    if iBeg = 0 then Break;
    iEnd := NextStrPos(RTF_FLDRSLT, FSource, iBeg);
    if iEnd = 0 then Break;
    iIndex := iBeg + Length(RTF_FLDINST);
    while (iIndex < iEnd) and (FSource[iIndex] <= SP) do inc(iIndex);
    if not StrAtPos(RTF_IMGLINK, FSource, iIndex) then begin
      iBeg := iEnd + Length(RTF_FLDRSLT);
      Continue;
    end;
    inc(iIndex, Length(RTF_IMGLINK));
    while (iIndex < iEnd) and (FSource[iIndex] <= SP) do inc(iIndex);
    if FSource[iIndex] = QU then begin
      iNameBeg := succ(iIndex);
      iNameEnd := iNameBeg;
      while (iNameEnd < iEnd) and (FSource[iNameEnd] <> QU) do inc(iNameEnd);
      if iNameEnd = iEnd then begin
	iBeg := iEnd + Length(RTF_FLDRSLT);
	Continue;
      end;
    end
    else begin
      iNameBeg := iIndex;
      iNameEnd := iNameBeg;
      while (iNameEnd < iEnd) and not (FSource[iNameEnd] in [SP, RTF_RB]) do inc(iNameEnd);
      if iNameEnd = iEnd then begin
	iBeg := iEnd + Length(RTF_FLDRSLT);
	Continue;
      end;
    end;
    if LoStrAtPos(CID_MARK, FSource, iNameBeg) then inc(iNameBeg, Length(CID_MARK));
    while (FSource[pred(iNameEnd)] <= SP ) do dec(iNameEnd);
    iBeg := PrevStrPos(RTF_FIELD, FSource, iBeg);
    while (iBeg > 0) and (FSource[iBeg] <> RTF_LB) do dec(iBeg);
    inc(iBeg);
    inc(iEnd, Length(RTF_FLDRSLT));
    iIndex := 0;
    repeat
      while (iEnd < iLen) and (FSource[iEnd] <> RTF_RB) do inc(iEnd);
      if FSource[iEnd] = RTF_RB then begin
	inc(iIndex);
	inc(iEnd);
      end
      else Break;
    until (iIndex = 2);
    dec(iEnd);
    sFileName := RtfUnescape(Unwrap(Copy(FSource, iNameBeg, iNameEnd - iNameBeg)));
    if sFileName <> E then AddItem(sFileName, iBeg, iEnd);
    iBeg := succ(iEnd);
  until iBeg >= iLen;
end;

function TCIDList.GetCIDString(Index: Integer): String;
begin
  Result := PCIDItem(Items[Index])^.CIDStr;
end;

function TCIDList.GetSectionBeg(Index: Integer): Integer;
begin
  Result := PCIDItem(Items[Index])^.SectionBeg;
end;

function TCIDList.GetSectionEnd(Index: Integer): Integer;
begin
  Result := PCIDItem(Items[Index])^.SectionEnd;
end;


// TMimeMsgTree ----------------------------------------------------------------------------------------

constructor TMimeMsgTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Images := T32BitImages.Create(nil);
  FMessageText := TTextList.Create;
  FMessageText.FMimeMsgTree := Self;
  if AOwner = nil then CreateTreeView;
  FillByte(FMsgTreeData, SizeOf(TMsgTreeData), 0);
end;

destructor TMimeMsgTree.Destroy;
begin
  if FMessageText <> nil then FMessageText.Free;
  if Images <> nil then Images.Free;
  inherited Destroy;
end;

procedure TMimeMsgTree.BeginUpdate;
begin
  Items.BeginUpdate;
end;

procedure TMimeMsgTree.EndUpdate;
begin
  Items.EndUpdate;
end;

procedure TMimeMsgTree.AddIcons;
var
  iIndex: Integer;
  hIcon: THandle;
begin
  Images.BkColor := Color;
  Images.BlendColor := clHighlight;
  Images.CreateHandle(GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), Count * 2, Brush, Themed);
  for iIndex := 0 to pred(Count) do begin
    with SectionNodes[iIndex] do begin
      hIcon := 0;
      case SectionType of
	mdtMultipartUnknown..mdtMultipartDigest: hIcon := FileTypeIcon(FOLDER, E, true, false);
	mdtMainHeaders, mdtPlainText, mdtAltText: hIcon := FileTypeIcon(TXT, E, false, false);
	mdtHtmlText: hIcon := FileTypeIcon(HTM, E, false, false);
	mdtRtfText: hIcon := FileTypeIcon(RTF, E, false, false);
	mdtSignature: hIcon := ExtractIcon(GetSysDir + 'CryptUI.dll', 3421, 32, true);
	mdtFile: begin
	  if Encoding in [ePlainKeys..eCerts] then
	    hIcon := GetDefaultIcon(E, E, 'PGP Armored File')
	  else hIcon := FileTypeIcon(FileName, SectionData, false, false);
	end;
      end;
      ImageIndex := Images.AddIconByHandle(hIcon, true, true);
      if ImageIndex = -1 then
	SelectedIndex := ImageIndex
      else SelectedIndex := succ(ImageIndex);
    end;
  end;
end;

procedure TMimeMsgTree.Clear;
begin
  FBodyEncoding := eNone;
  FFileAttachments := 0;
  FFirstTextSection := nil;
  FGetRawData := false;
  FHasAttachments := false;
  FIsMessageDigest := false;
  FUnencodedMimeData := false;
  FFindHeaderOptions := [];
  FMimeDecodeOptions := [];
  with FPSMimeBounds do begin
    BegPos := 0;
    EndPos := 0;
  end;
  FMessageText.Clear;
  Images.Clear;
  Items.Clear;
end;

function TMimeMsgTree.AddSection(Parent: TSectionNode; const SecName, SecData: String; MDT: TMsgDataType): TSectionNode;
begin
  Result := TSectionNode(Items.AddChild(Parent, SecName));
  if Result <> nil then with Result do begin
    FSectionType := MDT;
    if not (mdoAnalyseEncoding in FMimeDecodeOptions) then FSectionData := SecData;
  end;
end;

function TMimeMsgTree.InsertSectionAfter(Section: TSectionNode; const Name, Data: String; MDT: TMsgDataType): TSectionNode;
var
  snNext: TSectionNode;
begin
  if Section <> nil then begin
    snNext := TSectionNode(Section.GetNext);
    if snNext <> nil then
      Result := TSectionNode(Items.Insert(snNext, Name))
    else Result := TSectionNode(Items.Add(Section, Name));
    if Result <> nil then begin
      Result.FSectionType := MDT;
      Result.FSectionData := Data;
    end;
  end
  else Result := nil;
end;

function TMimeMsgTree.FindPSEncodedSection(Encoding: TEncoding): TSectionNode;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do begin
    if SectionNodes[iIndex].Encoding = Encoding then begin
      Result := SectionNodes[iIndex];
      Exit;
    end;
  end;
  Result := nil;
end;

function TMimeMsgTree.GetAttachmentList: String;
const
  ARROW_W = EQ + NULL + GT + NULL + SP + NULL + wNUL;
  ARROW_A = EQ + GT + SP;
var
  iLen, iIndex: Integer;
  pResult: PChar;
  sSectionName: String;
begin
  if FFileAttachments > 0 then begin
    iLen := Count * (Length(ARROW_A) + MAX_PATH + Length(PARA));
    if mdoReturnUnicode in FMimeDecodeOptions then iLen := iLen shl 1 + Length(wNUL);
    SetString(Result, nil, iLen);
    pResult := Pointer(Result);
    for iIndex := 1 to pred(Count) do begin
      if mdoReturnUnicode in FMimeDecodeOptions then begin
	case SectionTypes[iIndex] of
	  mdtUnknownData: sSectionName := Utf8ToWide(SectionNames[iIndex]);
	  mdtMainHeaders: sSectionName := Utf8ToWide(FMailMessageName + SP + SectionNodes[iIndex].Utf8Subject);
	  mdtFile: sSectionName := Utf8ToWide(SectionNames[iIndex]);
	else
	  sSectionName := E;
	end;
	if sSectionName <> E then begin
	  sSectionName := Concat([ARROW_W, Abbreviate(sSectionName, MAX_PATH)]);
	  iLen := Length(sSectionName) - Length(wNUL);
	  Move(Pointer(sSectionName)^, pResult^, iLen);
	  inc(pResult, iLen);
	  PQChar(pResult)^ := wCRLF;
	  inc(pResult, Length(wCRLF));
	end;
      end
      else begin
	case SectionTypes[iIndex] of
	  mdtUnknownData: sSectionName := Utf8ToAnsi(SectionNames[iIndex]);
	  mdtMainHeaders: sSectionName := Utf8ToAnsi(FMailMessageName + SP + SectionNodes[iIndex].Utf8Subject);
	  mdtFile: sSectionName := Utf8ToAnsi(SectionNames[iIndex]);
	else
	  sSectionName := E;
	end;
	if sSectionName <> E then begin
	  sSectionName := Concat([ARROW_A, Abbreviate(sSectionName, MAX_PATH)]);
	  iLen := Length(sSectionName);
	  Move(Pointer(sSectionName)^, pResult^, iLen);
	  inc(pResult, iLen);
	  PDChar(pResult)^ := aCRLF;
	  inc(pResult, Length(aCRLF));
	end;
      end;
    end;
    if mdoReturnUnicode in FMimeDecodeOptions then begin
      PDChar(pResult)^ := wNUL;
      inc(pResult, Length(wNUL));
    end;
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function TMimeMsgTree.GetMessageText(IncludeImages: Longbool): TTextList;
var
  snParent: TSectionNode;
  iIndex: Integer;
begin
  FMessageText.Clear;
  if not FIsMessageDigest and (FFirstTextSection <> nil) then begin
    if IncludeImages then FFirstTextSection.IncludeImages := true;
    FMessageText.AddObject(FFirstTextSection.SectionText, FFirstTextSection);
    snParent := FFirstTextSection.Parent;
    if snParent <> nil then begin
      iIndex := succ(snParent.IndexOf(FFirstTextSection));
      repeat
	for iIndex := iIndex to pred(snParent.Count) do with snParent.Child[iIndex] do begin
	  if SectionType in [mdtPlainText..mdtAltText] then FMessageText.AddObject(SectionText, snParent.Child[iIndex]);
	end;
	snParent := snParent.Parent;
	iIndex := 0;
      until snParent = nil;
    end;
  end;
  Result := FMessageText;
end;

function TMimeMsgTree.GetPSMimeSignatureResult: TPSMimeSignatureResult;
begin
  if Count > 0 then with SectionNodes[0] do begin
    Result.Flags := SectionFlags;
    Result.Text := SectionSig;
    SpaceExtract(Result.Text);
  end;
end;

function TMimeMsgTree.IsPSMimeSignedSection(Section: TSectionNode): Longbool;
var
  snParent: TSectionNode;
begin
  if Section <> nil then begin
    snParent := Section.getParent;
    while snParent <> nil do begin
      Result := (snParent.HeaderDataEx.MultipartType in [mdtMultipartEncrypted..mdtMultipartSigned]);
      if not Result then
	snParent := snParent.getParent
      else Exit;
    end;
  end;
  Result := false;
end;

function TMimeMsgTree.IsUnknownSectionText(Index: Integer): Longbool;
begin
  with SectionNodes[Index] do begin
    Result := (SectionType = mdtUnknownData) and
	       LoStrAtPos(CT_TEXT, SectionName, Length(FUnknownSectionName) + Length(CL) + Length(SP) + 1);
  end;
end;

function TMimeMsgTree.RecreateMimeHeaders(const Headers: String; DateTime: TDateTime): String;
var
  hdNonMime: THeaderData;
  sDate, sFrom, sTo, sSubject, sReplyTo: String;
  iPos: Integer;
begin
  FillChar(hdNonMime, SizeOf(THeaderData), 0);
  hdNonMime.Text := Headers;
  sDate := DateTimeToMimeUTC(DateTime);
  sFrom := FindHeaderLine(FSenderHeaderName, hdNonMime, false).Text;
  sTo := FindHeaderLine(FRecipientHeaderName, hdNonMime, false).Text;
  sSubject := FindHeaderLine(FSubjectHeaderName, hdNonMime, false).Text;
  sReplyTo := FindHeaderLine(FReplyToHeaderName, hdNonMime, false).Text;
  iPos := succ(Length(sTo) - Length(ET_AL));
  if StrAtPos(ET_AL, sTo, iPos) then System.Delete(sTo, iPos, MAXINT);
  if IsExtended(sFrom) then sFrom := QPEncodeAddressHeader(sFrom, UTF_8);
  if IsExtended(sTo) then sTo := QPEncodeAddressHeader(sTo, UTF_8);
  if IsExtended(sSubject) then sSubject := QPEncode(sSubject, UTF_8);
  if IsExtended(sReplyTo) then sReplyTo := QPEncode(sReplyTo, UTF_8);
  if sReplyTo <> E then sReplyTo := CRLF + H_REPLY_TO + SP + sReplyTo;
  Result := H_DATE + SP + sDate + CRLF +
	    H_FROM + SP + sFrom + CRLF +
	    H_TO + SP + sTo + CRLF +
	    H_SUBJECT + SP + sSubject + sReplyTo;
end;

procedure TMimeMsgTree.MimeParse(const MailMessage: String; BegPos, EndPos: Integer; MDO: TMimeDecodeOptions);

  function FullParseOrDecode(FMDO: TMimeDecodeOptions): Longbool;
  begin
    Result := (([mdoFullParse..mdoDecode] * FMDO) <> []);
  end;

  function DecodePSMimeMessage(const PSMimeHD: THeaderData; var PSMimeBounds: TBounds;
			       var SignatureText: String; var SignatureFlags: Integer;
			       var UnencodedMimeData: Longbool): THeaderData;
  var
    hd: THeaderData;
    iSuffixPos: Integer;
    sDecoded, sDummy, sPrefix, sSuffix: String;
  begin
    Result := PSMimeHD;
    if not FGetRawData then with PSMimeBounds do begin
      BegPos := 0;
      EndPos := 0;
    end;
    UnencodedMimeData := false;
    if ContainsPSMimeMsg(PSMimeHD, hd) <= eCerts then Exit;
    if mdoAnalyseEncoding in FMimeDecodeOptions then with hd do begin
      Result.Encoding := Encoding;
      with PSMimeBounds do begin
	BegPos := Pos;
	EndPos := pred(Pos + Len);
      end;
      Exit;
    end;
    with hd do begin
      iSuffixPos := Pos + Len;
      sDecoded := FOnEncodedData(Copy(Text, Pos, Len), sDummy, SignatureText, SignatureFlags);
      if sDecoded <> E then begin
	sPrefix := Copy(PSMimeHD.Text, PSMimeHD.Pos, Pos - PSMimeHD.Pos);
	sSuffix := ClearBlankLineStr(Copy(Text, iSuffixPos, succ(MaxPos - iSuffixPos)));
      end
      else Exit;
    end;
    with Result do begin
      Text := sPrefix + sDecoded + sSuffix;
      Pos := 1;
      Len := 0;
      MaxPos := Length(Text);
      Encoding := hd.Encoding;
      if Encoding in [ePMSigned, eSMSigned] then
	Result := PSMimeHD
      else GetHeaderInfo(Result, FFindHeaderOptions);
    end;
    with PSMimeBounds do begin
      BegPos := succ(Length(sPrefix));
      EndPos := pred(BegPos) + Length(sDecoded);
    end;
    UnencodedMimeData := (sSuffix <> E);
  end;

  function PGPDecodeInlineData(Section: TSectionNode; const InlineData: String; FileData: Longbool;
			       var Charset, SignatureText: String; var SignatureFlags: Integer): String;
  var
    hd: THeaderData;
    iSuffixPos: Integer;
    sDecoded, sPrefix, sSuffix: String;
  begin
    if not FGetRawData then with hd do begin
      FillChar(hd, SizeOf(THeaderData), 0);
      Text := InlineData;
      Pos := 1;
      MaxPos := Length(InlineData);
      if not IsPSMimeSignedSection(Section)
      and (GetInlineEncoding(hd, hd) in [ePlainEncrypted..ePlainSigned]) then begin
	iSuffixPos := Pos + Len;
	sDecoded := FOnEncodedData(Copy(Text, Pos, Len), Charset, SignatureText, SignatureFlags);
	if sDecoded <> E then begin
	  if not FileData then begin
	    sPrefix := Copy(Text, 1, pred(Pos));
	    sSuffix := ClearBlankLineStr(Copy(Text, iSuffixPos, succ(MaxPos - iSuffixPos)));
	    Charset := CheckWinCharset(Charset);
	    if Charset <> E then
	      Result := Utf8ToString(sDecoded, Charset)
	    else Result := Utf8OrAnsi(sDecoded);
	    if Result = E then
	      Result := sPrefix + sDecoded + sSuffix
	    else Result := sPrefix + Result + sSuffix;
	  end
	  else Result := sDecoded;
	end
	else Result := InlineData;
      end
      else Result := InlineData;
    end
    else Result := InlineData;
  end;

  function GetBodyText(var HDE: THeaderDataEx; var MDT: TMsgDataType): String;

    function GetMsgTextType(const CTHeader: String; var Charset: String): TMsgDataType;
    var
      iBeg, iLen, iEnd: Integer;
    begin
      Charset := E;
      Result := mdtUnknownData;
      try
	if LoStrAtPos(CT_TEXT, CTHeader, 1) then begin
	  iBeg := NextNonSpaceChar(CTHeader, succ(Length(CT_TEXT)));
	  if LoStrAtPos(_PLAIN, CTHeader, iBeg) then
	    Result := mdtPlainText
	  else if LoStrAtPos(_HTML, CTHeader, iBeg) then
	    Result := mdtHtmlText
	  else if LoStrAtPos(_RTF, CTHeader, iBeg) then
	    Result := mdtRtfText
	  else Result := mdtAltText;
	  iBeg := NextNonSpaceChar(CTHeader, NextLoStrPos(CT_CHARSET, CTHeader, succ(iBeg)) + Length(CT_CHARSET));
	  if iBeg = 0 then Exit;
	  iLen := Length(CTHeader);
	  while (iBeg < iLen) and (CTHeader[iBeg] in WSPEQ) do inc(iBeg);
	  if CTHeader[iBeg] = QU then begin
	    inc(iBeg);
	    iEnd := NextCharPos(QU, CTHeader, iBeg);
	  end
	  else begin
	    iEnd := NextCharPos(SC, CTHeader, iBeg);
	    if iEnd = 0 then iEnd := succ(Length(CTHeader));
	  end;
	  Charset := UpperCase(Trim(Copy(CTHeader, iBeg, iEnd - iBeg)));
	  System.Delete(Charset, FirstCharPos(AST, Charset), MAXINT);
	end
	else if CTHeader = E then Result := mdtPlainText;
      finally
	if not FGetRawData then Charset := CheckWinCharset(Charset);
      end;
    end;

  var
    iSignatureFlags, iBeg, iEnd: Integer;
    sPGPCharset, sHeaderCharset, sSignatureText: String;
  begin
    Result := E;
    MDT := mdtPlainText;
    iSignatureFlags := 0;
    with HDE.HeaderData do begin
      if (CTHeader = E) or LoStrAtPos(CT_TEXT, CTHeader, 1) then begin
	iBeg := Pos + Len;
	if iBeg < MaxPos then begin
	  case Text[iBeg] of
	    LF: inc(iBeg, Length(LF) shl 1);
	    CR: inc(iBeg, Length(CRLF) shl 1);
	  end;
	end;
	if HDE.Boundary <> E then begin
	  iEnd := FirstRangeStrPos(DD + HDE.Boundary, Text, iBeg, MaxPos);
	  Result := Copy(Text, iBeg, iEnd - iBeg);
	end
	else Result := Copy(Text, iBeg, succ(MaxPos) - iBeg);
      end
      else Exit;
      if CTEHeader <> E then begin
	if LoStrAtPos(CTE_B64, CTEHeader, 1) then
	  Result := B64Convert(Result)
	else if LoStrAtPos(CTE_QP, CTEHeader, 1) then Result := QPConvert(Result, false);
      end;
      if mdoDecode in FMimeDecodeOptions then begin
	Result := PGPDecodeInlineData(Selected, Result, false, sPGPCharset,
				      sSignatureText, iSignatureFlags);
      end;
      MDT := GetMsgTextType(CTHeader, sHeaderCharset);
    end;
    with HDE do begin
      if sPGPCharset <> E then
	Charset := sPGPCharset
      else Charset := sHeaderCharset;
      if MDT = mdtRtfText then
	Result := StringToAnsi(Result, Charset)
      else Result := StringToUtf8(Result, Charset);
      Charset := sHeaderCharset;
      if MDT = mdtPlainText then Result := UnwrapArmoredKeys(Result);
    end;
  end;

  function GetTextSectionName(const SectionName: String; mdtSection: TMsgDataType): String;
  begin
    if (FTextSectionName <> E) and (SectionName <> E) then
      Result := FTextSectionName + CL + SP + SectionName
    else if FTextSectionName <> E then
      Result := FTextSectionName
    else Result := SectionName;
    case mdtSection of
      mdtHtmlText: Result := Result + SP + LB + _HTML + RB;
      mdtRtfText: Result := Result + SP + LB + _RTF + RB;
    end;
  end;

  function GetMultipartName(const CTHeader: String; var MultipartType: TMsgDataType): String;
  var
    iBeg, iEnd: Integer;
  begin
    MultipartType := mdtMultipartUnknown;
    iBeg := NextNonSpaceChar(CTHeader, succ(Length(CT_MULTIPART)));
    iEnd := NextCharPos(SC, CTHeader, iBeg);
    Result := LowerCase(Copy(CTHeader, iBeg, iEnd - iBeg));
    if Result = CT_MIXED then
      MultipartType := mdtMultipartMixed
    else if Result = CT_RELATED then
      MultipartType := mdtMultipartRelated
    else if Result = CT_ALTERNATIVE then
      MultipartType := mdtMultipartAlternative
    else if Result = CT_ENCRYPTED then
      MultipartType := mdtMultipartEncrypted
    else if Result = CT_Signed then
      MultipartType := mdtMultipartSigned
    else if Result = CT_DIGEST then
      MultipartType := mdtMultipartDigest;
  end;

  function IsPMimeSignature(const CTHeader: String): Longbool;
  begin
    Result := (FirstLoStrPos(IS_PGP + DA + PM_SIG, CTHeader) > 0);
  end;

  function IsSMimeSignature(const CTHeader: String): Longbool;
  begin
    Result := (FirstLoStrPos(PKCS7 + DA + PM_SIG, CTHeader) > 0);
  end;

  function IsPSMimeSignature(const CTHeader: String): Longbool;
  begin
    Result := IsPMimeSignature(CTHeader) or IsSMimeSignature(CTHeader);
  end;

  function PosInBounds(Pos: Integer; const SectionBounds: TBounds): Longbool;
  begin
    with SectionBounds do Result := (Pos >= BegPos) and (Pos <= EndPos);
  end;

  procedure ParseSubSections(const HDE: THeaderDataEx);

    function GetSectionBounds(const HD: THeaderData; const Boundary: String; MainSection, Digest: Longbool): TBounds;
    var
      iLen: Integer;
    begin
      with HD, Result do begin
	iLen := Length(DD) + Length(Boundary);
	BegPos := Pos + iLen;
	if MainSection then BegPos := NextStrPos(PARA, Text, BegPos) + Length(PARA);
	while (BegPos < MaxPos) and (Text[BegPos] in LFCR) do inc(BegPos);
	if Digest then
	  EndPos := pred(NextLoStrPos(DD + Boundary, Text, BegPos))
	else EndPos := pred(NextLoStrPos(DD + Boundary + DD, Text, BegPos));
      end;
    end;

    function GetSectionName(const Kind, Name: String): String;
    begin
      if not FGetRawData then begin
	if (Kind <> E) and (Name <> E) then
	  Result := Kind + CL + SP + Name
	else if Kind <> E then
	  Result := Kind
	else Result := Name;
      end
      else Result := Name;
    end;

    function ExtractSection(const SectionHDE: THeaderDataEx): String;
    var
      iBeg, iEnd: Integer;
    begin
      with SectionHDE.HeaderData do begin
	iBeg := Pos + Len;
	while (iBeg <= MaxPos) and (Text[iBeg] < SP) do inc(iBeg);
	iEnd := FirstRangeLoStrPos(DD + SectionHDE.Boundary, Text, succ(iBeg), MaxPos);
	if iEnd > iBeg then Result := Copy(Text, iBeg, iEnd - iBeg);
      end;
    end;

  var
    hdeSection: THeaderDataEx;
    mdtSection: TMsgDataType;
    hdSection: THeaderData;
    hdInline: THeaderData;
    snAdded: TSectionNode;
    iSignatureFlags, iDataStrLen: Integer;
    sSignatureText, sSectionData, sSectionName, sSectionType, sFileName, sFileType: String;
  begin
    hdeSection := HDE;
    with hdeSection do begin
      while HeaderData.Len > 0 do begin
	Charset := E;
	mdtSection := mdtUnknownData;
	if MultipartType <> mdtMainHeaders then begin
	  HeaderData := GetNextHeaderInfo(HeaderData, Boundary, FFindHeaderOptions);
	  if FullParseOrDecode(FMimeDecodeOptions) then begin
	    if not FGetRawData then begin
	      if (mdoDecode in FMimeDecodeOptions) and (ContainsPSMimeMsg(HeaderData, hdSection) > eCerts) then begin
		with GetSectionBounds(hdSection, Boundary, false, false) do begin
		  FSubMimeDecoding := true;
		  MimeParse(hdSection.Text, BegPos, EndPos, FMimeDecodeOptions);
		  Continue;
		end;
	      end;
	    end
	    else with HeaderData do if Pos > 0 then AddSection(Selected, E, Copy(Text, Pos, Len), mdtSubHeaders);
	  end;
	end;
	snAdded := nil;
	sSignatureText := E;
	iSignatureFlags := 0;
	with HeaderData do if Pos > 0 then begin
	  sSectionData := E;
	  // digest or message
	  if (MultipartType = mdtMultipartDigest) or LoStrAtPos(CT_MESSAGE, CTHeader, 1) then begin
	    if (Selected <> nil) and (Selected.Level = 0) and (MultipartType = mdtMultipartDigest) then
	      FIsMessageDigest := true
	    else inc(FFileAttachments);
	    with GetSectionBounds(HeaderData, Boundary, true, MultipartType = mdtMultipartDigest) do begin
	      MimeParse(Text, BegPos, EndPos, FMimeDecodeOptions);
	    end;
	  end
	  // other multipart sections
	  else if LoStrAtPos(CT_MULTIPART, CTHeader, 1) then begin
	    GetMultipartName(CTHeader, mdtSection);
	    with GetSectionBounds(HeaderData, Boundary, false, mdtSection = mdtMultipartDigest) do begin
	      MimeParse(Text, BegPos, EndPos, FMimeDecodeOptions);
	    end;
	  end
	  // subsections
	  else begin
	    if FGetRawData then
	      sSectionName := FindHeaderLine(CTD_HEADER, HeaderData, true).Text
	    else sSectionName := MimeDecodeHeaders(FindHeaderLine(CTD_HEADER, HeaderData, true).Text);
	    // alternative text
	    if MultipartType = mdtMultipartAlternative then begin
	      if LoStrAtPos(_PLAIN, CTHeader, NextNonSpaceChar(CTHeader, succ(Length(CT_TEXT))))
	      or (FFirstTextSection = nil) or (FMimeDecodeOptions * [mdoFullParse, mdoFancyText] <> []) then begin
		sSectionData := GetBodyText(hdeSection, mdtSection);
		if sSectionData <> E then begin
		  snAdded := AddSection(Selected, GetTextSectionName(sSectionName, mdtSection), sSectionData, mdtSection);
		  if Recursion = 0 then begin
		    if ((FFirstTextSection = nil) or (FFirstTextSection.SectionText = E))
		    or ((mdoFancyText in FMimeDecodeOptions) and (mdtSection in [mdtHtmlText..mdtAltText])) then begin
		      FFirstTextSection := snAdded;
		    end;
		  end;
		end;
	      end
	      else Continue;
	    end
	    // other text
	    else if ((CTHeader = E) or LoStrAtPos(CT_TEXT, CTHeader, 1))
	    and not LoStrAtPos(CDP_ATTACHMENT, CDPHeader, 1) then begin
	      sSectionData := GetBodyText(hdeSection, mdtSection);
	      if sSectionData <> E then begin
		snAdded := AddSection(Selected, GetTextSectionName(sSectionName, mdtSection), sSectionData, mdtSection);
		if (FFirstTextSection = nil) or (FFirstTextSection.SectionText = E) then FFirstTextSection := snAdded;
	      end;
	    end
	    // attachment
	    else begin
	      sSectionType := CharCopy(CTHeader, SC);
	      if not ((mdoFancyText in FMimeDecodeOptions) and (MultipartType in [mdtMultipartRelated]))
	      and not (MultipartType in [mdtMultipartEncrypted..mdtMultipartSigned]) then begin
		inc(FFileAttachments);
	      end;
	      GetContentHeaders(HeaderData, [fhoContentLocation, fhoContentID], true);
	      if CTHeader <> E then sSectionName := MimeGetParameterValue(CTHeader, CT_NAME, nil, FGetRawData);
	      if sSectionName = E then sSectionName := MimeGetParameterValue(CDPHeader, CDP_FILENAME, nil, FGetRawData);
	      if (sSectionName = E) and (CIDHeader <> E) then sSectionName := DelAngles(Unescape(CIDHeader, false, PC));
	      if (sSectionName = E) and (CLCHeader <> E) then sSectionName := GetUrlFileName(Unescape(CLCHeader, false, PC));
	      if sSectionName = E then sSectionName := sSectionType;
	      sFileName := sSectionName;
	      sFileType := FileTypeInfo(sSectionName, sSectionType);
	      if sFileType = E then sFileType := FUnknownSectionName;
	      sSectionName := GetSectionName(sFileType, sSectionName);
	      if (MultipartType = mdtMultipartSigned) and IsPSMimeSignature(CTHeader) then
		mdtSection := mdtSignature
	      else if sSectionName <> E then mdtSection := mdtFile;
	      if FullParseOrDecode(FMimeDecodeOptions) then begin
		FillChar(hdInline, SizeOf(THeaderData), 0);
		if mdtSection <> mdtSignature then begin
		  if LoStrAtPos(CTE_B64, CTEHeader, 1) then
		    sSectionData := B64Convert(ExtractSection(hdeSection))
		  else if LoStrAtPos(CTE_QP, CTEHeader, 1) then
		    sSectionData := QPConvert(ExtractSection(hdeSection), false)
		  else sSectionData := ExtractSection(hdeSection);
		  if mdoDecode in FMimeDecodeOptions then begin
		    sSectionData := PGPDecodeInlineData(Selected, sSectionData, true, Charset,
							sSignatureText, iSignatureFlags);
		    if (sSectionData <> E) and (FirstLoStrPos(ExtractFileExt(sFileName), PGP + ASC) <> 0)
		    and (GetInlineEncoding(HeaderData, hdInline) in [ePlainEncrypted..eCerts]) then begin
		      HeaderData.Encoding := hdInline.Encoding;
		      sFileName := ExtractBaseName(sFileName);
		      if ExtractFileExt(sFileName) <> E then begin
			sFileType := FileTypeInfo(sFileName, E);
			if sFileType = E then sFileType := FUnknownSectionName;
			sSectionName := GetSectionName(sFileType, sFileName);
		      end;
		    end;
		  end;
		end
		else sSectionData := ExtractSection(hdeSection);
		snAdded := AddSection(Selected, sSectionName, sSectionData, mdtSection);
		Charset := FDefaultCharset;
		if snAdded <> nil then begin
		  if hdInline.Encoding <> eNone then snAdded.Encoding := hdInline.Encoding;
		  if CIDHeader <> E then
		    snAdded.SectionCID := CIDHeader
		  else snAdded.SectionCID := CLCHeader;
		  snAdded.SectionType := mdtSection;
		  snAdded.Boundary := Boundary;
		end;
	      end
	      else AddSection(Selected, sSectionName, E, mdtSection);
	    end;
	  end;
	end
	else if FGetRawData and (Selected <> nil) then AddSection(Selected, E, DD + Boundary + DD, mdtSubHeaders);
	if snAdded <> nil then begin
	  if mdtSection = mdtSignature then with snAdded do begin
	    sSectionData := GetSectionText;
	    iDataStrLen := Length(sSectionData);
	    if iDataStrLen > 0 then with FHeaderDataEx do begin
	      HeaderData.Text := sSectionData;
	      if IsPMimeSignature(hdeSection.HeaderData.CTHeader) then
		HeaderData.Encoding := ePMSigned
	      else HeaderData.Encoding := eSMSigned;
	      FSectionData := Parent.FSectionData;
	    end;
	  end;
	  snAdded.Charset := Charset;
	end;
	if MultipartType = mdtMainHeaders then Break;
      end;
    end;
  end;

  procedure ExtractBody(const HD: THeaderData);

    function GetMultipartData(var HDE: THeaderDataEx): String;
    begin
      with HDE do begin
	Boundary := GetBoundary(HeaderData);
	if Boundary <> E then Result := GetMultipartName(HeaderData.CTHeader, MultipartType);
      end;
    end;

  var
    fIsFirstMsgSection: Longbool;
    snAdded: TSectionNode;
    hde: THeaderDataEx;
    bdsPSMime: TBounds;
    iSignatureFlags: Integer;
    sSignatureText, sMultipart: String;
  begin
    FFindHeaderOptions := [fhoContentType, fhoContentTransferEncoding, fhoContentDisposition];
    fIsFirstMsgSection := (Count = 0) or (HD.Recursion > 0);
    snAdded := nil;
    with hde do begin
      MultipartType := mdtUnknownData;
      // root header section(s)
      repeat
	if fIsFirstMsgSection then begin
	  HeaderData := HD;
	  HeaderData.Len := 0;
	  sSignatureText := E;
	  iSignatureFlags := 0;
	  GetHeaderInfo(HeaderData, FFindHeaderOptions);
	  sMultipart := GetMultipartData(hde);
	  if (HeaderData.Recursion = 0) or FullParseOrDecode(FMimeDecodeOptions) then begin
	    if mdoDecode in FMimeDecodeOptions then begin
	      HeaderData := DecodePSMimeMessage(HeaderData, bdsPSMime, sSignatureText,
						iSignatureFlags, FUnencodedMimeData);
	      sMultipart := GetMultipartData(hde);
	      if (HeaderData.Recursion = 0)
	      or ((mdoAnalyseEncoding in FMimeDecodeOptions) and (FBodyEncoding = eNone)) then begin
		with HeaderData do begin
		  if Encoding > eNone then FPSMimeBounds := bdsPSMime;
		  FBodyEncoding := Encoding;
		end;
	      end;
	      HeaderData.Encoding := eNone;
	    end;
	  end;
	  if (HeaderData.Recursion = 0) or ((MultipartType <> mdtMultipartDigest) and not FSubMimeDecoding) then begin
	    with HeaderData do snAdded := AddSection(Selected, FHeaderSectionName, Copy(Text, Pos, Len), mdtMainHeaders);
	    if snAdded <> nil then with snAdded do begin
	      FSectionFlags := iSignatureFlags;
	      FSectionSig := sSignatureText;
	      HeaderDataEx := hde;
	    end;
	  end;
	  FSubMimeDecoding := false;
	end
	else begin
	  HeaderData := GetSubHeaderInfo(SectionNodes[pred(Count)].HeaderData, FFindHeaderOptions, fmoFirstSubHeader);
	  sMultipart := GetMultipartData(hde);
	end;
	if Boundary <> E then with HeaderData do begin
	  if FGetRawData and not fIsFirstMsgSection then AddSection(Selected, E, Copy(Text, Pos, Len), mdtSubHeaders);
	  if MultipartType <> mdtMultipartSigned then
	    snAdded := AddSection(Selected, LB + sMultipart + RB, E, MultipartType)
	  else snAdded := AddSection(Selected, LB + sMultipart + RB, sSignatureText, MultipartType);
	  if snAdded <> nil then begin
	    snAdded.FHeaderDataEx := hde;
	    Selected := snAdded;
	  end;
	end;
	fIsFirstMsgSection := false;
      until Boundary = E;
      if snAdded = nil then Exit;
      // root data section
      if (HeaderData.Recursion = 0) or (mdoFullParse in FMimeDecodeOptions) then begin
	with snAdded do if Boundary = E then begin
	  MultipartType := mdtMainHeaders;
	  ParseSubSections(HeaderDataEx);
	end;
      end;
    end;
    // subsections
    repeat
      with snAdded do if SectionType in [mdtMultipartUnknown..mdtMultipartDigest] then ParseSubSections(HeaderDataEx);
      if (snAdded.Parent <> nil) and (snAdded.HeaderData.Recursion = snAdded.Parent.HeaderData.Recursion) then
	Selected := snAdded.Parent
      else Selected := nil;
      if Selected <> nil then with Selected do CopyHeaderPosLen(snAdded.HeaderData);
      snAdded := Selected;
    until (Selected = nil) or (Selected.SectionType = mdtMultipartDigest);
  end;

var
  hd: THeaderData;
begin
  if MailMessage <> E then with hd do begin
    BeginUpdate;
    FMimeDecodeOptions := MDO;
    FillChar(hd, SizeOf(THeaderData), 0);
    if mdoFancyText in FMimeDecodeOptions then
      FMimeDecodeOptions := FMimeDecodeOptions + [mdoReturnRtf, mdoFullParse]
    else if mdoAnalyseEncoding in FMimeDecodeOptions then
      FMimeDecodeOptions := FMimeDecodeOptions + [mdoFullParse, mdoDecode];
    if not Assigned(FOnEncodedData) then Exclude(FMimeDecodeOptions, mdoDecode);
    FGetRawData := (FMimeDecodeOptions = [mdoFullParse, mdoDecode]);
    if EndPos = 0 then EndPos := Length(MailMessage);
    Text := MailMessage;
    Pos := BegPos;
    Len := 0;
    MaxPos := EndPos;
    Encoding := eNone;
    inc(hd.Recursion, ord(Count > 0));
    ExtractBody(hd);
    dec(hd.Recursion, ord(Count > 0));
    FHasAttachments := (FFileAttachments > 0);
    EndUpdate;
  end;
end;

function TMimeMsgTree.SingleCharExtensionMatch(const FileExt: String): Longbool;
begin
  Result := (Length(FileExt) = 2) and (FirstLoStrPos(FileExt, FSingleCharExtensions) > 0);
end;

function TMimeMsgTree.DoubleCharExtensionMatch(const FileExt: String): Longbool;
begin
  Result := (Length(FileExt) = 3) and (FirstLoStrPos(FileExt, FDoubleCharExtensions) > 0);
end;

function TMimeMsgTree.TripleCharExtensionMatch(const FileExt: String): Longbool;
begin
  Result := (Length(FileExt) = 4) and (FirstLoStrPos(FileExt, FTripleCharExtensions) > 0);
end;

function TMimeMsgTree.CreateNode: TTreeNode;
begin
  Result := TSectionNode.Create(Items);
end;

procedure TMimeMsgTree.CreateTreeView;
begin
  WindowHandle := CreateWindowEx(0, WC_TREEVIEW, nil, 0, 0, 0, 0, 0, 0, 0, hInstance, nil);
end;

function TMimeMsgTree.Get32BitImages: T32BitImages;
begin
  Result := T32BitImages(inherited Images);
end;

procedure TMimeMsgTree.Set32BitImages(Value: T32BitImages);
begin
  inherited Images := Value;
end;

function TMimeMsgTree.GetContainsData: Longbool;
begin
  Result := (Items.Count > 0);
end;

function TMimeMsgTree.GetCount: Integer;
begin
  Result := Items.Count;
end;

function TMimeMsgTree.GetAnsiHeaders: String;
begin
  Result := TSectionNode(Items[0]).AnsiMsgHeaders;
end;

function TMimeMsgTree.GetMainHeaders: String;
begin
  if mdoReturnUnicode in FMimeDecodeOptions then
    Result := Utf8ToWide(TSectionNode(Items[0]).Utf8MainHeaders)
  else Result := Utf8ToAnsi(TSectionNode(Items[0]).Utf8MainHeaders);
end;

function TMimeMsgTree.GetMessageData: String;
var
  iLength, iIndex, iSize: Integer;
  pResult: PChar;
  sData: String;
begin
  if Count > 0 then begin
    iLength := 0;
    for iIndex := 0 to pred(Count) do inc(iLength, Length(SectionData[iIndex]) + Length(PARA));
    SetString(Result, nil, iLength);
    pResult := Pointer(Result);
    for iIndex := 0 to pred(Count) do begin
      sData := SectionData[iIndex];
      iSize := Length(sData);
      if iSize > 0 then begin
	Move(Pointer(sData)^, pResult^, iSize);
	inc(pResult, iSize);
	if SectionTypes[iIndex] in [mdtMainHeaders..mdtSubHeaders] then begin
	  PQChar(pResult)^ := PARA;
	  inc(pResult, Length(PARA));
	end
	else begin
	  if (iSize < 2) or (sData[iSize] <> LF) then begin
	    PQChar(pResult)^ := PARA;
	    inc(pResult, Length(PARA));
	  end
	  else if (iSize < 4) or (sData[iSize - 2] <> LF) then begin
	    PDChar(pResult)^ := CRLF;
	    inc(pResult, Length(CRLF));
	  end;
	end;
      end;
    end;
    System.Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := E;
end;

function TMimeMsgTree.GetUTCDateTime: TDateTime;
begin
  Result := TSectionNode(Items[0]).UTCDateTime;
end;

function TMimeMsgTree.GetCharset(Index: Integer): String;
begin
  Result := TSectionNode(Items).Charset;
end;

function TMimeMsgTree.GetFileName(Index: Integer): String;
begin
  Result := TSectionNode(Items).FileName;
end;

function TMimeMsgTree.GetFirstTextIsFormatted: Longbool;
begin
  Result := (FirstTextSection <> nil) and FirstTextSection.HasFormattedText;
end;

function TMimeMsgTree.GetSectionNode(Index: Integer): TSectionNode;
begin
  Result := TSectionNode(Items[Index]);
end;

function TMimeMsgTree.GetSectionData(Index: Integer): String;
begin
  Result := TSectionNode(Items[Index]).SectionData;
end;

procedure TMimeMsgTree.SetSectionData(Index: Integer; Value: String);
begin
  TSectionNode(Items[Index]).SectionData := Value;
end;

function TMimeMsgTree.GetSectionName(Index: Integer): String;
begin
  Result := TSectionNode(Items[Index]).Text;
end;

procedure TMimeMsgTree.SetSectionName(Index: Integer; Value: String);
begin
  TSectionNode(Items[Index]).Text := Value;
end;

function TMimeMsgTree.GetSectionType(Index: Integer): TMsgDataType;
begin
  Result := TSectionNode(Items[Index]).SectionType;
end;

procedure TMimeMsgTree.SetSectionType(Index: Integer; Value: TMsgDataType);
begin
  TSectionNode(Items[Index]).SectionType := Value;
end;

function TMimeMsgTree.GetSelection: TSectionNode;
begin
  Result := TSectionNode(inherited Selected);
end;

procedure TMimeMsgTree.SetSelection(const Value: TSectionNode);
begin
  inherited Selected := Value;
end;

procedure TMimeMsgTree.CNNotify(var Message: TWMNotify);
var
  snToDisplay: TSectionNode;
begin
  inherited;
  with Message.NMHdr^ do if code = TVN_GETDISPINFO then begin
    with PTVDispInfo(Pointer(Message.NMHdr))^ do begin
      if (item.mask and TVIF_TEXT) <> 0 then begin
	snToDisplay := TSectionNode(Items.GetNode(item.hItem));
	if snToDisplay <> nil then StrPLCopy(item.pszText, Utf8ToAnsi(snToDisplay.SectionName), item.cchTextMax);
      end;
    end;
  end;
end;


// TDatedMessages ----------------------------------------------------------------------------------

const
  ANSI_SEPARATOR	=  '';

  ANSI_END_MARKER	= ANSI_SEPARATOR + SP + SP + ANSI_SEPARATOR + aPARA;

  MIME_UTF8_HEADERS	= MIME_VERSION + SP + MIME_V10 + CRLF +
			  CT_TEXT_UTF8 + CRLF +
			  CTE_TEXT_8BIT + CRLF +
			  CTD_HEADER + SP + CTD_BODY;

function TDatedMessages.CreateMimeHeaders(const MailData: String; DateTime: TDateTime): String;
var
  iHeaderLen, iPos: Integer;
  sHeaders: String;
begin
  iHeaderLen := pred(FirstStrPos(PARA, MailData));
  sHeaders := FMsgTree.RecreateMimeHeaders(Copy(MailData, 1, iHeaderLen), DateTime);
  Result := UNIX_MAIL_START + CRLF + sHeaders + CRLF + MIME_UTF8_HEADERS + Copy(MailData, succ(iHeaderLen), MAXINT);
  iPos := LastStrPos(SP + SP, Result);
  Delete(Result, iPos, pred(Length(Result) - iPos));
  if (Result <> E) and (Result[Length(Result)] <> LF) then Result := Result + CRLF;
end;

procedure TDatedMessages.SetAttachmentHint(const Value: String);
begin
  if mdoReturnUnicode in FMDO then
    FAttachmentHint := AnsiToWide(ANSI_SEPARATOR + aCRLF + Value + aCRLF)
  else FAttachmentHint := ANSI_SEPARATOR + aCRLF + Value + aCRLF;
end;

procedure TDatedMessages.SetDigestHint(const Value: String);
begin
  if mdoReturnUnicode in FMDO then
    FDigestHint := AnsiToWide(ANSI_SEPARATOR + aCRLF + Value + aCRLF)
  else FDigestHint := ANSI_SEPARATOR + aCRLF + Value + aCRLF;
end;

constructor TDatedMessages.Create(MsgList: TDatedStrings);
begin
  FMsgList := MsgList;
  if FMsgList <> nil then begin
    FMsgTree := TMimeMsgTree.Create(nil);
    if FMsgTree <> nil then begin
      inherited Create;
      if FMsgList is TDatedWideStrings then begin
	FDisplayEndMarker := AnsiToWide(ANSI_END_MARKER);
	FMDO := [mdoReturnUnicode, mdoDecode];
      end
      else begin
	FDisplayEndMarker := ANSI_END_MARKER;
	FMDO := [mdoDecode];
      end;
    end;
  end;
end;

destructor TDatedMessages.Destroy;
begin
  FMsgTree.Free;
  inherited;
end;

procedure TDatedMessages.ProcessFolder(const Data: String; Unix: Longbool);
var
  sEndMarker, sHeaders, sBodyText, sHintText, sMessageText: String;
  iBeg, iEnd, iLen: Integer;
begin
  if (FMsgList <> nil) and (FMsgTree <> nil) and (Data <> E) then begin
    FMsgTree.MsgTreeData := FMsgTreeData;
    FTotalBytes := Length(Data);
    FMsgList.Clear;
    iBeg := 1;
    if Unix then
      sEndMarker := LF + UNIX_MAIL_START
    else sEndMarker := PM_MAIL_END;
    repeat
      iEnd := NextStrPos(sEndMarker, Data, iBeg);
      if iEnd = 0 then iEnd := Length(Data);
      if Unix then inc(iEnd, Length(LF));
      iLen := iEnd - iBeg;
      if iLen > 0 then with FMsgTree do begin
	Clear;
	MimeParse(Data, iBeg, pred(iEnd), FMDO);
	if ContainsData then begin
	  sHeaders := MainHeaders;
	  if mdoReturnUnicode in FMDO then
	    sBodyText := GetMessageText(false).UnicodeText
	  else sBodyText := GetMessageText(false).AnsiText;
	  if HasAttachments then
	    sHintText := FAttachmentHint
	  else if IsMessageDigest then
	    sHintText := FDigestHint
	  else sHintText := E;
	  if sHintText <> E then begin
	    sMessageText := Concat([sHeaders, sBodyText, AppendBlankLine(sBodyText),
				    sHintText, GetAttachmentList, FDisplayEndMarker]);
	  end
	  else sMessageText := Concat([sHeaders, sBodyText, AppendBlankLine(sBodyText), FDisplayEndMarker]);
	  FMsgList.AddDated(sMessageText, UTCDateTime);
	  if (TMethod(FOnShowProgress).Code <> nil) and not FOnShowProgress(iLen + ord(Unix = false)) then Break;
	end;
      end;
      iBeg := succ(iEnd);
    until iBeg > FTotalBytes;
  end;
end;

function TDatedMessages.SaveAsUnixMailbox(const FileName: String): Longbool;
var
  fsUnixMailbox: TFileStream;
  sMessage: String;
  iIndex: Integer;
begin
  Result := false;
  if FMsgList.Count = 0 then Exit;
  fsUnixMailbox := TFileStream.Create(FileName, fmCreate);
  if fsUnixMailbox <> nil then with FMsgList do begin
    try
      try
	for iIndex := 0 to pred(Count) do begin
	  if mdoReturnUnicode in FMDO then
	    sMessage := CreateMimeHeaders(WideToUtf8(Strings[iIndex]), DateTimes[iIndex])
	  else sMessage := CreateMimeHeaders(Strings[iIndex], DateTimes[iIndex]);
	  fsUnixMailbox.WriteBuffer(Pointer(sMessage)^, Length(sMessage));
	end;
	Result := true;
      finally
	fsUnixMailbox.Free;
      end;
    except
      DeleteFile(FileName);
    end;
  end;
end;

function ExtractHeaderAddresses(const Headers, BCCRecipients: String; var RecipientCommaList: String): Integer;
const
  CMSP = CM + SP;
var
  hdAddresses: THeaderData;
  sAddresses, sAddress, sEmail: String;
  iMaxLen, iStart, iLen: Integer;
  pRecipientCommaList: PChar;
begin
  Result := 0;
  RecipientCommaList := E;
  if Headers <> E then begin
    FillByte(hdAddresses, SizeOf(THeaderData), 0);
    with hdAddresses do begin
      Text := Headers;
      Pos := 1;
      Len := Length(Headers);
      MaxPos := Len;
    end;
    with FindHeaderLine(H_BCC, hdAddresses, true) do begin
      if Pos > 0 then begin
	if Text <> E then
	  sAddresses := Text
	else sAddresses := BCCRecipients;
      end
      else sAddresses := FindHeaderLine(H_TO, hdAddresses, true).Text + CMSP + FindHeaderLine(H_CC, hdAddresses, true).Text;
    end;
    if sAddresses <> E then begin
      iMaxLen := Length(sAddresses);
      while (iMaxLen > 0) and (sAddresses[iMaxLen] in [SP, CM]) do dec(iMaxLen);
      Delete(sAddresses, succ(iMaxLen), MAXINT);
      if iMaxLen > 0 then begin
	SetString(RecipientCommaList, nil, iMaxLen + succ(GetCharCount(CM, sAddresses)) * Length(CMSP));
	pRecipientCommaList := Pointer(RecipientCommaList);
	iStart := 1;
	while iStart <= iMaxLen do begin
	  iLen := GetAddressLength(sAddresses, iStart);
	  if iLen > 0 then begin
	    sAddress := Copy(sAddresses, iStart, iLen);
	    sEmail := ExtractEmail(sAddress);
	    if sEmail = E then sEmail := sAddress;
	    inc(iStart, iLen);
	    iLen := Length(sEmail);
	    if iLen > 0 then begin
	      Move(Pointer(sEmail)^, pRecipientCommaList^, iLen);
	      inc(pRecipientCommaList, iLen);
	      PDChar(pRecipientCommaList)^ := CMSP;
	      inc(pRecipientCommaList, Length(CMSP));
	      inc(Result);
	    end;
	  end;
	  while (iStart <= iMaxLen) and (sAddresses[iStart] <> CM) do inc(iStart);
	  inc(iStart);
	end;
	Delete(RecipientCommaList, pred(pRecipientCommaList - Pointer(RecipientCommaList)), MAXINT);
      end;
    end;
  end;
end;


end.

